/*	QWICAP (Quick Web Interface for Conventional Applications) by Chris W. Johnson
	Copyright (C) 2003-2011 The University of Texas at Austin ("U.T. Austin"). All rights reserved.
	
	U.T. Austin's copyright notice in this evolving version reflects our original authorship. It does not assert 
	copyright over changes contributed by others.
	
	This library is free software; you can redistribute it and/or modify it under the terms of the GNU Lesser General 
	Public License as published by the Free Software Foundation; either version 2.1 of the License, or (at your option) 
	any later version.
	
	This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied 
	warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for 
	more details.
	
	You should have received a copy of the GNU Lesser General Public License along with this library; if not, write to 
	the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/


package edu.utexas.its.eis.tools.qwicap.servlet;


import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import javax.servlet.Servlet;
import javax.servlet.ServletOutputStream;
import java.io.PrintWriter;


/**	<p>Classes that implement this interface provide Qwicap with its authentication &amp; authorization 
 *	(Auth<sup><small>2</small></sup>) functionality. Every hit is passed to the 
 * 	{@link #qwicapAuth2SchemeHit qwicapAuth2SchemeHit} method for approval. Approved hits are passed along to your 
 * 	Qwicap web application, while unapproved hits do nothing. Implementations will probably want to indicate 
 * 	disapproval to users by one of the following means:
 * 	</p>
 * 
 * 	<ul>
 * 		<li>Return a standard HTTP error code (probably 401, {@link HttpServletResponse#SC_UNAUTHORIZED}), and an 
 * 		informative error message, by using the {@link HttpServletResponse#sendError(int, String)} method.
 * 		</li>
 * 
 * 		<li>Transmit an explanatory page directly to the client by using the {@link PrintWriter}, or 
 * 		{@link ServletOutputStream}, object available from the methods {@link HttpServletResponse#getWriter} 
 * 		and {@link HttpServletResponse#getOutputStream}, respectively. The HTTP connection is all yours in this 
 * 		case, so don't forget to begin by invoking {@link HttpServletResponse#setContentType(String)} to identify the 
 * 		<a href='http://www.iana.org/assignments/media-types/'>MIME type</a> of the content that you will be 
 * 		transmitting to the client.
 * 		</li>
 * 
 * 		<li>Redirect the user to one, or more, static, explanatory pages using the {@link HttpServletResponse#sendRedirect(String)} 
 * 		method. Note that if those static pages are part of this web application, then the 
 * 		{@link Auth2SchemeReply#qwicapCanSendStaticFiles} method must return <code>true</code>, or Qwicap will refuse 
 * 		the user access to those pages.
 * 		</li>
 * 	</ul>
 * 
 * 	<p>Instances of <code>Auth2Scheme</code> are provided to Qwicap by an instance of a class implementing the 
 * 	{@link Auth2SchemeFactory} interface. Qwicap discovers that factory class through an entry in the web application
 * 	deployment descriptor ("web.xml" file). See the <code>Auth2SchemeFactory</code> documentation for complete details.
 * 	</p>
 *
 * 	@see Auth2SchemeFactory
 * 	@see Auth2SchemeReply
 * 
 *	@author Chris W. Johnson
 */
public interface Auth2Scheme {

//	todo See if we can find a practical means of allowing the qwicapAuth2SchemeHit method to check for the presence of required, and still valid nonces in the data set.

/**	Implements an authentication &amp; authorization scheme that determines how Qwicap will respond to each hit.
 * 
 * @param Serv			The active servlet.
 * @param Req			The current request.
 * @param Resp			The current response.
 * @param FDS			The <code>FormDataSet</code> instance describing the form data set, or URI attributes, that were
 * 						sent with, or otherwise associated with, the latest hit. Note: Modification of the contents of 
 * 						this <code>FormDataSet</code> instance is possible, but strongly discouraged, because, in 
 * 						general, a {@link FormDataSetFilter} should be used for that purpose.
 * @return				An object implementing the {@link Auth2SchemeReply} interface.
 * @throws Exception	Thrown if anything goes wrong. The hit is discarded in such cases.
 */
	public Auth2SchemeReply qwicapAuth2SchemeHit
	(
		final Servlet				Serv,
		final HttpServletRequest	Req,
		final HttpServletResponse	Resp,
		final FormDataSet			FDS
	) 
	throws Exception;
		
/**	Returns <code>true</code> to inform the servlet that this <code>Auth2Scheme</code> instance should be retained as 
 *	part of the current user's session state, and should be used to evaluate the next hit received from this user. 
 *	Returns <code>false</code> to indicate that this <code>Auth2Scheme</code> instance should be discarded, in which 
 *	case a new instance will be created to evaluate the next hit.
 * 
 * @return	<code>true</code>, if this instance should be stored and reused by this session, <code>false</code> if it 
 *			should be discarded.
 */
	public boolean qwicapShouldRetainThisAuth2SchemeInstance();
}

