/*	QWICAP (Quick Web Interface for Conventional Applications) by Chris W. Johnson
	Copyright (C) 2003-2010 The University of Texas at Austin ("U.T. Austin"). All rights reserved.
	
	U.T. Austin's copyright notice in this evolving version reflects our original authorship. It does not assert 
	copyright over changes contributed by others.
	
	This library is free software; you can redistribute it and/or modify it under the terms of the GNU Lesser General 
	Public License as published by the Free Software Foundation; either version 2.1 of the License, or (at your option) 
	any later version.
	
	This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied 
	warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for 
	more details.
	
	You should have received a copy of the GNU Lesser General Public License along with this library; if not, write to 
	the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/


package edu.utexas.its.eis.tools.qwicap.servlet;


import java.util.Collection;
import java.util.HashMap;
import java.util.LinkedHashMap;
import java.util.logging.Logger;
import java.util.logging.Level;
import java.nio.charset.Charset;
import java.io.UnsupportedEncodingException;


/**	Represents all that we need to understand about the character sets supported within this particular Java VM.
 * 
 * 	@author Chris W. Johnson
 */
final class CharacterSetRecognizer {
	
	private static final Logger 		Log = Logger.getLogger(CharacterSetRecognizer.class.getName());
	
//	The reference string assumes that it will never be the first control in the list. The way the Page class is 
//	implemented, this will be true.
	
	static final String					kReferenceString = Qwicap.kQwicapCharSetControlName;
		
	private final CharacterSetFamily[]	AllFamilies;
	
	
	CharacterSetRecognizer() throws UnsupportedEncodingException {
		final HashMap<ByteArray, CharacterSetFamily> AllEncodingsList = new LinkedHashMap<ByteArray, CharacterSetFamily>();
		
	//	Seed the supported character sets list with some basic character sets, so that we can assign meaningful 
	//	names to the families, and allocate special family classes where necessary.
		
		try {
			addInitialCharFamily(AllEncodingsList, new CharacterSetFamilyASCII("ISO-8859-1", "a single-byte, ASCII-derived character set"));
		} catch (UnsupportedEncodingException e) {
			Log.log(Level.WARNING, "One of the \"basic\" character sets is not present on this platform.", e);	
		}
			
		try {
			addInitialCharFamily(AllEncodingsList, new CharacterSetFamily("IBM1047", "an EBCDIC-derived character set"));
		} catch (UnsupportedEncodingException e) {
			Log.log(Level.WARNING, "The IBM1047 character set is not present on this platform, so Qwicap will not be able to handle any of the EBCDIC character sets (code pages).", e);	
		}
		
		try {
			addInitialCharFamily(AllEncodingsList, new CharacterSetFamilyUTF16BE("UTF-16BE"));
		} catch (UnsupportedEncodingException e) {
			Log.log(Level.WARNING, "One of the \"basic\" character sets is not present on this platform.", e);	
		}
		
		try {
			addInitialCharFamily(AllEncodingsList, new CharacterSetFamilyUTF16LE("UTF-16LE"));
		} catch (UnsupportedEncodingException e) {
			Log.log(Level.WARNING, "One of the \"basic\" character sets is not present on this platform.", e);	
		}
		
		try {
			addInitialCharFamily(AllEncodingsList, new CharacterSetFamilyUTF32BE("UTF-32BE"));
		} catch (UnsupportedEncodingException e) {
			Log.log(Level.WARNING, "One of the \"basic\" character sets is not present on this platform.", e);	
		}
		
		try {
			addInitialCharFamily(AllEncodingsList, new CharacterSetFamilyUTF32LE("UTF-32LE"));
		} catch (UnsupportedEncodingException e) {
			Log.log(Level.WARNING, "One of the \"basic\" character sets is not present on this platform.", e);	
		}
			
		final Collection<Charset>	CharSets = Charset.availableCharsets().values();

		for (final Charset CharSet : CharSets) {

			if (CharSet.isRegistered() && CharSet.canEncode()) {
				final String			CharSetName = CharSet.name();
				
			//	Don't register the character set "UTF-16", because it is already handled by "UTF-16BE" and "UTF-16LE". 
			//	
			//	Don't register the character set "UTF-32", because it is already handled by "UTF-32BE" and "UTF-32LE".
			//	
			//	*Do* register everything else.
			
				if ("UTF-16".equals(CharSetName) == false && "UTF-32".equals(CharSetName) == false)
					addCharSet(AllEncodingsList, CharSetName);
			}
		}
		
		AllFamilies = AllEncodingsList.values().toArray(new CharacterSetFamily[AllEncodingsList.size()]);
		
		if (Log.isLoggable(Level.FINEST))
			Log.log(Level.FINEST, "The following IANA-registered character sets are available on this machine:\n{0}", toString());
	}
	
	String identifyCharSetOf
	(
		final FormControlBytes[]	FCBs	//	Must have at least one element.
	)
	throws UnsupportedEncodingException
	{
		String					CharSetName = null;
			
		for (final CharacterSetFamily Family : AllFamilies) {
			
			CharSetName = Family.identifyCharSet(FCBs);
			if (CharSetName != null)
				break;
		}
		
		return CharSetName;
	}

	@Override public String toString() {
		final StringBuilder		Buff = new StringBuilder(8 * 1024);
		
		for (final CharacterSetFamily Family : AllFamilies)
			Buff.append(Family).append('\n');
		
		return Buff.toString();
	}
	
	private static void addInitialCharFamily
	(
		final HashMap<ByteArray, CharacterSetFamily>	AllFamilies,
		final CharacterSetFamily						NewSeries
	) 
	{
		final ByteArray				NewSeriesKey = NewSeries.getReferenceBytes();
		final CharacterSetFamily	Series       = AllFamilies.get(NewSeriesKey);
		
		if (Series == null)
			AllFamilies.put(NewSeriesKey, NewSeries);
		else
			throw new IllegalStateException("A character series (\"" + Series.getName() + "\") is already registered, so the new character series (\"" + NewSeries.getName() + "\") cannot be registered.");
	}
	
	private static void addCharSet
	(
		final HashMap<ByteArray, CharacterSetFamily>	AllFamilies,
		final String									CharSetName
	) 
	throws UnsupportedEncodingException
	{
		final ByteArray			NewKey = new ByteArray(kReferenceString.getBytes(CharSetName));
		CharacterSetFamily		Family = AllFamilies.get(NewKey);
		
		if (Family == null) {
			
			Family = new CharacterSetFamily(CharSetName, "one that overlaps with " + CharSetName);
			AllFamilies.put(NewKey, Family);
		
		} else
			Family.add(CharSetName);
	}
}
