/*	QWICAP (Quick Web Interface for Conventional Applications) by Chris W. Johnson
	Copyright (C) 2003-2011 The University of Texas at Austin ("U.T. Austin"). All rights reserved.
	
	U.T. Austin's copyright notice in this evolving version reflects our original authorship. It does not assert 
	copyright over changes contributed by others.
	
	This library is free software; you can redistribute it and/or modify it under the terms of the GNU Lesser General 
	Public License as published by the Free Software Foundation; either version 2.1 of the License, or (at your option) 
	any later version.
	
	This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied 
	warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for 
	more details.
	
	You should have received a copy of the GNU Lesser General Public License along with this library; if not, write to 
	the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/


package edu.utexas.its.eis.tools.qwicap.servlet;


import java.io.*;
import java.util.ArrayList;


/**	<code>ClassFile</code> parses Java class files and provides methods for retrieving information about them, notably 
 *	the list of methods in the class, and the other classes on which the class depends.
 * 
 * 	<p>One could use the Apache Jakarta Project's BCEL (Byte Code Engineering Library) library for this purpose (and a 
 * 	great deal more), but it is currently 524 KB long - almost as large as Qwicap, itself.
 * 	</p>
 * 
 * 	@author Chris W. Johnson
 *  @see <a href='http://java.sun.com/docs/books/vmspec/2nd-edition/html/ClassFile.doc.html'>VM Spec: The Class File Format</a>
 */
final class ClassFile {

	private final int				MinorVersion;
	private final int				MajorVersion;
	private final Constant[]		Constants;
	private final int				AccessFlags;
	private final int				ClassNameIndex;
	private final int				SuperclassNameIndex;
	
	private final ClassConstant[]	Interfaces;
	private final FieldOrMethod[]	Fields;
	private final FieldOrMethod[]	Methods;
	
	private transient InputStream	In;


/*	public static void main(final String[] Args) throws Exception {
		
		for (int Index = 0, Count = Args.length; Index < Count; Index++)
			dumpClassFiles(new File(Args[Index]));
	}
	
	private static void dumpClassFiles
	(
		final File				F
	)
	throws IOException
	{
		if (F.isDirectory()) {
			final File[]			List = F.listFiles();
			
			for (int Index = 0, Count = List.length; Index < Count; Index++)
				dumpClassFiles(List[Index]);
		} else
			dumpClassFileInfo(F);
	}
	
	
	private static void dumpClassFileInfo
	(
		final File				F
	)
	throws IOException
	{
		System.out.println();
		System.out.println();
		
		try {
			final ClassFile			CF = new ClassFile(new FileInputStream(F), true);
				
			System.out.println(CF.getClassName() + " (" + CF.getMajorVersion() + "." + CF.getMinorVersion() + ")");
			System.out.println();
						
			final String[]			Classes = CF.getClasses();
			
			System.out.println("\tClasses:");
			for (int Index = 0, Count = Classes.length; Index < Count; Index++)
				System.out.println("\t\t" + Classes[Index]);

			System.out.println();
			
			final String[]			Interfaces = CF.getInterfaces();
			
			System.out.println("\tInterfaces:");
			for (int Index = 0, Count = Interfaces.length; Index < Count; Index++)
				System.out.println("\t\t" + Interfaces[Index]);
			
			System.out.println();
			
			final String[]			Fields = CF.getFieldStrings();
			
			System.out.println("\tFields:");
			for (int Index = 0, Count = Fields.length; Index < Count; Index++)
				System.out.println("\t\t" + Fields[Index]);
			
			System.out.println();
			
			final String[]			Methods = CF.getMethodStrings();
			
			System.out.println("\tMethods:");
			for (int Index = 0, Count = Methods.length; Index < Count; Index++)
				System.out.println("\t\t" + Methods[Index]);
				
		} catch (ClassFileFormatException e) {
			
			System.out.println("NOT A CLASS: " + F.getAbsolutePath());
			e.printStackTrace();
		}
	}
*/	

	ClassFile
	(
		final InputStream		In,
		final boolean			CloseIn
	)
	throws ClassFileFormatException
	{
		try {
			this.In = In;

			if (readInt() != 0xCAFEBABE)
				throw new ClassFileFormatException("Bad magic.");
			
			MinorVersion = readShort();
			MajorVersion = readShort();
			Constants    = new Constant[readShort()];

			for (int Index = 1, Count = Constants.length; Index < Count; Index++) {
				final int				TagType = readByte();

				switch (TagType) {

					case 1:		//	UTF8

						Constants[Index] = new UTF8Constant();
						break;

					case 3:		//	Integer

						Constants[Index] = new IntegerConstant();
						break;

					case 4: 	//	Float

						Constants[Index] = new FloatConstant();
						break;

					case 5:		//	Long

						Constants[Index] = new LongConstant();
						Index++;
						break;

					case 6:		//	Double

						Constants[Index] = new DoubleConstant();
						Index++;
						break;

					case 7:		//	Class

						Constants[Index] = new ClassConstant();
						break;

					case 8:		//	String

						Constants[Index] = new StringConstant();
						break;

					case 9:		//	Fieldref

						Constants[Index] = new FieldrefConstant();
						break;

					case 10:	//	Methodref

						Constants[Index] = new MethodrefConstant();
						break;

					case 11:	//	Interface Methodref

						Constants[Index] = new InterfaceMethodrefConstant();
						break;

					case 12:	//	NameAndType

						Constants[Index] = new NameAndTypeConstant();
						break;

					default:
						
						throw new ClassFileFormatException("Constant pool tag type " + TagType + " is not recognized.");
				}
			}
			
			AccessFlags         = readShort();
			ClassNameIndex      = readShort();
			SuperclassNameIndex = readShort();
			
			Interfaces = new ClassConstant[readShort()];
			
			for (int Index = 0, Count = Interfaces.length; Index < Count; Index++)
				Interfaces[Index] = (ClassConstant) Constants[readShort()];
			
			Fields = new FieldOrMethod[readShort()];
			
			for (int Index = 0, Count = Fields.length; Index < Count; Index++)
				Fields[Index] = new FieldOrMethod();
			
			Methods = new FieldOrMethod[readShort()];
			
			for (int Index = 0, Count = Methods.length; Index < Count; Index++)
				Methods[Index] = new FieldOrMethod();
			
		} catch (ClassFileFormatException e) {
			
			throw e;
			
		} catch (Exception e) {
			
			throw new ClassFileFormatException(e);
			
		} finally {
			
			if (CloseIn)
				try { In.close(); } catch (IOException e) { /* We can do nothing useful here. */ }
			
			this.In = null;
		}
	}
	
	int getMajorVersion() {
		return MajorVersion;
	}
	
	int getMinorVersion() {
		return MinorVersion;
	}
	
	int getAccessFlags() {
		return AccessFlags;
	}
	
	String getClassName() {
		return ((ClassConstant) Constants[ClassNameIndex]).getClassName();
	}
	
	String getSuperclassName() {
		
		if (SuperclassNameIndex == 0)
			return null;	//	Only java.lang.Object is allowed to have no superclass!
		
		return ((ClassConstant) Constants[SuperclassNameIndex]).getClassName();
	}
	
	boolean dependsOnClass
	(
		final String			ClassName
	)
	{
		for (int Index = 1, Count = Constants.length; Index < Count; Index++) {
			
			if (Index == ClassNameIndex)
				continue;
			
			final Constant			C = Constants[Index];
			
			if (C != null && C.getTagType() == 7 && ClassName.equals(((ClassConstant) C).getClassName()))
				return true;
		}
		
		return false;
	}

	String[] getClasses() {
		final ArrayList<String>	List = new ArrayList<String>(Constants.length);
		
		for (int Index = 1, Count = Constants.length; Index < Count; Index++) {
			
			if (Index == ClassNameIndex)
				continue;	//	Don't include this class' name.
			
			final Constant			C = Constants[Index];
			
			if (C != null && C.getTagType() == 7)
				List.add(((ClassConstant) C).getClassName());
		}
		
		return List.toArray(new String[List.size()]);
	}

	String[] getInterfaces() {
		final String[]			Faces = new String[Interfaces.length];
		
		for (int Index = 0, Count = Interfaces.length; Index < Count; Index++)
			Faces[Index] = Interfaces[Index].getClassName();
		
		return Faces;
	}
	
	boolean hasMethod
	(
		final String			Name,
		final String			Descriptor,
		final int				RequiredAccessFlags
	)
	{
		for (final FieldOrMethod M : Methods) {
			
			if (Name.equals(M.getName()) && Descriptor.equals(M.getDescriptor()) && (RequiredAccessFlags & M.AccessFlags) == RequiredAccessFlags)
				return true;
		}
		
		return false;
	}
	
	String[] getMethodStrings() {
		final String[]			S = new String[Methods.length];
		
		for (int Index = 0, Count = Methods.length; Index < Count; Index++) {
			final FieldOrMethod		M = Methods[Index];
			
			S[Index] = M.getName() + ' ' + M.getDescriptor();
		}
		
		return S;
	}
	
	String[] getFieldStrings() {
		final String[]			S = new String[Fields.length];
		
		for (int Index = 0, Count = Fields.length; Index < Count; Index++) {
			final FieldOrMethod		M = Fields[Index];
			
			S[Index] = M.getName() + ' ' + M.getDescriptor();
		}
		
		return S;
	}
	
	private int readByte() throws IOException {
		final int				Value = In.read();

		if (Value < 0)
			throw new EOFException();

		return Value;
	}

	private byte[] readBytes
	(
		final byte[]			Bytes
	)
	throws IOException
	{
		for (int Index = 0, Count = Bytes.length; Index < Count; Index++)
			Bytes[Index] = (byte) readByte();

		return Bytes;
	}

	private int readShort() throws IOException {
		int						Value = 0;

		for (int Count = 2; --Count >= 0; )
			Value = Value << 8 | readByte();

		return Value;
	}

	private int readInt() throws IOException {
		int						Value = 0;

		for (int Count = 4; --Count >= 0; )
			Value = Value << 8 | readByte();

		return Value;
	}

	private long readLong() throws IOException {
		long					Value = 0;

		for (int Count = 8; --Count >= 0; )
			Value = Value << 8 | readByte();

		return Value;
	}
	
	private void skip(long Count) throws IOException {
		
		while (Count != 0) {
			final long			Skipped = In.skip(Count);
			
			if (Skipped > 0)
				Count -= Skipped;
		}
	}
	
	final class FieldOrMethod {
		
		final int				AccessFlags;
		final int				NameIndex;
		final int				DescriptorIndex;
		final Attribute[]		Attrs;
		
		FieldOrMethod() throws IOException {
			
			AccessFlags     = readShort();
			NameIndex       = readShort();
			DescriptorIndex = readShort();
			Attrs           = new Attribute[readShort()];
			
			for (int Index = 0, Count = Attrs.length; Index < Count; Index++)
				Attrs[Index] = new Attribute();
		}
		
		String getName() {
			return ((UTF8Constant) Constants[NameIndex]).Value;
		}
		
		String getDescriptor() {
			return ((UTF8Constant) Constants[DescriptorIndex]).Value;
		}
		
		Attribute[] getAttributes() {
			return Attrs;
		}
	}
	
	final class Attribute {
		
		final int				NameIndex;
		final long				Length;		//	Length is an unsigned 32-bit value.
		
		Attribute() throws IOException {
			
			NameIndex = readShort();
			Length    = readInt() & 0x0FFFFFFFFL;	//	Convert the signed int into an unsigned long.
			
			skip(Length);	//	Skip the contents of the attribute.
		}
		
		String getName() {
			return ((UTF8Constant) Constants[NameIndex]).Value;
		}
	}
	
	abstract static class Constant {
		final int				TagType;

		protected Constant(final int TagType) { this.TagType = TagType; }
		int getTagType() { return TagType; }
	}

	final class UTF8Constant extends Constant {
		final String			Value;

		UTF8Constant() throws IOException { super(1); Value = new String(readBytes(new byte[readShort()]), "UTF-8"); }
	}

	final class IntegerConstant extends Constant {
		final int				Value;

		IntegerConstant() throws IOException { super(3); Value = readInt(); }
	}

	final class FloatConstant extends Constant {
		final float				Value;

		FloatConstant() throws IOException { super(4); Value = Float.intBitsToFloat(readInt()); }
	}

	final class LongConstant extends Constant {
		final long				Value;

		LongConstant() throws IOException { super(5); Value = readLong(); }
	}

	final class DoubleConstant extends Constant {
		final double			Value;

		DoubleConstant() throws IOException { super(6); Value = Double.longBitsToDouble(readLong()); }
	}

	final class ClassConstant extends Constant {
		final int				UTF8Index;

		ClassConstant() throws IOException { super(7); UTF8Index = readShort(); }
		String getClassName() { return ((UTF8Constant) Constants[UTF8Index]).Value; }
	}

	final class StringConstant extends Constant {
		final int				UTF8Index;

		StringConstant() throws IOException { super(8); UTF8Index = readShort(); }
		String getString() { return ((UTF8Constant) Constants[UTF8Index]).Value; }
	}
	
	final class FieldrefConstant extends Constant {
		final int				ClassIndex;
		final int				NameAndTypeIndex;

		FieldrefConstant() throws IOException { super(9); ClassIndex = readShort(); NameAndTypeIndex = readShort(); }
		String              getClassName()   { return ((ClassConstant) Constants[ClassIndex]).getClassName(); }
		NameAndTypeConstant getNameAndType() { return (NameAndTypeConstant) Constants[ClassIndex]; }
	}

	final class MethodrefConstant extends Constant {
		final int				ClassIndex;
		final int				NameAndTypeIndex;

		MethodrefConstant() throws IOException { super(10); ClassIndex = readShort(); NameAndTypeIndex = readShort(); }
		String              getClassName()   { return ((ClassConstant) Constants[ClassIndex]).getClassName(); }
		NameAndTypeConstant getNameAndType() { return (NameAndTypeConstant) Constants[ClassIndex]; }
	}

	final class InterfaceMethodrefConstant extends Constant {
		final int				ClassIndex;
		final int				NameAndTypeIndex;

		InterfaceMethodrefConstant() throws IOException { super(11); ClassIndex = readShort(); NameAndTypeIndex = readShort(); }
		String              getClassName()   { return ((ClassConstant) Constants[ClassIndex]).getClassName(); }
		NameAndTypeConstant getNameAndType() { return (NameAndTypeConstant) Constants[ClassIndex]; }
	}

	final class NameAndTypeConstant extends Constant {
		final int				NameIndex;	//	UTF8
		final int				TypeIndex;	//	UTF8

		NameAndTypeConstant() throws IOException { super(12); NameIndex = readShort(); TypeIndex = readShort(); }
		String getName() { return ((UTF8Constant) Constants[NameIndex]).Value; }
		String getType() { return ((UTF8Constant) Constants[TypeIndex]).Value; }
	}
}
