/*	QWICAP (Quick Web Interface for Conventional Applications) by Chris W. Johnson
	Copyright (C) 2008 Chris W. Johnson. All rights reserved.
	
	Chris W. Johnson's copyright notice in this evolving version reflects his original authorship. It does not assert 
	copyright over changes contributed by others.
	
	This library is free software; you can redistribute it and/or modify it under the terms of the GNU Lesser General 
	Public License as published by the Free Software Foundation; either version 2.1 of the License, or (at your option) 
	any later version.
	
	This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied 
	warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for 
	more details.
	
	You should have received a copy of the GNU Lesser General Public License along with this library; if not, write to 
	the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/


package edu.utexas.its.eis.tools.qwicap.servlet;


import edu.utexas.its.eis.tools.qwicap.util.ArrayToString;

import javax.servlet.ServletContext;
import javax.servlet.http.HttpServlet;
import java.util.*;
import java.util.zip.ZipEntry;
import java.util.jar.JarInputStream;
import java.util.logging.Level;
import java.util.logging.Logger;
import java.io.IOException;


/**	In Tomcat 6.0.14 and 6.0.19 (and possibly many others), it has been observed that the re-deployment of a web 
 * 	application (but, oddly enough, not the shutdown of the Tomcat server) can trigger <code>NoClassDefFoundError</code>s 
 * 	during execution of the servlet's <code>destroy</code> method. This appears to apply only to classes that 
 * 	have not been loaded prior to the execution of <code>destroy</code>, but which are used by the <code>destroy</code> 
 * 	code (directly or indirectly), and implies that the web application's class loader has been instructed to suspend 
 * 	class loading prior to the invocation of <code>destroy</code>.
 * 
 * 	A means to change this behavior of the web application's class loader has not been identified. One workaround, 
 * 	however, is to force loading of the problematic classes before <code>destroy</code> is invoked. We do that here, 
 * 	hoping that we will always pre-load enough classes to prevent the <code>NoClassDefFoundError</code> from 
 * 	manifesting under all circumstances in which the <code>destroy</code> method may execute. (We retain references to 
 * 	the pre-loaded classes in order to ensure that they cannot be garbage-collected prior to servlet destruction.)
 *	
 *	@author Chris W. Johnson
 */
final class ClassPreloaderAndHolder {
	
	private static final Logger				Log = Logger.getLogger(ClassPreloaderAndHolder.class.getName());
	
	
	private final String[]					DesiredPackages;
	private final HashMap<String, Class<?>>	LoadedClasses = new HashMap<String, Class<?>>(1000);
	
	
	ClassPreloaderAndHolder
	(
		final HttpServlet		Servlet,
		final String...			PackageNameRoots
	)
	{
		DesiredPackages = new String[PackageNameRoots.length];
		
		for (int Index = 0, Count = PackageNameRoots.length; Index < Count; Index++)
			DesiredPackages[Index] = PackageNameRoots[Index].replace('.', '/');
		
		preloadPackages(Servlet);
	}
	
	private void preloadPackages
	(
		final HttpServlet		Servlet
	) 
	{
		
		try {
			final ServletContext	SCtxt = Servlet.getServletContext();
			final Set<String>		Paths = SCtxt.getResourcePaths("/WEB-INF/lib/");
			
			if (Paths == null) {
				
				Log.log(Level.FINE, "Failed to get the path to this application's \"/WEB-INF/lib/\" subdirectory.");
				return;
			}
			
			final HashSet<String>	Dependencies = new HashSet<String>(1000);
						
			for (final String Path : Paths) {
				
				if (Path.endsWith("/"))
					continue;
				
				try {
					final JarInputStream Jar = new JarInputStream(SCtxt.getResourceAsStream(Path));

					try {

						for (ZipEntry Entry = Jar.getNextEntry(); Entry != null; Entry = Jar.getNextEntry()) {
							final String			Name = Entry.getName();

						//	Skip anything that isn't a class file.

							if (Name.endsWith(".class") == false)
								continue;

						//	Ignore all classes in packages other than the ones we're meant to be preloading.

							if (isDesiredPackageOrSubpackage(Name) == false)
								continue;

							try {
								final ClassFile			CF = new ClassFile(Jar, false);
								
								recordClassAndDependencies(CF, Dependencies);

							} catch (ClassFileFormatException e) {
							//	Ignore.
							}
						}

					} finally {
						Jar.close();
					}
					
				} catch (IOException e) {
				//	We don't care about IOExceptions here. We just want to move along to the next file.
					Log.log(Level.FINE, "Failed to read the resource \"" + Path + "\" as a JAR file.", e);
				} catch (Exception e) {
				//	We care a little more about exceptions other than the expected IOExceptions, so we log them at a 
				//	slightly higher level.	
					Log.log(Level.INFO, "Failed to read the resource \"" + Path + "\" as a JAR file due to an unexpected exception.", e);
				}
			}
			
		//	Load all of the classes.
			
			final ArrayList<Class<?>>	ClassList  = new ArrayList<Class<?>>(Dependencies.size());
			final ArrayList<String>		Unresolved = new ArrayList<String>();
			
			for (final String ClassToLoad : Dependencies) {
				final String			ClassName   = ClassToLoad.replace('/', '.');
				final Class<?>			LoadedClass = loadClass(ClassName);
				
				if (LoadedClass != null)
					ClassList.add(LoadedClass);
				else
					Unresolved.add(ClassName);
			}
			
			Log.log(Level.INFO, "Preloaded {0} of {1} classes associated with the packages, and sub-packages of: {2}.", new Object[] { ClassList.size(), Dependencies.size(), new ArrayToString(DesiredPackages).alwaysQuote() });
			
			if (Unresolved.size() > 0)
				Log.log(Level.INFO, "Failed to preload the following {0} classes:\n{1}", new Object[] { Unresolved.size(), new ArrayToString(Unresolved).alwaysQuote().setSeparator("\n").setFinalSeparator("\n") });
			
		} catch (Exception e) {
			Log.log(Level.WARNING, "A problem was encountered during the package pre-loading process.", e);
		}
	}
	
	private static void recordClassAndDependencies
	(
		final ClassFile			CF,
		final HashSet<String>	Dependencies
	)
	{
		Dependencies.add(CF.getClassName());
		
		for (final String Dependency : CF.getClasses()) {
			
			if (Dependency.startsWith("java/") || Dependency.startsWith("javax/") || Dependency.startsWith("sun/"))
				continue;
			
			Dependencies.add(Dependency);
		}
	}
	
	private static Class<?> loadClass
	(
		final String			ClassName
	)
	{
		try {
			
			return Class.forName(ClassName);
			
		} catch (LinkageError e) {
		
			Log.log(Level.SEVERE, "Could not load class \"{0}\", because there was a linkage error.", ClassName);
		
		} catch (ClassNotFoundException e) {
		
			Log.log(Level.SEVERE, "Could not load class \"{0}\", because the class was not found.", ClassName);
		}

		return null;
	}
	
	private boolean isDesiredPackageOrSubpackage
	(
		final String			PackageName
	)
	{
		for (final String DesiredPackage : DesiredPackages) {
			
			if (PackageName.startsWith(DesiredPackage))
				return true;
		}
		
		return false;
	}
}
