/*	QWICAP (Quick Web Interface for Conventional Applications) by Chris W. Johnson
	Copyright (C) 2010-2011 Chris W. Johnson. All rights reserved.
	
	Chris W. Johnson's copyright notice in this evolving version reflects original authorship. It does not assert 
	copyright over changes contributed by others.
	
	This library is free software; you can redistribute it and/or modify it under the terms of the GNU Lesser General 
	Public License as published by the Free Software Foundation; either version 2.1 of the License, or (at your option) 
	any later version.
	
	This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied 
	warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for 
	more details.
	
	You should have received a copy of the GNU Lesser General Public License along with this library; if not, write to 
	the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/


package edu.utexas.its.eis.tools.qwicap.servlet;


import java.util.regex.Pattern;
import java.util.regex.Matcher;
import java.util.ArrayList;
import java.util.logging.Logger;
import java.util.logging.Level;
import java.nio.charset.Charset;
import java.nio.charset.IllegalCharsetNameException;
import java.nio.charset.UnsupportedCharsetException;


/*	About The "Content-Type" Header and HTTP

	RFCs 822, 2045, 2046 and 2616 share a substantially common set of rules for the their "Content-Type" identifiers. 
	Seen from the perspective of the HTTP 1.1 standard (RFC 2616), which is the most derivative of the standards with 
	regard to "Content-Type", the ultimate set of rules are derived in the following manner:
	
	The HTTP 1.1 "Content-Type" header is discussed in section 14.17 of RFC 2616. The "Content-Type" header specifies 
	a media type. Media types are discussed in section 3.7 of RFC 2616. Section 3.7 mentions several important facts:
	
		* "The type, subtype, and parameter attribute names are case-insensitive."
		* "Parameter values might or might not be case-sensitive, depending on the semantics of the parameter name."
		* "Linear white space (LWS) MUST NOT be used between the type and subtype, nor between an attribute and its 
		  value."
	
	Section 3.7.1 of RFC 2616, "Canonicalization and Text Defaults", also has something interesting to say:
	
		When no explicit charset parameter is provided by the sender, media subtypes of the "text" type are defined to 
		have a default charset value of "ISO-8859-1" when received via HTTP.
	
	Section 3.7.2 of RFC 2616, "Multipart Types", states that multi-part HTTP uses the same syntax defined in RFC 2046, 
	section 5.1.1. The "abstract" at the start of RFC 2046 states:
	
		The initial document in this set, RFC 2045, specifies the various headers used to describe the structure of 
		MIME messages. This second document defines the general structure of the MIME media typing system and defines 
		an initial set of media types. [....]
	
	Thus, the HTTP 1.1 standard (RFC 2616) bases its handling of mutlipart data on RFC 2046, and RFC 2046 depends on 
	RFC 2045 to specify the relevant headers. Section 5 of RFC 2045, defines the "Content-Type" header, and thus seems 
	to be the ultimate source of the definition of the "Content-Type" header of HTTP 1.1 (RFC 2616). Section 5.1 of 
	RFC 2045 provides the detailed defintion of the "Content-Type" header syntax (using the augmented BNF of RFC 822):
	
		content := "Content-Type" ":" type "/" subtype
				   *(";" parameter)
			
	Where "parameter" is defined as:
	
		parameter := attribute "=" value
			
	And "value" is defined as:
	
		value := token / quoted-string
			
	However, while "token" is defined in section 5.1 of RFC 2045, "quoted-string" is not defined in RFC 2045 at all. 
	To find the definition of "quoted-string" we have to refer back to RFC 822, section 3.3, "Lexical Tokens", which 
	supplies the following definition:
	
		quoted-string = <"> *(qtext/quoted-pair) <">; Regular qtext or
													;   quoted chars.
			
	Where "qtext" is defined as:
	
		qtext       =  <any CHAR excepting <">,     ; => may be folded
						"\" & CR, and including
						linear-white-space>
	
	And "quoted-pair" is defined as:
	
		quoted-pair =  "\" CHAR                     ; may quote any char
	
	So, a "quoted-pair" identifies a conventional backslash ('\') based single-character escaping mechanism, as 
	explained in detail in RFC 822 section 3.4.1, where it is referred to as "quoting". That's a misleading choice of 
	terminology, because the escaping mechanism can apply to any character, not just quotes, but that's the term RFC 
	822 uses, so we're stuck with it.
	
	By the way, RFC 2045, section 5.1, helpfully includes the following examples of "Content-Type" headers:
	
		Content-type: text/plain; charset=us-ascii (Plain text)
	
		Content-type: text/plain; charset="us-ascii"
			
	These do not illustrate the quoted-pair mechanism, but do illustrate the concepts of parameters whose values are 
	tokens (charset=us-ascii), parameters whose values are quoted-strings (charset="us-ascii"), and parenthetical 
	comments ((Plain text)). Comments are not included in the RFC 2045 BNF defining "parameter", but the text of 
	section 5.1 states: "comments are allowed in accordance with RFC 822 rules for structured header fields". So, back 
	we go to RFC 822, where we find the following definition of "comment" in section 3.3:
	   
		comment     =  "(" *(ctext / quoted-pair / comment) ")"
			
	Thus it seems that an RFC 2616 "Content-Type" header must support all of the features and syntax defined for it in 
	RFC 2045 and RFC 822. That support includes: an unlimited number of parameters following the type/subtype; 
	parameters including a trailing, parenthetical comment; quoted parameter values; and a backslash-based escaping 
	mechanism for use within quoted parameter values and comments.
*/


/* Created by IntelliJ IDEA.
 * User: chrisj
 * Date: Feb 7, 2010
 * Time: 6:55:24 PM
 * To change this template use File | Settings | File Templates.
 */


/**	<p>The <code>ContentType</code> class parses the values of the "Content-type" headers in standard email messages 
 * 	(<a href='http://www.ietf.org/rfc/rfc0822.txt'>RFC 822</a>), MIME email messages (RFC 
 * 	<a href='http://www.ietf.org/rfc/rfc2045.txt'>2045</a> and <a href='http://www.ietf.org/rfc/rfc2046.txt'>2046</a>), 
 * 	and HTTP transactions (<a href='http://www.w3.org/Protocols/rfc2616/rfc2616.html'>RFC 2616</a>). Because those
 * 	standards differ with regard to their default character sets, this class is incomplete unless its constructor is
 * 	supplied with an implementation of {@link ContentTypeDefaultCharSet} that knows the default character set rule(s)
 * 	for a particular standard.
 *	</p>
 * 	<hr/>
 * 	<h3>About The "Content-Type" Header and HTTP</h3>
 * 	<p>RFCs <a href='http://www.ietf.org/rfc/rfc0822.txt'>822</a>, 
 * 	<a href='http://www.ietf.org/rfc/rfc2045.txt'>2045</a>, <a href='http://www.ietf.org/rfc/rfc2046.txt'>2046</a> and 
 * 	<a href='http://www.w3.org/Protocols/rfc2616/rfc2616.html'>2616</a> share a substantially common set of rules 
 * 	for the their "Content-Type" identifiers. Seen from the perspective of the HTTP 1.1 standard (RFC 2616), which is 
 * 	the most derivative of the standards with regard to "Content-Type", the ultimate set of rules are derived in 
 * 	the following manner:
 * 	</p>
 *	<p>The HTTP 1.1 "Content-Type" header is discussed in 
 * 	<a href='http://www.w3.org/Protocols/rfc2616/rfc2616-sec14.html#sec14.17'>section 14.17 of RFC 2616</a>. The 
 * 	"Content-Type" header specifies a media type. Media types are discussed in 
 * 	<a href='http://www.w3.org/Protocols/rfc2616/rfc2616-sec3.html#sec3.7'>section 3.7 of RFC 2616</a>. Section 3.7 
 * 	mentions several important facts:</p>
 *	<ul>	
 *	<li>"The type, subtype, and parameter attribute names are case-insensitive."</li>
 *	<li>"Parameter values might or might not be case-sensitive, depending on the semantics of the parameter name."</li>
 * 	<li>"Linear white space (LWS) MUST NOT be used between the type and subtype, nor between an attribute and its 
 * 	value."</li>
 *	</ul>	
 *	<p>Section 3.7.1 of RFC 2616, "Canonicalization and Text Defaults", also has something interesting to say:</p>
 *	<blockquote>When no explicit charset parameter is provided by the sender, media subtypes of the "text" type are 
 * 	defined to have a default charset value of "ISO-8859-1" when received via HTTP.
 * 	</blockquote>	
 *	<p>Section 3.7.2 of RFC 2616, "Multipart Types", states that multi-part HTTP uses the same syntax defined in 
 * 	<a href='http://www.ietf.org/rfc/rfc2046.txt'>RFC 2046</a>, section 5.1.1. The "abstract" at the start of RFC 2046 
 * 	states:
 * 	</p>
 *	<blockquote>	
 *	The initial document in this set, RFC 2045, specifies the various headers used to describe the structure of MIME 
 *	messages. This second document defines the general structure of the MIME media typing system and defines an initial 
 *	set of media types. [....]
 *	</blockquote>
 *	<p>Thus, the HTTP 1.1 standard (RFC 2616) bases its handling of mutlipart data on RFC 2046, and RFC 2046 depends on 
 *	RFC 2045 to specify the relevant headers. Section 5 of <a href='http://www.ietf.org/rfc/rfc2045.txt'>RFC 2045</a>, 
 *	defines the "Content-Type" header, and thus seems to be the ultimate source of the definition of the "Content-Type" 
 *	header of HTTP 1.1 (RFC 2616). Section 5.1 of RFC 2045 provides the detailed defintion of the "Content-Type" header 
 * 	syntax (using the augmented BNF of <a href='http://www.ietf.org/rfc/rfc0822.txt'>RFC 822</a>):
 *	</p>
 * 	<pre>
    content := "Content-Type" ":" type "/" subtype
               &#42;(";" parameter)
	</pre>
 *	<p>Where "<code>parameter</code>" is defined as:</p>
 *	<pre>
    parameter := attribute "=" value
	</pre>
 *	<p>And "<code>value</code>" is defined as:</p>
 *	<pre>
    value := token / quoted-string
	</pre>
 *	<p>However, while "<code>token</code>" is defined in section 5.1 of RFC 2045, "<code>quoted-string</code>" is 
 * 	<i>not</i> defined in RFC 2045 at all. To find the definition of "<code>quoted-string</code>" we have to refer back 
 * 	to RFC 822, section 3.3, "Lexical Tokens", which supplies the following definition:
 *	</p>
 * 	<pre>
    quoted-string = &lt;"&gt; *(qtext/quoted-pair) &lt;"&gt;; Regular qtext or
                                                ;   quoted chars.
	</pre>	
 *	<p>Where "<code>qtext</code>" is defined as:
 *	</p>
 * 	<pre> 
    qtext       =  &lt;any CHAR excepting &lt;"&gt;,     ; => may be folded
                    "\" & CR, and including
                    linear-white-space>
	</pre>
 *	<p>And "<code>quoted-pair</code>" is defined as:
 * 	</p>
 *	<pre>
    quoted-pair =  "\" CHAR                     ; may quote any char
	</pre>
 *	<p>So, a "<code>quoted-pair</code>" identifies a conventional backslash ('\') based single-character escaping 
 * 	mechanism, as explained in detail in RFC 822 section 3.4.1, where it is referred to as "quoting". That's a 
 * 	misleading choice of terminology, because the escaping mechanism can apply to any character, not just quotes, but 
 * 	that's the term RFC 822 uses, so we're stuck with it.
 *	</p>	
 *	<p>By the way, RFC 2045, section 5.1, helpfully includes the following examples of "Content-Type" headers:
 *	</p>
 * 	<pre> 
    Content-type: text/plain; charset=us-ascii (Plain text)

    Content-type: text/plain; charset="us-ascii"
	</pre>
 *	<p>These do not illustrate the quoted-pair mechanism, but do illustrate the concepts of parameters whose values are 
 *	tokens (<code>charset=us-ascii</code>), parameters whose values are quoted-strings (<code>charset="us-ascii"</code>), 
 * 	and parenthetical comments (<code>(Plain text)</code>). Comments are not included in the RFC 2045 BNF defining 
 * 	"<code>parameter</code>", but the text of section 5.1 states: "comments are allowed in accordance with RFC 822 rules 
 * 	for structured header fields". So, back we go to RFC 822, where we find the following definition of 
 * 	"<code>comment</code>" in section 3.3:
 *	<pre>	
    comment     =  "(" *(ctext / quoted-pair / comment) ")"
	</pre>
 *	<p>Thus it seems that an <a href='http://www.w3.org/Protocols/rfc2616/rfc2616-sec14.html#sec14.17'>RFC 2616 
 * 	"Content-Type"</a> header must support all of the features and syntax defined for it in 
 * 	<a href='http://www.ietf.org/rfc/rfc2045.txt'>RFC 2045</a> and 
 * 	<a href='http://www.ietf.org/rfc/rfc0822.txt'>RFC 822</a>. That support includes: an unlimited number of parameters 
 * 	following the type/subtype; parameters including a trailing, parenthetical comment; quoted parameter values; and a 
 * 	backslash-based escaping mechanism for use within quoted parameter values and comments.
 * 	</p>
 * 	
 * 	@author Chris W. Johnson
 */
public class ContentType {
	
	private static final Logger					Log                  = Logger.getLogger(ContentType.class.getName());
	private static final Pattern				kMIMETypePat         = Pattern.compile("^(\\S+?)/(\\S+?)$");
	private static final ContentTypeParameter[]	kEmptyParameterArray = {};
	
	private final String						Type;
	private final String						Subtype;
	private final String						CharSet;
	private final boolean						CharSetSpecified;
	private final String						OrigStr;
	private final ContentTypeParameter[]		Params;


/**	Creates a <code>ContentType</code> instance which is a parsed representation of the value of a "Content-Type"
 * 	header.
 * 
 * @param ContentTypeStr				The value of a "Content-Type" header. For example: "<code>text/html; 
 * 										charset="UTF-8" (Unicode 8-bit Encoding)</code>".
 * @param DefaultCharSet				An instance of a class that can determine the default character set appropriate
 * 										to a particular standard in the absence of a "charset" parameter in the content
 * 										type value. Can be <code>null</code>, in which case the "default" charset will
 * 										also be <code>null</code>.
 */
	public ContentType
	(
		final String					ContentTypeStr,	//	Can be null (but shouldn't be).
		final ContentTypeDefaultCharSet	DefaultCharSet	//	Can be null (but shouldn't be).
	)
	{
		if (ContentTypeStr == null || ContentTypeStr.trim().length() == 0) {
			
			Type             = null;
			Subtype          = null;
			CharSet          = DefaultCharSet != null ? DefaultCharSet.get(null, null) : null;
			CharSetSpecified = false;
			OrigStr          = ContentTypeStr;
			Params           = kEmptyParameterArray;
			
			return;
		}
		
		final Subpart[][]		Parts = new Parser(ContentTypeStr).getParts();
		
		if (Parts.length == 0)
			throw new IllegalArgumentException("No content-type information was found in \"" + ContentTypeStr + "\".");
		
		final Subpart[]			FirstPart = Parts[0];
		
		if (FirstPart.length == 0)
			throw new IllegalArgumentException("The first part of the content type \"" + ContentTypeStr + "\" was empty.");
		
		if (FirstPart[0].Type != SubpartType.Plain)
			throw new IllegalArgumentException("The first piece of a content type must be plain text. Instead, the first piece is of type " + FirstPart[0].Type + " and contains the text: \"" + FirstPart[0].Value + "\".");
		
	//	Regarding case sensitivity, RFC 2045, <http://www.ietf.org/rfc/rfc2045.txt>, section 5.1, "Syntax of the 
	//	Content-Type Header Field", states:
	//	
	//	"The type, subtype, and parameter names are not case sensitive. For example, TEXT, Text, and TeXt are all 
	//	equivalent top-level media types. Parameter values are normally case sensitive, but sometimes are interpreted 
	//	in a case-insensitive fashion, depending on the intended use."
		
		final int				FirstParamIndex;
		
		{
			final Matcher			Match = kMIMETypePat.matcher(FirstPart[0].Value);
			
			if (Match.matches()) {
				
			//	With regard to the following lower-case conversions, see the definition of the "content" type in 
			//	RFC 2045, <http://www.ietf.org/rfc/rfc2045.txt>, subsection 5.1, "Syntax of the Content-Type Header 
			//	Field", where it states: "Matching of media type and subtype is ALWAYS case-insensitive".
			//
			//	Conversion to lower case does not guarantee case-insensitivity, of course. What it does guarantee is 
			//	that all classes that use this class can use lower-case type and subtype values as the basis for their
			//	comparisons. That, ultimately, is how case-insensitivity is achieved.
				
				Type            = Match.group(1).toLowerCase();
				Subtype         = Match.group(2).toLowerCase();
				FirstParamIndex = 1;
				
			} else {
				
				Type            = null;
				Subtype         = null;
				FirstParamIndex = 0;
				
				Log.log(Level.INFO, "The MIME type is missing from the content type specifier \"{0}\". This is illegal, and the MIME type and subtype will be recorded as null, but the processing of any attributes will be allowed to proceed normally.", ContentTypeStr);
				
			//	throw new IllegalArgumentException("The string \"" + FirstPart[0] + "\" is not formatted correctly to specify a MIME type.");
			}
		}
		
		Params = new ContentTypeParameter[Parts.length - FirstParamIndex];
		
		for (int PartIndex = FirstParamIndex, ParamIndex = 0; PartIndex < Parts.length; PartIndex++, ParamIndex++)
			Params[ParamIndex] = new ContentTypeParameter(Parts[PartIndex]);
		
	//	Get the character set (if any) for this MIME type, possibly resorting to a protocol-specific default supplied
	//	by the ContentTypeDefaultCharSet implementation optionally specified by the DefaultCharSet parameter.
		
		final ContentTypeParameter	CharSetParam = getParameter("charset");
		String						CharSet      = null;
		
		if (CharSetParam != null) {
			
			CharSetSpecified = true;	//	todo Consider: If the value of the "charset" attribute is blank, or (if it is possible) null, should CharSetSpecified be set to false?
			
			CharSet = CharSetParam.Value;
			
		//	Enclosing the characater set name in single quotes is not legal, according to the standards. However, we
		//	have encountered it. So, while the single quotes are ignored for purposes of parsing, we will take notice
		//	of them here in order to remove them from a place where they definitely do not belong. (The altnerative is
		//	throwing an Exception, which doesn't seem nearly as useful.)
			
			if (CharSet.startsWith("'") && CharSet.endsWith("'")) {
				
				CharSet = CharSet.substring(1, CharSet.length() - 1);
				Log.log(Level.INFO, "Content-Type attribute values may only be quoted using double quote characters (''\"''). In the content type specifier \"{0}\", however, the value of the \"charset\" attribute (\"{1}\") is enclosed in single quote characters (\"''\"). This is illegal, but the quotes will be removed and the error overlooked.", new Object[] { ContentTypeStr, CharSet });
			}
			
		} else {
			
			CharSetSpecified = false;
			
			if (DefaultCharSet != null)
				CharSet = DefaultCharSet.get(Type, Subtype);
		}
		
	//	RFC 2046, <http://www.ietf.org/rfc/rfc2046.txt>, section 4.1.2, "Charset Parameter", states: "Unlike some other
	//	parameter values, the values of the charset parameter are NOT case sensitive." Thus the conversion to canonical
	//	capitalization that is performed below is valid, and can't violate any expectations about case, since no 
	//	expectations are allowed.
	
		if (CharSet != null) {
			
			try {
				
			//	Attempt to canonicalize the character set name. Nothing requires that we do this, it just seems like a
			//	nice feature. If it doesn't work, we revert to reporting the name as it was originally specified.
				
				CharSet = Charset.forName(CharSet).displayName();
				
			} catch (IllegalCharsetNameException e) {
			//	Do nothing. It's not our job to make sure that the character set name is valid, just to report it.
			} catch (UnsupportedCharsetException e) {
			//	Do nothing. It's not our job to make sure that the character set is supported, just to report it.
			}
		}
		
	//	Record the character set name (canonicalized, if possible), and the original content type string, once and for
	//	all.
		
		this.CharSet = CharSet;
		this.OrigStr = ContentTypeStr;
	}

/**	The content type string passed to the constructor.
 * 
 * @return						The content type string passed to the constructor.
 */
	@Override public String toString() {
		return OrigStr;	
	}
	
/**	The <a href='http://www.iana.org/assignments/media-types/'>MIME media type</a> of this content. For example, if the 
 * 	MIME media type is "text/plain", this method returns "text/plain".
 * 
 * @return						The MIME type of this content, or <code>null</code> if the MIME type was missing.
 */
	public String getMIMEType() {
		return Type != null ? Type + '/' + Subtype : null;
	}
	
/**	The MIME type of this content. For example, if the <a href='http://www.iana.org/assignments/media-types/'>MIME 
 * 	media type</a> is "text/plain", the type is "text".
 * 
 * @return						The MIME type of this content, or <code>null</code> if the MIME type was missing.
 */
	public String getType() {
		return Type;
	}

/**	The MIME subtype of this content. For example, if the <a href='http://www.iana.org/assignments/media-types/'>MIME 
 * 	media type</a> is "text/plain", the subtype is "plain".
 * 
 * @return						The MIME subtype of this content, or <code>null</code> if the MIME type was missing.
 */
	public String getSubtype() {
		return Subtype;
	}

/**	Returns the canonicalized name of the character set identified by the "charset" parameter, if that parameter was 
 * 	present.
 * 
 * @return						The canonicalized name of the character set identified by the "charset" parameter, or
 * 								<code>null</code> if the parameter was absent, and the default character set could not
 * 								be determined.
 */
	public String getCharacterSet() {
		return CharSet;
	}

/**	Returns <code>true</code> if the character set was explicity identified in the content type, and <code>false</code>
 * 	if it was missing. The ability, or inability, to obtain a definitive character set identification from other 
 * 	sources, like a protocol's defaults, is irrelevant to the value returned by this method.
 * 
 * @return						<code>true</code> if the character set was explicity identified in the content type, or
 * 								<code>false</code> otherwise.
 */
	public boolean getCharacterSetWasSpecified() {
		return CharSetSpecified;
	}
	
/**	Returns the parameter that has the specified, case-insensitive name, or <code>null</code> if this content type did
 * 	not include the specified parameter.
 * 
 * @param ParamName				The case-insensitive name of the parameter to retrieve.
 * @return						The requested parameter, or <code>null</code>, if there was no such parameter.
 */	
	public ContentTypeParameter getParameter
	(
		final String			ParamName	
	) 
	{
		for (final ContentTypeParameter Param : Params) {
			
			if (ParamName.equalsIgnoreCase(Param.Name))
				return Param;
		}
		
		return null;
	}

/**	Returns all of the parameters included in this content type.
 * 
 * @return						The parameters included in this content type, or an empty array if there were no 
 * 								parameters.
 */
	public ContentTypeParameter[] getParameters() {
		return Params;	//	todo Consider: Should we clone this array for safety?
	}
	
	
	private static class Parser {
		
		final ArrayList<Subpart[]>	Parts    = new ArrayList<Subpart[]>();
		final ArrayList<Subpart>	Subparts = new ArrayList<Subpart>();
		final StringBuilder			Buff;
		boolean						Escaped;
		SubpartType					SubType;
		
		
		Parser
		(
			final String		Str
		)
		{
			this.Buff    = new StringBuilder(Str.length());
			this.SubType = SubpartType.Plain;
			
			for (final char Char : Str.toCharArray()) {
			
				if (Escaped) {
				
					Buff.append(Char);
					Escaped = false;
					continue;
				}
			
				if (SubType != SubpartType.Plain) {
					
					if (Char == SubType.Terminator)
						endSubpart();
					else if (Char == '\\')
						Escaped = true;
					else
						Buff.append(Char);
					
				} else {
					
					switch (Char) {
							
						case '"':
							
							beginSubpart(SubpartType.Quote);
							break;
							
						case '(':
							
							beginSubpart(SubpartType.Comment);
							break;
							
						case ';':
							
							addPart();
							break;
							
						default:
							
							Buff.append(Char);
					}
				}		
			}
			
			addPart();
		}
		
		private void beginSubpart
		(
			final SubpartType		SubType
		)
		{
			endSubpart();
			this.SubType = SubType;
			
		//	Note: We would clear Buff here, but the endSubpart method (invoked above) has already done so.
		}
		
		private void endSubpart() {
			String					Str = Buff.toString();
			
			Buff.setLength(0);
			
			if (SubType == SubpartType.Plain) {
				
				Str = Str.trim();
				if (Str.length() == 0)
					return;	
			}
			
			Subparts.add(new Subpart(SubType, Str));
			SubType = SubpartType.Plain;	//	Restore the default value. (Otherwise, the parser malfunctions.)
		}
		
		private void addPart() {
			
			endSubpart();
			
			final int				SubpartCount = Subparts.size();
			
			if (SubpartCount != 0) {
				
				Parts.add(Subparts.toArray(new Subpart[SubpartCount]));
				Subparts.clear();
			}
			
			beginSubpart(SubpartType.Plain);
		}
		
		Subpart[][] getParts() {
			return Parts.toArray(new Subpart[Parts.size()][]);
		}
	}
	
	enum SubpartType { 
		
		Plain(';'), 
		Quote('"'),
		Comment(')');
		
		final char				Terminator;
		
		SubpartType(final char Terminator) {
			this.Terminator = Terminator;
		}
	}
	
	static class Subpart {
		
		final SubpartType			Type;
		final String				Value;
		
		
		Subpart
		(
			final SubpartType		Type,
			final String			Value
		)
		{
			this.Type  = Type;
			this.Value = Value;
		}
		
		@Override public String toString() {
			return Value;	
		}
	}
}
