/*	QWICAP (Quick Web Interface for Conventional Applications) by Chris W. Johnson
	Copyright (C) 2010 Chris W. Johnson. All rights reserved.
	
	Chris W. Johnson's copyright notice in this evolving version reflects original authorship. It does not assert 
	copyright over changes contributed by others.
	
	This library is free software; you can redistribute it and/or modify it under the terms of the GNU Lesser General 
	Public License as published by the Free Software Foundation; either version 2.1 of the License, or (at your option) 
	any later version.
	
	This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied 
	warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for 
	more details.
	
	You should have received a copy of the GNU Lesser General Public License along with this library; if not, write to 
	the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/


package edu.utexas.its.eis.tools.qwicap.servlet;


/**
 * Created by IntelliJ IDEA.
 * User: chrisj
 * Date: Feb 7, 2010
 * Time: 11:06:09 PM
 * To change this template use File | Settings | File Templates.
 */
public final class ContentTypeParameter {
	
/**	The name of the parameter. */
	
	public final String				Name;
	
/**	The value of the parameter, decoded if necessary. */
	
	public final String				Value;
	
/**	The optional comment associated with the parameter, decoded if necessary. If no comment was present, the value of
 * 	this field is <code>null</code>. 
 */
	public final String				Comment;	//	null if there was no comment.
	
	
	ContentTypeParameter
	(
		final ContentType.Subpart[]	Pieces
	)
	{
		if (Pieces.length == 0)
			throw new IllegalArgumentException("An array containing at least one element must be supplied.");
		
		int							PieceIndex = 0;
		final ContentType.Subpart	FirstPart  = Pieces[PieceIndex];
		
		if (FirstPart.Type != ContentType.SubpartType.Plain)
			throw new IllegalArgumentException("The type of \"" + FirstPart.Value + "\" is \"" + FirstPart.Type + "\" where plain text was expected.");
		
		final int							FirstPartEqualIndex = FirstPart.Value.indexOf('=');
		
		if (FirstPartEqualIndex == -1)
			throw new IllegalArgumentException("The equal sign ('=') expected in \"" + FirstPart.Value + "\" is not present.");
	
	//	Regarding case sensitivity RFC 2045, <http://www.ietf.org/rfc/rfc2045.txt>, section 5.1, "Syntax of the 
	//	Content-Type Header Field", states:
	//	
	//	"The type, subtype, and parameter names are not case sensitive. For example, TEXT, Text, and TeXt are all 
	//	equivalent top-level media types. Parameter values are normally case sensitive, but sometimes are interpreted 
	//	in a case-insensitive fashion, depending on the intended use."
		
		Name = FirstPart.Value.substring(0, FirstPartEqualIndex).trim().toLowerCase();
		
		if (FirstPartEqualIndex == FirstPart.Value.length() - 1) {
			
		//	The first piece ends with an equal sign, so its value must be in the next piece, which will be of type 
		//	"Quote".
			
			if (++PieceIndex >= Pieces.length)
				throw new IllegalArgumentException("The second piece of the parameter \"" + FirstPart + "\" is missing.");
			if (Pieces[PieceIndex].Type != ContentType.SubpartType.Quote)
				throw new IllegalArgumentException("The second piece of the parameter \"" + FirstPart + "\" is not quoted, as expected. Instead it is of type \"" + Pieces[1].Type + "\".");
			
			Value = Pieces[PieceIndex].Value;
			
		} else {
			
		//	The first piece includes characters after the equal sign. Those characters will be the value.
			
			Value = FirstPart.Value.substring(FirstPartEqualIndex + 1).trim();
		}
		
	//	If there is another piece, it should be a comment....
		
		if (++PieceIndex < Pieces.length) {
			final ContentType.Subpart	SecondPart = Pieces[PieceIndex];
			
			if (SecondPart.Type != ContentType.SubpartType.Comment)
				throw new IllegalArgumentException("Where a comment was expected, a \"" + SecondPart.Type + "\" was found. (Its value was \"" + SecondPart + "\".)");
			
			Comment = SecondPart.Value;
			
		} else
			Comment = null;
	}
	
	@Override public String toString() {
	
		if (Comment != null) {
		
			return "Name:    \"" + Name    + "\"\n" +
				   "Value:   \"" + Value   + "\"\n" + 
				   "Comment: \"" + Comment + "\"\n";
		}
	
		return "Name:  \"" + Name  + "\"\n" +
			   "Value: \"" + Value + "\"\n"; 
	}
}
