/*	QWICAP (Quick Web Interface for Conventional Applications) by Chris W. Johnson
	Copyright (C) 2003-2007 The University of Texas at Austin ("U.T. Austin"). All rights reserved.
	
	U.T. Austin's copyright notice in this evolving version reflects our original authorship. It does not assert 
	copyright over changes contributed by others.
	
	This library is free software; you can redistribute it and/or modify it under the terms of the GNU Lesser General 
	Public License as published by the Free Software Foundation; either version 2.1 of the License, or (at your option) 
	any later version.
	
	This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied 
	warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for 
	more details.
	
	You should have received a copy of the GNU Lesser General Public License along with this library; if not, write to 
	the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/


package edu.utexas.its.eis.tools.qwicap.servlet;


import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import javax.servlet.ServletContext;
import java.io.File;
import java.io.FileNotFoundException;
import java.net.URL;
import java.net.MalformedURLException;


/**	<code>Context</code> provides a few utility methods, as well as convenient access to hit-specific information, like 
 * 	the {@link HttpServletRequest} and {@link HttpServletResponse} objects. Most of the time, Qwicap applications won't 
 * 	need this class, but you can obtain the current instance of it by invoking the {@link Qwicap#getContext} method.
 *	
 *	@author Chris W. Johnson
 */
public class Context {

	private final Qwicap		Q;


	Context
	(
		final Qwicap			Q
	)
	{
		this.Q = Q;
	}

/**	Returns the current <code>HttpServlet</code> instance.
 * 
 * @return						The current <code>HttpServlet</code>.
 */
	public QwicapServlet getServlet() {
		return Q.getHitContext().getServlet();
	}

/**	Returns the current <code>HttpServletRequest</code> instance.
 * 
 * @return						The current <code>HttpServletRequest</code>.
 */
	public HttpServletRequest getRequest() {
		return Q.getHitContext().getRequest();
	}

/** Returns the current <code>HttpServletResponse</code> instance.
 * 
 * @return						The current <code>HttpServletResponse</code>.
 */
	public HttpServletResponse getResponse() {
		return Q.getHitContext().getResponse();
	}

/* Returns the current {@link FormDataSet} instance.
 * 
 * @return						The current {@link FormDataSet}.
 * @deprecated					Use {@link Qwicap#getFormDataSet} instead.
 *
	public FormDataSet getFormDataSet() {
		return Q.getFormDataSet();
	}
*/

/**	Returns a <code>File</code> object corresponding to the specified file (or directory), which may, or may not, exist.
 * 
 * @param FileName				The name/path of a local file.
 * @return						A <code>File</code> object corresponding to the specified local file, whether it 
 *								actually exists, or not.
 * @see #getExistingFile
 */
	public File getFile
	(
		final String			FileName
	)
	throws FileNotFoundException
	{
		final QwicapServlet		Srv = getServlet();

		if (Srv == null)
			throw new IllegalStateException("Cannot complete getFile(\"" + FileName + "\") because the QwicapServlet reference has become null, which should only happen after the web application has exited.");

		final ServletContext	Ctxt = Srv.getServletContext();

		if (Ctxt == null)
			throw new IllegalStateException("Cannot complete getFile(\"" + FileName + "\") because HttpServlet.getContext returned null.");

		return new File(Ctxt.getRealPath(FileName));
	}

/**	Returns a <code>File</code> object corresponding to the specified file (or directory), if it exists. If it does not
 *	exist, a <code>FileNotFoundException</code> is thrown. The message in that exception will state the file/path that
 *	was requested, and the absolute path of the location where this method believed the file would be found.
 * 
 * @param FileName					The name/path of the local file that the caller asserts must exist.
 * @return							A <code>File</code> object corresponding to the specified local file, if the 
 *									file exists.
 * @throws FileNotFoundException	If the file does not exist.
 */
	public File getExistingFile
	(
		final String			FileName
	)
	throws FileNotFoundException
	{
		final File				TheFile = getFile(FileName);

		if (TheFile.exists() == false)
			throw new FileNotFoundException("The file \"" + FileName + "\" was not found at absolute path \"" + TheFile.getAbsolutePath() + "\".");

		return TheFile;
	}

/**	Returns the {@link URL} for a resource supplied by the servlet container, for instance an HTML file within your web
 * 	application. The name must begin with a slash character ('/'), so, if you wanted the URL for a file named 
 * 	"index.html" located at the top level of your web application, you would pass the string <samp>"/index.html"</samp>.
 * 
 * @param ResourceName				The name/path of the resource to be supplied by the servlet container.
 * @return							The URL of the resource supplied by the servlet container, or <code>null</code> if 
 * 									the resource could not be found.
 * @throws MalformedURLException	If <code>ResourceName</code> does not begin with a slash character ('/'), or if 
 * 									more conventional URL formation rules have somehow been violated.
 * @see ServletContext#getResource
 */
	public URL getResource
	(
		final String			ResourceName
	)
	throws MalformedURLException
	{
		final QwicapServlet		Srv = getServlet();

		if (Srv == null)
			throw new IllegalStateException("Cannot complete getResource(\"" + ResourceName + "\") because the QwicapServlet reference has become null, which should only happen after the web application has exited.");

		final ServletContext	Ctxt = Srv.getServletContext();

		if (Ctxt == null)
			throw new IllegalStateException("Cannot complete getResource(\"" + ResourceName + "\") because HttpServlet.getContext returned null.");
		
		return Ctxt.getResource(ResourceName);
	}

/**	Returns the {@link URL} for a resource supplied by the servlet container, for instance an HTML file within your web
 * 	application. The name must begin with a slash character ('/'), so, if you wanted the URL for a file named 
 * 	"index.html" located at the top level of your web application, you would pass the string <samp>"/index.html"</samp>.
 * 
 * @param ResourceName				The name/path of the resource to be supplied by the servlet container.
 * @return							The URL of the resource supplied by the servlet container.
 * @throws MalformedURLException	If <code>ResourceName</code> does not begin with a slash character ('/'), or if 
 * 									more conventional URL formation rules have somehow been violated.
 * @throws FileNotFoundException	If the requested resource could not be found.
 * @see ServletContext#getResource
 */
	public URL getExistingResource
	(
		final String			ResourceName
	)
	throws FileNotFoundException, MalformedURLException
	{
		final URL				RsrcURL = getResource(ResourceName);
		
		if (RsrcURL == null)
			throw new FileNotFoundException("The resource \"" + ResourceName + "\" could not be located by the servlet container.");
		
		return RsrcURL;
	}

/**	If your web application is using session attributes, use this method to retrieve them, in order to avoid naming 
 *	conflicts.
 * 
 * @param Name					The name of the session attribute to be retrieved.
 * @return						The value of the session attribute.
 */
	public Object getSessionAttribute
	(
		final String			Name
	)
	{
		if (Name.startsWith("qwicap"))
			throw new IllegalArgumentException("Session attribute names are not allowed to begin with \"qwicap\"; those names are reserved for Qwicap's use.");

		return Q.getSessionContext().getSessionAttribute(Name);
	}

/**	If your web application is using session attributes, use this method to set them, in order to avoid naming 
 *	conflicts.
 * 
 * @param Name					The name of the session attribute to be retrieved.
 * @param Value					The value of the session attribute.
 */
	public void setSessionAttribute
	(
		final String			Name,
		final Object			Value
	)
	{
		if (Name.startsWith("qwicap"))
			throw new IllegalArgumentException("Session attribute names are not allowed to begin with \"qwicap\"; those names are reserved for Qwicap's use.");

		Q.getSessionContext().setSessionAttribute(Name, Value);
	}

/**	If your web application is using session attributes, use this method to remove them, in order to avoid naming 
 *	conflicts.
 * 
 * @param Name					The name of the session attribute to be removed.
 */
	public void removeSessionAttribute
	(
		final String			Name
	)
	{
		if (Name.startsWith("qwicap"))
			throw new IllegalArgumentException("Session attribute names are not allowed to begin with \"qwicap\"; those names are reserved for Qwicap's use.");

		Q.getSessionContext().removeSessionAttribute(Name);
	}
}
