/*	QWICAP (Quick Web Interface for Conventional Applications) by Chris W. Johnson
	Copyright (C) 2007-2011 Chris W. Johnson. All rights reserved.
	
	Chris W. Johnson's copyright notice in this evolving version reflects his original authorship. It does not assert 
	copyright over changes contributed by others.
	
	This library is free software; you can redistribute it and/or modify it under the terms of the GNU Lesser General 
	Public License as published by the Free Software Foundation; either version 2.1 of the License, or (at your option) 
	any later version.
	
	This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied 
	warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for 
	more details.
	
	You should have received a copy of the GNU Lesser General Public License along with this library; if not, write to 
	the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/


package edu.utexas.its.eis.tools.qwicap.servlet;


import edu.utexas.its.eis.tools.qwicap.template.xml.TagException;
import edu.utexas.its.eis.tools.qwicap.template.xml.immutable.XMLDocument;
import edu.utexas.its.eis.tools.table.Table;
import edu.utexas.its.eis.tools.table.lltable.MutableLLTable;

import javax.servlet.http.HttpServletResponse;
import java.io.File;
import java.io.IOException;
import java.io.OutputStream;
import java.net.MalformedURLException;
import java.net.URL;


/**
 * Created by IntelliJ IDEA.
 * User: chrisj
 * Date: Sep 26, 2007
 * Time: 9:36:34 AM
 * To change this template use File | Settings | File Templates.
 */
class DocumentCache {
	
	private static final int							kArbitraryDocCountIncrease = 10;
	
	private final LRUCacheMap<String, DocumentSource>	Cache;
	private final MIMETypeGuesser						MIMEGuesser;
	private final DocumentCensus						Census;
	
	
	DocumentCache
	(
		final MIMETypeGuesser	MIMEGuesser,
		final DocumentCensus	DocCensus,
		final int				LocalizationCount	//	Note that the default localization is NOT counted.
	)
	{
		final int				PotentialDocCount = DocCensus.getDocumentNameCount() * (LocalizationCount + 1) + kArbitraryDocCountIncrease;
		
		this.MIMEGuesser = MIMEGuesser;
		this.Cache       = new LRUCacheMap<String, DocumentSource>(PotentialDocCount, PotentialDocCount * 2);
		this.Census      = DocCensus;
	}
	
	DocumentCensus getCensus() {
		return Census;	
	}
	
	int getDiscardCount
	(
		final boolean			Reset
	) 
	{
		return Cache.getDiscardCount(Reset);
	}
	
	String guessMIMEType
	(
		final String			URLStr
	)
	{
		return MIMEGuesser.getGuess(URLStr);	//	Returns null, if no guess can be made.
	}
	
	DocumentSource getDocumentSource
	(
		final File				DocFile
	)
	throws MalformedURLException
	{
		return getDocumentSource(DocFile.toURI().toURL());
	}

	DocumentSource getDocumentSource
	(
		final URL				DocURL
	)
	{
		final String			DocURLStr = DocURL.toString();
		DocumentSource			DocSrc;
		
		synchronized (Cache) {
			
			DocSrc = Cache.get(DocURLStr);
	
			if (DocSrc == null) {
	
				DocSrc = new DocumentSource(this, DocURL);
				Cache.put(DocURLStr, DocSrc);	
			}
		}
		
	//	We can, and do, have cache entries for non-existent documents, so that we don't have to waste time establishing
	//	that they don't exist more than once. So, we invoke DocSrc.checkExists() which will return DocSrc if it the
	//	document exists, or will throw a FileNotFoundException, if it doesn't.
		
		return DocSrc.incrementAccessCount();
	}

	XMLDocument getXMLDocument
	(
		final URL				DocURL
	)
	throws TagException, IOException
	{
		return getDocumentSource(DocURL).checkExists().getXML();
	}

	XMLDocument getXMLDocument
	(
		final File				DocFile
	)
	throws TagException, IOException
	{
		return getDocumentSource(DocFile).checkExists().getXML();
	}

	byte[] getBytes
	(
		final URL				DocURL
	)
	throws IOException
	{
		return getDocumentSource(DocURL).checkExists().getBytes();
	}

	byte[] getBytes
	(
		final File				DocFile
	)
	throws IOException
	{
		return getDocumentSource(DocFile).checkExists().getBytes();
	}

	void writeTo
	(
		final URL					DocURL,
		final HttpServletResponse	Resp
	)
	throws IOException
	{
		getDocumentSource(DocURL).checkExists().writeTo(Resp);
	}

	void writeTo
	(
		final URL				DocURL,
		final OutputStream		Out
	)
	throws IOException
	{
		getDocumentSource(DocURL).checkExists().writeTo(Out);
	}

	void writeTo
	(
		final File					DocFile,
		final HttpServletResponse	Resp
	)
	throws IOException
	{
		getDocumentSource(DocFile).checkExists().writeTo(Resp);
	}

	void writeTo
	(
		final File				DocFile,
		final OutputStream		Out
	)
	throws IOException
	{
		getDocumentSource(DocFile).checkExists().writeTo(Out);
	}
	
	Table getStatus
	(
		final boolean			ResetCounters
	) 
	{
		final MutableLLTable	T = new MutableLLTable().setColNames("Document URL", "State", "Content\nType", "Bytes", "Last Modified", "Access\nCount", "Cache\nHits", "Cache\nMisses", "getXML\nCount", "getBytes\nCount", "writeTo\nCount", "Cached\nForms");
		
		synchronized (Cache) {
			
			for (final DocumentSource DocSrc : Cache.values())
				DocSrc.addStatusData(T.addRow(), ResetCounters);
		}
		
		return T.sortRowsAscending(0);
	}
}
