/*	QWICAP (Quick Web Interface for Conventional Applications) by Chris W. Johnson
	Copyright (C) 2007-2011 Chris W. Johnson. All rights reserved.
	
	Chris W. Johnson's copyright notice in this evolving version reflects his original authorship. It does not assert 
	copyright over changes contributed by others.
	
	This library is free software; you can redistribute it and/or modify it under the terms of the GNU Lesser General 
	Public License as published by the Free Software Foundation; either version 2.1 of the License, or (at your option) 
	any later version.
	
	This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied 
	warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for 
	more details.
	
	You should have received a copy of the GNU Lesser General Public License along with this library; if not, write to 
	the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/


package edu.utexas.its.eis.tools.qwicap.servlet;


import edu.utexas.its.eis.tools.qwicap.template.css.Results;
import edu.utexas.its.eis.tools.qwicap.template.xml.Match;
import edu.utexas.its.eis.tools.qwicap.template.xml.TagException;
import edu.utexas.its.eis.tools.qwicap.template.xml.immutable.XMLDocument;
import edu.utexas.its.eis.tools.qwicap.util.ArrayToString;
import edu.utexas.its.eis.tools.qwicap.util.Null;
import edu.utexas.its.eis.tools.table.MutableTable;

import javax.servlet.http.HttpServletResponse;
import java.io.*;
import java.lang.ref.SoftReference;
import java.net.URL;
import java.net.URLConnection;
import java.text.DecimalFormat;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.logging.Level;
import java.util.logging.Logger;


/*
 * Created by IntelliJ IDEA.
 * User: chrisj
 * Date: Sep 25, 2007
 * Time: 10:21:13 PM
 * To change this template use File | Settings | File Templates.
 */


final class DocumentSource {

	private static final Logger				Log        = Logger.getLogger(DocumentSource.class.getName());
	
	private static final SimpleDateFormat	LastModFmt = new SimpleDateFormat("yyyy-MM-dd HH:mm:ss");
	
	private static final DecimalFormat		BytesFmt   = new DecimalFormat("#,##0 bytes");
	private static final DecimalFormat		IntegerFmt = new DecimalFormat("#,##0");
	
	private static final int				kUpperLimitOnCachedDocumentSize = 1024 * 1024;
	

	enum State { UNKNOWN, EXISTS, NOT_FOUND }
	
	private final DocumentCache		Cache;
	
	private State					DocState = State.UNKNOWN;

	private final String			SourceURLStr;
	private final URL 				SourceURL;
	private int						Length = -1;		//	  -1 if not known.
	private long					LastModified;		//	   0 if not known.
	private String					MIMEType;			//	null if not known.
	private String					CharSet;			//	null if there wasn't even an applicable default.
	private boolean					CharSetSpecified;	//	true if the character set wasn't specified and had to be inferred.
	
	private XMLDocument				CachedXMLDoc;
	private byte[]					CachedBytesHard;
	private SoftReference<byte[]>	CachedBytesSoft;
	
	private final Object			CounterSynchObj = new Object();
	private int						AccessCount;	//	Number of times that DocumentCache has returned this DocumentSource instance.
	private int						CacheHits;		//	Number of times an in-memory representation of this document has satisfied a getXML, getBytes or writeTo invocation.
	private int						CacheMisses;	//	Number of times there was no in-memory representation of this document with which to satisfy a getXML, getBytes or writeTo invocation.
	private int						GetXMLCount;	//	Number of times getXML has been invoked.
	private int						GetBytesCount;	//	Number of times getBytes has been invoked.
	private int						WriteToCount;	//	Number of times writeTo has been invoked.


	DocumentSource
	(
		final DocumentCache		Cache,
		final URL				SourceURL
	)
	{
		this.Cache        = Cache;
		this.SourceURLStr = SourceURL.toString();
		this.SourceURL    = SourceURL;

	//	Attempt to load the document right now, because we need to know whether or not it exists. This effort isn't 
	//	wasted, because, if the document does exist, its bytes are retained in our hard cache, which will be used by 
	//	subsequent invocations of getXMLDocument, writeTo, and getBytes. 
	//
	//	However, this could be a bad strategy if, for instance, the document is something like a multi-gigabyte movie 
	//	file, which might not fit into the JVM's memory, or, if it would fit, might still cause a critical shortage of 
	//	memory while it remains hard-cached.
	//
	//	To solve that problem we impose an upper limit on the size of the items we'll cache.
		
		try {
			final InputStream		In = getInputStream();

			try {
				
				DocState = State.EXISTS;
				
			//	While we have a connection to the document, we might as well read and cache it, provided it is of
			//	a reasonable size.
				
				if (Length >= 0 && Length <= kUpperLimitOnCachedDocumentSize)
					getBytesLow(In);
				
			} finally {
				In.close();
			}

		} catch (IOException e) {

			DocState = State.NOT_FOUND;
			Log.log(Level.FINE, "Attempt to get the bytes of <{0}> failed. The document will be regarded as unavailable, and that fact will be cached.", SourceURL);
			
		//	Discard this exception. It needs to be possible to create DocumentSource instances that represent 
		//	non-existent documents, so the we can cache the fact of their non-existence, rather than wasting time
		//	establishing their non-existence over and over again.
		}
		
	}
	
	DocumentSource incrementAccessCount() {
		
		synchronized (CounterSynchObj) {
			AccessCount++;
		}
		
		return this;
	}

	String getUniqueIDString() {
		return SourceURLStr;
	}
	
	URL getURL() {
		return SourceURL;
	}

	boolean exists() {
		return DocState == State.EXISTS;
	}

	DocumentSource checkExists() throws FileNotFoundException {

		if (!exists())
			throw new FileNotFoundException("The document <" + SourceURLStr + "> does not exist.");
		
		return this;
	}

	State getState() {
		return DocState;
	}

	int getLength() {
		return Length;
	}
	
	String getMIMEType() {
		return MIMEType;
	}

	private int getBufferLength() {
		final int				LengthHint = getLength();

		if (LengthHint <= 0)
			return 64 * 1024;

		if (LengthHint > 1024 * 1024)
			return 1024 * 1024;

		return LengthHint;
	}
	
	private InputStream getInputStream() throws IOException {
		
		synchronized (CounterSynchObj) {
			CacheMisses++;	
		}
		
		final URLConnection 	Conn = SourceURL.openConnection();

		Conn.setDoInput(true);
		Conn.setDoOutput(false);
		
		Length       = Conn.getContentLength();
		LastModified = Conn.getLastModified();
		
	//	Note: Whether the HTTPContentType or XMLContenType class is used, if the content is not XML, the content type 
	//	will not be modified, and no default character set will be supplied (in the absence of an explicit character 
	//	set declaration), so using HTTPContentType and XMLContentType here are, at worst, harmless.
		
		final ContentType		ContentType = Null.toEmptyString(SourceURL.getProtocol()).startsWith("http") ? new HTTPContentType(Conn.getContentType()) : new XMLContentType(Conn.getContentType());
		final String			MIMEType    = ContentType.getMIMEType();
		
	//	todo Consider the implications of the Content-Type header. For instance, if it tells us that the content-type is
	//	     "gzip", then the bytes we read and cache might be gzip compressed (or not, if the URLConnection is 
	//	     automatically uncompressing the content). If the bytes we read do represent gzip compressed content, we 
	//	     we need to pass that information along to clients when the getBytes and writeTo methods are used to send
	//	     send the bytes to a client. We also need to decompress the bytes before attempting to parse them, in the
	//	     case of getXML.
		
		Log.log(Level.FINE, "Connected to document: <{0}>; Length: {1,number,integer}; Content-Type: {2}; Content-Encoding: {3}; Last-Modified: {4,date,yyyy-MM-dd HH:mm:ss}", new Object[] { SourceURLStr, Length, ContentType, Conn.getContentEncoding(), LastModified });
				
		if (Null.orEmpty(MIMEType) || "content/unknown".equals(MIMEType))
			this.MIMEType = Cache.guessMIMEType(SourceURLStr);
		else
			this.MIMEType = MIMEType;
		
		this.CharSet          = ContentType.getCharacterSet();
		this.CharSetSpecified = ContentType.getCharacterSetWasSpecified();
		
		final int				BuffLen;
		
		if (Length < 0)
			BuffLen =   64 * 1024;
		else if (Length >= 1024 * 1024)
			BuffLen = 1024 * 1024;
		else
			BuffLen = Length;

		return new BufferedInputStream(Conn.getInputStream(), BuffLen);
	}

	@Override public int hashCode() {
		return SourceURLStr.hashCode();
	}

	@Override public boolean equals
	(
		final Object			Other
	)
	{
		return Other instanceof DocumentSource && SourceURLStr.equals(((DocumentSource) Other).SourceURLStr);
	}

	XMLDocument getXML() throws IOException, TagException {
		
		synchronized (CounterSynchObj) {
			GetXMLCount++;
		}
		
		checkExists();

		if (CachedXMLDoc != null) {
			
			synchronized (CounterSynchObj) { CacheHits++; }
			
			return CachedXMLDoc;
		}
		
	//	If we have a hard cache of the bytes of this document, use that, instead of reading the document from its source.

		if (CachedBytesHard != null) {
			final ByteArrayInputStream	In = new ByteArrayInputStream(CachedBytesHard);
				
			try {
				CachedXMLDoc = CharSetSpecified ? new XMLDocument(new InputStreamReader(In, CharSet), getLength(), SourceURLStr) : new XMLDocument(In, getLength(), SourceURLStr);
			} finally {
				In.close();
			}
			
			updateContentType();
			
		//	Because we're probably only going to use this document in its parsed, XML form, we should switch from 
		//	maintaining a hard-referenced cache of its bytes to s soft-referenced cache.

			CachedBytesSoft = new SoftReference<byte[]>(CachedBytesHard);
			CachedBytesHard = null;
			
			synchronized (CounterSynchObj) { CacheHits++; }

			return CachedXMLDoc;
		}

	//	If we have a soft cache of the bytes of this document, use that, instead of reading the document from its source.

		if (CachedBytesSoft != null) {
			final byte[]				Bytes = CachedBytesSoft.get();

			if (Bytes != null) {
				final ByteArrayInputStream	In = new ByteArrayInputStream(Bytes);

				try {
					CachedXMLDoc = CharSetSpecified ? new XMLDocument(new InputStreamReader(In, CharSet), getLength(), SourceURLStr) : new XMLDocument(In, getLength(), SourceURLStr);
				} finally {
					In.close();
				}
				
				updateContentType();
				
				synchronized (CounterSynchObj) { CacheHits++; }
				
				return CachedXMLDoc;
				
			} else
				CachedBytesSoft = null;	//	It lost its reference, so it's of no further use to us.
		}

	//	We have no cache of the bytes of this document, so read it from its source.

		final InputStream		In = getInputStream();

		try {
			CachedXMLDoc = CharSetSpecified ? new XMLDocument(new InputStreamReader(In, CharSet), getLength(), SourceURLStr) : new XMLDocument(In, getLength(), SourceURLStr);
		} finally {
			In.close();
		}
		
		updateContentType();
		
		return CachedXMLDoc;
	}
	
	private void updateContentType() {
		final XMLDocument		Doc = CachedXMLDoc;
		
		if (Doc == null)
			return;
		
		if (this.MIMEType == null)
			this.MIMEType = Cache.guessMIMEType(SourceURLStr);
		
		if ("text/html".equalsIgnoreCase(MIMEType)) {
			
		//	A lot of code, from web servers to the URLConnection class, assumes that documents with the extension 
		//	".html" are properly classified as "text/html". However, as there is no single, mandatory extension for
		//	XHTML documents, they commonly use the ".html" extension. Such documents are most correctly identified by
		//	the MIME type "application/xhtml+xml" (see RFC 3236 at <http://www.rfc-editor.org/rfc/rfc3236.txt>),
		//	although they may include "<meta http-equiv="Content-Type" content="..."/>" tags in their HTML "head"
		//	sections that should be respected.
		//
		//	Here we examine documents that are supposedly "text/html" to try to obtain a more correct content type for
		//	them.
			
		//	We already know that this document is essentially valid XML, so we know it's not mere "text/html". So,
		//	for starters, we can identify it as "application/xhtml+xml", which is Qwicap's default document content
		//	type these days.
			
			MIMEType = QwicapServlet.DefaultContentType;	//	These days, the default is: "application/xhtml+xml"
			
		//	Now, see if the document overrides that MIME type using a "<meta http-equiv="Content-Type" content="..."/>"
		//	tag in its HTML "head" section.
			
			final Results			MetaTags = Doc.get("@namespace xhtml url(http://www.w3.org/1999/xhtml); xhtml|html head meta[http-equiv]");
			
			for (final Match MetaTag : MetaTags) {
				
				if ("Content-Type".equalsIgnoreCase(MetaTag.getAttribute("http-equiv"))) {
					
					MIMEType = new XMLContentType(MetaTag.getAttribute("content")).getMIMEType();
					break;
				}
			}
		}
		
		final String			Charset = Doc.getCharacterSet();
		
		if (this.CharSet == null || this.CharSet.equalsIgnoreCase(Charset) == false) {
			
			if (Charset != null) {
				
				this.CharSet          = Charset;
				this.CharSetSpecified = true;	//	We presume that it was specified, or that a definite determination was made by Doc.
			}
		}
	}
	
	private String generateContentType() {
		return generateContentType(false);
	}
	
	private String generateContentType
	(
		final boolean				OmitUnspecifiedCharSet
	) 
	{
		if (MIMEType != null) {
			
			if (CharSet != null) {
				
				if (CharSetSpecified || !OmitUnspecifiedCharSet)
					return MIMEType + ";charset=" + CharSet;
			}
			
			return MIMEType;
		}
		
		return null;
	}
	
	long writeTo
	(
		final HttpServletResponse	Resp
	)
	throws IOException
	{
		final String				ContentType = generateContentType();
		
		if (ContentType != null)
			Resp.setContentType(ContentType);

		if (Length >= 0)
			Resp.setContentLength(Length);
		
		if (LastModified != 0)
			Resp.setDateHeader("Last-Modified", LastModified);
		
		Resp.addHeader("Cache-Control", "max-age=120");
		
	//	todo Define the "Content-Language" header whenever possible.
		
		return writeTo(Resp.getOutputStream());
	}
	
	long writeTo
	(
		final OutputStream		Out
	)
	throws IOException
	{
		synchronized (CounterSynchObj) {
			WriteToCount++;
		}
		
		checkExists();

	//	If we have a hard cache of the bytes of this document, use that, instead of reading the document from its source.

		if (CachedBytesHard != null) {

			Out.write(CachedBytesHard);
			
			synchronized (CounterSynchObj) { CacheHits++; }
			
			return CachedBytesHard.length;
		}

	//	If we have a soft cache of the bytes of this document, use that, instead of reading the document from its source.

		if (CachedBytesSoft != null) {
			final byte[]			Bytes = CachedBytesSoft.get();

			if (Bytes != null) {

				Out.write(Bytes);
				
				synchronized (CounterSynchObj) { CacheHits++; }
				
				return Bytes.length;
				
			} else
				CachedBytesSoft = null;	//	It lost its reference, so it's of no further use to us.
		}
		
	//	If we have a parsed, XML version of the document cached, use that.
	
		if (CachedXMLDoc != null) {
			final PrintWriter		POut = new PrintWriter(Out);
			
			CachedXMLDoc.write(POut);
			POut.flush();
			
			synchronized (CounterSynchObj) { CacheHits++; }
			
			return Length;
		}

	//	We have no cache of the bytes of this document, so read it from its source.

		final InputStream		In = getInputStream();

		try {
			final byte[]			Buff           = new byte[getBufferLength()];
			int						BytesRead      = In.read(Buff);
			long					TotalBytesRead = 0;

			while (BytesRead > 0) {

				TotalBytesRead += BytesRead;

				Out.write(Buff, 0, BytesRead);

				BytesRead = In.read(Buff);
			}
			
			return TotalBytesRead;

		} finally {
			In.close();
		}
	}

	byte[] getBytes() throws IOException {
		
		synchronized (CounterSynchObj) {
			GetBytesCount++;
		}

	//	If we have a hard cache of the bytes of this document, use that, instead of reading the document from its source.

		if (CachedBytesHard != null) {
			
			synchronized (CounterSynchObj) { CacheHits++; }
			
			return CachedBytesHard;
		}
		
	//	If we have a soft cache of the bytes of this document, use that, instead of reading the document from its source.

		if (CachedBytesSoft != null) {
			final byte[]			Bytes = CachedBytesSoft.get();

			if (Bytes != null) {
				
				synchronized (CounterSynchObj) { CacheHits++; }
				
				return Bytes;
				
			} else
				CachedBytesSoft = null;	//	It lost its reference, so it's of no further use to us.
		}

	//	We have no cache of the bytes of this document, so read it from its source, and create a hard cache of those bytes.

		final InputStream		In = getInputStream();

		try {
			return getBytesLow(In);
		} finally {
			In.close();
		}
	}
	
	private byte[] getBytesLow
	(
		final InputStream		In
	)
	throws IOException
	{
		final int				Length         = getLength();
		final byte[]			Buff           = new byte[Length];
		int						TotalBytesRead = 0;
		int						BytesRead      = In.read(Buff);

		while (BytesRead > 0) {

			TotalBytesRead += BytesRead;

			BytesRead = In.read(Buff, TotalBytesRead, Length - TotalBytesRead);
		}

		CachedBytesHard = Buff;
		
		return Buff;
	}
	
	@Override public String toString() {
		final StringBuilder		Buff = new StringBuilder(256);

		Buff.append("Document URL:    <").append(SourceURLStr).append(">\n");
		Buff.append("Document State:  " ).append(DocState).append('\n');
		Buff.append("Content-Type:    " ).append(Null.toEmptyString(generateContentType(true))).append('\n');
		
		if (exists()) {
			
			synchronized (BytesFmt) {
				Buff.append("Document Length: ").append(Length >= 0 ? BytesFmt.format(Length) : "").append('\n');
			}
			
			synchronized (LastModFmt) {
				Buff.append("Last Modified:   ").append(LastModified != 0 ? LastModFmt.format(LastModified) : "").append('\n');
			}
		}
		
		synchronized (CounterSynchObj) {
			
			Buff.append("Cache Hits:      ").append(CacheHits    ).append('\n');
			Buff.append("Cache Misses:    ").append(CacheMisses  ).append('\n');
			Buff.append("getXML Count:    ").append(GetXMLCount  ).append('\n');
			Buff.append("getBytes Count:  ").append(GetBytesCount).append('\n');
			Buff.append("writeTo Count:   ").append(WriteToCount ).append('\n');
		}
		
		if (!exists())
			return Buff.toString();
		
		Buff.append("Cached Forms:    ");

		final ArrayList<String>		Reps = new ArrayList<String>(3);

		if (CachedBytesHard != null)
			Reps.add("Hard Bytes");

		if (CachedBytesSoft != null && CachedBytesSoft.get() != null)
			Reps.add("Soft Bytes");

		if (CachedXMLDoc != null)
			Reps.add("XML");

		Buff.append(new ArrayToString(Reps).neverQuote()).append('\n');
		
		return Buff.toString();
	}
	
	void addStatusData
	(
		final MutableTable		Table,
		final boolean			ResetCounters
	)
	{
		String					LengthStr      = "";
		String					LastModStr     = "";
		String					CachedFormsStr = "";
		
		if (exists()) {
			
			if (Length >= 0) {
				
				synchronized (IntegerFmt) {
					LengthStr = IntegerFmt.format(Length);
				}
			}
			
			if (LastModified != 0) {
				
				synchronized (LastModFmt) {
					LastModStr = LastModFmt.format(LastModified);
				}
			}
			
			{
				final ArrayList<String>	Reps = new ArrayList<String>(3);
				
				if (CachedBytesHard != null)
					Reps.add("Hard Bytes");

				if (CachedBytesSoft != null && CachedBytesSoft.get() != null)
					Reps.add("Soft Bytes");

				if (CachedXMLDoc != null)
					Reps.add("XML");

				CachedFormsStr = new ArrayToString(Reps).neverQuote().toString();
			}	
		}
		
		synchronized (CounterSynchObj) {
			
			Table.add(SourceURLStr)
				 .add(DocState)
				 .add(Null.toEmptyString(generateContentType(true)))
				 .add(LengthStr)
				 .add(LastModStr)
				 .add(AccessCount)
				 .add(CacheHits)
				 .add(CacheMisses)
				 .add(GetXMLCount)
				 .add(GetBytesCount)
				 .add(WriteToCount)
				 .add(CachedFormsStr);
				 
			if (ResetCounters) {
				
				AccessCount   = 0;
				CacheHits     = 0;
				CacheMisses   = 0;
				GetXMLCount   = 0;
				GetBytesCount = 0;
				WriteToCount  = 0;
			}
		}
	}
	
/*	Example status data:
	
	+---------------------------------------------+-----------+-----------+---------+---------------------+--------+-------+--------+--------+----------+---------+--------+
	| Document URL                                | State     | Content   | Length  | Last Modified       | Access | Cache | Cache  | getXML | getBytes | writeTo | Cached |
	|                                             |           | Type      | (Bytes) |                     | Count  | Hits  | Misses | Count  | Count    | Count   | Forms  |
	+---------------------------------------------+-----------+-----------+---------+---------------------+--------+-------+--------+--------+----------+---------+--------+
	| jndi:/localhost/wumpus/en/map-config.html   | NOT_FOUND | text/html |         |                     | 3      | 0     | 1      | 0      | 0        | 0       |        |
	| jndi:/localhost/wumpus/en/wumpus-intro.html | NOT_FOUND | text/html |         |                     | 2      | 0     | 1      | 0      | 0        | 0       |        |
	| jndi:/localhost/wumpus/en/wumpus.html       | NOT_FOUND | text/html |         |                     | 2      | 0     | 1      | 0      | 0        | 0       |        |
	| jndi:/localhost/wumpus/map-config.html      | EXISTS    | text/html | 1,925   | 2007-05-29 07:55:38 | 3      | 3     | 1      | 3      | 0        | 0       | XML    |
	| jndi:/localhost/wumpus/wumpus-intro.html    | EXISTS    | text/html | 4,187   | 2007-05-29 07:55:38 | 2      | 2     | 1      | 2      | 0        | 0       | XML    |
	| jndi:/localhost/wumpus/wumpus.html          | EXISTS    | text/html | 1,491   | 2007-04-25 20:27:54 | 2      | 2     | 1      | 2      | 0        | 0       | XML    |
	+---------------------------------------------+-----------+-----------+---------+---------------------+--------+-------+--------+--------+----------+---------+--------+
*/
}
