/*	QWICAP (Quick Web Interface for Conventional Applications) by Chris W. Johnson
	Copyright (C) 2005-2010 Chris W. Johnson. All rights reserved.
	
	Chris W. Johnson's copyright notice in this evolving version reflects his original authorship. It does not assert 
	copyright over changes contributed by others.
	
	This library is free software; you can redistribute it and/or modify it under the terms of the GNU Lesser General 
	Public License as published by the Free Software Foundation; either version 2.1 of the License, or (at your option) 
	any later version.
	
	This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied 
	warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for 
	more details.
	
	You should have received a copy of the GNU Lesser General Public License along with this library; if not, write to 
	the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/


package edu.utexas.its.eis.tools.qwicap.servlet;


import javax.servlet.http.HttpServletResponse;
import java.io.IOException;
import java.io.OutputStream;
import java.util.Random;


/**	Represents a downloadable entity, typically generated dynamically, that is being, or will be, made available to 
 *	the client via the {@link Qwicap#addDownloadable(Downloadable) Qwicap.addDownloadable} method.
 * 
 * 	@author Chris W. Johnson
 */
public class Downloadable {

	private static final String		kLocalURIPrefix = "qwicapdl/";
	private static final String		kURIPrefix      = '/' + kLocalURIPrefix;

	private final String			Name;
	private final String			MIMEType;
	private final byte[]			Contents;
	private final String			URI;
	private final long				LastModified = System.currentTimeMillis();


	static boolean isDownloadableURI
	(
		final String			URI
	)
	{
		return URI.contains(kURIPrefix);
	}

	static String simplifyURI
	(
		final String			URI
	)
	{
		final int				SimpleStartIndex = URI.indexOf(kLocalURIPrefix);

		if (SimpleStartIndex < 0)
			return null;

		return URI.substring(SimpleStartIndex);
	}

/**	Creates an instance of <code>Downloadable</code> to store the information triplet representing any downloadable 
 *	item: name, <a href='http://www.iana.org/assignments/media-types/'>MIME type</a>, and content bytes. Creating an 
 * 	instance of <code>Downloadable</code> does cause a unique, relative URI to be assigned to the instance, but does 
 * 	not instruct Qwicap to make it available for download. That is done by invoking 
 * 	{@link Qwicap#addDownloadable(Downloadable) Qwicap.addDownloadable}.
 * 
 * @param Rndm					A random number generator which will provide the random number included in this 
 * 								<code>Downloadable</code> instance's URI.
 * @param Name					The name of this downloadable, for example: "myimage.jpg".
 * @param MIMEType				The <a href='http://www.iana.org/assignments/media-types/'>MIME type</a> of this 
 *								downloadable, for example: "image/jpeg".
 * @param Contents				The contents of this download, represented as the array of bytes that will be 
 *								trasmitted to the client.
 */
	Downloadable
	(
		final Random			Rndm,
		final String			Name,
		final String 			MIMEType,
		final byte[]			Contents
	)
	{
		this.Name     = Name;
		this.MIMEType = MIMEType;
		this.Contents = Contents;
		this.URI      = kLocalURIPrefix + Integer.toHexString(Rndm.nextInt()) + '/' + Name;

	//	Ideally, the unique number we incorporate into the URIs of downloadables would be kept small (short) by 
	// 	starting it at zero, and by making it session specific. Unfortunately, browsers with bugs in their caching 
	//	logic (especially Internet Explorer for Mac, and probably the Windows flavor as well) will keep dredging-up 
	//	downloadables with the same URI from previous sessions, even though you've used every HTTP cache control header 
	//	there is in order to tell them not to cache those items at all. This is a disaster for the web application. So, 
	//	we use the following, vaguely unpleasant scheme to make the dredging-up of old downloadables by broken browsers 
	//	highly unlikely: In our URI we include unique numbers with a secure, random value 32-bit value.
	}

/**	Returns the unique, relative URI of this downloadable.
 * 
 * @return						This downloadable's unique URI.
 */
	@Override public String toString() {
		return URI;
	}

/**	Returns the unique, relative URI of this downloadable.
 * 
 * @return						This downloadable's unique URI.
 * @see #toString
 */
	public String getURI() {
		return URI;
	}

/**	Returns the name of this downloadable; if the client saves this image to disk, this will be its file name.
 * 
 * @return						Name of this downloadable.
 */
	public String getName() {
		return Name;
	}

/**	Returns the MIME type of this downloadable.
 * 
 * @return						The MIME type of this downloadable.
 */
	public String getMIMEType() {
		return MIMEType;
	}

/**	Returns the length of this downloadable.
 * 
 * @return						The length of this downloadable.
 */
	public int getContentLength() {
		return Contents.length;
	}
	
	void send
	(
		final HttpServletResponse	Resp
	)
	throws IOException
	{
		Resp.setContentType(MIMEType);
		Resp.setContentLength(Contents.length);
		Resp.addDateHeader("Last-Modified", LastModified);

	//	Prevent this download from being cached.... Except that, since the URLs of downloadables include large random
	//	numbers (which, in practice, proved to be the only way to prevent inappropriate caching behavior by some 
	//	browsers), cache control isn't crucial, and may not always be beneficial. For instance, if it causes a client to
	//	request the same downloadable over-and-over-again, when it will not have changed, that's a waste of time and
	//	resources. (Downloadables are regarded as immutable once they've been created, so re-requesting one will never
	// 	get the client anything new.) On the other hand, we don't want downloadables hanging-around in caches forever. 
	//	Someday, the same URL might be generated, and a stale response delivered. Also, it's probably easier on caches 
	//	if they have some idea of how long to hold onto any response.

		Resp.addHeader("Cache-Control", "max-age=1800");	//	Cache for no more than 30 minutes.

		final OutputStream			Out = Resp.getOutputStream();

		Out.write(Contents);
		Out.flush();
		Resp.flushBuffer();
	}
}
