/*	QWICAP (Quick Web Interface for Conventional Applications) by Chris W. Johnson
	Copyright (C) 2009-2010 The University of Texas at Austin ("U.T. Austin"). All rights reserved.
	
	U.T. Austin's copyright notice in this evolving version reflects our original authorship. It does not assert 
	copyright over changes contributed by others.
	
	This library is free software; you can redistribute it and/or modify it under the terms of the GNU Lesser General 
	Public License as published by the Free Software Foundation; either version 2.1 of the License, or (at your option) 
	any later version.
	
	This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied 
	warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for 
	more details.
	
	You should have received a copy of the GNU Lesser General Public License along with this library; if not, write to 
	the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/


package edu.utexas.its.eis.tools.qwicap.servlet;


import java.util.*;
import java.util.logging.Logger;
import java.util.logging.Level;
import java.net.InetAddress;
import java.io.File;
import java.math.BigDecimal;


/**	Generates entropy data for the seeding of a pseudo random number generator (PRNG). A test of this class that 
 * 	generated 200,000 256-bit entropy values showed that the entropy values had a worst case divergence from an equal 
 * 	number of 1 and 0 bits of 5.86%, a mean divergence of 0.497%, and that each bit is almost equally likely to be a 
 * 	1 as it is a 0, so no bits have pedictable states.
 * 
 * @author Chris W. Johnson
 * @version 1.1.2
 */
final class Entropy {
	
	private static final Logger	Log = Logger.getLogger(Entropy.class.getName());


/**	Generates entropy data for the seeding of a pseudo random number generator (PRNG).
 * 
 * @param InitialStateBits		An initial seed value represented as an array of <code>byte</code> primitives, where
 * 								the value of each <code>byte</code> represents a bit, and should, therefore, be either
 * 								a zero (0) or a one (1). This seed value is not required; supplying an array of 
 * 								<code>byte</code>s that are all set to zero (0) is perfectly acceptable. Note that the
 * 								length of this array determines the number of bits of entropy produced by this method.
 * @return						The <code>InitialState</code> array is returned, but with its contents altered to 
 * 								represent the requested entropy. Each <code>byte</code> of this array represents a 
 * 								single bit, and will, therefore, contain either a zero (0) or a one (1).
 */
	static byte[] getBits
	(
		final byte[]			InitialStateBits
	)
	{
		final ThreadAccumulator			TIA = new ThreadAccumulator();
		final EntropyBuilder			EnB = new EntropyBuilder(InitialStateBits);
		final RuntimeInfoAccumulator	RIA = new RuntimeInfoAccumulator();
		final HashCodeAccumulator		HCA = new HashCodeAccumulator();
		
		HCA.add(InitialStateBits, EnB, HCA, RIA);
		
		RIA.add();

		try {
			final Properties			Props = System.getProperties();
			final Enumeration<String>	Names = (Enumeration<String>) Props.propertyNames();
			
			HCA.add(Props, Names);
			
			while (Names.hasMoreElements()) {
				final String			Name  = Names.nextElement();
				final String			Value = Props.getProperty(Name);
				
				EnB.addStrings("System property name/value", Name, Value);
				HCA.add(Name, Value);
				RIA.add();
			}

		} catch (Exception e) {
			HCA.addException(e);
		}
		
		RIA.add();
		
		try {
			
			for (final InetAddress Addr : InetAddress.getAllByName(InetAddress.getLocalHost().getHostAddress())) {
				final byte[]			AddrBytes = Addr.getAddress();
				final String			AddrIP    = Addr.getHostAddress();
				final String			AddrName  = Addr.getHostName();
				
				EnB.addBytes("IP address bytes", AddrBytes);
				EnB.addStrings("IP address and DNS names string", AddrIP, AddrName);
				HCA.add(Addr, AddrBytes, AddrIP, AddrName);
				RIA.add();
			}
			
		} catch (Exception e) {
			HCA.addException(e);
		}
		
		RIA.add();
		
		try {
			final File				TempDir   = new File(System.getProperty("java.io.tmpdir"));
			final File[]			TempFiles = TempDir.listFiles();
			final long[]			Sizes     = new long[TempFiles.length];
			final long[]			ModDates  = new long[TempFiles.length];
			
			HCA.add(TempDir, TempFiles, Sizes, ModDates);
			
			for (int Index = 0, Count = TempFiles.length; Index < Count; Index++) {
				final File				TempFile     = TempFiles[Index];
				final String			TempFileName = TempFile.getName();
				
				Sizes[Index]    = TempFile.length();
				ModDates[Index] = TempFile.lastModified();
				
				EnB.addString("Temp file name", TempFileName);
				HCA.add(TempFile, TempFileName);
				RIA.add();
			}
			
			EnB.addLongs("Temp file sizes",    Sizes   );
			EnB.addLongs("Modification dates", ModDates);
			
		} catch (Exception e) {
			HCA.addException(e);
		}
		
		RIA.add();
			
		HCA.addException(new Throwable(new Date().toString()));
		
	//	Accumulate data from our thread timings for at least the 1/2 second used below. In practice, this provides much 
	//	more entropy data, and much more useful entropy data, than all of the code above. The code above is retained
	//	so that we won't be using just one means of acquiring potential entropy data. To the extent that the code above
	//	doesn't produce much entropy data, or may produce data that is predictable or controllable to vary degrees,
	//	those potential shortcomings are overwhelmed by the data collected below.
		
		for (int Count = 250; --Count >= 0; ) {
			
			RIA.add();
			
			try {
				Thread.sleep(2);	
			} catch (Exception e) {
			//	Ignore.
			}
		}
		
		HCA.addToStateBuilder(EnB);
		RIA.addToStateBuilder(EnB);
		
		TIA.pleaseDie();
		TIA.addToStateBuilder(EnB);
		
	//	Done. Report any extreme deviations from randomness.
		
		final byte[]			EntropyBits = EnB.getEntropyBits();	//	Each byte represents a single bit, and therefore has a value of either 0 or 1.
		final double			DeviationFromEqualOnesAndZeros = Math.abs(0.5 - EnB.getOneBitCount() / (double) EnB.getTotalBitCount());
		
		Log.log(Level.FINE, "The entropy generator produced {3} bits of entropy, which deviated from an equal number of ones and zeros by {0,number,0.##%}. (The number of bits provided to the entropy generator was {1,number,integer}, of which {2,number,integer} were deemed suitable for use in entropy generation.)", new Object[] { DeviationFromEqualOnesAndZeros, EnB.getInputBitRawCount(), EnB.getInputBitUsedCount(), InitialStateBits.length });
		
		return EntropyBits;
	}


/**	Accumulates data from various runtime variables. The accumulated data is fed into an {@link EntropyBuilder} 
 * 	instance by invoking the {@link #addToStateBuilder} method.
 * 	
 * @author Chris W. Johnson
 * @version 1.0
 */	
	private static final class RuntimeInfoAccumulator {
		
		private final Runtime			R         = Runtime.getRuntime();
		private final ArrayList<Long>	Times     = new ArrayList<Long>();
		private final ArrayList<Long>	Nanos     = new ArrayList<Long>();
		private final ArrayList<Long>	FreeMems  = new ArrayList<Long>();
		private final ArrayList<Long>	TotalMems = new ArrayList<Long>();
		
		private long					PrevTime;
		private long					PrevNanos;
		private long					PrevFreeMem;
		private long					PrevTotalMem;
		
		
		RuntimeInfoAccumulator() {
			
			PrevTime     = System.currentTimeMillis();
			PrevNanos    = System.nanoTime();
			PrevFreeMem  = R.freeMemory();
			PrevTotalMem = R.totalMemory();
			
			add();	
		}
		
		Runtime getRuntime() {
			return R;	
		}
		
		void add() {
			final long			CurTime     = System.currentTimeMillis();
			final long			CurNanos    = System.nanoTime();
			final long			CurFreeMem  = R.freeMemory();
			final long			CurTotalMem = R.totalMemory();
			
			if (CurTime != PrevTime) {
				
				Times.add(CurTime);
				PrevTime = CurTime;
			}
			
			if (CurNanos != PrevNanos) {
				
				Nanos.add(CurNanos);
				PrevNanos = CurNanos;
			}
			
			if (CurFreeMem != PrevFreeMem) {
				
				FreeMems.add(CurFreeMem);
				PrevFreeMem = CurFreeMem;
			}
			
			if (CurTotalMem != PrevTotalMem) {
				
				TotalMems.add(CurTotalMem);
				PrevTotalMem = CurTotalMem;
			}
		}
		
		void addToStateBuilder
		(
			final EntropyBuilder	ISB
		)
		{
			add();
			
			ISB.addNumbers("System.currentTimeMillis()", (List) Times    );
			ISB.addNumbers("System.nanoTime()",          (List) Nanos    );
			ISB.addNumbers("Free RAM",                   (List) FreeMems );
			ISB.addNumbers("Total RAM",                  (List) TotalMems);
		}
	}


/**	Accumulates data from the hash codes of arbitrary objects. The accumulated data is fed into an 
 * 	{@link EntropyBuilder} instance by invoking the {@link #addToStateBuilder} method.
 * 	
 * @author Chris W. Johnson
 * @version 1.0
 */	
	private static final class HashCodeAccumulator {
		
		private final ArrayList<Integer>	IdentityHashCodes = new ArrayList<Integer>();
		private final ArrayList<Integer>	NormalHashCodes   = new ArrayList<Integer>();
		private final ArrayList<Integer>	StackTraceLineNos = new ArrayList<Integer>();
		
		
		void add
		(
			final Object...			Objs
		)
		{
			for (final Object Obj : Objs)
				add(Obj);
		}
		
		void add
		(
			final Object			Obj
		)
		{
			if (Obj == null)
				return;
			
			final int				IHC = System.identityHashCode(Obj);
			final int				NHC = Obj.hashCode();
			
			IdentityHashCodes.add(IHC);
			
			if (NHC != IHC)
				NormalHashCodes.add(NHC);
		}
		
		void addException
		(
			final Throwable			T
		)
		{
			add(
				T,
				T.getMessage(),
				T.toString()
			);
			
			for (final StackTraceElement STE : T.getStackTrace()) {
				
				add(
					STE,
					STE.toString()
				);
				
				StackTraceLineNos.add(STE.getLineNumber());
			}
			
			final Throwable			Cause = T.getCause();
			
			if (Cause != null)
				addException(Cause);
		}
		
		void addToStateBuilder
		(
			final EntropyBuilder	ISB
		)
		{
			ISB.addNumbers("Identity hash code",      (List) IdentityHashCodes);
			ISB.addNumbers("Normal hash code",        (List) NormalHashCodes  );
			ISB.addNumbers("Stack trace line number", (List) StackTraceLineNos);
		}
	}
	
	
	private static final class ThreadAccumulator {
		
		private final int					ThreadCount;
		private final ThreadGroup			Group         = new ThreadGroup("Entropy");
		private final Object				ContentionObj = new Object();
		
		private final ArrayList<Long>		WaitTimes = new ArrayList<Long>();
		private final ArrayList<Short>		ThreadNos = new ArrayList<Short>();
		private final ArrayList<Long>		Nanos     = new ArrayList<Long>();
		
		private boolean						PleaseDie;
		
		
		ThreadAccumulator() {
			final int				Processors = Runtime.getRuntime().availableProcessors();
			
			ThreadCount = Processors * 10;
			
			for (int ThreadNo = 0; ThreadNo < ThreadCount; ThreadNo++)
				new Thread(Group, new EntropyRunnable(ThreadNo, this), "Entropy Thread " + ThreadNo, 64 * 1024).start();
		}
		
		boolean getPleaseDie() {
			return PleaseDie;	
		}
		
		void pleaseDie() {
			final Thread[]			Threads = new Thread[ThreadCount];
			
			Group.enumerate(Threads);
			PleaseDie = true;
			
			boolean					ThreadsRemain;
			
			do {
				
				ThreadsRemain = false;
				
				for (final Thread T : Threads) {
					
					if (T != null && T.isAlive()) {
							
						T.interrupt();
						ThreadsRemain = true;
					}
				}
				
			} while (ThreadsRemain);
			
			Group.destroy();
		}
		
		Object getContentionObject() {
			return ContentionObj;
		}
		
		void addNanos
		(
			final long				Ns
		)
		{
			final long				StartNanos = System.nanoTime();
			final long				EndNanos;
			
			synchronized (Nanos) {
				
				EndNanos = System.nanoTime();
				
				Nanos.add(Ns);
			}
			
			addWaitTime(EndNanos - StartNanos);
		}
		
		void addWaitTime
		(
			final long				Wait
		)
		{
			final long				StartNanos = System.nanoTime();
			
			synchronized (WaitTimes) {
				final long				EndNanos = System.nanoTime();
				
				WaitTimes.add(Wait);
				WaitTimes.add(EndNanos - StartNanos);
			}
		}
		
		void addThreadNo
		(
			final int				ThreadNo
		)
		{
			final long				StartNanos = System.nanoTime();
			final long				EndNanos;
			
			synchronized (ThreadNos) {
				
				EndNanos = System.nanoTime();
				
				ThreadNos.add((short) ThreadNo);
			}
			
			addWaitTime(EndNanos - StartNanos);
		}
		
		void addToStateBuilder
		(
			final EntropyBuilder	ISB
		)
		{
			ISB.addNumbers("Synchronization wait time", (List) WaitTimes);
			ISB.addNumbers("Thread number",             (List) ThreadNos);
			ISB.addNumbers("Nanosecond values",         (List) Nanos    );
		}
	}
	
	
	private static final class EntropyRunnable implements Runnable {
		
		private final int				ThreadNo;
		private final ThreadAccumulator	Accumulator;
		private final Object			ContentionObj;
		
		
		EntropyRunnable
		(
			final int				ThreadNo,
			final ThreadAccumulator	Parent
		)
		{
			this.ThreadNo      = ThreadNo;
			this.Accumulator   = Parent;
			this.ContentionObj = Parent.getContentionObject();
		}
		
		@Override public void run() {
			
			try {
				
				while (!Accumulator.getPleaseDie()) {
					final long				StartNanos1 = System.nanoTime();
					long					StopNanos1  = -1;
					long					StopNanos2  = -1;
					
					try {
						
						synchronized (ContentionObj) {
							
							StopNanos1 = System.nanoTime();
							
							ContentionObj.wait(0, (int) StopNanos1 & 0x7F);	//	Note: In Java 6 on Mac OS X 10.6.1, the low three bits System.nanoTime() are always zero.
							ContentionObj.notifyAll();
							
							StopNanos2 = System.nanoTime();
						}
						
						Thread.sleep(0, (int) System.nanoTime() & 0x3F);	//	Note: In Java 6 on Mac OS X 10.6.1, the low three bits System.nanoTime() are always zero.
						
					} catch (Exception e) {

					//	Ignore.
						
					} finally {
						
						if (StopNanos1 != -1) {
							
							Accumulator.addWaitTime(StopNanos1 - StartNanos1);
							
							if (StopNanos2 != -1)
								Accumulator.addWaitTime(StopNanos2 - StopNanos1);	
						}
						
						Accumulator.addThreadNo(ThreadNo);
						Accumulator.addNanos(System.nanoTime());
					}
				}
				
			} catch (Exception e) {
				
				if (!Accumulator.getPleaseDie())
					Log.log(Level.INFO, "Entropy thread died unexpectedly.", e);
			}
		}
	}
	
	
	
/**	Processes arbitrary input data to produce entropy data suitable for seeding a pseudo random number generator (PRNG).
 * 	
 * @author Chris W. Johnson
 * @version 1.0.2
 */	
	private static final class EntropyBuilder {

		private final int[]			OneBitCounts;		//	Each int represents the number of times a 1 bit was encountered in at its offset.
		private final int			BitCount;			//	The number of bits in the InitialState passsed to the constructor, and the length of the OneBitCounts array.
		private int					BitIndex;			//	The index of the next element in the OneBitCounts array that will be used. This value may be equal to the length of the array, at which point it should be reset to zero.
		private int					InputBitRawCount;	//	The total number of bits provided to this object.
		private int					InputBitUsedCount;	//	The number of bits actually judged as usable by this object.
		private int					FinalOneBitCount = -1;
		

		EntropyBuilder
		(
			final byte[]			InitialState
		)
		{
			this.BitCount     = InitialState.length;
			this.OneBitCounts = new int[BitCount];
			this.BitIndex     = 0;
			
			for (final byte B : InitialState)
				addBit((int) B);
		}
		
		byte[] getEntropyBits() {
			int						MinOneBits = Integer.MAX_VALUE;
			int						MaxOneBits = Integer.MIN_VALUE;
			
			for (final int OneBitCount : OneBitCounts) {
				
				if (MinOneBits > OneBitCount)
					MinOneBits = OneBitCount;
				if (MaxOneBits < OneBitCount)
					MaxOneBits = OneBitCount;
			}
			
			final Histogram			Hist = new Histogram(MinOneBits, MaxOneBits, BitCount);
			
			for (final int OneBitCount : OneBitCounts)
				Hist.addSample(OneBitCount);
			
			final int				MedianOneBitCount = Hist.getMedianValue();			
			final byte[]			EntropyBits       = new byte[BitCount];
			
			FinalOneBitCount = 0;
			
			for (int Index = 0; Index < BitCount; Index++)   {
				
				if (OneBitCounts[Index] > MedianOneBitCount) {
					
					EntropyBits[Index] = (byte) 1;
					FinalOneBitCount++;
			
				} else
					EntropyBits[Index] = (byte) 0;
			}
			
			return EntropyBits;
		}
		
		int getInputBitUsedCount() {
			return InputBitUsedCount;
		}
		
		int getInputBitRawCount() {
			return InputBitRawCount;	
		}
		
		int getOneBitCount() {
			return FinalOneBitCount;
		}
		
		int getTotalBitCount() {
			return BitCount;
		}

		void addNumbers
		(
			final String			Description,
			final List<Number>		Numbers
		)
		{
			final int				Size = Numbers.size();
			
			if (Size == 0)
				return;
			
			final Class<?>			ValueClass = Numbers.get(0).getClass();
			final long[]			Longs      = new long[Size];
			int						Index      = 0;
			
			if (isFloatingPoint(ValueClass)) {
				
				for (final Number No : Numbers)
					Longs[Index++] = Double.doubleToRawLongBits(No.doubleValue());
				
			} else {
				final long				Mask;
				
				if (ValueClass == Byte.class)
					Mask = 0xFF;
				else if (ValueClass == Short.class)
					Mask = 0xFFFF;
				else if (ValueClass == Integer.class)
					Mask = 0xFFFFFFFFL;
				else if (ValueClass == Long.class)
					Mask = -1L;
				else
					throw new IllegalStateException("Numeric class \"" + ValueClass.getName() + "\" not supported.");
				
				for (final Number No : Numbers)
					Longs[Index++] = Mask & No.longValue();
			}
			
			addLongs(Description, Longs);
		}
	
		private static boolean isFloatingPoint
		(
			final Class<?>			ValueClass
		)
		{
			return ValueClass == Double.class || ValueClass == Float.class || ValueClass == BigDecimal.class;
		}
		
		void addLongs
		(
			final String			Description,	//	Not used in production, but handy for debugging and/or analysis.
			final long...			Longs
		)
		{
			if (Longs.length == 0)
				return;
			
			long					SometimesOneBitsMask = 0;
			long					AlwaysOneBitsMask    = -1L;
			
			for (final long L : Longs) {
				
				SometimesOneBitsMask |= L;
				AlwaysOneBitsMask    &= L;
			}
			
			final long				AlwaysZeroBitsMask = ~SometimesOneBitsMask;
			final long				UselessBitsMask    = AlwaysZeroBitsMask | AlwaysOneBitsMask;
			
			for (final long Long : Longs) {
				
				for (long BitMask = 0x1; BitMask != 0; BitMask <<= 1) {
					
					if ((BitMask & UselessBitsMask) == 0)
						addBit((Long & BitMask) != 0 ? 1 : 0);
				}
			}
			
			InputBitRawCount += 64 * Longs.length;
		}
		
		void addStrings
		(
			final String			Description,
			final String...			Strs
		)
		{
			final int				Count   = Strs.length;
			final long[]			Lengths = new long[Count];
			
			for (int Index = 0; Index < Count; Index++) {
				final String			Str = Strs[Index];
				
				addString(Description, Str);
				Lengths[Index] = Str.length();
			}
			
			addLongs(Description, Lengths);
		}
		
		void addString
		(
			final String			Description,
			final String			Str
		)
		{
			final char[]			Chars = Str.toCharArray();
			final int				Count = Chars.length;
			final long[]			Longs = new long[Count];
			
			for (int Index = 0; Index < Count; Index++)
				Longs[Index] = Chars[Index];
			
			addLongs(Description, Longs);
		}
		
		void addBytes
		(
			final String			Description,
			final byte...			Bytes
		)
		{
			final int				Count = Bytes.length;
			final long[]			Longs = new long[Count];
			
			for (int Index = 0; Index < Count; Index++)
				Longs[Index] = Bytes[Index] & 0xFF;
			
			addLongs(Description, Longs);
		}
		
		void addBooleans
		(
			final Collection<Boolean>	Booleans
		)
		{
			for (final Boolean B : Booleans)
				addBit(B == Boolean.TRUE ? 1 : 0);
		}
	
		private void addBit
		(
			final int				Bit		//	Must be only zero or 1.
		)
		{
			if (BitIndex >= BitCount)
				BitIndex = 0;
			
			OneBitCounts[BitIndex++] += Bit;
			
			InputBitUsedCount++;
		}
	}
	
	private static class Histogram {
		
		private final int			MinValue;
		private final int[]			Counters;
		private final boolean		OneToOne;
		private final double		ManyToOneMultiplier;
		private int					TotalSamples;
		
		
		Histogram
		(
			final int				MinValue,
			final int				MaxValue,
			final int				MaxResolution
		)
		{
			final int				DiscreteValueCount = (MaxValue - MinValue) + 1;
			
			if (DiscreteValueCount <= MaxResolution) {
				
				Counters            = new int[DiscreteValueCount];
				OneToOne            = true;
				ManyToOneMultiplier = 1.0;
				
			} else {
				
				Counters            = new int[MaxResolution + 1];
				OneToOne            = false;
				ManyToOneMultiplier = MaxResolution / (double) DiscreteValueCount;
			}
			
			this.MinValue = MinValue;
		}
		
		void addSample
		(
			final int				NewValue
		)
		{
			final int				RelativeValue = NewValue - MinValue;
			
			if (OneToOne)
				Counters[RelativeValue]++;
			else	
				Counters[(int) Math.round(ManyToOneMultiplier * RelativeValue)]++;
			
			TotalSamples++;
		}
		
		int getMedianValue() {
			final int				HalfOfTotalSamples = (int) Math.round(TotalSamples / 2.0);
			int						SampleCount        = 0;
			int						MedianIndex        = -1;
			
			for (int Index = 0, Count = Counters.length; Index < Count; Index++) {
				
				SampleCount += Counters[Index];
				
				if (SampleCount >= HalfOfTotalSamples) {
				
					if (SampleCount > HalfOfTotalSamples) {
						final int				PrevSampleCount = SampleCount - Counters[Index];
						
						if (HalfOfTotalSamples - PrevSampleCount <= SampleCount - HalfOfTotalSamples)
							MedianIndex = Index - 1;
						else
							MedianIndex = Index;
					} else
						MedianIndex = Index;
					
					break;
				}
			}
			
			if (MedianIndex < 0)
				throw new IllegalStateException();
			
			if (OneToOne)
				return MinValue + MedianIndex;
			
			return MinValue + (int) Math.round(MedianIndex / ManyToOneMultiplier);
		}
	}
}
