/*	QWICAP (Quick Web Interface for Conventional Applications) by Chris W. Johnson
	Copyright (C) 2003-2011 The University of Texas at Austin ("U.T. Austin"). All rights reserved.
	
	U.T. Austin's copyright notice in this evolving version reflects our original authorship. It does not assert 
	copyright over changes contributed by others.
	
	This library is free software; you can redistribute it and/or modify it under the terms of the GNU Lesser General 
	Public License as published by the Free Software Foundation; either version 2.1 of the License, or (at your option) 
	any later version.
	
	This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied 
	warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for 
	more details.
	
	You should have received a copy of the GNU Lesser General Public License along with this library; if not, write to 
	the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/


package edu.utexas.its.eis.tools.qwicap.servlet;


import edu.utexas.its.eis.tools.qwicap.template.css.CSSPatterns;
import edu.utexas.its.eis.tools.qwicap.template.css.Results;
import edu.utexas.its.eis.tools.qwicap.template.xml.Markup;
import edu.utexas.its.eis.tools.qwicap.template.xml.Match;
import edu.utexas.its.eis.tools.qwicap.template.xml.TagException;
import edu.utexas.its.eis.tools.qwicap.template.xml.immutable.ImmutableMarkup;
import edu.utexas.its.eis.tools.qwicap.template.xml.immutable.XMLString;
import edu.utexas.its.eis.tools.qwicap.template.xml.mutable.AttributeUtilities;
import edu.utexas.its.eis.tools.qwicap.template.xml.mutable.MutableMarkup;
import edu.utexas.its.eis.tools.qwicap.util.ArrayToString;
import edu.utexas.its.eis.tools.qwicap.util.Null;
import edu.utexas.its.eis.tools.table.MutableTable;
import edu.utexas.its.eis.tools.table.Table;
import edu.utexas.its.eis.tools.table.lltable.MutableLLTable;

import java.lang.reflect.Array;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Comparator;
import java.util.logging.Logger;
import java.util.logging.Level;
import java.util.regex.Pattern;
import java.net.URL;
import java.net.MalformedURLException;
import java.nio.charset.Charset;


/**	The <code>Form</code> class provides a set of high-level utility methods for operating on the controls within an 
 *	XHTML <a href='http://www.w3.org/TR/html4/interact/forms.html'>"<samp>form</samp>" element</a>.
 *	
 *	<p>If it looks like this class is implemented a lot like a large mass of procedural code, rather than as an
 *	elegant hierarchy of classes representing the various types of form controls, you're right. And it's not by 
 *	accident. The elegant hierarchy of classes would be conceptually cleaner, probably easier to support, and more fun 
 *	to write, but it would also introduce assumptions into the code that the <code>Form</code> class does not, 
 *	because building a set of control objects to represent the contents of a "<samp>form</samp>" element creates a
 *	representation of the state of the form that is independent of the underlying markup. Which means that the 
 *	representation can become out-of-synch with the markup, if, for instance, the programmer prefers to do some
 *	manipulation of form contents by directly using the templating engine, e.g. 
 *	<code>MyMarkup.get("select[name='foo'] option[name='bar']").delete()</code>, or if the representation is retained
 *	across <code>prompt</code> invocations and Qwicap updates the form markup by embedding user-supplied values, etc.
 *	</p>	
 *
 *	<p>In contrast, the <code>Form</code> class maintains no state information about the form (other than the assumption 
 *	that the form still exists), so it can't become out-of-synch with the markup, programmers don't receive rude 
 *	surprises (so, no need for extra documentation, notes and warnings), and they remain free to choose the method of 
 *	manipulating the form markup that works best for them in any given situation (doing it directly, going through the 
 *	<code>Form</code> class, or both). Put simply, the advantage of the <code>Form</code> class as it is presently 
 *	constituted is that it is too dumb to die.
 *	</p>	
 *
 *	<p>Finding, Examining, Transforming, and Deleting Controls:</p>
 *	
 *	<ul>
 *		<li>{@link #deleteControl(Match, Object)} [static]</li>
 *		<li>{@link #deleteControl(String, Object)}</li>
 *		<li>{@link #getControl(String, Object)}</li>
 * 		<li>{@link #getControlByIndex(String, int)}</li>
 * 		<li>{@link #getControlByIndexAndValue(String, int, Object)}</li>
 *		<li>{@link #getControls(String, Object)}</li>
 *		<li>{@link #getControlsByName(String)}</li>
 *		<li>{@link #getControlName(Match)} [static]</li>
 * 		<li>{@link #getSubmitControls}</li>
 *		<li>{@link #convertSubmitButtonsToInputs}</li>
 *		<li>{@link #convertSubmitInputsToButtons}</li>
 *	</ul>
 *
 *	<p>Control Value Related Methods:</p>
 *
 *	<ul>
 *		<li>{@link #addToControlValue(Match, Object)} [static]</li>
 *		<li>{@link #addToControlValue(String, Object)}</li>
 *		<li>{@link #clearControlValue(Match)} [static]</li>
 *		<li>{@link #clearControlValue(String)}</li>
 *		<li>{@link #clearControlValues}</li>
 *		<li>{@link #getControlValue(Match)} [static]</li>
 *		<li>{@link #getControlValue(String)}</li>
 *		<li>{@link #getControlValues(Match)} [static]</li>
 *		<li>{@link #getControlValues(Match, boolean, boolean, boolean)} [static]</li>
 *		<li>{@link #getControlValues(String)}</li>
 *		<li>{@link #setControlValue(Match, Object)} [static]</li>
 *		<li>{@link #setControlValue(String, Object)}</li>
 *		<li>{@link #setControlValues(Object...)}</li>
 *	</ul>
 *
 *	<p>Control State Related Methods:</p>
 *	
 *	<ul>
 *		<li>{@link #enableControl(Match)} [static]</li>
 *		<li>{@link #enableControl(String)}</li>
 *		<li>{@link #enableControl(Match, Object)} [static]</li>
 *		<li>{@link #enableControl(String, Object)}</li>
 *		<li>{@link #enableControls(String...)}</li>
 *		<li>{@link #enableAllControls}</li>
 *		<li>{@link #disableControl(Match)} [static]</li>
 *		<li>{@link #disableControl(String)}</li>
 *		<li>{@link #disableControl(Match, Object)} [static]</li>
 *		<li>{@link #disableControl(String, Object)}</li>
 *		<li>{@link #disableControls(String...)}</li>
 *		<li>{@link #disableAllControls}</li>
 *		<li>{@link #isControlEnabled(Match)} [static]</li>
 *		<li>{@link #setControlEnable(Match, boolean)} [static]</li>
 *		<li>{@link #setControlEnable(String, boolean)}</li>
 *	</ul>
 *	
 *	<p>Control Group (Checkbox, Radio, Select) Related Methods:</p>
 *
 *	<ul>
 *		<li>{@link #addChoice(boolean, String, Object, boolean, String)}</li>
 *		<li>{@link #addFirstChoice(String, Object, boolean, String)}</li>
 *		<li>{@link #addLastChoice(String, Object, boolean, String)}</li>
 *		<li>{@link #addToControlValue(Match, Object)} [static]</li>
 *		<li>{@link #addToControlValue(String, Object)}</li>
 *		<li>{@link #createChoice(String, String, Object, boolean, String)} [static]</li>
 *		<li>{@link #deleteChoices(String)}</li>
 *	</ul>
 *	
 *	<p>Form Data Set Methods:</p>
 *	
 *	<ul>
 *		<li>{@link #getAction}</li>
 *		<li>{@link #getActionURL}</li>
 *		<li>{@link #getCharacterSet}</li>
 *		<li>{@link #setCharacterSet}</li>
 *		<li>{@link #getCharacterSets}</li>
 *		<li>{@link #getDataSet}</li>
 *		<li>{@link #getDataSet(boolean)}</li>
 *		<li>{@link #setDataSet(FormDataSet)}</li>
 *		<li>{@link #getMethod}</li>
 *	</ul>
 *
 *	<p>Form Utility Methods:</p>
 *	
 *	<ul>
 *		<li>{@link #convertSubmitButtonsToInputs}</li>
 * 		<li>{@link #convertSubmitInputsToButtons}</li>
 *		<li>{@link #get(String)}</li>
 *		<li>{@link #get(CSSPatterns)}</li>
 *		<li>{@link #getMatch}</li>
 *		<li>{@link #toResults}</li>
 *	</ul>
 *
 *	<p>Debugging Methods:</p>
 *	
 *	<ul>
 *		<li>{@link #getControlsTable}</li>
 *	</ul>
 *
 *	@author Chris W. Johnson
 */
public final class Form {
	
	private static final Logger		Log = Logger.getLogger(Form.class.getName());
	
	private static final Pattern	kSpaceOrCommaPat  = Pattern.compile("[,\\s]+");
	private static final String[]	kEmptyStringArray = {};	//	Used by getControlValues(Match, boolean, boolean)
	
	private final Match				TheForm;
	
	
/**	Creates a <code>Form</code> object for manipulating the form that produced the supplied <code>FormDataSet</code>.
 * 
 * @param ThePage					The page of markup in which the form is to be found.
 * @param DataSet					The <code>FormDataSet</code> that was produced by one of the forms on this page.
 * @throws FormNotFoundException	If the markup of this page does not contain a form that could have produced the
 * 									supplied <code>FormDataSet</code>.
 */
	public Form
	(
		final MutableMarkup		ThePage,
		final FormDataSet		DataSet
	)
	throws FormNotFoundException
	{
		this(findRelevantForm(ThePage, DataSet, false));
	}
	
/**	Creates a <code>Form</code> object for manipulating the specified form.
 * 
 * @param ThePage					The page of markup in which the form is to be found.
 * @param FormIDStr					The Qwicap-assigned unique identifier for the relevant form.
 * @throws FormNotFoundException	If the markup of this page does not contain the form identified by the 
 *									<code>FormIDStr</code> parameter.
 */
	Form
	(
		final MutableMarkup		ThePage,
		final String			FormIDStr
	)
	throws FormNotFoundException
	{
		this(findFormByID(ThePage, FormIDStr));
	}
	
/**	Creates a <code>Form</code> object for operating on the specified form.
 * 
 * @param TheForm					The relevant "form" tag within that markup.
 * @throws FormNotFoundException	If the <code>TheForm</code> does not refer to a "<samp>form</samp>" element.
 */
	public Form
	(
		final Match				TheForm
	)
	throws FormNotFoundException
	{
		if (!"form".equals(TheForm.getType()))
			throw new FormNotFoundException("\"Form\" objects can only be instantiated to represent \"form\" elements. This \"Form\" object is being illegally asked to refer to: " + TheForm);
		
		this.TheForm = TheForm;
	}
	
	@Override public String toString() {
		return TheForm.toString(false);
	}
	
/**	Returns the <code>Match</code> object that uniquely identifies the "<samp>form</samp>" element represented by this
 *	<code>Form</code> object.
 * 
 * @return						The <code>Match</code> object that uniquely identifies the "<samp>form</samp>" element 
 *								represented by this <code>Form</code> object.
 */	
	public Match getMatch() {
		return TheForm;
	}

/**	Returns the value of the <a href='http://www.w3.org/TR/html4/interact/forms.html#h-17.3'>form's "action" 
 * 	attribute</a>, or <code>null</code> if there is no such attribute.
 * 
 * @return						The value of the form's "action" attribute, or <code>null</code> if there is no such
 * 								attribute.
 */
	public String getAction() {
		return TheForm.getAttribute("action");
	}

/**	Attempts to construct a fully-qualified action URL from the value of the 
 * 	<a href='http://www.w3.org/TR/html4/interact/forms.html#h-17.3'>form's "action" attribute</a>, the URL of the page
 * 	in which this form occurs, and any <a href='http://www.w3.org/TR/html4/struct/links.html#h-12.4'>"base" element</a>
 *  that the page may contain.
 * 
 * @return							This "form" element's action URL.
 * @throws MalformedURLException	If a potentially valid action URL could not be constructed.
 */
	public URL getActionURL() throws MalformedURLException {
		final String			Action  = Null.toEmptyString(TheForm.getAttribute("action"));
		final String			BaseURL = TheForm.getMarkup().get("head base").getAttribute("href");
		final String			PageURL = TheForm.getMarkup().getMarkupName();
		
		if (BaseURL != null) {
			
			return new URL(new URL(BaseURL), Action);
				
		} else if (PageURL != null) {		
			
			return new URL(new URL(PageURL), Action);
			
		} else {
			
		//	In this case, we have to hope that Action is a fully qualified URL, or this won't work. (But we have no 
		//	other option, so it's worth a try.)
			
			return new URL(Action);
		}
	}

/**	Returns the value of the <a href='http://www.w3.org/TR/html4/interact/forms.html#h-17.3'>form's "method" 
 * 	attribute</a> ("get" or "post"), or <code>null</code> if there is no such attribute.
 * 
 * @return						The value of the form's "method" attribute, or <code>null</code> if there is no such
 * 								attribute.
 */	
	public String getMethod() {
		return TheForm.getAttribute("method");
	}
	
/**	Sets the character set in which the data set created by this form will be encoded.
 * 
 * @param CharacterSet			The official <a href='http://www.iana.org/assignments/character-sets'>IANA name of the
 * 								character set</a> in which the form's data set will be encoded.
 * @return						This <code>Form</code> object.
 * @throws TagException			If a problem is encountered with the manipulation of the markup.
 */	
	public Form setCharacterSet
	(
		final String			CharacterSet
	)
	throws TagException
	{
		TheForm.setAttribute("accept-charset", CharacterSet);
		return this;
	}

/**	Returns the first character set name from the list of character set names in the "form" element's "accept-charset"
 * 	attribute. If that attribute is missing or empty, the character set of the enclosing document is returned.
 * 
 * @return						The official <a href='http://www.iana.org/assignments/character-sets'>IANA name of the
 * 								character set</a> in which the form's data set will be encoded.
 * @see #getCharacterSets
 */
	public String getCharacterSet() {
		final String[]			CharacterSets = getCharacterSets();
		
		return CharacterSets[0];
	}
	
/**	Returns the name of the character sets in which this form's data set may be encoded, based on the list in the
 * 	"form" element's "accept-charset" attribute. If that attribute is missing or empty, the character set of the 
 * 	enclosing document is returned.
 * 
 * @return						The official <a href='http://www.iana.org/assignments/character-sets'>IANA name of the
 * 								character set</a> in which the form's data set will be encoded.
 */
	public String[] getCharacterSets() {
		final String			CharacterSetList = TheForm.getAttribute("accept-charset");	//	Space- or comma-delimited list.
		
		if (CharacterSetList != null) {
			final String[]			CharacterSets = kSpaceOrCommaPat.split(CharacterSetList);
			
			if (CharacterSets.length > 0 && CharacterSets[0].length() > 0)
				return CharacterSets;
		}
		
	//	There was no list of character sets available from the "accept-charset" attribute, so we default to the 
	//	document's character set.
	
		return new String[] { TheForm.getMarkup().getCharacterSet() };
	}
	
/**	Searches this "<samp>form</samp>" element, and its children, for matches with the specified CSS2 pattern(s). If 
 *	there are no matches, the <code>Results</code> object returned is empty.
 * 
 * @param CSSPatterns	One or more CSS2 patterns to be applied to this element and its children.
 * @return				A <code>Results</code> object containing a list of elements that matched the specified 
 *						pattern(s), if any.
 */
	public Results get
	(
		final String			CSSPatterns
	)
	{
		return TheForm.get(CSSPatterns);
	}
	
/**	Searches this "<samp>form</samp>" element, and its children, for matches with the specified CSS2 pattern(s). If 
 *	there are no matches, the <code>Results</code> object returned is empty.
 *
 * @param CSSPatterns	One or more CSS2 patterns to be applied to this element and its children.
 * @return				A <code>Results</code> object containing a list of elements that matched the specified 
 *						pattern(s), if any.
 */
	public Results get
	(
		final CSSPatterns		CSSPatterns
	)
	{
		return TheForm.get(CSSPatterns, null);
	}
	
/**	Creates <code>Results</code> object containing a reference to this "<samp>form</samp>" element.
 * 
 * @return				A <code>Results</code> object containing a reference to this "<samp>form</samp>" element.
 */	
	public Results toResults() {
		return TheForm.toResults();
	}
	
/**	Deletes all choices associated with the control of the specified name. Note that for "input" controls of type 
 *	"radio" and "checkbox" this call is not reversible, i.e. you can't invoke <code>addChoice</code> to build-up a 
 *	new list of choices to replace the ones you've deleted, because once an array of radio or checkbox buttons has
 *	been deleted, there's no trace of them left in the markup. That's because the way HTML forms were designed, radio
 *	and checkbox controls have no parent control to hold their place, unlike "select" controls which are parents to 
 *	all of the "option" controls that represent the choices they offer. So, don't use <code>deleteChoices</code> on 
 *	"input" controls of types "radio" or "checkbox" unless you want to be rid of them, but feel free to use it on
 *	"select" controls.
 * 
 * @param ControlName		The name of the control whose associated choices should be deleted.
 * @return					This <code>Form</code> object.
 * @throws TagException		If a problem is encountered with the manipulation of the markup.
 */
	public Form deleteChoices
	(
		final String			ControlName
	)
	throws TagException
	{
		for (final Match Control : getControlsByName(ControlName)) {
			final String			Type = Control.getType();
			
			if ("input".equals(Type)) {
				final String			Subtype = Control.getAttribute("type");
				
				if ("radio".equals(Subtype) || "checkbox".equals(Subtype))
					Control.delete();
				else
					throw new IllegalArgumentException("deleteChoices can only operate on \"input\" controls of type \"radio\" and \"checkbox\", \"" + Control + "\" is not a valid target.");
				
			} else if ("select".equals(Type)) {
				
				Control.get("option").delete();
				
			} else
				throw new IllegalArgumentException("deleteChoices can only operate on the controls \"input\" (types \"radio\" and \"checkbox\") and \"select\". The control \"" + Control + "\" is not a valid target.");
		}
		
		return this;
	}
	
/**	Adds a new first choice to a pre-existing "<samp>select</samp>" control, group of radio buttons, or checkboxes. If 
 *	there is more than one <samp>select</samp> in this form with the specified name, the choice is added to all of them. 
 *	We would do the same for groups of radio buttons and checkboxes, except that those elements lack an enclosing 
 *	element, so there's no dependable way to identify distinct groups of them.
 * 
 * @param ControlName			Name of the control.
 * @param ControlValue			Value of this choice.
 * @param Selected				<code>true</code> if the choice is selected by default.
 * @param Label					Human-interface label for the new choice.
 * @return						A new <code>Results</code> object that references the new choice control.
 * @throws TagException			If a problem is encountered with the manipulation of the markup.
 * @see #addChoice(boolean, String, Object, boolean, String)
 * @see #addLastChoice(String, Object, boolean, String)
 */
	public Results addFirstChoice
	(
		final String			ControlName,
		final Object			ControlValue,
		final boolean			Selected,
		final String			Label
	)
	throws TagException
	{
		return addChoice(true, ControlName, ControlValue, Selected, Label);
	}
	
/**	Adds a new last choice to a pre-existing "<samp>select</samp>" control, group of radio buttons, or checkboxes. If 
 *	there is more than one <samp>select</samp> in this form with the specified name, the choice is added to all of them. 
 *	We would do the same for groups of radio buttons and checkboxes, except that those elements lack an enclosing 
 *	element, so there's no dependable way to identify distinct groups of them.
 * 
 * @param ControlName			Name of the control.
 * @param ControlValue			Value of this choice.
 * @param Selected				<code>true</code> if the choice is selected by default.
 * @param Label					Human-interface label for the new choice.
 * @return						A new <code>Results</code> object that references the new choice control.
 * @throws TagException			If a problem is encountered with the manipulation of the markup.
 * @see #addFirstChoice(String, Object, boolean, String)
 * @see #addChoice(boolean, String, Object, boolean, String)
 */
	public Results addLastChoice
	(
		final String			ControlName,
		final Object			ControlValue,
		final boolean			Selected,
		final String			Label
	)
	throws TagException
	{
		return addChoice(false, ControlName, ControlValue, Selected, Label);
	}
	
/**	Adds a choice to a pre-existing "<samp>select</samp>" control, group of radio buttons, or checkboxes. If there is
 *	more than one <samp>select</samp> in this form with the specified name, the choice is added to all of them. We
 *  would do the same for groups of radio buttons and checkboxes, except that those elements lack an enclosing element,
 *  so there's no dependable way to identify distinct groups of them.
 * 
 * @param InsertBefore			<code>true</code> if the new choice should be inserted before the existing choices, or 
 *								<code>false</code> if the new choice should be inserted after the existing choices.
 * @param ControlName			Name of the control.
 * @param ControlValue			Value of this choice.
 * @param Selected				<code>true</code> if the choice is selected by default.
 * @param Label					Human-interface label for the new choice.
 * @return						A new <code>Results</code> object that references the new choice control.
 * @throws TagException			If a problem is encountered with the manipulation of the markup.
 * @see #addFirstChoice(String, Object, boolean, String)
 * @see #addLastChoice(String, Object, boolean, String)
 */
	public Results addChoice
	(
		final boolean			InsertBefore,
		final String			ControlName,
		final Object			ControlValue,
		final boolean			Selected,
		final String			Label
	)
	throws TagException
	{
		final Results			NewChoices = new Results();
		final Results			Controls   = getControlsByName(ControlName);
		final String			ElemType   = Controls.getType();
		
		if (Controls.isEmpty())
			throw new IllegalArgumentException("There is no control named \"" + ControlName + "\" in this form.");
		
		if ("input".equals(ElemType)) {
			final Results			BaseElem  = InsertBefore ? Controls.first() : Controls.last();
			final String			InputType = BaseElem.get(0).getAttribute("type");
			
			NewChoices.addResults(BaseElem.add(InsertBefore, createChoice(InputType, ControlName, ControlValue, Selected, Label)));
			
		//	Remove any pre-existing controls that had the same value as the new control.
			
			Controls.get("input[value='" + ControlValue + "']").delete();
			
		} else if ("select".equals(ElemType)) {
			
			for (int Index = 0, Count = Controls.size(); Index < Count; Index++) {
				final Results			ThisSelect       = Controls.nth(Index);
				final Results			RedundantOptions = getControlsByValue(ThisSelect, Null.toEmptyString(ControlValue)).delete();
				
				NewChoices.addResults(ThisSelect.addContent(InsertBefore, createChoice("option", ControlName, ControlValue, Selected, Label)));
				
			//	Remove any pre-existing controls that had the same value as the new control.
				
				RedundantOptions.delete();
			}
			
		} else
			throw new IllegalArgumentException("addChoice can only operate on \"input\" (type \"radio\" and \"checkbox\") and \"select\". The control \"" + Controls.get(0) + "\" is not a valid target.");
		
		return NewChoices;
	}
		
/**	Creates the markup for a new choice ("option", "radio", or "checkbox").
 * 
 * @param Type					The type of control to be created: "option", "radio", or "checkbox".
 * @param ControlName			Name of the control.
 * @param ControlValue			Value of this choice.
 * @param Selected				<code>true</code> if the choice is selected by default.
 * @param Label					Human-interface label for the new choice.
 * @return						An <code>ImmutableMarkup</code> object containing the markup for the new choice.
 * @throws TagException			If a problem is encountered with the manipulation of the markup.
 */
	public static ImmutableMarkup createChoice
	(
		final String			Type,	//	checkbox, radio, option
		final String			ControlName,
		final Object			ControlValue,
		final boolean			Selected,
		final String			Label
	)
	throws TagException
	{
		final ImmutableMarkup	Markup;
		final String 			EncodedLabel = Qwicap.htmlEncode(Label);
		final String			EncodedValue = ControlValue != null ? AttributeUtilities.encodeAttributeValue(ControlValue.toString()) : "";
		
		if ("radio".equals(Type))
			Markup = new XMLString("<label><input type='radio' name='" + ControlName + "' value='" + EncodedValue + '\'' + (Selected ? " checked='checked'" : "") + " />" + EncodedLabel + "</label>");
		else if ("checkbox".equals(Type))
			Markup = new XMLString("<label><input type='checkbox' name='" + ControlName + "' value='" + EncodedValue + '\'' + (Selected ? " checked='checked'" : "") + " />" + EncodedLabel + "</label>");
		else if ("option".equals(Type))
			Markup = new XMLString("<option value='" + EncodedValue + '\'' + (Selected ? " selected='selected'" : "") + '>' + EncodedLabel + "</option>");
		else
			throw new TagException("The supported choice types are 'radio', 'checkbox' and 'option'. The type '" + Type + "' is not valid.");
		
		return Markup;
	}
	
/**	A debugging function that prints out a list of all controls in this form, and the labels associated with them.
 * 
 * @return	This <code>Form</code> object.
 * @deprecated Use {@link #getControlsTable} instead.
 */
	@Deprecated public Form printAllLabels() {
		
		System.out.println();
		System.out.println(getControlsTable());	
		return this;
	}
	
/**	A debugging method that produces a table of all controls in this form, the labels associated with them, and the 
 * 	value(s) of each control. The controls are listed in the same order in which they occur in the markup.
 * 
 * @return	A <code>MutableTable</code> object listing all of the controls in this form, and the labels associated with 
 * 			them.
 */
	public MutableTable getControlsTable() {
		
	//	Get every control in this form.
		
		final Results			AllControls = get("input, select, option, textarea, button");
		final Match[]			ControlsInMarkupOrder = new Match[AllControls.size()];
		
		for (int Index = 0; Index < ControlsInMarkupOrder.length; Index++)
			ControlsInMarkupOrder[Index] = AllControls.get(Index);
		
	//	Sort them according to their order in the markup.
		
		Arrays.sort(ControlsInMarkupOrder, new Comparator<Match>() {
			@Override public int compare(final Match A, final Match B) { return A.getIndex() - B.getIndex(); }
		});
		
	//	Build the table.
		
		final Table				T = new MutableLLTable().addCols("Form Control", "Name", "Value(s)", "Label (Tidy)");
		
		for (final Match Elem : ControlsInMarkupOrder) {
			final FormControlInfo	FCI = new FormControlInfo(this, null, null, Elem);
			
			T.addRow().addAll(Elem, FCI.getName(), new ArrayToString(FCI.getValues()).setFinalSeparator(", "), FCI.getLabelText(true));
		}
		
		return (MutableTable) T;
	}
	
/**	Get the unique identifier that Qwicap automatically assigned to this form.
 * 
 * @return	The unique identifier that Qwicap automatically assigned to this form.
 */
	private String getFormID() {
		final Results			FormIDs = get("input[type='hidden'][name='" + Qwicap.kQwicapFormIDControlName + "']");
		
		if (FormIDs.isEmpty())
			return null;
		
		return FormIDs.get(0).getAttribute("value");
	}
	
	boolean formIDEquals
	(
		final String			FormIDStr
	)
	{
		final String			ThisFormIDStr = getFormID();
		
		return ThisFormIDStr != null && ThisFormIDStr.equals(FormIDStr);
	}
	
	private static Match findFormByID
	(
		final Markup			Doc,
		final String			FormIDStr
	)
	throws FormNotFoundException
	{
		final Results			FormIDs = Doc.get("input[type='hidden'][name='" + Qwicap.kQwicapFormIDControlName + "'][value='" + FormIDStr + "']");
		
		if (FormIDs.isEmpty())
			throw new FormNotFoundException(Doc, FormIDStr);
		
		Match					Elem = FormIDs.get(0);
		
		while (Elem != null && "form".equals(Elem.getType()) == false)
			Elem = Elem.getParent();
		
		if (Elem == null)
			throw new FormNotFoundException(Doc, FormIDStr);
		
		return Elem;
	}

/**	Searches the XHTML page most recently sent to the user for the form that must have supplied the most recently
 *	received form data set.
 * 
 * @param ThePage								The page that is to be searched for a matching form.
 * @param DataSet								The form data set whose form of origin is to be found.
 * @param IfThereIsOnlyOneFormAssumeItIsTheOne	If <code>true</code>, and the document contains only one form, that one 
 *												form is assumed to the be the form that supplied the data set. No 
 *												verification of that assumption is attempted.
 * @return										A <code>Match</code> referencing the relevant form.
 * @throws FormNotFoundException				If a form that could supply the specified form data set can't be
 *												found in the specified markup.
 */
	private static Match findRelevantForm
	(
		final Markup			ThePage,
		final FormDataSet		DataSet,
		final boolean			IfThereIsOnlyOneFormAssumeItIsTheOne
	)
	throws FormNotFoundException
	{
		try {
			
			return findRelevantForm(ThePage, DataSet, IfThereIsOnlyOneFormAssumeItIsTheOne, false);
			
		} catch (FormNotFoundException e) {
			
		//	We couldn't find a matching form when respecting Qwicap parameters, so try again while ignoring Qwicap
		//	parameters.
			
			if (DataSet.hasControlNamesStartingWith("qwicap"))
				return findRelevantForm(ThePage, DataSet, IfThereIsOnlyOneFormAssumeItIsTheOne, true);
			else
				throw e;
		}
	}

/**	Searches an XHTML page for the form that must have supplied the a specified set of parameters.
 * 
 * @param ThePage								The markup that is to be searched for a matching form.
 * @param DataSet								The form data set whose form of origin is to be found.
 * @param IfThereIsOnlyOneFormAssumeItIsTheOne	If <code>true</code>, and the document contains only one form, that one 
 *												form is assumed to the be the form that supplied the data set. No 
 *												verification of that assumption is attempted.
 * @param IgnoreQwicapControls					If <code>true</code>, the various parameters that Qwicap automatically 
 *												embeds in forms before sending a page to the user are ignored. If 
 *												<code>false</code>, the Qwicap parameters are respected, and their 
 *												absence from a form will cause the form to be rejected as a possible 
 *												source of the supplied data set.
 * @return										A <code>Match</code> referencing the relevant form.
 * @throws FormNotFoundException				If a form that could supply the specified form data set can't be
 *												found in the specified page.
 */
	private static Match findRelevantForm
	(
		final Markup			ThePage,
		final FormDataSet		DataSet,
		final boolean			IfThereIsOnlyOneFormAssumeItIsTheOne,
		final boolean			IgnoreQwicapControls		//	Ignore any parameter whose name begins with "qwicap"
	)
	throws FormNotFoundException
	{
	//	todo Consider caching pre-compiled CSS selectors for all known method values in a HashMap, and using those instead of the following dynamically generated CSS pattern.
		
		final Results			Forms = ThePage.get("form[method='" + DataSet.getMethod() + "']");
		
		if (IfThereIsOnlyOneFormAssumeItIsTheOne && Forms.size() == 1)
			return Forms.get(0);
			
		final String			Action = DataSet.getContextPath() + DataSet.getServletPath();
		
	//	For each form in the document...
		
		for (final Match Form : Forms) {
			
		//	Get the form's "action" attribute to see if it refers to this web application.
		//	(It would be counter-productive to examine, or operate-on, forms that could not
		//	have sent us the parameter(s) we're rejecting.)
			
			final String			FormAction = Form.getAttribute("action");
			
			if (FormAction.endsWith(Action) || actionRefersToThisWebApp(FormAction, null)) {
			
			//	Verify that all of the parameters we just received could have been supplied by this form.
			//	There's room to make this check more rigorous, for instance by checking for matching values from 
			//	hidden input items, but this should be good enough for now.
				
				boolean					HasAllCtrls = true;
				
				for (final FormControlValue Ctrl : DataSet) {
					final String			CtrlName = Ctrl.getName();
					
					if (IgnoreQwicapControls && Ctrl.isQwicapControl())
						continue;
					
					if (getControlsByName(Form, CtrlName).isEmpty()) {
						
						HasAllCtrls = false;
						break;
					}
				}
				
				if (HasAllCtrls)
					return Form;
			}
		}
		
		throw new FormNotFoundException(ThePage, DataSet, IgnoreQwicapControls);
	}
	
//	Note that the Page.addPageIDToLinks method still uses its own logic rather than the Form.actionRefersToThisWebApp
//	method. Ideally, that logic would be replaced with an invocation of this method.
	
	static boolean actionRefersToThisWebApp
	(
		final String			Action,		//	Can be null.
		final URL				WebAppURL	//	Can be null.
	)
	{
	
	//	Forms without "action" attributes (which is the case that would allow Action to be null here) do submit their
	//	data sets to the current web application.
	
		if (Action == null)
			return true;
		
		if ("".equals(Action) || ".".equals(Action) || "./".equals(Action))
			return true;
		
		if (WebAppURL != null) {
			final String			WebAppURLStr = WebAppURL.toString();
			
			if (Action.equals(WebAppURLStr))
				return true;
			
			if (Action.startsWith("http")) {	//	Remember that this test encompasses both "http" and "https".
				
				try {
					
					if (WebAppURL.equals(new URL(Action)))
						return true;
					
				} catch (MalformedURLException e) {
				//	Discard exception.
				}
				
			} else {
				
			//	If WebAppURL is "http://127.0.0.1:8080/wumpus/", the getPath method will return "/wumpus/".
				
				if (Action.equals(WebAppURL.getPath()))
					return true;			
			}
		}
		
		return false;
	}
		
	private static Match getControlByValue
	(
		final Results			Elems,		//	A list of control elements.
		final String			ValueStr	//	Can be null
	)
	{
		if (ValueStr != null) {
			
			for (final Match Elem : Elems) {
				
				if ("select".equals(Elem.getType())) {
					final Match				Option = getControlByValue(Elem.get("option"), ValueStr);
					
					if (Option != null)
						return Option;
					
				} else if (ValueStr.equals(getControlValueLow(Elem)))
					return Elem;
			}
		}
		
		return null;
	}
	
	private static Results getControlsByValue
	(
		final Results			Elems,
		final String			ValueStr	//	Can be null
	)
	{
			
	//	If Value is null, return all the items in Elems, since we can't differentiate between them without a value 
	//	(and maybe even *with* a value).
		
		if (ValueStr == null)	
			return Elems;
		
		final Results			Matches = new Results();
		
		for (final Match Elem : Elems) {
			
			if ("select".equals(Elem.getType())) {
				final Match				Option = getControlByValue(Elem.get("option"), ValueStr);
				
				if (Option != null) {
					
					Matches.addMatch(Option);
					break;	//	If we care about the possibility of multiple controls supplying the same value, we shouldn't break here.
				}
				
			} else if (ValueStr.equals(getControlValueLow(Elem))) {
				
				Matches.addMatch(Elem);
				break;		//	If we care about the possibility of multiple controls supplying the same value, we shouldn't break here.
			}
		}
		
	//	At worst, we'll return an empty Results object.
		
		return Matches;
	}
		
/**	Finds the form controls having the specified name. The returned controls may be of type "input", "select",
 *	"button", or "textarea". They will not be of type "option", because those are always subsumed by a "select" control.
 * 
 * @param ControlName			The name of the control to be searched-for.
 * @return						A list of zero or more form controls having the specified name.
 */
	public Results getControlsByName
	(
		final String			ControlName
	)
	{
		return getControlsByName(TheForm, ControlName);
	}

/**	Finds the form controls having the specified name. The returned controls may be of type "input", "select",
 *	"button", or "textarea". They will not be of type "option", because those are always subsumed by a "select" control.
 * 
 * @param Form					A <code>Match</code> object referencing the form to be searched.
 * @param ControlName			The name of the control to be searched-for.
 * @return						A list of zero or more form controls having the specified name.
 */
	private static Results getControlsByName
	(
		final Match				Form,
		final String			ControlName
	)
	{
	//	Get every control with the specified name. Note that controls of different types can have the same name. So we
	//	have to examine every type of control to be sure that we build a complete list.
	
		final Results			FormControls;
		
	//	Note: If there are form controls other than input, select, button and textarea that can supply parameters, we 
	//	need to add them here.
		
		FormControls = Form.get(
			"input[name='"    + ControlName + "']," + 
			"select[name='"   + ControlName + "']," + 
			"button[name='"   + ControlName + "']," +
			"textarea[name='" + ControlName + "']"
		);
						
	//	Search for "input" controls of type "submit" that are using Qwicap's hack for representing a
	//	true name/value pair within the "name" attribute.
		
		final Results			InputSubmits = Form.get("input[type='submit']");
		
		if (InputSubmits.notEmpty()) {
			
			for (final Match Elem : InputSubmits) {
				final InputSubmitButton	ISB = new InputSubmitButton(Elem);
				
				if (ControlName.equals(ISB.getName()))
					FormControls.addMatch(Elem);
			}
		}
			
		return FormControls;
	}	
	
/**	Returns a <code>FormControlInfo</code> object describing the form control with the specified name (which is 
 * 	understood to be shared with other controls in the form), and index number. In other words, it gets the Nth control
 * 	of the specified name from the form. The controls are considered in the order in which they occur in the markup, 
 * 	with index 0 considered to be the first control.
 * 
 * @param ControlName		Name of the parameter whose originating form control should be found.
 * @param ControlIndex		Index number of the originating form control.
 * @return					A <code>FormControlInfo</code> instance describing the relevant form control, otherwise 
 *							<code>null</code>.
 */
	public FormControlInfo getControlByIndex
	(
		final String			ControlName,
		final int				ControlIndex
	)
	{
		final Results			Controls = getControlsByName(TheForm, ControlName);
		
		if (ControlIndex < 0)
			throw new IllegalArgumentException("A negative index number cannot be used to identify a control (name \"" + ControlName + "\".");
		if (ControlIndex >= Controls.size())
			throw new IllegalArgumentException("There are only " + Controls.size() + " controls in this form that are named \"" + ControlName + "\", so " + ControlIndex + " is not a valid index.");
		
		final Match				Control = Controls.get(ControlIndex);
		
		return new FormControlInfo(this, ControlName, getControlValue(Control), Control);
	}
	
/**	Returns a <code>FormControlInfo</code> object describing the form control with the specified name (which is 
 * 	understood to be shared with other controls in the form), index number, and value. In other words, it gets the Nth 
 * 	control of the specified name from the form, provided that the supplied value matches it, or some part of it. The 
 * 	controls are considered in the order in which they occur in the markup, with index 0 considered to be the first 
 * 	control.
 * 
 * @param ControlName		Name of the parameter whose originating form control should be found.
 * @param ControlIndex		The index number of the originating form control.
 * @param ControlValue		The value whose originating form control should be found.
 * @return					A <code>FormControlInfo</code> instance describing the relevant form control, otherwise 
 *							<code>null</code>.
 */
	public FormControlInfo getControlByIndexAndValue
	(
		final String			ControlName,
		final int				ControlIndex,
		final Object			ControlValue
	)
	{
		final String			ControlValueStr = ControlValue == null ? null : ControlValue.toString();
		final Results			Controls        = getControlsByName(TheForm, ControlName);
		
		if (ControlIndex < 0)
			throw new IllegalArgumentException("A negative index number cannot be used to identify a control (name \"" + ControlName + "\".");
		if (ControlIndex >= Controls.size())
			throw new IllegalArgumentException("There are only " + Controls.size() + " controls in this form that are named \"" + ControlName + "\", so " + ControlIndex + " is not a valid index.");
		
		final Match				Control = getControlByValue(Controls.nth(ControlIndex), ControlValueStr);
		
		return new FormControlInfo(this, ControlName, ControlValueStr, Control);
	}
	
/**	Returns a <code>FormControlInfo</code> object describing the specific form control that supplied, or could have
 * 	supplied, a particular name/value pair in a form data set. Since groups of radio buttons and checkboxes consist of 
 *	sets of controls that all submit name/value pairs with the same name, this function has to match the value as well
 * 	in order to ensure that it returns the control specifically responsible for the value.
 *
 *	<p>Note that a specific form control can't necessarily be identified from a control name and value, because some 
 *	controls (input for text and passwords, for example) have their values supplied client-side by the user, in other
 *	words, they can't be identified because the value is not in the document markup. In such cases, the first (and 
 *	presumably the only) form control that supplies a parameter of the specified name is returned. This is also the 
 *	behavior when, for whatever reason, the supplied control value is <code>null</code>.
 * 
 * @param ControlName		Name of the parameter whose originating form control should be found.
 * @param ControlValue		Value of the parameter whose originating form control should be found. Can be 
 *							<code>null</code>; see method description.
 * @return					A <code>FormControlInfo</code> instance describing the relevant form control, otherwise 
 *							<code>null</code>.
 */
	public FormControlInfo getControl
	(
		final String			ControlName,
		final Object			ControlValue	//	Can be null
	) 
	{
		final String			ControlValueStr = ControlValue == null ? null : ControlValue.toString();
		final Results			Controls        = getControlsByName(TheForm,  ControlName    );
		Match					Control         = getControlByValue(Controls, ControlValueStr);
		
	//	We can't necessarily identify a control by its value, because some controls (input for text and passwords) have
	//	their value supplied client-side by the user, i.e. because the value is not in the document markup.
	//	In such cases, getControlByValue will have returned null, and we simply return the first (and 
	//	presumably the only) control in the list of form controls that supply parameters of the specified name.
		
		if (Control == null && Controls.notEmpty())
			Control = Controls.get(0);
		
	//	If, for any reason, we haven't found a relevant form control, leave now.
		
		if (Control == null)
			return null;
		
		return new FormControlInfo(this, ControlName, ControlValueStr, Control);
	}
	
/**	Returns an array of {@link FormControlValue} objects describing the form controls that supplied, or could have 
 *	supplied, a particular name/value pair in a form data set. Since groups of radio buttons and checkboxes consist of 
 *	sets of controls that all submit values of the same name, this function has to match the value to ensure that it 
 *	returns the control specifically responsible for the value.
 *
 *	<p>Note that a specific form control can't necessarily be identified from a control name and value, because some 
 *	controls (input for text and passwords, for example) have their values supplied client-side by the user, in other
 *	words, they can't be identified because the value is not in the document markup. In such cases, all the form 
 *	controls of the specified name are returned. This is also the behavior when, for whatever reason, the supplied 
 *	control value is <code>null</code>.
 * 
 * @param ControlName		Name of the control.
 * @param ControlValue		Value of the control. Can be <code>null</code>; see method description.
 * @return					An array of <code>FormControlInfo</code> instances describing the relevant form controls, 
 *							or an empty array if no relevant form controls could be found.
 */
	public FormControlInfo[] getControls
	(
		final String			ControlName,
		final Object			ControlValue	//	Can be null.
	) 
	{
		final String						ControlValueStr              = ControlValue == null ? null : ControlValue.toString();
		final ArrayList<FormControlInfo>	InfoList                     = new ArrayList<FormControlInfo>();
		final Results						ControlsMatchingName         = getControlsByName(TheForm, ControlName);
		final Results						ControlsMatchingNameAndValue = getControlsByValue(ControlsMatchingName, ControlValueStr);
		
		if (ControlsMatchingNameAndValue.isEmpty()) {
			
		//	We found no form controls whose name AND value matched the supplied name and value, so return
		//	all the controls that supply a value of the specified name.
			
			for (final Match Elem : ControlsMatchingName)
				InfoList.add(new FormControlInfo(this, ControlName, ControlValueStr, Elem));
			
		} else {
		
		//	We found one or more controls whose name AND value matched the supplied parameter name and value; 
		//	return all of them.
			
			for (final Match Elem : ControlsMatchingNameAndValue)
				InfoList.add(new FormControlInfo(this, ControlName, ControlValueStr, Elem));
		}
		
	//	At worst, we'll return a zero-length array.
		
		return InfoList.toArray(new FormControlInfo[InfoList.size()]);
	}

/**	Returns an array of {@link FormControlValue} objects describing the "submit" controls in this form. 
 * 
 * @return					Descriptions of the "submit" controls in this form.
 */
	public FormControlInfo[] getSubmitControls() {
		final Results			Submits   = TheForm.get("input[type='submit'],input[type='image'],button[type='submit']");
		final int				CtrlCount = Submits.size();
		final FormControlInfo[]	CtrlInfo  = new FormControlInfo[CtrlCount];
		
		for (int Index = 0; Index < CtrlCount; Index++) {
			final Match 			Ctrl = Submits.get(Index);
			
			CtrlInfo[Index] = new FormControlInfo(this, getControlName(Ctrl), null, Ctrl);
		}
		
		return CtrlInfo;
	}
	
/**	Adds to the value of a form control. In the case of most single-value controls, the new value is appended to the 
 *	old value. In the case of a multi-value "select" control, an "option" tag with the specified value is located 
 *	within it, and that option is selected. In the case of a single-value "select" control, the existing value (if any)
 *  is cleared, then an "option" tag with the specified value is located within it, and that option is selected (this 
 *	is no different from calling {@link #setControlValue(String, Object)}, but it's a single-value control, so we have 
 *	no choice). In the case of "input" controls of type "checkbox", the control with the specified value is located and
 *	selected ("checked"). In the case of "input" controls of type "radio", the control with the specified value 
 *  is located and selected ("checked"), and all of the other controls are de-selected (their "checked" attributes are
 *	deleted). Where those radio buttons are concerned, this method is no different from calling 
 *	{@link #setControlValue(String, Object)}, but a group of radio buttons can represent only one value, so we have no 
 *	choice.
 * 
 * @param ControlName			The name of a form control of type "input", "select", "textarea" or "button".
 * @param ControlValue			Value of the form control.
 * @return 						This <code>Form</code> object.
 * @throws TagException			If a problem is encountered with the manipulation of the markup.
 */
	public Form addToControlValue
	(
		final String			ControlName,
		final Object			ControlValue
	) 
	throws TagException
	{
		for (final Match Elem : getControlsByName(ControlName))
			addToControlValue(Elem, ControlValue);
		
		return this;
	}
	
/**	Adds to the value of a form control. In the case of most single-value controls, the new value is appended to the 
 *	old value. In the case of a multi-value "select" control, an "option" tag with the specified value is located 
 *	within it, and that option is selected. In the case of a single-value "select" control, the existing value (if any)
 *  is cleared, then an "option" tag with the specified value is located within it, and that option is selected (this 
 *	is no different from calling {@link #setControlValue(Match, Object)}, but it's a single-value control, so we have 
 *	no choice). In the case of an "input" control of type "checkbox", the control is selected ("checked") if its value 
 *	matches the specified value, and it is left untouched otherwise. In the case of an "input" control of type "radio", 
 *	the control is selected ("checked") if its value matches the specified value, and it is unselected (its "checked" 
 *	attribute is deleted) otherwise.
 * 
 * @param Control				A <code>Match</code> instance that refers to an "input", "select", "textarea" or 
 *								"button" control.
 * @param ControlValue			Value of the form control.
 * @return 						This <code>Form</code> object.
 * @throws TagException			If a problem is encountered with the manipulation of the markup.
 */
	public static Match addToControlValue
	(
		final Match				Control,
		final Object			ControlValue
	)
	throws TagException
	{
		final String			ElementType = Control.getType();
		final String			Value       = ControlValue.toString();
		
		if ("input".equals(ElementType)) {
			
			addToInputControlValue(Control, Value);	//	Ultimately, setAttribute calls are made, and they encode the value.
			
		} else if ("select".equals(ElementType)) {
			
			if (Control.hasAttribute("multiple") == false)
				Control.get("option[selected]").deleteAttribute("selected");
			
			setSelectControlValue(Control, Value);
			
		} else if ("textarea".equals(ElementType)) {
			
			Control.addContent(false, Value);
			
		} else if ("button".equals(ElementType)) {
			final String			OrigValue = Control.getAttribute("value");
			
			if (OrigValue != null)
				Control.setAttribute("value", OrigValue + Value);
			else
				Control.setAttribute("value", Value);
		}
		
		return Control;
	}
	
/**	Sets the value of a form control, identified by name. In the case of checkboxes, radio buttons, and "select" 
 *	controls, the value is used in addition to the name in order to identify the relevant form control; when a form 
 *	control with a matching name and value is found, a "checked" or "selected" attribute is added, as appropriate. 
 *	Note that you cannot repeatedly invoke this method in order to select more than one checkbox, or option, in a 
 *	multiple selection control, because it removes the "checked" or "selected" attributes from all controls with the 
 *	specified name before it checks/selects the control with the matching value. However, you may make multiple 
 *	selections with a <i>single</i> <code>setControlValue</code> invocation by passing an array of values to be 
 *	selected as the <code>ControlValue</code> parameter. The array elements may be of any type; the 
 *	<code>toString</code> methods of the non-null elements will be invoked, and the controls whose values match the 
 *	returned <code>String</code>s (if any) will be checked/selected. Alternately, if you prefer to set the values of
 *	multi-value controls by using multiple method invocations, start by invoking {@link #clearControlValue(String)} to 
 * 	remove all pre-existing checks/selects, then use invocations of {@link #addToControlValue(String, Object)} to 
 * 	check/select as many items as necessary. For single-value controls, this <code>setControlValue</code> method is 
 * 	always appropriate.
 * 
 * @param ControlName			Name of the form control.
 * @param ControlValue			Value of the form control. If <code>null</code>, the value of the control is cleared 
 *								with the {@link #clearControlValue(Match)} method. If the value is an array, and the
 *								control supports multiple values, the control is given each of the values in the array.
 *								The array may be of any type; its elements are converted to strings using their 
 *								<code>toString</code> methods.
 * @return						A reference to this <code>Form</code> object.
 * @throws TagException			If a problem is encountered with the manipulation of the markup.
 * @see #addToControlValue(String, Object)
 * @see #clearControlValue(String)
 */
	public Form setControlValue
	(
		final String			ControlName,
		final Object			ControlValue
	) 
	throws TagException
	{
		for (final Match Control : getControlsByName(ControlName))
			setControlValue(Control, ControlValue);
		
		return this;
	}
	
/**	Sets the value of a form control, identified by name. In the case of checkboxes, radio buttons, and "select" 
 *	controls, the value is used in addition to the name in order to identify the relevant form control; when a form 
 *	control with a matching name and value is found, a "checked" or "selected" attribute is added, as appropriate. 
 *	Note that you cannot repeatedly invoke this method in order to select more than one checkbox, or option, in a 
 *	multiple selection control, because it removes the "checked" or "selected" attributes from all controls with the 
 *	specified name before it checks/selects the control with the matching value. However, you may make multiple 
 *	selections with a <i>single</i> <code>setControlValue</code> invocation by passing an array of values to be 
 *	selected as the <code>ControlValue</code> parameter. The array elements may be of any type; the 
 *	<code>toString</code> methods of the non-null elements will be invoked, and the controls whose values match the 
 *	returned <code>String</code>s (if any) will be checked/selected. Alternately, if you prefer to set the values of
 *	multi-value controls by using multiple method invocations, start by invoking {@link #clearControlValue(String)} to 
 * 	remove all pre-existing checks/selects, then use invocations of {@link #addToControlValue(String, Object)} to 
 * 	check/select as many items as necessary. For single-value controls, this <code>setControlValue</code> method is 
 * 	always appropriate.
 * 
 * @param Control				A <code>Match</code> object corresponding to a tag of type "input", "select", 
 *								"textarea", or "button".
 * @param ControlValue			Value of the form control. If <code>null</code>, the value of the control is cleared 
 *								with the {@link #clearControlValue(Match)} method. If the value is an array, and the
 *								control supports multiple values, the control is given each of the values in the array.
 *								The array may be of any type; its elements are converted to strings using their 
 *								<code>toString</code> methods.
 * @return						A reference to this <code>Form</code> object.
 * @throws TagException			If a problem is encountered with the manipulation of the markup.
 * @see #addToControlValue(String, Object)
 * @see #clearControlValue(String)
 */
	public static Match setControlValue
	(
		final Match				Control,
		final Object			ControlValue
	)
	throws TagException
	{
		if (ControlValue == null) {
			
			clearControlValue(Control);
			
		} else {
			final String			ControlType = Control.getType();
			boolean					IsMultiElementControl = false;
			
			if ("input".equals(ControlType) && ("checkbox".equals(Control.getAttribute("type")) || "radio".equals(Control.getAttribute("type"))))
				IsMultiElementControl = true;
			else if ("select".equals(ControlType))
				IsMultiElementControl = true;
			
			if (IsMultiElementControl)
				clearControlValue(Control);	//	This is only necessary for controls made up of multiple elements.
			
			if (ControlValue.getClass().isArray()) {
				final int				ValueCount = Array.getLength(ControlValue);
				
				if (IsMultiElementControl == false && ValueCount > 1)
					throw new IllegalArgumentException("The control named \"" + getControlName(Control) + "\" in document \"" + Control.getMarkup().getMarkupName() + "\" can have only one value, but an array of " + ValueCount + " values has been supplied: " + new ArrayToString(ControlValue).alwaysQuote());
				
				for (int Index = 0; Index < ValueCount; Index++) {
					final Object			Elem = Array.get(ControlValue, Index);
					
					if (Elem != null)
						setControlValueLow(Control, Elem.toString());
				}
				
			} else
				setControlValueLow(Control, ControlValue.toString());
		}
		
		return Control;
	}
	
/**	Sets the value of multiple form controls, as identified by the names of the parameters the controls will submit. In 
 *	the case of checkboxes, radio buttons, and "select" controls, the parameter value is used in addition to the name to 
 *	identify the relevant form control; when a form control with a matching name and value is found, a "checked" or 
 *	"selected" attribute is added, as appropriate. Note that this method does not support selecting more than one 
 *	checkbox, or option, in a multiple selection control, because it removes the "checked" or "selected" attributes 
 *	from all controls of the specified name before it checks/selects the control with the matching value. Instead, use 
 *	{@link #clearControlValue(String)} to remove all pre-existing checks/selects, then use invocations of 
 * 	{@link #addToControlValue(String, Object)} to check/select as many items as necessary. This <code>setControlValues</code>
 * 	method is appropriate for all other form controls, including radio buttons and single-selection lists.
 * 
 * @param NameValuePairs		An array composed of control name/value pairs such that the even-numbered elements 
 *								are the names of controls, and the odd-numbered elements are the values, e.g. 
 *								"name1", "value1", "name2", "value2", "name3", "value3". The values may be of any type
 *								(they will ultimately be converted to <code>String</code>s using their 
 *								<code>toString</code> methods). If a value is <code>null</code>, the value of the 
 *								control is cleared with the {@link #clearControlValue(Match)} method. If a value is an
 * 								array, the control is given all the values specified by the array. (The array may be of 
 *								any type; it's elements are converted to strings using their 
 *								<code>toString</code> methods.) The control names in the even-numbered elements of this 
 *								parameter will ordinarily be <code>String</code>s, but can actually be of any type (as 
 *								with values, their <code>toString</code> methods are invoked).
 * @return						This <code>Form</code> object.
 * @throws TagException			If a problem is encountered with the manipulation of the markup.
 */
	public Form setControlValues
	(
		final Object...			NameValuePairs
	)
	throws TagException
	{
		final int				NVLength = NameValuePairs.length;
		
		if (NVLength % 2 != 0)
			throw new IllegalArgumentException("The array passed to setControlValues is meant to be composed of name/value pairs, and therefore should contain an even number of elements. The array passed in this case, however, contains an odd number of elements: " + NVLength + '.');
		
		for (int Index = 0; Index < NVLength; Index += 2)
			setControlValue(NameValuePairs[Index].toString(), NameValuePairs[Index + 1]);
		
		return this;
	}
	
/**	Enables all controls in this form, optionally including even the "submit" button(s).
 * 
 * @param IncludesSubmitButtons	<code>true</code> if all submit buttons in this form should also be enabled, 
 *								<code>false</code> otherwise.
 * @return						This <code>Form</code> object.
 * @throws TagException			If a problem is encountered with the manipulation of the markup.
 */
	public Form enableAllControls
	(
		final boolean			IncludesSubmitButtons
	) 
	throws TagException 
	{
		final Results			UserControls = get("input[type='text'], input[type='password'], input[type='file'], input[type='radio'], input[type='checkbox'], select, option, textarea");
		
		if (IncludesSubmitButtons)
			UserControls.addResults(get("input[type='submit'], button[type='submit']"));
		
		for (final Match Elem : UserControls)
			enableControl(Elem);
		
		return this;
	}
	
/**	Disables all controls in this form, optionally including even the "submit" button(s).
 * 
 * @param IncludesSubmitButtons	<code>true</code> if all submit buttons in this form should also be disabled, 
 *								<code>false</code> otherwise.
 * @return						This <code>Form</code> object.
 * @throws TagException			If a problem is encountered with the manipulation of the markup.
 */
	public Form disableAllControls
	(
		final boolean			IncludesSubmitButtons
	) 
	throws TagException 
	{
		final Results			UserControls = get("input[type='text'], input[type='password'], input[type='file'], input[type='radio'], input[type='checkbox'], select, option, textarea");
		
		if (IncludesSubmitButtons)
			UserControls.addResults(get("input[type='submit'], button[type='submit']"));
		
		for (final Match Elem : UserControls)
			disableControl(Elem);
		
		return this;
	}
	
/**	Enables or disables the named control according to a <code>boolean</code> parameter.
 * 
 * @param ControlName			Name of the form control.
 * @param Enable				<code>true</code> if the control should be enabled, <code>false</code> if it should be
 * 								disabled.
 * @return						This <code>Form</code> object.
 * @throws TagException			If a problem is encountered with the manipulation of the markup.
 * @see #enableControl(String)
 * @see #disableControl(String)
 */
	public Form setControlEnable
	(
		final String			ControlName,
		final boolean			Enable
	)
	throws TagException
	{
		if (Enable)
			enableControl(ControlName);
		else
			disableControl(ControlName);
		
		return this;
	}
	
/**	Enables or disables the specified control according to a <code>boolean</code> parameter.
 * 
 * @param Control				A <code>Match</code> object corresponding to a tag of type "input", "select", 
 *								"textarea", or "button".
 * @param Enable				<code>true</code> if the control should be enabled, <code>false</code> if it should be
 * 								disabled.
 * @return						The specified control.
 * @throws TagException			If a problem is encountered with the manipulation of the markup.
 */
	public static Match setControlEnable
	(
		final Match				Control,
		final boolean			Enable
	)
	throws TagException
	{
		if (Enable)
			enableControl(Control);
		else
			disableControl(Control);
		
		return Control;
	}
	
/**	Enables a form control identified by the name of the parameter that the control will submit in the form data set. In the case of 
 *	multiple selection controls (checkboxes, radio buttons, and "select" controls) all controls are enabled. (To
 *	enable individual selections use the {@link #enableControl(String, Object)} method.) Note that "enabling" 
 *	means nothing more than deleting the "disabled" attribute from applicable form controls.
 * 
 * @param ControlName			Name of the form control.
 * @return 						This <code>Form</code> object.
 * @throws TagException			If a problem is encountered with the manipulation of the markup.
 * @see #enableControl(String, Object)
 * @see <a href='http://www.w3.org/TR/html4/interact/forms.html#adef-disabled'>Forms in HTML Documents, sec. 17.12.1, Disabled controls</a>
 */	
	public Form enableControl
	(
		final String			ControlName
	)
	throws TagException
	{
		return enableControl(ControlName, null);
	}
	
/**	Enables multiple form controls as identified by their names. In the case of multiple selection controls 
 *	(checkboxes, radio buttons, and "select" controls) all selections in the controls are ensabled. (To ensable 
 *	individual selections use the {@link #enableControl(String, Object)} method.) Note that "enabling" means nothing 
 *	more than deleting the "disabled" attribute from applicable form controls.
 * 
 * @param ControlNames			The names of the form controls to enable.
 * @return 						This <code>Form</code> object.
 * @throws TagException			If a problem is encountered with the manipulation of the markup.
 * @see #enableControl(String)
 * @see <a href='http://www.w3.org/TR/html4/interact/forms.html#adef-disabled'>Forms in HTML Documents, sec. 17.12.1, Disabled controls</a>
 */	
	public Form enableControls
	(
		final String...			ControlNames
	)
	throws TagException
	{
		for (int Index = 0, Count = ControlNames.length; Index < Count; Index++)
			enableControl(ControlNames[Index]);
		
		return this;
	}

/**	Enables the named control, if its value matches the specified value, or the specified value is 
 *	<code>null</code>. In the case of checkboxes, radio buttons, and "select" controls, the parameter value is used in 
 *	addition to the name to identify the relevant form control. In all other cases the value may be omitted. (Note that 
 *	"enabling" means nothing more than removing the "disabled" attribute from applicable form controls, if present.)
 * 
 * @param ControlName			Name of the form control.
 * @param ControlValue			Value of the form control, or <code>null</code>. The value is only needed when 
 *								enabling specific items in multiple selection controls (checkboxes, radio buttons, and 
 *								"select" controls). When dealing with those controls, passing <code>null</code> in 
 *								this parameter causes all selections in the group to be enabled.
 * @return 						This <code>Form</code> object.
 * @throws TagException			If a problem is encountered with the manipulation of the markup.
 * @see <a href='http://www.w3.org/TR/html4/interact/forms.html#adef-disabled'>Forms in HTML Documents, sec. 17.12.1, Disabled controls</a>
 */	
	public Form enableControl
	(
		final String			ControlName,
		final Object			ControlValue	//	Can be null.
	) 
	throws TagException
	{
		for (final Match Control : getControlsByName(ControlName))
			enableControl(Control, ControlValue);
		
		return this;
	}
	
/**	Enables the specified control, if its value matches the specified value, or the specified value is 
 *	<code>null</code>.
 * 
 * @param Control				A <code>Match</code> object corresponding to a tag of type "input", "select", 
 *								"textarea", or "button".
 * @param ControlValue			The value the control must have in order to be enabled, or <code>null</code> if it 
 *								should be enabled regardless of its value.
 * @return						The specified control.
 * @throws TagException			If a problem is encountered with the manipulation of the markup.
 * @see <a href='http://www.w3.org/TR/html4/interact/forms.html#adef-disabled'>Forms in HTML Documents, sec. 17.12.1, Disabled controls</a>
 */
	public static Match enableControl
	(
		final Match				Control,
		final Object			ControlValue
	)
	throws TagException
	{
		if (ControlValue == null || ControlValue.toString().equals(getControlValueLow(Control)))
			enableControl(Control);
		
		return Control;	
	}
	
/**	Disables the named control. In the case of multiple selection controls (checkboxes, radio buttons, and "select" 
 *	controls) all selections in the controls are disabled. (To disable individual selections use the 
 *	{@link #disableControl(String, Object)} method.) Note that "disabling" means nothing more than adding the "disabled" 
 *	attribute to applicable form controls.
 * 
 * @param ControlName			Name of the form control to disable.
 * @return 						This <code>Form</code> object.
 * @throws TagException			If a problem is encountered with the manipulation of the markup.
 * @see #disableControl(String, Object)
 * @see <a href='http://www.w3.org/TR/html4/interact/forms.html#adef-disabled'>Forms in HTML Documents, sec. 17.12.1, Disabled controls</a>
 */	
	public Form disableControl
	(
		final String			ControlName
	)
	throws TagException
	{
		return disableControl(ControlName, null);
	}
	
/**	Disables multiple form controls as identified by their names. In the case of multiple selection controls 
 *	(checkboxes, radio buttons, and "select" controls) all selections in the controls are disabled. (To disable 
 *	individual selections use the {@link #disableControl(String, Object)} method.) Note that "disabling" means nothing 
 *	more than adding the "disabled" attribute to applicable form controls.
 * 
 * @param ControlNames			The names of the form controls to disable.
 * @return 						This <code>Form</code> object.
 * @throws TagException			If a problem is encountered with the manipulation of the markup.
 * @see #disableControl(String)
 * @see <a href='http://www.w3.org/TR/html4/interact/forms.html#adef-disabled'>Forms in HTML Documents, sec. 17.12.1, Disabled controls</a>
 */	
	public Form disableControls
	(
		final String...			ControlNames
	)
	throws TagException
	{
		for (int Index = 0, Count = ControlNames.length; Index < Count; Index++)
			disableControl(ControlNames[Index]);
		
		return this;
	}
	
/**	Disables the named control, if its value matches the specified value, or the specified value is 
 *	<code>null</code>. In the case of checkboxes, radio buttons, and "select" controls, the control value is used in 
 *	addition to the name to identify the relevant form control. In all other cases the value may be omitted. (Note that 
 *	"disabling" means nothing more than adding the "disabled" attribute to applicable form controls.)
 * 
 * @param ControlName			Name of the form control.
 * @param ControlValue			Value of the form control, or <code>null</code>. The value is only needed when 
 *								disabling specific items in multiple selection controls (checkboxes, radio buttons, and 
 *								"select" controls). When dealing with those controls, passing <code>null</code> in 
 *								this parameter causes all selections in the group to be disabled.
 * @return 						This <code>Form</code> object.
 * @throws TagException			If a problem is encountered with the manipulation of the markup.
 * @see <a href='http://www.w3.org/TR/html4/interact/forms.html#adef-disabled'>Forms in HTML Documents, sec. 17.12.1, Disabled controls</a>
 */	
	public Form disableControl
	(
		final String			ControlName,
		final Object			ControlValue
	) 
	throws TagException
	{
		for (final Match Control : getControlsByName(ControlName))
			disableControl(Control, ControlValue);
		
		return this;
	}
	
/**	Disables the specified control, if its value matches the specified value, or the specified value is 
 *	<code>null</code>.
 * 
 * @param Control				A <code>Match</code> object corresponding to a tag of type "input", "select", 
 *								"textarea", or "button".
 * @param ControlValue			The value the control must have in order to be disabled, or <code>null</code> if it 
 *								should be disabled regardless of its value.
 * @return						The specified control.
 * @throws TagException			If a problem is encountered with the manipulation of the markup.
 * @see <a href='http://www.w3.org/TR/html4/interact/forms.html#adef-disabled'>Forms in HTML Documents, sec. 17.12.1, Disabled controls</a>
 */
	public static Match disableControl
	(
		final Match				Control,
		final Object			ControlValue
	)
	throws TagException
	{
		if (ControlValue == null || ControlValue.toString().equals(getControlValueLow(Control)))
			disableControl(Control);
		
		return Control;	
	}
	
	
/**	Deletes the specified control, if its value matches the specified value, or the specified value is 
 *	<code>null</code>.
 * 
 * @param ControlName			Name of the form control.
 * @param ControlValue			Value of the form control, or <code>null</code>. The value is only needed when 
 *								deleting specific items in multiple selection controls (checkboxes, radio buttons, and 
 *								"select" controls). When dealing with those controls, passing <code>null</code> in  
 *								this parameter causes all selections in the group to be deleted.
 * @return						This <code>Form</code> object.
 * @throws TagException			If a problem is encountered with the manipulation of the markup.
 */
	public Form deleteControl
	(
		final String			ControlName,
		final Object			ControlValue
	) 
	throws TagException
	{
		for (final Match Control : getControlsByName(ControlName))
			deleteControl(Control, ControlValue);
		
		return this;
	}
	
/**	Deletes the specified control, if its value matches the specified value, or the specified value is 
 *	<code>null</code>.
 * 
 * @param Control				A <code>Match</code> object corresponding to a tag of type "input", "select", 
 *								"textarea", or "button".
 * @param ControlValue			The value the control must have in order to be deleted, or <code>null</code> if it 
 *								should be deleted regardless of its value.
 * @throws TagException			If a problem is encountered with the manipulation of the markup.
 */
	public static void deleteControl
	(
		final Match				Control,
		final Object			ControlValue
	)
	throws TagException
	{
		if (ControlValue == null || ControlValue.toString().equals(getControlValueLow(Control)))
			Control.delete();
	}
	
/**	Get the values of all of the enabled, successful controls in this form. In other words, if we sent this form to a 
 *	user, and they clicked "submit" without doing anything else to the form, what control values would we receive?
 *	That's the form data set that this method should return. Well, almost. There are special cases: (1) If there's more 
 * 	than one "submit" control in a form, none of their values are included, because we can't know which one the user 
 * 	would have picked. (2) If the only "submit" control in a form is an image map (an "input" control of type "image"), 
 * 	the "x" and "y" values supplied by it are both assumed to be zero. (3) File upload controls supply no values, 
 * 	because we can't know what file the user would have chosen.
 *	
 *	@return						A form data set made up of name/value(s) pairs supplied by all of the enabled, 
 * 								successful controls in this form, except those noted in the method description.
 */
	public FormDataSet getDataSet() {
		return getDataSet(true);
	}
	
/**	Get the values of all of the enabled, successful controls in this form. In other words, if we sent this form to a 
 *	user, and they clicked "submit" without doing anything else to the form, what control values would we receive?
 *	That's the form data set that this method should return. Well, almost. File upload controls supply no values, 
 * 	because we can't know what file the user would have chosen. Also, if the <code>IncludeSolitarySubmitControl</code>
 * 	parameter is <code>false</code> the values of all submit control(s) are omitted. If the 
 * 	<code>IncludeSolitarySubmitControl</code> parameter is <code>true</code>, and there's only one submit control, its
 * 	value is included, but if there are multiple "submit" controls, all of their values are omitted, because we cannot
 *	know which one the user would have picked.
 *	
 * 	@param IncludeSolitarySubmitControl	If <code>true</code> and the form includes only one "submit" control, the value
 * 										of that control is included in the data set. If <code>false</code>, all 
 * 										"submit" contorls are ignored.
 *	@return								A form data set made up of name/value(s) pairs supplied by all of the enabled, 
 * 										successful controls in this form, except those noted in the method description.
 */
	public FormDataSet getDataSet
	(
		final boolean			IncludeSolitarySubmitControl
	) 
	{
		final FormDataSet		DataSet = new FormDataSet();
		
	//	Determine the form's method ("get" or "post"), and pass it along to the FormDataSet instance.
		
		{
			final String			Method = TheForm.getAttribute("method");
			
			if (Method != null)
				DataSet.setMethod(Method);
		}
		
	//	Determine the action URL, and pass it along to the FormDataSet instance.
	
		try {
			final URL				ActionURL = getActionURL();
			
			if (ActionURL != null)
				DataSet.setActionURL(ActionURL.toString());
		
		} catch (MalformedURLException e) {
			Log.log(Level.FINE, "Could not determine a valid action URL from a form in document \"" + TheForm.getMarkup().getMarkupName() + "\" to pass along to the data set.", e);
		}
		
	//	Determine the form's character set.
		
		{
		//	In the absence of the "accept-charset" attribute (or a valid character set name among that attribute's 
		//	values), the character set used to encode the characters of this form data set will be the same character 
		//	set used by the document in which this "form" exists.
			
			String					DataSetCharSet = TheForm.getMarkup().getCharacterSet();
			final String			AcceptCharset  = TheForm.getAttribute("accept-charset");
					
			if (AcceptCharset != null) {
				
				for (final String CharSetName : kSpaceOrCommaPat.split(AcceptCharset)) {
					
					try {
						
						Charset.forName(CharSetName);
						DataSetCharSet = CharSetName;
					
					} catch (Exception e) {
					//	If an exception was thrown, the character set is not a valid choice on this platform, so we 
					//	ignore it. 
					}
				}
			}
			
			DataSet.setCharacterSetName(DataSetCharSet);
		}
		
	//	Get the control values that populate this form data set.	
			
		for (final Match Control : get("input, button, select, textarea")) {
			
		//	Note that FormDataSet.addData automatically ignores calls that supply zero-length value arrays,
		//	which is a distinct possibility as it is used below. And we *want* such calls to be ignored, so this is 
		//	perfect.
			
			final String			CtrlName = getControlName(Control);
			
		//	Note that some controls do not have names. For example: <input type='submit' value='Guess'/>
			
			if (CtrlName != null)
				DataSet.addData(CtrlName, (Object[]) getControlValues(Control, true, false, false));
		}
		
		if (IncludeSolitarySubmitControl) {
			final FormControlInfo[]	Submits = getSubmitControls();
			
			if (Submits.length == 1) {
				final String			Name = Submits[0].getName();
			 	
			 	if (Name != null) {
			 		String[]				Values = Submits[0].getValues();
			 		
			 		if (Values == null || Values.length == 0)
			 			Values = new String[] { "" };
			 		
					DataSet.addData(Name, Values);
				}
			}
		}
		
		return DataSet;
	}
	
	
//	todo The following setDataSet() method is the only method we had to eliminate in order to sever this class' dependence on the Page class. Consider whether/how this method should be restored.
	
/*	Takes the values from the most recently submitted form data set, and writes them into this form, where applicable.
 *	Having done so, the user will not have to retype any of the values as he/she does further work with this form, like 
 *	correcting fields with invalid values.
 * 
 * @return 						This <code>Form</code> object.
 * @throws TagException
 *
	public Form setDataSet() throws TagException {
		return setDataSet(ThePage.getFormDataSet());
	}
*/
	
/*	Takes the values of the form data set present in an <code>HttpServletRequest</code>, and writes them into this form, 
 *	where applicable. Having done so, the user will not have to retype any of the values as he/she does further work 
 *	with this form, like correcting fields with invalid values.
 *
 * @param Req					The <code>HttpServletRequest</code> object containing the form data set to be written 
 *								into this form.
 * @return 						This <code>Form</code> object.
 * @throws TagException
 *
	public Form setDataSet
	(
		final HttpServletRequest	Req
	) 
	throws TagException
	{
		return setDataSet(new FormDataSet(Req));
	}
*/

/**	Takes the values of the supplied form data set, and writes them into this form, where applicable.
 *	Having done so, the user will not have to retype all the values as he/she does further work with this form, like 
 *	correcting fields with invalid values.
 *
 * @param DataSet				The form data set to be written into this form.
 * @return 						This <code>Form</code> object.
 * @throws TagException			If a problem is encountered with the manipulation of the markup.
 */
	public Form setDataSet
	(
		final FormDataSet		DataSet
	)
	throws TagException
	{
		
	//	Start with a clean slate (no values in the form).
	//
	//	Here's a tricky bit: Disabled controls in the form. Since they can't submit values, they won't be represented
	//	in WebParams, but they can hold and display values, which is a useful behavior. If we clear the value of all
	//	form controls, disabled controls will loose their values, and embedding the submitted values won't give them
	//	back their values. So, we invoke clearControlValues and pass false, to tell it not to clear the values of 
	//	disabled controls.
		
		clearControlValues(false);
		
	//	Add to the form the values provided by the current parameters.
		
		for (final FormControlValue FCV : DataSet) {
			final String			ControlName = FCV.getName();
			
			if (Qwicap.kQwicapFormIDControlName.equals(ControlName) || Qwicap.kQwicapPageIDControlName.equals(ControlName))
				continue;
			
			final String[]			ControlValues;
			
			if (FCV instanceof FormFileUpload)
				ControlValues = new String[] { ((FormFileUpload) FCV).getFileItem().getName() };
			else
				ControlValues = FCV.getValues();
			
		//	Embed the values.
			
			final Results			Controls = getControlsByName(ControlName);
			
			if (Controls.isEmpty())
				continue;
			
			if (isSingleValueControl(Controls)) {
				
			//	We are dealing with a type of control that can have only one value. If there are multiple values, it 
			//	must be because there are several of these single-value controls with the same name.
				
				for (int ControlIndex = 0, ControlCount = Controls.size(), ValueIndex = 0, ValueCount = ControlValues.length; ControlIndex < ControlCount; ControlIndex++, ValueIndex++) {
					final Match				Control = Controls.get(ControlIndex);
					
					if (isControlValueUserModifiable(Control)) {
						
						if (ValueIndex < ValueCount)
							setControlValueLow(Control, ControlValues[ValueIndex]);
						else
							clearControlValue(Control);	//	We ran out of values before we ran out of controls, so clear the values of the remaining controls.
					}
				}
				
			} else {
				
			//	We are dealing with a type of control that can have multiple values, so, if we have multiple values for
			//	the control, that won't be at all surprising.
				
				for (final Match Control : Controls) {
					
					if (isControlValueUserModifiable(Control)) {
						
						for (final String ControlValue : ControlValues)
							setControlValueLow(Control, ControlValue);
					}
				}
			}
		}
		
		return this;
	}
	
	private static boolean isSingleValueControl
	(
		final Results			Control
	)
	{
		final Match				Ctrl = Control.get(0);
		final String			Type = Ctrl.getType();
		
		if ("input".equals(Type)) {
			final String			InputType = Ctrl.getAttribute("type");
			
			return "text".equals(InputType) || "password".equals(InputType) || "hidden".equals(InputType) || "submit".equals(InputType) || "image".equals(InputType);
			
		} else if ("button".equals(Type)) {
			
			return true;
			
		} else if ("textarea".equals(Type)) {
		
			return true;
			
		} else if ("select".equals(Type)) {
			
			return Ctrl.hasAttribute("multiple") == false;
		}
		
		throw new IllegalArgumentException("Unknown control type: \"" + Type + "\". The known control types are: \"input\", \"button\", \"textarea\" and \"select\" (which includes \"option\").");
	}
	
	private static boolean isControlValueUserModifiable
	(
		final Match				Ctrl
	)
	{
		final String			Type = Ctrl.getType();
		
		if ("input".equals(Type)) {
			final String			InputType = Ctrl.getAttribute("type");
			
		//	Controls are only considered "user modifiable" if (1) the user can choose the control's value, and (2) the 
		//	value submitted from a form can be embedded into the control markup. Thus "file" inputs are not considered 
		//	user modifiable, because the contents of the file cannot be embedded into the control markup. Same thing 
		//	with "image" inputs.
			
			return "text".equals(InputType) || "password".equals(InputType) || "checkbox".equals(InputType) || "radio".equals(InputType);
			
		} else if ("button".equals(Type)) {
			
			return false;
		
		} else if ("textarea".equals(Type)) {
			
			return true;
			
		} else if ("select".equals(Type)) {
			
			return true;
		}
		
		throw new IllegalArgumentException("Unknown control type: \"" + Type + "\". The known control types are: \"input\", \"button\", \"textarea\" and \"select\" (which includes \"option\").");
	}
		
/**	Removes the values from the controls in this form whose values can be altered by the user. The values of buttons, 
 *	and hidden input controls are not affected, and the values of disabled controls of all types may be spared, if you 
 *	wish.	
 * 
 * @param AlsoClearDisabledControls		If <code>true</code>, the values of disabled controls of any type are also 
 *										cleared. Otherwise, the values of disabled controls are left as they are.
 * @return								This <code>Form</code> object.
 * @throws TagException					If a problem is encountered with the manipulation of the markup.
 */	
	public Form clearControlValues
	(
		final boolean			AlsoClearDisabledControls
	)
	throws TagException
	{
		clearInputControls_TextPasswordFile(AlsoClearDisabledControls);
		clearInputControls_RadioCheckbox(AlsoClearDisabledControls);
		clearSelectControls(AlsoClearDisabledControls);
		clearTextAreaControls(AlsoClearDisabledControls);
		
		return this;
	}
	
	private void clearInputControls_TextPasswordFile
	(
		final boolean			AlsoClearDisabledControls
	)
	throws TagException
	{
		final Results			Controls = get("input[type='text'][value], input[type='password'][value], input[type='file'][value]");
		
		if (AlsoClearDisabledControls) {
			
			Controls.deleteAttribute("value");
			
		} else {
			
			for (final Match Control : Controls) {
				
				if (isControlEnabled(Control))
					Control.deleteAttribute("value");
			}	
		}
	}
	
	private void clearInputControls_RadioCheckbox
	(
		final boolean			AlsoClearDisabledControls
	)
	throws TagException
	{
		final Results			Controls = get("input[type='radio'][checked], input[type='checkbox'][checked]");
		
		if (AlsoClearDisabledControls) {
			
			Controls.deleteAttribute("checked");
			
		} else {
			
			for (final Match Control : Controls) {
				
				if (isControlEnabled(Control))
					Control.deleteAttribute("checked");
			}	
		}
	}
	
	private void clearSelectControls
	(
		final boolean			AlsoClearDisabledControls
	)
	throws TagException
	{
		final Results			Controls = get("select option[selected]");
		
		if (AlsoClearDisabledControls) {
			
			Controls.deleteAttribute("selected");
			
		} else {
			
			for (final Match Control : Controls) {
				
				if (isControlEnabled(Control))
					Control.deleteAttribute("selected");
			}	
		}
	}
	
	private void clearTextAreaControls
	(
		final boolean			AlsoClearDisabledControls
	)
	throws TagException
	{
		final Results			Controls = get("textarea");
		
		if (AlsoClearDisabledControls) {
			
			Controls.deleteContent();
			
		} else {
			
			for (final Match Control : Controls) {
				
				if (isControlEnabled(Control))
					Controls.deleteContent();
			}	
		}
	}
	
/**	Removes the value of the specified form control.
 * 
 * @param ControlName			Name of the form control.
 * @return 						This <code>Form</code> object.
 * @throws TagException			If a problem is encountered with the manipulation of the markup.
 */
	public Form clearControlValue
	(
		final String			ControlName
	)
	throws TagException
	{
		for (final Match Elem : getControlsByName(ControlName))
			clearControlValue(Elem);
		
		return this;
	}

/**	Clears the value of the supplied from control, even if it is of a type whose value cannot be altered by the user.
 * 
 * @param Control				A <code>Match</code> object corresponding to a tag of type "input", "select", 
 *								"textarea", or "button".
 * @throws TagException			If a problem is encountered with the manipulation of the markup.
 */	
	public static void clearControlValue
	(
		final Match				Control
	)
	throws TagException
	{
		final String			ControlType = Control.getType();
		
		if ("input".equals(ControlType)) {
			final String			Subtype = Control.getAttribute("type");
			
			if ("text".equals(Subtype) || "password".equals(Subtype) || "hidden".equals(Subtype)) {
				
				Control.deleteAttribute("value");
				
			} else if ("checkbox".equals(Subtype) || "radio".equals(Subtype)) {
				
				Control.deleteAttribute("checked");
				
			} else if ("submit".equals(Subtype)) {
				
				Control.setAttribute("name", new InputSubmitButton(Control).setValue("").toControlNameString());
				
			} else if ("file".equals(Subtype)) {
				
			//	The value of "input" controls of type "file" can't be determined server-side, because the value of the
			//	control is the content of the file, if any, that the user selects. However, such controls can have an
			//	optional "value" attribute which specifies the default name of the file on the client machine that is 
			//	to be uploaded.
				
				Control.deleteAttribute("value");
				
			} else if ("image".equals(Subtype) || "reset".equals(Subtype) || "button".equals(Subtype)) {
				
			//	"image" -- There's no user to click on this, so it has no values.
			//	"reset" -- Never supplies a value.
			//	"button" -- Has no defined behavior. (Used to trigger client-side scripts.) Never successful.
				
			} else
				throw new IllegalArgumentException("The supplied Match does not represent a known type of \"input\" control, it represents \"" + Control + "\".");
			
		} else if ("select".equals(ControlType)) {
			
			Control.get("option[selected]").deleteAttribute("selected");
			
		} else if ("option".equals(ControlType)) {	//	Always found inside a "select" control.
			
		//	The value of an "option" control is set in one of two ways: (1) By a "value" attribute, or (2) by the text 
		//	content of the option control.
			
			Control.deleteAttribute("selected");
			
		} else if ("textarea".equals(ControlType)) {
			
			Control.deleteContent();
			
		} else if ("button".equals(ControlType)) {
			
		//	The "value" attribute is optional. If the button has a "name" attribute, but no "value" attribute, a blank 
		//	string is, in effect, submitted. If it has no "name" attribute, no name/value pair is added to the form data
		//	set.
			
			Control.deleteAttribute("value");
			
		} else
			throw new IllegalArgumentException("The supplied Match does not represent a known form control element ('input', 'select', 'option', 'button' or 'textarea'), it represents \"" + Control + "\".");
	}
	
/**	Returns the name of the specified control, or <code>null</code> if the control has no name.
 * 
 * @param Control				A <code>Match</code> object corresponding to a tag of type "input", "select", 
 *								"textarea", or "button".
 * @return						The name of the specified control, or <code>null</code> if the control has no name.
 */
	public static String getControlName
	(
		final Match				Control
	)
	{
		final String			Type = Control.getType();	//	input, select, button or textarea
		final String			ControlName;
		
		if ("input".equals(Type)) {
			final String			TypeAttr = Control.getAttribute("type");
			
		//	Note that not all "submit" buttons have names. For example: <input type='submit' value='Guess'/>
			
			if ("submit".equals(TypeAttr))
				ControlName = new InputSubmitButton(Control).getName();
			else
				ControlName = Control.getAttribute("name");
			
		} else if ("select".equals(Type) || "button".equals(Type) || "textarea".equals(Type)) {
			
			ControlName = Control.getAttribute("name");
			
		} else if ("option".equals(Type)) {
			
			ControlName = getControlName(Control.getAncestorOfType("select"));
			
		} else
			throw new IllegalArgumentException("The supplied Match does not represent a form control element ('input', 'select', 'button' or 'textarea'), it represents \"" + Control + "\".");
		
		return ControlName;
	}
	
/**	Returns the current value of the named control as a <code>String</code>. If the control has more than one value,
 * 	this method will only return the first; use {@link #getControlValues(String)} for multi-value controls. If the 
 *	control has no value, or would not be <a href='http://www.w3.org/TR/html4/interact/forms.html#successful-controls'>successful</a>, 
 *	<code>null</code> is returned.
 * 
 * @param ControlName			The name of the control whose value should be retrieved.
 * @return						The value of the control, or <code>null</code> if the control does not exist, has no 
 *								value, or would be unsuccessful.
 */
	public String getControlValue
	(
		final String			ControlName
	)
	{
		final String[]			Values = getControlValues(ControlName);
		
		return Values.length == 0 ? null : Values[0];
	}
	
/**	Returns the current value(s) of the named control as a <code>String</code> array. If the control has no 
 *	values (maybe it is a group of radio buttons and none are selected), a zero-length array is returned. The exceptions
 *	to that rule are "input" controls of type "text" and "password", because without a value they will submit an empty
 *	string, so this method returns an empty string in their case. In other words, this method differentiates between
 *	<a href='http://www.w3.org/TR/html4/interact/forms.html#successful-controls'>successful</a> and unsuccessful 
 *	controls; returning <code>null</code> for a control that would be unsuccessful as it stands; returning an empty 
 *	array for a control that would be successful, but has no value; and returning an array of one or more values for a
 *	control that would be successful and has one or more values.
 * 
 * @param ControlName			The name of the control whose value(s) should be retrieved.
 * @return						The value(s) of the control, or <code>null</code> if the control does not exist, or
 *								would be unsuccessful. 
 */
	public String[] getControlValues
	(
		final String			ControlName
	)
	{
		final Results			Controls     = getControlsByName(TheForm, ControlName);
		final int				ControlCount = Controls.size();
		
		if (ControlCount == 1) {
			
			return getControlValues(Controls.get(0));	//	Consider the misnomer value situation.
			
		} else if (ControlCount > 1) {
			final ArrayList<String>	AllValues = new ArrayList<String>();
			
			for (final Match Control : Controls)
				AllValues.addAll(Arrays.asList(getControlValues(Control)));
			
			return AllValues.toArray(new String[AllValues.size()]);
		}
		
		return null;
	}
	
/**	Returns the value of the specified control as a <code>String</code>. If the control has more than one value,
 * 	this method will only return the first; use {@link #getControlValues(Match)} for multi-value controls. (In the 
 *	context of this method, which can only operate on an individual control element, only a "select" control can have 
 *	multiple values. To avoid that limitation, use {@link #getControlValue(String)}.) If the control has no value (maybe 
 *	it is a "select" control with none of its "option" elements selected), <code>null</code> is returned. The exceptions
 *	to that rule are "input" controls of type "text" and "password", because without a value they will submit an empty
 *	string, so this method returns an empty string in their case. In other words, this method differentiates between
 *	<a href='http://www.w3.org/TR/html4/interact/forms.html#successful-controls'>successful</a> and unsuccessful 
 *	controls; returning <code>null</code> for a control that would be unsuccessful as it stands; returning an empty 
 *	array for a control that would be successful, but has no value; and returning an array of one or more values for a
 *	control that would be successful and has one or more values.
 * 
 * @param Control					A <code>Match</code> object corresponding to a tag of type "input", "select", 
 *									"option", "textarea", or "button".
 * @return							The value of the control, or <code>null</code> if the control does not exist, or 
 *									has no value.
 */
	public static String getControlValue
	(
		final Match				Control	
	)
	{
		final String[]			Values = getControlValues(Control);
		
		return Values.length == 0 ? null : Values[0];
	}
	
/**	Returns the value of the specified control as a <code>String</code>. If the control has more than one value,
 * 	this method will only return the first; use {@link #getControlValues(Match)} for multi-value controls. (In the 
 *	context of this method, which can only operate on an individual control element, only a "select" control can have 
 *	multiple values. To avoid that limitation, use {@link #getControlValue(String)}.) If the control has no value (maybe 
 *	it is a "select" control with none of its "option" elements selected), <code>null</code> is returned. The exceptions
 *	to that rule are "input" controls of type "text" and "password", because without a value they will submit an empty
 *	string, so this method returns an empty string in their case. In other words, this method differentiates between
 *	<a href='http://www.w3.org/TR/html4/interact/forms.html#successful-controls'>successful</a> and unsuccessful 
 *	controls; returning <code>null</code> for a control that would be unsuccessful as it stands; returning an empty 
 *	array for a control that would be successful, but has no value; and returning an array of one or more values for a
 *	control that would be successful and has one or more values.
 * 
 * @param Control					A <code>Match</code> object corresponding to a tag of type "input", "select", 
 *									"option", "textarea", or "button".
 * @return							The value(s) of the control, or <code>null</code> if the control would be 
 *									unsuccessful. 
 */
	public static String[] getControlValues
	(
		final Match				Control
	)
	{
		return getControlValues(Control, true, true, false);
	}
	
/**	Returns the value(s) of the specified control, optionally limiting the values to those supplied by the controls
 *	that will be successful by default. If the specified control is an "input" control of type "checkbox" or "radio", 
 *	or an "option" control, only the value of the specified control is considered; other controls in the same group
 *	(of the same name) are not considered. However, if the specified control is a "select" control, all of the "option"
 *	controls it contains <i>are</i> considered. Which means that, if you want to build-up a complete picture of the 
 *	values that will be supplied by a form, you get all controls of type "input", "select", "textarea" and "button" and 
 *	feed them to this method. Ordinarily, your would <i>not</i> bother to ask this method about controls of type 
 *	"option", since they are dealt with when this method is called for their enclosing "select" controls.
 * 
 * @param Control					A <code>Match</code> object corresponding to a tag of type "input", "select", 
 *									"option", "textarea", or "button".
 * @param SuccessfulControlsOnly	If <code>true</code>, only controls that would supply values (be successful) if a 
 *									user were to submit the form as-is are allowed to contribute values to the list of
 *									values that is returned. If <code>false</code>, all controls supply their values.
 * @param IncludeSubmitControls		If <code>true</code>, the value(s) of submit button(s) (either "input" or "button"
 *									controls of type "submit"), are included in the list of values that is returned. 
 *									If <code>false</code> they are omitted.
 * @param IncludeMisnomerValues		If <code>true</code>, the values of "input" controls of type "file" are returned.
 *									The real value of such controls can't be determined server-side, because the real
 *									value of the control is the content of the file, if any, that the user selects. 
 *									However, such controls do have a "value" attribute which specifies the default name 
 *									of the file on the client machine that is to be uploaded. For lack of a better name, 
 *									I call that default file name in the "value" attribute a "misnomer value".
 * @return							An array of <code>String</code>s, containing zero or more elements, representing
 *									the value(s) of the specified control. If the control has no value, the array is
 *									empty.
 */
	public static String[] getControlValues
	(
		final Match				Control,
		final boolean			SuccessfulControlsOnly,		//	Well, controls that would be successful if the form were submitted as-is.
		final boolean			IncludeSubmitControls,
		final boolean			IncludeMisnomerValues		//	The value of "input" controls of type "file".
	)
	{
		if (SuccessfulControlsOnly && isControlEnabled(Control) == false)
			return kEmptyStringArray;
		
		final String			ControlType = Control.getType();
		
		if ("input".equals(ControlType)) {
			final String			Subtype = Control.getAttribute("type");
			
			if ("text".equals(Subtype) || "password".equals(Subtype)) {
				
			//	All of these controls are always successful. The absence of a default value simply means their value 
			//	is an empty string, since these are always successful and an empty string is, in effect, what would be 
			//	submitted by an empty text/password control.
				
				return new String[] { Null.toEmptyString(Control.getAttribute("value")) };
				
			} else if ("hidden".equals(Subtype)) {
				final String			Value = Control.getAttribute("value");
				
			//	The "value" attribute is optional here, although the control is utterly useless without one.
				
			//	Section 17.13.2 of the HTML 4 specification: "If a control doesn't have a current value when the form 
			//	is submitted, user agents are not required to treat it as a successful control."
				
				if (Value != null)
					return new String[] { Value };	//	If we have a value, we're successful.
				else
					return kEmptyStringArray;
				
			} else if ("checkbox".equals(Subtype) || "radio".equals(Subtype)) {
				
			//	These controls must have a "value" attribute. Section 17.4 of the HTML 4 spec. states: "[The value] 
			//	attribute specifies the initial value of the control. It is optional except when the type attribute 
			//	has the value 'radio' or 'checkbox'."
				
				if (SuccessfulControlsOnly) {	
					
				//	If this control doesn't have a "checked" attribute, it won't be successful and therefore doesn't 
				//	have a value.
					
					if (Control.hasAttribute("checked"))
						return new String[] { Control.getAttribute("value") };	//	"value" is not optional.
					else
						return kEmptyStringArray;
					
				} else {
					
				//	Return the value of this control whether it has a "checked" attribute, or not.
					
					return new String[] { Control.getAttribute("value") };		//	"value" is not optional.
				}
				
			} else if ("submit".equals(Subtype)) {
				
			//	The "value" attribute is optional here
				
				if (IncludeSubmitControls)
					return new String[] { new InputSubmitButton(Control).getValue() };
				else
					return kEmptyStringArray;
				
			} else if ("file".equals(Subtype)) {
				
			//	The value of "input" controls of type "file" can't be determined server-side, because the value of the
			//	control is the content of the file, if any, that the user selects. However, such controls can have an
			//	optional "value" attribute which specifies the default name of the file on the client machine that is 
			//	to be uploaded. For lack of a better name, I'll call that default file name a "misnomer value". Yes,
			//	it lives in the standard "value" attribute, but, no, it's not the value that the control will supply.
				
				if (IncludeMisnomerValues) {
					final String			Value = Control.getAttribute("value");
					
					if (Value != null)
						return new String[] { Value };
				}
				
				return kEmptyStringArray;
				
			} else if ("image".equals(Subtype)) {
			
			//	"image" -- There's no user to click on this, so we don't know exactly what values (the X and Y 
			//	coordinates) to supply, but, if we've been asked to include submit buttons, we go ahead and supply the
			//	coordinates (0, 0) on the grounds that any image should have at least one pixel.
				
				if (IncludeSubmitControls)
					return new String[] { "0", "0" };
				else
					return kEmptyStringArray;
					
			} else if ("reset".equals(Subtype) || "button".equals(Subtype)) {
				
			//	"reset" -- Never supplies a value.
			//	"button" -- Has no defined behavior. (Used to trigger client-side scripts.) Never successful.
			//	"file" -- What the hell do we do with these? They have a value (a default file name), but the real value
			//			  submitted by this element is an actual file. For the current purpose, the only option seems to
			//			  be to treat them as if they have no value.
				
				return kEmptyStringArray;	
			}
			
		} else if ("select".equals(ControlType)) {
			final Results			Options;
			
			if (SuccessfulControlsOnly) {
				final Results			Selected = Control.get("option[selected]");
				
				if (Selected.notEmpty())
					Options = Selected;
				else
					Options = Control.get("option").first();	//	In the absence of selected "option" elements, the first "option" element is implicitly selected, and therefore "successful".
			} else
				Options = Control.get("option");
			
			if (Control.hasAttribute("multiple")) {
				final String[]			Values = new String[Options.size()];
				int						Index  = 0;
				
				for (final Match Option : Options)
					Values[Index++] = getTrimmedOptionValue(Option);
				
				return Values;
				
			} else {
				
				if (Options.notEmpty())
					return new String[] { getTrimmedOptionValue(Options.get(0)) };
				else
					return kEmptyStringArray;
			}
			
		} else if ("option".equals(ControlType)) {	//	Always found inside a "select" control.
			
		//	The value of an "option" control is set in one of two ways: (1) By a "value" attribute, or (2) by the text 
		//	content of the option control.
			
			if (SuccessfulControlsOnly) {
				
				if (Control.hasAttribute("selected"))
					return new String[] { getTrimmedOptionValue(Control) };
				else
					return kEmptyStringArray;
			} else
				return new String[] { getTrimmedOptionValue(Control) };
			
		} else if ("textarea".equals(ControlType)) {
			
			return new String[] { Control.getText() };		//	Should this be trimmed? No.
		
		} else if ("button".equals(ControlType)) {
			
			if (IncludeSubmitControls && "submit".equals(Control.getAttribute("type"))) {
			
			//	Without a name, the button does not supply a name/value pair in the form data set, and therefore has 
			//	no value, even if it has a "value" attribute.
				
				if (Control.hasNonEmptyAttribute("name")) {
					final String			Value = Control.getAttribute("value");
					
				//	Judging from experience, and the W3C validator when operating on "XHTML 1.0 Strict", the "value" 
				//	attribute is optional.
					
					if (Value != null)
						return new String[] { Value };
				}
			}
			
			return kEmptyStringArray;
		}
		
		throw new IllegalArgumentException("The supplied Match does not represent a known form control element ('input', 'select', 'option', 'button' or 'textarea'), it represents \"" + Control + "\".");
	}
	
	private static String getTrimmedOptionValue
	(
		final Match				Option
	)
	{
		String					Value;
		
	//	The value of an "option" element can be represented as a "value" attribute, or as the text contained between
	//	a pair of "<option>" and "</option>" tags. The "value" attribute takes precedence.
		
		Value = Option.getAttribute("value");
		if (Value == null)
			Value = Option.getText().trim();	//	Will never be null.  Should this be trimmed?
		
		return Value;
	}
	
/**	Returns the value of an "input", "option", "textarea", or "button" control. This method does not concern
 *	itself with whether or not "option", checkbox and radio controls are actually marked as selected; if asked for
 *	their value, this method returns it.
 * 
 * @param Control			The control to be examined.
 * @return					Value of the control, or <code>null</code>.
 */	
	private static String getControlValueLow
	(
		final Match				Control
	)
	{
		final String			Value;
		final String			ElementType = Control.getType();
		
		if ("input".equals(ElementType)) {
			
			if ("submit".equals(Control.getAttribute("type"))) {
				
				Value = new InputSubmitButton(Control).getValue();
				
			} else
				Value = Control.getAttribute("value");
		
		} else if ("option".equals(ElementType)) {	//	Always found inside a "select" control.
			
			Value = getTrimmedOptionValue(Control);
			
		} else if ("textarea".equals(ElementType)) {
			
			Value = Control.getText();				//	Should this be trimmed?
		
		} else if ("button".equals(ElementType)) {
			
			Value = Control.getAttribute("value");
			
		} else
			throw new IllegalArgumentException("The supplied Match does not represent a supported form control element ('input', 'option', 'button' or 'textarea'), it represents \"" + Control + "\".");

		return Value;
	}

	
/*	Returns the value of an "input", "option", "textarea", "button" or "select" control. In the case of "select" 
 *	controls that can have multiple values, only the first potential value is returned. This method does not concern
 *	itself with whether or not "option", checkbox and radio controls are actually marked as selected; if asked for
 *	their value, this method returns it. In the case of a "select" control, the values of all of its subsidiary 
 *	"option" controls are acquired, but only the first is returned, which makes this method pretty useless where 
 *	"select" controls are concerned.
 * 
 * @param Control			The control to be examined.
 * @return					Value of the control, or <code>null</code>.
 *
	static String getControlValueLow
	(
		final Match				Control
	)
	{
		final String[]			Values = getControlValues(Control, false, true, true);
		
		return Values.length == 0 ? null : Values[0];
	}
*/
	
/**	Add to the value of the specified form control. For controls that can only have one value, this sets the value of 
 *	the control. For multi-value controls like radio buttons and checkboxes, if the supplied value matches the 
 *	"value" attribute of the control, the radio button or checkbox is checked. If not, nothing happens. Essentially the 
 *	same thing happens with the "option" controls within "select" blocks.
 * 
 * @param FormElement			Reference to an "input", "select", "textarea" or "button" control.
 * @param Value					Value to be assigned to the control.
 * @return						A Match object supplying a fresh reference to the form control. (If the control was 
 *								converted from immutable to mutable as a result of having its value set, the original 
 *								Match object becomes invalid. Or used to; this isn't a problem anymore.)
 * @throws TagException			If a problem is encountered with the manipulation of the markup.
 * @see #clearControlValue
 */
	private static Match setControlValueLow
	(
		final Match				FormElement,	//	input, select or textarea
		final String			Value
	) 
	throws TagException
	{
		final String			ElementType = FormElement.getType();
		
		if ("input".equals(ElementType)) {
			
			setInputControlValue(FormElement, Value);	//	Ultimately, setAttribute calls are made, and they encode the value.
			
		} else if ("select".equals(ElementType)) {
			
			setSelectControlValue(FormElement, Value);	//	value encoding isn't relevant, since the value isn't inserted into the markup. It is important that Value NOT be encoded, however.
			
		} else if ("textarea".equals(ElementType)) {
			
			FormElement.setContent(Value);				//	setContent encodes the value.
		
		} else if ("button".equals(ElementType)) {
			
			FormElement.setAttribute("value", Value);	//	setAttribute encodes the value.
		}
		
		return FormElement;
	}
	
/**	Sets the value of an "input" form control. For radio buttons and checkboxes, if the supplied value matches the 
 *	"value" attribute of the control, the radio button or checkbox is checked. If not, nothing happens.
 * 
 * @param Elem					A reference to an "input" form control.
 * @param Value					The new value for that control, if applicable.
 * @return						A Match object supplying a fresh reference to the form control. (If the control was 
 *								converted from immutable to mutable as a result of having its value set, the original 
 *								Match object becomes invalid.)
 * @throws TagException			If a problem is encountered with the manipulation of the markup.
 */
	private static Match setInputControlValue
	(
		final Match				Elem,
		final String			Value
	)
	throws TagException
	{
		final String			Type = Elem.getAttribute("type");
		
	//	text|password|checkbox|radio|submit|reset|file|hidden|image|button
		
		if ("text".equals(Type) || "password".equals(Type) || "hidden".equals(Type) || "file".equals(Type) || "button".equals(Type)) {
			
			Elem.setAttribute("value", Value);
			
		} else if ("checkbox".equals(Type) || "radio".equals(Type)) {
			
			if (Value.equals(Elem.getAttribute("value")))
				Elem.setAttribute("checked", "checked");
			
		} else if ("submit".equals(Type)) {
			
		//	An input control of type "submit" can't supply a normal name/value pair as a parameter, because the HTML
		//	specification conflated the human-interface name (placed in the "value" attribute of the control), with
		//	the client/server-interface name. Thus, if a "name" attribute is present, the submit button sends its 
		//	human interface name as the value of the relevant parameter. So, if you want to maintain the name of the 
		//	button independently of the code that processes its input, you can't. Nor can you have the button supply a
		//	value that you wouldn't want the user to see. In short, that part of the HTML spec. is incredibly badly 
		//	designed. Here we use the InputSubmitButton class to get around the aformentioned problems.
			
			Elem.setAttribute("name", new InputSubmitButton(Elem).setValue(Value).toControlNameString());
		}
		
		return Elem;
	}
	
/**	Adds to the value of an "input" form control. For radio buttons and checkboxes, if the supplied value matches the 
 *	"value" attribute of the control, the radio button or checkbox is checked. If not, nothing happens. For other 
 *	types of "input" control, the new value is appended to the existing value.
 * 
 * @param Control				A reference to an "input" form control.
 * @param Value					The new value for that control, if applicable.
 * @return						A <code>Match</code> object supplying a fresh reference to the form control. (If the 
 *								control was converted from immutable to mutable as a result of having its value set, 
 *								the original <code>Match</code> object becomes invalid.)
 * @throws TagException			If a problem is encountered with the manipulation of the markup.
 */
	private static Match addToInputControlValue
	(
		final Match				Control,
		final String			Value
	)
	throws TagException
	{
		final String			Type = Control.getAttribute("type");
		
	//	text|password|checkbox|radio|submit|reset|file|hidden|image|button
		
		if ("text".equals(Type) || "password".equals(Type) || "hidden".equals(Type) || "file".equals(Type) /* || "button".equals(Type) */) {
			final String			OrigValue = Control.getAttribute("value");
			
			if (OrigValue != null)
				Control.setAttribute("value", OrigValue + Value);
			else
				Control.setAttribute("value", Value);
			
		} else if ("checkbox".equals(Type)) {
			
			if (Value.equals(Control.getAttribute("value")))
				Control.setAttribute("checked", "checked");
			
		} else if ("radio".equals(Type)) {
			
			if (Value.equals(Control.getAttribute("value")))
				Control.setAttribute("checked", "checked");
			else
				Control.deleteAttribute("checked");
			
		} else if ("submit".equals(Type)) {
			
		//	An input control of type "submit" can't supply a normal name/value pair as a parameter, because the HTML
		//	specification conflated the human-interface name (placed in the "value" attribute of the control), with
		//	the client/server-interface name. Thus, if a "name" attribute is present, the submit button sends its 
		//	human interface name as the value of the relevant parameter. So, if you want to maintain the name of the 
		//	button independently of the code that processes its input, you can't. Nor can you have the button supply a
		//	value that you wouldn't want the user to see. In short, that part of the HTML spec. is incredibly badly 
		//	designed. Here we use the InputSubmitButton class to get around the aformentioned problems.
			
			final InputSubmitButton		ISB = new InputSubmitButton(Control);
			final String				OrigValue = ISB.getValue();
			
			if (OrigValue != null)
				Control.setAttribute("name", ISB.setValue(OrigValue + Value).toControlNameString());
			else
				Control.setAttribute("name", ISB.setValue(Value).toControlNameString());
		}
		
		return Control;
	}

/**	Sets the value of a "select" form control. The "option" controls that match the supplied value are maked as 
 *	"selected".
 * 
 * @param Elem					A reference to a "select" form control.
 * @param Value					A new value for the control, if applicable.
 * @return						A Match object supplying a fresh reference to the form control. (If the control was 
 *								converted from immutable to mutable as a result of having its value set, the original 
 *								Match object becomes invalid.)
 * @throws TagException			If a problem is encountered with the manipulation of the markup.
 */
	private static Match setSelectControlValue
	(
		final Match				Elem,
		final String			Value
	)
	throws TagException
	{
	//	final boolean			Multiple = Elem.hasAttribute("multiple");
	//	
	//	If this "select" control supports only one selection at a time, clear all of the selections, before
	//	selecting the relevant option. Strictly speaking, this should be unnecessary due to the action of the 
	//	clearForm method that we haven't written yet.
	//	
	//	if (!Multiple)
	//		Elem.get("option").deleteAttribute("selected");
		
	//	The value of "option" controls may by specified by a "value" attribute, or by the content bounded by the option
	//	control (for example: "<option>Fred Was Here</option>").
		
		for (final Match Option : Elem.get("option"))
			setOptionControlValue(Option, Value);
		
		return Elem;
	}
	
	private static Match setOptionControlValue
	(
		final Match				Elem,	//	Must refer to an "option" control.
		final String			Value
	)
	throws TagException
	{
		String					OptionValue = Elem.getAttribute("value");
		
		if (OptionValue == null)
			OptionValue = Elem.getContent().getText();
		
		if (Value.equals(OptionValue))
			Elem.setAttribute("selected", "selected");
		
		return Elem;
	}
	
/**	Returns <code>true</code> if the specified control is enabled, <code>false</code> otherwise.
 * 
 * @param Control				A <code>Match</code> instance that refers to an "input", "select", "textarea" or 
 *								"button" control.
 * @return						<code>true</code> if the specified control is enabled, <code>false</code> otherwise.
 */
	public static boolean isControlEnabled
	(
		final Match				Control
	)
	{
		return Control.hasAttribute("disabled") == false;
	}
	
/**	Disables the specified control.
 * 
 * @param Control				A <code>Match</code> instance that refers to an "input", "select", "textarea", 
 *								"button", "option", or "optgroup" tag.
 * @return						The specified control.
 * @throws TagException			If a problem is encountered with the manipulation of the markup.
 */
	public static Match disableControl
	(
		final Match				Control		//	input, select or textarea
	) 
	throws TagException
	{
		final String			ElementType = Control.getType();
		
	//	These are the only valid targets. See "Forms in HTML Documents", sec. 17.12.1, <http://www.w3.org/TR/html4/interact/forms.html#adef-disabled>.
		
		if ("input".equals(ElementType)
		||  "button".equals(ElementType)
		||  "optgroup".equals(ElementType)
		||  "option".equals(ElementType)
		||  "select".equals(ElementType)
		||  "textarea".equals(ElementType)) {
			
			Control.setAttribute("disabled", "disabled");
		}
		
		return Control;
	}
	
/**	Enables the specified control.
 * 
 * @param Control				A <code>Match</code> instance that refers to an "input", "select", "textarea", 
 *								"button", "option", or "optgroup" tag.
 * @return						The specified control.
 * @throws TagException			If a problem is encountered with the manipulation of the markup.
 */
	public static Match enableControl
	(
		final Match				Control		//	input, select or textarea
	) 
	throws TagException
	{
		final String			ElementType = Control.getType();
		
	//	These are the only valid targets. See "Forms in HTML Documents", sec. 17.12.1, <http://www.w3.org/TR/html4/interact/forms.html#adef-disabled>.
	
		if ("input".equals(ElementType)
		||  "button".equals(ElementType)
		||  "optgroup".equals(ElementType)
		||  "option".equals(ElementType)
		||  "select".equals(ElementType)
		||  "textarea".equals(ElementType)) {
			
			Control.deleteAttribute("disabled");
		}
		
		return Control;
	}
	
/**	Converts all "button" controls of type "submit" to "input" controls of type "submit", in order to work around a 
 * 	hideous bug in Internet Explorer (Windows and Mac OS).	
 * 
 * @return						The number of controls converted.
 * @throws TagException			If a problem is encountered with the manipulation of the markup.
 */
	public int convertSubmitButtonsToInputs() throws TagException {
		
	//	@todo Consider: Does Internet Explorer screw-up only when there's more than one submit button in a form, or 
	//	does it also screw-up forms that only contain a single submit button? If single-submit-button forms are OK,
	//	we could leave them alone.
		
		final XMLString			InputSubmitMarkup = new XMLString("<input type='submit'/>");
		int						Conversions       = 0;
		
		for (final Match Submit : get("button[type='submit']")) {
			final InputSubmitButton	ISB       = new InputSubmitButton(Submit.getAttribute("name"), Submit.getAttribute("value"));
			final String			HIName    = Submit.getText();
			final String			ID        = Submit.getAttribute("id");
			final String			Classes   = Submit.getAttribute("class");
			final Match				NewSubmit = Submit.replace(InputSubmitMarkup);
			
			NewSubmit.setAttribute("value", HIName);
			NewSubmit.setAttribute("name",  ISB.toControlNameString());
			
		//	If they were present, preserve the 'id' and 'class' attributes of the original "button" control.
			
			if (ID != null)
				NewSubmit.setAttribute("id", ID);
			if (Classes != null)
				NewSubmit.setAttribute("class", Classes);
			
			Conversions++;
		}
		
		return Conversions;
	}
	
/**	Converts all "input" controls of type "submit" that were "button" controls before 
 *	{@link #convertSubmitButtonsToInputs} was invoked back into "button" controls of type "submit". In other words, 
 *	this method undoes what {@link #convertSubmitButtonsToInputs} does.
 * 
 * @return						The number of controls converted.
 * @throws TagException			If a problem is encountered with the manipulation of the markup.
 */
	public int convertSubmitInputsToButtons() throws TagException {
		final XMLString			ButtonMarkup = new XMLString("<button type='submit'>Name</button>");
		int						Conversions  = 0;
		
		for (final Match Input : get("input[type='submit']")) {
			final String			Name = Input.getAttribute("name");
			
			if (Name != null && Name.startsWith(Qwicap.kQwicapInputSubmitControlName)) {
				final InputSubmitButton	ISB       = new InputSubmitButton(Input);
				final String			ID        = Input.getAttribute("id");
				final String			Classes   = Input.getAttribute("class");
				final Match				NewSubmit = Input.replace(ButtonMarkup);
				
				NewSubmit.setAttribute("name",  ISB.getName());
				NewSubmit.setAttribute("value", ISB.getValue());
				NewSubmit.setContent(ISB.getHIName());
				
				if (ID != null)
					NewSubmit.setAttribute("id", ID);
				if (Classes != null)
					NewSubmit.setAttribute("class", Classes);
				
				Conversions++;
			}
		}
		
		return Conversions;
	}
}
