/*	QWICAP (Quick Web Interface for Conventional Applications) by Chris W. Johnson
	Copyright (C) 2003-2010 The University of Texas at Austin ("U.T. Austin"). All rights reserved.
	
	U.T. Austin's copyright notice in this evolving version reflects our original authorship. It does not assert 
	copyright over changes contributed by others.
	
	This library is free software; you can redistribute it and/or modify it under the terms of the GNU Lesser General 
	Public License as published by the Free Software Foundation; either version 2.1 of the License, or (at your option) 
	any later version.
	
	This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied 
	warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for 
	more details.
	
	You should have received a copy of the GNU Lesser General Public License along with this library; if not, write to 
	the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/


package edu.utexas.its.eis.tools.qwicap.servlet;


import edu.utexas.its.eis.tools.qwicap.template.css.Results;
import edu.utexas.its.eis.tools.qwicap.template.xml.Match;
import edu.utexas.its.eis.tools.qwicap.template.xml.structure.NamedItem;
import edu.utexas.its.eis.tools.qwicap.util.Characters;
import edu.utexas.its.eis.tools.qwicap.util.ArrayToString;
import edu.utexas.its.eis.tools.table.lltable.MutableLLTable;

import java.util.ArrayList;


/**	<code>FormControlInfo</code> associates a form control's name and value with the markup element that supplied them, 
 *	and the label(s) in the markup that apply to the control.
 *	
 *	@author Chris W. Johnson
 */
public final class FormControlInfo {
	
	private final Form					TheForm;
	private final String				ControlName;
	private final String[]				ControlValues;
	private final Match					Control;
	private Match						EnclosingElem;	//	"form" or "fieldset"
	private ArrayList<LabelElementInfo>	Labels;			//	Contains LabelElementInfo objects
	
	
/**	Creates a <code>FormControlInfo</code> object that associates a form control's name and value with the markup 
 * 	markup that supplied them, and the label(s) in the markup that apply to the control.
 * 
 * @param TheForm			The <code>Form</code> in which this control is located.
 * @param ControlName		The name of this form control.
 * @param ControlValue		The value of this form control. If <code>null</code>, the form is examined to determine the
 *							value(s) of this control.
 * @param Control			The <code>Match</code> object that uniquely identifies the form control that is to
 *							be represented by this <code>FormControlInfo</code> object.
 *
 */
 	FormControlInfo
	(
		final Form				TheForm,
		final String			ControlName,
		final String			ControlValue,	//	Can be null.
		final Match				Control
	)
	{
		this.TheForm       = TheForm;
		this.ControlName   = ControlName;
		this.ControlValues = ControlValue != null ? new String[] { ControlValue } : Form.getControlValues(Control, false, true, false);
		this.Control       = Control;
	}
	
	@Override public String toString() {
		return new MutableLLTable().setColNames("Enclosing Element", "Form Element", "Control Element", "Control Name", "Control Value(s)", "Label Element(s)", "Label Text")
		       .addRow().addAll(getEnclosingElement(), TheForm.getMatch().toString(true), Control, ControlName, new ArrayToString(ControlValues).alwaysQuote(), new ArrayToString(getLabels()).alwaysQuote(), getLabelText(false))
		       .toString(); 
	}
	
/**	Returns the <code>Form</code> object representing the "<samp>form</samp>" element in which this form control is
 * 	located.	
 * 
 * @return		The <code>Form</code> object representing the "<samp>form</samp>" element in which this form control is
 * 				located.
 */	
	public Form getForm() {
		return TheForm;
	}
	
/**	Returns the name of this control, as it was supplied to the constructor.
 * 
 * @return		The name of this control.
 */
	public String getName() {
		return ControlName;
	}
	
/**	Returns the value of this control. If it has multiple values, as is possible with some "select" controls, only the
 *	first value is returned.
 * 
 * @return		The first value of this control, or an empty string, if this control has no value.
 */
	public String getValue() {
		return ControlValues.length == 0 ? "" : ControlValues[0];
	}
	
/**	Returns the values of this control.	
 * 
 * @return		The values of this control, or an empty array, if this control has no values.
 */	
	public String[] getValues() {
		return ControlValues;
	}
	
/**	Returns the <code>Match</code> object that uniquely identifies the form control element represented by this
 *	<code>FormControlInfo</code> object.
 * 
 * @return		The <code>Match</code> object that uniquely identifies the form control element represented by 
 *				this <code>FormControlInfo</code> object.
 */	
	public Match getMatch() {
		return Control;
	}
	
/**	Returns the "<samp>form</samp>" element which encloses the element described by this object, or, when the form is 
 *	itself enclosed in a "<samp>fieldset</samp>" element, this method returns that "<samp>fieldset</samp>" element. In 
 *	the case of multiple enclosing "<samp>fieldset</samp>" elements, the outer-most one is returned.
 * 
 * @return		The outermost markup element, either "<samp>form</samp>" or "<samp>fieldset</samp>", that encloses the 
 *				form control described by this object.
 */	
	public Match getEnclosingElement() {
		
		if (Labels == null)
			findLabels();
		
		return EnclosingElem;
	}
	
/**	Returns <code>true</code> if label information was found for this control, <code>false</code> otherwise.
 * 
 * @return		<code>true</code> if label information was found for this control, <code>false</code> otherwise.
 */
	public boolean hasLabelText() {
		
		if (Labels == null)
			findLabels();
		
		return Labels.size() != 0;
	}
	
/**	Returns the text of all the labels that apply to this control, with each individual label separated by a "/" 
 *	character.
 * 
 * @param Tidy	<code>true</code> if steps should be taken to tidy-up the label text, <code>false</code> if the text
 *				should be used exactly as it was found in the markup. Tidying is performed by 
 *				<code>LabelElementInfo.getLabelTextTidy</code> and currently consists of removing trailing colons 
 *				(':') from labels.
 * @return		The text of all the labels that apply to this control.
 */
	public String getLabelText
	(
		final boolean			Tidy
	) 
	{
		if (Labels == null)
			findLabels();
		
		String					LabelText = null;
		final int				Count     = Labels.size();
		
		if (Count == 1) {
			
			LabelText = Labels.get(0).getLabelText(Tidy);
			
		} else if (Count > 1) {
			final StringBuilder		Buff = new StringBuilder(256);
			
			for (int Index = Count; --Index >= 0;) {
				
				Buff.append(Labels.get(Index).getLabelText(Tidy));
				
				if (Index > 0)
					Buff.append(" / ");
			}
			
			LabelText = Buff.toString();
		}
		
		return LabelText;
	}
	
/**	Returns a <code>Results</code> object containing references to all markup elements that supply label information
 * 	related to this control.
 * 
 * @return		A <code>Results</code> object containing references to all markup elements that supply label 
 *				information related to this control.
 */
	public Results getLabels() {
		
		if (Labels == null)
			findLabels();
		
		final Results			Lbls = new Results();
		
		for (final LabelElementInfo Label : Labels)
			Lbls.addMatch(Label.Element);
		
		return Lbls;
	}
	
/**	Returns the text of the label that is most immediately associated with this form control.
 * 
 * @param Tidy	<code>true</code> if steps should be taken to tidy-up the label text, <code>false</code> if the text
 *				should be used exactly as it was found in the markup. Tidying is performed by 
 *				<code>LabelElementInfo.getLabelTextTidy</code> and currently consists of removing trailing colons 
 *				(':') from labels.
 * @return		The text of the label that is most immediately associated with this form control.
 */
	public String getFirstLabel
	(
		final boolean			Tidy
	) 
	{
		if (Labels == null)
			findLabels();
		
		if (hasLabelText())
			return Labels.get(0).getLabelText(Tidy);
		else
			return "";
	}
	
//	Locate the form element's label by examining the tag hierarchy within which the element is located.
//	This takes some work, particularly because several different mechanisms for associating a label with
//	a form element via the tag hierarchy.
	
//	Consider the possibility that the form element is embedded within the label statement, along with
//	the label text, for example:
//	
//	<label>
//		First Name
//		<input type="text" name="firstname">
//	</label>
//
//	If that doesn't work-out, consider the possibility that the form element is within a "fieldset" 
//	element, in which case the label text will be the content of a "legend" element, also within the 
//	"fieldset" element. For example:
//	
//	<fieldset>
//		<legend>Your Species</legend>
//		<input name="species" type="checkbox" value="wombat"> Wombat
//		<input name="species" type="checkbox" value="platypus"> Platypus
//	</fieldset>
//
//	If that doesn't work, consider the possibility that the form element is whithin an "optgroup" element
//	within a "select" element, in which case the label text will be provided by the "label" attribute of
//	of the "optgroup" element.
//	
//	<select name="ComOS">
//		<option selected label="none" value="none">None</option>
//		<optgroup label="PortMaster 3">
//			<option label="3.7.1" value="pm3_3.7.1">PortMaster 3 with ComOS 3.7.1</option>
//			<option label="3.7" value="pm3_3.7">PortMaster 3 with ComOS 3.7</option>
//			<option label="3.5" value="pm3_3.5">PortMaster 3 with ComOS 3.5</option>
//		</optgroup>
//	</select>
	
//	The process must go something like this:
//
//	1. Verify that the name of the parameter supplied by this element matches the supplied parameter name.
//
//	2. Verify that the value of the parameter--IF ANY--supplied by this element matches the supplied parameter name.
//
//	3. Consider our immediate parent: It could be "label", "fieldset", "optgroup", or "select". 
//		A. If it's "label", call getText on it. That should be the label text.
//		B. If it's "fieldset", check for an "id" attribute. If present, use getLabelByID to look for an associated
//			label. If there isn't such a label, search the fieldset for a "legend" tag and return its text. If
//			that doesn't work, see if any of the "fieldset" element's parents are a "label" tag.
//		C. If it's "optgroup", check for a "label" attribute. If present, return that. If not present, find the
//			the parent "select" element and hand processing over to the "select" handler. (Would a separate "label"
//			element ever be associated with an "optgroup" via an "id" attribute?)
//		D. If it's "select", use getLabelByID to look for an associated label. If there isn't such a label, check 
//			to see if the "select" element is contained in a "label" element. Note that the containing "label"
//			element might not be the "select" element's immediate parent.

	private boolean findLabels() {
		
		Labels = new ArrayList<LabelElementInfo>(4);
	
		final Match				Control = getMatch();
		
		if (Control != null) {
			Match					OutermostEnclosingElement = null;
			
			for (final Match Elem : Control.getAncestors(true, null)) {
				final Characters		ElemType = ((NamedItem) Elem.get()).getName();
				
			//	NOTE: We used to exit this loop when we found a "form" element, on the assumption that there could be
			//	no more identifying information for a form element outside of its enclosing "form". However, it turns out
			//	to be legal to use "fieldset" elements outside of forms, and they certainly provide some identifying
			//	information, so we now analyze all ancestors, even after we go beyond the enclosing "form" element. We
			//	could test for the "body" element and exit then, but it seems like an "optimization" that will consume as
			//	much time as it saves.
			//	
			//	Note, too, that the Java practice of comparing a String constant to a variable, e.g. "label".equals(ElemType),
			//	is not followed below because: (1) The String class doesn't know how to perform this comparison directly, 
			//	and (2) the Characters class can be more efficient about making this comparison than using the String.equals
			// 	methods on the the String version of ElemType, e.g. "label".equals(ElemType.toString()).
				
				if (ElemType.equals("label")) {
					
					addLabel(Elem, Elem.getText());
					
				} else if (getLabelByID(Elem) == false) {
					
					if (ElemType.equals("fieldset")) {
						final Results			Legend = Elem.get("legend").first();
						
						if (Legend.notEmpty())
							addLabel(Legend.get(0), Legend.getText());
						
						OutermostEnclosingElement = Elem;
						
					} else if (ElemType.equals("option")) {
						String					LabelText = Elem.getAttribute("label");
						
						if (LabelText == null && Elem.get().isStartTag()) {
							
							LabelText = Elem.getText().trim();
							if (LabelText.length() == 0)
								LabelText = null;
						}
						
						if (LabelText != null)
							addLabel(Elem, LabelText);
								
					} else if (ElemType.equals("input")) {
						final String			Type = Elem.getAttribute("type");
						
						if ("submit".equals(Type))
							addLabel(Elem, Elem.getAttribute("value"));
						else if ("reset".equals(Type))
							addLabel(Elem, "Reset");
						
					} else if (ElemType.equals("optgroup")) {
						final String			Text = Elem.getAttribute("label");
						
						if (Text != null)
							addLabel(Elem, Text);
						
					} else if (ElemType.equals("button")) {
						
						addLabel(Elem, Elem.getText());
						
					} else if (ElemType.equals("form")) {
						
						OutermostEnclosingElement = Elem;
					} // else if (AncestorType.equals("select")) {	//	"select" either has a label-by-id, or it's embedded in a "label" statement; either way there's nothing to do here.
				}
			}
			
			EnclosingElem = OutermostEnclosingElement;
			
		} else
			EnclosingElem = TheForm.getMatch();
		
		return hasLabelText();
	}
	
	private void addLabel
	(
		final Match				Element,
		final String			LabelStr
	) 
	{
		for (final LabelElementInfo Label : Labels) {
		
			if (Label.Element == Element)
				return;	//	We've already discovered this label element, so don't add it again.
		}
		
		Labels.add(new LabelElementInfo(Element, LabelStr));
	}
	
	private boolean getLabelByID
	(
		final Match				LabelledElem
	) 
	{
		boolean					Found = false;
		final String			ID    = LabelledElem.getAttribute("id");
			
		if (ID != null) {
		
		//	Locate the form element's label by its unique ID. This is easy.
		
		//	Find the associated label(s) whose "for" attribute matches that "id" attribute (if any), and mark
		//	them with a reserved class name that can be used by the style sheet to change their appearance in some
		//	way that will draw attention to them (typically, the label text will be drawn in red).
			
			final Results			Labels = LabelledElem.getMarkup().get("label[for='" + ID + "']").first();
			
			if (Labels.notEmpty()) {
				
				addLabel(Labels.get(0), Labels.getText());
				Found = true;
			}
		}
		
		return Found;
	}
}




