/*	QWICAP (Quick Web Interface for Conventional Applications) by Chris W. Johnson
	Copyright (C) 2003-2010 The University of Texas at Austin ("U.T. Austin"). All rights reserved.
	
	U.T. Austin's copyright notice in this evolving version reflects our original authorship. It does not assert 
	copyright over changes contributed by others.
	
	This library is free software; you can redistribute it and/or modify it under the terms of the GNU Lesser General 
	Public License as published by the Free Software Foundation; either version 2.1 of the License, or (at your option) 
	any later version.
	
	This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied 
	warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for 
	more details.
	
	You should have received a copy of the GNU Lesser General Public License along with this library; if not, write to 
	the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/


package edu.utexas.its.eis.tools.qwicap.servlet;


import edu.utexas.its.eis.tools.qwicap.util.ArrayIterator;
import edu.utexas.its.eis.tools.qwicap.util.Null;

import java.util.Iterator;


/**	Represents the control-name/current-value(s) pair for one or more 
 *	<a href='http://www.w3.org/TR/html4/interact/forms.html#h-17.13.2'>successful controls</a> of the <i>same</i> name 
 *	in an XHTML form.
 *	
 *	@author Chris W. Johnson
 */
public class FormControlValue implements Comparable<FormControlValue>, Iterable<String> {

	private final String		Name;
	private String[]			Values;
	private final boolean		IsQwicapControl;


/**	Constructs a <code>FormControlValue</code> object with any number of values.
 * 
 * @param ControlName			The name of the controls.
 * @param ControlValues			All of the values supplied by controls of the specified name. There must be at least
 *								one value.
 */
	FormControlValue
	(
		final String			ControlName,
		final Object[]			ControlValues
	)
	{
		final String[]			ValueStrs  = new String[ControlValues.length];
		int						WriteIndex = 0;
		
		for (final Object Value : ControlValues)
			ValueStrs[WriteIndex++] = Value != null ? Value.toString() : "";
		
		Name            = filterName(ControlName);
		Values          = ValueStrs;
		IsQwicapControl = Name.startsWith(Qwicap.kQwicapControlNamePrefix);
	}

	@Override public int hashCode() {
		return Name.hashCode();
	}

/** Tests the name and value(s) of this form control for equality with the name and value(s) of another 
 *	<code>FormControlValue</code> object. The order of the values in each object will not affect the outcome of this
 *	comparison.
 * 
 * @param Other					Another object, or <code>null</code>.
 * @return						<code>true</code> if the other object is an instance of <code>FormControlValue</code>,
 *								and its name and value(s) match the name of value(s) of this 
 *								<code>FormControlValue</code> instance, <code>false</code> otherwise.
 */
	@Override public boolean equals
	(
		final Object			Other
	)
	{
		if (Other == null)
			return false;

		if (Other == this)
			return true;

		if (Other instanceof FormControlValue == false)
			return false;

		final FormControlValue	O = (FormControlValue) Other;

		if (Name.equals(O.Name) == false)
			return false;
		
	//	Note that it's important that we do NOT directly access our Values field in the code below, because we want
	//	this method to work for the FormFileUpload subclass, and it does not store its value in our Values field.
		
		final String[]			OValues = O.getValues();
		final String[]			Values  = getValues();
		final int				Length  = Values.length;
		
		if (Length != OValues.length)
			return false;

		if (Length == 1)
			return Values[0].equals(OValues[0]);

		final boolean[]			Matched = new boolean[Length];
		
		for (final String Value : Values) {
			boolean					FoundMatch = false;

			for (int OIndex = 0; OIndex < Length; OIndex++) {

				if (Matched[OIndex])
					continue;

				if (Value.equals(OValues[OIndex])) {

					Matched[OIndex] = true;
					FoundMatch      = true;
					break;
				}
			}

			if (FoundMatch == false)
				return false;
		}

		return true;
	}

/**	Compares the name of this form control with the name of another form control, specified as a 
 * 	<code>FormControlValue</code> object.
 * 
 * @param Other					A <code>FormControlValue</code> object whose name should be compared to the name of 
 * 								this <code>FormControlValue</code> object.
 * @return						A negative integer, zero, or a positive integer as this form control's name is less 
 * 								than, equal to, or greater than the name of the specified form control object.
 */
	@Override public int compareTo
	(
		final FormControlValue	Other
	)
	{
	//	The Comparable interface requires the following NullPointerException. It states: "Note that null is not an 
	//	instance of any class, and e.compareTo(null) should throw a NullPointerException even though e.equals(null) 
	//	returns false."

		if (Other == null)
			throw new NullPointerException();	//	The compareTo contract requires this behavior.

		return Name.compareTo(Other.Name);		//	The compareTo contract requires the ClassCastException that this might throw.
	}

/**	Corrects an old bug in some browsers that corrupted control names. To explain: When parameters are embedded in a 
 *	URL in a web page, the ampersands that separate parameters must be encoded in standard HTML fashion as "&amp;". 
 *	Some browser authors haven't acquainted themselves with the spec, however, and fail to decode URLs before passing 
 *	them along to web servers. In such cases, the parameter names arrive with "amp;" prepended to them. Here we correct 
 *	for that by stripping "amp;" from the start of parameter names. Since nobody is likely to begin their form 
 *	parameter names with that string (the semicolon alone seems to assure that), this should be a safe fix.
 * 
 * @param Name	The name of the parameter.
 * @return		The name of the parameter stripped of its "amp;" prefix, if it had one.
 */
	private static String filterName
	(
		final String			Name
	)
	{
		if (Name.startsWith("amp;"))
			return Name.substring(4);
		else
			return Name;
	}

/**	Returns the name of the control that supplied this value, or values.	
 * 
 * @return		The name of the control that supplied this value, or values.
 */
	public final String getName() {
		return Name;
	}

/**	Returns the first value in the relevant form data set that was supplied by this control, or controls. This method 
 *	is unsuitable for controls like checkboxes which may provide multiple values; use {@link #getValues} when dealing 
 *	such controls.
 * 
 * @return		The first value supplied by this control.
 * @see			#getValues
 */
	public String getValue() {
		return Values[0];
	}

/**	Returns all of the values in the relevant form data set that were supplied by this control, or controls. Multiple 
 *	values are supplied by controls like checkboxes, or by any set of simultaneously successful controls of the same 
 * 	name.
 * 
 * @return		The values of this parameter represented as <code>String</code> array. For efficiency, the array 
 *				returned is this object's own master copy, and <b>its contents should not be modified</b>.
 */
	public String[] getValues() {
		return Values;
	}

/**	Returns an <code>Iterator</code> of the values in the relevant form data set that were supplied by this control, 
 * 	or controls. Multiple values are supplied by controls like checkboxes, or by any set of simultaneously successful
 * 	controls of the same name.
 * 
 * @return		An <code>Iterator</code> of the strings representing the values of this control.
 */
	@Override public Iterator<String> iterator() {
		return new ArrayIterator<String>(Values);
	}

	void addValues
	(
		final Object[]			AddValues
	)
	{
		final int				OldValueCount = Values.length;
		final int				AddValueCount = AddValues.length;
		final String[]			NewValues     = new String[OldValueCount + AddValueCount];

		System.arraycopy(Values, 0, NewValues, 0, OldValueCount);
		
		int						WriteIndex = OldValueCount;
		
		for (final Object Value : AddValues)
			NewValues[WriteIndex++] = Null.toSomeString(Value, "");
		
		Values = NewValues;
	}

	void addValue
	(
		final Object			Value
	)
	{
		final int				OldValueCount = Values.length;
		final String[]			NewValues     = new String[OldValueCount + 1];

		System.arraycopy(Values, 0, NewValues, 0, OldValueCount);
		NewValues[OldValueCount] = Null.toSomeString(Value, "");

		Values = NewValues;
	}
	
	void setValues
	(
		final Object[]			NewValues
	)
	{
		final int				ValueCount = NewValues.length;
		
		Values = new String[ValueCount];
		
		for (int Index = 0; Index < ValueCount; Index++)
			Values[Index] = Null.toSomeString(NewValues[Index], "");
	}

/**	Returns <code>true</code> if the control that supplied this value, or values, is owned by Qwicap and therefore
 *	reserved for its internal use, or <code>false</code> if the control was a normal part of the form. To explain: 
 * 	Qwicap automatically adds hidden "input" controls to all forms for its internal use. The names of those controls 
 * 	are prefixed with the string "qwicap-" in order to avoid collisions with the names of the controls placed in the 
 * 	form by the web developer. This method makes it easy to distinguish between the values supplied by Qwicap's hidden 
 * 	controls, and the web application's own controls.
 * 
 * @return	<code>true</code> if the control that supplied this value, or values, is owned by Qwicap, and 
 * 			<code>false</code> if the control was a normal part of the form.
 * @deprecated Use {@link #isQwicapControl} instead.
 */
	public boolean isQwicapParameter() {
		return IsQwicapControl;
	}

/**	Returns <code>true</code> if the control that supplied this value, or values, is owned by Qwicap and therefore
 *	reserved for its internal use, or <code>false</code> if the control was a normal part of the form. To explain: 
 * 	Qwicap automatically adds hidden "input" controls to all forms for its internal use. The names of those controls 
 * 	are prefixed with the string "qwicap-" in order to avoid collisions with the names of the controls placed in the 
 * 	form by the web developer. This method makes it easy to distinguish between the values supplied by Qwicap's hidden 
 * 	controls, and the web application's own controls.
 * 
 * @return	<code>true</code> if the control that supplied this value, or values, is owned by Qwicap, and 
 * 			<code>false</code> if the control was a normal part of the form.
 */
	public boolean isQwicapControl() {
		return IsQwicapControl;
	}
}
