/*	QWICAP (Quick Web Interface for Conventional Applications) by Chris W. Johnson
	Copyright (C) 2003-2011 The University of Texas at Austin ("U.T. Austin"). All rights reserved.
	
	U.T. Austin's copyright notice in this evolving version reflects our original authorship. It does not assert 
	copyright over changes contributed by others.
	
	This library is free software; you can redistribute it and/or modify it under the terms of the GNU Lesser General 
	Public License as published by the Free Software Foundation; either version 2.1 of the License, or (at your option) 
	any later version.
	
	This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied 
	warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for 
	more details.
	
	You should have received a copy of the GNU Lesser General Public License along with this library; if not, write to 
	the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/


package edu.utexas.its.eis.tools.qwicap.servlet;


import edu.utexas.its.eis.tools.qwicap.util.ArrayToString;
import edu.utexas.its.eis.tools.qwicap.util.Null;
import edu.utexas.its.eis.tools.table.lltable.MutableLLTable;
import edu.utexas.its.eis.tools.table.MutableTable;
import edu.utexas.ots.chrisj.util.Hexadecimal;
import org.apache.commons.fileupload.FileItem;
import org.apache.commons.fileupload.FileUploadException;
import org.apache.commons.fileupload.FileUploadBase;
import org.apache.commons.fileupload.disk.DiskFileItemFactory;
import org.apache.commons.fileupload.servlet.ServletFileUpload;
import org.apache.commons.fileupload.servlet.ServletRequestContext;

import javax.servlet.http.HttpServletRequest;
import java.io.*;
import java.util.*;
import java.util.logging.Level;
import java.util.logging.Logger;
import java.util.regex.Pattern;


/**	Provides a complete representation of a set of web parameters (a "Form Data Set" in official 
 * 	<a href='http://www.w3.org/'>W3C</a> nomenclature) that is independent of the {@link HttpServletRequest} that 
 * 	supplied them. Thus, this form data set may be retained between requests. Also stored is information on the 
 * 	origin of the data set including the method by which it was submitted (GET, POST, etc.), and the path to 
 * 	the servlet that received them.
 *	
 *	@see <a href='http://www.w3.org/TR/html4/interact/forms.html#h-17.13.3.2'>Form Data Set as defined in section 17.13.3 of the HTML 4 Specification</a>
 *	@author Chris W. Johnson
 */
public final class FormDataSet implements Iterable<FormControlValue> {

	private static final Logger		Log = Logger.getLogger(FormDataSet.class.getName());

/**	Data sets received with matching page ID values within this length of time will be considered to be the result of 
 *	an accidental multi-click (bounce) of a form submit button, link, or whatever.
 */
 	private static final int		kBounceThresholdMillis = 1000;
	private static final byte[]		kEmptyByteArray        = {};
	private static final Pattern	kMatchAllPattern       = Pattern.compile(".*");

	private final long				HitTime = System.currentTimeMillis();
	private String					ActualCharSet;
	private String					Method;
	private final String			ContextPath;
	private final String			ServletPath;
	private Exception				DataSetLoadingException;
	private String					ActionURL;
	
/**	Note that we use a <code>LinkedHashMap</code> for <code>DataSet</code>, because we need to preserve the order of 
 * 	the data in the data set. The order is important to the character set analysis code that examines the first bytes 
 * 	of the data set for UTF byte-order-marks (BOMs).
 */
	private final HashMap<String, FormControlValue>	DataSet = new LinkedHashMap<String, FormControlValue>();
	

/**	Creates an empty <code>FormDataSet</code> object.
 */
	public FormDataSet() {
		
		ActualCharSet = "UTF-8";
		Method        = "n/a";
		ContextPath   = "";
		ServletPath   = "";
	}

/**	Creates a <code>FormDataSet</code> instance populated with the data set extracted from a string encoded in the
 * 	"<a href='http://www.w3.org/TR/html4/interact/forms.html#h-17.13.4.1'>application/x-www-form-urlencoded</a>" format.
 * 	
 * @param WWWFormURLEncodedString	A <code>String</code> in "application/x-www-form-urlencoded" format.
 * @param DefaultCharSetName		The name of the character set used to encode the characters of this data set. Note
 * 									that standards require that the characters of which the <code>WWWFormURLEncodedString</code> 
 * 									parameter is composed must all be encoded using the "US-ASCII" character set, but 
 * 									that the characters represented by those characters may be encoded using any 
 * 									character set. The <code>DefaultCharSetName</code> parameter identifies the latter 
 * 									character set.
 */
	public FormDataSet
	(
		final String			WWWFormURLEncodedString,
		final String			DefaultCharSetName
	)
	{
		final FormDataSetDecoder	Decoder = new FormDataSetDecoder(WWWFormURLEncodedString, DefaultCharSetName).populateFormDataSet(this);
		
		ActualCharSet = Decoder.getCharacterSet();
		Method        = "n/a";
		ContextPath   = "";
		ServletPath   = "";
	}
	
/**	Creates a <code>FormDataSet</code> object representing the data set contained in the supplied <code>byte</code> 
 * 	array.
 * 
 * @param Method				The value returned by the {@link HttpServletRequest#getMethod} method of the current 
 * 								request.
 * @param ContextPath			The value returned by the {@link HttpServletRequest#getContextPath} method of the 
 * 								current request.
 * @param ServletPath			The value returned by the {@link HttpServletRequest#getServletPath} method of the 
 * 								current request.
 * @param ReqBytes				The bytes of the current request in <code>application/x-www-form-urlencoded</code> 
 * 								format. Can be <code>null</code>.
 * 
 * @see <a href='http://www.ietf.org/rfc/rfc1738.txt'>RFC 1738, section 2.2, "URL Character Encoding Issues"</a>
 */
	public FormDataSet
	(
		final String			Method,
		final String			ContextPath,
		final String			ServletPath,
		final byte[]			ReqBytes
	)
	{
	//	For details of the "application/x-www-form-urlencoded" content type, see:
	//
	//	See: <http://www.ietf.org/rfc/rfc1738.txt>
	//	See: <http://www.w3.org/TR/html4/interact/forms.html#didx-applicationx-www-form-urlencoded>
		
		this.Method      = Method.toLowerCase();
		this.ContextPath = ContextPath;
		this.ServletPath = ServletPath;

	//	Handle ordinary form parameters. These are sent by all forms whose 'enctype' is omitted, or is something 
	//	other than 'multipart/...'.
		
		if (Log.isLoggable(Level.FINEST))
			Log.log(Level.FINEST, "Form data set bytes ({1}):\n{0}", new Object[] { Hexadecimal.andASCIIFromBytes(ReqBytes), ReqBytes.length });
		
		if (ReqBytes != null && ReqBytes.length > 0) {
			final FormDataSetDecoder	FDSDecoder = new FormDataSetDecoder(ReqBytes, null);
			
			ActualCharSet = FDSDecoder.getCharacterSet();
			FDSDecoder.populateFormDataSet(this);
		
		} else
			ActualCharSet = "";
		
		if (Log.isLoggable(Level.FINE))
			Log.log(Level.FINE, "FormDataSet received at {0,date,yyyy-MM-dd HH:mm:ss.SSS}:\n{1}", new Object[] { HitTime, toString() });
	}

/**	Creates a <code>FormDataSet</code> object representing the data set available from the supplied 
 * 	<code>HttpServletRequest</code> object.	
 * 
 * @param Req				The current <code>HttpServletRequest</code> object.
 */
	public FormDataSet
	(
		final HttpServletRequest	Req
	)
	{
		Method      = Req.getMethod().toLowerCase();
		ContextPath = Req.getContextPath();
		ServletPath = Req.getServletPath();
		ActionURL   = Req.getRequestURL().toString();
		
		String					ActualCharSet = "US-ASCII";
		
		if (FileUploadBase.isMultipartContent(new ServletRequestContext(Req))) {
			
		//	Handle multi-part form parameters. These are typically sent when HTTP file upload is used.

		//	todo Consider character set encoding issues here. Does the Apache Commons File Upload code already handle it, or do we have to do so?
			
			final DiskFileItemFactory	DFIF = new DiskFileItemFactory();
			final ServletFileUpload		SFU  = new ServletFileUpload(DFIF);
			
			{
				final SessionContext		SCtxt = SessionContext.get(Req);
				final Qwicap				Q     = SCtxt != null ? SCtxt.getQwicap() : null;
	
				if (Q != null) {
					
					DFIF.setSizeThreshold(Q.getUploadedFileRAMLimit());
					SFU.setSizeMax(Q.getUploadedFileSizeLimit());
		
					if (Q.getUploadedFileRepository() != null)
						DFIF.setRepository(new File(Q.getUploadedFileRepository()));
					
				} else {
					
					DFIF.setSizeThreshold(Qwicap.kDefaultFileUploadRAMLimit);
					SFU.setSizeMax(Qwicap.kDefaultFileUploadSizeLimit);
				}
			}
			
			try {
				final ArrayList<FormControlBytes>	FCBs   = new ArrayList<FormControlBytes>();
				final List<FileItem>				FilesL = (List<FileItem>) SFU.parseRequest(Req);

				for (final FileItem FI : FilesL) {
					
					if (FI.isFormField()) {

						Log.log(Level.FINEST, "Control \"{0}\" has content type \"{1}\".", new Object[] {FI.getFieldName(), FI.getContentType()});

						FCBs.add(new FormControlBytes(new ByteArray(FI.getFieldName().getBytes("ISO-8859-1")), new ByteArray(FI.get())));

					} else {
						final FormFileUpload Upload = new FormFileUpload(FI);

						DataSet.put(Upload.getName(), Upload);
					}
				}
				
				new FormDataSetDecoder(FCBs, null).populateFormDataSet(this);
				
			} catch (UnsupportedEncodingException e) {
				
				DataSetLoadingException = e;
				Log.log(Level.SEVERE, "The ISO-8859-1 character set is unavailable. This should never happen.", e);
				
			} catch (FileUploadException e) {

				DataSetLoadingException = e;
				Log.log(Level.WARNING, "File upload failed: " + e.getMessage(), e);
			}

		} else {
			
		//	Handle ordinary form parameters. These are sent by all forms whose 'enctype' is omitted, or is something 
		//	other than 'multipart/...'.
			
			byte[]					ReqBytes = null;
			
			if ("GET".equalsIgnoreCase(Req.getMethod())) {
				
				try {
					final String			QueryStr = Req.getQueryString();
					
					ReqBytes = QueryStr != null ? QueryStr.getBytes("US-ASCII") : kEmptyByteArray;
					
				} catch (UnsupportedEncodingException e) {
	
					DataSetLoadingException = e;
					Log.log(Level.WARNING, "Query string getBytes failed: " + e.getMessage(), e);
				}
					
			} else {
				
				try {
					final InputStream		In = Req.getInputStream();
				
					try {
						ReqBytes = getInputBytes(In, Req.getContentLength());
					} finally {
						In.close();
					}
				
				} catch (IOException e) {
	
					DataSetLoadingException = e;
					Log.log(Level.WARNING, "Form data set read failed: " + e.getMessage(), e);
				}
			}
			
			if (ReqBytes != null) {
				
				if (Log.isLoggable(Level.FINEST))
					Log.log(Level.FINEST, "Form data set bytes ({1}):\n{0}", new Object[] { Hexadecimal.andASCIIFromBytes(ReqBytes), ReqBytes.length });
				
				if (ReqBytes.length > 0) {
					final FormDataSetDecoder	FDSDecoder = new FormDataSetDecoder(ReqBytes, null);
					
					ActualCharSet = FDSDecoder.getCharacterSet();
					FDSDecoder.populateFormDataSet(this);
				}
			}
		}
		
		this.ActualCharSet = ActualCharSet;
		
		if (Log.isLoggable(Level.FINE))
			Log.log(Level.FINE, "FormDataSet received at {0,date,yyyy-MM-dd HH:mm:ss.SSS}:\n{1}", new Object[] { HitTime, toString() });
	}

/**	Compares two data sets to determine whether or not one is likely to be an accidental replication of the other, of
 * 	the sort caused by a user accidentally (or foolishly) double-clicking on a form's "submit" button.
 * 
 * @param Other				Another data set to compared with this one.
 * @return					<code>true</code> if either of the data sets appears to be an accidental replication of the
 * 							other. <code>false</code> otherwise.
 */
	boolean isAccidentalDuplicateOf
	(
		final FormDataSet		Other
	)
	{
		if (Other == null)
			return false;
		
		long					HitTimeDelta = HitTime - Other.HitTime;
		
		if (HitTimeDelta < 0)
			HitTimeDelta = -HitTimeDelta;
		
		if (HitTimeDelta > kBounceThresholdMillis)
			return false;
		
		return Null.safeEquals(getDatum(Qwicap.kQwicapPageIDControlName), Other.getDatum(Qwicap.kQwicapPageIDControlName));
	}
	
/**	Removes all data from this data set.
 */
	public void clear() {
		DataSet.clear();
	}
	
	
/**	Adds control name/value(s) pairs to this form data set. If the named control is already represented in this data
 * 	set, the additional value(s) supplied to this method are appended to its existing list of values (rather than 
 *	replacing the existing values, for example). This method is used internally for building-up the data set, and user 
 *	code shouldn't need it, ordinarily.
 * 
 * @param Name				The name of the control.
 * @param Values			The value(s) of the control.
 * @return					This <code>FormDataSet</code> object.
 * @deprecated				Use {@link #addData} instead.
 */
	@Deprecated public FormDataSet addControlValues
	(
		final String			Name,
		final String...			Values
	)
	{
		return addData(Name, (Object[]) Values);
	}
	
/**	Adds control name/value(s) pairs to this form data set. If the named control is already represented in this data
 * 	set, the additional value(s) supplied to this method are appended to its existing list of values (rather than 
 *	replacing the existing values, for example). This method is used internally for building-up the data set, and user 
 *	code shouldn't need it, ordinarily.
 * 
 * @param Name				The name of the control.
 * @param Values			The value(s) of the control.
 * @return					This <code>FormDataSet</code> object.
 */
	public FormDataSet addData
	(
		final String			Name,
		final Object...			Values
	)
	{
	//	Can't add a control with no name (like certain "input" controls of type "submit") to the data set.
		
		if (Name == null)
			return this;
		
	//	"no values" is not an acceptable state.	//	todo Check this conclusion. Do browsers supply empty strings as the values of such controls? In that case we're doing it wrong here.
		
		if (Values.length == 0)
			return this;

		if (InputSubmitButton.producedThisControl(Name)) {
			
			for (final Object Value : Values) {
				final InputSubmitButton	ISB = new InputSubmitButton(Name, Value);
				
				Log.log(Level.FINER, "Translating control \"{0}\" with value \"{1}\" to a control named \"{2}\" with value \"{3}\".", new Object[] { Name, Value, ISB.getName(), ISB.getValue() });
				
				addData(ISB.getName(), ISB.getValue());
			}

		} else {
			final FormControlValue	Param = DataSet.get(Name);

			if (Param != null) {

				Param.addValues(Values);
				Log.log(Level.FINER, "Added the value(s) {0} to the existing control named \"{1}\".", new Object[] { new ArrayToString(Values).alwaysQuote(), Name });

			} else {

				DataSet.put(Name, new FormControlValue(Name, Values));
				Log.log(Level.FINER, "Added the control \"{0}\" with value(s) {1} to the form data set.", new Object[] { Name, new ArrayToString(Values).alwaysQuote() });
			}
		}

		return this;
	}
	
/**	Adds control name/value(s) pairs to this form data set. If the named control is already represented in this data
 * 	set, the new value(s) replace its existing value(s).
 * 
 * @param Name				The name of the control.
 * @param Values			The value(s) of the control.
 * @return					This <code>FormDataSet</code> object.
 */
	public FormDataSet setData
	(
		final String			Name,
		final Object...			Values
	)
	{
		if (InputSubmitButton.producedThisControl(Name)) {
			
			for (final Object Value : Values) {
				final InputSubmitButton	ISB = new InputSubmitButton(Name, Value);

				Log.log(Level.FINER, "Translating control \"{0}\" with value \"{1}\" to a control named \"{2}\" with value \"{3}\".", new Object[] { Name, Value, ISB.getName(), ISB.getValue() });

				setData(ISB.getName(), ISB.getValue());
			}

		} else {
			final FormControlValue	Param = DataSet.get(Name);

			if (Param != null) {
				
				Param.setValues(Values);
				Log.log(Level.FINER, "Changed the value(s) of the existing control named \"{1}\" to: {0}.", new Object[] { new ArrayToString(Values).alwaysQuote(), Name });

			} else {
				
				DataSet.put(Name, new FormControlValue(Name, Values));
				Log.log(Level.FINER, "Created the control \"{0}\" with value(s) {1}.", new Object[] { Name, new ArrayToString(Values).alwaysQuote() });
			}
		}
		
		return this;
	}

/**	If this data set was instantiated from a <code>String</code>, or <code>byte</code> array, of input, this method
 * 	returns the name of the character set in which the input was originally represented. (In its encoded form, the 
 * 	input is required by <a href='http://www.ietf.org/rfc/rfc1738.txt'>RFC 1738</a> to be in the US-ASCII character set, 
 * 	but the encoded bytes could represent characters in <i>any</i> character set.) If this data set was instantiated 
 * 	using the default constructor, a general-purpose character set (almost certainly UTF-8) will have been selected as 
 * 	the default, and this method will return the name of that character set.
 * 
 * @return						The name of the character set used by this data set.
 */
	public String getCharacterSetName() {
		return ActualCharSet;
	}

/**	Sets the name of the character set to be used to encode the characters of this form data set.
 * 
 * @param NewCharSetName		The name of the character set used to be used to encode the characters of this form 
 * 								data set.
 * @return						A reference to this <code>FormDataSet</code> instance.
 */
	public FormDataSet setCharacterSetName
	(
		final String			NewCharSetName
	) 
	{
		this.ActualCharSet = NewCharSetName;
		return this;
	}

/**	Returns the name of the HTTP mechanism used to send this form data set to the server: 
 *	"<a href='http://www.w3.org/Protocols/rfc2616/rfc2616-sec9.html#sec9.3'>GET</a>" or 
 *	"<a href='http://www.w3.org/Protocols/rfc2616/rfc2616-sec9.html#sec9.5'>POST</a>".
 * 
 * @return	The name of the HTTP mechanism used to send this form data set to the server: "GET" or "POST".
 */
	public String getMethod() {
		return Method;
	}

/**	Sets the name of the HTTP mechanism to be used to send this form data set to a server: 
 *	"<a href='http://www.w3.org/Protocols/rfc2616/rfc2616-sec9.html#sec9.3'>GET</a>" or 
 *	"<a href='http://www.w3.org/Protocols/rfc2616/rfc2616-sec9.html#sec9.5'>POST</a>".
 * 
 * @param Method				The name of the HTTP mechanism ("get" or "post") to be used to send this form data set 
 * 								to a server.
 * @return						A reference to this <code>FormDataSet</code> instance.
 */
	public FormDataSet setMethod
	(
		final String			Method
	) 
	{
		this.Method = Method;
		return this;
	}

/**	Returns the action URL of this data set, if available. If not available, <code>null</code> is returned. The action
 * 	URL of a data set received from an <code>HttpServletRequest</code> is the URL to which the data set was sent. The
 * 	action URL of a data set extracted from an XHTML page is the URL to which the data set should be sent. If the 
 * 	method used by this data set is "get" (see {@link #getMethod}), this data set will be appropriately encoded and
 * 	appended to the URL as URI attributes, for example: "<samp>http://localhost/app/?hello=world&foo=bar</samp>".
 * 
 * @return						The action URL of this data set, or <code>null</code>.
 */
	public String getActionURL() {
	
		if ("get".equalsIgnoreCase(Method) && notEmpty())
			return ActionURL + '?' + toWWWFormURLEncodedString(getCharacterSetName(), '&');
		else
			return ActionURL;
	}

/**	Sets the action URL of this data set. In the case of a data set received from an <code>HttpServletRequest</code>,
 * 	the action URL should indicate the URL to which the data set was sent (see 
 * 	{@link HttpServletRequest#getRequestURL()}). In the case of a data set extract from an XHTML page, the action URL
 * 	should specify where the data set is meant to be sent.
 * 
 * @param ActionURL				The action URL of this data set.
 * @return						A reference to this <code>FormDataSet</code> instance.
 */
	public FormDataSet setActionURL
	(
		final String			ActionURL
	)
	{
		this.ActionURL = ActionURL;
		return this;
	}
	
/** As stated in the <code><a href='http://java.sun.com/products/servlet/2.3/javadoc/javax/servlet/http/HttpServletRequest.html#getContextPath()'>HttpServletRequest.getContextPath</a></code> documentation: 
 *	<blockquote>Returns the portion of the request URI that indicates the context of the request. The context path 
 *	always comes first in a request URI. The path starts with a "/" character but does not end with a "/" character. 
 *	For servlets in the default (root) context, this method returns "". The container does not decode this string.
 *	</blockquote>
 * 
 * @return	"[T]he portion of the request URI that indicates the context of the request."
 */
	public String getContextPath() {
		return ContextPath;
	}

/**	As stated in the <code><a href='http://java.sun.com/products/servlet/2.3/javadoc/javax/servlet/http/HttpServletRequest.html#getServletPath()'>HttpServletRequest.getServletPath</a></code> documentation:
 *	<blockquote>Returns the part of this request's URL that calls the servlet. This includes either the servlet name 
 *	or a path to the servlet, but does not include any extra path information or a query string.
 *	</blockquote>
 * 
 * @return	The portion of the request's URL that called the servlet.
 */
	public String getServletPath() {
		return ServletPath;
	}

/**	Returns the <code>Exception</code>, if any, that was produced when this object's constructor retrieved, or attempted
 * 	to retrieve, this form data set.
 * 
 * @return	The <code>Exception</code> produced when this form data set was built, or <code>null</code> if everything
 *			worked correctly.
 */
	public Exception getProcessingException() {
		return DataSetLoadingException;
	}

/**	Returns <code>true</code> if this form data set is empty, <code>false</code> otherwise.
 * 
 * @return	<code>true</code> if this form data set is empty, <code>false</code> otherwise.
 */
	public boolean isEmpty() {
		return DataSet.isEmpty();
	}

/**	Returns <code>true</code> if this form data set contains data, <code>false</code> if it is empty.
 * 
 * @return	<code>true</code> if this form data set contains data, <code>false</code> if it is empty.
 */
	public boolean notEmpty() {
		return !isEmpty();
	}

/**	Returns a <code>FormControlValue</code> object containing all of the values in this form data set that were supplied by 
 *	controls having the specified name.
 * 
 * @param ControlName	The name of the control(s) whose value(s) should be returned.
 * @return				A <code>FormControlValue</code> object containing all of the values supplied by controls having the 
 *						specified name, or <code>null</code> if there are no such values.
 */
	public FormControlValue get
	(
		final String			ControlName
	)
	{
		return DataSet.get(ControlName);
	}

/** Returns an array of <code>FormControlValue</code> objects containing all of the values in this form data set that 
 *	were supplied by controls whose names matched the supplied regular expression.
 * 
 * @param Regex					The regular expression compiled as a {@link Pattern} object.
 * @return						A <code>FormControlValue</code> object for each value (or set of values) whose name 
 *								matched the regular expression, or an empty array if there are no such controls.
 */
	public FormControlValue[] get
	(
		final Pattern			Regex
	)
	{
		final ArrayList<FormControlValue>	Matches = new ArrayList<FormControlValue>();	//	Contains FormControlValue objects.

		for (final FormControlValue Param : this) {

			if (Regex.matcher(Param.getName()).matches())
				Matches.add(Param);
		}

		return Matches.toArray(new FormControlValue[Matches.size()]);
	}

/**	Returns the first value in this form data set that was supplied by a control having the specified name.
 * 
 * @param ControlName			The name of the control(s) whose first value should be returned.
 * @return						The first value of the named control(s), or <code>null</code> if there are no such 
 *								values.
 * @deprecated 					Use {@link #getDatum(String)} instead.
 */
	@Deprecated public String getValue
	(
		final String			ControlName
	)
	{
		return getDatum(ControlName);
	}
	
/**	Returns the first value in this form data set that was supplied by a control having the specified name.
 * 
 * @param ControlName			The name of the control(s) whose first value should be returned.
 * @return						The first value of the named control(s), or <code>null</code> if there are no such 
 *								values.
 */
	public String getDatum
	(
		final String			ControlName
	)
	{
		final FormControlValue	Param = get(ControlName);
		String					Value = null;

		if (Param != null)
			Value = Param.getValues()[0];

		return Value;
	}

/** Returns the first value in this form data set that was supplied by a control whose names matches the supplied 
 *	regular expression.
 * 
 * @param Regex					The regular expression compiled as a {@link Pattern} object.
 * @return						The first value of the matching control(s), or <code>null</code> if there are no such 
 *								values.
 * @deprecated 					Use {@link #getDatum(Pattern)} instead.
 */
	@Deprecated public String getValue
	(
		final Pattern			Regex
	)
	{
		return getDatum(Regex);
	}

/** Returns the first value in this form data set that was supplied by a control whose names matches the supplied 
 *	regular expression.
 * 
 * @param Regex					The regular expression compiled as a {@link Pattern} object.
 * @return						The first value of the matching control(s), or <code>null</code> if there are no such 
 *								values.
 */
	public String getDatum
	(
		final Pattern			Regex
	)
	{
		for (final FormControlValue Param : this) {

			if (Regex.matcher(Param.getName()).matches())
				return Param.getValues()[0];
		}

		return null;
	}

/**	Returns <code>true</code> if there is data from the named control in this data set, or <code>false</code> otherwise.
 * 
 * @param ControlName			Control name.
 * @return						<code>true</code> if there is data from the named control in this data set, or 
 * 								<code>false</code> otherwise.
 */
	public boolean has
	(
		final String			ControlName
	)
	{
		return DataSet.get(ControlName) != null;
	}

/**	Returns <code>true</code> if this form data set includes values from controls whose names begin with the specified 
 *	string. The substring comparison is case-sensitive.
 * 
 * @param ControlNamePrefix		Control name prefix for which to test.
 * @return						<code>true</code> if this form data set includes values from controls whose names begin 
 *								with the specified string, <code>false</code> otherwise.
 */
	public boolean hasControlNamesStartingWith
	(
		final String			ControlNamePrefix
	)
	{
		for (final FormControlValue Elem : this) {

			if (Elem.getName().startsWith(ControlNamePrefix))
				return true;
		}

		return false;
	}

/** Returns <code>true</code> if this form data set includes vlaues from controls whose names match the supplied 
 *	regular expression.
 * 
 * @param Regex					The regular expression compiled as a {@link Pattern} object.
 * @return						<code>true</code> if this form data set includes values from controls whose names 
 *								are matched by the specified regular expression, <code>false</code> otherwise.
 */
	public boolean hasControlNamesMatching
	(
		final Pattern			Regex
	)
	{
		for (final FormControlValue Elem : this) {

			if (Regex.matcher(Elem.getName()).matches())
				return true;
		}

		return false;
	}

/**	Removes from this form data set all values associated with the control named by the <code>ControlName</code>
 * 	parameter. If there is no such control in the data set, nothing happens.
 * 
 * @param ControlName			The name of the control whose values should be removed from this form data set.
 * @return						A reference to this <code>FormDataSet</code> instance, in order to enable method 
 * 								invocation chaining.
 */
	public FormDataSet deleteData
	(
		final String			ControlName
	)
	{
		DataSet.remove(ControlName);
		return this;
	}

/**	Removes from this form data set all values associated with each of the controls identified by a name in the 
 * 	<code>ControlNames</code> parameter. When there is no control corresponding to a specified name, nothing happens.
 * 
 * @param ControlNames			The names of the controls whose values should be removed from this form data set.
 * @return						A reference to this <code>FormDataSet</code> instance, in order to enable method 
 * 								invocation chaining.
 */
	public FormDataSet deleteData
	(
		final String...			ControlNames
	)
	{
		for (final String ControlName : ControlNames)
			DataSet.remove(ControlName);
		
		return this;
	}

/**	Returns an iterator for the complete set of <code>FormControlValue</code> objects in this form data set. Don't be 
 *	surprised to find more than just the values from your own form controls, because the full-up Qwicap system (as 
 *	opposed to the templater) adds hidden "input" controls to all forms for its internal use. Their names are always 
 *	prefixed with "qwicap" to avoid name collisions. You can use the <code>FormControlValue.isQwicapControl</code>
 *	method to conveniently identify the control values that Qwicap has added.
 * 
 * @return				An iterator for the complete set of {@link FormControlValue} objects in this form data set.
 */
	@Override public Iterator<FormControlValue> iterator() {
		return DataSet.values().iterator();
	}

/**	Returns a human-readable representation of this data set.
 * 
 * @return				A human-readable representation of this data set.
 */
	@Override public String toString() {
		return toTable().toString();	
	}

/**	Returns this data set represented as a <code>String</code> in 
 * 	"<a href='http://www.w3.org/TR/html4/interact/forms.html#didx-applicationx-www-form-urlencoded'>application/x-www-form-urlencoded</a>" 
 * 	format. Like any <code>String</code> in Java, the characters comprising this string are Unicode, but, as required by 
 * 	<a href='http://www.ietf.org/rfc/rfc1738.txt'>RFC 1738</a> all of the Unicode characters used in this string 
 * 	correspond directly to "the graphic printable characters of the US-ASCII coded character set", excluding those 
 * 	characters that RFC 1738 labels "unsafe" (see section 2.2). Thus, while this string is not actually in the 
 * 	required US-ASCII character set (it's in UTF-16, like all Java <code>String</code> objects), it is guaranteed to be
 * 	perfectly convertable to US-ASCII, in keeping with all of the requirements of RFC 1738.
 * 
 *  <p>The characters comprising the actual names and values of the data in this data set 
 * 	are encoded in the original/default character set for this data set (see {@link #getCharacterSetName()}). The 
 * 	bytes produced by that encoding are the bytes that are subsequently encoded in the "application/x-www-form-urlencoded" 
 * 	format using only the prescribed portions of the US-ASCII character set as specified by RFC 1738. Thus, the data of 
 * 	this data set may be encoded in any character set, but still be represented entirely by US-ASCII characters, as 
 * 	required.
 * 	</p>
 * 
 * @return				A <code>String</code> representation of this data set in the "application/x-www-form-urlencoded" 
 * 						format.
 */
	public String toWWWFormURLEncodedString() {
		return toWWWFormURLEncodedString(getCharacterSetName(), '&');
	}
	
/**	Returns this data set represented as a <code>String</code> in 
 * 	"<a href='http://www.w3.org/TR/html4/interact/forms.html#didx-applicationx-www-form-urlencoded'>application/x-www-form-urlencoded</a>" 
 * 	format. Like any <code>String</code> in Java, the characters comprising this string are Unicode, but, as required by 
 * 	<a href='http://www.ietf.org/rfc/rfc1738.txt'>RFC 1738</a> all of the Unicode characters used in this string 
 * 	correspond directly to "the graphic printable characters of the US-ASCII coded character set", excluding those 
 * 	characters that RFC 1738 labels "unsafe" (see section 2.2). Thus, while this string is not actually in the 
 * 	required US-ASCII character set (it's in UTF-16, like all Java <code>String</code> objects), it is guaranteed to be 
 * 	perfectly convertable to US-ASCII, in keeping with all of the requirements of RFC 1738.
 * 
 *  <p>The characters comprising the actual names and values of the data in this data set are encoded in the character 
 * 	set specified by the <code>CharSet</code> parameter. The bytes produced by that encoding are the bytes that are 
 * 	subsequently encoded in the "application/x-www-form-urlencoded" format using only the prescribed portions of the 
 * 	US-ASCII character set as specified by RFC 1738. Thus, the data of this data set may be encoded in any character 
 * 	set, but still be represented entirely by US-ASCII characters, as required.
 * 	</p>
 * 
 * @param CharSet		The name of the character set in which the names and values of the data in this data set are to
 * 						be encoded.
 * @return				A <code>String</code> representation of this data set in the "application/x-www-form-urlencoded" 
 * 						format.
 */
	public String toWWWFormURLEncodedString
	(
		final String			CharSet	
	) 
	{
		return toWWWFormURLEncodedString(CharSet, '&');
	}
	
/**	Returns this data set represented as a <code>String</code> in 
 * 	"<a href='http://www.w3.org/TR/html4/interact/forms.html#didx-applicationx-www-form-urlencoded'>application/x-www-form-urlencoded</a>" 
 * 	format. Like any <code>String</code> in Java, the characters comprising this string are Unicode, but, as required by 
 * 	<a href='http://www.ietf.org/rfc/rfc1738.txt'>RFC 1738</a> all of the Unicode characters used in this string 
 * 	correspond directly to "the graphic printable characters of the US-ASCII coded character set", excluding those 
 * 	characters that RFC 1738 labels "unsafe" (see section 2.2). Thus, while this string is not actually in the 
 * 	required US-ASCII character set (it's in UTF-16, like all Java <code>String</code> objects), it is guaranteed to be 
 * 	perfectly convertable to US-ASCII, in keeping with all of the requirements of RFC 1738.
 * 
 *  <p>The characters comprising the actual names and values of the data in this data set are encoded in the character 
 * 	set specified by the <code>CharSet</code> parameter. The bytes produced by that encoding are the bytes that are 
 * 	subsequently encoded in the "application/x-www-form-urlencoded" format using only the prescribed portions of the 
 * 	US-ASCII character set as specified by RFC 1738. Thus, the data of this data set may be encoded in any character 
 * 	set, but still be represented entirely by US-ASCII characters, as required.
 * 	</p>
 * 
 * 	<p>When a data set is encoded in "application/x-www-form-urlencoded" format for use in an HTTP URL, the name/value 
 * 	pairs of the data set may be separated by either ampersand ('&amp;'), or semicolon (';'), characters. (See 
 * 	<a href='http://www.w3.org/TR/html4/appendix/notes.html#ampersands-in-uris'>HTML 4 Specification, Section B.2.2, 
 * 	"Ampersands in URI Attribute Values"</a>.) The ampersand is the most common separator, but the semicolon is the 
 * 	most covenient when the string is to be placed into HTML markup, because a semicolon does not have to be encoded as 
 * 	an <a href='http://www.w3.org/TR/html4/sgml/entities.html'>HTML character entity reference</a>, unlike an ampersand
 * 	(which ends-up as either "&amp;amp;", or "&amp;#59;"). The <code>URIAttributeSeparator</code> parameter allows you 
 * 	to specify whether an ampersand or a semicolon should be used.
 * 	</p>
 * 
 * @param CharSet					The name of the character set in which the names and values of the data in this 
 * 									data set are to be encoded.
 * @param URIAttributeSeparator		Either an ampersand ('&amp;'), or a semicolon (';'), character.
 * @return							A <code>String</code> representation of this data set in the 
 * 									"application/x-www-form-urlencoded" format.
 */
	public String toWWWFormURLEncodedString
	(
		final String			CharSet,
		final char				URIAttributeSeparator
	) 
	{
		try {
			return new FormDataSetEncoder(this, CharSet, URIAttributeSeparator).toString();
		} catch (IOException e) {
			throw new RuntimeException(e);	
		}
	}
	
/**	Returns this data set represented as a <code>byte</code> array in 
 * 	"<a href='http://www.w3.org/TR/html4/interact/forms.html#didx-applicationx-www-form-urlencoded'>application/x-www-form-urlencoded</a>" 
 * 	format. As required by 	<a href='http://www.ietf.org/rfc/rfc1738.txt'>RFC 1738</a>, all bytes in the array 
 * 	correspond directly to "the graphic printable characters of the US-ASCII coded character set", excluding those 
 * 	characters that RFC 1738 labels "unsafe" (see section 2.2).
 * 
 *  <p>The characters comprising the actual names and values of the data in this data set 
 * 	are encoded in the original/default character set for this data set (see {@link #getCharacterSetName()}). The 
 * 	bytes produced by that encoding are the bytes that are subsequently encoded in the "application/x-www-form-urlencoded" 
 * 	format using only the prescribed portions of the US-ASCII character set as specified by RFC 1738. Thus, the data of 
 * 	this data set may be encoded in any character set, but still be represented entirely by US-ASCII character codes, 
 * 	as required.
 * 	</p>
 * 
 * @return				A <code>byte</code> array representing this data set in the "application/x-www-form-urlencoded" 
 * 						format.
 */
	public byte[] toWWWFormURLEncodedBytes() {
		return toWWWFormURLEncodedBytes(getCharacterSetName(), '&');
	}

/**	Returns this data set represented as a <code>byte</code> array in 
 * 	"<a href='http://www.w3.org/TR/html4/interact/forms.html#didx-applicationx-www-form-urlencoded'>application/x-www-form-urlencoded</a>" 
 * 	format. As required by 	<a href='http://www.ietf.org/rfc/rfc1738.txt'>RFC 1738</a>, all bytes in the array 
 * 	correspond directly to "the graphic printable characters of the US-ASCII coded character set", excluding those 
 * 	characters that RFC 1738 labels "unsafe" (see section 2.2).
 * 
 *  <p>The characters comprising the actual names and values of the data in this data set are encoded in the character 
 * 	set specified by the <code>CharSet</code> parameter. The bytes produced by that encoding are the bytes that are 
 * 	subsequently encoded in the "application/x-www-form-urlencoded" format using only the prescribed portions of the 
 * 	US-ASCII character set as specified by RFC 1738. Thus, the data of this data set may be encoded in any character 
 * 	set, but still be represented entirely by US-ASCII character codes, as required.
 * 	</p>
 * 
 * @param CharSet		The name of the character set in which the names and values of the data in this data set are to
 * 						be encoded.
 * @return				A <code>byte</code> array representing this data set in the "application/x-www-form-urlencoded" 
 * 						format.
 */
	public byte[] toWWWFormURLEncodedBytes
	(
		final String			CharSet	
	) 
	{
		return toWWWFormURLEncodedBytes(CharSet, '&');
	}
	
/**	Returns this data set represented as a <code>byte</code> array in 
 * 	"<a href='http://www.w3.org/TR/html4/interact/forms.html#didx-applicationx-www-form-urlencoded'>application/x-www-form-urlencoded</a>" 
 * 	format. As required by 	<a href='http://www.ietf.org/rfc/rfc1738.txt'>RFC 1738</a>, all bytes in the array 
 * 	correspond directly to "the graphic printable characters of the US-ASCII coded character set", excluding those 
 * 	characters that RFC 1738 labels "unsafe" (see section 2.2).
 * 
 *  <p>The characters comprising the actual names and values of the data in this data set are encoded in the character 
 * 	set specified by the <code>CharSet</code> parameter. The bytes produced by that encoding are the bytes that are 
 * 	subsequently encoded in the "application/x-www-form-urlencoded" format using only the prescribed portions of the 
 * 	US-ASCII character set as specified by RFC 1738. Thus, the data of this data set may be encoded in any character 
 * 	set, but still be represented entirely by US-ASCII character codes, as required.
 * 	</p>
 * 
 * 	<p>When a data set is encoded in "application/x-www-form-urlencoded" format for use in an HTTP URL, the name/value 
 * 	pairs of the data set may be separated by either ampersand ('&amp;'), or semicolon (';'), characters. (See 
 * 	<a href='http://www.w3.org/TR/html4/appendix/notes.html#ampersands-in-uris'>HTML 4 Specification, Section B.2.2, 
 * 	"Ampersands in URI Attribute Values"</a>.) The ampersand is the most common separator, but the semicolon is the 
 * 	most covenient when the string is to be placed into HTML markup, because a semicolon does not have to be encoded as 
 * 	an <a href='http://www.w3.org/TR/html4/sgml/entities.html'>HTML character entity reference</a>, unlike an ampersand
 * 	(which ends-up as either "&amp;amp;", or "&amp;#59;"). The <code>URIAttributeSeparator</code> parameter allows you 
 * 	to specify whether an ampersand or a semicolon should be used.
 * 	</p>
 * 
 * @param CharSet					The name of the character set in which the names and values of the data in this 
 * 									data set are to be encoded.
 * @param URIAttributeSeparator		Either an ampersand ('&amp;'), or a semicolon (';'), character.
 * @return							A <code>byte</code> array representing this data set in the 
 * 									"application/x-www-form-urlencoded" format.
 */
	public byte[] toWWWFormURLEncodedBytes
	(
		final String			CharSet,
		final char				URIAttributeSeparator
	) 
	{
		try {
			return new FormDataSetEncoder(this, CharSet, URIAttributeSeparator).toBytes();
		} catch (IOException e) {
			throw new RuntimeException(e);	
		}
	}
	
	MutableTable toTable() {
		final FormControlValue[]	All = get(kMatchAllPattern);
		final MutableTable			DSTbl;
		
		if (All.length > 0) {
			
			Arrays.sort(All);
			
			DSTbl = new MutableLLTable().setColNames("Control Name", "Control Value");
			
			for (final FormControlValue Elem : All) {
				final String			Name = Elem.getName();
				
				for (final String Value : Elem.getValues())
					DSTbl.addRow().add(Name).add('"' + Value + '"');
			}
		
		} else
			DSTbl = null;
		
		final MutableTable		HdrTbl = new MutableLLTable();
		
		HdrTbl.addRow("Form Method  ").add(Method                     );
		HdrTbl.addRow("Action URL"   ).add(getActionURL()             );
		HdrTbl.addRow("Context Path ").add(ContextPath                );
		HdrTbl.addRow("Servlet Path ").add(ServletPath                );
		HdrTbl.addRow("Character Set").add(ActualCharSet              );
		HdrTbl.addRow("DataSet"      ).add(DSTbl != null ? DSTbl : "" );
		
		return HdrTbl;
	}
	
/**	A debugging method that prints to <code>System.out</code> all information about this form data set and its contents.
 * 	Note that, in general, the best way to see this information is not by modifying your code to invoke this 
 * 	<code>print</code> method, but by setting the logging level for this class to <code>FINE</code>.
 * 
 * @return						A reference to this <code>FormDataSet</code> instance.
 */
	public FormDataSet print() {
		
		try {
			final Writer			Out = new OutputStreamWriter(System.out, "UTF-8");
			
			Out.write(toString());
			Out.flush();
			
		} catch (IOException e) {
			throw new RuntimeException(e);
		}
		
		return this;
	}

/**	Reads the entire contents of an <code>InputStream</code> and returns it as a <code>byte</code> array. 
 * 
 * @param In					The bytes to be read.
 * @param ContentLength			The probable number of bytes to be read, or <code>-1</code> if you can't even guess.
 * @return						The bytes read from the <code>InputStream</code>.
 * @throws IOException			If there was a problem reading the data.
 */
	static byte[] getInputBytes
	(
		final InputStream		In,
		final int				ContentLength	//	-1 if not known. 
	)
	throws IOException
	{
		if (ContentLength < 0) {
			int						InitialBuffSize = In.available();
			
			if (InitialBuffSize < 1024)
				InitialBuffSize = 1024;
			
			final ByteArrayOutputStream	BOut      = new ByteArrayOutputStream(InitialBuffSize);
			final byte[]				Buff      = new byte[InitialBuffSize];
			int							BytesRead = In.read(Buff);
			
			while (BytesRead >= 0) {
				
				BOut.write(Buff, 0, BytesRead);
				
				BytesRead = In.read(Buff);
			}
			
			return BOut.toByteArray();
			
		} else {
			final byte[]			Bytes     = new byte[ContentLength];
			int						Index     = 0;
			int						BytesRead = In.read(Bytes);
			
			while (BytesRead >= 0) {
				
				Index    += BytesRead;
				BytesRead = In.read(Bytes, Index, ContentLength - Index);
			}
			
			if (Index != ContentLength)
				Log.log(Level.SEVERE, "There should have been {0} bytes to read, but only {1} were actually read.", new Object[] { ContentLength, Index });
			
			return Bytes;
		}
	}
}



