/*	QWICAP (Quick Web Interface for Conventional Applications) by Chris W. Johnson
	Copyright (C) 2007-2008 The University of Texas at Austin ("U.T. Austin"). All rights reserved.
	
	U.T. Austin's copyright notice in this evolving version reflects our original authorship. It does not assert 
	copyright over changes contributed by others.
	
	This library is free software; you can redistribute it and/or modify it under the terms of the GNU Lesser General 
	Public License as published by the Free Software Foundation; either version 2.1 of the License, or (at your option) 
	any later version.
	
	This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied 
	warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for 
	more details.
	
	You should have received a copy of the GNU Lesser General Public License along with this library; if not, write to 
	the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/


package edu.utexas.its.eis.tools.qwicap.servlet;


import edu.utexas.ots.chrisj.util.Hexadecimal;

import java.io.UnsupportedEncodingException;
import java.nio.charset.CharacterCodingException;
import java.nio.charset.Charset;
import java.nio.charset.CharsetDecoder;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Set;
import java.util.List;
import java.util.logging.Level;
import java.util.logging.Logger;


/**	Translates bytes into a <code>FormDataSet</code>, while taking account of the character set of the information 
 * 	represented by those bytes.
 * 	
 * 	@author Chris W. Johnson
 * 	@see <a href='http://www.w3.org/TR/html4/appendix/notes.html#h-B.2'>HTML 4 Specification, Section B.2, Special characters in URI attribute values</a>
 * 	@see <a href='http://www.rfc-editor.org/rfc/rfc1738.txt'>RFC 1738 - Uniform Resource Locators</a>
 */
final class FormDataSetDecoder {

	private static final Logger				Log = Logger.getLogger(FormDataSetDecoder.class.getName());

	private static CharacterSetRecognizer 	QwicapCharSetControlNameRecognizer;

	private final FormControlBytes[]		FCBs;
	private final String					CharSetName;
	
	
	FormDataSetDecoder
	(
		final List<FormControlBytes>	FCBs,
		final String					DefaultCharSetName	//	Can be null.
	)
	{
		this.FCBs        = FCBs.toArray(new FormControlBytes[FCBs.size()]);
		this.CharSetName = identifyCharSet(this.FCBs, DefaultCharSetName);
	}
	
	FormDataSetDecoder
	(
		final String			Encoded,
		final String			DefaultCharSetName	//	Can be null.
	)
	{
	//	The characters in the string are all Unicode characters, but URLs are required to use only a subset of the 
	//	printable portion of the US-ASCII character set. (See RFC 1738, section 2.2.) Therefore, there should be no 
	//	Unicode characters in the string that are not also present in the legal subset of the US-ASCII character set 
	//	used by all URLs.
	//
	//	Note that while URLs are expressed in a subset of US-ASCII, those US-ASCII characters may be representing data
	//	in any character set. Character codes in the actual character set that happen to correspond to "safe" US-ASCII 
	//	characters can be placed directly into the URL. Charactes outside of the safe range are hex-encoded character 
	//	codes of the actual chracter set.
		
		this(stringToUSASCII(Encoded), DefaultCharSetName);
	}
	
	private static byte[] stringToUSASCII
	(
		final String			Encoded
	)
	{
		try {
			return Encoded.getBytes("US-ASCII");
		} catch (UnsupportedEncodingException e) {
			throw new RuntimeException(e);	
		}
	}
	
	FormDataSetDecoder
	(
		final byte[]			Encoded,
		final String			DefaultCharSetName	//	Can be null.
	)
	{
	//	For details of the "applicationx/www-form-urlencoded" content type, see:
	//
	//	See: <http://www.w3.org/TR/html4/interact/forms.html#didx-applicationx-www-form-urlencoded>
	//	See: <http://www.ietf.org/rfc/rfc1738.txt>, section 2.2, "URL Encoding Issues."

		final int								EncodedLength    = Encoded.length;
		final byte[]							Decoded          = new byte[EncodedLength];
		final ArrayList<SignificantCharacter>	SignificantChars = new ArrayList<SignificantCharacter>(EncodedLength >> 4);
		int										DecodedIndex     = 0;

		for (int EncodedIndex = 0; EncodedIndex < EncodedLength; EncodedIndex++) {
			final int				ASCIIChar = Encoded[EncodedIndex] & 0xFF;

			switch (ASCIIChar) {
				
				case ';':
					
				//	RFC 1866, section 8.2.1, suggests that semicolon (';') characters should be supported as synonyms 
				//	for ampersands in the "search" portion of HTTP URLs. See: <http://www.ietf.org/rfc/rfc1866.txt>. 
				//	Also, RFC 1738 states that the semicolon character has a reserved meaning within that portion of 
				//	URLs, which is consistent with the notion that the suggestion regarding semicolons is meant to be 
				//	supported in practice.
				//
				//	This behavior is unambiguously recommended in the HTML 4 specification. See: 
				//	<http://www.w3.org/TR/html4/appendix/notes.html#ampersands-in-uris>
					
				case '&':
					
				// 	We pass an ampersand, even if we received a semicolon, in order to simplify the code that 
				//	interprets the SignificantCharacter objects.
					
					SignificantChars.add(new SignificantCharacter('&', DecodedIndex));
					break;
				
				case '=':

					SignificantChars.add(new SignificantCharacter('=', DecodedIndex));
					break;

				case '+':

					Decoded[DecodedIndex++] = ' ';
					break;

				case '%':

					final int			EncodedBytesRemaining = EncodedLength - EncodedIndex;

					if (EncodedBytesRemaining >= 2) {
						final int			ASCIIChar2 = Encoded[++EncodedIndex] & 0xFF;

						try {
							Decoded[DecodedIndex++] = (byte) (Hexadecimal.toNibble(ASCIIChar2) << 4 | Hexadecimal.toNibble(Encoded[++EncodedIndex] & 0xFF));
						} catch (NumberFormatException e) {
							try {
								Log.log(Level.WARNING, "The following encoded form data set appears to be improperly constructed: " + new String(Encoded, "US-ASCII"), e);
							} catch (UnsupportedEncodingException e2) {
								Log.log(Level.SEVERE, "The \"US-ASCII\" character set is not supported. This should never happen.", e2);
							}
						}
					}
					break;

				default:

					Decoded[DecodedIndex++] = (byte) ASCIIChar;
			}
		}

		final byte[]			DecodedBytes = new byte[DecodedIndex];
		
		System.arraycopy(Decoded, 0, DecodedBytes, 0, DecodedIndex);

		FCBs        = getFormControlBytesArray(DecodedBytes, SignificantChars);
		CharSetName = identifyCharSet(FCBs, DefaultCharSetName);
	}
	
	private static String identifyCharSet
	(
		final FormControlBytes[]	FCBs,
		final String				DefaultCharSetName	//	Can be null.
	)
	{
		String					CharSetName = null;
		
		try {
			if (QwicapCharSetControlNameRecognizer == null)
				QwicapCharSetControlNameRecognizer = new CharacterSetRecognizer();
			
			if (FCBs.length > 0)
				CharSetName = QwicapCharSetControlNameRecognizer.identifyCharSetOf(FCBs);
			
			if (CharSetName == null && DefaultCharSetName == null)
				Log.log(Level.SEVERE, "The character set of the form data set could not be identified; defaulting to ISO-8859-1.");

		} catch (UnsupportedEncodingException e) {
			
			if (DefaultCharSetName != null)
				Log.log(Level.SEVERE, "The character set of the form data set is not supported; defaulting to " + DefaultCharSetName + '.', e);
			else
				Log.log(Level.SEVERE, "The character set of the form data set is not supported; defaulting to ISO-8859-1.", e);
		}
		
		if (CharSetName == null)
			CharSetName = DefaultCharSetName != null ? DefaultCharSetName : "ISO-8859-1";
		
		return CharSetName;
	}
	
	String getCharacterSet() {
		return CharSetName;
	}
	
	FormDataSetDecoder populateFormDataSet
	(
		final FormDataSet		DataSet
	)	
	{
		try {
			final HashMap<String, ArrayList<String>>	NamesAndValues = new HashMap<String, ArrayList<String>>(FCBs.length);
			
			{
				final CharsetDecoder	Decoder = Charset.forName(CharSetName).newDecoder();
				
				for (final FormControlBytes FCB : FCBs) {
					final String			CtrlName  = FCB.Name.toString(Decoder);		//	Decoder.decode(FCB.Name.getByteBuffer() ).toString();
					final String			CtrlValue = FCB.Value.toString(Decoder);	//	Decoder.decode(FCB.Value.getByteBuffer()).toString();
					ArrayList<String>		Values    = NamesAndValues.get(CtrlName);
					
					if (Values == null) {
					
						Values = new ArrayList<String>();
						NamesAndValues.put(CtrlName, Values);
					}
					
					Values.add(CtrlValue);
				}
			}
			
			final Set<String>		Keys = NamesAndValues.keySet();
			
			for (final String CtrlName : Keys) {
				final ArrayList<String>	CtrlValues = NamesAndValues.get(CtrlName);
				
				DataSet.addData(CtrlName, (Object[]) CtrlValues.toArray(new String[CtrlValues.size()]));
			}
			
			Log.log(Level.FINEST, "Form data set decoded using the {0} character set:\n{1}", new Object[] { CharSetName, DataSet });
			
		} catch (CharacterCodingException e) {
			Log.log(Level.INFO, "Failed to decode a control name or value.", e);
		}
		
		return this;
	}

	private static FormControlBytes[] getFormControlBytesArray
	(
		final byte[]							DecodedBytes,
		final ArrayList<SignificantCharacter>	SigChars
	)
	{
		final int							SigCharsCount = SigChars.size();
		final ArrayList<FormControlBytes>	FCBs          = new ArrayList<FormControlBytes>(SigCharsCount >> 1);
		int									AmpIndex      = 0;
		int									EqIndex       = 0;

		for (final SignificantCharacter SigChar : SigChars) {
			final int				NewIndex = SigChar.Offset;
			
			if (SigChar.ASCIIChar == '&') {

				if (EqIndex != 0)
					FCBs.add(new FormControlBytes(DecodedBytes, AmpIndex, EqIndex, NewIndex));
				else	//	the control name is present, but the value is missing.
					FCBs.add(new FormControlBytes(DecodedBytes, AmpIndex, NewIndex, NewIndex));

				AmpIndex = NewIndex;
				EqIndex  = 0;

			} else if (EqIndex == 0)	//	ASCIIChar must be '='.
				EqIndex = NewIndex;
		}

		if (EqIndex != 0)
			FCBs.add(new FormControlBytes(DecodedBytes, AmpIndex, EqIndex, DecodedBytes.length));
		else	//	the control name is present, but the value is missing.
			FCBs.add(new FormControlBytes(DecodedBytes, AmpIndex, DecodedBytes.length, DecodedBytes.length));

		return FCBs.toArray(new FormControlBytes[FCBs.size()]);
	}

	private static class SignificantCharacter {
		
		final int				ASCIIChar;
		final int				Offset;

		SignificantCharacter
		(
			final int				ASCIIChar,
			final int				Offset
		)
		{
			this.ASCIIChar = ASCIIChar;
			this.Offset    = Offset;
		}
	}
}
