/*	QWICAP (Quick Web Interface for Conventional Applications) by Chris W. Johnson
	Copyright (C) 2008-2010 Chris W. Johnson. All rights reserved.
	
	Chris W. Johnson's copyright notice in this evolving version reflects his original authorship. It does not assert 
	copyright over changes contributed by others.
	
	This library is free software; you can redistribute it and/or modify it under the terms of the GNU Lesser General 
	Public License as published by the Free Software Foundation; either version 2.1 of the License, or (at your option) 
	any later version.
	
	This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied 
	warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for 
	more details.
	
	You should have received a copy of the GNU Lesser General Public License along with this library; if not, write to 
	the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/


package edu.utexas.its.eis.tools.qwicap.servlet;


import edu.utexas.ots.chrisj.util.Hexadecimal;

import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.nio.charset.Charset;
import java.nio.charset.CharsetEncoder;
import java.nio.CharBuffer;
import java.nio.ByteBuffer;


/**	Performs the "applicationx/www-form-urlencoded" encoding of the contents a <code>FormDataSet</code> whose contents
 * 	are represented in an arbitrary character set. The result is the data set reduced to a safe subset of the printable
 * 	portion of the US-ASCII character set, as specified in RFC 1738, section 2.2.
 * 
 * 	@author Chris W. Johnson
 * 	@see <a href='http://www.w3.org/TR/html4/appendix/notes.html#h-B.2'>HTML 4 Specification, Section B.2, Special characters in URI attribute values</a>
 * 	@see <a href='http://www.rfc-editor.org/rfc/rfc1738.txt'>RFC 1738 - Uniform Resource Locators</a>
 */
final class FormDataSetEncoder {
	
	private static final byte[][] SafeASCIIByteEncodings = new byte[256][];
	
	static {
		
	//	From RFC 1738, section 2.2: ``only alphanumerics, the special characters "$-_.+!*'(),", and reserved characters 
	//	used for their reserved purposes may be used unencoded within a URL.''
		
		for (int Byte = 0; Byte < 256; Byte++) {
			
			if (Byte >= 'A' && Byte <= 'Z' 
			||  Byte >= 'a' && Byte <= 'z'
			||  Byte >= '0' && Byte <= '9'
			||  Byte == '$' || Byte == '-'  || Byte == '_' || Byte == '.' || Byte == '!' 
			||	Byte == '*' || Byte == '\'' || Byte == '(' || Byte == ')' || Byte == ',') {
				
			//	Note that we omit the plus character (+) from the list of characters above, which means that it will be
			//  encoded when found in a control name or valur. This preserves its use as a replacement for the space 
			//	character.
				
				SafeASCIIByteEncodings[Byte] = new byte[] { (byte) Byte };
				
			} else
				SafeASCIIByteEncodings[Byte] = new byte[] { (byte) '%', (byte) Hexadecimal.fromNibble(Byte >> 4), (byte) Hexadecimal.fromNibble(Byte & 0x0F) };
		}
		
	//	Override the general-purpose hex encodings for characters that have special encodings.
		
		SafeASCIIByteEncodings[' '] = new byte[] { (byte) '+' };
		
	//	Note thae the HTML 4 specification, section 17.13.4, specifies that line breaks should encoded as "CR LF" 
	//	pairs (i.e., `%0D%0A'). So, what do we consider to be line breaks? And how do we allow for the fact that the
	//	byte(s) we're encoding may not represent characters in the US-ASCII character set, and therefore what appears to
	//	be a line break (when the bytes are interpreted as US-ASCII characters) may not, in fact be a line break. In 
	//	such a case encoding what appears to be a line break, but isn't, as "%0D%0A" would actually corrupt the data.
	}
	
	
	private final String		CharSet;
	private final byte[]		EncodedBytes;	//	Composed only of printable US-ASCII characters.
	private final String		EncodedString;	//	Composed only of Unicode characters corresponding to printable US-ASCII characters.
	
	
	FormDataSetEncoder
	(
		final FormDataSet		FDS,
		final String			CharSetName,
		final char				URIAttrSepChar	//	Usually '&', but ';' is also acceptable to properly written web servers and/or CGIs.
	)
	throws IOException
	{
		final ByteArrayOutputStream	BAOS      = new ByteArrayOutputStream(1024);
		final CharsetEncoder		Encoder   = Charset.forName(CharSetName).newEncoder();
		int							AttrIndex = 0;
		
		for (final FormControlValue FCV : FDS) {
			final String			Name      = FCV.getName();
			final ByteBuffer		NameBytes = Encoder.encode(CharBuffer.wrap(Name));
						
			for (final String Value : FCV.getValues()) {
				final ByteBuffer		ValueBytes = Encoder.encode(CharBuffer.wrap(Value));
				
				if (AttrIndex++ != 0)
					BAOS.write(URIAttrSepChar);	//	We could use an ampersand (&) here, but it requires encoding when used within HTML documents.
				
				writeAsSafeASCII(NameBytes, BAOS);
				BAOS.write('=');
				writeAsSafeASCII(ValueBytes, BAOS);
			}
		}
		
		CharSet       = CharSetName;
		EncodedBytes  = BAOS.toByteArray();
		EncodedString = BAOS.toString("US-ASCII");
	}
	
	private static void writeAsSafeASCII
	(
		final ByteBuffer			Buff,
		final ByteArrayOutputStream	BAOS
	)
	throws IOException
	{
		final int				Length = Buff.limit();
		
		if (Buff.hasArray()) {
			final byte[]			Bytes  = Buff.array();
			
			for (int Index = 0; Index < Length; Index++)
				BAOS.write(SafeASCIIByteEncodings[Bytes[Index] & 0xFF]);
	
		} else {
			
		//	This case should never arise, but we'll implement it for the sake of good form.
			
			for (int Index = 0; Index < Length; Index++)
				BAOS.write(SafeASCIIByteEncodings[Buff.get(Index) & 0xFF]);
		}
	}
	
	public String getDataSetCharSetName() {
		return CharSet;	
	}
	
	public byte[] toBytes() {
		return EncodedBytes;	
	}
	
	@Override public String toString() {
		return EncodedString;	
	}
}
