/*	QWICAP (Quick Web Interface for Conventional Applications) by Chris W. Johnson
	Copyright (C) 2003-2010 The University of Texas at Austin ("U.T. Austin"). All rights reserved.
	
	U.T. Austin's copyright notice in this evolving version reflects our original authorship. It does not assert 
	copyright over changes contributed by others.
	
	This library is free software; you can redistribute it and/or modify it under the terms of the GNU Lesser General 
	Public License as published by the Free Software Foundation; either version 2.1 of the License, or (at your option) 
	any later version.
	
	This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied 
	warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for 
	more details.
	
	You should have received a copy of the GNU Lesser General Public License along with this library; if not, write to 
	the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/


package edu.utexas.its.eis.tools.qwicap.servlet;


/**	Defines the actions Qwicap is to take after executing a {@link FormDataSetFilter} implementation. By default,
 * 	instances of this class specify no actions, and therefore cause nothing to happen. By invoking the public methods 
 * 	of an instance, you construct a sequential list of actions to be performed when your {@link FormDataSetFilter#filter} 
 * 	method exits. Consider, however, that not all combinations of actions will produce useful, desirable or 
 * 	comprehensible results. For instance, having more than one <code>goBack...</code> action in a response would be 
 * 	senseless, and should not be considered a supported option. (Going backward, in sequence, through a series of pages 
 * 	would work, but would produce exactly the same result as going back directly to the final page in the series. Going 
 * 	backward out of sequence would produce an exception.)
 * 
 * 	@author Chris W. Johnson
 */
public class FormDataSetFilterResponse {
	
	private final PageStack		PgStack;
	private final FormDataSet	DataSet;	//	If we need this at all, we need it so we can have actions clear the data set.
	
	private FilterAction		Actions;
	private FilterAction		Last;
	
	
	FormDataSetFilterResponse
	(
		final PageStack			PgStack,
		final FormDataSet		DataSet
	)
	{
		this.PgStack = PgStack;
		this.DataSet = DataSet;
	}

/**	Immediately executes the set of actions described by this response object.
 * 
 * @throws QwicapAbandonmentException	If Qwicap needs to assert control over the client application.
 */
	void execute() throws QwicapAbandonmentException {
		
		if (Actions != null) {
			
			Actions.executeAll();
			Actions = null;
		}
	}

/**	Removes all actions from this response.
 * 
 * @return						A reference to this object.
 */
	public FormDataSetFilterResponse clear() {
		Actions = null;
		return this;
	}
	
//	public FormDataSetFilterResponse exit() {
//		
//		addAction(new FilterActionExit());
//		return this;
//	}

/**	Instructs Qwicap to back to the first (earliest) page that is still "in play".
 * 
 * @return						A reference to this object.
 * @throws QwicapException		Thrown if something goes wrong, like there being no pages "in play" at all.
 */
	public FormDataSetFilterResponse goBackToFirstPage() throws QwicapException {
		
		addAction(new FilterActionGoBackToFirstPage(PgStack));
		return this;
	}

/**	Instructs Qwicap to go back to a page that is still "in play". The page is identified by the name of the markup
 * 	it is using.
 * 
 * @param MarkupName			The name of the markup used by the page to which Qwicap should return.
 * @return						A reference to this object.
 * @throws QwicapException		Thrown if something goes wrong, like there being no "in play" page whose markup has
 * 								the specified name.
 */
	public FormDataSetFilterResponse goBackToPage
	(
		final String			MarkupName
	)
	throws QwicapException
	{
		addAction(new FilterActionGoBackToNamedMarkup(PgStack, MarkupName));
		return this;
	}
	
/**	Instructs Qwicap to go back to a page that is still "in play". The page is identified by the class, and the 
 *	name of the method, that implements the page's prompt pattern.
 * 
 * @param PromptClass			The class containing the method that implements the page's prompt pattern.
 * @param PromptMethodName		The name of the method in <code>PromptClass</code> that implements the page's prompt 
 *								pattern. Note that the name of constructors is always "<code>&lt;init&gt;</code>".
 * @return						A reference to this object.
 * @throws QwicapException		Thrown if something goes wrong, like there being no "in play" page whose implementing
 * 								class and method name match those supplied to this method.
 */
	public FormDataSetFilterResponse goBackToPage
	(
		final Class<?>			PromptClass,
		final String			PromptMethodName
	)
	throws QwicapException
	{
		addAction(new FilterActionGoBackToMethod(PgStack, PromptClass, PromptMethodName));
		return this;
	}
	
/**	Instructs Qwicap to go back to the previous page that is "in play".
 * 
 * @return						A reference to this object.
 * @throws QwicapException		Thrown if something goes wrong, like there being no "in play" page whose markup has
 * 								the specified name.
 */
	public FormDataSetFilterResponse goBackOnePage() throws QwicapException {
		
		addAction(new FilterActionGoBackOnePage(PgStack));
		return this;
	}
	
/**	Instructs Qwicap to invoke the <code>run</code> method of the supplied object. That method is allowed to do 
 *	anything that the rest of your client application's code can do. That includes prompting with a new page. 
 * 
 * @param RunMe					The object whose <code>run</code> method should be invoked.
 * @return						A reference to this object.
 */
	public FormDataSetFilterResponse run
	(
		final FilterRunnable	RunMe
	)
	{
		addAction(new FilterActionRun(PgStack, RunMe));
		return this;
	}
	
/**	Instructs Qwicap to ignore the current form data set. Ultimately, this is accomplished by clearing the current 
 *	form data set.
 * 
 * @return						A reference to this object.
 */
	public FormDataSetFilterResponse ignoreDataSet() {
		
		addAction(new FilterActionIgnoreDataSet(DataSet));
		return this;
	}
	
/**	Instructs Qwicap to process the current form data set normally, and to ignore any further actions specified in this
 *	response object.
 * 
 * @return						A reference to this object.
 */
	public FormDataSetFilterResponse proceedNormally() {
		
		addAction(new FilterActionProceedNormally());
		return this;
	}
	
	private void addAction
	(
		final FilterAction		Action
	) 
	{
		if (Last != null)
			Last.Next = Action;
		else
			Actions = Action;
		
		Last = Action;
	}
	
	
	
/*
	Exit

	Exit, then Run...

	Go Back To Page _____

	Go Back To Page _____, then Run...

	Go Back One Page

	Go Back One Page, then Run...

	Run...

	Ignore Data Set

	Proceed Normally



	Pages can be identified by the Class and method implementing their prompt pattern,
	or by the name of the file from which the page markup was originally read.
	
	The "Go To Page _____" feature can't be implemented in a useful and convenient manner,
	since an arbitrary method will expect arbitrary parameters. We can let the caller 
	supply an array of parameters and reflect accordingly, but it will be more direct to
	just invoke a "go wherever you want" method on the filter, and let the filter implementer
	branch to his method of choice, passing whatever parameters his method requires.
	
	
*/	
}
