/*	QWICAP (Quick Web Interface for Conventional Applications) by Chris W. Johnson
	Copyright (C) 2003-2010 The University of Texas at Austin ("U.T. Austin"). All rights reserved.
	
	U.T. Austin's copyright notice in this evolving version reflects our original authorship. It does not assert 
	copyright over changes contributed by others.
	
	This library is free software; you can redistribute it and/or modify it under the terms of the GNU Lesser General 
	Public License as published by the Free Software Foundation; either version 2.1 of the License, or (at your option) 
	any later version.
	
	This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied 
	warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for 
	more details.
	
	You should have received a copy of the GNU Lesser General Public License along with this library; if not, write to 
	the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/


package edu.utexas.its.eis.tools.qwicap.servlet;


import org.apache.commons.fileupload.FileItem;

import java.util.Iterator;

import edu.utexas.its.eis.tools.qwicap.util.ArrayIterator;


/**	Represents files uploaded via "<samp>&lt;input type='file' &hellip; /&gt;</samp>" elements in 
 * 	<a href='http://www.w3.org/TR/html4/interact/forms.html'>web page forms</a>.
 *	<code>FormFileUpload</code> is a subclass of {@link FormControlValue}, so client code can receive and process 
 *	uploaded files in exactly the same way it processes all other control values. In cases where client code needs to 
 *	be aware that a parameter represents an uploaded file, it can use "<code>instanceof FormFileUpload</code>" to 
 *	differentiate a <code>FormFileUpload</code> object from a <code>FormControlValue</code> object. If the parameter is an 
 *	instance of <code>FormFileUpload</code>, the {@link #getFileItem} method can be used to gain access to the underlying 
 *	{@link FileItem} object from the <a href='http://jakarta.apache.org/commons/fileupload/'>Jakarta Commons FileUpload 
 * 	library</a>.
 *	
 *	@author Chris W. Johnson
 */
public final class FormFileUpload extends FormControlValue {
	
	private static final String[]	kEmptyStringArray = {};
	
	private final FileItem			FI;
	private String					Value;
	
	
/**	Constructs a <code>FormFileUpload</code> object to represent a file uploaded using an 
 *	<a href='http://www.w3.org/TR/html4/interact/forms.html#h-17.4.1'>"input" control</a> of type "file".
 * 
 * @param FI			The underlying <code><a href='http://jakarta.apache.org/commons/fileupload/apidocs/org/apache/commons/fileupload/FileItem.html'>FileItem</a></code>
 *						object that describes, and provides access to the data of, the uploaded file.
 */
	FormFileUpload
	(
		final FileItem			FI
	)
	{
		super(FI.getFieldName(), kEmptyStringArray);
		this.FI = FI;
	}
	
	@Override public String getValue() {
		
		if (Value == null)
			Value = FI.getString();
		
		return Value;
	}
	
	@Override public String[] getValues() {
		
		if (Value == null)
			Value = FI.getString();
		
		return new String[] { Value };
	}
	
	@Override public Iterator<String> iterator() {
		return new ArrayIterator<String>(getValues());
	}
	
	@Override void addValue
	(
		final Object			Value
	)
	{
	//	This makes no sense with file uploads, so do nothing.
	}
	
	@Override void addValues
	(
		final Object[]			Value
	)
	{
	//	This makes no sense with file uploads, so do nothing.
	}
	
/**	Returns the underlying <code><a href='http://jakarta.apache.org/commons/fileupload/apidocs/org/apache/commons/fileupload/FileItem.html'>FileItem</a></code>
 *	object that describes, and provides access to the data of, the uploaded file. For files whose contents can usefully
 *	be treated as strings, using {@link #getValue} will be simpler.
 * 
 * @return				The <code>FileItem</code> object that describes, and provides access to the data of, the 
 *						uploaded file.
 */
	public FileItem getFileItem() {
		return FI;
	}
}
