/*	QWICAP (Quick Web Interface for Conventional Applications) by Chris W. Johnson
	Copyright (C) 2003-2010 The University of Texas at Austin ("U.T. Austin"). All rights reserved.
	
	U.T. Austin's copyright notice in this evolving version reflects our original authorship. It does not assert 
	copyright over changes contributed by others.
	
	This library is free software; you can redistribute it and/or modify it under the terms of the GNU Lesser General 
	Public License as published by the Free Software Foundation; either version 2.1 of the License, or (at your option) 
	any later version.
	
	This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied 
	warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for 
	more details.
	
	You should have received a copy of the GNU Lesser General Public License along with this library; if not, write to 
	the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/


package edu.utexas.its.eis.tools.qwicap.servlet;


import edu.utexas.its.eis.tools.qwicap.template.css.Results;
import edu.utexas.its.eis.tools.qwicap.template.xml.Match;
import edu.utexas.its.eis.tools.qwicap.template.xml.TagException;
import edu.utexas.its.eis.tools.qwicap.util.ArrayToString;

import java.util.ArrayList;


/**
 * Created by IntelliJ IDEA.
 * User: chrisj
 * Date: Aug 23, 2004
 * Time: 7:46:22 PM
 * To change this template use Options | File Templates.
 */
final class FormInputRejectionHelper {
	
	private static final boolean	kTidyUpLabelNames = true;	//	If true, trailing colons are stripped, and maybe other things.
	
	private Match					EarliestEnclosingElem;
	private int						EarliestEnclosingElemIndex = Integer.MAX_VALUE;
	private final ArrayList<String>	Labels = new ArrayList<String>();


	FormInputRejectionHelper
	(
		final Page				P,
		final String[]			ControlNames,
		final Object[]			ControlValues		//	Must not be null, but all of its elements may be null.
	)
	throws TagException, FormNotFoundException
	{
	//	Get the labels for the controls that supplied the bad parameters, then substitute them into the error
	//	message. Also mark the elements that contained those labels with the kQwicapBadInputFieldClass class, so that
	//	they can be highlighted, if desired, via the document's stylesheet.
		
		final Form				MostRecentForm = P.getMostRecentForm();
		final Results			OtherForms     = P.getMarkup().get("form");
		
		if (MostRecentForm != null)
			OtherForms.xor(MostRecentForm.toResults());
		
		for (int Index = 0, Count = ControlNames.length; Index < Count; Index++) {
			final String			ControlName  = ControlNames[Index];
			final Object			ControlValue = ControlValues[Index];
			
		//	If the most recent form includes a form element that supplies the specified parameter, mark the element
		//	as bad, and proceed to the next iteration of the enclosing loop.
			
			if (MostRecentForm != null && markFormElementsAsBad(MostRecentForm, ControlName, ControlValue))
				continue;
				
		//	The most recent form does NOT include a control that supplies the specified parameter. Therefore,
		//	we search all of the other Forms in the document for such controls and mark all of them as bad. It is 
		//	possible that we'll find controls in more than one form that could have supplied a given parameter.
		//	We'll mark them all as bad. This may be confusing to the user, but it correctly reflects the ambiguity
		//	with which Qwicap is faced. Web application developers can eliminate this ambiguity by uniquely naming
		//	the controls in all of the forms on any given page.
			
			boolean					FoundControl = false;
			
			for (final Match Elem : OtherForms) {
				
				if (markFormElementsAsBad(new Form(Elem), ControlName, ControlValue))
					FoundControl = true;
			}
			
			if (!FoundControl)
				throw new TagException("There is no control named \"" + ControlNames[Index] + "\" in any of the forms on this page.");
		}
		
		if (EarliestEnclosingElem == null)
			throw new TagException("FormInputRejectionHelper was unable to find an enclosing element for the " +
				"control(s) " + new ArrayToString(ControlNames).alwaysQuote() + " having the value(s) " + 
				new ArrayToString(ControlValues).alwaysQuote()  + ". In other words, at least one form control " +
				"appears not to be enclosed within a \"form\" element in document \"" + P.getMarkup().getMarkupName() + 
				"\"."
			);
	}
	
	Match getEarliestEnclosingElement() {
		return EarliestEnclosingElem;
	}
	
	int getLabelCount() {
		return Labels.size();
	}
	
	String getLabelsAsString() {
		return getLabelsAsString(true);
	}
	
	String getLabelsAsString
	(
		final boolean			UseQuotes
	) 
	{
		final ArrayToString		Lbls = new ArrayToString(Labels);
		
		if (UseQuotes)
			Lbls.setQuotes(Qwicap.kLeftDoubleQuote, Qwicap.kRightDoubleQuote);
		else
			Lbls.neverQuote();
		
		return Lbls.trim().toString();
	}
	
	private boolean markFormElementsAsBad
	(
		final Form				TargetForm,
		final String			ControlName,
		final Object			ControlValue
	)
	throws TagException
	{
		final FormControlInfo[]	ElemInfo = TargetForm.getControls(ControlName, ControlValue);
		
		for (int Index = 0, Count = ElemInfo.length; Index < Count; Index++) {
			final FormControlInfo	Info = ElemInfo[Index];
			
			Info.getLabels().addClass(Qwicap.kQwicapBadInputFieldClass);
			
			rememberLabel(Info);
			
			final Match				EnclosingElem      = Info.getEnclosingElement();
			final int				EnclosingElemIndex = EnclosingElem.getIndex();
			
			if (EarliestEnclosingElemIndex > EnclosingElemIndex) {
				
				EarliestEnclosingElemIndex = EnclosingElemIndex;
				EarliestEnclosingElem      = EnclosingElem;
			}
		}
		
		return ElemInfo.length > 0;
	}
	
	private void rememberLabel
	(
		final FormControlInfo		Info
	)
	{
		final String				NewLabel = Info.getLabelText(kTidyUpLabelNames);
		
		if (NewLabel != null && NewLabel.length() > 0) {
			
		//	Add the new label to our list of label strings, unless there is an identical label string already in the list.
			
			for (int Index = 0, Count = Labels.size(); Index < Count; Index++) {
				
				if (Labels.get(Index).equals(NewLabel))
					return;
			}
			
			Labels.add(NewLabel);
		}
	}
}
