/*	QWICAP (Quick Web Interface for Conventional Applications) by Chris W. Johnson
	Copyright (C) 2007-2011 The University of Texas at Austin ("U.T. Austin"). All rights reserved.
	
	U.T. Austin's copyright notice in this evolving version reflects our original authorship. It does not assert 
	copyright over changes contributed by others.
	
	This library is free software; you can redistribute it and/or modify it under the terms of the GNU Lesser General 
	Public License as published by the Free Software Foundation; either version 2.1 of the License, or (at your option) 
	any later version.
	
	This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied 
	warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for 
	more details.
	
	You should have received a copy of the GNU Lesser General Public License along with this library; if not, write to 
	the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/


package edu.utexas.its.eis.tools.qwicap.servlet;


import javax.servlet.http.HttpServletRequest;


/**	Represents the contents of the HTTP "Accept-Charset" header.
 * 
 * 	@see <a href='http://www.w3.org/Protocols/rfc2616/rfc2616-sec14.html#sec14.2'>RFC 2616, Section 14.2</a>
 * 	@author Chris W. Johnson
 */
final class HTTPAcceptCharsetList extends HTTPAcceptList<HTTPCharsetIdentifier> {
	
	private static final HTTPAcceptCharsetFactory	Factory = new HTTPAcceptCharsetFactory();
	
	private final boolean			AutomaticallyAddedISO_8859_1;
	
	
	HTTPAcceptCharsetList() {
		
		super(Factory);
		AutomaticallyAddedISO_8859_1 = false;
	}

	HTTPAcceptCharsetList
	(
		final HttpServletRequest	Req
	)
	{
		this(Req.getHeader("Accept-Charset"));
	}

	HTTPAcceptCharsetList
	(
		final String			AcceptCharsetStr
	)
	{
		super(Factory, AcceptCharsetStr);

		if (AcceptCharsetStr != null) {
			
		//	If not specified, the ISO-8859-1 character set is assumed to be accepted, and is assigned a quality of 1.0.
			
			if (explicitlyIncludes("ISO-8859-1") == false) {
				
				add("ISO-8859-1", 1.0);
				AutomaticallyAddedISO_8859_1 = true;
			
			} else
				AutomaticallyAddedISO_8859_1 = false;

		} else {

		//	"If no Accept-Charset header is present, the default is that any character set is acceptable."
			
			add("*");
			AutomaticallyAddedISO_8859_1 = false;
		}
	}
	
	@Override protected boolean explicitlyIncludes
	(
		final String			CharSet
	)
	{
		if (AutomaticallyAddedISO_8859_1 && "ISO-8859-1".equalsIgnoreCase(CharSet))
			return false;
		
		return super.explicitlyIncludes(CharSet);
	}

	String getBestDefault() {

	//	Unicode character sets are always our first choice, because they provide our only guarantee of being able to 
	//	represent all characters. If one is explicity mention in the "Accept-Charset" list, we use that one.

		if (explicitlyAccepts("UTF-8"   ))
			return "UTF-8";

		if (explicitlyAccepts("UTF-16"  ))
			return "UTF-16";

		if (explicitlyAccepts("UTF-16BE"))
			return "UTF-16BE";

		if (explicitlyAccepts("UTF-16LE"))
			return "UTF-16LE";

	//	None of the Unicode character sets is explicity listed as supported. But, if all character sets have been 
	//	declared acceptable, we're justified in using our default choice (UTF-8).

		if (acceptsEverything())
			return QwicapServlet.DefaultCharset;

	//	If all else fails, there's always ISO-8859-1; the HTTP specification requires that it be supported.

		return "ISO-8859-1";
	}
}
