/*	QWICAP (Quick Web Interface for Conventional Applications) by Chris W. Johnson
	Copyright (C) 2007-2011 The University of Texas at Austin ("U.T. Austin"). All rights reserved.
	
	U.T. Austin's copyright notice in this evolving version reflects our original authorship. It does not assert 
	copyright over changes contributed by others.
	
	This library is free software; you can redistribute it and/or modify it under the terms of the GNU Lesser General 
	Public License as published by the Free Software Foundation; either version 2.1 of the License, or (at your option) 
	any later version.
	
	This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied 
	warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for 
	more details.
	
	You should have received a copy of the GNU Lesser General Public License along with this library; if not, write to 
	the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/


package edu.utexas.its.eis.tools.qwicap.servlet;


import java.util.regex.Pattern;


/**	Represents an element in the lists conveyed in the HTTP headers "Accept-Charset", "Accept-Language", etc.
 * 
 * 	@author Chris W. Johnson
 */
class HTTPAcceptElement<T extends HTTPAcceptIdentifier> implements Comparable<HTTPAcceptElement<T>> {
	
	private static final Pattern	kQualityPattern    = Pattern.compile("\\s*;\\s*[qQ]=\\s*");
	private static final int		kQualityMultiplier = 10000;	//	Preserve four decimal places, and discard the rest.
	private static final int		kQualityOne        = kQualityMultiplier;
	
	private final int				ListIndex;
	private final T					Identifier;
	private final int				Quality;
//	private final BigDecimal		Quality;
	private final boolean			QualityIsGreaterThanZero;
	private final String			CanonicalStr;


	HTTPAcceptElement
	(
		final HTTPAcceptFactory<T>	Factory,
		final int					ListIndex,
		final String				Code
	)
	{
		this.ListIndex = ListIndex;

		final String[]			Parts = kQualityPattern.split(Code);

		Identifier = Factory.createIdentifier(Parts[0]);

		if (Parts.length > 1)
			Quality = (int) (Double.parseDouble(Parts[1]) * kQualityMultiplier);
		else
			Quality = kQualityOne;
			
		QualityIsGreaterThanZero = Quality > 0;
		
		CanonicalStr = createString(Identifier, Quality);
	}

	HTTPAcceptElement
	(
		final HTTPAcceptFactory<T>	Factory,
		final int					ListIndex,
		final String				IdentifierStr,
		final double				Quality
	)
	{
		this.ListIndex                = ListIndex;
		this.Identifier               = Factory.createIdentifier(IdentifierStr);
		this.Quality                  = (int) (Quality * kQualityMultiplier);
		this.QualityIsGreaterThanZero = Quality > 0;
		this.CanonicalStr             = createString(Identifier, this.Quality);
	}
	
	private static <T extends HTTPAcceptIdentifier> String createString
	(
		final T					Identifier,
		final int				Quality
	)
	{
		if (Quality == kQualityOne)
			return Identifier.toString();

		return Identifier + ";q=" + (double) Quality / kQualityMultiplier;
	}
	
	int getListIndex() {
		return ListIndex;
	}
	
	double getQuality() {
		return (double) Quality / kQualityMultiplier;
	}
	
	int getQualityInt() {
		return Quality;
	}
	
	boolean qualityIsGreaterThanZero() {
		return QualityIsGreaterThanZero;
	}
	
	T getIdentifier() {
		return Identifier;
	}
	
	@Override public String toString() {
		return CanonicalStr;
	}
	
	@Override public int hashCode() {
		return toString().hashCode();
	}
	
	@Override public boolean equals
	(
		final Object			Other
	)
	{
		return Other instanceof HTTPAcceptElement && compareTo((HTTPAcceptElement<T>) Other) == 0;
	}
	
	@Override public int compareTo
	(
		final HTTPAcceptElement<T>	Other
	)
	{
		int						Difference;

		Difference = signum(Quality, Other.Quality);
		if (Difference != 0)
			return Difference;
		
		Difference = signum(Other.ListIndex, ListIndex);
		if (Difference != 0)
			return Difference;
		
		return Identifier.compareTo(Other.Identifier);
	}
	
	private static int signum
	(
		final int				A,
		final int				B
	)
	{
		if (A == B)
			return 0;
		
		if (A < B)
			return -1;
		
		return 1;
	}
}