/*	QWICAP (Quick Web Interface for Conventional Applications) by Chris W. Johnson
	Copyright (C) 2007 The University of Texas at Austin ("U.T. Austin"). All rights reserved.
	
	U.T. Austin's copyright notice in this evolving version reflects our original authorship. It does not assert 
	copyright over changes contributed by others.
	
	This library is free software; you can redistribute it and/or modify it under the terms of the GNU Lesser General 
	Public License as published by the Free Software Foundation; either version 2.1 of the License, or (at your option) 
	any later version.
	
	This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied 
	warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for 
	more details.
	
	You should have received a copy of the GNU Lesser General Public License along with this library; if not, write to 
	the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/


package edu.utexas.its.eis.tools.qwicap.servlet;


import java.util.Locale;


/**	Represents the contents of the HTTP "Accept-Language" header.
 * 
 * 	@see <a href='http://www.w3.org/Protocols/rfc2616/rfc2616-sec14.html#sec14.4'>RFC 2616, Section 14.4</a>
 * 	@author Chris W. Johnson
 */
final class HTTPAcceptLanguageList extends HTTPAcceptList<HTTPLanguageRange> {

	private static final HTTPAcceptLanguageFactory	Factory = new HTTPAcceptLanguageFactory();


	HTTPAcceptLanguageList() {
		super(Factory);
	}

	HTTPAcceptLanguageList
	(
		final String			CodeAndQualitiesListStr
	)
	{
		super(Factory, CodeAndQualitiesListStr);
	}

	HTTPLanguageTag getBestLanguageTag
	(
		final i18nLocalizationsList	WebAppLocales
	)
	{
		final HTTPLanguageTagsWithDefault	WebAppTranslations = WebAppLocales.getTranslations();
		final HTTPLanguageTag[]				WebAppLangTags     = WebAppTranslations.toArray();

	//	Apply the HTTP language ranges in this list to locales supported by the web application. Return the first match,
	//	which will also be the highest-quality match.

		if (acceptsEverything() == false) {

		//	First, search for exact matches.

			for (final HTTPAcceptElement<HTTPLanguageRange> AcceptElem : getDescendingElementList()) {

				if (!AcceptElem.qualityIsGreaterThanZero())
					break;

				final HTTPLanguageTag	MatchedTag = AcceptElem.getIdentifier().getBestExactMatch(WebAppLangTags);

				if (MatchedTag != null)
					return MatchedTag;
			}

		//	Second, settle for partial matches.

			for (final HTTPAcceptElement<HTTPLanguageRange> AcceptElem : getDescendingElementList()) {

				if (!AcceptElem.qualityIsGreaterThanZero())
					break;

				final HTTPLanguageTag	MatchedTag = AcceptElem.getIdentifier().getBestPartialMatch(WebAppLangTags);

				if (MatchedTag != null)
					return MatchedTag;
			}
		}
		
	//	None of the HTTP language ranges in this list matched a locale supported by the web application.
		
		HTTPLanguageTag			Match;
		
	//	First, see if the web application marked one of its translations as the default. If so, use that.
		
		Match = WebAppTranslations.getDefault();
		if (Match != null)
			return Match;
		
	//	Unfortunately, no translation was marked as the default, and we still must choose some locale to work in, so 
	//	try using the JVM's default locale as a language range, and see if that matches any locale supported by the web 
	//	application.

		final HTTPLanguageRange	DefaultLocaleRange = new HTTPLanguageRange(Locale.getDefault());
		
		Match = DefaultLocaleRange.getBestExactMatch(WebAppLangTags);
		if (Match == null)
			Match = DefaultLocaleRange.getBestPartialMatch(WebAppLangTags);		//	Settle for a partial match with the default locale, if necessary.
		
		if (Match == null) {
		
		//	The JVM's default locale matched no locale supported by the web application, so settle for our last chance 
		//	default, which is likely to be the first translation of any kind that was found in this web application.
			
			Match = WebAppLocales.getTranslations().getLastChanceDefault();
		}
		
		return Match;	//	One way or another, Match will never be null.
	}
}
