/*	QWICAP (Quick Web Interface for Conventional Applications) by Chris W. Johnson
	Copyright (C) 2007 The University of Texas at Austin ("U.T. Austin"). All rights reserved.
	
	U.T. Austin's copyright notice in this evolving version reflects our original authorship. It does not assert 
	copyright over changes contributed by others.
	
	This library is free software; you can redistribute it and/or modify it under the terms of the GNU Lesser General 
	Public License as published by the Free Software Foundation; either version 2.1 of the License, or (at your option) 
	any later version.
	
	This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied 
	warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for 
	more details.
	
	You should have received a copy of the GNU Lesser General Public License along with this library; if not, write to 
	the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/


package edu.utexas.its.eis.tools.qwicap.servlet;


import junit.framework.TestCase;

import java.util.Set;
import java.util.LinkedHashSet;
import java.util.Locale;


/**
 * Created by IntelliJ IDEA.
 * User: chrisj
 * Date: Oct 8, 2007
 * Time: 9:47:34 PM
 * To change this template use File | Settings | File Templates.
 */
public class HTTPAcceptLanguageListJUnit extends TestCase {


//	What we need to do is to determine what translations are present within Qwicap and/or the web application, 
//	then use that list to narrow down the selections from the "Accept-Language" header. The selection should be
//	weighted for languages for which there are translations of "qwicap.properties.xml" AND the web application's 
//	pages and other documents.

	public void test1() {
		final i18nLocaleList		KnownLocales   = new i18nLocaleList();
		final Set<String>			TopLevelRsrcs  = new LinkedHashSet<String>();

		TopLevelRsrcs.add("/de/");
		TopLevelRsrcs.add("/en-us/");
		TopLevelRsrcs.add("/en/");
		TopLevelRsrcs.add("/es/");
		TopLevelRsrcs.add("/META-INF/");
		TopLevelRsrcs.add("/number-guess.css");
		TopLevelRsrcs.add("/number-guess.html");
		TopLevelRsrcs.add("/WEB-INF/");

		final i18nLocalizationsList	Localizations  = new i18nLocalizationsList(KnownLocales, TopLevelRsrcs);
		
		
	//	Ensure that quality levels are being respected by offering many valid (and some invalid) language choices with 
	//	different quality levels.
		
		HTTPAcceptLanguageList		AcceptLangsHdr = new HTTPAcceptLanguageList("en-US, en-gb;q=0.95, en;q=0.9, de;q=0.6, es;q=0.4, da;q=0.3, ru;q=0.1");
		HTTPLanguageTag				BestLocale     = AcceptLangsHdr.getBestLanguageTag(Localizations);
		
		assertTrue("en-us".equals(HTTPLanguageTag.normalize(BestLocale.toString())));
		
		
	//	Ensure that quality levels are being respected, without regard to list item ordering, by offering two valid 
	//	language choices with different quality levels.
		
		AcceptLangsHdr = new HTTPAcceptLanguageList("de;q=0.8, en");
		BestLocale     = AcceptLangsHdr.getBestLanguageTag(Localizations);

		assertTrue("en".equals(BestLocale.toString()));
		
		
	//	Ensure that quality levels are being respected, without regard to list item ordering, by offering two valid 
	//	language choices with different quality levels.
		
		AcceptLangsHdr = new HTTPAcceptLanguageList("de, en;q=0.8");
		BestLocale     = AcceptLangsHdr.getBestLanguageTag(Localizations);

		assertTrue("de".equals(BestLocale.toString()));

		
	//	The following test faces the language selection process with a range that could match two of the locales 
	//	supported by the web application. The HTTP specification isn't clear on how such a case should be handled, so
	//	we've selected the behavior that seems most intuitive: if possible, choose the locale that matches exactly.
		
		AcceptLangsHdr = new HTTPAcceptLanguageList("en");
		BestLocale     = AcceptLangsHdr.getBestLanguageTag(Localizations);
		
		assertTrue("en".equals(BestLocale.toString()));
		
		
	//	The following test faces the language selection process with nothing but a wildcard. Faced with multiple, 
	//	equally correct choices, the choices are narrowed-down by looking for the supported locale that is most 
	//	similar to the JVM's default locale.
		
		AcceptLangsHdr = new HTTPAcceptLanguageList("*");
		BestLocale     = AcceptLangsHdr.getBestLanguageTag(Localizations);

		assertTrue(BestLocale.toString().equals(HTTPLanguageTag.normalize(Locale.getDefault().toString())));
		
		
	//	The following test faces the language selection process with a wildcard, and a choice ('z@') that can never be 
	//	correct, and which can never correspond to a real locale (and thus cannot be the default locale). Therefore it
	//	must use the wildcard to match one of the web application's supported locales. Faced with multiple, equally
	//	correct choices, the choices are narrowed-down by looking for the supported locale that is most similar to the
	//	JVM's default locale.
		
		AcceptLangsHdr = new HTTPAcceptLanguageList("z@, *");
		BestLocale     = AcceptLangsHdr.getBestLanguageTag(Localizations);

		assertTrue(BestLocale.toString().equals(HTTPLanguageTag.normalize(Locale.getDefault().toString())));
		
		
	//	The following test faces the language selection process with an impossible choice; only "z@" is even potentially 
	//	acceptable, but the web application doesn't support it (nothing ever will, because it's a bogus language code). 
	//	When all else fails, our behavior is to use the JVM's default locale. That's what we're checking for here.
		
		AcceptLangsHdr = new HTTPAcceptLanguageList("z@, *;q=0");
		BestLocale     = AcceptLangsHdr.getBestLanguageTag(Localizations);
		
		assertTrue(BestLocale.toString().equals(HTTPLanguageTag.normalize(Locale.getDefault().toString())));
	}
}
