/*	QWICAP (Quick Web Interface for Conventional Applications) by Chris W. Johnson
	Copyright (C) 2007-2011 The University of Texas at Austin ("U.T. Austin"). All rights reserved.
	
	U.T. Austin's copyright notice in this evolving version reflects our original authorship. It does not assert 
	copyright over changes contributed by others.
	
	This library is free software; you can redistribute it and/or modify it under the terms of the GNU Lesser General 
	Public License as published by the Free Software Foundation; either version 2.1 of the License, or (at your option) 
	any later version.
	
	This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied 
	warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for 
	more details.
	
	You should have received a copy of the GNU Lesser General Public License along with this library; if not, write to 
	the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/


package edu.utexas.its.eis.tools.qwicap.servlet;


import edu.utexas.its.eis.tools.qwicap.util.ArrayIterator;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Iterator;
import java.util.regex.Pattern;


/**	Implements all of the functionality common to the HTTP "Accept" headers, specifically the headers "Accept-Charset" 
 * 	and "Accept-Language".
 * 
 * 	@see HTTPAcceptLanguageList
 * 	@see HTTPAcceptCharsetList
 * 	@see <a href='http://www.w3.org/Protocols/rfc2616/rfc2616-sec14.html#sec14.2'>RFC 2616, Section 14.2</a>
 * 	@see <a href='http://www.w3.org/Protocols/rfc2616/rfc2616-sec14.html#sec14.4'>RFC 2616, Section 14.4</a>
 * 
 * 	@author Chris W. Johnson
 */
abstract class HTTPAcceptList<T extends HTTPAcceptIdentifier> implements Iterable<HTTPAcceptElement<T>> {

	private static final Pattern					kCommaPattern = Pattern.compile("\\s*,\\s*");
	
	private final HTTPAcceptFactory<T>				Factory;
	private final ArrayList<HTTPAcceptElement<T>>	CodesInOriginalOrder = new ArrayList<HTTPAcceptElement<T>>();
	private HTTPAcceptElement<T>[]					CodesInDescendingOrder;
	private boolean									AcceptEverything;


//	See <http://www.w3.org/Protocols/rfc2616/rfc2616-sec14.html#sec14.2>.
//	See <http://www.w3.org/Protocols/rfc2616/rfc2616-sec14.html#sec14.4>.

	
	protected HTTPAcceptList
	(
		final HTTPAcceptFactory<T>	Factory
	) 
	{
		
		this.Factory = Factory;
	}

	protected HTTPAcceptList
	(							
		final HTTPAcceptFactory<T>	Factory,
		final String				CodeAndQualitiesListStr
	)
	{
		this.Factory = Factory;
		
		if (CodeAndQualitiesListStr != null) {
			final String[]			Codes = kCommaPattern.split(CodeAndQualitiesListStr);

			for (int Index = 0, Count = Codes.length; Index < Count; Index++)
				add(Codes[Index]);

		} else
			add("*");
	}
	
	HTTPAcceptList<T> acceptEverything() {
		
		AcceptEverything = true;	//	This is only the case when the "Accept" list is omitted.
		return this;
	}
	
	boolean acceptsEverything() {
		return AcceptEverything;
	}
	
	int size() {
		return CodesInOriginalOrder.size();
	}
	
	void add
	(
		final String			Name,
		final double			Quality
	)
	{
		addLow(new HTTPAcceptElement<T>(Factory, CodesInOriginalOrder.size(), Name, Quality));
	}
	
	void add
	(
		final String			CodeAndQuality
	)
	{
		addLow(new HTTPAcceptElement<T>(Factory, CodesInOriginalOrder.size(), CodeAndQuality));
	}
	
	private void addLow
	(
		final HTTPAcceptElement<T>	NewCode
	)
	{
		if (explicitlyIncludes(NewCode.getIdentifier().toString()))
			return;
		
		CodesInOriginalOrder.add(NewCode);
		CodesInDescendingOrder = null;
	
	//	AcceptEverything can only be true when the list is empty due to omission of the relevant HTTP header from 
	//	the HTTP request. If the header is present, but blank, then nothing is acceptable, because everything not in 
	//	the list (which, in the case of blank list, is everything) is implicitly assigned a quality of zero.
		
		AcceptEverything = false;
	}
	
	private HTTPAcceptElement<T>[] getCodesInDescendingOrder() {
		
		if (CodesInDescendingOrder == null) {
			
			CodesInDescendingOrder = CodesInOriginalOrder.toArray(Factory.createElementArray(CodesInOriginalOrder.size()));
			Arrays.sort(CodesInDescendingOrder, new ReverseComparator<HTTPAcceptElement<T>>());
		}
		
		return CodesInDescendingOrder;
	}
	
	boolean accepts
	(
		final String			Name
	)
	{
		return AcceptEverything || explicitlyAccepts(Name);

	}
	
	boolean explicitlyAccepts
	(
		final String			Name
	)
	{
		for (final HTTPAcceptElement<T> Code : getCodesInDescendingOrder()) {
			
		//	If we reach items with qualities of zero (or less, though that should never happen), we've run out of 
		//	acceptable options, so we're done.
			
			if (Code.getQualityInt() <= 0)
				return false;
			
			if (Code.getIdentifier().matches(Name))
				return true;
		}
		
		return false;
	}

	protected boolean explicitlyIncludes
	(
		final String			Name
	)
	{
		for (final HTTPAcceptElement<T> Code : getCodesInDescendingOrder()) {
			
			if (Code.getIdentifier().toString().equals(Name))
				return true;
		}
		
		return false;
	}
	
	HTTPAcceptElement<T>[] getDescendingElementList() {
		return getCodesInDescendingOrder();
	}
	
	T[] getDescendingIdentifierList() {
		final HTTPAcceptElement<T>[]	Codes      = getCodesInDescendingOrder();
		final int						CodesCount = Codes.length;
		final T[]						SortedList = Factory.createIdentifierArray(CodesCount);
		
		for (int Index = 0; Index < CodesCount; Index++)
			SortedList[Index] = Codes[Index].getIdentifier();
		
		return SortedList;
	}
	
	@Override public Iterator<HTTPAcceptElement<T>> iterator() {
		return new ArrayIterator<HTTPAcceptElement<T>>(getCodesInDescendingOrder());
	}

	@Override public String toString() {
		final StringBuilder		Buff  = new StringBuilder(64);
		int						Index = 0;

		for (final HTTPAcceptElement<T> Code : CodesInOriginalOrder) {

			if (Index++ > 0)
				Buff.append(", ");

			Buff.append(Code);
		}

		return Buff.toString();
	}
}
