/*	QWICAP (Quick Web Interface for Conventional Applications) by Chris W. Johnson
	Copyright (C) 2007-2010 The University of Texas at Austin ("U.T. Austin"). All rights reserved.
	
	U.T. Austin's copyright notice in this evolving version reflects our original authorship. It does not assert 
	copyright over changes contributed by others.
	
	This library is free software; you can redistribute it and/or modify it under the terms of the GNU Lesser General 
	Public License as published by the Free Software Foundation; either version 2.1 of the License, or (at your option) 
	any later version.
	
	This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied 
	warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for 
	more details.
	
	You should have received a copy of the GNU Lesser General Public License along with this library; if not, write to 
	the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/


package edu.utexas.its.eis.tools.qwicap.servlet;


import java.util.Locale;


/**	Represents an HTTP language range, which is described in RFC 2616, section 14.4, as follows: "A language-range 
 * 	matches a language-tag if it exactly equals the tag, or if it exactly equals a prefix of the tag such that the 
 * 	first tag character following the prefix is '-'. The special range '*', if present in the Accept-Language field, 
 * 	matches every tag not matched by any other range present in the Accept-Language field."
 * 
 * 	@see <a href='http://www.w3.org/Protocols/rfc2616/rfc2616-sec14.html#sec14.4'>RFC 2616, Section 14.4</a>
 * 	@author Chris W. Johnson
 */
final class HTTPLanguageRange extends HTTPLanguageTag implements HTTPAcceptIdentifier<HTTPLanguageRange> {
	
	final boolean				IsWildcard;
	final HTTPLanguageRange		WildcardDefaultRange;
	
	
	HTTPLanguageRange
	(
		final Locale			Loc
	)
	{
		super(Loc);
		
		IsWildcard           = false;
		WildcardDefaultRange = null;
	}

	HTTPLanguageRange
	(
		final String			Str
	)
	{
		super(Str);
		
		IsWildcard           = "*".equals(Str);
		WildcardDefaultRange = IsWildcard ? new HTTPLanguageRange(Locale.getDefault()) : null;
	}
	
	HTTPLanguageRange
	(
		final String[]			ComponentStrs,
		final int				Offset,
		final int				Count
	)
	{
		super(ComponentStrs, Offset, Count);
		
		IsWildcard           = Count == 1 && "*".equals(ComponentStrs[Offset]);
		WildcardDefaultRange = IsWildcard ? new HTTPLanguageRange(Locale.getDefault()) : null;
	}
	
	@Override public boolean equals
	(
		final Object			Other
	)
	{
		if (IsWildcard)
			return Other instanceof HTTPLanguageTag;	//	We match any language tag.
		
		return super.equals(Other);
	}
	
	@Override public int compareTo
	(
		final HTTPLanguageRange	Other
	)
	{
		return toString().compareTo(Other.toString());
	}
	
	@Override public boolean matches
	(
		final String			LanguageTagStr		//	Consider this to be a language tag.
	)
	{
		return matches(new HTTPLanguageTag(LanguageTagStr));
	}
	
	boolean matches
	(
		final HTTPLanguageTag	LanguageTag
	)
	{
		return getMatchQuality(LanguageTag) != 0;
	}
	
	@Override int getMatchQuality
	(
		final HTTPLanguageTag	LanguageTag
	)
	{
		int						MatchPercentage = super.getMatchQuality(LanguageTag);
		
		if (IsWildcard && MatchPercentage == 0)
			MatchPercentage = 1;	//	Acknowledge a match, but the lowest-quality match possible.
		
		return MatchPercentage;
	}	
	
	HTTPLanguageTag getBestExactMatch
	(
		final HTTPLanguageTag[]		Tags
	)
	{
		if (IsWildcard) {
			
		//	The wildcard can select anything, but, if it is faced with multiple choices, which one should it
		//	choose? I believe that it should choose, if possible, the locale that is most similar to the JVM's
		//	default locale.
			
			return WildcardDefaultRange.getBestExactMatch(Tags);
			
		} else {
			
			for (final HTTPLanguageTag Tag : Tags) {

				if (equals(Tag))
					return Tag;
			}
		}
		
		return null;
	}
	
	HTTPLanguageTag getBestPartialMatch
	(
		final HTTPLanguageTag[]		Tags
	)
	{
		if (IsWildcard) {
			
		//	The wildcard can select anything, but, if it is faced with multiple choices, which one should it
		//	choose? I believe that it should choose, if possible, the locale that is most similar to the JVM's
		//	default locale.
			
			return WildcardDefaultRange.getBestPartialMatch(Tags);
			
		} else {
			
			for (final HTTPLanguageTag Tag : Tags) {

				if (matches(Tag.toString()))
					return Tag;
			}
		}
		
		return null;
	}
}
