/*	QWICAP (Quick Web Interface for Conventional Applications) by Chris W. Johnson
	Copyright (C) 2007-2010 The University of Texas at Austin ("U.T. Austin"). All rights reserved.
	
	U.T. Austin's copyright notice in this evolving version reflects our original authorship. It does not assert 
	copyright over changes contributed by others.
	
	This library is free software; you can redistribute it and/or modify it under the terms of the GNU Lesser General 
	Public License as published by the Free Software Foundation; either version 2.1 of the License, or (at your option) 
	any later version.
	
	This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied 
	warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for 
	more details.
	
	You should have received a copy of the GNU Lesser General Public License along with this library; if not, write to 
	the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/


package edu.utexas.its.eis.tools.qwicap.servlet;


import java.util.regex.Pattern;
import java.util.Locale;


/**	Represents an HTTP language tag. As RFC 2616, Section 3.10, explains: "A language tag identifies a natural language 
 * 	spoken, written, or otherwise conveyed by human beings for communication of information to other human beings."
 * 
 * 	@see <a href='http://www.w3.org/Protocols/rfc2616/rfc2616-sec3.html#sec3.10'>RFC 2616, Section 3.10</a>
 * 	@see <a href='ftp://ftp.rfc-editor.org/in-notes/rfc4646.txt'>RFC 4646</a> which obsoletes RFC 3066, which obsoletes
 * 	     RFC 1766, which is the RFC describing language tags that is referenced by the HTTP specification, RFC 2616.
 * 
 * 	@author Chris W. Johnson
 */
class HTTPLanguageTag {

	private static final Pattern	LanguageSplitter = Pattern.compile("-");

	private final String			HTTPLangStr;
	private HTTPLanguageTag[]		Variants;
	private String[]				Subtags;


	static String normalize
	(
		final String			LangStr
	)
	{
		return LangStr.toLowerCase().replace('_', '-');
	}

	HTTPLanguageTag
	(
		final Locale			Loc
	)
	{
		this(Loc.toString());
	}

	HTTPLanguageTag
	(
		final String			Str
	)
	{
		HTTPLangStr = normalize(Str);
	}

	HTTPLanguageTag
	(
		final String[]			ComponentStrs,
		int						Offset,
		int						Count
	)
	{
		final StringBuilder		Buff = new StringBuilder(3 * Count);

		for (int ComponentNo = 0; --Count >= 0; Offset++, ComponentNo++) {

			if (ComponentNo != 0)
				Buff.append('-');

			Buff.append(ComponentStrs[Offset].toLowerCase());
		}

		HTTPLangStr = Buff.toString();
	}

	@Override public boolean equals
	(
		final Object			Other
	)
	{
		return Other instanceof HTTPLanguageTag && HTTPLangStr.equals(((HTTPLanguageTag) Other).HTTPLangStr);
	}

	@Override public int hashCode() {
		return HTTPLangStr.hashCode();
	}

	@Override public String toString() {
		return HTTPLangStr;
	}
	
	HTTPLanguageTag[] getSubtagCombinations() {

		if (Variants != null)
			return Variants;

		final String[]			Subtags          = getSubtags();
		int						SubtagCount      = Subtags.length;
		final HTTPLanguageTag[]	Combinations     = new HTTPLanguageTag[SubtagCount];
		int						CombinationIndex = 0;

		Combinations[CombinationIndex++] = this;

		while (--SubtagCount > 0)
			Combinations[CombinationIndex++] = new HTTPLanguageRange(Subtags, 0, SubtagCount);

		this.Variants = Combinations;

		return Combinations;
	}
	
	String[] getSubtags() {
		
		if (Subtags == null)
			Subtags = LanguageSplitter.split(HTTPLangStr);
		
		return Subtags;
	}
	
	String getSubtag
	(
		final int				SubtagIndex
	)
	{
		final String[]			Subtags = getSubtags();
		
		if (SubtagIndex < Subtags.length)
			return Subtags[SubtagIndex];
		
		return null;
	}
	
	int getMatchQuality
	(
		final HTTPLanguageTag	LanguageTag
	)
	{
	//	From RFC 2616, Section 14.4, <http://www.w3.org/Protocols/rfc2616/rfc2616-sec14.html#sec14.4>:
	//
	//	"A language-range matches a language-tag if it exactly equals the tag, or if it exactly equals a prefix of the 
	//	tag such that the first tag character following the prefix is '-'."
		
		final String[]			TagSubtags      = LanguageTag.getSubtags();
		final String[]			RangeSubtags    = getSubtags();
		int						MatchingSubtags = 0;
		
		if (RangeSubtags.length <= TagSubtags.length) {		
			
			for (int Index = 0, Length = RangeSubtags.length; Index < Length; Index++) {
				
				if (RangeSubtags[Index].equals(TagSubtags[Index]) == false)
					break;
				
				MatchingSubtags++;
			}
		}
		
		return (MatchingSubtags * 100) / TagSubtags.length;	//	Return a percentage value.
	}
}
