/*	QWICAP (Quick Web Interface for Conventional Applications) by Chris W. Johnson
	Copyright (C) 2003-2007 The University of Texas at Austin ("U.T. Austin"). All rights reserved.
	
	U.T. Austin's copyright notice in this evolving version reflects our original authorship. It does not assert 
	copyright over changes contributed by others.
	
	This library is free software; you can redistribute it and/or modify it under the terms of the GNU Lesser General 
	Public License as published by the Free Software Foundation; either version 2.1 of the License, or (at your option) 
	any later version.
	
	This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied 
	warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for 
	more details.
	
	You should have received a copy of the GNU Lesser General Public License along with this library; if not, write to 
	the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/


package edu.utexas.its.eis.tools.qwicap.servlet;


import javax.servlet.http.HttpServletResponse;
import javax.servlet.http.HttpServletRequest;
import java.util.logging.Level;
import java.util.logging.Logger;


/**	Represents the context unique to each hit.
 *
 *	@author Chris W. Johnson
 */
final class HitContext {

	private static final Logger				Log = Logger.getLogger(HitContext.class.getName());
	
	private final QwicapServlet				Srv;
	private final HttpServletRequest		Req;
	private final HttpServletResponse		Resp;
	private final Auth2SchemeReply			A2SReply;
	private final FormDataSet				DataSet;

	private boolean							Processing = true;


	HitContext
	(
		final QwicapServlet			Srv,
		final HttpServletRequest	Req,
		final FormDataSet			FDS,
		final HttpServletResponse	Resp,
		final Auth2SchemeReply		A2SReply
	)
	{
		if (Srv == null)
			throw new NullPointerException("The QwicapServlet reference cannot be null.");
		if (Req == null)
			throw new NullPointerException("The HttpServletRequest reference cannot be null.");
		if (Resp == null)
			throw new NullPointerException("The HttpServletResponse reference cannot be null.");

		this.Srv      = Srv;
		this.Req      = Req;
		this.Resp     = Resp;
		this.A2SReply = A2SReply;
		this.DataSet  = FDS;
	//	this.DataSet  = ReqBytes == null ? new FormDataSet(Req) : new FormDataSet(Req.getMethod(), Req.getContextPath(), Req.getServletPath(), ReqBytes);
	}

	void waitForProcessingToComplete() throws InterruptedException {
		
		Log.log(Level.FINEST, "HIT: waitForProcessingToComplete entered.");
		
		try {
			
			synchronized (this) {
	
				Log.log(Level.FINEST, "HIT: Servlet waiting for Qwicap thread to complete processing.");
	
				while (Processing)
					wait();
	
				Log.log(Level.FINEST, "HIT: Servlet notified by Qwicap thread that processing is complete.");
			}
		
		} finally {
			
			Log.log(Level.FINEST, "HIT: waitForProcessingToComplete exited.");
		}
	}

	void processingComplete() {
		
		Log.log(Level.FINEST, "HIT: processingComplete entered.");
		
		try {
			
		//	Notify our SessionContext object that it no longer needs to wait (in its "go" method) for the processing of 
		//	this hit to complete.
	
			synchronized (this) {
				
				Log.log(Level.FINEST, "HIT: Notifying servlet that processing is complete.");
	
				Processing = false;
				notify();
	
				Log.log(Level.FINEST, "HIT: Servlet notified that processing is complete.");
				
			//	Note that the thread that invoked waitForProcessingToComplete can't resume executing until this 
			//	"synchronized" block has exited.
			}
		
		} finally {
			
			Log.log(Level.FINEST, "HIT: processingComplete exited.");
		}
	}

	QwicapServlet getServlet() {
		return Srv;
	}

	HttpServletRequest getRequest() {
		return Req;
	}

	HttpServletResponse getResponse() {
		return Resp;
	}

	Auth2SchemeReply getAuth2SchemeReply() {
		return A2SReply;
	}

	FormDataSet getFormDataSet() {
		return DataSet;
	}
}
