/*	QWICAP (Quick Web Interface for Conventional Applications) by Chris W. Johnson
	Copyright (C) 2008-2010 The University of Texas at Austin ("U.T. Austin"). All rights reserved.
	
	U.T. Austin's copyright notice in this evolving version reflects our original authorship. It does not assert 
	copyright over changes contributed by others.
	
	This library is free software; you can redistribute it and/or modify it under the terms of the GNU Lesser General 
	Public License as published by the Free Software Foundation; either version 2.1 of the License, or (at your option) 
	any later version.
	
	This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied 
	warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for 
	more details.
	
	You should have received a copy of the GNU Lesser General Public License along with this library; if not, write to 
	the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/


package edu.utexas.its.eis.tools.qwicap.servlet;


import edu.utexas.its.eis.tools.qwicap.template.xml.immutable.XMLString;
import edu.utexas.its.eis.tools.qwicap.template.xml.TagException;
import edu.utexas.its.eis.tools.qwicap.template.css.Results;
import edu.utexas.its.eis.tools.qwicap.util.SmartStringConverter;

import java.util.logging.Logger;
import java.util.logging.Level;


/**	Encapsulates all of the information about an arbitrary problem that should be reported to the user. Additional 
 * 	debugging information may be included in a comment block within the page, if--and only if--the web application has 
 * 	Qwicap operating in its low-security mode (<code>ReportExceptionToUser</code> is <code>false</code>). In any case, 
 * 	the error message and optional debugging information are supplied to the server's log.
 * 
 * 	@author Chris W. Johnson
 */
class IncidentProblem extends Incident {

	private static final Logger 	Log = Logger.getLogger(IncidentProblem.class.getName());
	private static final XMLString	kCommentMarkup;
	
	static {
		try {
			kCommentMarkup = new XMLString("<!-- -->");
		} catch (TagException e) {
			throw new RuntimeException(e);	
		}
	}
	
	private final Object			HumanInterfaceMessage;
	private final Object			DebugInformation;
	
	
	IncidentProblem
	(
		final i18nServices		i18n,
		final Object			HumanInterfaceMessage,
		final Object			DebugInformation,
		final boolean			IncludeDebugInformation
	)
	{
		this.HumanInterfaceMessage = HumanInterfaceMessage;
		this.DebugInformation      = IncludeDebugInformation ? DebugInformation : null;
		
	//	It may be excessive to log the types of problems that will typically be reported through this class, but, by
	//	setting the log level to FINE, we can keep these log messages out of the way in most configurations, while 
	//	preserving the ability to have them logged if the server administrator wants to know about them.
		
		if (Log.isLoggable(Level.FINE)) {
			final SmartStringConverter	StrCon       = new SmartStringConverter(i18n);
			final String				HIMessageStr = StrCon.toString(HumanInterfaceMessage);
			
			if (DebugInformation != null) {
				
				try {
					final String				DebugInfoStr = StrCon.toString(DebugInformation);
					final String				LineBreakStr;
					
				//	If DebugInfoStr contains multiple lines of text, it will probably be more readable in the logs, if 
				//	it is logged beginning on a new line. So, here we determine whether or not DebugInfoStr is a multi-
				//	line string, and supply either an actual line break string to the logging message, or an empty 
				//	string. Supplying the empty string causes the debug info to be written on the same line as the log
				//	message.
					
					if (DebugInfoStr.contains("\n") || DebugInfoStr.contains("\r"))
						LineBreakStr = "\n";
					else
						LineBreakStr = " ";
					
					Log.log(Level.FINE, "The problem \"{0}\" will be reported to the user. Debug data follows:{1}{2}", new Object[] { HIMessageStr, LineBreakStr, DebugInfoStr } );
					
				} catch (Exception e) {
					Log.log(Level.WARNING, "The problem \"" + HIMessageStr + "\" will be reported to the user. Unfortunately, the attempt to convert the supplied debugging information into a String failed, due to the following exception.", e);
				}
				
			} else 
				Log.log(Level.FINE, "The problem \"{0}\" will be reported to the user.", HIMessageStr);
		}
	}
	
	@Override Object getHumanInterfaceMessage() {
		return HumanInterfaceMessage;
	}

	@Override Object getDebugInformation() {
		return DebugInformation;
	}
	
	@Override void addReportToPage
	(
		final Results			ErrorReportDiv
	)
	throws TagException
	{
			
	//	Unlike the handling of exception reports, we always place the human-readable error message from IncidentProblem 
	//	classes into the web page.
		
		ErrorReportDiv.setContent(getHumanInterfaceMessage());
		
	//	Note that the constructor has already made the determination whether or not Qwicap is operating
	//	in its default high-security mode, or its optional, low-security "debug" mode. In the former case, 
	//	getDebugInformation will return null. In the latter case, it will return whatever debugging information 
	//	object was provided by the web application.
		
		if (getDebugInformation() != null) {
			
		//	Including the debugging information in the page (as an HTML comment) is very convenient during development 
		//	and debugging, but contravenes good security practices by providing attackers with information about the 
		//	internal state of the web application.
			
			ErrorReportDiv.addToEndOfContent(kCommentMarkup.getMutable().getComments().setContent(getDebugInformation()));
		}
	}
}
