/*	QWICAP (Quick Web Interface for Conventional Applications) by Chris W. Johnson
	Copyright (C) 2008-2010 The University of Texas at Austin ("U.T. Austin"). All rights reserved.
	
	U.T. Austin's copyright notice in this evolving version reflects our original authorship. It does not assert 
	copyright over changes contributed by others.
	
	This library is free software; you can redistribute it and/or modify it under the terms of the GNU Lesser General 
	Public License as published by the Free Software Foundation; either version 2.1 of the License, or (at your option) 
	any later version.
	
	This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied 
	warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for 
	more details.
	
	You should have received a copy of the GNU Lesser General Public License along with this library; if not, write to 
	the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/


package edu.utexas.its.eis.tools.qwicap.servlet;


import edu.utexas.ots.chrisj.util.Hexadecimal;
import edu.utexas.its.eis.tools.qwicap.template.xml.mutable.MutableMarkup;
import edu.utexas.its.eis.tools.qwicap.template.xml.TagException;
import edu.utexas.its.eis.tools.qwicap.template.css.Results;

import java.util.logging.Logger;
import java.util.logging.Level;
import java.util.Random;
import java.util.Date;
import java.text.SimpleDateFormat;


/**	Encapsulates all of the information about an exception that needs to be reported. If the web application has Qwicap
 * 	operating in its high-security mode, only an opaque incident code is placed in the web page that reports this 
 * 	exception. In low-security (debug) mode, the exception's message is displayed to the user, and its stack trace is 
 * 	embedded in the same web page as an HTML comment block. In either case, the exception (including stack trace) and 
 * 	the incident code are supplied to the server's log.
 * 
 * 	@author Chris W. Johnson
 */
final class IncidentThrowable extends Incident {

	private static final Logger Log = Logger.getLogger(IncidentThrowable.class.getName());
	
	private final i18nServices	i18n;
	private final String		IncidentCode;
	private final Object		HumanInterfaceMessage;	//	Could be a String (incident code), or a Throwable.
	private final Throwable		DebugInformation;
	
	
	IncidentThrowable
	(
		final i18nServices		i18n,
		final Throwable			ErrorOrException,
		final boolean			ShowExceptionToUser,
		final Random			Rndm
	)
	{
		this.i18n         = i18n;
		this.IncidentCode = new SimpleDateFormat("yyyyMMdd-").format(new Date()) + Hexadecimal.fromInt(Rndm.nextInt(1 << 24)).substring(2);
		
		if (ErrorOrException instanceof QwicapSessionWillNotDieError)
			Log.log(Level.SEVERE, "Incident code \"{0}\" refers to the following exception: {1}", new Object[] { IncidentCode, ErrorOrException.getMessage() }); 
		else
			Log.log(Level.SEVERE, "Incident code \"" + IncidentCode + "\" refers to the following exception:", ErrorOrException);
		
		if (ShowExceptionToUser) {
			
			HumanInterfaceMessage = ErrorOrException;
			DebugInformation      = ErrorOrException;
			
		//	Don't bother to introduce a random delay if we're reporting exceptions to users, because such reporting renders 
		//	a security measure like the random delay pointless. (We've already told an attacker everything they could want 
		//	to know; adding a delay won't confuse them in the least.)
			
			introduceRandomDelay(Rndm);
			
		} else {
			
			HumanInterfaceMessage = IncidentCode;
			DebugInformation      = null;
		}
	}
	
	String getIncidentCode() {
		return IncidentCode;
	}
	
	@Override Object getHumanInterfaceMessage() {
		return HumanInterfaceMessage;
	}
	
	@Override Object getDebugInformation() {
		return DebugInformation;	
	}
	
	@Override void addReportToPage
	(
		final Results			ErrorReportDiv
	)
	throws TagException 
	{
		final MutableMarkup		Message;
		
	//	Note that the IncidentThrowable class has already made the determination whether or not Qwicap is operating
	//	in its default high-security mode, or its optional, low-security "debug" mode. In the former case, 
	//	getDebugInformation will return null. In the latter case, it will return the exception that was thrown.
		
		if (getDebugInformation() != null) {
			
		//	Including the stack trace in the page (as an HTML comment) is very convenient during development and 
		//	debugging, but contravenes some security guidelines by giving attackers information about the point in the 
		//	code at which their attempted attack is being caught.
			
			Message = i18n.getXML("PageError-Exception").getMutable();
			Message.get("span." + Qwicap.kQwicapExceptionMessageClass).setContent(getDebugInformation());
			
		} else {
			
		//	Provide only an incident code. This gives attackers no information about what went wrong, but does give
		//	users a specific problem to report - one that can be matched with an entry in the server's logs.
			
			Message = i18n.getXML("PageError-Incident").getMutable();
			Message.get("span." + Qwicap.kQwicapIncidentCodeClass).setContent(getHumanInterfaceMessage());
		}
		
		ErrorReportDiv.setContent(Message);
	}
	
	private static void introduceRandomDelay
	(
		final Random			Rndm
	)
	{
		try {
			Thread.sleep(Rndm.nextInt(2000));
		} catch (InterruptedException e) {
			Log.log(Level.WARNING, "The random delay security measure stopped prematurely.", e);
		} catch (Exception e) {
			Log.log(Level.WARNING, "The random delay security measure failed.", e);
		}
	}
}
