/*	QWICAP (Quick Web Interface for Conventional Applications) by Chris W. Johnson
	Copyright (C) 2003-2008 The University of Texas at Austin ("U.T. Austin"). All rights reserved.
	
	U.T. Austin's copyright notice in this evolving version reflects our original authorship. It does not assert 
	copyright over changes contributed by others.
	
	This library is free software; you can redistribute it and/or modify it under the terms of the GNU Lesser General 
	Public License as published by the Free Software Foundation; either version 2.1 of the License, or (at your option) 
	any later version.
	
	This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied 
	warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for 
	more details.
	
	You should have received a copy of the GNU Lesser General Public License along with this library; if not, write to 
	the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/


package edu.utexas.its.eis.tools.qwicap.servlet;


import edu.utexas.its.eis.tools.qwicap.template.xml.Match;
import edu.utexas.its.eis.tools.qwicap.template.xml.TagException;


/**	Encapsulates the logic that allows Qwicap to setup "input" tags of type "submit" with a parameter value other than
 *	the human-interface name of the button, which the HTML specifications otherwise make impossible. Background: An 
 *	input element of type "submit" can't supply a normal name/value pair as a parameter, because the HTML specification 
 *	conflated the human-interface name (placed in the "value" attribute of the element), with the 
 *	client/server-interface name. Thus, if a "name" attribute is present, the submit button sends its human interface 
 *	name as the value of the relevant parameter. So, if you want to maintain the name of the button independently of 
 *	the code that processes its input, you can't. Nor can you have the button supply a value that you wouldn't want the 
 *	user to see. In short, that part of the HTML spec. is incredibly badly designed.
 *	
 *	<p>In the post-HTML4 world, sensible people would avoid the aforementioned problem by using the "button" tag, but 
 *	Microsoft's browsers flagrantly violate the relevant specifications, thereby making the "button" tag unuseable in 
 *	many circumstances. Take, for example, a form with multiple button tags of type "submit": Clicking on any of those 
 *	buttons in the Windows version of Internet Explorer will cause all of them to transmit their names and 
 *	human-interface names to the server (the intended value supplied in their "value" attributes is ignored); in the 
 *	Macintosh version of IE, clicking on any of those buttons will submit everything from the relevant form except the 
 *	name/value pair of the submit button. The behavior clearly set-out in the HTML4 specification (and thus inherited 
 *	by XHTML) for "button" tags of type "submit" is to transmit a parameter whose name is taken from the "name" 
 *	attribute of the "button" tag that the user clicked-on, and whose value is taken from the "value" attribute of the 
 *	tag. Any other submit buttons in the same form are ignored. See section 17.13.2, "Successful controls", of the HTML 
 *	4.01 specification at <a href='http://www.w3.org/TR/html4/interact/forms.html#submit-format'>
 *	http://www.w3.org/TR/html4/interact/forms.html#submit-format</a> where it states: "If a form contains more than one 
 *	submit button, only the activated submit button is successful."
 *	</p>
 *
 *	@author Chris W. Johnson
 */
final class InputSubmitButton {
	
	private String				HIName;	//	Human-interface name.
	private String				CName;	//	Name of parameter submitted by button.
	private String				CValue;	//	Value of parameter submitted by button. 
	
	private Match				Tag;
	
	
/**	Returns <code>true</code> if the supplied control name is of the type produced by this class when it is 
 * 	embedding a control name and value into an input/submit button's name, and <code>false</false> 
 *	otherwise.	
 * 
 * @param ControlName	The name of a form control.
 * @return				<code>true</code> if this class is required to decode the supplied control name, 
 *						<code>false</code> otherwise.
 */	
	static boolean producedThisControl
	(
		final String			ControlName
	)
	{
		return ControlName.startsWith(Qwicap.kQwicapInputSubmitControlName);
	}
	
	InputSubmitButton
	(
		final String			ControlName,
		final Object			ControlValue
	)
	{
		initNameAndValue(ControlName, ControlValue);
	}
	
	InputSubmitButton
	(
		final Match				Elem
	)
	{
		final String			ElemType = Elem.getType();
		final String			TypeAttr = Elem.getAttribute("type");
		
		if ("input".equals(ElemType) == false || "submit".equals(TypeAttr) == false)
			throw new IllegalArgumentException("The supplied markup element is not an 'input' tag of type 'submit', it is \"" + Elem + "\".");	
		
		Tag    = Elem;
		HIName = Elem.getAttribute("value");
		
		initNameAndValue(Elem.getAttribute("name"), Elem.getAttribute("value"));
	}
	
	private void initNameAndValue
	(
		final String			ControlName,
		final Object			ControlValueObj
	)
	{
		final String			ControlValue = ControlValueObj != null ? ControlValueObj.toString() : "";
		
		if (ControlName == null) {
			
			CName  = null;
			CValue = null;
			HIName = "Submit";
					
		} else if (ControlName.startsWith(Qwicap.kQwicapInputSubmitControlName)) {
			final int				PrefixLen = Qwicap.kQwicapInputSubmitControlName.length();
			final int				EqIndex   = ControlName.indexOf('=', PrefixLen);
			
			if (EqIndex < 0)
				throw new IllegalArgumentException("The input button's name attribute is improperly formatted; the equal sign (=) is missing.");
			
			CName  = ControlName.substring(PrefixLen, EqIndex);
			CValue = ControlName.substring(EqIndex + 1);
			HIName = ControlValue;
			
		} else {
			
			CName  = ControlName;
			CValue = ControlValue;
			HIName = ControlValue;
		}
	}
	
	String getName() {
		return CName;
	}
	
	String getValue() {
		return CValue;
	}
	
	String getHIName() {
		return HIName;
	}
	
	InputSubmitButton setName
	(
		final String			ParamName
	)
	throws TagException
	{
		CName = ParamName;
		updateTag();
		
		return this;
	}
	
	InputSubmitButton setValue
	(
		final Object			ParamValue
	)
	throws TagException
	{
		CValue = ParamValue != null ? ParamValue.toString() : "";
		updateTag();
		
		return this;
	}
	
	InputSubmitButton setHIName
	(
		final String			HIName
	)
	throws TagException
	{
		this.HIName = HIName;
		updateTag();
		
		return this;
	}
	
	private void updateTag() throws TagException {
		
		if (Tag != null) {
			
			if (CValue.equals(HIName)) {
				
				Tag.setAttribute("value", CValue);
				Tag.setAttribute("name",  CName);
				
			} else {
				
				Tag.setAttribute("value", HIName);
				Tag.setAttribute("name",  toControlNameString());
			}
		}
	}
	
/**	Because input/submit elements use their 'value' attribute for the button's human-interface name, there's
 *	no way to supply an actual value for such elements ... at least not without implementing our own protocol
 *	for representing such values. It is that representation that is constructed here. Specifically, we encode 
 *	the value into the control name, by prefixing the name with a constant, reserved string ("qwicap-input-submit"), 
 *	adding a question mark (?) character, and appending the desired value to that. We then have the 
 *	<code>FormDataSet</code> class special-case any parameters whose names begin with "qwicap-input-submit" and we're 
 *	in business, albeit an ugly business.	
 * 
 * @return		A name for the this input/submit button that encodes a separate name and value pair which will be
 *				decoded automatically upon receipt by Qwicap's data set handling code.
 */	
	String toControlNameString() {
		return Qwicap.kQwicapInputSubmitControlName + CName + '=' + CValue;
	}
}


/*	Operation of 'button' tags of type 'submit' versus 'input' tags of type 'submit':
	
	<button type='submit' name='param-name' value='param-value'>HI Name</button>	OUTPUT: 'param-name'='param-value'
	<input  type='submit' value='HI Name' name='param-name'/>                       OUTPUT: 'param-name'='HI Name'
*/


/*	From the HTML 3.2 specification at <http://www.w3.org/TR/REC-html32.html#fields>:
	
	type=submit
		This defines a button that users can click to submit the form's contents to the server. The button's label is 
		set from the value attribute. If the name attribute is given then the submit button's name/value pair will be 
		included in the submitted data. You can include several submit buttons in the form. See type=image for 
		graphical submit buttons.
		
			<input type=submit value="Party on ...">
*/