/*	QWICAP (Quick Web Interface for Conventional Applications) by Chris W. Johnson
	Copyright (C) 2005-2011 Chris W. Johnson. All rights reserved.
	
	Chris W. Johnson's copyright notice in this evolving version reflects his original authorship. It does not assert 
	copyright over changes contributed by others.
	
	This library is free software; you can redistribute it and/or modify it under the terms of the GNU Lesser General 
	Public License as published by the Free Software Foundation; either version 2.1 of the License, or (at your option) 
	any later version.
	
	This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied 
	warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for 
	more details.
	
	You should have received a copy of the GNU Lesser General Public License along with this library; if not, write to 
	the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/


package edu.utexas.its.eis.tools.qwicap.servlet;


import javax.servlet.http.*;
import java.util.TimerTask;
import java.util.Map;


/**
 * Created by IntelliJ IDEA.
 * User: chrisj
 * Date: Dec 30, 2005
 * Time: 2:58:23 AM
 * To change this template use File | Settings | File Templates.
 */
final class LingeringDownloadables implements HttpSessionBindingListener {
	
	private static final long				kLingerMillis             = 30 * 1000L;
	private static final String				kSessionAttributeNameRoot = "qwicap-final-downloadables";
	
	private final QwicapApplicationServices	AppServices;
	private final HttpSession				Sess;
	private final String					SessionAttributeName;
	private final Map<String, Downloadable>	Downloadables;
	private TimerTask						DisposeTask;
	private int								DisposeCount;
	
	
	static boolean install
	(
		final SessionContext	Ctxt
	)
	{
		final Map<String, Downloadable>	Downloadables = Ctxt.getQwicap().getPageStack().getAllDownloadables();
		
		if (!Downloadables.isEmpty()) {
			
			new LingeringDownloadables(Ctxt, Downloadables);
			return true;
		}
		
		return false;
	}
	
	static void remove
	(
		final SessionContext	Ctxt
	)
	{
		Ctxt.removeSessionAttribute(kSessionAttributeNameRoot);
	}
	
	static Downloadable getDownloadable
	(
		final HttpServletRequest	Req,
		final String				RelativeURI		//	Does not include servlet name.
	)
	{
	//	The relevant Qwicap application exited some time prior to this request. If it had any downloadables when it
	//	exited, they may still be available from a Hashtable object stored in a session attribute. (Typically, that
	//	attribute is deleted something like 30 seconds after the application exits, which should be plenty of time
	//	for the client that received the application's final page to request any embedded downloadables, like 
	//	images. It doesn't matter whether or not the downloads have actually completed by the time the session 
	//	attribute is deleted; as long they were begun before then, they'll complete normally, and then the 
	//	downloadables will become available for garbage collection.)
		
		final HttpSession		Sess = Req.getSession(false);	//	Get the session, if any. Do not create a new session.
		
		if (Sess != null) {
			final Object			LD = Sess.getAttribute(SessionContext.createSessionAttributeName(Req.getContextPath(), kSessionAttributeNameRoot));
			
			if (LD != null)
				return ((LingeringDownloadables) LD).getDownloadable(RelativeURI);
		}
		
		return null;
	}
	
	private LingeringDownloadables
	(
		final SessionContext			Ctxt,
		final Map<String, Downloadable>	Downloadables
	)
	{
		this.AppServices          = Ctxt.getApplicationServices();
		this.Sess                 = Ctxt.getSession();
		this.Downloadables        = Downloadables;
		this.SessionAttributeName = Ctxt.setSessionAttribute(kSessionAttributeNameRoot, this);
	}
	
	void dispose() {
		
	//	Perpetual loops of dispose calls shouldn't be possible, but a call-within-a-call is possible, and, well,
	//	better safe than sorry.
		
		if (DisposeCount++ == 0) {
			
			try {
				Sess.removeAttribute(SessionAttributeName);
			} catch (Exception e) {
			//	Note: an IllegalArgumentException is thrown if the session has already been invalidated.
			}
			
			if (DisposeTask != null)
				DisposeTask.cancel();
			
		//	Note that invalidating the session at this point is only valid while Qwicap 1.4 remains limited to one 
		//	application in each WAR. Actually, it isn't even safe then, because the user may have re-entered this
		//	web app before the timer could fire and execute this "dispose" method, in which case the this session is
		//	reused for the new run of the web app. Invalidating it causes the web app that is in-progress to 
		//	mysteriously stop working and start back at its beginning while the user is trying to use the web app.
	
			
		//	Sess.invalidate();
		//	System.out.println("LingeringDownloadables disposed of.");
		}
	}
	
	private Downloadable getDownloadable
	(
		final String			RelativeURI
	)
	{
		return Downloadables.get(RelativeURI);
	}
	
	@Override public void valueBound
	(
		final HttpSessionBindingEvent	Event
	)
	{
	//	We don't need a dispose task until we're bound to the session, so we don't bother to setup such a task until
	//	we arrive here.
		
		DisposeTask = AppServices.scheduleTimerTask(new RemoveLingeringDownloadablesTask(this), kLingerMillis);
	}
	
	@Override public void valueUnbound
	(
		final HttpSessionBindingEvent	Event
	)
	{
	//	Invoking our dispose method will trigger this method, but so will other relevant happenings, like our session
	//	attribute being overwritten with a new value.
		
		dispose();
	}
	
	static class RemoveLingeringDownloadablesTask extends TimerTask {
		
		private final LingeringDownloadables	LD;
		
		RemoveLingeringDownloadablesTask
		(
			final LingeringDownloadables	LD
		)							
		{
			this.LD = LD;
		}
		
		@Override public void run() {
			LD.dispose();
		}
	}
}