/*	QWICAP (Quick Web Interface for Conventional Applications) by Chris W. Johnson
	Copyright (C) 2003-2011 The University of Texas at Austin ("U.T. Austin"). All rights reserved.
	
	U.T. Austin's copyright notice in this evolving version reflects our original authorship. It does not assert 
	copyright over changes contributed by others.
	
	This library is free software; you can redistribute it and/or modify it under the terms of the GNU Lesser General 
	Public License as published by the Free Software Foundation; either version 2.1 of the License, or (at your option) 
	any later version.
	
	This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied 
	warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for 
	more details.
	
	You should have received a copy of the GNU Lesser General Public License along with this library; if not, write to 
	the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/


package edu.utexas.its.eis.tools.qwicap.servlet;


import edu.utexas.its.eis.tools.qwicap.template.css.CSSPatterns;
import edu.utexas.its.eis.tools.qwicap.template.css.Results;
import edu.utexas.its.eis.tools.qwicap.template.xml.Markup;
import edu.utexas.its.eis.tools.qwicap.template.xml.Match;
import edu.utexas.its.eis.tools.qwicap.template.xml.TagException;
import edu.utexas.its.eis.tools.qwicap.template.xml.immutable.ImmutableMarkup;
import edu.utexas.its.eis.tools.qwicap.template.xml.immutable.XMLString;
import edu.utexas.its.eis.tools.qwicap.template.xml.mutable.MutableMarkup;
import edu.utexas.its.eis.tools.qwicap.template.xml.structure.Declaration;
import edu.utexas.its.eis.tools.qwicap.template.xml.structure.TagWithAttributes;
import edu.utexas.its.eis.tools.qwicap.template.xml.structure.XMLDeclaration;
import edu.utexas.its.eis.tools.qwicap.util.Null;
import edu.utexas.its.eis.tools.table.Table;
import edu.utexas.its.eis.tools.table.readonly.ReadOnlyTable;

import javax.servlet.http.HttpServletResponse;
import java.io.IOException;
import java.io.PrintWriter;
import java.net.URL;
import java.util.*;
import java.util.logging.Level;
import java.util.logging.Logger;
import java.util.regex.Pattern;


/**	<code>Page</code> stores all of Qwicap's per-page context, including: the markup which defines the page, the most 
 *	recent form data set submitted from the page, and a reference to the form within the page from which the form data 
 *	set originated. This class is not part of Qwicap's public API.
 *	
 *	@author Chris W. Johnson
 */
final class Page {

	private static final Logger				Log = Logger.getLogger(Page.class.getName());

	private static final XMLString			kNavigationErrorDiv;
	static final XMLString					kNavigationErrorDialog;
	private static final XMLString			kExceptionReportDiv;
	
	static {

		try {

			kNavigationErrorDiv    = new XMLString("<div class='" + Qwicap.kQwicapNavigationErrorClass + ' ' + Qwicap.kQwicapAutoDeleteClass + "'></div>");
			kNavigationErrorDialog = new XMLString("<div class='qwicap-dialog-overlay " + Qwicap.kQwicapNavigationErrorClass + ' ' + Qwicap.kQwicapAutoDeleteClass + "' id='qwicap-dialog-overlay'>\n" +
			                                       "<div class='qwicap-barrier'></div>\n" +
			                                       "<div class='qwicap-dialog'>\n" +
			                                       "<p class='qwicap-page-navigation-error-message-text'>Error message goes here.</p>\n" +
				                                   "<p class='qwicap-dialog-buttons'><input type='button' value='OK' onclick=\"window['edu.utexas.its.eis.tools.qwicap.deleteByID']('qwicap-dialog-overlay');\"/></p>\n" +
				                                   "</div>\n" +
			                                       "</div>\n");
			
			kExceptionReportDiv    = new XMLString("<div class='" + Qwicap.kQwicapExceptionReportClass + ' ' + Qwicap.kQwicapAutoDeleteClass + "'></div>");
			
		} catch (TagException e) {
			throw new RuntimeException(e);
		}
	}

	static final boolean					kAutoPopulateMostRecentFormDefault = true;
	
	private static final CSSPatterns		kCSSQwicapTitleClass             = new CSSPatterns("*.qwicap-title");
	private static final CSSPatterns		kCSSTitleTag                     = new CSSPatterns("head title");
	
	private static final CSSPatterns		kCSSQwicapBadInputFieldClass     = new CSSPatterns("[class~='" + Qwicap.kQwicapBadInputFieldClass  + "']");
	private static final CSSPatterns		kCSSQwicapAutoDeleteClass        = new CSSPatterns("[class~='" + Qwicap.kQwicapAutoDeleteClass     + "']");
	
	private static final CSSPatterns		kCSSQwicapDialogOverlayDiv       = new CSSPatterns("div.qwicap-dialog-overlay");
	private static final CSSPatterns		kCSSBodyTag                      = new CSSPatterns("body");
	private static final CSSPatterns		kCSSInputOKButton                = new CSSPatterns("input[type='button'][value='ok']");
	private static final CSSPatterns		kCSSQwicapNavErrorDiv            = new CSSPatterns("div." + Qwicap.kQwicapNavigationErrorClass);
	private static final CSSPatterns		kCSSQwicapNavErrorClass          = new CSSPatterns("*." + Qwicap.kQwicapNavigationErrorClass);
	
	private static final CSSPatterns		kCSSQwicapExceptionReportDiv     = new CSSPatterns("div." + Qwicap.kQwicapExceptionReportClass);
	
	private static final CSSPatterns		kCSSFormAcceptCharset            = new CSSPatterns("body form[accept-charset]");
	
	private static final CSSPatterns		kCSSBodyH1Tag                    = new CSSPatterns("body h1");
	
	private static final CSSPatterns		kCSSInputsPageIDAndFormID        = new CSSPatterns("input[name='" + Qwicap.kQwicapPageIDControlName + "'], input[name='" + Qwicap.kQwicapFormIDControlName + "']");
	private static final CSSPatterns		kCSSInputsPageIDFormIDAndCharSet = new CSSPatterns("input[name='" + Qwicap.kQwicapPageIDControlName + "'], input[name='" + Qwicap.kQwicapFormIDControlName + "'], input[name='" + Qwicap.kQwicapCharSetControlName + "']");
	private static final CSSPatterns		kCSSAHrefAndAreaHref             = new CSSPatterns("a[href], area[href]");
	
	private static final CSSPatterns		kCSSInputTag                     = new CSSPatterns("input");
	private static final CSSPatterns		kCSSSelectTag                    = new CSSPatterns("select");
	private static final CSSPatterns		kCSSTextAreaTag                  = new CSSPatterns("textarea");
	private static final CSSPatterns		kCSSButtonTag                    = new CSSPatterns("button");
	
	private static final CSSPatterns		kCSSFormTag                      = new CSSPatterns("form");
	
	private final i18nServices				i18n;

	private final URL						WebAppURL;
	private final MutableMarkup				ThePage;
	private PageMIMETypeAndCharSet			PageType;
	private PageID							PageID;
	private String							Title;

	private Form							MostRecentForm;
	private FormDataSet						MostRecentDataSet = new FormDataSet();

	private final SimplifiedStackTrace		StackTrace;

	private PromptFlags						Flags;

	private Map<String, Downloadable>		Downloadables;


	Page
	(
		final URL				WebAppURL,
		final Random			Rndm,
		final i18nServices		i18n,
		final MutableMarkup		XHTMLPage
	)
	throws QwicapException
	{
		this.WebAppURL  = WebAppURL;
		this.i18n       = i18n;
		this.ThePage    = XHTMLPage;
		this.PageType   = new PageMIMETypeAndCharSet(XHTMLPage);
		this.StackTrace = new SimplifiedStackTrace();

	//	If there's an element in the document with the class "qwicap-title", set the content of the document's
	//	"title" element to the text of that element.

		final String			PageTitle = XHTMLPage.get(kCSSQwicapTitleClass).getText();

		if (PageTitle.length() > 0)
			XHTMLPage.get(kCSSTitleTag).setContent(PageTitle);

	//	Add/update the unique IDs to every form in this page.

		setPageAndFormIDs(Rndm);
	}

	SimplifiedStackTrace getStackTrace() {
		return StackTrace;
	}

	Page write
	(
		final Qwicap			Q
	)
	{
		final HitContext				HCtxt = Q.getHitContext();
		final HttpServletResponse		Resp  = HCtxt.getResponse();

	/*	{
			final String					OrigPageCharSet = ThePage.getCharacterSet();
			final CharacterSetsAcceptable	AccCharSets = HCtxt.getAcceptableCharacterSets();
			
			if (AccCharSets.explicitlyIncludes(OrigPageCharSet) == false) {
				final String			AcceptableCharSet = AccCharSets.getBestDefault();
				
				if (OrigPageCharSet.equalsIgnoreCase(AcceptableCharSet) == false) {
					
					ThePage.setCharacterSet(AcceptableCharSet);
					PageType = new PageContentTypeAndCharSet(ThePage);
					
					Log.log(Level.FINEST, "The page \"{0}\" specifies that it should be encoded using the \"{1}\" character set, however the client does not explicitly state that it accepts that character set, so the page will be sent encoded with the \"{2}\" character set, instead.", new Object[] { ThePage.getMarkupName(), OrigPageCharSet, AcceptableCharSet });
				}
			}
		}
	*/

	//	The setContentType invocation in QwicapServlet.doQWICAp has already set the content type and default character 
	//	set for this response ("text/html" and "UTF-8"). We only re-specify the content-type if we've found that the 
	//	page markup specifies a different character set for this page.

		if (PageType.IsDefault == false)
			Resp.setContentType(PageType.toString());

	//	Setup other standard headers in the response.

		{
			final HTTPAcceptCharsetList	AcceptList = new HTTPAcceptCharsetList();

			for (final String CharSet : getFormCharacterSets())
				AcceptList.add(CharSet, 1.0);
			
			Resp.addHeader("Accept-Charset", AcceptList.toString());

			Log.log(Level.FINEST, "Added to the response an \"Accept-Charset\" HTTP header with the value: {0}", AcceptList);
		}

		Resp.addDateHeader("Last-Modified", System.currentTimeMillis());
		Resp.addHeader("Cache-Control", "must-revalidate, no-cache");
		
	//	todo Consider what other standard headers we should be supplying here.

		try {
			final PrintWriter			Out = Resp.getWriter();
			
			try {
			
				ThePage.write(Out);	//	The setContentType invocation either in this method, or in QwicapServlet.doQWICAp, determines the character set used by this writer.
				Log.log(Level.FINE, "Transmitted page \"{0}\" to client.", PageID);
		
			} finally {
				Out.flush();	//	Commit the response right now.
				Out.close();
				Resp.flushBuffer();
			}
		
		} catch (IOException e) {
			Log.log(Level.INFO, "Failed to transmit page \"" + PageID + "\" to client.", e);
		}

		return this;
	}

	Page cleanup() throws TagException {

		ThePage.get(kCSSQwicapBadInputFieldClass).deleteClass(Qwicap.kQwicapBadInputFieldClass);
		ThePage.get(kCSSQwicapAutoDeleteClass).delete();

		return this;
	}

	Page insertNavigationError
	(
		final boolean			ScriptsWorking,
		String					Message
	)
	throws TagException
	{
		Message = Message.trim();
		
		Log.log(Level.FINE, "Inserting navigation error: \"{0}\"", Message);
		
		if (ScriptsWorking) {
			Results					NavErrDiv = ThePage.get(kCSSQwicapDialogOverlayDiv);
			
			if (NavErrDiv.isEmpty()) {
				
				NavErrDiv = ThePage.get(kCSSBodyTag).addToEndOfContent(kNavigationErrorDialog);
				NavErrDiv.get(kCSSInputOKButton).setAttribute("value", i18n.getString("NavigationError-OK").trim());
			}
			
			NavErrDiv.get("p.qwicap-page-navigation-error-message-text").setContent(Message);
			
		} else {
			Results					NavErrDiv = ThePage.get(kCSSQwicapNavErrorDiv);
	
			if (NavErrDiv.isEmpty())
				NavErrDiv = ThePage.get(kCSSBodyTag).addToStartOfContent(kNavigationErrorDiv);
	
			NavErrDiv.setContent(Message);
		}
		
		return this;
	}

	Page removeNavigationErrors() throws TagException {

		ThePage.get(kCSSQwicapNavErrorClass).delete();
		return this;
	}

	Page insertIncidentReport
	(
		final Incident			Problem
	)
	{
		try {
			Results					ErrDiv = ThePage.get(kCSSQwicapExceptionReportDiv);
	
			if (ErrDiv.isEmpty())
				ErrDiv = ThePage.get(kCSSBodyTag).addToStartOfContent(kExceptionReportDiv);
			
			Problem.addReportToPage(ErrDiv);
		
		} catch (TagException e) {
			Log.log(Level.SEVERE, "An exception was produced while trying to add an incident report to a page. (\"" + Problem + "\")", e);
		}
		
		return this;
	}

	Page doButtonConversions() throws QwicapException {

		ThePage.convertSubmitButtonsToInputs();
		return this;
	}

	Page undoButtonConversions() throws QwicapException {

		ThePage.convertSubmitInputsToButtons();
		return this;
	}

	void setPageAndFormIDs
	(
		final Random			Rndm
	)
	throws TagException
	{
		final Results			Forms = ThePage.get(kCSSFormTag);

	//	Remove any old page and form IDs from the document. If someone constructs a bogus XHTML document that 
	//	includes nested forms, the following line will run into problems. One answer would be to call
	//	"distinct()" before the call to "delete()", but I hate to burden the implementation with such behaviors
	//	just to make bad XHTML appear to be good for a little while.
		
		Forms.get(kCSSInputsPageIDFormIDAndCharSet).delete();

	//	Install a "qwicap-page-id" and "qwicap-form-id" in all forms. The page ID will be the same for every form,
	//	while the form ID will be different for each form.

		PageID = new PageID(Rndm);

		final ImmutableMarkup	PageIDElem = new XMLString("<input type='hidden' name='" + Qwicap.kQwicapPageIDControlName + "' value='" + PageID + "' />");

		for (final Match CurForm : Forms) {
			
		//	It is possible that the form we're examining refers to a web application other than this one. Here we 
		//	detect that case, and leave such forms unmodified.
			
			if (Form.actionRefersToThisWebApp(CurForm.getAttribute("action"), WebAppURL) == false)
				continue;
			
		//	CurForm.setAttribute("action", WebAppURL);	//	This is a response to a bug in Safari (3.0.4) that allows the GET parameters in the page URL, and the POST parameters from a form in that page, to become intermixed in the resulting data set.
			
			String					CharSet = CurForm.getAttribute("accept-charset");

			if (CharSet == null)
				CharSet = getCharacterSet();

			addPageIDToForm(
				CurForm,
				PageIDElem,
				new XMLString("<input type='hidden' name='" + Qwicap.kQwicapFormIDControlName  + "' value='" + new FormID(Rndm) + "' />"),
				new XMLString("<input type='hidden' name='" + Qwicap.kQwicapCharSetControlName + "' value='" + CharSet          + "' />")
			);
			
		//	todo Safari has some brain damage that makes it desirable (probably) to set the form's "action" attribute to the canonical URL of this web application. However, we may have made that irrelevant with the QwicapServlet code that redirects GET requests to turn them into POST requests.
		}

	//	Install a "qwicap-page-id" in all links that appear to reference this servlet, and are trying to send 
	//	parameters.

		addPageIDToLinks(ThePage.get(kCSSAHrefAndAreaHref), "href", ThePage.getCharacterSet(), PageID);
	}
	
	void removePageAndFormIDs() throws TagException {
		
		ThePage.get(kCSSInputsPageIDAndFormID).delete();
		removePageIDFromLinks(ThePage.get(kCSSAHrefAndAreaHref), "href", ThePage.getCharacterSet());
	}
	
	private static void addPageIDToForm
	(
		final Match				Form,
		final ImmutableMarkup	PageIDElem,
		final ImmutableMarkup	FormIDElem,
		final ImmutableMarkup	CharSetElem
	)
	throws TagException
	{

	//	Note that we can't just add our hidden "input" control as the first element in a "form", because XHTML
	//	doesn't permit that. According to the W3C's HTML validator, XHTML form controls must be within: "p", 
	//	"h1", "h2", "h3", "h4", "h5", "h6", "div", "pre", "address", "fieldset", "ins", "del" or a start-tag.
	//	Presumably, the common denominator is that they're all block-level elements.
	//	
	//	So we have to find one of the existing form controls and insert ours just before (or after) it. That way,
	//	conformance to the spec is entirely the responsibility of the page author.
		
		Results					FirstFormControl;

		FirstFormControl = Form.get(kCSSInputTag).first();
		if (FirstFormControl.isEmpty()) {

			FirstFormControl = Form.get(kCSSSelectTag).first();
			if (FirstFormControl.isEmpty()) {

				FirstFormControl = Form.get(kCSSTextAreaTag).first();
				if (FirstFormControl.isEmpty()) {

					FirstFormControl = Form.get(kCSSButtonTag).first();
				}
			}
		}

		if (FirstFormControl.isEmpty() == false) {

		//	Firefox 1.0 had a bug (and later versions may still) that caused them some problem when a form control
		//	(even a hidden control) was inserted inside a "label" element. So, here we check for the case where the
		//	first form control is located inside a label. If so, we insert our ID controls immediately before the 
		//	label element. If not, we insert our ID controls immediately before the first form control.

			Match					Label = null;

			for (final Match Elem : FirstFormControl.getAncestors()) {
				final String			ElemType = Elem.getType();

				if ("label".equals(ElemType)) {

					Label = Elem;
					break;

				} else if ("form".equals(ElemType))
					break;
			}

			if (Label != null) {

				Label.add(true, CharSetElem);
				Label.add(true, PageIDElem );
				Label.add(true, FormIDElem );

			} else {

				FirstFormControl.addBefore(CharSetElem);
				FirstFormControl.addBefore(PageIDElem );
				FirstFormControl.addBefore(FormIDElem );
			}
		}
	}

	MutableMarkup getMarkup() {
		return ThePage;
	}

	String getCharacterSet() {
		return PageType.Charset;
	}

	private String[] getFormCharacterSets() {
		final Results			Forms     = ThePage.get(kCSSFormAcceptCharset);
		final int				FormCount = Forms.size();
		final HashSet<String>	Charsets  = new HashSet<String>(FormCount + 1);

		Charsets.add(getCharacterSet());

		for (final Match Form : Forms)
			Charsets.add(Form.getAttribute("accept-charset"));

		return Charsets.toArray(new String[Charsets.size()]);
	}

	PageID getPageID() {
		return PageID;
	}
		
	String getTitle() {
		
		if (Title == null) {
			
			Title = ThePage.get(kCSSTitleTag).getText().trim();
			if (Null.orEmpty(Title))
				Title = ThePage.get(kCSSBodyH1Tag).first().getText().trim();
		}
		
		return Title;
	}
	
	PromptFlags getFlags() {
		return Flags;
	}

	void setFlags
	(
		final PromptFlags		Flags
	)
	{
		this.Flags = (PromptFlags) Flags.clone();
	}

/**	Takes the form data set most recently received from this page and embeds its values into the XHTML markup of the 
 *	form controls from which they originated in this page. This happens automatically when <code>setFormDataSet</code>
 *	is invoked, unless <code>setAutoPopulateMostRecentForm(false)</code> has been invoked to disable that behavior. 
 *	If there is no most-recent-form, this method does nothing.
 * 
 * @throws TagException				If there was a problem with manipulation of the page markup.
 * @throws FormNotFoundException	If the <samp>form</samp> element has disappeared from the markup.
 */
	void populateMostRecentForm() throws TagException {
		final Form				CurForm = getMostRecentForm();

		if (CurForm != null)
			CurForm.setDataSet(MostRecentDataSet);
	}

/**	Removes all values from the XHTML markup of the form in this page from which a form data set was most recently 
 *	received. If there is no most-recent-form, this method does nothing.
 * 
 * @throws TagException				If there was a problem with manipulation of the page markup.
 * @throws FormNotFoundException	If the <samp>form</samp> element has disappeared from the markup.
 */
	void clearMostRecentForm() throws TagException {
		final Form				CurForm = getMostRecentForm();

		if (CurForm != null)
			CurForm.clearControlValues(false);
	}

	Form getMostRecentForm() throws FormNotFoundException {

		if (MostRecentForm != null && MostRecentForm.getMatch().exists() == false)
			setMostRecentForm(new Form(getMarkup(), MostRecentDataSet));

		return MostRecentForm;
	}

	private void setMostRecentForm
	(
		final Form				CurForm		//	Can be null.
	)
	{
		MostRecentForm = CurForm;
	}

	FormDataSet getFormDataSet() {
		return MostRecentDataSet;
	}

/**	Tell this page that a new <code>FormDataSet</code> has arrived for it.
 * 
 * @param PgStack						A reference to the <code>PaqeStack</code> instance of which this page is an 
 * 										element.
 * @param DataSet						The <code>FormDataSet</code>.
 * @return								Returns this <code>Page</code> object.
 * @throws QwicapFormAbandonedException	If the data set does not correspond to a <samp>form</samp> element in this page.
 * @throws TagException					If there was a problem with manipulation of the page markup.
 * @throws FormNotFoundException		If the <samp>form</samp> element has disappeared from the markup.
 */
	Page setFormDataSet
	(
		final PageStack			PgStack,
		final FormDataSet		DataSet
	)
	throws QwicapFormAbandonedException, TagException
	{
		if (DataSet.has(Qwicap.kQwicapNoOpControlName))
			return this;	//	NoOps affect the PageStack, but never affect a page (other than causing it to be popped off of the stack).
		
		if (PageID == null) {

			MostRecentDataSet = DataSet;

		} else {
			final String			NewPageIDStr = DataSet.getDatum(Qwicap.kQwicapPageIDControlName);

		//	Don't accept the new parameters unless they were sent to this page...

			if (PageID.equals(NewPageIDStr) == false) {	//	PageID can comapare against other PageID instances, and against Strings, like NewPageIDStr.
				final Page				OtherPage     = PgStack.getByPageID(NewPageIDStr);
				final String			OtherPageName = OtherPage != null ? OtherPage.getMarkup().getMarkupName() : "<name unknown>";

				throw new IllegalArgumentException("Can't give page \"" + getMarkup().getMarkupName() + "\", currently assigned ID \"" + PageID + "\", the form data set that was sent to page \"" + OtherPageName + "\", currently assigned ID \"" + NewPageIDStr + "\".");
			}

			final String			NewFormIDStr = DataSet.getDatum(Qwicap.kQwicapFormIDControlName);

			if (NewFormIDStr != null) {
				final Form				CurForm = getMostRecentForm();

				if (CurForm != null && CurForm.formIDEquals(NewFormIDStr)) {

					MostRecentDataSet = DataSet;

				} else {

					try {
						setMostRecentForm(new Form(getMarkup(), NewFormIDStr));
					} catch (FormNotFoundException e) {

					//	We couldn't locate the form based on its ID. (If the client application has rebuilt the page
					//	markup, the hidden input control that supplies the form ID would be missing.) So, try to locate
					//	the correct form based on the contents of the form data set.

						try {
							setMostRecentForm(new Form(getMarkup(), DataSet));
						} catch (FormNotFoundException e2) {
							throw new QwicapFormAbandonedException(this /*, NewFormIDStr */);
						}
					}

					MostRecentDataSet = DataSet;
				}

			} else {

			//	These parameters were produced by an HTML link (e.g. "a href='.../?name=value'), and therefore do not
			//	correspond to a form in this page (even if the link happened to be within a form, the form had nothing
			//	to do with producing these parameters).

				MostRecentDataSet = DataSet;
				setMostRecentForm(null);
			}
		}

		if (Flags.shouldPromptWithFormData())
			populateMostRecentForm();

		return this;
	}

	FormControlValue getFormControlValue
	(
		final String			Name
	)
	{
		return MostRecentDataSet.get(Name);
	}

	boolean hasNoPageID() {
		return PageID == null;
	}

	boolean hasMatchingPageID
	(
		final String			PageIDStr
	)
	{
		if (this.PageID == null)
			return PageIDStr == null;

		return this.PageID.equals(PageIDStr);	//	PageID can comapare against other PageID instances, and against Strings, like PageIDStr.
	}

	boolean isSameInvocationAs
	(
		final Page				Other
	)
	{
		return StackTrace.equals(Other.StackTrace);
	}

	boolean isSameInvocationAs
	(
		final SimplifiedStackTrace	Other
	)
	{
		return StackTrace.equals(Other);
	}

	boolean isSameMethodAs
	(
		final Page				Other
	)
	{
		return StackTrace.isSameMethodAs(Other.StackTrace);
	}

	boolean isSameMethodAs
	(
		final SimplifiedStackTrace	Other
	)
	{
		return StackTrace.isSameMethodAs(Other);
	}

	boolean isStillInPlay
	(
		final Page				LatestPage
	)
	{
		return StackTrace.isSubsetOf(LatestPage.StackTrace);
	}

	boolean isStillInPlay
	(
		final SimplifiedStackTrace	LatestTrace
	)
	{
		return StackTrace.isSubsetOf(LatestTrace);
	}

	boolean isSamePromptAs
	(
		final Page				Other
	)
	{
	//	Check the two pages to determine if their calls into Qwicap (an invocation of some flavor of the prompt method) 
	//	occurred at exactly the same place in the call-chain. This is a slow check, but reliably correct, unlike the 
	//	markup reference equality check we once used as an "optimization".

		return isSameMethodAs(Other);
//		return isSameInvocationAs(Other);
	}

	@Override public String toString() {
		return toTable().toString();
	}

	Table toTable() {
		
		return new ReadOnlyTable(new Object[][] {
			{ "Page ID",       PageID                      },
			{ "Markup",        ThePage.getMarkupName()     },
			{ "Form Data Set", MostRecentDataSet.toTable() },
			{ "Stack Trace",   StackTrace                  }
		});
	}

	Page assumeDutiesOf
	(
		final Page				Other
	)
	{
		Downloadables = Other.Downloadables;
		return this;
	}
	
	private static void removePageIDFromLinks
	(
		final Results			Elems,
		final String			AttrName,
		final String			CharSetName	//	Ordinarily, this'll be the same as the page's character set, but there's no guarantee.
	)
	throws TagException
	{
		for (final Match Elem : Elems) {
			final String			Link    = Elem.getAttribute(AttrName);
			final int				LinkLen = Link.length();	
			
			if (LinkLen == 0)
				continue;
			
			final char				FirstChar = Link.charAt(0);
		
		//	If we aren't dealing with a relative URL, the link probably isn't one we should playing with.
		//	todo How can we improve on this check? The goal is to establish whether this link will invoke this servlet.
		//	Note: There is very closely related logic in the Form.actionRefersToThisWebApp method.
			
			if (FirstChar != '.' && FirstChar != '/' && FirstChar != '?')
				continue;
			
			final int				QuestionIndex = Link.indexOf('?');
			
			if (QuestionIndex == -1 || QuestionIndex == LinkLen - 1)
				continue;
			
			final String			BeforeQuestion   = Link.substring(0, QuestionIndex);
			final String			AfterQuestion    = Link.substring(QuestionIndex + 1);
			final String			NewAfterQuestion = new FormDataSet(AfterQuestion, CharSetName).deleteData(Qwicap.kQwicapPageIDControlName).toWWWFormURLEncodedString(CharSetName);
		
			Elem.setAttribute(AttrName, BeforeQuestion + '?' + NewAfterQuestion);
		}
	}	

	private static void addPageIDToLinks
	(
		final Results			Elems,
		final String			AttrName,
		final String			CharSetName,
		final PageID			PageID
	)
	throws TagException
	{
		final String			PageIDStr = PageID.toString();

		for (final Match Elem : Elems) {
			final String			Link = Elem.getAttribute(AttrName);

			if (Link.length() == 0) {
				
				Elem.setAttribute(AttrName, '?' + new FormDataSet().addData(Qwicap.kQwicapCharSetControlName, CharSetName).addData(Qwicap.kQwicapPageIDControlName, PageIDStr).toWWWFormURLEncodedString(CharSetName));
				continue;
			}

			final char				FirstChar = Link.charAt(0);

		//	If we aren't dealing with a relative URL, the link probably isn't one we should playing with.
		//	todo How can we improve on this check? The goal is to establish whether this link will invoke this servlet.
		//	Note: There is very closely related logic in the Form.actionRefersToThisWebApp method.
		
			if (FirstChar != '.' && FirstChar != '/' && FirstChar != '?')
				continue;

		//	This link provides no parameters, so it is irrelevant to us.
			
			final int				QuestionIndex = Link.indexOf('?');

			if (Link.indexOf('?') == -1)
				continue;
			
			final String			BeforeQuestion = Link.substring(0, QuestionIndex);
			final String			AfterQuestion  = Link.substring(QuestionIndex + 1);
			final FormDataSet		FDS            = new FormDataSet(AfterQuestion, CharSetName);
			
			FDS.deleteData(Qwicap.kQwicapCharSetControlName, Qwicap.kQwicapPageIDControlName);
			FDS.addData(   Qwicap.kQwicapCharSetControlName, CharSetName);
			FDS.addData(   Qwicap.kQwicapPageIDControlName,  PageIDStr);
			
			Elem.setAttribute(AttrName, BeforeQuestion + '?' + FDS.toWWWFormURLEncodedString(CharSetName));
		}
	}

	boolean hasDownloadables() {
		return Downloadables != null && !Downloadables.isEmpty();
	}

	void addDownloadables
	(
		final ArrayList<Downloadable>	NewDownloads
	)
	{

	//	If we don't already have one, create a hashtable to hold the downloadables for this page.

		if (Downloadables == null)
			Downloadables = Collections.synchronizedMap(new HashMap<String, Downloadable>());

	//	Get the names and URIs of the existing downloadables for this page.

		final int				CurCount = Downloadables.size();
		final String[]			CurNames = new String[CurCount];
		final String[]			CurURIs  = new String[CurCount];

		if (CurCount > 0) {
			int						Index = 0;
			
			for (final Downloadable Download : Downloadables.values()) {

				CurNames[Index] = Download.getName();
				CurURIs[Index]  = Download.getURI();
				Index++;
			}
		}

	//	Add the new downloadables.

		for (int Index = 0, Count = NewDownloads.size(); Index < Count; Index++) {
			final Downloadable		NewDownload = NewDownloads.get(Index);
			final String			NewName     = NewDownload.getName();

		//	Search this page's existing downloadables for a downloadable of the same name as the new item. 
		//	If there is an existing item with the same name, replace it with the new download, which assumes the URI
		//	of the downloadable that it replaces.

			for (int CurIndex = 0; CurIndex < CurCount; CurIndex++) {

				if (CurNames[CurIndex].equals(NewName)) {
					final String			CurURI = CurURIs[Index];

				//	System.out.println("Replacing downloadable \"" + CurURI + "\" with \"" + NewDownload.getURI() + "\"");

					Downloadables.remove(CurURI);
					break;
				}
			}

		//	If there's no downloadable of the same name as the new item, insert the new downloadable with its unique URI.

			Downloadables.put(NewDownload.getURI(), NewDownload);
		}
	}

	Map<String, Downloadable> getDownloadables() {
		return Downloadables;
	}

	Downloadable getDownloadable
	(
		final String			URI
	)
	{
		if (Downloadables == null)
			return null;

		return Downloadables.get(URI);
	}

	static class PageMIMETypeAndCharSet {
		
	//	See <http://www.w3.org/TR/2000/REC-xhtml1-20000126/#docconf> for the list of valid XHTML DOCTYPEs.
		
		private static final Pattern		kDocTypeXHTMLPat            = Pattern.compile("\\s*(html|HTML)\\s+PUBLIC\\s+\"-//W3C//DTD XHTML .*", Pattern.DOTALL);
		private static final CSSPatterns	kCSSXHTMLNamespaceTag       = new CSSPatterns("html[xmlns='http://www.w3.org/1999/xhtml']");
		private static final CSSPatterns	kCSSMetaHttpEquivContentTag = new CSSPatterns("html > head > meta[http-equiv][content]");
		
		private final boolean		IsDefault;
	
		private final String		MIMEType;
		private final boolean		MIMETypeIsDefault;

		private final String		Charset;
		private final boolean		CharsetIsDefault;


		PageMIMETypeAndCharSet
		(
			final Markup			ThePage
		)
		{
			final Object			FirstElement      = ThePage.get(0);
			final boolean			HasXMLDeclaration = FirstElement instanceof XMLDeclaration;
			String					ContentType       = null;
			String					CharSet           = null;
			
			if (HasXMLDeclaration)
				CharSet = ((XMLDeclaration) FirstElement).getEncoding();
			
		//	Note: A page is considered to have the content type "application/xhtml+xml" only if all of the following
		//	      conditions are met: (1) The page has an XML declaration. (2) The page contains an "html" tag that
		//	      specifies the XHTML namespace. (3) There's a DOCTYPE declaration that identifies the page as XHMTL.
		//
		//	      Ideally, since we're always delivering XHMTL pages, we'd always use the "application/xhtml+xml" type
		//	      (see "Sending XHTML as text/html Considered Harmful" at <http://www.hixie.ch/advocacy/xhtml>), but
		//	      various browsers are too dumb to handle that type, so we leave the choice up to the developer by
		//	      looking for *all* of the aformentioned conditions to be met. If they're not, we use the incorrect, but
		//	      universally understood, "text/html" type.
		//
		//	      Be aware that the page could also contain a "<meta http-equiv='Content-Type' .... />" element that 
		//	      disagrees with whatever conclusion we draw from the three items discussed above. In that case, what's 
		//	      the correct answer? Presenting browsers with contradictory information is a bad idea - browsers
		//	      are overwhelmed with bad data, as-is, and Qwicap should always strive to do things exactly right, 
		//	      rather than forcing other things to work around mistakes. (This includes fixing mistakes made
		//	      by developers using Qwicap, which may sound presumptuous, but the alternate presumption is that every
		//	      developer can learn every applicable rule, keep them all in his/her head all the time, and always 
		//	      apply them correctly. I think it's clear that the latter presumption is actually the least tenable of
		//        the two, and therefore correcting mistakes wherever we can do so safely is a good philosophy.) Can we, 
		//	      therefore, identify a safe method for resolving whatever contradictions might exist?
			
			if (HasXMLDeclaration && ThePage.get(kCSSXHTMLNamespaceTag).notEmpty()) {
				final Results			DecElems = ThePage.getDeclarations();
				
				if (DecElems.notEmpty()) {
					final Match				DecElem = DecElems.get(0);
					final Declaration		Dec     = (Declaration) DecElem.get();
					
					if (Dec.getName().equals("DOCTYPE")) {	//	This is correct - "DOCTYPE" should be an argument to equals() in this case. Doing the reverse is traditionally correct, but won't work in this case.
						final String			DecContent = Dec.getContent().toString();
						
						if (kDocTypeXHTMLPat.matcher(DecContent).matches())
							ContentType = "application/xhtml+xml";
					}
				}
			}
			
		//	Allow the value supplied by a "<meta http-equiv='Content-Type' content='...'/>" tag to override the content
		//	type, and to identify the character set, if the character set hasn't been identified already.
			
			for (final Match Tag : ThePage.get(kCSSMetaHttpEquivContentTag)) {
				final TagWithAttributes	HttpEquiv = (TagWithAttributes) Tag.get();
				
				if ("Content-Type".equalsIgnoreCase(HttpEquiv.getAttribute("http-equiv"))) {
					final XMLContentType	CT = new XMLContentType(HttpEquiv.getAttribute("content"));
					
					ContentType = CT.getMIMEType();
					
					if (CharSet == null)
						CharSet = CT.getCharacterSet();

				//	todo In the future, it might be helpful to add detection of disagreements between the charset specified in the XML declaration, and the charset specified by the optional "meta http-equiv='Content-Type'" tag.

					break;
				}
			}
			
			if (CharSet == null)
				CharSet = ThePage.getCharacterSet();

			this.MIMEType          = ContentType != null ? ContentType : "text/html";
			this.MIMETypeIsDefault = QwicapServlet.DefaultContentType.equalsIgnoreCase(this.MIMEType);	//	MIME types are case insensitive. See RFC 2045, section 2.
			
			if (CharSet == null) {

				this.Charset          = QwicapServlet.DefaultCharset;
				this.CharsetIsDefault = true;

			} else {

				this.Charset          = CharSet;
				this.CharsetIsDefault = QwicapServlet.DefaultCharset.equalsIgnoreCase(CharSet);	//	Character set name are case insensitive.
			}

			IsDefault = MIMETypeIsDefault && CharsetIsDefault;
		}
		
		@Override public String toString() {
			return MIMEType + "; charset=" + Charset;
		}
	}
}

