/*	QWICAP (Quick Web Interface for Conventional Applications) by Chris W. Johnson
	Copyright (C) 2003-2011 The University of Texas at Austin ("U.T. Austin"). All rights reserved.
	
	U.T. Austin's copyright notice in this evolving version reflects our original authorship. It does not assert 
	copyright over changes contributed by others.
	
	This library is free software; you can redistribute it and/or modify it under the terms of the GNU Lesser General 
	Public License as published by the Free Software Foundation; either version 2.1 of the License, or (at your option) 
	any later version.
	
	This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied 
	warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for 
	more details.
	
	You should have received a copy of the GNU Lesser General Public License along with this library; if not, write to 
	the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/


package edu.utexas.its.eis.tools.qwicap.servlet;


import edu.utexas.its.eis.tools.qwicap.template.xml.mutable.MutableMarkup;
import edu.utexas.its.eis.tools.qwicap.template.xml.TagException;
import edu.utexas.its.eis.tools.qwicap.template.xml.Markup;
import edu.utexas.its.eis.tools.table.MutableTable;
import edu.utexas.its.eis.tools.table.lltable.MutableLLTable;

import java.util.*;
import java.net.URL;


/**	<code>PageStack</code> stores all of the {@link Page}s currently in-play on a stack, while providing front-ends 
 *	for the <code>Page</code> accessor methods that always apply to the <code>Page</code> on the top of the stack.
 *	
 *	@author Chris W. Johnson
 */
final class PageStack implements Iterable<Page> {
	
	private final URL			WebAppURL;
	private final Random		Rndm;
	private final i18nServices	i18n;
	private final Stack<Page>	PgStack = new Stack<Page>();	//	Contains Page objects. Needs to be this synchronized type (Stack), in order for downloads to be hanlded properly.
	private boolean				Goodbye;


	PageStack
	(
		final URL				WebAppURL,
		final Random			Rndm,
		final i18nManager		i18nMgr
	)
	{
		this.WebAppURL = WebAppURL;
		this.Rndm      = Rndm;
		this.i18n      = i18nMgr.getServicesForClass(Qwicap.class);
	}

	MutableMarkup getMarkup() {
		return isEmpty() ? null : peek().getMarkup();
	}

	Form getMostRecentForm() throws FormNotFoundException {
		return isEmpty() ? null : peek().getMostRecentForm();
	}

	boolean getAutoPopulateMostRecentForm() {
		return getFlags().shouldPromptWithFormData();
	}

	PromptFlags getFlags() {
		return isEmpty() ? new PromptFlags() : peek().getFlags();
	}

	Page updatePage() throws QwicapException {

		if (isEmpty())
			return null;

		final Page				CurPage = peek();

		CurPage.setPageAndFormIDs(Rndm);

		return CurPage;
	}

/**	Takes the form data set most recently received from the current page and embeds its values into the XHTML markup of 
 *	the form controls from which they originated in that page. This happens automatically when <code>setFormDataSet</code>
 *	is invoked, unless <code>setAutoPopulateMostRecentForm(false)</code> has been invoked to disable that behavior. 
 *	If there is no current page, or most-recent-form, this method does nothing.
 * 
 * @throws TagException				If there was a problem with manipulation of the page markup.
 * @throws FormNotFoundException	If the <samp>form</samp> element has disappeared from the markup.
 */
	void populateMostRecentForm() throws TagException {

		if (notEmpty()) {
			final Page				CurPage = peek();

			if (CurPage != null)
				CurPage.populateMostRecentForm();
		}
	}

/**	Removes all values from the XHTML markup of the form in the current page from which parameters were most recently 
 *	received. If there is no current page, or most-recent-form, this method does nothing.
 * 
 * @throws TagException				If there was a problem with manipulation of the page markup.
 * @throws FormNotFoundException	If the <samp>form</samp> element has disappeared from the markup.
 */
	void clearMostRecentForm() throws TagException {

		if (notEmpty()) {
			final Page				CurPage = peek();

			if (CurPage != null)
				CurPage.clearMostRecentForm();
		}
	}

	boolean setFormDataSet
	(
		final Qwicap			Q,		//	todo Change this to be a reference to the InputFilter that should be executed.
		final FormDataSet		DataSet
	)
	throws QwicapException
	{
	//	"NoOp" operations change the state of the page stack (popping to a specified page), but do NOTHING else, 
	//	including updating a page's current data set.
		
		final boolean			NoOp = DataSet.has(Qwicap.kQwicapNoOpControlName);
		
		if (!NoOp && Q != null) {
			
		//	NOTE: The input filter that is executed below may implement its own prompt pattern, or invoke code that does.
		//        This OK, and we need to make sure that it continues to be OK in the future.
			
			Q.invokeInputFilter(DataSet);
		}

	//	Get on with our normal duties.

		if (isEmpty())
			return false;

		boolean					IsGoodDataSet = !NoOp;
		final Page				CurPage       = peek();

		if (CurPage.hasNoPageID()) {
			
			CurPage.setFormDataSet(this, DataSet);

		} else {
			final String			NewPageIDStr = DataSet.getDatum(Qwicap.kQwicapPageIDControlName);

		//	See that the form data set is sent to the correct page, if any.

			if (CurPage.hasMatchingPageID(NewPageIDStr)) {

			//	The form data set originated from the current page (the one on the top of the page stack).
				
				CurPage.setFormDataSet(this, DataSet);

			} else {
				final Page				DestPage = getByPageID(NewPageIDStr);

				if (DestPage != null) {

				//	The page from which the form data set originated is still present on the page stack.

				//	Make sure there are no modal pages between the top of the page stack, and the page that the user
				//	is trying to go back to.
					
					synchronized (PgStack) {
						
						for (int Index = PgStack.size(); --Index >= 0; ) {
							final Page				Pg = PgStack.get(Index);
	
							if (Pg.hasMatchingPageID(NewPageIDStr))
								break;
	
						//	We can't go back beyond the most recent modal page on the stack.
	
							if (Pg.getFlags().shouldPromptModally()) {
	
							//	todo Verify that the page is still in play.
								
								Pg.insertNavigationError(Q.getScriptsWorking(), i18n.getString("NavigationError-CannotLeaveModalPage"));
							//	Pg.insertNavigationError("Sorry, you cannot return to the requested page, because the current page still requires your attention.");
	
							//	Discard the new form data set, and abandon pages until we get back to the most recent modal
							//	page.
	
							//	IsGoodDataSet = false;
								throw new QwicapPageAbandonedException(Pg, false);
							}
						}
					}
					
				//	Go back to the page the user has gone back to.

				//	todo Verify that the page is still in play.
					
					DestPage.setFormDataSet(this, DataSet);
					
					throw new QwicapPageAbandonedException(DestPage, !NoOp);

				} else {

				//	The page from which the form data set originated is NOT present on the page stack, so we discard
				//	the data set, and add an error message to the current page.

					IsGoodDataSet = false;
					
					if (NewPageIDStr != null)
						CurPage.insertNavigationError(Q.getScriptsWorking(), i18n.getString("NavigationError-CannotGoToInvalidPage"));
					//	CurPage.insertNavigationError("Sorry, you cannot return to the requested page, because it is no longer active.");
				}
			}
		}

		if (IsGoodDataSet)
			CurPage.cleanup();

		return IsGoodDataSet;
	}

	FormDataSet getFormDataSet() {
		return isEmpty() ? null : peek().getFormDataSet();
	}

	FormControlValue getFormControlValue
	(
		final String			Name
	)
	{
		return isEmpty() ? null : peek().getFormControlValue(Name);
	}

	boolean notEmpty() {
		return !isEmpty();
	}

	boolean isEmpty() {
		return PgStack.isEmpty();
	}

	int size() {
		return PgStack.size();
	}

	Page get
	(
		final int				Index
	)
	{
		return PgStack.get(Index);
	}

	void popTo
	(
		final Page				TopOfTheStackPage
	)
	{
		synchronized (PgStack) {
						
			if (PgStack.contains(TopOfTheStackPage) == false)
				throw new IllegalArgumentException("The page \"" + TopOfTheStackPage.getMarkup().getMarkupName() + "\", assigned ID \"" + TopOfTheStackPage.getPageID() + "\", is not on the page stack, so we can't pop to it.");
	
			while (peek() != TopOfTheStackPage)	//	Yes, a reference equality test is correct.
				pop();
		}
	}

	private Page push
	(
		final MutableMarkup		XHTMLPage
	)
	throws QwicapException
	{
		return push(new Page(WebAppURL, Rndm, i18n, XHTMLPage));
	}

	private Page push
	(
		final Page				NewPage
	)
	throws QwicapException
	{

	//	When we're pushing the very last page of a session, the following protective check is unnecessary, and can 
	//	cause needless complications for the client application, so we skip it.

		if (!Goodbye) {
			final Page				PrevPage = getByMarkup(NewPage.getMarkup());

			if (PrevPage != null)
				throw new QwicapException("The same instance of markup (\"" + NewPage.getMarkup().getMarkupName() + "\") is " +
					"still in use by an earlier, active (\"in-play\") page whose prompt pattern is implemented at the " +
					"following location:\n\n" +
					PrevPage.getStackTrace() +
					"\nIf that page should NOT be regarded as active, add an invocation of Qwicap.promptComplete to the " +
					"method that implements its \"prompt pattern\". That invocation should be added at a point prior to " +
					"the invocation that ultimately led to this attempt to reuse that markup.\n\n" +
					"If that page should still be active, you can work around this shared-markup problem by cloning the " +
					"markup, and using the clone here."
				);
		}

		PgStack.push(NewPage);
		return NewPage;
	}

	Page peek() {
		return PgStack.peek();
	}

	private Page pop() {
		return PgStack.pop();
	}

/**	If the page on the top of the stack (after just-in-case pruning) corresponds to the application method that invoked 
 *	Qwicap, pop it off the stack. 
 */
	void popPageAssociatedWithInvokingMethod() {
		final SimplifiedStackTrace	LatestTrace = new SimplifiedStackTrace();

		prune(LatestTrace);

		if (notEmpty()) {

			if (peek().isSameMethodAs(LatestTrace))
				pop();
		}
	}

	Page getByClassAndMethod
	(
		final Class<?>			PromptPatClass,
		final String			PromptPatMethodName
	)
	{
		synchronized (PgStack) {
				
			for (int Index = PgStack.size(); --Index >= 0; ) {
				final Page				Pg = PgStack.get(Index);
	
				if (Pg.getStackTrace().endsWithMethod(PromptPatClass, PromptPatMethodName))
					return Pg;
			}
		}
		
		return null;
	}

	Page getByMethod
	(
		final SimplifiedStackTrace	LatestTrace
	)
	{
		synchronized (PgStack) {
				
			for (int Index = PgStack.size(); --Index >= 0; ) {
				final Page				Pg = PgStack.get(Index);
	
				if (Pg.isSameMethodAs(LatestTrace))
					return Pg;
			}
		}
		
		return null;
	}

	Page getByPageID
	(
		final String			PageID
	)
	{
		synchronized (PgStack) {
			
			for (int Index = PgStack.size(); --Index >= 0; ) {
				final Page				Pg = PgStack.get(Index);
	
				if (Pg.hasMatchingPageID(PageID))
					return Pg;
			}
		}
		
		return null;
	}

	Page getByMarkupName
	(
		final String			MarkupName
	)
	{
		synchronized (PgStack) {
			
			for (int Index = PgStack.size(); --Index >= 0; ) {
				final Page				Pg = PgStack.get(Index);
	
				if (Pg.getMarkup().getMarkupName().equals(MarkupName))
					return Pg;
			}
		}
		
		return null;
	}

	private Page getByMarkup
	(
		final Markup			Doc
	)
	{
		for (int Index = PgStack.size(); --Index >= 0; ) {
			final Page				Pg = PgStack.get(Index);

			if (Pg.getMarkup() == Doc)	//	Yes, a reference equality test is correct.
				return Pg;
		}

		return null;
	}

/**	Sets the final page. This works just like {@link #setPage} but disables some error checking that is unnecessary for
 *	the final page.
 * 
 * @param XHTMLPage						A document sent to the client by the {@link Qwicap#goodbye} method.
 * @return								The <code>Page</code> object that represents the supplied XHTML page, the most
 *										recent form data set it has received (if any), and other good things.
 * @throws QwicapException				If there is a problem wiht using the specified page as the "goodbye" page.
 */
	Page setGoodbyePage
	(
		final MutableMarkup		XHTMLPage
	)
	throws QwicapException
	{
		Goodbye = true;
		return setPage(XHTMLPage, new PromptFlags());
	}

/*	Sets the current page, updating the page stack accordingly.
 * 
 * @param XHTMLPage						A document sent to the client by the {@link Qwicap#prompt} method.
 * @param AutoEmbedFormControlValues	<code>true</code> if form control values returned by the user should be 
 *										automatically embedded into the supplied XHTML page, <code>false</code>
 *										otherwise.
 * @param Modeless						<code>true</code> if the page is modeless, <code>false</code> if it is modal.
 * @return								The <code>Page</code> object that represents the supplied XHTML page, the most
 *										recent form data set it has received (if any), and other good things.
 *
	Page setPage
	(
		final MutableMarkup		XHTMLPage,
		final boolean			AutoEmbedFormControlValues,
		final boolean			Modeless
	)
	throws QwicapException
	{
		final int				StackSize = PgStack.size();
		final Page				NewPage;
		
		if (StackSize == 0)
			NewPage = push(XHTMLPage);
		else
			NewPage = autoPushPopReplace(XHTMLPage);
		
		NewPage.setAutoPopulateMostRecentForm(AutoEmbedFormControlValues);
		NewPage.setModeless(Modeless);
		
		return peek();
	}
*/

/**	Sets the current page, updating the page stack accordingly.
 * 
 * @param XHTMLPage					A document sent to the client by the {@link Qwicap#prompt} method.
 * @param Flags						Flags that modify the behavior of this page.
 * @return							The <code>Page</code> object that represents the supplied XHTML page, the most
 *									recent form data set it has received (if any), and other good things.
 * @throws QwicapException			If there is a problem making the specified page into the current page.
 */
	Page setPage
	(
		final MutableMarkup		XHTMLPage,
		final PromptFlags		Flags
	)
	throws QwicapException
	{
		synchronized (PgStack) {
			final int				StackSize = PgStack.size();
			final Page				NewPage;
	
			if (StackSize == 0)
				NewPage = push(XHTMLPage);
			else
				NewPage = autoPushPopReplace(XHTMLPage);
	
			NewPage.setFlags(Flags);
	
			return peek();
		}
	}

	private Page autoPushPopReplace
	(
		final MutableMarkup		XHTMLPage
	)
	throws QwicapException
	{
	/*	If at all possible, we want to determine automatically whether the page just supplied to the Qwicap.prompt
		method is re-prompting from the same place as the previous prompt, or whether this is a new prompt. 
		
		From a simplicity stand-point, we want to avoid creating methods like promptPush and promptReplace, since
		that would require the prompting loop for each page to promptPush on its first iteration, and promptReplace 
		on all subsequent iterations. Further, some sort of "pop" invocation would be necessary to clean-up afterwards.
		
		The hope, therefore is to identify whether the new page should replace the previous page on the stack, or 
		merely be added to the top of the stack, by looking for identical MutableMarkup instances, and, failing that,
		for calls to prompt() from exactly the same place in the call-chain. If either of those conditions are met,
		the new page should replace the previous page on the top of the stack. Otherwise, we can assume we're seeing 
		a new prompt, and we should push the new page onto the top of page stack.
	*/
		final Page				NewPage = new Page(WebAppURL, Rndm, i18n, XHTMLPage);

	//	Strip abandoned pages from the top of the stack. (The application code may have exited from any number of 
	//	page-supporting methods between the previous invocation of a prompt method, and the latest invocation.)

		prune(NewPage);

	//	Work down the stack, examining every page on it, looking for a previous page provided by the same caller. 
	//	If we find such a previous page, we need to pop the page stack back to that item and replace it with the new
	//	page (or pop to the item before that one and do a push).

		for (int Index = PgStack.size(); --Index >= 0; ) {
			final Page				OldPage = PgStack.get(Index);

		//	Was OldPage supplied by exactly the same prompt invocation as NewPage?

			if (NewPage.isSamePromptAs(OldPage)) {

			//	If we're replacing a pre-existing page, move its downloads, etc. to the new page, since they're both 
			//	effectively the same page.

				NewPage.assumeDutiesOf(OldPage);

			//	Truncate the stack to remove OldPage and everything pushed on top of it. 

				PgStack.setSize(Index);
				break;
			}
		}

	//	Push the new page onto the stack.

		push(NewPage);

	//	Done. Return a reference to the new Page object we've created.

		return NewPage;
	}

	private void prune
	(
		final SimplifiedStackTrace	LatestTrace
	)
	{
	//	Work down the stack, examining every page on it, checking for pages that are no longer in-play (their 
	//	implementing method is no longer in the call chain).

		while (notEmpty()) {

			if (peek().isStillInPlay(LatestTrace))
				break;

			pop();
		}
	}

	private void prune
	(
		final Page				LatestPage
	)
	{
	//	Work down the stack, examining every page on it, checking for pages that are no longer in-play (their 
	//	implementing method is no longer in the call chain).

		while (notEmpty()) {

			if (peek().isStillInPlay(LatestPage))
				break;

			pop();
		}
	}
	
	@Override public String toString() {
			
		if (isEmpty())
			return "Page stack is empty.";
		
		final MutableTable		Tbl   = new MutableLLTable();
		int						Index = 0;
		
		for (final Page P : this)
			Tbl.addRow().add("Page " + Index++).add(P.toTable());
		
		return Tbl.toString();
	}
	
	@Override public Iterator<Page> iterator() {
		return PgStack.iterator();
	}
	
	Map<String, Downloadable> getAllDownloadables() {
		final Map<String, Downloadable>	All = Collections.synchronizedMap(new HashMap<String, Downloadable>());
		
		synchronized (PgStack) {
			
			for (int Index = 0, Count = PgStack.size(); Index < Count; Index++) {
				final Page						Pg                    = PgStack.get(Index);
				final Map<String, Downloadable>	DownloadablesFromPage = Pg.getDownloadables();
	
				if (DownloadablesFromPage != null)
					All.putAll(DownloadablesFromPage);
			}
		}
		
		return All;
	}

	Downloadable getDownloadable
	(
		final String			RelativeURI	//	The servlet name, and the slash character that immediately follows it, have been removed.	
	)
	{
//		final int				SimpleStartIndex = URI.indexOf(Downloadable.kLocalURIPrefix);
//		
//		if (SimpleStartIndex < 0)
//			return null;
//		
//		final String			SimplifiedURI = URI.substring(SimpleStartIndex);

	//	System.out.println("Searching for download: " + SimplifiedURI);

	//	Work from the top of the stack, down to the bottom, asking each page if it has a Downloadable corresponding to
	//	the specified URI. Thus, we check with the current page first, and work our way down to the oldest page on the
	//	stack.
		
		synchronized (PgStack) {
			
			for (int Index = PgStack.size(); --Index >= 0; ) {
				final Downloadable		Download = PgStack.get(Index).getDownloadable(RelativeURI);
	
				if (Download != null)
					return Download;
			}
		}
		
		return null;
	}

	void addDownloadables
	(
		final ArrayList<Downloadable>	NewDownloads
	)
	{
		if (notEmpty())
			peek().addDownloadables(NewDownloads);
	}
	
	PageInfo[] getInfo() {
		
		synchronized (PgStack) {
			final int				Count  = PgStack.size();
			final PageInfo[]		PgInfo = new PageInfo[Count];
			
			for (int Index = 0; Index < Count; Index++) {
				final Page				Pg = PgStack.get(Index);
				
				PgInfo[Index] = new PageInfo(Pg.getPageID(), Pg.getFlags(), Pg.getTitle());
			}
		
			return PgInfo;
		}
	}
	
	String getPageState
	(
		final String			PageIDStr
	)
	{
		final Page				Pg = getByPageID(PageIDStr);

		if (Pg == null)
			return "unknown";

		return "known";
	}
}
