/*	QWICAP (Quick Web Interface for Conventional Applications) by Chris W. Johnson
	Copyright (C) 2003-2010 The University of Texas at Austin ("U.T. Austin"). All rights reserved.
	
	U.T. Austin's copyright notice in this evolving version reflects our original authorship. It does not assert 
	copyright over changes contributed by others.
	
	This library is free software; you can redistribute it and/or modify it under the terms of the GNU Lesser General 
	Public License as published by the Free Software Foundation; either version 2.1 of the License, or (at your option) 
	any later version.
	
	This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied 
	warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for 
	more details.
	
	You should have received a copy of the GNU Lesser General Public License along with this library; if not, write to 
	the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/


package edu.utexas.its.eis.tools.qwicap.servlet;


import edu.utexas.its.eis.tools.qwicap.template.xml.mutable.MutableMarkup;


/**	<code>PromptFlags</code> represents the many options that allow you to modify the behavior of Qwicap's 
 * 	<code>prompt</code> and <code>rejectInput</code> methods. It will generally be used in-line within the invocation of
 *	<code>prompt</code> with a chained series of invocations specifying all of the relevant behaviors. For example:
 *
 *	<p><code>Q.prompt(MyMarkup, new PromptFlags().promptWithoutFormData().rejectWithoutFormData().modal());</code></p>
 * 
 *	<p>By default, a new instance of <code>PromptFlags</code> is configured as though you have already made the 
 *	following invocations:
 *	</p>
 *
 *	<p><code>new PromptFlags().promptWithFormData().rejectWithFormData().modal().synchronous()</code></p>
 *
 *	<p>...so there's no need to make those invocations yourself, unless you just want those behaviors to be requested
 *	explicitly for code-clarity purposes.
 *	</p>
 *
 * 	@author Chris W. Johnson
 */
final class PromptFlags implements Cloneable {

	private boolean			PromptWithFormData = true;
	private boolean			RejectWithFormData = true;
	private boolean			ModalPage;
	private boolean			Synchronous        = true;


/* <p>Creates a new <code>PromptFlags</code> instance that is configured to embed the most recent, relevant form data 
 * 	set into the appropriate markup in the page, and to operate modally and synchronously. In other words, it operates 
 * 	as if the following methods had already been invoked:
 * 	</p>
 * 	
 * 	<ul>
 * 		<li>{@link #promptWithFormData}</li>
 * 		<li>{@link #rejectWithFormData}</li>
 * 		<li>{@link #modal}</li>
 * 	</ul>
 *
	PromptFlags() { }
*/
	
/**	Instructs Qwicap to embed the form control values submitted by the user into the relevant page when
 * 	{@link Qwicap#prompt(MutableMarkup, PromptFlags)} is invoked.
 * 	This is the default setting.	
 * 
 * @return					A reference to this <code>PromptFlags</code> instance.
 * @see #promptWithoutFormData()
 */
	PromptFlags promptWithFormData() {

		PromptWithFormData = true;
		return this;
	}

/**	Instructs Qwicap <b>not</b> to embed the form control values submitted by the user into the relevant page when
 * 	{@link Qwicap#prompt(MutableMarkup, PromptFlags)} is invoked.
 * 
 * @return					A reference to this <code>PromptFlags</code> instance.
 * @see #promptWithFormData()
 */
	PromptFlags promptWithoutFormData() {

		PromptWithFormData = false;
		return this;
	}

/**	Instructs Qwicap to embed the form control values submitted by the user into the relevant page when any of the 
 * 	<code>Qwicap.rejectInput</code> methods are invoked. If {@link #promptWithoutFormData} has been invoked, this
 * 	embedding is temporary, and the page is restored to its original state when <code>rejectInput</code> is complete.
 * 	This is the default setting.	
 * 
 * @return					A reference to this <code>PromptFlags</code> instance.
 * @see #rejectWithoutFormData()
 */
	PromptFlags rejectWithFormData() {

		RejectWithFormData = true;
		return this;
	}

/**	Instructs Qwicap <b>not</b> to embed the form control values submitted by the user into the relevant page when any 
 * 	of the <code>Qwicap.rejectInput</code> methods are invoked. Note that this setting is relevant even when 
 * 	{@link #promptWithoutFormData} has been invoked.
 * 
 * @return					A reference to this <code>PromptFlags</code> instance.
 * @see #rejectWithFormData
 */
	PromptFlags rejectWithoutFormData() {

		RejectWithFormData = false;
		return this;
	}

/**	Instructs Qwicap to prompt in a modal manner.
 * 
 * @return					A reference to this <code>PromptFlags</code> instance.
 * @see #modeless
 */
	PromptFlags modal() {

		ModalPage = true;
		return this;
	}

/**	Instructs Qwicap to prompt in a modeless manner. This is the default setting.
 * 
 * @return					A reference to this <code>PromptFlags</code> instance.
 * @see #modal
 */
	PromptFlags modeless() {

		ModalPage = false;
		return this;
	}

/*	Instructs Qwicap to prompt in a synchronous manner. This is the default setting.
 * 
 * @return					A reference to this <code>PromptFlags</code> instance.
 * @see #asynchronous()
 *
	public PromptFlags synchronous() {

		Synchronous = true;
		return this;
	}
*/
/*	Instructs Qwicap to prompt in an asynchronous manner. <b>Not supported, yet.</b>
 * 
 * @return					A reference to this <code>PromptFlags</code> instance.
 * @see #synchronous()
 *
	public PromptFlags asynchronous() {

		throw new UnsupportedOperationException("Sorry, asynchronous prompting has not yet been implemented.");

//		Synchronous = false;
//		return this;
	}
*/
	boolean shouldPromptWithFormData() {
		return PromptWithFormData;
	}

	boolean shouldRejectWithFormData() {
		return RejectWithFormData;
	}

	boolean shouldPromptModally() {
		return ModalPage;
	}

	boolean shouldPromptSynchronously() {
		return Synchronous;
	}

	public Object clone() {
		try {
			return super.clone();
		} catch (CloneNotSupportedException e) {
		//	This will never happen, because "clone" is supported.
			throw new RuntimeException(e);
		}
	}

	@Override public String toString() {
		final StringBuilder		Buff = new StringBuilder(256);

		Buff.append("Prompting ");
		Buff.append(shouldPromptSynchronously() ? "synchronously" : "asynchronously").append(" and ");
		Buff.append(shouldPromptModally() ? "modally " : "modelessly ");

		if (shouldPromptWithFormData()) {

			if (shouldRejectWithFormData())
				Buff.append("with form data always embedded in the page.");
			else
				Buff.append("with form data embedded in the page, but without embedding rejected input.");

		} else {

			if (shouldRejectWithFormData())
				Buff.append("without form data embedded in the page, except when input is rejected.");
			else
				Buff.append("without form data ever being embedded in the page.");
		}
		
		return Buff.toString();
	}
}



