/*	QWICAP (Quick Web Interface for Conventional Applications) by Chris W. Johnson
	Copyright (C) 2003-2010 The University of Texas at Austin ("U.T. Austin"). All rights reserved.
	
	U.T. Austin's copyright notice in this evolving version reflects our original authorship. It does not assert 
	copyright over changes contributed by others.
	
	This library is free software; you can redistribute it and/or modify it under the terms of the GNU Lesser General 
	Public License as published by the Free Software Foundation; either version 2.1 of the License, or (at your option) 
	any later version.
	
	This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied 
	warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for 
	more details.
	
	You should have received a copy of the GNU Lesser General Public License along with this library; if not, write to 
	the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/


package edu.utexas.its.eis.tools.qwicap.servlet;


import edu.utexas.its.eis.tools.qwicap.template.xml.mutable.MutableMarkup;


/**	See {@link Qwicap#prompt()} for an explanation, and examples, of the use of this class.
 * 	
 * 	@author Chris W. Johnson
 */
public class PromptModifiers {

	private final Qwicap		Q;
	private final PromptFlags	Flags;
	private MutableMarkup		Page;


	PromptModifiers
	(
		final Qwicap			Q
	)
	{
		this.Q     = Q;
		this.Flags = new PromptFlags();
	}
	
//	When we support asynchronous prompting, we'll need these methods, but not until then. Synchronous operation will
//	always the be the default, so this presents no backwards-compatibility problems.
//
//	public PromptModifiers synchronously() {
//		Flags.synchronous();
//		return this;
//	}
//
//	public PromptModifiers asynchronously() {
//		Flags.asynchronous();
//		return this;
//	}

/**	Instructs Qwicap to prompt in a modal manner.
 * 
 * @return					A reference to this <code>PromptModifiers</code> instance.
 * @see #modelessly
 */
	public PromptModifiers modally() {
		Flags.modal();
		return this;
	}

/**	Instructs Qwicap to prompt in a modeless manner. This is the default setting, so you never have to invoke it. It 
 * 	is only included for clarity.
 * 
 * @return					A reference to this <code>PromptModifiers</code> instance.
 * @see #modally
 */
	public PromptModifiers modelessly() {
		Flags.modeless();
		return this;
	}

/**	Instructs Qwicap to embed the form control values submitted by the user into the relevant page before it is sent to
 * 	the client. This is the default setting, so you never have to invoke it. It is only included for clarity.
 * 
 * @return					A reference to this <code>PromptModifiers</code> instance.
 * @see #withoutFormData
 */
	public PromptModifiers withFormData() {
		Flags.promptWithFormData();
		return this;
	}

/**	Instructs Qwicap <b>not</b> to embed the form control values submitted by the user into the relevant page.
 * 
 * @return					A reference to this <code>PromptModifiers</code> instance.
 * @see #withFormData
 */
	public PromptModifiers withoutFormData() {
		Flags.promptWithoutFormData();
		return this;
	}

/**	Instructs Qwicap to embed the form control values submitted by the user into the relevant page when any of the 
 * 	<code>Qwicap.rejectInput</code> methods are invoked. If {@link #withoutFormData} has been invoked, this
 * 	embedding is temporary, and the page is restored to its original state when <code>rejectInput</code> is complete.
 * 	This is the default setting, so you never have to invoke it. It is only included for clarity.
 * 
 * @return					A reference to this <code>PromptModifiers</code> instance.
 * @see #rejectingWithoutFormData
 */
	public PromptModifiers rejectingWithFormData() {
		Flags.rejectWithFormData();
		return this;
	}

/**	Instructs Qwicap <b>not</b> to embed the form control values submitted by the user into the relevant page when any 
 * 	of the <code>Qwicap.rejectInput</code> methods are invoked. Note that this setting is relevant even when 
 * 	{@link #withoutFormData} has been invoked.
 * 
 * @return						A reference to this <code>PromptModifiers</code> instance.
 * @see #rejectingWithFormData
 */
	public PromptModifiers rejectingWithoutFormData() {
		Flags.rejectWithoutFormData();
		return this;
	}

/**	Prompts the user with the specified page, in a manner consistent with the previously specified modifiers. Unless 
 * 	this method is invoked, no prompting will take place.
 * 
 * @param XHTMLPage				XHTML page to be sent to the user.
 * @return						A reference to the current <code>Qwicap</code> object, just in case it's convenient.
 * @throws QwicapException		If anything goes wrong, or if the page is abandoned.
 */
	public Qwicap usingPage
	(
		final MutableMarkup		XHTMLPage
	)
	throws QwicapException
	{
		this.Page = XHTMLPage;
		return Q.prompt(XHTMLPage, Flags);
	}

	@Override public String toString() {
		final StringBuilder		Buff = new StringBuilder(256);

		Buff.append(Flags.toString());

		if (Page != null)
			Buff.append(" The page displayed will be \"").append(Page.getMarkupName()).append("\".");
		else
			Buff.append(" At least, that is what will happen when the \"PromptModifiers.usingPage\" method is invoked to tell Qwicap what page you want sent to the client.");
		
		return Buff.toString();
	}
}



/*	

	Q.prompt().withFormData().modally().usingPage(MyMarkup);
	
	Q.prompt().withoutFormData().modelessly().usingPage(MyMarkup);
	
	Q.prompt().rejectingWithoutFormData().usingPage(MyMarkup);
	
	Q.prompt().withoutFormData().rejectingWithFormData().usingPage(MyMarkup);

	Q.prompt().synchronously().usingPage(MyMarkup);

	Q.prompt().asynchronously().usingPage(MyMarkup);


*/