/*	QWICAP (Quick Web Interface for Conventional Applications) by Chris W. Johnson
	Copyright (C) 2003-2011 The University of Texas at Austin ("U.T. Austin"). All rights reserved.
	
	U.T. Austin's copyright notice in this evolving version reflects our original authorship. It does not assert 
	copyright over changes contributed by others.
	
	This library is free software; you can redistribute it and/or modify it under the terms of the GNU Lesser General 
	Public License as published by the Free Software Foundation; either version 2.1 of the License, or (at your option) 
	any later version.
	
	This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied 
	warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for 
	more details.
	
	You should have received a copy of the GNU Lesser General Public License along with this library; if not, write to 
	the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/


//  Qwicap.java
//  qwicap
//
//  Created by Chris W. Johnson on Wed Jun 04 2003.


package edu.utexas.its.eis.tools.qwicap.servlet;


import edu.utexas.its.eis.tools.qwicap.template.css.Results;
import edu.utexas.its.eis.tools.qwicap.template.xml.Markup;
import edu.utexas.its.eis.tools.qwicap.template.xml.Match;
import edu.utexas.its.eis.tools.qwicap.template.xml.TagException;
import edu.utexas.its.eis.tools.qwicap.template.xml.immutable.XMLDocument;
import edu.utexas.its.eis.tools.qwicap.template.xml.immutable.XMLString;
import edu.utexas.its.eis.tools.qwicap.template.xml.mutable.MutableMarkup;
import edu.utexas.its.eis.tools.qwicap.util.ArrayToString;
import edu.utexas.its.eis.tools.qwicap.util.HTMLEntityCodec;
import edu.utexas.its.eis.tools.qwicap.Version;
import org.apache.commons.fileupload.FileItem;
import org.apache.commons.fileupload.FileUploadBase;
import org.apache.commons.fileupload.FileUploadException;

import javax.servlet.ServletContext;
import javax.servlet.ServletOutputStream;
import javax.servlet.http.HttpServletResponse;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.lang.reflect.InvocationTargetException;
import java.math.BigDecimal;
import java.math.BigInteger;
import java.net.MalformedURLException;
import java.net.URL;
import java.security.SecureRandom;
import java.security.Security;
import java.util.ArrayList;
import java.util.Enumeration;
import java.util.Random;
import java.util.logging.Level;
import java.util.logging.Logger;
import java.util.regex.Matcher;
import java.util.regex.Pattern;


/**	The <code>Qwicap</code> class provides the core functionality of the Qwicap (Quick Web Interface for Conventional
 *	Applications) system, including the methods that send web pages to clients, and those that retrieve and verify the 
 *	input returned by those pages, while automatically requesting corrections from users as necessary. It is also the 
 *	thread within which each client's instance of the web application is run.
 *	
 *	<p>Form Input (<a href='http://www.w3.org/TR/html4/interact/forms.html#h-17.13.3.2'>Form Data Set</a>) -Oriented Methods:</p>
 *	<ul>
 *		<li>{@link #clearMostRecentForm}</li>
 *		<li>{@link #populateMostRecentForm}</li>
 *		<li>{@link #get(String)}</li>
 *		<li>{@link #getAll(String)}</li>
 *		<li>{@link #getBigDecimal(String, BigDecimal, BigDecimal, boolean, BigDecimal)}</li>
 *		<li>{@link #getBigDecimals(String, BigDecimal, BigDecimal, boolean, BigDecimal[])}</li>
 *		<li>{@link #getBigInteger(String, BigInteger, BigInteger, boolean, BigInteger)}</li>
 *		<li>{@link #getBigIntegers(String, BigInteger, BigInteger, boolean, BigInteger[])}</li>
 *		<li>{@link #getBoolean(String, boolean, boolean)}</li>
 *		<li>{@link #getBoolean(String, String, String, boolean, boolean)}</li>
 *		<li>{@link #getDouble(String, double, double, boolean, double)}</li>
 *		<li>{@link #getDoubles(String, double, double, boolean, double[])}</li>
 *		<li>{@link #getFloat(String, float, float, boolean, float)}</li>
 *		<li>{@link #getFloats(String, float, float, boolean, float[])}</li>
 *		<li>{@link #getFormDataSet}</li>
 *		<li>{@link #getInt(String, int, int, boolean, int)}</li>
 *		<li>{@link #getInts(String, int, int, boolean, int[])}</li>
 *		<li>{@link #getLong(String, long, long, boolean, long)}</li>
 *		<li>{@link #getLongs(String, long, long, boolean, long[])}</li>
 *		<li>{@link #getString(String, int, int, boolean)}</li>
 *		<li>{@link #getString(String, int, int, boolean, String, Object)}</li>
 *		<li>{@link #getUploadedFile(String, String, boolean)}</li>
 *		<li>{@link #getUploadedFile(String, String[], boolean)}</li>
 *		<li>{@link #getUploadedFileRAMLimit}</li>
 *		<li>{@link #getUploadedFileRepository}</li>
 *		<li>{@link #getUploadedFileSizeLimit}</li>
 *		<li>{@link #setUploadedFileLimits(int, long, String)}</li>
 *		<li>{@link #has(String)}</li>
 *		<li>{@link #hasFormDataSet}</li>
 *		<li>{@link #rejectInput(Object)}</li>
 *		<li>{@link #rejectInput(String, Object, Object)}</li>
 *		<li>{@link #rejectInput(String[], Object[], Object)}</li>
 *	</ul>
 *	
 *	<p>Form-Oriented Methods:</p>
 *	<ul>
 *		<li>{@link #clearMostRecentForm}</li>
 *		<li>{@link #populateMostRecentForm}</li>
 *		<li>{@link #convertSubmitButtonsToInputs(String[])}</li>
 *		<li>{@link #getForm}</li>
 *	</ul>
 *
 *	<p>Page-Oriented Methods:</p>
 *	<ul>
 *		<li>{@link #addDownloadable(String, String, byte[])}</li>
 *		<li>{@link #addDownloadable(String, byte[])}</li>
 *		<li>{@link #convertSubmitButtonsToInputs(String[])}</li>
 *		<li>{@link #getDocument(String)}</li>
 *		<li>{@link #getDocument(URL)}</li>
 *		<li>{@link #goodbye(Downloadable)}</li>
 *		<li>{@link #goodbye(Markup)}</li>
 *		<li>{@link #goodbye(String)}</li>
 * 		<li>{@link #goodbye(String MIMEType, byte[] Data)}</li>
 *		<li>{@link #htmlEncode(String)}</li>
 *		<li>{@link #htmlDecode(String)}</li>
 *		<li>{@link #printPageStack}</li>
 *		<li>{@link #prompt()}</li>
 *		<li>{@link #prompt(MutableMarkup)}</li>
 *		<li>{@link #promptComplete()}</li>
 *		<li>{@link #redirect(String)}</li>
 *		<li>{@link #reportException(Throwable)}</li>
 *		<li>{@link #reportProblem(Object)}</li>
 *		<li>{@link #reportProblem(Object, Object)}</li>
 * 		<li>{@link #showPage(MutableMarkup)}</li>
 *	</ul>
 *
 *	<p>Download-Oriented Methods:</p>
 *	<ul>
 *		<li>{@link #addDownloadable(String, String, byte[])}</li>
 *		<li>{@link #addDownloadable(String, byte[])}</li>
 *		<li>{@link #createDownloadable(String, String, byte[])}</li>
 *	</ul>
 *
 *	<p>Upload-Oriented Methods:</p>
 *	<ul>
 *		<li>{@link #getUploadedFile(String, String, boolean)}</li>
 *		<li>{@link #getUploadedFile(String, String[], boolean)}</li>
 *		<li>{@link #getUploadedFileRAMLimit}</li>
 *		<li>{@link #getUploadedFileRepository}</li>
 *		<li>{@link #getUploadedFileSizeLimit}</li>
 *		<li>{@link #setUploadedFileLimits(int, long, String)}</li>
 *	</ul>
 *
 *	<p>Other Methods:</p>
 *	<ul>
 * 		<li>{@link #addBlockingListener(QwicapBlockingListener)}</li>
 *		<li>{@link #getAuth2SchemeReply}</li>
 *		<li>{@link #getContext}</li>
 *		<li>{@link #getCurrentInstance}</li>
 *		<li>{@link #getInputFilter}</li>
 * 		<li>{@link #getRandom}</li>
 * 		<li>{@link #getStatusReport(boolean)}</li>
 * 		<li>{@link #getStatusReport(Class, boolean)}</li>
 *		<li>{@link #getVersionString}</li>
 *		<li>{@link #invokeInputFilter()}</li>
 * 		<li>{@link #removeBlockingListener(QwicapBlockingListener)}</li>
 *		<li>{@link #setInputFilter(FormDataSetFilter)}</li>
 *	</ul>
 *
 *	@author Chris W. Johnson
 */
public final class Qwicap {

//	todo As a part of the internationalization work, we should probably add getYear and getCurrency methods, since both of those are numbers that are handled differently from other numbers.

	private static final Logger		Log                                = Logger.getLogger(Qwicap.class.getName());

//	todo Consider that these double quote constants may need to be internationalizable.

	static final char				kLeftDoubleQuote                   = '\u201C';	//	201C == Unicode equivalent of "&ldquo;"
	static final char				kRightDoubleQuote                  = '\u201D';	//	201D == Unicode equivalent of "&rdquo;"

	static final HTMLEntityCodec	HTML                               = new HTMLEntityCodec().setEncoding('\n', "<br/>\n");	//	todo Consider whether or not newline encoding is really appropriate.

	private static final String[]	kEmptyStringArray                  = {};

	static final String				kQwicapPackage                     = "edu.utexas.its.eis.tools.qwicap";
	
	static final String				kQwicapWebAppDefaultName           = "Qwicap Web Application";	//	As found in the default "web.xml" file for Qwicap applications.

//	Qwicap HTML classes.

	static final String				kQwicapAutoDeleteClass             = "qwicap-auto-delete";						//	Used only to mark hit-specific elements in web pages; elements that should be automatically deleted after the hit is complete.
	static final String				kQwicapInputProcessingFailureClass = "qwicap-input-processing-failure";			//	Class used to identify "div" elements used for error reports associated with input processing failures.
	static final String				kQwicapBadInputFieldClass          = "qwicap-bad-field";						//	Used to highlight the label for the form control that supplied bad input.
	private static final String		kQwicapBadFormMessageClass         = "qwicap-form-with-bad-field-error-message";//	Class assigned to the error message "div" that's inserted just before the form that supplied the bad input.
	static final String				kQwicapExceptionReportClass        = "qwicap-exception";						//	Class assigned to the exception report "div" that's inserted at the start of the page.
	static final String				kQwicapExceptionMessageClass       = "qwicap-exception-message";				//	Class assigned to elements, typically to "span" elements, whose contents will display an exception message.
	static final String				kQwicapIncidentCodeClass           = "qwicap-incident-code";					//	Class assigned to elements, typically to "span" elements, whose contents will dispay Qwicap incident codes.
	static final String				kQwicapNavigationErrorClass        = "qwicap-page-navigation-error-message";	//	Class assigned to the error report "div" that's inserted at the start of the page when there is a problem with page navigation (page not in-play, modal page, etc.).

//	Qwicap control names.

	static final String				kQwicapControlNamePrefix           = "qwicap";
	static final String				kQwicapPageIDControlName           = "qwicap-page-id";
	static final String				kQwicapFormIDControlName           = "qwicap-form-id";
	static final String				kQwicapCharSetControlName          = "qwicap-charset";
	static final String				kQwicapNoOpControlName             = "qwicap-noop";
	static final String				kQwicapInputSubmitControlName      = "qwicap-input-submit?";
	
	static final int				kDefaultFileUploadSizeLimit        = 16 * 1024;	//	Default maximum size of uploaded file that will be held in memory; larger files are written to the repository.
	static final int				kDefaultFileUploadRAMLimit         = 16 * 1024;	//	Default maximum acceptable size of uploaded files.

	private static final XMLString	InputProcessingErrorMarkup;
	private static final XMLString	RejectedInputErrorMarkup;

	static {
		try{
			InputProcessingErrorMarkup = new XMLString("<div class='" + kQwicapBadFormMessageClass + ' ' + kQwicapAutoDeleteClass + ' ' + kQwicapInputProcessingFailureClass + "'></div>");
			RejectedInputErrorMarkup   = new XMLString("<div class='" + kQwicapBadFormMessageClass + ' ' + kQwicapAutoDeleteClass + "'></div>");
		} catch (TagException e) {
			throw new RuntimeException(e);
		}
	}

	private final QwicapGlobalServices		GlobalServices;
	private final QwicapApplicationServices	ApplicationServices;
	
	private final String					QwicapToString;
	private final QwicapThread				QThread;		//	The thread in which this Qwicap instance is running.
	private final SessionContext			SessCtxt;
	volatile boolean						ShouldWait;		//	Altered by the servlet thread from code in SessionContext.go
	volatile HitContext						HitCtxt;		//	Altered by the servlet thread from code in SessionContext.go
	private Context							PublicCtxt;		//	This is the context we expose through the public API.
	
	private final i18nManager 				i18nMgr;		//	Internationalization manager for this session.
	private final i18nServices				i18n;			//	Internationalization services for this class.

	private Incident						IncidentToReport;

	private Page							NoPromptPage;

	private boolean							IsDone;

	private boolean							PleaseDie;
	private String							PleaseDieReason;
	
	private boolean							GoodbyeInvoked;

	private final PageStack					Pages;

//	Unfortunately, we can't just ask an instance of the DiskFileUpload object for the default value of its various 
//	settings, because it lacks some of them, like the repository path, and will throw a NullPointerException if we
//	"get" them. So, we set our own defaults, and we try to pick them such that the repository setting will be
//	irrelevant at first.
		
	private int								UFInMemorySizeLimit = kDefaultFileUploadRAMLimit;	//	Maximum size of uploaded file that will be held in memory; larger files are written to the repository.
	private long							UFAbsoluteSizeLimit = kDefaultFileUploadSizeLimit;	//	Maximum acceptable size of uploaded files.
	private String							UFRepositoryPath;									//	Location on disk where uploaded files can be written.

	private Matcher[]						ConvertSubmitButtonsToInputsUserAgentPatterns;

	private ArrayList<Downloadable>			NewDownloadables;

	private boolean							InputFilterIsNew;
	private FormDataSetFilter				InputFilter;

	private ArrayList<QwicapBlockingListener> BlockingListeners;

	private final QwicapAJAX				AJAX;

	private int								SessionKillingAttemptsCount;
	private int								SessionDeathExceptionsThrown;
	private int								SessionAbandonedExeptionsThrown;

	private FormDataSet						PrevDataSet = new FormDataSet();	//	This variable is for the exclusive use of the synchronize method.


	Qwicap
	(
		final QwicapGlobalServices		GlobalServices,
		final QwicapApplicationServices ApplicationServices,
		final QwicapThread				QThread,
		final SessionContext			SessCtxt,
		final HitContext				HitCtxt
	)
	throws MalformedURLException
	{
		this.GlobalServices      = GlobalServices;
		this.ApplicationServices = ApplicationServices;
		this.QThread             = QThread;
		this.SessCtxt            = SessCtxt;
		this.HitCtxt             = HitCtxt;
	//	this.Main                = ApplicationServices.getClientClass().getMethod("main", String[].class);
		this.i18nMgr             = ApplicationServices.createi18nManager(HitCtxt.getRequest());
		this.i18n                = i18nMgr.getServicesForClass(Qwicap.class);
		this.Pages               = new PageStack(new URL(HitCtxt.getRequest().getRequestURL().toString()), GlobalServices.getRandom(), i18nMgr);
		this.AJAX                = new QwicapAJAX(this, i18nMgr);
		this.QwicapToString      = "Qwicap, version " + Version.getString() + ", executing client class \"" + ApplicationServices.getClientClass().getName() + "\" in servlet context \"" + ApplicationServices.getServletContextPath() + '"';
		
	//	todo If the ClassLoader used by the servlet is different from this class' ClassLoader, consider changing the ClassLoader used by our i18nManager to the servlet's ClassLoader.
	}

/**	Returns a string identifying this version of Qwicap. The version numbering convention is: "M.N[.R][SB]".
 *	
 *	<ul>
 *	<li>M - Major Version Number
 *	<li>N - Minor Version Number
 *	<li>R - Revision Number, added/incremented when features haven't really changed, but bugs, etc. have been fixed.
 *	<li>S - Build Status:
 *		<ul>
 *		<li>'d' - Development build. Often doesn't leave the developer's machine. Will seldom go beyond the members of 
 *				  the development team.
 *		<li>'a' - Alpha test. Not feature complete, but suitable for early testing, and much more stable than a 
 *				  development build.
 *		<li>'b' - Beta test. Feature complete in every important respect. Available to general testing audience.
 * 		<li>none - Release.
 *		</ul>
 *	<li>B - Build Number. Only used while code is in development, alpha or beta test.
 *	</ul> 
 *
 *	<p>Typically, the date of the build is appended, for example: "1.1 of 29-Apr-2004". The date is in DD-MMM-YYYY 
 *	format because it's fairly easy to read, and it's fairly unambiguous in international contexts.
 *
 * @return Version number string for this build of Qwicap.
 */
	public static String getVersionString() {
		return Version.getString();
	}

/**	Returns a string containing a human-readable report on Qwicap's current (and past) status. Due to the changeable 
 * 	nature of the data gathered, no API for programmatically accessing this data has been defined.
 * 
 * @param WebAppMainClass		The class that contains the <code>public static void main(String[])</code> method which
 * 								is this web application's entry point. This reference to the entry point class is 
 *								required in order to determine which application's status should be returned. (In cases 
 * 								where Qwicap is deployed as a shared library, it may be maintaining status information 
 * 								for any number of web applications.)
 * @param Clear					<code>true</code> if the data in the report should be reset after the report is 
 * 								generated, <code>false</code> if the data should continue to accumulate.
 * @return						A <code>String</code> containing the human-readable status report.
 */
	public static String getStatusReport
	(
		final Class<?>			WebAppMainClass,
		final boolean			Clear
	)
	{
		final QwicapGlobalServices	QGS = QwicapServlet.getGlobalServices();

		if (QGS == null)
			return "N/A";

		final ServiceDataRecorder	SDR = QGS.getSDR(WebAppMainClass.getClassLoader());

		if (SDR == null)
			return "N/A";

		return SDR.getGatheredData(Clear).toString();
	}

/**	Returns a string containing a human-readable report on Qwicap's current (and past) status. Due to the changeable 
 * 	nature of the data gathered, no API for programmatically accessing this data has been defined.
 * 
 * @param Clear					<code>true</code> if the data in the report should be reset after the report is 
 * 								generated, <code>false</code> if the data should continue to accumulate.
 * @return						A <code>String</code> containing the human-readable status report.
 */
	public String getStatusReport
	(
		final boolean			Clear
	)
	{
		return ApplicationServices.getServiceDataRecorder().getGatheredData(Clear).toString();
	}

/**	<p>Returns the pseudo-random number generator (PRNG) used internally by Qwicap. This PRNG is shared by all Qwicap
 * 	threads in a particular web application. It is initialized when Qwicap is first loaded, using various
 * 	system-specific information as its source of entropy, in a manner inspired by the <code>SecureRandom</code> class.
 *	The algorithm is based on a "rule 30", one-dimensional, cellular automaton, as suggested by Stephen Wolfram in his
 * 	book <cite>A New Kind of Science</cite>, <a href='http://www.wolframscience.com/nksonline/page-317'>page 317</a>
 * 	(see also <a href='http://www.wolframscience.com/nksonline/page-974b-text'>page 974</a>). As Wolfram points-out,
 * 	this is the same method that Mathematica has always used to generate its random numbers.
 *	</p>	
 * 
 * 	<p>By using this PRNG in your web applications, you avoid the overhead of creating a new PRNG, and the unpredictable
 * 	interleaving of usage by the various concurrent sessions (threads) in your web application makes the sequence
 * 	obtained by any given session even more unpredictable than it would be anyway.
 *	</p>
 * 
 * 	<p><b>Note:</b> If the application server's security manager permits it, this PRNG implementation is made available 
 * 	through the "Provider" mechanism of the {@link Security} class. Thus, a {@link SecureRandom} instance which utilizes 
 * 	this PRNG implementation, and which, by default, is seeded from Qwicap's internal instance of this PRNG, may be 
 * 	obtained by using the following invocation: <code>SecureRandom.getInstance("Rule30", "Qwicap");</code>. Note that 
 * 	this particular "rule 30" PRNG implementation has not been subjected to FIPS 140-2 compliance testing, though it has 
 * 	been subjected to similar tests, and other "rule 30" PRNG implementations appear to have satisfied FIPS 140-2
 * 	requirements.
 * 	</p>
 * 
 * @return						Qwicap's internal pseudo-random number generator instance.
 */
	public Random getRandom() {
		return GlobalServices.getRandom();
	}

/**	Returns a reference to the current Qwicap thread.	
 * 
 * @return							Current Qwicap thread.
 * @throws IllegalStateException	If invoked from a non-Qwicap thread.
 */
	public static Qwicap getCurrentInstance() {
		try {
			return ((QwicapThread) Thread.currentThread()).getApplication();
		} catch (ClassCastException e) {
			throw new IllegalStateException("This code is not executing within a Qwicap thread, so there is no current instance of Qwicap. The current thread is of type \"" + Thread.currentThread().getClass().getName() + "\".");
		}
	}

/**	Returns the {@link HitContext} object associated with this Qwicap thread.
 * 
 * @return	The HitContext object associated with this Qwicap thread.
 */
	HitContext getHitContext() {

		if (PleaseDie)
			throwSessionKillingException();

		if (HitCtxt == null)
			throw new IllegalStateException("There is no hit context at this time.");

		return HitCtxt;
	}

	SessionContext getSessionContext() {
		return SessCtxt;
	}

/**	Returns the {@link Context} object associated with this Qwicap thread.
 * 
 * @return	The Context object associated with this Qwicap thread.
 */
	public Context getContext() {

		if (PublicCtxt == null)
			PublicCtxt = new Context(this);

		return PublicCtxt;
	}

/**	Returns the web browser (user agent) identity string that was provided by the client with its most recent request.
 * 
 * @return The web browser (user agent) identity string that was provided by the client with its most recent request,
 *         or an empty <code>String</code>, if there was none.
 */
	public String getUserAgent() {

		if (PleaseDie)
			throwSessionKillingException();

		final Enumeration<String>	UserAgents = getHitContext().getRequest().getHeaders("User-Agent");

		if (UserAgents.hasMoreElements())
			return UserAgents.nextElement();
		else
			return "";
	}

/**	Returns the reply object from the authentication &amp; authorization scheme ({@link Auth2Scheme}), if any, that is
 * 	associated with the current hit.
 * 
 * @return	The {@link Auth2SchemeReply} instance associated with the current hit.
 */
	 public Auth2SchemeReply getAuth2SchemeReply() {
		return getHitContext().getAuth2SchemeReply();
	}

	i18nManager geti18n() {
		return i18nMgr;
	}

	public i18nServices geti18nServicesForClass
	(
		final Class<?>			ClientClass
	)
	{
		return i18nMgr.getServicesForClass(ClientClass);
	}

/**	This method is temporary and unsupported. It is part of a potential new feature that is still a work-in-progress.
 * 
 * @return							<code>true</code> if Qwicap is placing scripts in web pages, <code>false</code> if
 * 									not.
 */
	public boolean getScriptsEnabled() {
		return AJAX.getEnabled();
	}

/**	This method is temporary and unsupported. It is part of a potential new feature that is still a work-in-progress.
 * 
 * @param ScriptsEnabled			<code>true</code> if Qwicap is allowed to place scripts in web pages, 
 * 									<code>false</code> if not.
 * @return							A reference to this <code>Qwicap</code> instance, to facilitate method invocation
 * 									chaining.
 */
	public Qwicap setScriptsEnabled
	(
		final boolean				ScriptsEnabled
	)
	{
		AJAX.setEnabled(ScriptsEnabled);
		return this;
	}

	boolean getScriptsWorking() {
		return AJAX.getWorking();
	}

/**	Adds a {@link QwicapBlockingListener} to the list of listeners that will be notified by this instance of
 * 	<code>Qwicap</code> just before and after it blocks to wait for any user action. If this instance of 
 * 	<code>Qwicap</code> is in the process of blocking/unblocking (which can only be the case if this method is invoked 
 * 	from within a <code>QwicapBlockingListener</code>), the specified listener will not be added until after
 * 	the unblocking is complete.
 * 
 * @param Listener					An instance of a class that implements the {@link QwicapBlockingListener} interface.
 * @return							A reference to this <code>Qwicap</code> instance, to facilitate method invocation
 * 									chaining.
 */
	public Qwicap addBlockingListener
	(
		final QwicapBlockingListener	Listener
	)
	{
		if (PleaseDie)
			throwSessionKillingException();

		if (Listener != null) {

			if (BlockingListeners == null)
				BlockingListeners = new ArrayList<QwicapBlockingListener>();

			if (BlockingListeners.contains(Listener) == false)
				BlockingListeners.add(Listener);
		}

		return this;
	}

/**	Removes a {@link QwicapBlockingListener} from the list of listeners that will be notified by this instance of 
 * 	<code>Qwicap</code> just before and after it blocks to wait for any user action. If this instance of 
 * 	<code>Qwicap</code> is in the process of blocking/unblocking (which can only be the case if this method is invoked 
 * 	from within a <code>QwicapBlockingListener</code>), the specified listener will not be removed until after
 * 	the unblocking is complete.
 * 
 * @param Listener					An instance of a class that implements the {@link QwicapBlockingListener} interface,
 * 									and which has previously been passed to the {@link #addBlockingListener addBlockingListener} 
 * 									method of this <code>Qwicap</code> instance.
 * @return							A reference to this <code>Qwicap</code> instance, to facilitate method invocation
 * 									chaining.
 */
	public Qwicap removeBlockingListener
	(
		final QwicapBlockingListener	Listener
	)
	{
		if (BlockingListeners != null) {

			BlockingListeners.remove(Listener);

			if (BlockingListeners.isEmpty())
				BlockingListeners = null;
		}

		return this;
	}

/**	Sets an object that will filter all input received from the client machine before Qwicap passes it along to the 
 *	client application. There can be only one filter at a time. To remove the current filter, pass <code>null</code>.
 * 	
 * 	<p>The data set (if any) sent with the current hit is filtered the next time an invocation of a <code>prompt</code> 
 * 	method, or the {@link #redirect redirect} method, occurs, unless {@link #invokeInputFilter invokeInputFilter} is 
 * 	used before then.
 * 	</p>
 * 
 * @param Filter					The object which will filter all input received from the client machine. To 
 *									specify that there should be no filter, pass <code>null</code>.
 * @return							A reference to this <code>Qwicap</code> instance, to facilitate method invocation
 * 									chaining.
 */
	public Qwicap setInputFilter
	(
		final FormDataSetFilter	Filter
	)
	{
		if (PleaseDie)
			throwSessionKillingException();

		InputFilter      = Filter;
		InputFilterIsNew = true;

		return this;
	}

/**	Invokes the {@link FormDataSetFilter} filter that was set using {@link #setInputFilter setInputFilter}, if any. The 
 * 	filter is sent the data set that arrived with the current hit.
 *
 * @return							A reference to this <code>Qwicap</code> instance, to facilitate method invocation
 * 									chaining.
 * @throws QwicapException			If execution of the filter produces any exceptions.
 */
	public Qwicap invokeInputFilter() throws QwicapException {

		if (PleaseDie)
			throwSessionKillingExceptions();

	//	Run the filter on the current hit's data set.

		invokeInputFilter(getHitContext().getFormDataSet());

		return this;
	}

	void invokeInputFilter
	(
		final FormDataSet		DataSet
	)
	throws QwicapException
	{
		InputFilterIsNew = false;

		if (InputFilter != null) {
			final FormDataSetFilterResponse Actions = new FormDataSetFilterResponse(Pages, DataSet);

			InputFilter.filter(DataSet, new FormDataSetFilterContext(getHitContext()), Actions);

			Actions.execute();
		}
	}

/**	Returns the object that is currently filtering all input received from the client machine.
 * 
 * @return							The current filter object.
 */
	public FormDataSetFilter getInputFilter() {
		return InputFilter;
	}

	PageStack getPageStack() {
		return Pages;
	}

	void setNoPromptPage
	(
		final Page				NoPromptPage
	)
	{
		this.NoPromptPage = NoPromptPage;
	}

/**	Loads from disk, or retrieves from the document cache, an immutable representation of the XML document located on 
 *	the local file system at the specified path. When a requested document is found in the cache, the file on disk from 
 *	which it was originally read is checked for modification by examining its size and modification date. If found to 
 *	have been modified, the document is reloaded and the cache updated. In order to pass the document to the 
 *	{@link #prompt(MutableMarkup)} or {@link #goodbye(Markup)} methods, or to do anything else that modifies the 
 *	document, you'll need to acquire a mutable version of the document by using the {@link XMLDocument#getMutable} 
 *	method.
 *
 * @param FileName					Path to the document on the local file system.
 * @return							An immutable representation of the requested document.
 * @throws FileNotFoundException	If the specified file was not found.
 * @throws TagException				If there was a problem with parsing the document.
 * @throws IOException				If there was a problem with reading the document.
 * @see XMLDocument#getMutable
 */
	public XMLDocument getDocument
	(
		final String			FileName
	)
	throws TagException, IOException
	{
		return i18n.getDocument(FileName);
	}

/**	Loads from the specified URL, or retrieves from the document cache, an immutable representation of an XML document. 
 *	Documents originally loaded from URLs are not cecked for modifications when they are found in the cache, so the 
 *	cache can become stale. In order to pass the document to the {@link #prompt(MutableMarkup)} or 
 *	{@link #goodbye(Markup)} methods, or to do anything else that modifies the document, you'll need to acquire a 
 *	mutable version of the document by using the {@link XMLDocument#getMutable} method.
 * 
 * @param DocumentURL				URL specifying the location of the document.
 * @return							An immutable representation of the requested document.
 * @throws TagException				If there was a problem with parsing the document.
 * @throws IOException				If there was a problem with reading the document.
 * @see XMLDocument#getMutable
 */
	public XMLDocument getDocument
	(
		final URL				DocumentURL
	)
	throws TagException, IOException
	{
		return i18n.getDocument(DocumentURL);
	}

/**	Requests that this Qwicap thread exit in an orderly fashion, and as soon as possible. Specifically, it wakes up
 *	this thread if it was sleeping, and causes <code>SessionDeathException</code>s to be thrown at it until it finally 
 *	agrees to die.
 * 
 * 	@param Reason					A <code>String</code> explaining why the application has been asked to exit.
 */
	void pleaseDie
	(
		final String			Reason
	)
	{

	//	Only set PleaseDieReason the FIRST time pleaseDie is invoked, otherwise subsequent invocations will 
	//	obscure the real reason the application was asked to die.

		if (!PleaseDie) {

			PleaseDie       = true;
			PleaseDieReason = Reason;
		}

	//	Log.log(Level.INFO, "pleaseDie invoked. Reason: {0}", Reason);

		if (!IsDone)
			QThread.interrupt();
	}

	boolean shouldDie() {
		return PleaseDie;
	}

	boolean isDone() {
		return IsDone || !QThread.isAlive();
	}

/**	When <code>PleaseDie</code> is <code>true</code>, this method should be called by any Qwicap public method that 
 * 	supports user interaction, but cannot throw <code>QwicapAbandonmentException</code>.
 * 	
 *	@throws QwicapSessionDeathException
 */
	private void throwSessionKillingException() {
		
	//	Emergency means of killing a session, when the normal means have failed many times.
		
		if (SessionKillingAttemptsCount == 25)
			throw new QwicapSessionWillNotDieError(this, SessionKillingAttemptsCount);
		
		SessionKillingAttemptsCount++;
		
	//	Normal means of killing a session.
		
		SessionDeathExceptionsThrown++;
		throw new QwicapSessionDeathException(PleaseDieReason);
	}

/**	When <code>PleaseDie</code> is <code>true</code>, this method should be called by any Qwicap public method that 
 * 	supports user interaction, and is allowed to throw <code>QwicapAbandonmentException</code>.
 * 	
 *	@throws QwicapSessionDeathException			Our first choice for an exception to force an application to exit.
 *	@throws QwicapSessionAbandonedException		Our second choice for an exception to force an application to exit.
 *  @throws QwicapSessionWillNotDieError		Our last-ditch method for compelling an application to exit.
 */
	private void throwSessionKillingExceptions() throws QwicapSessionAbandonedException {
		
	//	Emergency means of killing a session, when the normal means have failed many times.
		
		if (SessionKillingAttemptsCount == 25)
			throw new QwicapSessionWillNotDieError(this, SessionKillingAttemptsCount);
		
		SessionKillingAttemptsCount++;
		
	//	Normal means of killing a session.
		
		if (SessionAbandonedExeptionsThrown >= 3) {

		//	We've reached the end of a cycle. Reset both counters.

			SessionAbandonedExeptionsThrown = 0;
			SessionDeathExceptionsThrown    = 0;
		}

		if (SessionDeathExceptionsThrown < 3) {

		//	First, we make three tries to kill this web application session by throwing QwicapSessionDeathExceptions.

			SessionDeathExceptionsThrown++;
			throw new QwicapSessionDeathException(PleaseDieReason);

		} else if (SessionAbandonedExeptionsThrown < 3) {

		//	Second, we make three tries to kill this web application session by throwing QwicapSessionAbandonedExceptions.
		//	The Qwicap "prompt pattern" gives these a very good chance of getting through even when 
		//	QwicaSessionDeathExceptions are being caught. Unfortunately, not all of the potentially relevant methods 
		//	are declared to throw QwicapAbandonmentExceptions, so we can't rely solely on this mechanism.

			SessionAbandonedExeptionsThrown++;
			throw new QwicapSessionAbandonedException(PleaseDieReason);
		}
	}

/**	Only invoked by the <code>QwicapServlet</code> class.
 */
	void run() {
		
		Log.log(Level.FINEST, "RUN: Invoking \"public static void main(String[])\" method of {0} using thread <{1}>. (HitContext = {2})", new Object[] { ApplicationServices.getClientClass(), Thread.currentThread(), HitCtxt });
		
		Throwable				Failure = null;
		
		try {
			
		//	Invoke the client applications "public static void main(String[] Args)" method.
			
			ApplicationServices.getClientMain().invoke(null, new Object[] { ApplicationServices.getClientMainArgsCopy() });
			
		} catch (InvocationTargetException e) {
			final Throwable			Target = e.getTargetException();

			if (Target instanceof QwicapSessionDeathException)
				Log.log(Level.FINE, "RUN: Application session killed (QwicapSessionDeathException). Reason: {0}", Target.getMessage());
			else if (Target instanceof QwicapSessionAbandonedException)
				Log.log(Level.FINE, "RUN: Application session killed (QwicapSessionAbandonedException). Reason: {0}", Target.getMessage());
			else
				Failure = Target;
		
		} catch (Exception e) {

			Failure = e;

		} finally {

			IsDone = true;
			
			try {
				
				if (Failure != null) {
					final IncidentThrowable	Incident = new IncidentThrowable(i18n, Failure, ApplicationServices.isInDevelopmentMode(), GlobalServices.getRandom());
					
					Log.log(Level.SEVERE, "{0}, exited due to incident {1}.", new Object[] { this, Incident.getIncidentCode() });
				
					if (HitCtxt != null)
						HitCtxt.getServlet().showCrashPage(HitCtxt.getRequest(), HitCtxt.getResponse(), Incident);
				
				} else if (!GoodbyeInvoked) {
					
					if (!PleaseDie)
						Log.log(Level.WARNING, "{0}, exited without invoking the Qwicap.goodbye method.", toString());
					
					if (HitCtxt != null)
						HitCtxt.getServlet().showExitPage(HitCtxt.getRequest(), HitCtxt.getResponse());
				}
	
				Log.log(Level.FINEST, "RUN: Session complete for application implemented by {0} using thread <{1}>. (HitContext = {2})", new Object[] { ApplicationServices.getClientClass(), Thread.currentThread(), HitCtxt });
			
			} finally {	
				
			//	Our SessionContext object (SessCtxt) is waiting for a signal that we've finished processing the latest 
			//	hit (HitCtxt). This is that signal (and the last one we'll ever send).
	
				if (HitCtxt != null)
					HitCtxt.processingComplete();
			}
		}
	}
	
	@Override public String toString() {
		return QwicapToString;
	}
	
	
/**	Instructs Qwicap to insert a problem report message into the page next passed to either the <code>prompt</code>, 
 * 	or the <code>goodbye</code>, methods. The problem description will also be provided to the server's log (level 
 * 	"FINE", class "edu.utexas.its.eis.tools.qwicap.servlet.IncidentProblem").
 * 	
 * @param ProblemDescription            The object that will supply the text of the problem report that is to be 
 * 										inserted into the page markup. Generally, this will be a <code>String</code> 
 * 										object, but it can be any type of object. Certain common classes (for example 
 * 										<code>Markup</code> and its subclasses, <code>Results</code>, etc.) receive 
 * 										special handling appropriate to their functions. Other objects are handled 
 * 										merely by invoking their <code>toString</code> methods and inserting the 
 * 										resulting <code>String</code> object into the page markup. <b>Note:</b> Passing 
 * 										a {@link Throwable}, or any of its subclasses, will cause 
 * 										{@link #reportException} to be invoked. This behavior prevents 
 * 										<code>reportException</code>'s security features from being bypassed merely by 
 * 										the expedient of passing exceptions to <code>reportProblem</code>, instead. To
 * 										maintain the security of their applications, it is important for developers to
 * 										carefully consider the nature of the information that they allow their 
 * 										applications to pass along to users (who may be attackers). 
 * @return								A reference to this <code>Qwicap</code> instance, to facilitate method 
 * 										invocation chaining.
 * @throws QwicapSessionDeathException	If you attempt to report a <code>QwicapSessionDeathException</code>, it will
 *										be rethrown, instead. This exception should not be caught, let alone reported.
 * @throws QwicapAbandonmentException	If you attempt to report any subclass of <code>QwicapAbandonmentException</code>,
 *										it will be rethrown, instead. This exception should not be reported, and 
 *										shouldn't even be caught, except when implementing the standard Qwicap "prompt"
 *										pattern.
 */
	public Qwicap reportProblem
	(
		final Object			ProblemDescription
	)
	throws QwicapSessionDeathException, QwicapAbandonmentException
	{
		if (ProblemDescription instanceof Throwable)
			return reportException((Throwable) ProblemDescription);
		
		IncidentToReport = new IncidentProblem(i18n, ProblemDescription, null, false);
		
		if (PleaseDie)
			throwSessionKillingExceptions();
		
		return this;
	}

/**	Instructs Qwicap to insert a problem report message into the page next passed to either the <code>prompt</code>, 
 * 	or the <code>goodbye</code>, methods. The problem description and associated debugging information will also be 
 * 	provided to the server's log (level "FINE", class "edu.utexas.its.eis.tools.qwicap.servlet.IncidentProblem"). 
 * 	Additionally, when Qwicap is operating in its low-security "development" mode (the "qwicap-development-mode" 
 * 	context parameter in the web application's "web.xml" file is set to "true"), the debug information will also be 
 * 	placed into an HTML comment block and embedded into the page along with the user-visible problem report message. 
 * 	Conversely, when Qwicap is operating in its default, high-security mode, the debugging information will <i>never</i> 
 * 	be included in <i>any</i> web page.
 * 	
 * @param ProblemDescription            The object that will supply the text of the problem report that is to be 
 * 										inserted into the page markup. Generally, this will be a <code>String</code> 
 * 										object, but it can be any type of object. Certain common classes (for example 
 * 										<code>Markup</code> and its subclasses, <code>Results</code>, etc.) receive 
 * 										special handling appropriate to their functions. Other objects are handled 
 * 										merely by invoking their <code>toString</code> methods and inserting the 
 * 										resulting <code>String</code> object into the page markup. <b>Note:</b> Passing 
 * 										a {@link Throwable}, or any of its subclasses, will cause 
 * 										{@link #reportException} to be invoked. This behavior prevents 
 * 										<code>reportException</code>'s security features from being bypassed merely by 
 * 										the expedient of passing exceptions to <code>reportProblem</code>, instead. To
 * 										maintain the security of their applications, it is important for developers to
 * 										carefully consider the nature of the information that they allow their 
 * 										applications to pass along to users (who may be attackers). 
 * @param DebugInformation	            An object that, when converted to a <code>String</code> will supply debugging 
 * 										information relevant to this problem.
 * @return								A reference to this <code>Qwicap</code> instance, to facilitate method 
 * 										invocation chaining.
 * @throws QwicapSessionDeathException	If you attempt to report a <code>QwicapSessionDeathException</code>, it will
 *										be rethrown, instead. This exception should not be caught, let alone reported.
 * @throws QwicapAbandonmentException	If you attempt to report any subclass of <code>QwicapAbandonmentException</code>,
 *										it will be rethrown, instead. This exception should not be reported, and 
 *										shouldn't even be caught, except when implementing the standard Qwicap "prompt"
 *										pattern.
 */
	public Qwicap reportProblem
	(
		final Object			ProblemDescription,
		final Object			DebugInformation
	)
	throws QwicapSessionDeathException, QwicapAbandonmentException
	{
		if (ProblemDescription instanceof Throwable)
			return reportException((Throwable) ProblemDescription);
	
		IncidentToReport = new IncidentProblem(i18n, ProblemDescription, DebugInformation, ApplicationServices.isInDevelopmentMode());
	
		if (PleaseDie)
			throwSessionKillingExceptions();
	
		return this;
	}
	
/**	Instructs Qwicap to insert an error message relating to the specified exception into the page next passed to either 
 * 	the <code>prompt</code>, or the <code>goodbye</code>, methods. The exception will be provided to the server's 
 * 	log (level "SEVERE", class "edu.utexas.its.eis.tools.qwicap.servlet.IncidentThrowable"). Additionally, when Qwicap 
 * 	is operating in its low-security "development" mode (the "qwicap-development-mode" context parameter in the web 
 * 	application's "web.xml" file is set to "true"), the exception's full stack trace will be placed into an HTML comment 
 * 	block and embedded into the page along with the user-visible error message. Conversely, when Qwicap is operating in 
 * 	its default, high-security mode, neither the exception's message, nor its stack trace, will <i>ever</i> be included 
 * 	in <i>any</i> web page. Instead, only an "incident code"&mdash;comprised of the date and a random hexadecimal 
 * 	sequence, for example "20080131-A79DF0"&mdash;will be reported to the user. The incident code is logged with the 
 * 	exception, so that when a user reports a problem and cites its incident code, personnel with access to the server's 
 * 	logs can easily identify the relevant log entry.
 * 	
 * @param T								The <code>Throwable</code> to be reported.
 * @return								A reference to this <code>Qwicap</code> instance, to facilitate method 
 * 										invocation chaining.
 * @throws QwicapSessionDeathException	If you attempt to report a <code>QwicapSessionDeathException</code>, it will
 *										be rethrown, instead. This exception should not be caught, let alone reported.
 * @throws QwicapAbandonmentException	If you attempt to report any subclass of <code>QwicapAbandonmentException</code>,
 *										it will be rethrown, instead. This exception should not be reported, and 
 *										shouldn't even be caught, except when implementing the standard Qwicap "prompt"
 *										pattern.
 * @see #reportProblem
 */
	public Qwicap reportException
	(
		final Throwable			T
	)
	throws QwicapSessionDeathException, QwicapAbandonmentException
	{
		if (T instanceof QwicapSessionDeathException)
			throw (QwicapSessionDeathException) T;
		if (T instanceof QwicapSessionWillNotDieError)
			throw (QwicapSessionWillNotDieError) T;

		QwicapAbandonmentException.rethrowIfThisPageWasAbandoned(T);

		IncidentToReport = new IncidentThrowable(i18n, T, ApplicationServices.isInDevelopmentMode(), getRandom());
	//	Log.log(Level.WARNING, "Qwicap.reportException was invoked to report the following exception.", T);
		
		if (PleaseDie)
			throwSessionKillingExceptions();
		
		return this;
	}

	private void insertIncidentReport
	(
		final Page				Pg
	)
	{
	//	If there is an exception to report, insert it.

		if (IncidentToReport != null)
			Pg.insertIncidentReport(IncidentToReport);
		
	//	Done. Clear the exception.

		IncidentToReport = null;
	}

/**	Instructs Qwicap that when certain browsers are encountered, each "button" element with a "type" attribute set
 *	to "submit" in any page should be converted to an "input" element with a "type" attribute set to "submit". This
 *	works-around hideous bugs in Internet Explorer for both Macintosh and Windows. (Ordinarily, Internet Explorer for
 * 	either platform can be identified using the regular expression <samp>"; MSIE [0-9]+[.][0-9]+([a-z]+[0-9]*)?;"</samp>.)
 *
 *	<p>This conversion works hard to be transparent. For one thing, once it is enabled using this method, the conversion 
 * 	takes place automatically, every time you invoke a <code>prompt</code> method, and the conversion is then 
 * 	automatically <i>undone</i> before the <code>prompt</code> method returns control to your code. The goal is that 
 * 	your code shouldn't be affected by the conversion at all. However, it isn't practical to make the process perfect. 
 * 	For instance, a "button" element like "<code>&lt;button name='shootto' value='17' type='submit' id='shoot3'&gt;Shoot 
 * 	Into &lt;span class='room'&gt;17&lt;/span&gt;&lt;/button&gt;</code>" will loose the "span" element inside its button 
 * 	label during the conversion, and it won't get it back when the conversion is reversed. By way of illustration, the 
 * 	"button" element shown above will be converted into the following "input" element: "<code>&lt;input type='submit' 
 * 	value='Shoot Into 17' name='qwicap-input-submit?shootto=17' id='shoot3'/&gt;</code>". When the conversion is 
 * 	reversed, that will yield the following "button" element: "<code>&lt;button type='submit' name='shootto' value='17' 
 * 	id='shoot3'&gt;Shoot Into 17&lt;/button&gt;</code>". If your code manipulates that same markup in a manner that 
 * 	depends on the presence of the "span" elements in the "button" titles, it will not acheive the desired result.
 *	</p>
 * 
 * @param UserAgentRegularExpressions	An array of regular expression strings that will be used to identify the 
 *										browser(s) for which this conversion should be performed. If any of them match
 *										any portion of the browser's user agent string, the conversion will be 
 *										performed.
 * @return								A reference to this <code>Qwicap</code> instance, to facilitate method 
 * 										invocation chaining.
 */
	public Qwicap convertSubmitButtonsToInputs
	(
		final String[]			UserAgentRegularExpressions
	)
	{
		final int				Count = UserAgentRegularExpressions.length;

		ConvertSubmitButtonsToInputsUserAgentPatterns = new Matcher[Count];

		for (int Index = 0; Index < Count; Index++)
			ConvertSubmitButtonsToInputsUserAgentPatterns[Index] = Pattern.compile(UserAgentRegularExpressions[Index]).matcher("");

		return this;
	}

/**	If you have a web page which has been constructed such that the user cannot possibly interact with it incorrectly,
 * 	this method can be used to send that page to the client instead of <code>prompt</code>, and without implementing the 
 * 	"prompt pattern". Typically, such pages are used to non-interactively display information to the user, and the user
 * 	is given no option except to acknowledge the information. Therefore, there is no possibility of having to display 
 * 	the page again in response to an input error. Nor does your code need any input from the page to determine what to 
 * 	do next. Whatever input the user supplies to the page, typically by clicking on an acknowledgement button (for 
 * 	example "OK"), is automatically discarded.
 * 
 * @param XHTMLPage						XHTML page to be sent to the user.
 * @return								A reference to this <code>Qwicap</code> instance, to facilitate method 
 * 										invocation chaining.
 * @throws QwicapException				If anything goes wrong, or if the page is abandoned.
 */
	public Qwicap showPage
	(
		final MutableMarkup		XHTMLPage
	)
	throws QwicapException
	{
		try {
			return prompt(XHTMLPage, new PromptFlags());
		} finally {
			promptComplete();
		}
	}

/** This method allows a method that displays a page (by implementing the "prompt" pattern for it) to explicitly inform 
 *	Qwicap when it has finished with the page. Qwicap then regards the page as dead, and knows to ignore input 
 *	from cached copies of it. Frequently, it is isn't necessary to invoke this method, because Qwicap will automatically 
 *	figure-out that a page is dead if its supporting method (the one that implemented the "prompt" pattern for it) has 
 *	exited before the next invocation of one of the "prompt" methods. Nonetheless, there are situations where the 
 *	automatic dead page detection is not sufficient, and this method will let you disambiguate those situations. 
 *	Failure to use this method when it is necessary can lead to some sub-optimal page management, but won't lead to a 
 *	crisis, so think of this method as a page navigation optimization, rather than a necessity.
 * 
 * @return								A reference to this <code>Qwicap</code> instance, to facilitate method 
 * 										invocation chaining.
 */
	public Qwicap promptComplete() {
		Pages.popPageAssociatedWithInvokingMethod();
		return this;
	}

/**	Sends an XHTML page to the user, and waits for their response. The form control values returned by the user are
 *	automatically embedded into the supplied XHTML page.
 * 
 * @param XHTMLPage						XHTML page to be sent to the user.
 * @return								A reference to this <code>Qwicap</code> instance, to facilitate method 
 * 										invocation chaining.
 * @throws QwicapException				If anything goes wrong, or if the page is abandoned.
 * @see #goodbye(Markup)
 */
	public Qwicap prompt
	(
		final MutableMarkup		XHTMLPage
	)
	throws QwicapException
	{
		return prompt(XHTMLPage, new PromptFlags());
	}

/**	Begins the process of setting-up a prompt to send an XHTML page to the user. An instance of the {@link PromptModifiers}
 * 	class is returned. By invoking various methods on it, the behavior of the prompt operation may be fully specified,
 * 	and the prompt is finally performed by invoking {@link PromptModifiers#usingPage PromptModifiers.usingPage}. This 
 * 	sounds cumbersome, but it's intended that all of this be performed in-line, and in an IDE that will show you the 
 * 	available methods at any given time. Used in that context it should be convenient and yield highly readable code. 
 * 	See the examples below:
 * 
 * 	<pre>Q.prompt().modelessly().withFormData().rejectingWithFormData().usingPage(MyMarkup);
 *Q.prompt().modally().withoutFormData().rejectingWithoutFormData().usingPage(MyMarkup);
 *</pre>
 * 
 * 	<p>The first example actually represents the default <code>prompt</code> configuration, and would ordinarily be 
 * 	written merely as "<code>Q.prompt().usingPage(MyMarkup);</code>". The second example represents the opposite of the
 * 	default configuration, and is, therefore, the worst case. The {@link PromptModifiers} methods may be invoked in any 
 * 	order, except for {@link PromptModifiers#usingPage(MutableMarkup) PromptModifiers.usingPage}, which must always come 
 * 	last, because it triggers the actual "prompt" operation. Like the various <code>Qwicap.prompt</code> methods, it 
 * 	returns a reference to your Qwicap instance, just in case it's convenient.
 * 	</p>
 * 
 * @return								A <code>PromptModifiers</code> object to allow the behavior of the prompt to 
 * 										be customized.
 */
	public PromptModifiers prompt() {
		return new PromptModifiers(this);
	}

/**	Sends an XHTML page to the user, and waits for their response.
 * 
 * @param XHTMLPage						XHTML page to be sent to the user.
 * @param Flags							A {@link PromptFlags} object supplying various configuration information that 
 * 										will affect the way this page behaves.
 * @return								A reference to this <code>Qwicap</code> object, just in case it's convenient.
 * @throws QwicapException				If anything goes wrong, or if the page is abandoned.
 * @see #prompt()
 * @see #goodbye(Markup)
 */
	Qwicap prompt
	(
		final MutableMarkup		XHTMLPage,
		final PromptFlags		Flags
	)
	throws QwicapException
	{
		if (PleaseDie)
			throwSessionKillingExceptions();

		if (NoPromptPage != null) {

		//	If we've come back to this page as the result of a page abandonment, we don't want to prompt again (the 
		//	user has already provided us with a fresh response to the previous version of the page we sent to them, 
		//	so prompting would be worse than redundant). We want to do nothing, while letting the caller believe it 
		//	has prompted so it will retrieve and process the new form data set (the one that caused the page
		//	abandonment exception to be thrown).

			if (NoPromptPage.isSameInvocationAs(new SimplifiedStackTrace())) {

				NoPromptPage = null;
				return this;
			}
		}

	//	Set the current page.

		promptLow(Pages.setPage(XHTMLPage, Flags));

		return this;
	}

/**	Marks the form control that supplied the rejected input, inserts an error message immediately before the form,
 *	and sends the page containing the form back to the user for correction. All of the user's input is embedded into the 
 *	form when it is sent back, so the user will not have to re-enter the form contents in order to correct the rejected 
 *	input. The web application blocks when <code>rejectInput</code> is invoked, and unblocks only after potentially 
 * 	valid input is received from the user.
 *	
 *	<p>If the ErrorMsg parameter contains one or more XML 
 *	elements marked with the class "qwicap-label-text", the contents of those elements will be replaced with the labels 
 *	of the form elements that supplied the rejected parameters. By the same token, the contents of elements marked with 
 *	the class "qwicap-param-value" will be replaced with the values that are being rejected. For example, if a 
 *	form element labelled "Wombat Burrow" was rejected, and the ErrorMsg markup was as follows:
 *	<br>
 *	<br><samp>The input &lt;span class='qwicap-param-value'&gt;&lt;/span&gt is not a valid value for the field 
 *		&lt;span class='qwicap-label-text'&gt;&lt;/span&gt;. Please fix it.</samp>
 * 	<br>
 *	<br>...then the error message that appeared above the form which supplied the parameter would end-up reading:
 *	<br>
 *	<br><samp>The input "42" is not a valid value for the field Wombat Burrow. Please fix it.</samp>
 *	</p>
 *	
 *	<p>A side-effect of this feature, in which the error message is treated as an XHTML fragment, is that the error 
 *	message string must not contain reserved characters like chevrons ('&lt;' and '&gt;') unless they are parts of 
 *	properly formed XHTML tags. When including arbitrary strings in the error message, encode them using 
 *	the {@link #htmlEncode htmlEncode} method before incorporating them into the message.
 *	</p>
 *
 * @param ControlName				Name of the control whose input is being rejected.
 * @param ControlValue				The input that is being rejected. Can be <code>null</code>; see 
 * 									{@link Form#getControl Form.getControl}.
 * @param ErrorMsgIsXHTMLFragment	<code>true</code> if the <code>ErrorMsg</code> string contains XHTML markup. In 
 *									this case, the string is not run through the {@link #htmlEncode htmlEncode} method. 
 * 									If <code>false</code>, the string is encoded with {@link #htmlEncode htmlEncode}, 
 * 									which prevents arbitrary strings from causing parsing exceptions, or being used in 
 * 									cross-site scripting attacks. Encoding of strings is part of the default behavior of 
 * 									the other <code>rejectInput</code> method, {@link #rejectInput(String, Object, Object)},
 *									so using this method with <code>ErrorMsgIsXHTMLFragment</code> set to 
 *									<code>false</code> is unnecessary; {@link #rejectInput(String, Object, Object)}
 *									can be used instead.
 * @param ErrorMsg					Error message to be placed above the form. If this string is a valid XHTML fragment,
 *									set the parameter <code>ErrorMsgIsXHTMLFragment</code> to <code>true</code>. In all
 *									other cases, pass <code>false</code>.
 * @throws QwicapException			Always thrown by this method in order to restart the caller's input processing 
 * 									logic. (The Qwicap "prompt pattern" ensures this behavior on the part of the 
 * 									client.)
 * @deprecated 						Use {@link #rejectInput(String, Object, Object)} instead, passing an 
 * 									{@link XMLString} if the error message is an XHTML fragment, or a 
 * 									<code>String</code>, if it is not XHTML.
 */
	@Deprecated public void rejectInput
	(
		final String			ControlName,
		final Object			ControlValue,
		final boolean			ErrorMsgIsXHTMLFragment,
		final String			ErrorMsg
	)
	throws QwicapException
	{
		if (ErrorMsgIsXHTMLFragment)
			rejectInput(ControlName, ControlValue, new XMLString(ErrorMsg));
		else
			rejectInput(ControlName, ControlValue, ErrorMsg);
	}

/**	Marks the form control that supplied the rejected input, inserts an error message immediately before the form,
 *	and sends the page containing the form back to the user for correction. All of the user's input is embedded into the 
 *	form when it is sent back, so the user will not have to re-enter the form contents in order to correct the rejected 
 *	input. The web application blocks when <code>rejectInput</code> is invoked, and unblocks only after potentially 
 * 	valid input is received from the user.
 *	
 *	<p>If the ErrorMsg parameter is any subclass of {@link Markup}, and it contains one or more XML 
 *	elements marked with the class "qwicap-label-text", the contents of those elements will be replaced with the labels 
 *	of the form elements that supplied the rejected parameters. By the same token, the contents of elements marked with 
 *	the class "qwicap-param-value" will be replaced with the values that are being rejected. For example, if a 
 *	form element labelled "Wombat Burrow" was rejected, and the ErrorMsg markup was as follows:
 *	<br>
 *	<br><samp>The input &lt;span class='qwicap-param-value'&gt;&lt;/span&gt is not a valid value for the field 
 *		&lt;span class='qwicap-label-text'&gt;&lt;/span&gt;. Please fix it.</samp>
 * 	<br>
 *	<br>...then the error message that appeared above the form which supplied the parameter would end-up reading:
 *	<br>
 *	<br><samp>The input "42" is not a valid value for the field Wombat Burrow. Please fix it.</samp>
 *	</p>
 *	
 *	<p>If the ErrorMsg parameter is of any other type, the rules documented in {@link MutableMarkup#insert MutableMarkup.insert} 
 * 	will be applied. (Strictly speaking, those rules always apply, even when the message is a subclass of {@link Markup}.)
 *	</p>
 *
 * @param ControlName				Name of the control whose input is being rejected.
 * @param ControlValue				The input that is being rejected. Can be <code>null</code>; see 
 * 									{@link Form#getControl Form.getControl}.
 * @param ErrorMsg					Error message to be placed above the form. This parameter can be any class of 
 *									object. See {@link MutableMarkup#insert MutableMarkup.insert} for the specific rules 
 * 									employed in the handling of various classes.
 * @throws QwicapException			Always thrown by this method in order to restart the caller's input processing 
 * 									logic. (The Qwicap "prompt pattern" ensures this behavior on the part of the 
 * 									client.)
 */
	public void rejectInput
	(
		final String			ControlName,
		final Object			ControlValue,
		final Object			ErrorMsg
	)
	throws QwicapException
	{
		rejectInput(new String[] { ControlName }, new Object[] { ControlValue }, ErrorMsg);
	}

/**	Marks the form controls that supplied the rejected input, inserts an error message immediately before the form,
 *	and sends the page containing the form back to the user for correction. All of the user's input is embedded into the 
 *	form when it is sent back, so the user will not have to re-enter the form contents in order to correct the rejected 
 *	input. The web application blocks when <code>rejectInput</code> is invoked, and unblocks only after potentially 
 * 	valid input is received from the user.
 *	
 *	<p>If the ErrorMsg parameter is any subclass of {@link Markup}, and it contains one or more XML 
 *	elements marked with the class "qwicap-label-text", the contents of those elements will be replaced with the labels 
 *	of the form elements that supplied the rejected parameters. By the same token, the contents of elements marked with 
 *	the class "qwicap-param-value" will be replaced with the values that are being rejected. For example, if a 
 *	form element labelled "Wombat Burrow" was rejected, and the ErrorMsg markup was as follows:
 *	<br>
 *	<br><samp>The input &lt;span class='qwicap-param-value'&gt;&lt;/span&gt is not a valid value for the field 
 *		&lt;span class='qwicap-label-text'&gt;&lt;/span&gt;. Please fix it.</samp>
 * 	<br>
 *	<br>...then the error message that appeared above the form which supplied the parameter would end-up reading:
 *	<br>
 *	<br><samp>The input "42" is not a valid value for the field Wombat Burrow. Please fix it.</samp>
 *	</p>
 *	
 *	<p>If the ErrorMsg parameter is of any type other than <code>Markup</code>, the rules documented in 
 * 	{@link MutableMarkup#insert MutableMarkup.insert} will be applied. (Strictly speaking, those rules always apply, 
 * 	even when the message is a subclass of {@link Markup}.)
 *	</p>
 * 
 * @param ControlNames				Names of the controls whose input is being rejected.
 * @param ControlValues				The input values that are being rejected, in the same order as the names in the 
 *									<code>ControlNames</code> array. This array must be the same length as the 
 *									<code>ControlNames</code> array, or <code>null</code>. Any element of the array may
 * 									also be <code>null</code>. Where <code>null</code> is encountered, the input from 
 *									all form controls of the specified name is considered rejected, regardless of the
 *									particular values.
 * @param ErrorMsg					Error message to be placed above the form. This parameter can be any class of 
 *									object. See {@link MutableMarkup#insert MutableMarkup.insert} for the specific rules 
 * 									employed in the handling of various classes.
 * @throws QwicapException			Always thrown by this method in order to restart the caller's input processing 
 * 									logic. (The Qwicap "prompt pattern" ensures this behavior on the part of the 
 * 									client.)
 */
	public void rejectInput
	(
		final String[]			ControlNames,
		Object[]				ControlValues,	//	Must not be null, but all elements may be null. Must be same length as ControlNames.
		final Object			ErrorMsg
	)
	throws QwicapException
	{
		if (PleaseDie)
			throwSessionKillingExceptions();

		final boolean			ControlValuesWasNull = ControlValues == null;

	//	ControlValues can't be allowed to remain null, but it's fine for it to be an array of nulls.

		if (ControlValuesWasNull)
			ControlValues = new Object[ControlNames.length];

	//	Convert the array of control names to a decently formatted string.

		final ArrayToString		ControlNameList = new ArrayToString(ControlNames);

	//	Log this event.

		Log.log(Level.FINE, "PARAMS: Rejecting form control(s): {0}", ControlNameList);

	//	Mark all of the implicated form elements as bad, while accumulating information about them, like their labels.

		final FormInputRejectionHelper	RPH = new FormInputRejectionHelper(Pages.peek(), ControlNames, ControlValues);

	//	Create a little markup to encapsulate the supplied error message.

		final MutableMarkup		ErrorMarkup = RejectedInputErrorMarkup.getMutable();

		ErrorMarkup.get("div").setContent(ErrorMsg);

		if (RPH.getLabelCount() > 0) {

			for (final Match LblTxt : ErrorMarkup.get("*.qwicap-label-text")) {
				final boolean			UseQuotes = !LblTxt.hasClass("qwicap-no-quotes");

				LblTxt.setContent(RPH.getLabelsAsString(UseQuotes));
			}
		}

		{
			for (final Match ValTxt : ErrorMarkup.get("*.qwicap-param-value")) {
				final boolean			UseQuotes = !ValTxt.hasClass("qwicap-no-quotes");
				final ArrayToString		Value     = ControlValuesWasNull ? null : arrayToStringAnd(ControlValues);

				if (Value != null) {

					if (UseQuotes)
						Value.alwaysQuote();
					else
						Value.neverQuote();
				}

				ValTxt.setContent(Value);
			}
		}

	//	Insert the error message into the page's markup immediately prior to the form that sent the bad input.

		RPH.getEarliestEnclosingElement().add(true, ErrorMarkup);

	//	If auto-populate forms is OFF, we still want to populate the current form with the data set that is being 
	//	rejected, so the user can conveniently correct it. So, we extract and retain the original state of the form, 
	//	then embed the rejected data set, send the page the to the user, and later-on we restore the form to its 
	//	original state, so the client code won't find unexepected values in its form markup.

		final Form				OrigForm = Pages.getMostRecentForm();
		final FormDataSet		OrigFormState;

		if (Pages.getFlags().shouldRejectWithFormData() && OrigForm != null) {

			OrigFormState = OrigForm.getDataSet(false);

		//	Insert into the markup of the most-recent form the values most recently submitted by the user, so that, 
		//	when the form is sent back to the client, the user won't have to re-enter all of the form's contents just 
		//	to correct the one erroneous field. If automatic from control value embedding is enabled, this has already 
		//	been done, but we need to be sure it is done in all cases.

			Pages.populateMostRecentForm();

		} else
			OrigFormState = null;

	//	Re-send the current page. 
	// 
	//	The advantage of doing this here, rather than letting prompting occur when the "prompt pattern" causes the 
	//	"prompt" method to be called again (as a result of this method throwing the QwicapInputAbandonedException)
	//	is that the caller doesn't have the chance to alter the page markup between when we reject the input,
	//	and when we prompt the user for a correction. If we let "prompt" do the prompting, any code within the
	//	page processing loop and ahead of the "prompt" call could tamper with the markup. This could cause confusion
	//	for the user, or waste time, or result in actual problems, since Qwicap is sensitive to some potential changes.

		try {

			promptLow();

		} finally {

		//	If we're meant to be preserving the form's original state, restore that state here.

			if (OrigFormState != null)
				OrigForm.setDataSet(OrigFormState);
		}

	//	Note that we do this even if the user just gave us what will turn-out to be correct values to replace the 
	//	bad ones we were complaining about. We do this because any time that the user has been allowed to submit a form,
	//	all of the input from that form must be re-examined, since the values of (almost) any of the controls could have 
	//	been changed by the user. Given the standard Qwicap "prompt" pattern, this exception will force the input 
	// 	evaluation already in progress to terminate, and will send us on a fresh pass through the "prompt" pattern 
	//	code. The "prompt" won't actually happen, since we want the client code to process the form data set 
	//	that we received by calling promptLow above, rather retransmitting the page to the client and acquiring yet 
	//	another form data set, but the client code doesn't need to know that. It just calls prompt, and is oblivious to 
	//	the fact that nothing happens, since the operation of prompt is opaque to it anyway.
	//
	//	There is another way to organize this: rejectInput does everything except the promptLow call, and signals
	//	prompt to send the page into which we just embedded the error message, parameters, etc.

		throw new QwicapInputAbandonedException(Pages.peek(), ControlNameList.toString());
	}

/**	Inserts an error message immediately before the form whose input was most recently submitted, and sends the page 
 *	containing that form back to the user for correction. All of the user's input is embedded into the form when it is 
 * 	sent back, so the user will not have to re-enter the form contents in order to correct the rejected input. The web 
 * 	application blocks when <code>rejectInput</code> is invoked, and unblocks only after potentially valid input is 
 * 	received from the user.
 *	
 *	<p>If the <code>ErrorMsg</code> parameter is any subclass of {@link Markup}, it is inserted as-is. If the 
 * 	<code>ErrorMsg</code> parameter is not an instance of {@link Markup}, the rules documented in 
 * 	{@link MutableMarkup#insert MutableMarkup.insert} will be applied. (Strictly speaking, those rules always apply, 
 * 	even when the message is a subclass of {@link Markup}.)
 *	</p>
 * 
 * @param ErrorMsg					Error message to be placed above the form. This parameter can be any class of 
 *									object. See {@link MutableMarkup#insert MutableMarkup.insert} for the specific rules 
 * 									employed in the handling of various classes.
 * @throws QwicapException			Always thrown by this method in order to restart the caller's input processing 
 * 									logic. (The Qwicap "prompt pattern" ensures this behavior on the part of the 
 * 									client.)
 */
	public void rejectInput
	(
		final Object			ErrorMsg
	)
	throws QwicapException
	{
		if (PleaseDie)
			throwSessionKillingExceptions();

	//	Log this event.

		Log.log(Level.FINE, "PARAMS: Rejecting all form input.");

	//	Create a little markup to encapsulate the supplied error message.

		final MutableMarkup		ErrorMarkup = RejectedInputErrorMarkup.getMutable();

		ErrorMarkup.get("div").setContent(ErrorMsg);

	//	Insert the error message into the page's markup immediately prior to the form that sent the bad input.

		final Form				OrigForm = Pages.getMostRecentForm();

		if (OrigForm != null) {
			final Results			Ancestors = OrigForm.getMatch().getAncestors(true, null);

			for (int Index = Ancestors.size(); --Index >= 0; ) {
				final Match				Ancestor = Ancestors.get(Index);
				final String			Type     = Ancestor.getType();

				if ("fieldset".equals(Type) || "form".equals(Type)) {

					Ancestor.add(true, ErrorMarkup);
					break;
				}
			}

		} else {
			final MutableMarkup		Page = Pages.getMarkup();

			if (Page != null)
				Page.get("body").addToStartOfContent(ErrorMarkup);
		}

	//	If auto-populate forms is OFF, we still want to populate the current form with the data set that is being 
	//	rejected, so the user can conveniently correct it. So, we extract and retain the original state of the form, 
	//	then embed the rejected data set, send the page the to the user, and later-on we restore the form to its 
	//	original state, so the client code won't find unexepected values in its form markup.

		final FormDataSet		OrigFormState;

		if (Pages.getAutoPopulateMostRecentForm() == false && OrigForm != null)
			OrigFormState = OrigForm.getDataSet(false);
		else
			OrigFormState = null;

	//	Insert into the markup of the most-recently used "form" element, the values most recently submitted by the user, 
	//	so when the "form" is sent back to the user they won't have to re-enter all of its contents just to correct the 
	//	erroneous field. If automatic from control value embedding is enabled, this has already been done, but we need 
	//	to be sure it is done in all cases.

		Pages.populateMostRecentForm();

	//	Resend the current page. 
	// 
	//	The advantage of doing this here, rather than letting prompting occur when the "prompt pattern" causes the 
	//	"prompt" method to be called again (as a result of this method throwing the QwicapInputAbandonedException)
	//	is that the caller doesn't have the chance to alter the page markup between when we reject the input,
	//	and when we prompt the user for a correction. If we let "prompt" do the prompting, any code within the
	//	page processing loop and ahead of the "prompt" call could tamper with the markup. This could cause confusion
	//	for the user, or waste time, or result in actual problems.

		try {

			promptLow();

		} finally {

		//	If we're meant to be preserving the form's original state, restore that state here.

			if (OrigFormState != null)
				OrigForm.setDataSet(OrigFormState);
		}

	//	Note that we do this even if the user just gave us what will turn-out to be correct values to replace the 
	//	bad ones we were complaining about. We do this because any time that the user has been allowed to submit a form,
	//	all of the input from that form must be re-examined, since the values of (almost) any of the controls could have 
	//	been changed by the user. Given the standard Qwicap "prompt" pattern, this exception will force the input 
	// 	evaluation already in progress to terminate, and will send us on a fresh pass through the "prompt" pattern 
	//	code. The "prompt" won't actually happen, since we want the client code to process the form data set 
	//	that we received by calling promptLow above, rather retransmitting the page to the client and acquiring yet 
	//	another form data set, but the client code doesn't need to know that. It just calls prompt, and is oblivious to 
	//	the fact that nothing happens, since the operation of prompt is opaque to it anyway.
	//
	//	There is another way to organize this: rejectInput does everything except the promptLow call, and signals
	//	prompt to send the page into which we just embedded the error message, parameters, etc.

		throw new QwicapInputAbandonedException(Pages.peek());
	}

/**	Redirects the client to the specified URL, and blocks until the user returns to this web application.
 * 
 * @param DestinationURL			The URL to which the client should be redirected.
 * @return							A reference to this <code>Qwicap</code> instance, to facilitate method 
 * 									invocation chaining.
 * @throws QwicapException			If anything goes wrong.
 * @see #goodbye(String)
 */
	public Qwicap redirect
	(
		final String			DestinationURL
	)
	throws QwicapException
	{
		if (PleaseDie)
			throwSessionKillingExceptions();

		if (InputFilterIsNew)
			invokeInputFilter(getHitContext().getFormDataSet());

	//	Send the redirect to the client.

		try {
			getHitContext().getResponse().sendRedirect(DestinationURL);
		} catch (IOException e) {
			throw new QwicapException(e);
		}

		final FormDataSet		DataSet;

	//	Synchronize and sleep until the user returns to this web application. There doesn't have to be any input;
	//	an HTTP "GET" to retrieve the page is an equally good indication that the user has returned.

		DataSet = synchronize(new RedirectDescription(this, DestinationURL));

	//	A hit has been received. Do any handling of the input that may be necessary before returning to the caller.

		if (DataSet.getProcessingException() != null)
			reportInputProcessingFailure(DataSet);

	//	If there was any input at all...

		if (DataSet.notEmpty()) {

		//	Update the form data set associated with the current page. Note that the PageStack.setFormDataSet 
		//	method throws a QwicapPageAbandonedException, or QwicapFormAbandonedException, if the current form data 
		//	set didn't originate from the current page and/or form.

			if (Pages.setFormDataSet(this, DataSet) == false)
				QThread.badDataSet();
		}

		return this;
	}

	private void promptLow() throws QwicapException {
		promptLow(Pages.updatePage());
	}

	private void promptLow
	(
		final Page				CurPage
	)
	throws QwicapException
	{
		if (InputFilterIsNew)
			invokeInputFilter(getHitContext().getFormDataSet());

	//	If we're correcting for hideous bugs in Internet Explorer, convert all the "button" elements of type "submit"
	//	into "input" elements of type "submit", if the browser's identification string matches one of the regular 
	//	expressions for which we've been told to watch.

		final boolean			DidButtonConversions = doButtonConversions(CurPage);

		try {

		//	Add any new downloadables to the current Page object.

			if (NewDownloadables != null && NewDownloadables.size() != 0) {

				Pages.addDownloadables(NewDownloadables);
				NewDownloadables.clear();
			}

		//	Prompt using the current page.

			promptLowest();

		} finally {

		//	Reverse the button conversions we performed above, so that the method that called "prompt" can't be affected
		//	by our conversions. (Well, it can be affected, but only if the method was holding references to Match objects
		//	that referred to the "button" elements and/or their content. This would be a very unusual way of working,
		//	and it would be straightforward to recode the client method to eliminate that practice.)

			if (DidButtonConversions)
				CurPage.undoButtonConversions();
		}
	}

//	promptLowest should only be invoked from promptLow, otherwise there's a chance that important markup house-keeping
//	won't take place.

	private void promptLowest() throws QwicapException {

		if (PleaseDie)
			throwSessionKillingExceptions();

		Page					CurPage = Pages.peek();

	//	Place a message in the page describing any exceptions that were reported to Qwicap by the application, 
	//	using the reportException method.

		insertIncidentReport(CurPage);

	//	If Qwicap features that require embedded scripts have been enabled, ensure that those scripts are embedded in
	//	in the page.

		AJAX.insertScripts(CurPage);

	//	Wait until a good data set arrives.

		boolean					DataSetIsGood;

		do {
			FormDataSet				DataSet;

			do {

			//	Send the supplied page to the client. We test whether or not the response has been committed because
			//	an input filter may have already generated a response of its own (think asynchronous AJAX-type 
			//	responses).

				if (getHitContext().getResponse().isCommitted() == false) {

					CurPage.write(this);
					CurPage.removeNavigationErrors();	//	todo Consider: Are there other errors that might be considered "short term"-enough to merit removal at this point? Maybe most errors that don't involve bad user input.
				}

			//	Block until input arrives.

				DataSet = synchronize(new PromptDescription(this, CurPage.getMarkup()));

			//	Input has arrived.

				if (DataSet.getProcessingException() != null)
					reportInputProcessingFailure(DataSet);

			//	Loop if we've received no form data set from the client, or if the form data set had such fundamental 
			//	problems that it couldn't be loaded (this only occurs in the context of file-uploads). In the case of 
			//	receiving no form data set, this presumably means that the client has requested a reload of the page.

			} while (DataSet.isEmpty());

		//	Loop until we get a good form data set, then update the form data set associated with the current page.
		//	Note that the PageStack.setFormDataSet method throws a QwicapPageAbandonedException, or
		//	QwicapFormAbandonedException, if the current form data set didn't originate from the current page and/or
		//	form.

			DataSetIsGood = Pages.setFormDataSet(this, DataSet);
			if (!DataSetIsGood) {

				if (DataSet.has(kQwicapNoOpControlName))
					CurPage = Pages.peek();
				else
					QThread.badDataSet();
			}

		} while (!DataSetIsGood);
	}

	private FormDataSet synchronize
	(
		final BlockingDescription	BlockInfo
	)
	{
		Log.finest("Qwicap.synchronize entered.");

		try {

		//	Invoke all of the blocking listeners, and retain any state objects that they may choose to return.
			
			final QwicapBlockingListenerStates	Listeners = QwicapBlockingListenerStates.blocked(BlockingListeners, BlockInfo);
			
		//	Complete processing of this hit, and wait for the next.

			try {
				FormDataSet				NewDataSet;
				boolean					WaitForNewDataSet;
				int						DuplicateDataSetCount = 0;

				do {

					synchronized (this) {

					//	Notify our HitContext object (HitCtxt) that we've finished processing this hit. (We must have
					// 	finished, because we're about to begin waiting for the next hit.) This tells the SessionContext
					//	that it can cease waiting (which it is currently doing in its "go" method).

						HitCtxt.processingComplete();
						HitCtxt = null;

					//	Wait for a new hit to arrive.

						do {
							
						//	If we've been told to die, we're going to exit this loop, and force the exit of lots of 
						// 	enclosing logic, by throwing the QwicapSessionDeathException, which is reserved for the 
						//	purpose of kiliing sessions.
						//
						//	Note that this is a particularly good place at which to trigger session death, because the
						//	HitCtxt.processingComplete invocation above has just released the blocked servlet thread
						//	to complete its work on the most recent hit. 
							
							if (PleaseDie)
								throwSessionKillingException();

						//	Now, wait for our SessionContent object (SessCtxt) to notify us (from its "go" method) when 
						//	the next hit is received. We'll retrieve the information about the hit (HitContext) from 
						//	the session's CurrentHit field.

							try {

								while (ShouldWait)
									wait();

							//	Set ShouldWait to true in preparation for the NEXT invocation of this method.

								ShouldWait = true;

							} catch (InterruptedException e) {
							//	We proceed the same way whether an InterruptedException was thrown, or not.
							}

						} while (HitCtxt == null);
					}

				//	Retrieve the new data set we just received from the user.

					NewDataSet        = getHitContext().getFormDataSet();
					WaitForNewDataSet = false;

				//	Determine whether or not it's a data set that should be processed by the client web application.

					if (AJAX.processCommand(NewDataSet)) {

						NewDataSet        = null;
						WaitForNewDataSet = true;

					} else if (NewDataSet.isAccidentalDuplicateOf(PrevDataSet)) {

					//	This situation usually arises when a user accidentally clicks the same submit button two or more 
					//	times before the web application can send back a new page. So, by ignoring duplicates, we're 
					//	effectively performing click de-bouncing, which is analogous to the long-vanished problem of 
					//	keyboard de-bouncing.

						PrevDataSet       = NewDataSet;
						NewDataSet        = null;
						WaitForNewDataSet = true;

						Log.log(Level.FINE, "Duplicate data set no. {0} detected and ignored. Current page will be retransmitted to client.", ++DuplicateDataSetCount);

//						try {
							Pages.peek().write(this);
//						} catch (QwicapException e) {
//							Log.log(Level.SEVERE, "Failed to retransmit current page to client, following detection of duplicate data set.", e);
//						}
					}

				//	NOTE: It might seem natural to test for empty data sets here, and to retransmit the current page
				//	      when one in detected. However, it is possible to return from redirects with an empty data set,
				//	      and if we handle the empty data set here, the Qwicap.redirect method will never get the 
				//	      opportunity to un-block. Also, there's already page retransmission logic in the promptLowest
				//	      method.

				//	Loop until we receive a data set that should be processed by the client web application.

				} while (WaitForNewDataSet);

				PrevDataSet = NewDataSet;

				return NewDataSet;

			} finally {

			//	Invoke all of the blocking listeners, returning to them any state objects that they may have previously
			//	provided.			

				Listeners.unblocked();
			}

		} finally {

			Log.finest("Qwicap.synchronize exited.");
		}
	}

	private void reportInputProcessingFailure
	(
		final FormDataSet		DataSet
	)
	throws TagException
	{
		final Exception			X            = DataSet.getProcessingException();
		final MutableMarkup		CurPage      = Pages.peek().getMarkup();
		Results					ErrorMsgDest = CurPage.get("body");
		boolean					InsertInside = true;	//	Otherwise, we insert above the destination.
		final XMLString			ErrorMsg;

		if (X != null) {

			if (X instanceof FileUploadException) {

				if (X instanceof FileUploadBase.InvalidContentTypeException) {
					
					ErrorMsg = i18n.getXML("getUploadedFile-UnknownTypeInvalid");

				} else if (X instanceof FileUploadBase.SizeLimitExceededException || X instanceof FileUploadBase.FileSizeLimitExceededException) {

					ErrorMsg = i18n.getMessageXML("getUploadedFile-UploadTooLarge", UFAbsoluteSizeLimit);

				} else
					ErrorMsg = i18n.getXML("getUploadedFile-UnknownProblem");

			//	Get a reference to the form tag that contains this input.

				final Results			FileInputs = CurPage.get("form input[type='file']");

				if (FileInputs.size() == 1) {
					Results					Parent = FileInputs.getParent();

					while (Parent.isEmpty() == false && "form".equals(Parent.getType()) == false)
						Parent = Parent.getParent();

					if ("form".equals(Parent.getType())) {

						ErrorMsgDest = Parent;
						InsertInside = false;	//	Insert immediately above the destination, instead.
					}
				}

			} else
				ErrorMsg = i18n.getXML("InputProcessing-UnknownProblem");
			
		//	todo This exception has already been logged by the FormDataSet class. Nonethless, would it be appropriate to
		//	     handle this exception in a manner similar to that of Qwicap.reportException - i.e. by placing the 
		//	     exception's stack trace into a comment block in the page, if (and only if) Qwicap is operating in its 
		//	     low-security "development" mode? In low-security mode we could also include the exception's message, if
		//	     any, in the user-visible error message.

		} else
			ErrorMsg = i18n.getXML("InputProcessing-UnknownProblem");

	//	Delete any input processing failure message that might already be in the page.

		CurPage.get("div." + kQwicapInputProcessingFailureClass).delete();

	//	Create a little markup to encapsulate the supplied error message.

		final MutableMarkup		ErrorMarkup = InputProcessingErrorMarkup.getMutable();

		ErrorMarkup.get("div").setContent(ErrorMsg);

	//	Insert the error message markup into the page.

		if (InsertInside)
			ErrorMsgDest.addToStartOfContent(ErrorMarkup);
		else
			ErrorMsgDest.addBefore(ErrorMarkup);
	}

/**	Returns a {@link Form} object which refers to the form that supplied the form data set most recently submitted to 
 *	the current web page.
 * 
 * @return							A {@link Form} object referencing the form that supplied the current form data set, 
 * 									or <code>null</code>, if no such form can be identified.
 * @throws FormNotFoundException	If the <samp>form</samp> element has disappeared from the page markup.
 */
	public Form getForm() throws FormNotFoundException {
		return Pages.getMostRecentForm();
	}

/**	Returns the form data set most recently submitted by the current web page.
 * 
 * @return		A <code>FormDataSet</code> object containing all of the input most recently submitted to the current 
 *				web page, or <code>null</code>, if no input is available.
 */
	public FormDataSet getFormDataSet() {

	//	todo If we're going to consider PleaseDie here, consider where else we should also consider it.

		if (PleaseDie)
			throwSessionKillingException();

		return Pages.getFormDataSet();
	}

/**	Returns <code>true</code> if a form data set is available, <code>false</code> otherwise.
 * 
 * @return		<code>true</code> if a form data set is available, <code>false</code> otherwise.
 */
	public boolean hasFormDataSet() {
		return getFormDataSet() != null;
	}

/**	Returns <code>true</code> if the specified form control supplied one or more values in the current form data set, 
 *	and <code>false</code> if it does not.
 * 
 * @param ControlName	Name of the form control.
 * @return				<code>true</code> if the control supplied one or more values in the current form data set, 
 *						<code>false</code> otherwise.
 */
	public boolean has
	(
		final String			ControlName
	)
	{
		return Pages.getFormControlValue(ControlName) != null;
	}

/**	Returns the value of the specified form control as a <code>String</code>. Since this can only return one value, 
 *	this method should not be used to retrieve the value of mutliple-selection form elements, for instance: groups of 
 *	checkboxes, or "select"s whose "multiple" attribute is set.
 * 
 * @param ControlName	Name of the form control whose value is to be returned.
 * @return				The value supplied by the specified form control, or <code>null</code>, if it supplied no value.
 * @see #getAll(String)
 * @see #rejectInput(String, Object, Object)
 */
	public String get
	(
		final String			ControlName
	)
	{
		final FormControlValue		P = Pages.getFormControlValue(ControlName);

		if (P != null)
			return P.getValue();
		else
			return null;
	}

/**	Returns the values of the specified form control as an array of <code>String</code>s.
 * 
 * @param ControlName	Name of the form control whose values are to be returned.
 * @return				The values supplied by the specified form control as an array of <code>String</code> objects. 
 * 						If there are no values associated with that control, a zero-length array is returned.
 * @see #rejectInput(String, Object, Object)
 */
	public String[] getAll
	(
		final String			ControlName
	)
	{
		final FormControlValue	P = Pages.getFormControlValue(ControlName);

		if (P != null)
			return P.getValues();
		else
			return kEmptyStringArray;
	}

/**	Specifies all the settings that affect the HTTP file upload functionality. By default, MaxSize and MaxSizeInRAM are 
 *	set to the same value, because this makes RepositoryPath irrelevant. That's useful because there's no single value 
 *	for RepositoryPath that will work across all platforms.
 * 
 * @param MaxSizeInRAM			The maximum size for uploaded files that are to be stored in memory; larger files will 
 *								be stored on disk.
 * @param MaxSize				The maximum acceptable size of an uploaded file. Larger files will be rejected.
 * @param RepositoryPath		The directory in which uploaded files are to be stored when they are too large to be 
 *								stored in memory. This parameter can be <code>null</code> when MaxSizeInRAM equals 
 *								MaxSize, since all uploads will be stored in memory.
 * @return						A reference to this <code>Qwicap</code> instance, to facilitate method invocation
 * 								chaining.
 * @see							FileItem#delete
 */
	public Qwicap setUploadedFileLimits
	(
		int 					MaxSizeInRAM,
		final long				MaxSize,
		final String			RepositoryPath	//	Can be null, sometimes.
	)
	{
		if (MaxSizeInRAM > MaxSize)
			MaxSizeInRAM = (int) MaxSize;		//	Unnecessary, but tidy.
		else if (MaxSizeInRAM < MaxSize && RepositoryPath == null)
			throw new IllegalArgumentException("The repository path cannot be null when allowing uploads larger than those that will be stored in memory (" + MaxSizeInRAM + " < " + MaxSize + " bytes).");

		UFInMemorySizeLimit = MaxSizeInRAM;
		UFAbsoluteSizeLimit = MaxSize;
		UFRepositoryPath    = RepositoryPath;

		return this;
	}

/**	Returns the maximum size of an uploaded file that will be memory-resident; larger files will be stored on disk (see
 *	{@link #getUploadedFileRepository}), while files beyond a certain size won't be accepted at all (see 
 *	{@link #getUploadedFileSizeLimit}).
 * 
 * @return				The maximum size of uploaded files that will be stored in memory.
 * @see 				#setUploadedFileLimits(int, long, String)
 */
	public int getUploadedFileRAMLimit() {
		return UFInMemorySizeLimit;
	}

/**	Returns the maximum acceptable size of an uploaded file. Larger files are rejected.
 * 
 * @return				The maximum acceptable size of an uploaded file.
 * @see 				#setUploadedFileLimits(int, long, String)
 */
	public long getUploadedFileSizeLimit() {
		return UFAbsoluteSizeLimit;
	}

/**	Returns the path of the directory in which uploaded files are stored when they are too large to be stored in 
 * 	memory; see {@link #getUploadedFileRAMLimit}.
 *	
 *	@return				Path of the directory which is the repository for uploaded files.
 *	@see 				#setUploadedFileLimits(int, long, String)
 *	@see				FileItem#delete
 */
	public String getUploadedFileRepository() {
		return UFRepositoryPath;
	}

/**	Returns a {@link FileItem} object describing the uploaded file associated with a specified control name. If the 
 *	file is absent, and the <code>AllowedToBeMissing</code> parameter is <code>false</code>, the XHTML page is sent 
 *	back to the user with a message describing the problem, and this method blocks until the user eventually uploads a 
 *	file.
 * 
 * @param ControlName			Name of the form control associated with the uploaded file.
 * @param ContentType			The required <a href='http://www.iana.org/assignments/media-types/'>MIME content type</a> 
 *								for the uploaded file, or <code>null</code> if all types are acceptable. If the uploaded 
 *								file is not of the required type, the file is deleted, and the user is prompted to 
 *								supply a file of the required type.
 * @param AllowedToBeMissing	<code>true</code>, if it's OK to return <code>null</code> when no file has been 
 *								uploaded. If <code>false</code>, and no file has been uploaded, the user is 
 *								automatically requested to upload a file, and this method blocks until a file is 
 *								uploaded.
 * @return						A FileItem object representing the uploaded file.
 * @throws QwicapException		If anything goes wrong.
 * @see 						#setUploadedFileLimits(int, long, String)
 * @see							FileItem#delete
*/
	public FileItem getUploadedFile
	(
		final String			ControlName,
		final String			ContentType,		//	Can be null.
		final boolean			AllowedToBeMissing
	)
	throws QwicapException
	{
		if (ContentType == null)
			return getUploadedFile(ControlName, (String[]) null, AllowedToBeMissing);
		else
			return getUploadedFile(ControlName, new String[] { ContentType }, AllowedToBeMissing);
	}

/**	Returns a {@link FileItem} object describing the uploaded file associated with a specified control name. If the 
 *	file is absent, and the <code>AllowedToBeMissing</code> parameter is <code>false</code>, the XHTML page is sent 
 *	back to the user with a message describing the problem, and this method blocks until the user eventually uploads a 
 *	file.
 * 
 * @param ControlName			Name of the form control associated with the uploaded file.
 * @param ContentTypes			A list of <a href='http://www.iana.org/assignments/media-types/'>MIME content types</a> 
 *								that are acceptable, or <code>null</code> if all types are acceptable. If the uploaded 
 *								file is not of an acceptable type, the file is deleted, and the user is prompted to 
 *								supply a file of one of the acceptable types.
 * @param AllowedToBeMissing	<code>true</code>, if it's OK to return <code>null</code> when no file has been 
 *								uploaded. If <code>false</code>, and no file has been uploaded, the user is 
 *								automatically requested to upload a file, and this method blocks until a file is 
 *								uploaded.
 * @return						A FileItem object representing the uploaded file.
 * @throws QwicapException		If anything goes wrong.
 * @see 						#setUploadedFileLimits(int, long, String)
 * @see							FileItem#delete
 */
	public FileItem getUploadedFile
	(
		final String			ControlName,
		final String[]			ContentTypes,		//	Can be null.
		final boolean			AllowedToBeMissing
	)
	throws QwicapException
	{
		if (PleaseDie)
			throwSessionKillingExceptions();

		final FormControlValue		P = Pages.getFormControlValue(ControlName);

		if (P != null) {

			if (P instanceof FormFileUpload == false)
				throw new QwicapException("The input supplied by the control \"" + ControlName + "\" is not an uploaded file.");

			final FileItem			FI          = ((FormFileUpload) P).getFileItem();
			final String			ContentType = FI.getContentType();

		//	Conclude that no file was uploaded if the file's size is zero, and one or both of the following 
		//	conditions are met: (1) the content-type is null, or (2) the file name is null or empty.

			if (FI.getSize() == 0 && (ContentType == null || FI.getName() == null || "".equals(FI.getName()))) {

			//	FileIsMissing = true;

			} else {

			//	We now know that a file really was uploaded; see if it's of a type we're willing to accept.

			//	If ContentTypes is null, any type is acceptable.

				if (ContentTypes == null)
					return FI;

			//	Search the ContentTypes array to determine whether the uploaded file is of an acceptable type.

				for (int Index = 0, Count = ContentTypes.length; Index < Count; Index++) {

					if (ContentTypes[Index].equals(ContentType))
						return FI;	//	The file's type is in the list of valid types, so we're done.
				}

			//	The supplied file must be rejected because of its invalid type. Delete the uploaded file and 
			//	explain the situation to the user.

				FI.delete();

				rejectInput(ControlName, "", i18n.getMessageXML("getUploadedFile-TypeInvalid", HTML.encode(ControlName), HTML.encode(ContentType), HTML.encode(arrayToStringOr(ContentTypes).alwaysQuote().toString())));
			}
		}

		if (AllowedToBeMissing)
			return null;
		else
			rejectInput(ControlName, "", i18n.getMessageXML("getUploadedFile-NoUpload", HTML.encode(ControlName)));

	//	Because rejectInput ALWAYS throws an exception, we will NEVER reach the following return statement.

		return null;
	}

/**	Returns the value of the specified control as a <code>boolean</code> primitive. If the value supplied by the user 
 *	is invalid, absent, or isn't one of the two acceptable values, the XHTML page is sent back to the user with a 
 *	message describing the problem, and this method blocks until the user eventually supplies a valid value.
 * 
 * @param ControlName		Name of the form control whose value is to be returned.
 * @param AllowedToBeBlank	<code>true</code> if the input can be missing or blank.
 * @param DefaultValue		Value to return if <code>AllowedToBeBlank</code> is <code>true</code>, and the input was 
 * 							missing or blank.
 * @return					The value of the form control represented as a <code>boolean</code>.
 * @throws QwicapException	If anything goes wrong.
 */
	public boolean getBoolean
	(
		final String			ControlName,
		final boolean			AllowedToBeBlank,
		final boolean			DefaultValue
	)
	throws QwicapException
	{
		return getBoolean(ControlName, "true", "false", AllowedToBeBlank, DefaultValue);
	}

/**	Returns the value of the specified control as a <code>boolean</code> primitive. If the value supplied by the user 
 *	is invalid, absent, or isn't one of the two acceptable values, the XHTML page is sent back to the user with a 
 *	message describing the problem, and this method blocks until the user eventually supplies a valid value.
 * 
 * @param ControlName		Name of the form control whose value is to be returned.
 * @param TrueStr			Value to be interpreted as <code>true</code>.
 * @param FalseStr			Value to be interpreted as <code>false</code>.
 * @param AllowedToBeBlank	<code>true</code> if the input can be missing or blank.
 * @param DefaultValue		Value to return if <code>AllowedToBeBlank</code> is <code>true</code>, and the input was 
 * 							missing or blank.
 * @return					The value of the form control represented as a <code>boolean</code>.
 * @throws QwicapException	If anything goes wrong.
 */
	public boolean getBoolean
	(
		final String			ControlName,
		final String			TrueStr,
		final String			FalseStr,
		final boolean			AllowedToBeBlank,
		final boolean			DefaultValue
	)
	throws QwicapException
	{
		final String			ValueStr = get(ControlName);
		final boolean			Blank    = ValueStr == null || ValueStr.length() == 0;

		if (Blank && AllowedToBeBlank)
			return DefaultValue;

		if (Blank)
			rejectInput(ControlName, ValueStr, i18n.getMessageXML("getBoolean-ValueBlank", HTML.encode(ControlName), HTML.encode(TrueStr), HTML.encode(FalseStr)));

	//	We already know that ValueStr is not null, but TrueStr and FalseStr might be, 
	//	so we use the ValueStr.equals below.

		if (TrueStr.equals(ValueStr))
			return true;
		else if (FalseStr.equals(ValueStr))
			return false;

	//	The input was the equivalent of neither true, nor false, so we reject the input.

		if (AllowedToBeBlank)
			rejectInput(ControlName, ValueStr, i18n.getMessageXML("getBoolean-ValueNotValidOrBlank", HTML.encode(ControlName), HTML.encode(TrueStr), HTML.encode(FalseStr), HTML.encode(ValueStr)));
		else
			rejectInput(ControlName, ValueStr, i18n.getMessageXML("getBoolean-ValueNotValid",        HTML.encode(ControlName), HTML.encode(TrueStr), HTML.encode(FalseStr), HTML.encode(ValueStr)));

	//	Because rejectInput ALWAYS throws an exception, we will NEVER reach the following return statement.

		return false;
	}

/**	Returns the value of the specified control as an <code>int</code>. If the value supplied by the user is invalid, 
 *	absent, or falls outside the range of acceptable values, the XHTML page is sent back to the user with a message 
 *	describing the problem, and this method blocks until the user eventually supplies a valid value.
 *
 *	<p>If you want to further constrain the acceptable values of this control, evaluate the value returned by this 
 *	method and, if it does not meet your constraints, invoke the {@link #rejectInput(String, Object, Object) rejectInput} 
 * 	method. Repeat as necessary, until you get a value that satisfies your constraints.
 *	</p> 
 *
 * @param ControlName		Name of the form control whose value is to be returned.
 * @param MinValue			Minimum acceptable value of the control (inclusive).
 * @param MaxValue			Maximum acceptable value of the control (inclusive).
 * @param AllowedToBeBlank	<code>true</code> if the input can be missing or blank.
 * @param DefaultValue		Value to return if <code>AllowedToBeBlank</code> is <code>true</code>, and the input was 
 * 							missing or blank. Note that this value need not be between <code>MinValue</code> and 
 * 							<code>MaxValue</code>. Also, if <code>AllowedToBeBlank</code> is <code>false</code>, 
 * 							you may pass <code>null</code> for this parameter.
 * @return					The value of the form control represented as an <code>int</code>.
 * @throws QwicapException	If anything goes wrong.
 */
	public int getInt
	(
		final String			ControlName,
		final int				MinValue,
		final int				MaxValue,
		final boolean			AllowedToBeBlank,
		final int				DefaultValue
	)
	throws QwicapException
	{
		return getBigInteger(ControlName, BigInteger.valueOf((long) MinValue), BigInteger.valueOf((long) MaxValue), AllowedToBeBlank, BigInteger.valueOf((long) DefaultValue)).intValue();
	}

/**	Returns the value of the specified control as a <code>long</code>. If the value supplied by the user is invalid, 
 *	absent, or falls outside the range of acceptable values, the XHTML page is sent back to the user with a message 
 *	describing the problem, and this method blocks until the user eventually supplies a valid value.
 *
 *	<p>If you want to further constrain the acceptable values of this control, evaluate the value returned by this 
 *	method and, if it does not meet your constraints, invoke the {@link #rejectInput(String, Object, Object) rejectInput} 
 * 	method. Repeat as necessary, until you get a value that satisfies your constraints.
 *	</p>
 * 
 * @param ControlName		Name of the form control whose value is to be returned.
 * @param MinValue			Minimum acceptable value of the control (inclusive).
 * @param MaxValue			Maximum acceptable value of the control (inclusive).
 * @param AllowedToBeBlank	<code>true</code> if the input can be missing or blank.
 * @param DefaultValue		Value to return if <code>AllowedToBeBlank</code> is <code>true</code>, and the input was 
 * 							missing or blank. Note that this value need not be between <code>MinValue</code> and 
 * 							<code>MaxValue</code>. Also, if <code>AllowedToBeBlank</code> is <code>false</code>, 
 * 							you may pass <code>null</code> for this parameter.
 * @return					The value of the form control represented as a <code>long</code>.
 * @throws QwicapException	If anything goes wrong.
 */
	public long getLong
	(
		final String			ControlName,
		final long				MinValue,
		final long				MaxValue,
		final boolean			AllowedToBeBlank,
		final long				DefaultValue
	)
	throws QwicapException
	{
		return getBigInteger(ControlName, BigInteger.valueOf(MinValue), BigInteger.valueOf(MaxValue), AllowedToBeBlank, BigInteger.valueOf(DefaultValue)).longValue();
	}

/**	Returns the value of the specified control as a <code>BigInteger</code>. If the value supplied by the user is 
 *	invalid, absent, or falls outside the range of acceptable values, the XHTML page is sent back to the user with a 
 *	message describing the problem, and this method blocks until the user eventually supplies a valid value.
 *
 *	<p>If you want to further constrain the acceptable values of this control, evaluate the value returned by this 
 *	method and, if it does not meet your constraints, invoke the {@link #rejectInput(String, Object, Object) rejectInput} 
 * 	method. Repeat as necessary, until you get a value that satisfies your constraints.
 *	</p>
 * 
 * @param ControlName		Name of the form control whose value is to be returned.
 * @param MinValue			Minimum acceptable value of the control (inclusive).
 * @param MaxValue			Maximum acceptable value of the control (inclusive).
 * @param AllowedToBeBlank	<code>true</code> if the input can be missing or blank.
 * @param DefaultValue		Value to return if <code>AllowedToBeBlank</code> is <code>true</code>, and the input was 
 * 							missing or blank. Note that this value need not be between <code>MinValue</code> and 
 * 							<code>MaxValue</code>. Also, if <code>AllowedToBeBlank</code> is <code>false</code>, 
 * 							you may pass <code>null</code> for this parameter.
 * @return					The value of the form control represented as a <code>BigInteger</code>.
 * @throws QwicapException	If anything goes wrong.
 */
	public BigInteger getBigInteger
	(
		final String			ControlName,
		final BigInteger		MinValue,
		final BigInteger		MaxValue,
		final boolean			AllowedToBeBlank,
		final BigInteger		DefaultValue
	)
	throws QwicapException
	{
		final String			ValueStr = get(ControlName);
		final boolean			Blank    = ValueStr == null || ValueStr.length() == 0;

		if (Blank && AllowedToBeBlank)
			return DefaultValue;

		if (!Blank) {
			XMLString				ErrorMsg;

			try {
				final BigInteger		Value = i18n.parseInteger(ValueStr);

				if (Value.compareTo(MinValue) >= 0 && Value.compareTo(MaxValue) <= 0)
					return Value;

				if (Value.compareTo(MinValue) < 0)
					ErrorMsg = i18n.getMessageXML("getInteger-ValueBelowRange", HTML.encode(ControlName), Value, MinValue, MaxValue);
				else
					ErrorMsg = i18n.getMessageXML("getInteger-ValueAboveRange", HTML.encode(ControlName), Value, MinValue, MaxValue);

			} catch (ArithmeticException e) {

				ErrorMsg = i18n.getMessageXML("getInteger-ValueNotAnInteger", HTML.encode(ControlName), HTML.encode(ValueStr));

			} catch (Exception e) {

				ErrorMsg = i18n.getMessageXML("getNumber-ValueNotANumber", HTML.encode(ControlName), HTML.encode(ValueStr));
			}

			rejectInput(ControlName, ValueStr, ErrorMsg);

		} else
			rejectInput(ControlName, "", i18n.getMessageXML("getNumber-ValueBlank", HTML.encode(ControlName)));

	//	Because rejectInput ALWAYS throws an exception, we will NEVER reach the following return statement.

		return null;
	}

/**	Returns the value of the specified control as a <code>float</code>. If the value supplied by the user is invalid, 
 *	absent, or falls outside the range of acceptable values, the XHTML page is sent back to the user with a message 
 *	describing the problem, and this method blocks until the user eventually supplies a valid value.
 *
 *	<p>If you want to further constrain the acceptable values of this control, evaluate the value returned by this 
 *	method and, if it does not meet your constraints, invoke the {@link #rejectInput(String, Object, Object) rejectInput} 
 * 	method. Repeat as necessary, until you get a value that satisfies your constraints.
 *	</p>
 * 
 * @param ControlName		Name of the form control whose value is to be returned.
 * @param MinValue			Minimum acceptable value of the control (inclusive).
 * @param MaxValue			Maximum acceptable value of the control (inclusive).
 * @param AllowedToBeBlank	<code>true</code> if the input can be missing or blank.
 * @param DefaultValue		Value to return if <code>AllowedToBeBlank</code> is <code>true</code>, and the input was 
 * 							missing or blank. Note that this value need not be between <code>MinValue</code> and 
 * 							<code>MaxValue</code>. Also, if <code>AllowedToBeBlank</code> is <code>false</code>, 
 * 							you may pass <code>null</code> for this parameter.
 * @return					The value of the form control represented as a <code>float</code>.
 * @throws QwicapException	If anything goes wrong.
 */
	 public float getFloat
	(
		final String			ControlName,
		final float				MinValue,
		final float				MaxValue,
		final boolean			AllowedToBeBlank,
		final float				DefaultValue
	)
	throws QwicapException
	{
	//	Note that floating-point precision issues apply here.
	//
	//	For instance specifying a MinValue of 0.1 results in a BigDecimal whose value is "0.100000001490116119384765625",
	//	which results in a value like 0.1 not being within the range it defines(!).
		
		final BigDecimal		CleanMinValue = new BigDecimal(Float.toString(MinValue));
		final BigDecimal		CleanMaxValue = new BigDecimal(Float.toString(MaxValue));
		
		return getBigDecimal(ControlName, CleanMinValue, CleanMaxValue, AllowedToBeBlank, new BigDecimal(DefaultValue)).floatValue();
	}
	
/**	Returns the value of the specified control as a <code>double</code>. If the value supplied by the user is invalid, 
 *	absent, or falls outside the range of acceptable values, the XHTML page is sent back to the user with a message 
 *	describing the problem, and this method blocks until the user eventually supplies a valid value.
 *
 *	<p>If you want to further constrain the acceptable values of this control, evaluate the value returned by this 
 *	method and, if it does not meet your constraints, invoke the {@link #rejectInput(String, Object, Object) rejectInput} 
 * 	method. Repeat as necessary, until you get a value that satisfies your constraints.
 *	</p>
 * 
 * @param ControlName		Name of the form control whose value is to be returned.
 * @param MinValue			Minimum acceptable value of the control (inclusive).
 * @param MaxValue			Maximum acceptable value of the control (inclusive).
 * @param AllowedToBeBlank	<code>true</code> if the input can be missing or blank.
 * @param DefaultValue		Value to return if <code>AllowedToBeBlank</code> is <code>true</code>, and the input was 
 * 							missing or blank. Note that this value need not be between <code>MinValue</code> and 
 * 							<code>MaxValue</code>. Also, if <code>AllowedToBeBlank</code> is <code>false</code>, 
 * 							you may pass <code>null</code> for this parameter.
 * @return					The value of the form control represented as a <code>double</code>.
 * @throws QwicapException	If anything goes wrong.
 */
	public double getDouble
	(
		final String			ControlName,
		final double			MinValue,
		final double			MaxValue,
		final boolean			AllowedToBeBlank,
		final double			DefaultValue
	)
	throws QwicapException
	{
	//	Note that floating-point precision issues apply here.
	//
	//	For instance specifying a MinValue of 0.1 results in a BigDecimal whose value is 
	//	"0.1000000000000000055511151231257827021181583404541", which results in a value like 0.1 not being within the 
	//	range it defines(!).
		
		final BigDecimal		CleanMinValue = new BigDecimal(Double.toString(MinValue));
		final BigDecimal		CleanMaxValue = new BigDecimal(Double.toString(MaxValue));
		
		return getBigDecimal(ControlName, CleanMinValue, CleanMaxValue, AllowedToBeBlank, new BigDecimal(DefaultValue)).doubleValue();
	}

/**	Returns the value of the specified control as a <code>BigDecimal</code>. If the value supplied by the user is 
 *	invalid, absent, or falls outside the range of acceptable values, the XHTML page is sent back to the user with a 
 *	message describing the problem, and this method blocks until the user eventually supplies a valid value.
 *
 *	<p>If you want to further constrain the acceptable values of this control, evaluate the value returned by this 
 *	method and, if it does not meet your constraints, invoke the {@link #rejectInput(String, Object, Object) rejectInput} 
 * 	method. Repeat as necessary, until you get a value that satisfies your constraints.
 *	</p>
 * 
 * @param ControlName		Name of the form control whose value is to be returned.
 * @param MinValue			Minimum acceptable value of the control (inclusive).
 * @param MaxValue			Maximum acceptable value of the control (inclusive).
 * @param AllowedToBeBlank	<code>true</code> if the input can be missing or blank.
 * @param DefaultValue		Value to return if <code>AllowedToBeBlank</code> is <code>true</code>, and the input was 
 * 							missing or blank. Note that this value need not be between <code>MinValue</code> and 
 * 							<code>MaxValue</code>. Also, if <code>AllowedToBeBlank</code> is <code>false</code>, 
 * 							you may pass <code>null</code> for this parameter.
 * @return					The value of the form control represented as a <code>BigDecimal</code>.
 * @throws QwicapException	If anything goes wrong.
 */
	public BigDecimal getBigDecimal
	(
		final String			ControlName,
		final BigDecimal		MinValue,
		final BigDecimal		MaxValue,
		final boolean			AllowedToBeBlank,
		final BigDecimal		DefaultValue
	)
	throws QwicapException
	{
		final String			ValueStr = get(ControlName);
		final boolean			Blank    = ValueStr == null || ValueStr.length() == 0;

		if (Blank && AllowedToBeBlank)
			return DefaultValue;

		if (!Blank) {
			XMLString				ErrorMsg;

			try {
				final BigDecimal		Value = i18n.parseDecimal(ValueStr);

				if (Value.compareTo(MinValue) >= 0 && Value.compareTo(MaxValue) <= 0)
					return Value;

				if (Value.compareTo(MinValue) < 0)
					ErrorMsg = i18n.getMessageXML("getFloatingPoint-ValueBelowRange", HTML.encode(ControlName), Value, MinValue, MaxValue);
				else
					ErrorMsg = i18n.getMessageXML("getFloatingPoint-ValueAboveRange", HTML.encode(ControlName), Value, MinValue, MaxValue);

			} catch (Exception e) {

				ErrorMsg = i18n.getMessageXML("getNumber-ValueNotANumber", HTML.encode(ControlName), HTML.encode(ValueStr));
			}

			rejectInput(ControlName, ValueStr, ErrorMsg);

		} else
			rejectInput(ControlName, "", i18n.getMessageXML("getNumber-ValueBlank", HTML.encode(ControlName)));

	//	Because rejectInput ALWAYS throws an exception, we will NEVER reach the following return statement.

		return null;
	}

/**	Returns the values of the specified controls as <code>int</code> primitives. If any of the values supplied by 
 * 	the user are invalid, absent, or fall outside the range of acceptable values, the XHTML page is sent back to the 
 * 	user with a message describing the problem, and this method blocks until the user eventually supplies valid values.
 *
 *	<p>If you want to further constrain the acceptable values of these controls, evaluate the values returned by this 
 *	method and, if they do not meet your constraints, invoke the {@link #rejectInput(String, Object, Object) rejectInput} 
 * 	method. Repeat as necessary, until you get a value that satisfies your constraints.
 *	</p>
 * 
 * @param ControlName		Name of the form control whose value is to be returned.
 * @param MinValue			Minimum acceptable value of the control (inclusive).
 * @param MaxValue			Maximum acceptable value of the control (inclusive).
 * @param AllowedToBeBlank	<code>true</code> if the input can be missing or blank.
 * @param DefaultValues		Values to return if <code>AllowedToBeBlank</code> is <code>true</code>, and the inputs were 
 * 							missing or blank. Note that this value need not be between <code>MinValue</code> and 
 * 							<code>MaxValue</code>. Also, if <code>AllowedToBeBlank</code> is <code>false</code>, 
 * 							you may pass <code>null</code> for this parameter.
 * @return					The values of these form controls represented as an array of <code>int</code> primitives.
 * @throws QwicapException	If anything goes wrong.
 */
	public int[] getInts
	(
		final String			ControlName,
		final int				MinValue,
		final int				MaxValue,
		final boolean			AllowedToBeBlank,
		final int[]				DefaultValues
	)
	throws QwicapException
	{
	//	Convert the array of default double values to an equivalent array of BigDecimal objects.

		final BigInteger[]		DefaultValuesBI = new BigInteger[DefaultValues.length];

		for (int Index = 0, Count = DefaultValues.length; Index < Count; Index++)
			DefaultValuesBI[Index] = BigInteger.valueOf(DefaultValues[Index]);

	//	Get the values of the controls.

		final BigInteger[]		ValuesBI = getBigIntegers(ControlName, BigInteger.valueOf(MinValue), BigInteger.valueOf(MaxValue), AllowedToBeBlank, DefaultValuesBI);

	//	Convert the array of BigDecimal control values to an equivalent array of double objects.

		final int[]				Values = new int[ValuesBI.length];

		for (int Index = 0, Count = ValuesBI.length; Index < Count; Index++)
			Values[Index] = ValuesBI[Index].intValue();

		return Values;
	}

/**	Returns the values of the specified controls as <code>long</code> primitives. If any of the values supplied by 
 * 	the user are invalid, absent, or fall outside the range of acceptable values, the XHTML page is sent back to the 
 * 	user with a message describing the problem, and this method blocks until the user eventually supplies valid values.
 *
 *	<p>If you want to further constrain the acceptable values of these controls, evaluate the values returned by this 
 *	method and, if they do not meet your constraints, invoke the {@link #rejectInput(String, Object, Object) rejectInput} 
 * 	method. Repeat as necessary, until you get a value that satisfies your constraints.
 *	</p>
 * 
 * @param ControlName		Name of the form control whose value is to be returned.
 * @param MinValue			Minimum acceptable value of the control (inclusive).
 * @param MaxValue			Maximum acceptable value of the control (inclusive).
 * @param AllowedToBeBlank	<code>true</code> if the input can be missing or blank.
 * @param DefaultValues		Values to return if <code>AllowedToBeBlank</code> is <code>true</code>, and the inputs were 
 * 							missing or blank. Note that this value need not be between <code>MinValue</code> and 
 * 							<code>MaxValue</code>. Also, if <code>AllowedToBeBlank</code> is <code>false</code>, 
 * 							you may pass <code>null</code> for this parameter.
 * @return					The values of these form controls represented as an array of <code>long</code> primitives.
 * @throws QwicapException	If anything goes wrong.
 */
	public long[] getLongs
	(
		final String			ControlName,
		final long				MinValue,
		final long				MaxValue,
		final boolean			AllowedToBeBlank,
		final long[]			DefaultValues
	)
	throws QwicapException
	{
	//	Convert the array of default double values to an equivalent array of BigDecimal objects.

		final BigInteger[]		DefaultValuesBI = new BigInteger[DefaultValues.length];

		for (int Index = 0, Count = DefaultValues.length; Index < Count; Index++)
			DefaultValuesBI[Index] = BigInteger.valueOf(DefaultValues[Index]);

	//	Get the values of the controls.

		final BigInteger[]		ValuesBI = getBigIntegers(ControlName, BigInteger.valueOf(MinValue), BigInteger.valueOf(MaxValue), AllowedToBeBlank, DefaultValuesBI);

	//	Convert the array of BigDecimal control values to an equivalent array of double objects.

		final long[]			Values = new long[ValuesBI.length];

		for (int Index = 0, Count = ValuesBI.length; Index < Count; Index++)
			Values[Index] = ValuesBI[Index].longValue();

		return Values;
	}

/**	Returns the values of the specified controls as <code>BigInteger</code> objects. If any of the values supplied by 
 * 	the user are invalid, absent, or fall outside the range of acceptable values, the XHTML page is sent back to the 
 * 	user with a message describing the problem, and this method blocks until the user eventually supplies valid values.
 *
 *	<p>If you want to further constrain the acceptable values of these controls, evaluate the values returned by this 
 *	method and, if they do not meet your constraints, invoke the {@link #rejectInput(String, Object, Object) rejectInput} 
 * 	method. Repeat as necessary, until you get a value that satisfies your constraints.
 *	</p>
 * 
 * @param ControlName		Name of the form control whose value is to be returned.
 * @param MinValue			Minimum acceptable value of the control (inclusive).
 * @param MaxValue			Maximum acceptable value of the control (inclusive).
 * @param AllowedToBeBlank	<code>true</code> if the input can be missing or blank.
 * @param DefaultValues		Values to return if <code>AllowedToBeBlank</code> is <code>true</code>, and the inputs were 
 * 							missing or blank. Note that this value need not be between <code>MinValue</code> and 
 * 							<code>MaxValue</code>. Also, if <code>AllowedToBeBlank</code> is <code>false</code>, 
 * 							you may pass <code>null</code> for this parameter.
 * @return					The values of these form controls represented as an array of <code>BigInteger</code> objects.
 * @throws QwicapException	If anything goes wrong.
 */
	public BigInteger[] getBigIntegers
	(
		final String			ControlName,
		final BigInteger		MinValue,
		final BigInteger		MaxValue,
		final boolean			AllowedToBeBlank,
		final BigInteger[]		DefaultValues
	)
	throws QwicapException
	{
		final String[]			ValueStrs  = getAll(ControlName);
		final int				ValueCount = ValueStrs.length;
		final boolean			Blank      = ValueCount == 0;

		if (Blank && AllowedToBeBlank)
			return DefaultValues;

		if (!Blank) {
			final BigInteger[]		Values = new BigInteger[ValueCount];

			for (int VIndex = 0; VIndex < ValueCount; VIndex++) {
				final String			ValueStr = ValueStrs[VIndex];
				XMLString				ErrorMsg;

				try {
					final BigInteger		Value = i18n.parseInteger(ValueStr);

					if (Value.compareTo(MinValue) >= 0 && Value.compareTo(MaxValue) <= 0) {

						Values[VIndex] = Value;
						continue;
					}

					if (Value.compareTo(MinValue) < 0)
						ErrorMsg = i18n.getMessageXML("getInteger-ValueBelowRange", HTML.encode(ControlName), Value, MinValue, MaxValue);
					else
						ErrorMsg = i18n.getMessageXML("getInteger-ValueAboveRange", HTML.encode(ControlName), Value, MinValue, MaxValue);

				} catch (ArithmeticException e) {

					ErrorMsg = i18n.getMessageXML("getInteger-ValueNotAnInteger", HTML.encode(ControlName), HTML.encode(ValueStr));

				} catch (Exception e) {

					ErrorMsg = i18n.getMessageXML("getNumber-ValueNotANumber", HTML.encode(ControlName), HTML.encode(ValueStr));
				}

				rejectInput(ControlName, ValueStr, ErrorMsg);	//	This method invocation will throw an exception that will terminate this loop.
			}

			return Values;

		} else
			rejectInput(ControlName, "", i18n.getMessageXML("getNumber-ValueBlank", HTML.encode(ControlName)));

	//	Because rejectInput ALWAYS throws an exception, we will NEVER reach the following return statement.

		return null;
	}

/**	Returns the values of the specified controls as <code>float</code> primitives. If any of the values supplied by 
 * 	the user are invalid, absent, or fall outside the range of acceptable values, the XHTML page is sent back to the 
 * 	user with a message describing the problem, and this method blocks until the user eventually supplies valid values.
 *
 *	<p>If you want to further constrain the acceptable values of these controls, evaluate the values returned by this 
 *	method and, if they do not meet your constraints, invoke the {@link #rejectInput(String, Object, Object) rejectInput} 
 * 	method. Repeat as necessary, until you get a value that satisfies your constraints.
 *	</p>
 * 
 * @param ControlName		Name of the form control whose value is to be returned.
 * @param MinValue			Minimum acceptable value of the control (inclusive).
 * @param MaxValue			Maximum acceptable value of the control (inclusive).
 * @param AllowedToBeBlank	<code>true</code> if the input can be missing or blank.
 * @param DefaultValues		Values to return if <code>AllowedToBeBlank</code> is <code>true</code>, and the inputs were 
 * 							missing or blank. Note that this value need not be between <code>MinValue</code> and 
 * 							<code>MaxValue</code>. Also, if <code>AllowedToBeBlank</code> is <code>false</code>, 
 * 							you may pass <code>null</code> for this parameter.
 * @return					The values of these form controls represented as an array of <code>float</code> primitives.
 * @throws QwicapException	If anything goes wrong.
 */
	public float[] getFloats
	(
		final String			ControlName,
		final float				MinValue,
		final float				MaxValue,
		final boolean			AllowedToBeBlank,
		final float[]			DefaultValues
	)
	throws QwicapException
	{
	//	Convert the array of default float values to an equivalent array of BigDecimal objects.

		final BigDecimal[]		DefaultValuesBD = new BigDecimal[DefaultValues.length];

		for (int Index = 0, Count = DefaultValues.length; Index < Count; Index++)
			DefaultValuesBD[Index] = new BigDecimal(DefaultValues[Index]);

	//	Get the values of the controls.

		final BigDecimal[]		ValuesBD = getBigDecimals(ControlName, new BigDecimal(MinValue), new BigDecimal(MaxValue), AllowedToBeBlank, DefaultValuesBD);

	//	Convert the array of BigDecimal control values to an equivalent array of float objects.

		final float[]			Values = new float[ValuesBD.length];

		for (int Index = 0, Count = ValuesBD.length; Index < Count; Index++)
			Values[Index] = ValuesBD[Index].floatValue();

		return Values;
	}

/**	Returns the values of the specified controls as <code>double</code> primitives. If any of the values supplied by 
 * 	the user are invalid, absent, or fall outside the range of acceptable values, the XHTML page is sent back to the 
 * 	user with a message describing the problem, and this method blocks until the user eventually supplies valid values.
 *
 *	<p>If you want to further constrain the acceptable values of these controls, evaluate the values returned by this 
 *	method and, if they do not meet your constraints, invoke the {@link #rejectInput(String, Object, Object) rejectInput} 
 * 	method. Repeat as necessary, until you get a value that satisfies your constraints.
 *	</p>
 * 
 * @param ControlName		Name of the form control whose value is to be returned.
 * @param MinValue			Minimum acceptable value of the control (inclusive).
 * @param MaxValue			Maximum acceptable value of the control (inclusive).
 * @param AllowedToBeBlank	<code>true</code> if the input can be missing or blank.
 * @param DefaultValues		Values to return if <code>AllowedToBeBlank</code> is <code>true</code>, and the inputs were 
 * 							missing or blank. Note that this value need not be between <code>MinValue</code> and 
 * 							<code>MaxValue</code>. Also, if <code>AllowedToBeBlank</code> is <code>false</code>, 
 * 							you may pass <code>null</code> for this parameter.
 * @return					The values of these form controls represented as an array of <code>double</code> primitives.
 * @throws QwicapException	If anything goes wrong.
 */
	public double[] getDoubles
	(
		final String			ControlName,
		final double			MinValue,
		final double			MaxValue,
		final boolean			AllowedToBeBlank,
		final double[]			DefaultValues
	)
	throws QwicapException
	{
	//	Convert the array of default double values to an equivalent array of BigDecimal objects.

		final BigDecimal[]		DefaultValuesBD = new BigDecimal[DefaultValues.length];

		for (int Index = 0, Count = DefaultValues.length; Index < Count; Index++)
			DefaultValuesBD[Index] = new BigDecimal(DefaultValues[Index]);

	//	Get the values of the controls.

		final BigDecimal[]		ValuesBD = getBigDecimals(ControlName, new BigDecimal(MinValue), new BigDecimal(MaxValue), AllowedToBeBlank, DefaultValuesBD);

	//	Convert the array of BigDecimal control values to an equivalent array of double objects.

		final double[]			Values = new double[ValuesBD.length];

		for (int Index = 0, Count = ValuesBD.length; Index < Count; Index++)
			Values[Index] = ValuesBD[Index].doubleValue();

		return Values;
	}

/**	Returns the values of the specified controls as <code>BigDecimal</code> objects. If any of the values supplied by 
 * 	the user are invalid, absent, or fall outside the range of acceptable values, the XHTML page is sent back to the 
 * 	user with a message describing the problem, and this method blocks until the user eventually supplies valid values.
 *
 *	<p>If you want to further constrain the acceptable values of these controls, evaluate the values returned by this 
 *	method and, if they do not meet your constraints, invoke the {@link #rejectInput(String, Object, Object) rejectInput} 
 * 	method. Repeat as necessary, until you get a value that satisfies your constraints.
 *	</p>
 * 
 * @param ControlName		Name of the form control whose value is to be returned.
 * @param MinValue			Minimum acceptable value of the control (inclusive).
 * @param MaxValue			Maximum acceptable value of the control (inclusive).
 * @param AllowedToBeBlank	<code>true</code> if the input can be missing or blank.
 * @param DefaultValues		Values to return if <code>AllowedToBeBlank</code> is <code>true</code>, and the inputs were 
 * 							missing or blank. Note that this value need not be between <code>MinValue</code> and 
 * 							<code>MaxValue</code>. Also, if <code>AllowedToBeBlank</code> is <code>false</code>, 
 * 							you may pass <code>null</code> for this parameter.
 * @return					The values of these form controls represented as an array of <code>BigDecimal</code> objects.
 * @throws QwicapException	If anything goes wrong.
 */
	public BigDecimal[] getBigDecimals
	(
		final String			ControlName,
		final BigDecimal		MinValue,
		final BigDecimal		MaxValue,
		final boolean			AllowedToBeBlank,
		final BigDecimal[]		DefaultValues
	)
	throws QwicapException
	{
		final String[]			ValueStrs  = getAll(ControlName);
		final int				ValueCount = ValueStrs.length;
		final boolean			Blank      = ValueCount == 0;

		if (Blank && AllowedToBeBlank)
			return DefaultValues;

		if (!Blank) {
			final BigDecimal[]		Values = new BigDecimal[ValueCount];

			for (int VIndex = 0; VIndex < ValueCount; VIndex++) {
				final String			ValueStr = ValueStrs[VIndex];
				XMLString				ErrorMsg;

				try {
					final BigDecimal		Value = i18n.parseDecimal(ValueStr);

					if (Value.compareTo(MinValue) >= 0 && Value.compareTo(MaxValue) <= 0) {

						Values[VIndex] = Value;
						continue;
					}

					if (Value.compareTo(MinValue) < 0)
						ErrorMsg = i18n.getMessageXML("getFloatingPoint-ValueBelowRange", HTML.encode(ControlName), Value, MinValue, MaxValue);
					else
						ErrorMsg = i18n.getMessageXML("getFloatingPoint-ValueAboveRange", HTML.encode(ControlName), Value, MinValue, MaxValue);

				} catch (Exception e) {

					ErrorMsg = i18n.getMessageXML("getNumber-ValueNotANumber", HTML.encode(ControlName), HTML.encode(ValueStr));
				}

				rejectInput(ControlName, ValueStr, ErrorMsg);	//	This method invocation will throw an exception that will terminate this loop.
			}

			return Values;

		} else
			rejectInput(ControlName, "", i18n.getMessageXML("getNumber-ValueBlank", HTML.encode(ControlName)));

	//	Because rejectInput ALWAYS throws an exception, we will NEVER reach the following return statement.

		return null;
	}

/**	Returns the value of the specified parameter as a <code>String</code>. If the value supplied by the user is too 
 *	long, or too short, the XHTML page is sent back to the user with a message describing the problem, and this method 
 *	blocks until the user eventually supplies a valid value.
 *
 *	<p>If you want to further constrain the acceptable values of this control, evaluate the value returned by this 
 *	method and if it does not meet your constraints, invoke the {@link #rejectInput(String, Object, Object) rejectInput} 
 * 	method. Repeat as necessary, until you get a value that satisfies your constraints.
 *	</p>
 * 
 * @param ControlName				Name of the form control whose value is to be returned.
 * @param MinLength					Minimum acceptable length of input, after optional trimming is performed.
 * @param MaxLength					Maximum acceptable length of input, after optional trimming is performed.
 * @param Trim						<code>true</code> if leading and trailing white space should be removed from the 
 * 									input string.
 * @return							The value of the specified control, after optional trimming, if it met all of the
 *									length and pattern requirements.
 * @throws QwicapException			If anything goes wrong.
 * @see #get(String)
 */
	public String getString
	(
		final String			ControlName,
		final int				MinLength,
		final int				MaxLength,
		final boolean			Trim
	)
	throws QwicapException
	{
		return getString(ControlName, MinLength, MaxLength, Trim, null, null);
	}

/**	Returns the value of the specified control as a <code>String</code>. If the value supplied by the user is too long, 
 *	too short, or fails to conform to the specified pattern, the XHTML page is sent back to the user with a message 
 *	describing the problem, and this method blocks until the user eventually supplies a valid value.
 *
 *	<p>If you want to further constrain the acceptable values of this control, evaluate the value returned by this 
 *	method and if it does not meet your constraints, invoke the {@link #rejectInput(String, Object, Object) rejectInput} 
 * 	method. Repeat as necessary, until you get a value that satisfies your constraints.
 * 	</p>
 * 
 * @param ControlName				Name of the form control whose value is to be returned.
 * @param MinLength					Minimum acceptable length of input, after optional trimming is performed.
 * @param MaxLength					Maximum acceptable length of input, after optional trimming is performed.
 * @param Trim						<code>true</code> if leading and trailing white space should be removed from the 
 *									input string.
 * @param ValidatorRegex			<code>null</code>, or a regular expression defining the acceptable pattern for the 
 *									input string.
 * @param ValidatorRegexErrorMsg	<code>null</code>, or an error message to be displayed when the string does not 
 *									conform to the pattern specified by the regular expression. If the error message
 * 									object is a subclass of {@link Markup}, it is inserted into the web page as-is. If
 * 									the error message is any other type of object, it is converted to a <code>String</code>, 
 * 									and HTML encoded before it is inserted.
 * @return							The value of the specified control, after optional trimming, if it met all of the
 *									length and pattern requirements.
 * @throws QwicapException			If anything goes wrong.
 * @see #get(String)
 */
	public String getString
	(
		final String			ControlName,
		final int				MinLength,
		final int				MaxLength,
		final boolean			Trim,
		final String			ValidatorRegex,			//	Can be null.
		final Object			ValidatorRegexErrorMsg	//	Can be null.
	)
	throws QwicapException
	{
		final String			OrigValueStr = get(ControlName);
		final Markup			ErrorMsg;

		if (OrigValueStr == null) {

			ErrorMsg = i18n.getMessageXML("getString-ValueMissing", HTML.encode(ControlName));

		} else {
			String					ValueStr = OrigValueStr;

			if (Trim)
				ValueStr = ValueStr.trim();

			final int				Length = ValueStr.length();

			if (MaxLength == MinLength && Length != MaxLength) {

				if (MaxLength < Length)
					ErrorMsg = i18n.getMessageXML("getString-ValueLengthTooHigh", HTML.encode(ControlName), HTML.encode(ValueStr), Length, MaxLength);
				else
					ErrorMsg = i18n.getMessageXML("getString-ValueLengthTooLow",  HTML.encode(ControlName), HTML.encode(ValueStr), Length, MaxLength);

			} else if (Length < MinLength) {

				ErrorMsg = i18n.getMessageXML("getString-ValueLengthBelowRange", HTML.encode(ControlName), HTML.encode(ValueStr), Length, MinLength, MaxLength);

			} else if (Length > MaxLength) {

				ErrorMsg = i18n.getMessageXML("getString-ValueLengthAboveRange", HTML.encode(ControlName), HTML.encode(ValueStr), Length, MinLength, MaxLength);

			} else if (ValidatorRegex != null && ValueStr.matches(ValidatorRegex) == false) {

				if (ValidatorRegexErrorMsg != null) {

					if (ValidatorRegexErrorMsg instanceof Markup)
						ErrorMsg = (Markup) ValidatorRegexErrorMsg;
					else
						ErrorMsg = new XMLString(HTML.encode(ValidatorRegexErrorMsg.toString()));

				} else
					ErrorMsg = i18n.getMessageXML("getString-ValueRegexNotMatched", HTML.encode(ControlName), HTML.encode(ValueStr), HTML.encode(ValidatorRegex));

			} else
				return ValueStr;
		}

		rejectInput(ControlName, OrigValueStr, ErrorMsg);

	//	Because rejectInput ALWAYS throws an exception, we will NEVER reach the following return statement.

		return null;
	}

/**	Creates an instance of <code>Downloadable</code> to store the information triplet representing any downloadable 
 *	item: name, <a href='http://www.iana.org/assignments/media-types/'>MIME type</a>, and content bytes. Creating an 
 * 	instance of <code>Downloadable</code> does cause a unique, relative URI to be assigned to the instance, but does 
 * 	not instruct Qwicap to make it available for download. That is done by invoking 
 * 	{@link Qwicap#addDownloadable(Downloadable) Qwicap.addDownloadable}.
 * 
 * @param Name					The name of this downloadable, for example: "myimage.jpg".
 * @param MIMEType				The <a href='http://www.iana.org/assignments/media-types/'>MIME type</a> of this 
 *								downloadable, for example: "image/jpeg".
 * @param Contents				The contents of this download, represented as the array of bytes that will be 
 *								trasmitted to the client.
 * @return						The new instance of <code>Downloadable</code>.
 */
	public Downloadable createDownloadable
	(
		final String			Name,
		final String 			MIMEType,
		final byte[]			Contents
	)
	{
		return new Downloadable(getRandom(), Name, MIMEType, Contents);
	}

/**	Associates a new downloadable item with the web page specified in the next <code>prompt</code> invocation. If that
 *	page already has a downloadable item of the same name, this item will replace it, and the previous item will become
 *	available for garbage collection. Note that this item is assigned a new URI, even if it replaces a previous item of 
 *	the same name, so any links in the document will still have to be updated. When the page is no longer in-play, all 
 *	downloadable items associated with it will be made available for garbage collection. 
 *	
 *	<p>The <a href='http://www.iana.org/assignments/media-types/'>MIME type</a>
 *	of the item is inferred from its name, using the {@link ServletContext#getMimeType} method.
 *	Because the outcome of the this inference is determined by the configuration of the underlying servlet engine, it
 *	will always be more dependable to explicity specify the MIME type using the 
 *	{@link #addDownloadable(String, String, byte[])} method.
 *	</p>
 * 
 * @param Name					The name to be given to this download, for example "mypicture.jpg".
 * @param Contents				The contents of this download, represented as the array of bytes that will be 
 *								trasmitted to the client.
 * @return						A relative URI for this downloadable item. Use this URI as the link to this item.
 */
	public String addDownloadable
	(
		final String			Name,
		final byte[]			Contents
	)
	{
		return addDownloadable(Name, getHitContext().getServlet().getServletContext().getMimeType(Name), Contents);
	}

/**	Associates a new downloadable item with the web page specified in the next <code>prompt</code> invocation. If that
 *	page already has a downloadable item of the same name, this item will replace it, and the previous item will become
 *	available for garbage collection. Note that this item is assigned a new URI, even if it replaces a previous item of 
 *	the same name, so any links in the document will still have to be updated. When the page is no longer in-play, all 
 *	downloadable items associated with it will be made available for garbage collection.
 * 
 * @param Name					The name to be given to this download, for example "mypicture.jpg".
 * @param MIMEType				The <a href='http://www.iana.org/assignments/media-types/'>MIME type</a> of the this 
 *								download, for example "image/jpeg".
 * @param Contents				The contents of this download, represented as the array of bytes that will be 
 *								trasmitted to the client.
 * @return						A relative URI for this downloadable item. Use this URI as the link to this item.
 */
	public String addDownloadable
	(
		final String			Name,
		final String			MIMEType,
		final byte[]			Contents
	)
	{
		return addDownloadable(new Downloadable(getRandom(), Name, MIMEType, Contents)).getURI();
	}

/**	Associates a new downloadable item with the web page specified in the next <code>prompt</code> invocation. If that
 *	page already has a downloadable item of the same name, this item will replace it, and the previous item will become
 *	available for garbage collection. Note that this item is assigned a new URI, even if it replaces a previous item of 
 *	the same name, so any links in the document will still have to be updated. When the page is no longer in-play, all 
 *	downloadable items associated with it will be made available for garbage collection.
 * 
 * @param Download				A <code>Downloadable</code> object to be made available for download.
 * @return						A relative URI for this downloadable item. Use this URI as the link to this item.
 */
	public Downloadable addDownloadable
	(
		final Downloadable		Download
	)
	{
		if (NewDownloadables == null)
			NewDownloadables = new ArrayList<Downloadable>();

		NewDownloadables.add(Download);

		return Download;
	}

/**	Encodes a string such that it can be safely embedded in an XHTML document without risk of the document's syntax
 * 	being corrupted, or the contents of the string being interpreted as XHTML markup (a potential security hole).
 * 
 * @param RawStr		The string to be encoded.
 * @return				A string with all reserved characters encoded, or the original string object if no encoding was
 *						necessary.
 */
	public static String htmlEncode
	(
		final String			RawStr
	)
	{
		return HTML.encode(RawStr);
	}

/**	Decodes a string such any HTML character entities in it are translated into corresponding Unicode characters.
 * 
 * @param PotentiallyEncodedStr	The string to be decoded.
 * @return						A string with all HTML characters entities decoded into corresponding Unicode characters.
 */
	public static String htmlDecode
	(
		final String			PotentiallyEncodedStr
	)
	{
		return HTML.decode(PotentiallyEncodedStr);
	}

/**	Takes the form data set most recently received from the client and embeds it into the XHTML markup of the form from 
 *	which it originated in the current page. This happens automatically when input is rejected by one of the 
 *	"get" methods, or when {@link #rejectInput rejectInput} is called. If there is no most-recent-form, this method does 
 *	nothing.
 * 
 * @throws TagException				If there was a problem with the manipulation of the page markup.
 * @throws FormNotFoundException	If the <samp>form</samp> element has disappeared from the markup.
 * @return							A reference to this <code>Qwicap</code> instance, to facilitate method invocation
 * 									chaining.
 */
	public Qwicap populateMostRecentForm() throws TagException, FormNotFoundException {

		Pages.populateMostRecentForm();
		return this;
	}

/**	Removes all values from the XHTML markup of the form in the current page from which input was most recently 
 *	received. If there is no most-recent-form, this method does nothing.
 * 
 * @throws TagException				If there was a problem with the manipulation of the page markup.
 * @throws FormNotFoundException	If the <samp>form</samp> element has disappeared from the markup.
 * @return							A reference to this <code>Qwicap</code> instance, to facilitate method invocation
 * 									chaining.
 */
	public Qwicap clearMostRecentForm() throws TagException, FormNotFoundException {

		Pages.clearMostRecentForm();
		return this;
	}

/**	Sends the final XHTML page of this session to the user, and returns without waiting for input from the user. 
 *	(Since it's the end of the session, there would be no point in the final page including a form that supplies any 
 *	input, and therefore there's no need to worry about the possibility of input.) Any further attempts by the 
 * 	application to retrieve input, send pages to the user, or to act like this session is still in progress will cause 
 *	<code>QwicapSessionDeathException</code>s to be thrown.
 * 
 * @param PageMarkup				Final XHTML page to be sent to the user.
 * @return							A reference to this <code>Qwicap</code> instance, to facilitate method 
 * 									invocation chaining.
 * @throws QwicapException			If anything went wrong.
 */
	public Qwicap goodbye
	(
		final Markup			PageMarkup
	)
	throws QwicapException
	{
		Log.log(Level.FINE, "Qwicap.goodbye({0}) invoked.", PageMarkup.getMarkupName());

		if (PleaseDie)
			throwSessionKillingExceptions();

		try {
			
			GoodbyeInvoked = true;

		//	Set the current page.

			final Page				CurPage = Pages.setGoodbyePage(PageMarkup.getMutable());

		//	If we're correcting for hideous bugs in Internet Explorer, convert all of the "button" elements of type 
		//	"submit" to "input" elements of type "submit", if the browser's identification string matches one of the 
		//	regular expressions for which we've been told to watch. (Note the we don't have to undo the conversion, 
		//	because the application is exiting.)

			doButtonConversions(CurPage);

		//	Place a message in the page describing any exceptions that were reported to Qwicap by the application, 
		//	using the reportException method.

			insertIncidentReport(CurPage);

		//	Remove the Qwicap scripts, because they don't apply to the "goodbye" page.

			AJAX.removeScripts(CurPage);

		//	Remove page IDs (and preferably all Qwicap nonces) from the forms and URLs on the "goodbye" page.
			
			CurPage.removePageAndFormIDs();
			
		//	Add any new downloadables to the current page.

			if (NewDownloadables != null && NewDownloadables.size() != 0) {

				Pages.addDownloadables(NewDownloadables);
				NewDownloadables.clear();
			}

		//	Transmit this page to the client.

			CurPage.write(this);
			
			return this;

		} finally {

		//	Tell this thread to die, and supply the error message it should include in any exceptions it has to throw
		//	to prevent the thread from continuing to operate as if it was still serving a user.

			pleaseDie("The \"goodbye\" page, \"" + PageMarkup.getMarkupName() + "\", has already been sent to the client.");
		}
	}

/**	Sends the final item of this session to the user, and returns without waiting for input from the user. Any further 
 * 	attempts by the application to retrieve input, send pages to the user, or to act like this session is still in 
 * 	progress will cause <code>QwicapSessionDeathException</code>s to be thrown.
 * 
 * @param ContentType				MIME type of data to be sent to client. Examples: "<samp>image/jpeg</samp>", 
 * 									"<samp>image/svg+xml; charset=UTF-8</samp>".
 * @param Data						The data to be sent to the client. If the data is text, it must already be encoded
 * 									in the character set specified in the <code>ContentType</code> parameter.
 * @return							A reference to this <code>Qwicap</code> instance, to facilitate method 
 * 									invocation chaining.
 * @throws QwicapException			If anything went wrong.
 */
	public Qwicap goodbye
	(
		final String			ContentType,
		final byte[]			Data
	)
	throws QwicapException
	{
		Log.log(Level.FINE, "Qwicap.goodbye({0}, {1} bytes) invoked.", new Object[] { ContentType, Data.length });

		if (PleaseDie)
			throwSessionKillingExceptions();

		try {
			
			GoodbyeInvoked = true;
			
			final HttpServletResponse	Resp = getHitContext().getResponse();
			
			Resp.setStatus(200);
			Resp.setContentType(ContentType);
			Resp.setContentLength(Data.length);
			Resp.setHeader("Cache-Control", "no-cache");
			
			final ServletOutputStream	Out = Resp.getOutputStream();
			
			Out.write(Data);
			Out.flush();	//	Commit the response.
		
			return this;

		} catch (IOException e) {
			throw new QwicapException(e);	
		} finally {

		//	Tell this thread to die, and supply the error message it should include in any exceptions it has to throw
		//	to prevent the thread from continuing to operate as if it was still serving a user.

			pleaseDie("The \"goodbye\" document (content type \"" + ContentType + "\", " + Data.length + " bytes) has already been sent to the client.");
		}
	}

/**	Sends the final item of this session to the user, and returns without waiting for input from the user. Any further 
 * 	attempts by the application to retrieve input, send pages to the user, or to act like this session is still in 
 * 	progress will cause <code>QwicapSessionDeathException</code>s to be thrown.
 * 
 * @param Download					A <code>Downloadable</code> to sent to the client.
 * @return							A reference to this <code>Qwicap</code> instance, to facilitate method 
 * 									invocation chaining.
 * @throws QwicapException			If anything went wrong.
 */
	public Qwicap goodbye
	(
		final Downloadable		Download
	)
	throws QwicapException
	{
		Log.log(Level.FINE, "Qwicap.goodbye(Downloadable({0}, {1} bytes)) invoked.", new Object[] { Download.getMIMEType(), Download.getContentLength() });

		if (PleaseDie)
			throwSessionKillingExceptions();

		try {
			
			GoodbyeInvoked = true;
			
			Download.send(getHitContext().getResponse());
		
			return this;

		} catch (IOException e) {
			throw new QwicapException(e);
		} finally {

		//	Tell this thread to die, and supply the error message it should include in any exceptions it has to throw
		//	to prevent the thread from continuing to operate as if it was still serving a user.

			pleaseDie("The \"goodbye\" document (content type \"" + Download.getMIMEType() + "\", " + Download.getContentLength() + " bytes) has already been sent to the client.");
		}
	}

/** Redirects to the client to an arbitrary URL, and returns immediately.  Any further attempts by the application to 
 * 	retrieve input, send pages to the user, or to act like this session is still in progress will cause 
 *	<code>QwicapSessionDeathException</code>s to be thrown.
 * 
 * @param RedirectURL				The URL to which the client should be redirected.
 * @return							A reference to this <code>Qwicap</code> instance, to facilitate method 
 * 									invocation chaining.
 * @throws QwicapException			If anything went wrong.
 * @see #redirect(String)
 */
	public Qwicap goodbye
	(
		final String			RedirectURL
	)
	throws QwicapException
	{
		Log.log(Level.FINE, "Qwicap.goodbye(<{0}>) invoked.", RedirectURL);

		if (PleaseDie)
			throwSessionKillingExceptions();

		try {
			
			GoodbyeInvoked = true;

			getHitContext().getResponse().sendRedirect(RedirectURL);

			return this;

		} catch (IOException e) {
			throw new QwicapException(e);
		} finally {
			pleaseDie("The \"goodbye\" redirect, <" + RedirectURL + ">, has already been sent to the client.");	//	It's time for this application to exit, even if we have to force it.
		}
	}

	private boolean doButtonConversions
	(
		final Page				Pg
	)
	throws QwicapException
	{

	//	If we're correcting for hideous bugs in Internet Explorer, convert all the "button" elements of type "submit"
	//	to "input" elements of type "submit", if the browser's identification string matches one of the regular 
	//	expressions for which we've been told to watch.

		boolean					DoButtonConversion = false;

		if (ConvertSubmitButtonsToInputsUserAgentPatterns != null) {
			final String			UserAgent = getUserAgent();

			for (int Index = 0, Count = ConvertSubmitButtonsToInputsUserAgentPatterns.length; Index < Count; Index++) {

				if (ConvertSubmitButtonsToInputsUserAgentPatterns[Index].reset(UserAgent).find()) {

					DoButtonConversion = true;
					break;
				}
			}

			if (DoButtonConversion)
				Pg.doButtonConversions();
		}

		return DoButtonConversion;
	}

//	private MutableMarkup doButtonConversions2
//	(
//		final MutableMarkup		XHTMLPage
//	)
//	throws QwicapException
//	{
//		
//	//	If we're correcting for hideous bugs in Internet Explorer, convert all the "button" elements of type "submit"
//	//	to "input" elements of type "submit", if the browser's identification string matches one of the regular 
//	//	expressions for which we've been told to watch.
//		
//		boolean					DoButtonConversion = false;
//		
//		if (ConvertSubmitButtonsToInputsUserAgentPatterns != null) {
//			final String			UserAgent = getUserAgent();
//			
//			for (int Index = 0, Count = ConvertSubmitButtonsToInputsUserAgentPatterns.length; Index < Count; Index++) {	
//				
//				if (ConvertSubmitButtonsToInputsUserAgentPatterns[Index].reset(UserAgent).find()) {
//					
//					DoButtonConversion = true;
//					break;
//				}
//			}
//			
//			if (DoButtonConversion) {
//				System.out.println("Converting buttons to inputs.");
//				
//				if (XHTMLPage.get("button").notEmpty()) {
//					final MutableMarkup			CopyOfPage = XHTMLPage.getImmutable().getMutable();
//					
//					CopyOfPage.convertSubmitButtonsToInputs();
//					return CopyOfPage;
//				}
//			}
//		}
//		
//		return XHTMLPage;
//	}

//	private void undoButtonConversions
//	(
//		final MutableMarkup		XHTMLPage
//	)
//	throws QwicapException
//	{
//	//	System.out.println("Undoing button conversions.");
//		
//	//	Reverse the button conversions we performed above, so that the method that called "prompt" can't be affected
//	//	by our conversions. (Well, it can be affected, but only if the method was holding references to Match objects
//	//	that referred to the "button" elements and/or their content. This would be a very unusual way of working,
//	//	and it would be straightforward to recode the client method to eliminate that practice.)
//		
//		XHTMLPage.convertSubmitInputsToButtons();
//	}
	
	
//	todo Deprecate printPageStack, and substitute a "logPageStack" method, or better, yet, a "pageStackToString" method.
	
/**	Prints to the console a summary of Qwicap's internal page stack as an aid to understanding Qwicap's automatic
 * 	page stack maintenance.
 * 
 * @return						A reference to this instance of the <code>Qwicap</code> class.
 */
	public Qwicap printPageStack() {
		System.out.println(Pages);
		return this;
	}

	private ArrayToString arrayToStringOr
	(
		final Object			Array
	)
	{
		return new ArrayToString(
			Array,
			i18n.getString("ArrayToString-Separator"),
			i18n.getString("ArrayToString-FinalSeparatorOr"),
			null
		).setQuotes(
			i18n.getString("ArrayToString-OpenQuote"),
			i18n.getString("ArrayToString-CloseQuote")
		);
	}

	private ArrayToString arrayToStringAnd
	(
		final Object			Array
	)
	{
		return new ArrayToString(
			Array,
			i18n.getString("ArrayToString-Separator"),
			i18n.getString("ArrayToString-FinalSeparatorAnd"),
			null
		).setQuotes(
			i18n.getString("ArrayToString-OpenQuote"),
			i18n.getString("ArrayToString-CloseQuote")
		);
	}
}


/*	Features Needed
	---------------
	
	* Asynchronous pages
	
	  - Can be used in modal or modeless fashions.
	  
	    + In the modeless case, users can go back to previous pages that are still "in play", and, having done so, 
	      updates to the asynchronous page, or attempts to retrieve user input from it, will throw 
	      QwicapPageAbandonmentExceptions.
	
	  - Updates to asynchronous pages must NOT cause page or form IDs to be updated; they should remain whatever they
	    were when the page was sent the first time. Otherwise, we couldn't reliably receive user input (think "cancel"
	    buttons) from asynchronous pages. This issue arises because we depend on the client to periodically retrieve
	    asynchronous pages, so we can't be sure which versions of an asynchronous the client will actually retrieve.
	  
	  - Thread safety (Servlet thread vs. Qwicap thread) relating to the non-thread-safe implementation of MutableMarkup
	    can be resolved by having the asynchronous prompt method render the MutableMarkup it is given into a String or 
	    char array. The Servlet should transmit that pre-rendered page, rather than attempting to write-out the
	    MutableMarkup as would usually happen.
	    
	    + Every time an asynchronous page is updated, the updated page must be rendered into a new buffer, because the
	      old buffer(s) may still be in use by a Servlet thread which is transmitting the buffer(s) to the client.
	    
	    + We wouldn't necessarilly have to re-render the page every time the asynchronous page is updated.
	        
	      * All mutable markup objects have a "change count" that would let us determine whether or not the page had
	        actually been modified since the clone was created. I'm not sure when this no-updates update case would
	        arise in practice, but it would be easy-enough to allow for.
	    
	    + The Page class will have to be modified to store a reference to the pre-rendered page, in addition to the 
	      usual MutableMarkup reference.
	    
	    + Asynchronous page updates will result from a new invocation of the asynchronous "prompt" method. Changes to
	      the asynchronous page's markup won't affect the page that'll be sent to the client, until a new invocation of 
	      the asynchronous "prompt" method is made.
	      
	      * This is caused by the Servlet's use of a pre-rendered page.
	      
	      * This is necessary to solve the thread-safety problem associated with MutableMarkup.
	      
	      * This would be the right way to do things even if there were no thread-safety issue, because it ensures that
	        the client will never receive a partially updated page (one that was in the middle of being updated by the
	        application when the client requested it).
	        
	      * Unless the asynchronous prompt invocations are always made from the same place in the application code, the
	        same MutableMarkup object should be supplied every time.
	        
	        - This is the same behavior already embraced by the "prompt" method, down in the "autoPushPopReplace" logic.
	        
	        - Alternately, we could insist that all asynchronous prompt invocations when an asynchronous page is on top 
	          of the page stack will be interpreted as updates to the existing asynchronous page.
	          
	          + That would prevent consecutive, nested asynchronous pages. 
	          
	            * WOULD THAT BE GOOD OR BAD?
	            
	          + We can add an "asynch" boolean to the Page class to support error checking in such cases. 
	       
	            * If we get an asynchronous page update, and the Page on the top of the stack is NOT marked as "asynch", 
	              we throw an exception.
	              
	        - If we use separate methods to enter asynchronous mode, and to update the current asynchronous page, we 
	          might be able to avoid the ambiguity discussed above.
	          
	          + THIS WOULD COMPLICATE THE API A BIT, but it would also eliminate the need to to document the rule of
	            passing the same MutableMarkup instance every time, or to always prompt from the same place in the
	            application code. Avoiding that is probably a big win.
	    
	  - We could provide support for a standard progress bar, OR NOT.
	  
	    + The application wouldn't update the page directly, it would invoke something like a "madeProgress" method
	      that would increment a progress number toward a known limit.
	      
	      * A comment string could be passed to that method, too.
	      
	    + When asked for a page refresh, the QwicapServlet could read that value, and update the page markup accordingly
	      before sending it to the client.
	      
	      * This would only work if we did NOT use the pre-rendered page scheme described above, so this feature should
	        probably be implemented in a different manner, if it is implemented at all.
	      
	    + We would also need an open-ended progress indicator. It would tell the user, in effect, "we don't know when
	      the process will be done, but things are still happening".
	  
	  - Life Cycle / Flow Control
	
	    + A period of asynchronous operation is terminated by calling the normal (synchronous) Qwicap.prompt methods. 
	  
	      * UNLESS WE IMPLEMENT THE IDEA OF SEPARATE METHODS to send the first version of an asynchronous page, and to
	        update the page during asynchronous processing.
	        
	        - In this case, the operation of one asynchronous page could be terminated by invoking the prompt method
	          even in asynchronous mode.
	          
	          + The new asynchronous page gets sent to the client the next time the client asks for a refresh.
	          
	            * The client might never see the new asynchronous page if the refresh time on the previous asynchronous
	              page doesn't happen to arrive before the application completes its work with new asynchronous page.
	              
	              - That's life.
	  
	      * The page passed to those prompt methods is made the markup that will be sent to the client when it 
	        next auto-refreshes.
	      
	        - At that point, client and server are back in synch, and the arrival of input from the user following the
	          transmission of that synchronous page causes the normal (synchronous) prompt method to un-block.

	  - Input to asynchronous pages:
	  
	    + We'll need a method like "hasNewInput" to allow the application to test for the presence of user input that 
	      it hasn't seen before.
	      
	      - A FormDataSetFilter might be a good place for processing asynchronous input, too, but we should support
	        both.
	      
	    + CAN WE SUPPORT AUTOMATIC INPUT VALIDATION AND USER INPUT CORRECTION REQUESTS?
	    
	      * Yes, if we can come-up with a viable reconsideration of the prompt pattern. 
	      
	        - We need to reconsider the prompt pattern for this case, because the application can't get caught in the
	          prompt pattern loop when it's trying to get processing done.
*/





