/*	QWICAP (Quick Web Interface for Conventional Applications) by Chris W. Johnson
	Copyright (C) 2007-2008 The University of Texas at Austin ("U.T. Austin"). All rights reserved.
	
	U.T. Austin's copyright notice in this evolving version reflects our original authorship. It does not assert 
	copyright over changes contributed by others.
	
	This library is free software; you can redistribute it and/or modify it under the terms of the GNU Lesser General 
	Public License as published by the Free Software Foundation; either version 2.1 of the License, or (at your option) 
	any later version.
	
	This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied 
	warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for 
	more details.
	
	You should have received a copy of the GNU Lesser General Public License along with this library; if not, write to 
	the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/


package edu.utexas.its.eis.tools.qwicap.servlet;


import edu.utexas.its.eis.tools.qwicap.template.xml.immutable.XMLString;
import edu.utexas.its.eis.tools.qwicap.template.xml.TagException;
import edu.utexas.its.eis.tools.qwicap.template.xml.mutable.MutableMarkup;
import edu.utexas.its.eis.tools.qwicap.template.css.Results;
import edu.utexas.its.eis.tools.qwicap.template.css.CSSPatterns;

import javax.servlet.http.HttpServletResponse;
import javax.servlet.http.HttpServletRequest;
import java.util.logging.Level;
import java.util.logging.Logger;
import java.io.IOException;
import java.io.Writer;


/**	Provides the implementation of Qwicap's optional AJAX-based features.
 * 
 * 	@author Chris W. Johnson
 */
final class QwicapAJAX {

	private static final Logger			Log = Logger.getLogger(QwicapAJAX.class.getName());
	
	private static final CSSPatterns	kCSSHTMLHeadTag            = new CSSPatterns("html head");
	private static final CSSPatterns	kCSSScriptIDGlobals        = new CSSPatterns("script#qwicap-script-globals");
	private static final CSSPatterns	kCSSScriptTag              = new CSSPatterns("script");
	private static final CSSPatterns	kCSSBodyTag                = new CSSPatterns("body");
	private static final CSSPatterns	kCSSScriptAndScriptGlobals = new CSSPatterns("script#qwicap-script, script#qwicap-script-globals");

	private static final PageInfo[]		kNoPageInfo = new PageInfo[0];
	private static final XMLString		kQwicapStdScriptGlobals;
	private static final XMLString		kQwicapStdScript;

	static {
		try {
			kQwicapStdScriptGlobals = new XMLString("\t<script type='text/javascript' id='qwicap-script-globals'></script>\n");
			kQwicapStdScript        = new XMLString("\t<script type='text/javascript' src='qwicap.js' charset='UTF-8' id='qwicap-script'></script>\n");
		} catch (TagException e) {
			throw new RuntimeException(e);
		}
	}
	
	static final String				kQwicapAJAXCommandControlName = "qwicap-cmd";
	
	private final i18nServices		i18n;
	private final Qwicap			Q;
	private boolean					Enabled;		//	True if Qwicap can place Javascript(s) in pages.
	private boolean					ScriptsWorking;	//	True if those Javascript(s) are reporting back successfully.
	
	
	QwicapAJAX
	(
		final Qwicap			Q,
		final i18nManager		i18nMgr
	) 
	{
		this.Q    = Q;
		this.i18n = i18nMgr.getServicesForClass(Qwicap.class);
	}
	
	boolean getEnabled() {
		return Enabled;
	}
	
	void setEnabled
	(
		final boolean			ScriptsEnabled
	)
	{
		this.Enabled = ScriptsEnabled;
	}
	
	boolean getWorking() {
		return ScriptsWorking;	
	}
	
	void insertScripts
	(
		final Page				CurPg
	)
	throws TagException
	{
		if (!Enabled)
			return;

		final String			GlblsCode      = "QwicapXMLURL='" + Q.getHitContext().getRequest().getRequestURL().toString().replace("'", "%27") + "'; QwicapPageID='" + CurPg.getPageID() + "';";
		final MutableMarkup		Doc            = CurPg.getMarkup();
		final Results			Head           = Doc.get(kCSSHTMLHeadTag);
		final Results			GlblsScriptTag = Head.get(kCSSScriptIDGlobals);

		if (GlblsScriptTag.isEmpty()) {
			final MutableMarkup		Glbls = kQwicapStdScriptGlobals.getMutable();

			Glbls.get(kCSSScriptTag).setContent(GlblsCode);

			Head.addToEndOfContent(Glbls);
			Head.addToEndOfContent(kQwicapStdScript);

		} else
			GlblsScriptTag.setContent(GlblsCode);
		
	//	If the page's "body" element includes "onload" and "onunload" attributes, references to the scripts they 
	//	contain will be written to window.onload and window.onunload, thereby eliminating the references to Qwicap's
	//	"onload" and "onunload" scripts. Therefore, if "onload" and/or "onunload" attributes exist, we append code to
	//	their scripts that invokes the Qwicap script's pageLoaded and/or pageUnloaded functions.
		
		final Results			Body = Doc.get(kCSSBodyTag);
		
		addToOptionalScript(Body, "onload",   "window['edu.utexas.its.eis.tools.qwicap.pageLoaded']();"  );
		addToOptionalScript(Body, "onunload", "window['edu.utexas.its.eis.tools.qwicap.pageUnloaded']();");
	}
	
	private static void addToOptionalScript
	(
		final Results			Element,
		final String			ScriptAttrName,
		final String			AdditionalScript
	)
	throws TagException
	{
		final String			BaseScript = Element.getAttribute(ScriptAttrName);
		
	//	If there is no script, do nothing. If there is a script, and it already contains AdditionalScript, do nothing.
		
		if (BaseScript == null || BaseScript.contains(AdditionalScript))
			return;
		
	//	Append AdditionalScript to the script that is already present in the specified attribute.
		
		if (BaseScript.endsWith(";"))
			Element.setAttribute(ScriptAttrName, BaseScript + AdditionalScript);
		else
			Element.setAttribute(ScriptAttrName, BaseScript + ';' + AdditionalScript);
	}

	void removeScripts
	(
		final Page				CurPg
	)
	throws TagException
	{
		if (!Enabled)
			return;

		final MutableMarkup		Doc = CurPg.getMarkup();

		Doc.get(kCSSHTMLHeadTag).get(kCSSScriptAndScriptGlobals).delete();
	}

//	Note: Keep this implementation of processQwicapAJAXCommand in synch with the static implementation that follows, 
//	where appropriate. (Becuase the static implementation deals with AJAX commands that arrive in the absence of a valid
//	session, its job is mostly to provide properly constructed command responses that deny knowledge of anything.)
	
	boolean processCommand
	(
		final FormDataSet		DataSet
	)
	{
	//	If this session is dying, treat the AJAX command as though it arrived when there was no session.
		
		if (Q.shouldDie()) {
			final HitContext		HitCtxt = Q.getHitContext();
			
			return processCommand(HitCtxt.getServlet().getApplicationServices(), HitCtxt.getRequest(), HitCtxt.getResponse(), DataSet);
		}
		
	//	If this session is not dying, process the AJAX command properly.
		
		final String			Command = DataSet.getDatum(kQwicapAJAXCommandControlName);

		if (Command == null)
			return false;	//	This data set was NOT created by a Qwicap script.
		
		if (!Enabled)
			return true;	//	This data set appears to have been created by Qwicap script, but AJAX is now diabled, so we do NOT process the data set.
		
		try {
			final HttpServletResponse	Resp        = Q.getHitContext().getResponse();
			boolean						GoodCommand = true;
			
			Resp.addHeader("Cache-Control", "no-cache");			//	Prevents these responses from being cached.
			Resp.setContentType("application/xml; charset=UTF-8");	//	Prevents IE from throwing the "Object Expected" error.

			if ("pgstat".equals(Command)) {
				final String			Event    = DataSet.getDatum("event");
				final String			PgID     = DataSet.getDatum("pageid");
				final PageStack			PgStack  = Q.getPageStack();
				final String			PgState  = PgStack.getPageState(PgID);
				final String			Message  = "unknown".equals(PgState) ? i18n.getString("NavigationError-CannotGoToInvalidPage") : null;
				final String			OK       = i18n.getString("NavigationError-OK");
				final String			ChoosePg = i18n.getString("NavigationError-ChooseActivePage");
				
				sendPageStateReply(Resp, PgID, PgState, Message, OK, ChoosePg, PgStack.getInfo());
				Log.log(Level.FINE, "AJAX: Requested page state for \"{0}\" due to event \"{2}\". State is: \"{1}\".", new Object[] { PgID, PgState, Event });
			
			} else if ("log".equals(Command)) {
				
				Log.log(Level.FINE, "AJAX: Log: {0}", DataSet.getDatum("message"));
				
			} else {
				
				GoodCommand = false;
				Log.log(Level.FINE, "AJAX: Unrecognized command: \"{0}\".", Command);
			}
			
			ScriptsWorking |= GoodCommand;
			
		} catch (IOException e) {
			Log.log(Level.SEVERE, "AJAX: Error processing Qwicap AJAX request.", e);
		}

		return true;	//	This data set was created by a Qwicap script.
	}
	
	static boolean processCommand
	(
		final QwicapApplicationServices	AppSrvcs,
		final HttpServletRequest		Req,
		final HttpServletResponse		Resp,
		final FormDataSet				DataSet
	) 
	{
		final String				Command = DataSet.getDatum(kQwicapAJAXCommandControlName);

		if (Command == null)
			return false;	//	This data set was NOT created by a Qwicap script.
		
		try {

			Resp.addHeader("Cache-Control", "no-cache");			//	Prevents these responses from being cached.
			Resp.setContentType("application/xml; charset=UTF-8");	//	Prevents IE from throwing the "Object Expected" error.

			if ("pgstat".equals(Command)) {
				final String			Event   = DataSet.getDatum("event");
				final String			PgID    = DataSet.getDatum("pageid");
				final String			PgState = "unknown";
				final i18nServices		i18n    = AppSrvcs.createi18nManager(Req).getServicesForClass(Qwicap.class);
				final String			Message = i18n.getString("NavigationError-CannotGoToInvalidPage");
				final String			OK      = i18n.getString("NavigationError-OK");
				
				sendPageStateReply(Resp, PgID, PgState, Message, OK, "", kNoPageInfo);
				Log.log(Level.FINE, "AJAX: (No Session.) Requested page state for \"{0}\" due to event \"{2}\". State is: \"{1}\".", new Object[] { PgID, PgState, Event });
			
			} else if ("log".equals(Command)) {
				
				Log.log(Level.FINE, "AJAX: (No Session.) Log: {0}", DataSet.getDatum("message"));
				
			} else
				Log.log(Level.FINE, "AJAX: (No Session.) Unrecognized command: \"{0}\".", Command);
			
		} catch (IOException e) {
			Log.log(Level.SEVERE, "AJAX: (No Session.) Error processing Qwicap AJAX request.", e);
		}

		return true;	//	This data set was created by a Qwicap script.
	}
	
	private static void sendPageStateReply
	(
		final HttpServletResponse	Resp,
		final String				PageID,
		final String				PageState,
		final String				MessageStr,	//	Only used when PageState is "unknown".
		final String				OKLabel,
		final String				ChooseActivePageLabel,
		final PageInfo[]			PagesInfo
	)
	throws IOException
	{
		final StringBuilder		Buff = new StringBuilder(640);
		
		Buff.append("<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n")
		    .append("<qwicap>\n")
			.append("\t<command>pgstat</command>\n")
			.append("\t<pgstat>\n")
			.append("\t\t<page-id>").append(PageID).append("</page-id>\n")
			.append("\t\t<page-state>").append(PageState).append("</page-state>\n");
		
		if ("unknown".equals(PageState)) {
		
			Buff.append("\t\t<message>").append(MutableMarkup.htmlEncode(MessageStr.trim())).append("</message>\n");
			Buff.append("\t\t<ok-label>").append(MutableMarkup.htmlEncode(OKLabel.trim())).append("</ok-label>\n");
		
			if (PagesInfo.length != 0) {
				
				Buff.append("\t\t<pages-label>").append(MutableMarkup.htmlEncode(ChooseActivePageLabel.trim())).append("</pages-label>\n");
				Buff.append("\t\t<pages>\n");
				
				for (final PageInfo Pg : PagesInfo) {
					
					Buff.append("\t\t\t<page>\n")
						.append("\t\t\t\t<id>"   ).append(Pg.ID   ).append("</id>\n")
						.append("\t\t\t\t<title>").append(Pg.Title).append("</title>\n")
						.append("\t\t\t\t<type>" ).append(Pg.Type != null && Pg.Type.shouldPromptModally() ? "modal" : "modeless").append("</type>\n")
						.append("\t\t\t</page>\n");
				}
				
				Buff.append("\t\t</pages>\n");
			}
		}
			
		Buff.append("\t</pgstat>\n")
			.append("</qwicap>");
		
	//	Send the reply.
		
		final Writer			Out = Resp.getWriter();
		
		try {
			Out.write(Buff.toString());
		} finally {
			Out.close();
		}
		
		Log.log(Level.FINE, "AJAX: Sent the following XML reply to the client:\n{0}", Buff);
	}
}
