/*	QWICAP (Quick Web Interface for Conventional Applications) by Chris W. Johnson
	Copyright (C) 2003-2011 The University of Texas at Austin ("U.T. Austin"). All rights reserved.
	
	U.T. Austin's copyright notice in this evolving version reflects our original authorship. It does not assert 
	copyright over changes contributed by others.
	
	This library is free software; you can redistribute it and/or modify it under the terms of the GNU Lesser General 
	Public License as published by the Free Software Foundation; either version 2.1 of the License, or (at your option) 
	any later version.
	
	This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied 
	warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for 
	more details.
	
	You should have received a copy of the GNU Lesser General Public License along with this library; if not, write to 
	the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/


package edu.utexas.its.eis.tools.qwicap.servlet;


import edu.utexas.its.eis.tools.qwicap.template.xml.TagException;

import java.util.logging.Level;
import java.util.logging.Logger;


/**	Subclasses of <code>QwicapAbandonmentException</code> are thrown in response to user abandonment of web pages, 
 *	forms on a web page, or invalid input to a form. Client code should handle all of these exceptions by catching 
 *	only <code>QwicapAbandonmentException</code>, and handling it as described in the Qwicap Prompt Pattern. 
 *	Therefore, the subclasses have package-level access only, so clients can't make the mistake of attempting to handle 
 *	them individually (at least, not without working at it). 
 *	
 *	@author Chris W. Johnson
 */
public abstract class QwicapAbandonmentException extends QwicapException {

	private static final Logger	Log = Logger.getLogger(QwicapAbandonmentException.class.getName());

	private final Qwicap		Q;
	private final Page			PageToRevertTo;

	
	protected QwicapAbandonmentException
	(
		final String			CauseOfDeathMsg
	)
	{
	//	This constructor is only for use by QwicapSessionAbandonedException.
		
		super(CauseOfDeathMsg);
		
		this.Q              = null;
		this.PageToRevertTo = null;
	}
	
	protected QwicapAbandonmentException
	(
		final Page				PageToRevertTo,
		final boolean			EvaluateNewInput
	)
	{
		this.Q              = Qwicap.getCurrentInstance();
		this.PageToRevertTo = setNoPromptPage(PageToRevertTo, EvaluateNewInput);
	}

	protected QwicapAbandonmentException
	(
		final Page				PageToRevertTo,
		final boolean			EvaluateNewInput,
		final String			Message
	)
	{
		super(Message);

		this.Q              = Qwicap.getCurrentInstance();
		this.PageToRevertTo = setNoPromptPage(PageToRevertTo, EvaluateNewInput);
	}

	protected Page setNoPromptPage
	(
		final Page				PageToRevertTo,
		final boolean			EvaluateNewInput
	)
	{
		if (EvaluateNewInput)
			Qwicap.getCurrentInstance().setNoPromptPage(PageToRevertTo);
		
		return PageToRevertTo;
	}

/**	Rethrows the <code>QwicapPageAbandonedException</code> if, and only if, the caller represents one of the pages that 
 *	has been abandoned by the user.
 * 
 * @throws QwicapPageAbandonedException	If the caller representes one of the pages that has been abandoned by the user.
 */
	public void rethrowIfThisPageWasAbandoned() throws QwicapAbandonmentException {
		final SimplifiedStackTrace	CallerTrace = new SimplifiedStackTrace();
		final PageStack				Pages       = Q.getPageStack();
		final Page					CurrentPage = Pages.getByMethod(CallerTrace);

		if (PageToRevertTo.isStillInPlay(CallerTrace)) {

			if (CurrentPage != null) {

				Pages.popTo(CurrentPage);

				if (CurrentPage != PageToRevertTo)	//	Yes, this is meant to be a *reference* equality check. This is not a mistake.
					throw this;
			}

		} else {

			if (CurrentPage != null) {
				try {
					CurrentPage.insertNavigationError(Q.getScriptsWorking(), "Sorry, the page you just used is no longer active. You have been returned to the previous active page.");
				} catch (TagException e) {
					Log.log(Level.INFO, "Failed to insert navigation error message into page.", e);
				}
			}

		//	The "no prompt" page (PageToRevertTo) isn't going to be found, so forget about it.

			Q.setNoPromptPage(null);
		}
	}

/**	Examines an arbitrary <code>Throwable</code> and all of its causes to determine if any of them were of type 
 *	<code>QwicapPageAbandonedException</code>, and, if so, it rethrows the <code>QwicapPageAbandonedException</code> 
 *	if, and only if, the caller represents one of the pages that has been abandoned by the user.
 * 
 * @param T								A <code>Throwable</code> that might be, or have been caused by, a 
 *										<code>QwicapPageAbandonedException</code>.
 * @throws QwicapPageAbandonedException	If the supplied <code>Throwable</code>, or any of its causes, was a 
 *										<code>QwicapPageAbandonedException</code>, and that exception's 
 *										<code>rethrowIfThisPageWasAbandoned</code> method believes that caller's page 
 *										was among those abandoned by the user.
 */
	public static void rethrowIfThisPageWasAbandoned
	(
		Throwable				T
	)
	throws QwicapAbandonmentException
	{
		while (T != null) {

			if (T instanceof QwicapAbandonmentException)
				((QwicapAbandonmentException) T).rethrowIfThisPageWasAbandoned();
			if (T instanceof QwicapSessionDeathException)
				throw (QwicapSessionDeathException) T;
			if (T instanceof QwicapSessionWillNotDieError)
				throw (QwicapSessionWillNotDieError) T;

			T = T.getCause();
		}
	}
}
