/*	QWICAP (Quick Web Interface for Conventional Applications) by Chris W. Johnson
	Copyright (C) 2007-2011 Chris W. Johnson. All rights reserved.
	
	Chris W. Johnson's copyright notice in this evolving version reflects his original authorship. It does not assert 
	copyright over changes contributed by others.
	
	This library is free software; you can redistribute it and/or modify it under the terms of the GNU Lesser General 
	Public License as published by the Free Software Foundation; either version 2.1 of the License, or (at your option) 
	any later version.
	
	This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied 
	warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for 
	more details.
	
	You should have received a copy of the GNU Lesser General Public License along with this library; if not, write to 
	the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/


package edu.utexas.its.eis.tools.qwicap.servlet;


import javax.servlet.ServletContext;
import javax.servlet.http.HttpServletRequest;
import java.util.logging.Logger;
import java.util.logging.Level;
import java.util.Timer;
import java.util.TimerTask;
import java.io.IOException;
import java.lang.reflect.Method;


/**	Encapsulates all of the objects representing services provided to a particular web application. If Qwicap is
 * 	deployed as a shared library, it may be serving multiple web applications. Each web application will have one (and
 * 	only one) instance of <code>QwicapApplicationServices</code>. In contrast, a single instance of 
 * 	{@link QwicapGlobalServices} would be shared by all of the web applications.
 * 
 * 	@author Chris W. Johnson
 */
final class QwicapApplicationServices {
	
	private static final Logger			Log = Logger.getLogger(QwicapApplicationServices.class.getName());
	
	private final String				ServletContextPath;
//	private final String				ServletContextName;
	private final String				ClientClassName;
	private final QwicapThreadPool		Pool;
	private final ServiceDataRecorder	SDR;
	private final DocumentCache			DocCache;
	private final i18nLocaleList		KnownLocales;
	private final i18nDocumentAccessor	i18nDocs;
	private final i18nLocalizationsList	WebAppLocalizations;
	private final Class<?>				ClientClass;
	private final Method				ClientMain;
	private final String[]				ClientMainArgs;
	private final Timer					DeferredTaskTimer;
	
//	In development mode, sensitive information (like exception messages and stack traces) is included in pages sent to 
// 	users. This can be very helpful while developing/debugging a web application, but is also a potential security hole. 
//	Development mode is disabled by default, and can only be enabled by using a context parameter in a web 
//	application's deployment descriptor ("web.xml" file).
	
	private final boolean				DevelopmentMode;
	
	
	QwicapApplicationServices
	(
		final String			ContextPath,
		final int				ThreadPoolThreadsMin,
		final int				ThreadPoolThreadsSpareMax,
		final int				ThreadPoolThreadsMax,
		final int				ThreadPoolThreadStackKB,
		final boolean			DevelopmentMode,
		final MIMETypeGuesser	MIMETyper,
		final ServletContext	SCtxt,				//	Ideally we'd abstract this out, but that'll have to wait.
		final Class<?>			ClientClass,
		final String[]			ClientMainArgs
	)
	throws NoSuchMethodException
	{
		this.ServletContextPath = ContextPath;
//		this.ServletContextName = SCtxt.getServletContextName();
		this.ClientClassName    = ClientClass.getName();
		this.DevelopmentMode    = DevelopmentMode;
		this.ClientClass        = ClientClass;
		this.ClientMain         = ClientClass.getMethod("main", String[].class);
		this.ClientMainArgs     = ClientMainArgs;
		this.DeferredTaskTimer  = new Timer("Qwicap \"" + ContextPath + "\" Deferred Task Timer", true);
		
	//	Setup the service data recorder.
		
		SDR = new ServiceDataRecorder(60 * 60 * 1000, 72);	//	Accumulate stats for one hour periods. Retain 72 of those hour periods.
		
	//	Setup the thread pool.

		Pool = new QwicapThreadPool(
			ServletContextPath,
			ThreadPoolThreadsMin,
			ThreadPoolThreadsSpareMax,
			ThreadPoolThreadsMax,
			ThreadPoolThreadStackKB * 1024,
			SDR
		);
		
	//	Setup internationalized/multi-lingual access to the document cache for GET requests of static content.
			
		try {
			
			KnownLocales        = new i18nLocaleList();
			WebAppLocalizations = new i18nLocalizationsList(KnownLocales, SCtxt);
			DocCache            = new DocumentCache(MIMETyper, new DocumentCensus(ContextPath, SCtxt), WebAppLocalizations.getTranslations().size());
			i18nDocs            = new i18nDocumentAccessor(DocCache, WebAppLocalizations, SCtxt);
		//	i18nMgr             = new i18nManager(ClientClass.getClassLoader(), KnownLocales, i18nDocs, HitCtxt.getRequest().getHeader("Accept-Language"));
			
		} catch (IOException e) {
			throw new RuntimeException(e);
		}
		
		SDR.setDocumentCache(DocCache);
		SDR.setLabel("context \"" + ServletContextPath + "\", executing client class \"" + ClientClassName + '"');
	}
	
	void close() {
		
		try {
			
		//	Destroy the DeferredCloseTimer that cleans-up lingering downloadables.
		
			DeferredTaskTimer.cancel();
			
		//	Destroy all of the threads in the thread pool.
			
			Pool.destroy();
			
		//	Log this servlet shutdown, and its final service data recorder report.
			
			Log.log(Level.INFO, "All Qwicap threads in context \"{0}\", executing client class \"{1}\", have been destroyed. Final status:\n{2}", new Object[] { ServletContextPath, ClientClassName, SDR.getGatheredData(false).toString() });
			
		} catch (Exception e) {
			Log.log(Level.WARNING, "Shutdown of QwicapApplicationServices was interrupted by an exception.", e);
		} catch (NoClassDefFoundError e) {
			Log.log(Level.SEVERE, "Shutdown of QwicapApplicationServices was interrupted by the inability to find a class.", e);
		}
	}
	
	String getServletContextPath() {
		return ServletContextPath;	
	}
	
	TimerTask scheduleTimerTask
	(
		final TimerTask			Task,
		final long				DelayMillis
	)
	{
		DeferredTaskTimer.schedule(Task, DelayMillis);
		return Task;
	}
	
	QwicapThreadPool getThreadPool() {
		return Pool;
	}
	
	i18nLocaleList getAllKnownLocales() {
		return KnownLocales;
	}
	
	i18nDocumentAccessor getDocumentAccessor() {
		return i18nDocs;	
	}
	
	i18nLocalizationsList getLocalizations() {
		return WebAppLocalizations;	
	}
	
	DocumentCache getDocumentCache() {
		return DocCache;
	}

	ServiceDataRecorder getServiceDataRecorder() {
		return SDR;
	}
	
	i18nManager createi18nManager
	(
		final HttpServletRequest Req
	)
	{
		return createi18nManager(Req.getHeader("Accept-Language"));
	}
	
	i18nManager createi18nManager
	(
		final String			AcceptLanguageHeaderStr
	)
	{
	//	i18nManager instances are session-specific, so we provide convenient means of creating them, but we do not
	//	retain references to them.
		
		return new i18nManager(ClientClass.getClassLoader(), KnownLocales, i18nDocs, AcceptLanguageHeaderStr);
	}
	
	boolean isInDevelopmentMode() {
		return DevelopmentMode;
	}
	
	Class<?> getClientClass() {
		return ClientClass;	
	}
	
	Method getClientMain() {
		return ClientMain;
	}
	
	String[] getClientMainArgsCopy() {
		return ClientMainArgs.clone();
	}
}
