/*	QWICAP (Quick Web Interface for Conventional Applications) by Chris W. Johnson
	Copyright (C) 2008 The University of Texas at Austin ("U.T. Austin"). All rights reserved.
	
	U.T. Austin's copyright notice in this evolving version reflects our original authorship. It does not assert 
	copyright over changes contributed by others.
	
	This library is free software; you can redistribute it and/or modify it under the terms of the GNU Lesser General 
	Public License as published by the Free Software Foundation; either version 2.1 of the License, or (at your option) 
	any later version.
	
	This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied 
	warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for 
	more details.
	
	You should have received a copy of the GNU Lesser General Public License along with this library; if not, write to 
	the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/


package edu.utexas.its.eis.tools.qwicap.servlet;


import java.util.ArrayList;
import java.util.logging.Level;
import java.util.logging.Logger;


/**	Manages a particular set of <code>QwicapBlockingListener</code> invocations, including holding any state information
 * 	that they return from their <code>qwicapBlocked</code> methods, and returning that state information to them when 
 * 	their <code>qwicapUnblocked</code> methods are invoked.
 * 
 * 	@author Chris W. Johnson
 */
final class QwicapBlockingListenerStates {

	private static final Logger			Log = Logger.getLogger(QwicapBlockingListenerStates.class.getName());
	
	private final BlockingDescription	BlockInfo;
	private QwicapBlockingListener[]	Listeners;
	private Object[]					ListenerUserObjs;
	private boolean[]					ListenerWorked;
	
	
	static QwicapBlockingListenerStates blocked
	(
		final ArrayList<QwicapBlockingListener>	BlockListeners,
		final BlockingDescription				BlockInfo
	)
	{
		return new QwicapBlockingListenerStates(BlockListeners, BlockInfo);
	}
	
	private QwicapBlockingListenerStates
	(
		final ArrayList<QwicapBlockingListener>	BlockListeners,
		final BlockingDescription				BlockInfo
	)
	{
		this.BlockInfo = BlockInfo;
		
	//	Invoke all of the blocking listeners, and retain any state objects that they may choose to return.

		if (BlockListeners != null) {
			final int				ListenerCount = BlockListeners.size();

		//	Copy the contents of BlockListeners to the Listeners array, to guard against the BlockListeners ArrayList 
		//	being modified by one of the qwicapBlocked methods of the filters.

			Listeners        = BlockListeners.toArray(new QwicapBlockingListener[ListenerCount]);
			ListenerUserObjs = new Object[ListenerCount];
			ListenerWorked   = new boolean[ListenerCount];

			for (int Index = 0; Index < ListenerCount; Index++) {

				try {
					ListenerUserObjs[Index] = Listeners[Index].qwicapBlocked(BlockInfo);
					ListenerWorked[Index]   = true;
				} catch (Exception e) {
					Log.log(Level.WARNING, "The filter method " + Listeners[Index].getClass().getName() + ".qwicapBlocked(" + BlockInfo + ") threw an exception. Qwicap caught it, logged it, and proceeded without interruption.", e);
				}
			}
		}
	}

	void unblocked() {

	//	Invoke all of the blocking listeners, returning to them any state objects that they may have previously
	//	provided.			

		if (Listeners != null) {

			for (int Index = 0, Count = Listeners.length; Index < Count; Index++) {

				if (ListenerWorked[Index]) {

					try {
						Listeners[Index].qwicapUnblocked(BlockInfo, ListenerUserObjs[Index]);
					} catch (Exception e) {
						Log.log(Level.WARNING, "The filter method " + Listeners[Index].getClass().getName() + ".qwicapUnblocked(" + BlockInfo + ") threw an exception. Qwicap caught it, logged it, and proceeded without interruption.", e);
					}

				} else
					Log.log(Level.WARNING, "The filter method " + Listeners[Index].getClass().getName() + ".qwicapUnblocked(" + BlockInfo + ") was not invoked, because the qwicapBlocked method of the same filter instance threw an exception.");
			}
		}
	}
}
