/*	QWICAP (Quick Web Interface for Conventional Applications) by Chris W. Johnson
	Copyright (C) 2003-2008 The University of Texas at Austin ("U.T. Austin"). All rights reserved.
	
	U.T. Austin's copyright notice in this evolving version reflects our original authorship. It does not assert 
	copyright over changes contributed by others.
	
	This library is free software; you can redistribute it and/or modify it under the terms of the GNU Lesser General 
	Public License as published by the Free Software Foundation; either version 2.1 of the License, or (at your option) 
	any later version.
	
	This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied 
	warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for 
	more details.
	
	You should have received a copy of the GNU Lesser General Public License along with this library; if not, write to 
	the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/


package edu.utexas.its.eis.tools.qwicap.servlet;


import edu.utexas.its.eis.tools.qwicap.util.ArrayToString;

import javax.servlet.http.HttpServlet;
import javax.servlet.ServletContext;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Set;
import java.util.logging.Logger;
import java.util.logging.Level;
import java.util.zip.ZipEntry;
import java.util.jar.JarInputStream;


/**
 * Created by IntelliJ IDEA.
 * User: chrisj
 * Date: Feb 23, 2007
 * Time: 10:48:10 PM
 * To change this template use File | Settings | File Templates.
 */
final class QwicapClientClass {

	private static final Logger	Log = Logger.getLogger(QwicapClientClass.class.getName());
	
	private static final String	kApplicationClassNameParam = "QwicapClientClassName";


	static Class<?> get
	(
		final HttpServlet		Srv
	)
	throws QwicapException, ClassNotFoundException
	{
		String					QwicapClientClassName = Srv.getServletConfig().getInitParameter(kApplicationClassNameParam);

		if (QwicapClientClassName == null) {

			QwicapClientClassName = discoverClientClassName(Srv);
			if (QwicapClientClassName == null) {

				throw new QwicapException("The initialization parameter \"" + kApplicationClassNameParam +
					"\" is not present in the current \"web.xml\" file. Without that parameter, Qwicap has no way of " +
					"knowing which class of yours should have its \"main\" method invoked when this web application " +
					"starts. (Qwicap has already exhausted its automatic entry-point detection options.)"
				);
			}
		}

		return Class.forName(QwicapClientClassName);
	}

	private static String discoverClientClassName
	(
		final HttpServlet		Srv
	)
	throws QwicapException
	{
		String					ClientClassName;

	//	Search for possible client classes (classes with "public static void main(String[])" methods).
	//	First, search in the WEB-INF/classes directory hierarchy.
	//	Second, search in the JAR files in the WEB-INF/lib directory.

		ClientClassName = discoverClientClassInClasses(Srv);
		if (ClientClassName == null)
			ClientClassName = discoverClientClassInLib(Srv);

		return ClientClassName;
	}

	private static String discoverClientClassInClasses
	(
		final HttpServlet		Srv
	)
	throws QwicapException
	{
		try {
			final ServletContext	SCtxt = Srv.getServletContext();
			final Set<String>		Paths = SCtxt.getResourcePaths("/WEB-INF/classes/");
			
			if (Paths == null) {
				
				Log.log(Level.FINE, "Failed to get the paths to the contents of this application's \"/WEB-INF/classes/\" subdirectory.");
				return null;
			}
			
			final ArrayList<String>	CandidateClasses = new ArrayList<String>();

			lookForCandidatesInClasses(SCtxt, Paths, "", CandidateClasses);

			return getTheOneCandidateClass(Srv, CandidateClasses);

		} catch (QwicapException e) {
			throw e;
		} catch (Exception e) {
			Log.log(Level.WARNING, "Failed to discover Qwicap client class among the \".class\" files in the \"/WEB-INF/classes/\" subdirectories.", e);
		}

		return null;
	}

	private static void lookForCandidatesInClasses
	(
		final ServletContext	SCtxt,
		final Set<String>		Paths,
		final String			Package,
		final ArrayList<String>	CandidateClasses
	)
	{
		if (Paths == null || Package.startsWith("edu.utexas.its.eis.tools.qwicap"))
			return;
		
		for (final String Path : Paths) {
						
			if (Path.endsWith("/")) {	//	If a path ends with '/', it's a directory.
				final int				LastSlashIndex         = Path.length() - 1;
				final int				SecondToLastSlashIndex = Path.lastIndexOf('/', LastSlashIndex - 1);
				final String			Name;
				
				if (SecondToLastSlashIndex >= 0)
					Name = Path.substring(SecondToLastSlashIndex + 1, LastSlashIndex);
				else
					Name = Path.substring(0, LastSlashIndex);
								
				lookForCandidatesInClasses(SCtxt, SCtxt.getResourcePaths(Path), Package + Name + '.', CandidateClasses);

			} else {					//	This isn't a directory, so assume it is a file.

			//	Skip anything that isn't a class file.

				if (Path.endsWith(".class") == false)
					continue;

				try {
					final ClassFile			CF = new ClassFile(SCtxt.getResource(Path).openStream(), true);

				//	See if the class depends on the Qwicap class.

					if (CF.dependsOnClass("edu/utexas/its/eis/tools/qwicap/servlet/Qwicap")) {

					//	See if the class has a "public static void main(String[])" method.

						if (CF.hasMethod("main", "([Ljava/lang/String;)V", 0x0009))
							CandidateClasses.add(CF.getClassName().replace('/', '.'));
					}

				} catch (IOException e) {
					Log.log(Level.INFO, "Examination of \"" + Path + "\" failed.", e);
				//	Ignore it and move on to the next class file.
				} catch (ClassFileFormatException e) {
				//	Ignore it and move on to the next class file.
				}
			}
		}
	}
	
	private static String discoverClientClassInLib
	(
		final HttpServlet		Srv
	)
	throws QwicapException
	{
		try {
			final ServletContext	SCtxt = Srv.getServletContext();
			final Set<String>		Paths = SCtxt.getResourcePaths("/WEB-INF/lib/");
			
			if (Paths == null) {
				
				Log.log(Level.FINE, "Failed to get the path to this application's \"/WEB-INF/lib/\" subdirectory.");
				return null;
			}

			final ArrayList<String>	CandidateClasses = new ArrayList<String>();
			
			for (final String Path : Paths) {
			
				try {
					final JarInputStream	Jar = new JarInputStream(SCtxt.getResourceAsStream(Path));

					try {

						for (ZipEntry Entry = Jar.getNextEntry(); Entry != null; Entry = Jar.getNextEntry()) {
							final String			Name = Entry.getName();

						//	Skip anything that isn't a class file.

							if (Name.endsWith(".class") == false)
								continue;

						//	Ignore the Qwicap package, and all sub-packages. (We can assume we know the name of the Qwicap JAR file, so we have to do this.)

							if (Name.startsWith("edu/utexas/its/eis/tools/qwicap/"))
								continue;

							try {
								final ClassFile			CF = new ClassFile(Jar, false);

							//	See if the class depends on the Qwicap class.

								if (CF.dependsOnClass("edu/utexas/its/eis/tools/qwicap/servlet/Qwicap")) {

								//	See if the class has a "public static void main(String[])" method.

									if (CF.hasMethod("main", "([Ljava/lang/String;)V", 0x0009))
										CandidateClasses.add(CF.getClassName().replace('/', '.'));
								}

							} catch (ClassFileFormatException e) {
							//	Ignore.
							}
						}

					} finally {
						Jar.close();
					}

				} catch (IOException e) {
				//	We don't care about IOExceptions here. We just want to move along to the next file.
					Log.log(Level.FINE, "Failed to read the resource \"" + Path + "\" as a JAR file.", e);
				}
			}

			return getTheOneCandidateClass(Srv, CandidateClasses);

		} catch (QwicapException e) {
			throw e;
		} catch (Exception e) {
		//	We accept the fact that a lot of things can go wrong in here. We indicate failure by returning null, rather than allowing an exception to propagate.
		}

		return null;
	}

/*	private static String discoverClientClassInClasses
	(
		final HttpServlet		Srv
	)
	throws QwicapException
	{
		try {
			final String			WebInfClasses    = Srv.getServletContext().getRealPath("WEB-INF/classes");
			
			if (WebInfClasses == null) {
				
				Log.log(Level.WARNING, "Failed to get the path to this application's \"WEB-INF/classes\" subdirectory.");
				return null;
			}
			
			final File[]			Files            = new File(WebInfClasses).listFiles();
			final ArrayList<String>	CandidateClasses = new ArrayList<String>();

			lookForCandidatesInClasses(Files, "", CandidateClasses);

			return getTheOneCandidateClass(Srv, CandidateClasses);

		} catch (QwicapException e) {
			throw e;
		} catch (Exception e) {
			Log.log(Level.WARNING, "Failed to discover Qwicap client class among the \".class\" files in the \"WEB-INF/classes\" subdirectories.", e);
		}

		return null;
	}

	private static void lookForCandidatesInClasses
	(
		final File[]			Files,
		final String			Package,
		final ArrayList<String>	CandidateClasses
	)
	{
		if (Files == null || Package.startsWith("edu.utexas.its.eis.tools.qwicap"))
			return;

		for (int Index = 0, Count = Files.length; Index < Count; Index++) {
			final File				F    = Files[Index];
			final String			Name = F.getName();

			if (F.isDirectory()) {

				lookForCandidatesInClasses(F.listFiles(), Package + F.getName() + '.', CandidateClasses);

			} else if (F.isFile()) {

			//	Skip anything that isn't a class file.

				if (Name.endsWith(".class") == false)
					continue;

			//	System.out.println(Package + Name.substring(0, Name.length() - ".class".length()));

				try {
					final ClassFile			CF = new ClassFile(new FileInputStream(F), true);

				//	See if the class depends on the Qwicap class.

					if (CF.dependsOnClass("edu/utexas/its/eis/tools/qwicap/servlet/Qwicap")) {

					//	See if the class has a "public static void main(String[])" method.

						if (CF.hasMethod("main", "([Ljava/lang/String;)V", 0x0009))
							CandidateClasses.add(CF.getClassName().replace('/', '.'));
					}

				} catch (FileNotFoundException e) {
				//	Ignore it and move on to the next class file.
				} catch (ClassFileFormatException e) {
				//	Ignore it and move on to the next class file.
				}
			}
		}
	}

	private static String discoverClientClassInLib
	(
		final HttpServlet		Srv
	)
	throws QwicapException
	{
		try {
			final String			WebInfLib = Srv.getServletContext().getRealPath("WEB-INF/lib");
			
			if (WebInfLib == null) {
				
				Log.log(Level.WARNING, "Failed to get the path to this application's \"WEB-INF/lib\" subdirectory.");
				return null;
			}
			
			final File[]			Libs = new File(WebInfLib).listFiles();

			if (Libs == null)
				return null;

		//	final ClassLoader		Loader = SessionContext.class.getClassLoader();
			final ArrayList<String>	CandidateClasses = new ArrayList<String>();

			for (int FileIndex = 0, FileCount = Libs.length; FileIndex < FileCount; FileIndex++) {
				final File				Lib = Libs[FileIndex];

			//	System.out.println("Examining JAR: " + Lib.getAbsolutePath());

				try {
					final JarFile			Jar = new JarFile(Lib);

					try {
						final Enumeration<JarEntry>	Entries = Jar.entries();

						while (Entries.hasMoreElements()) {
							final ZipEntry			Entry = Entries.nextElement();
							final String			Name  = Entry.getName();

						//	Skip anything that isn't a class file.

							if (Name.endsWith(".class") == false)
								continue;

						//	Ignore the Qwicap package, and all sub-packages. (We can assume we know the name of the Qwicap JAR file, so we have to do this.)

							if (Name.startsWith("edu/utexas/its/eis/tools/qwicap/"))
								continue;

							try {
								final ClassFile			CF = new ClassFile(Jar.getInputStream(Entry), true);

							//	See if the class depends on the Qwicap class.

								if (CF.dependsOnClass("edu/utexas/its/eis/tools/qwicap/servlet/Qwicap")) {

								//	See if the class has a "public static void main(String[])" method.

									if (CF.hasMethod("main", "([Ljava/lang/String;)V", 0x0009))
										CandidateClasses.add(CF.getClassName().replace('/', '.'));
								}

							} catch (ClassFileFormatException e) { 
							//	Ignore.
							}
						}

					} finally {
						Jar.close();
					}

				} catch (IOException e) {
				//	We don't care about IOExceptions here. We just want to move along to the next file.
					Log.log(Level.FINE, "Failed to read the file \"" + Lib.getAbsolutePath() + "\" as a JAR file.", e);
				}
			}

			return getTheOneCandidateClass(Srv, CandidateClasses);

		} catch (QwicapException e) {
			throw e;
		} catch (Exception e) {
		//	We accept the fact that a lot of things can go wrong in here. We indicate failure by returning null, rather than allowing an exception to propagate.
		}

		return null;
	}
*/
	private static String getTheOneCandidateClass
	(
		final HttpServlet		Srv,
		final ArrayList<String>	CandidateClasses	//	List of class names (strings).
	)
	throws QwicapException
	{
		final int				CandidateCount = CandidateClasses.size();

		if (CandidateCount == 1) {

			return CandidateClasses.get(0);

		} else if (CandidateCount > 1) {

			throw new QwicapException("The initialization parameter \"" + kApplicationClassNameParam +
				"\" is not present in our \"web.xml\" file (\"" + Srv.getServletContext().getRealPath("WEB-INF/web.xml") +
				"\"), so Qwicap tried to infer the main class for this web application. Unfortunately, several classes (" +
				new ArrayToString(CandidateClasses).alwaysQuote() + ") were found to contain methods with the signature " +
				"\"public static void main(String[])\", so Qwicap cannot proceed automatically. Please supply the " +
				"initialization parameter \"" + kApplicationClassNameParam + "\" in your \"web.xml\" in " +
				"order to solve this problem."
			);
		}

		return null;
	}
}
