/*	QWICAP (Quick Web Interface for Conventional Applications) by Chris W. Johnson
	Copyright (C) 2007-2010 Chris W. Johnson. All rights reserved.
	
	Chris W. Johnson's copyright notice in this evolving version reflects his original authorship. It does not assert 
	copyright over changes contributed by others.
	
	This library is free software; you can redistribute it and/or modify it under the terms of the GNU Lesser General 
	Public License as published by the Free Software Foundation; either version 2.1 of the License, or (at your option) 
	any later version.
	
	This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied 
	warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for 
	more details.
	
	You should have received a copy of the GNU Lesser General Public License along with this library; if not, write to 
	the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/


package edu.utexas.its.eis.tools.qwicap.servlet;


import edu.utexas.its.eis.tools.table.Table;
import edu.utexas.its.eis.tools.table.lltable.MutableLLTable;

import java.security.Security;
import java.security.Provider;
import java.util.logging.Level;
import java.util.logging.Logger;
import java.util.Collections;
import java.util.WeakHashMap;
import java.util.Map;
import java.lang.ref.WeakReference;


/**	Encapsulates all of the objects representing services common to every web application serviced by a particular copy 
 * 	of the Qwicap library. Ordinarily, a Qwicap-based web application includes its own copy of Qwicap, in which case 
 * 	these global services are used only by that application. However, if Qwicap is deployed as a shared library, these
 * 	global services may be shared by any number of separate web applications (each in its own class loader). In 
 * 	contrast, each web application has its own instance of {@link QwicapApplicationServices}.
 * 
 * 	@author Chris W. Johnson
 */
final class QwicapGlobalServices {

	private static final Logger			Log = Logger.getLogger(QwicapGlobalServices.class.getName());

	private final Rule30Random			Rndm;					//	It's useful for this to be global.
	private final Rule30RandomProvider	RndmProvider;
	private final boolean				RndmIsRegisteredProvider;//	It's useful for this to be global.

	private final Map<ClassLoader, WeakReference<ServiceDataRecorder>>	SDRList;


	QwicapGlobalServices() {

	//	Setup the pseudo-random number generator.

		Rndm         = new Rule30Random();
		RndmProvider = new Rule30RandomProvider(Rndm);
		
	//	Install the service provider that makes our pseudo-random number generator available to the SecureRandom class, 
	//	just in case someone wants to use it in a context that requires a SecureRandom object.

		boolean					IsProvider = false;

		try {
			Security.addProvider(RndmProvider);
			IsProvider = true;
		} catch (SecurityException e) {
			Log.log(Level.WARNING, "The security manager has prevented Qwicap from making its pseudo-random number generator available as a \"{0}\" service. (The error message was: \"{1}\".) If your code does not utilize the \"{2}\" provider''s SecureRandom \"{3}\" algorithm, this will not affect you.", new Object[] {  Rule30RandomProvider.kServiceName, e.getMessage(), Rule30RandomProvider.kProviderName, Rule30RandomProvider.kSecureRandomAlgorithmName });
		} finally {
			RndmIsRegisteredProvider = IsProvider;
		}
		
	//	Log all of the available Security providers, just in case we want to be able to see what is and isn't available.
		
		if (Log.isLoggable(Level.FINE)) {
			final Table				T = new MutableLLTable().setColNames("Service Type", "Service Algorithm", "Provider", "Provider\nVersion");
			
			for (final Provider P : Security.getProviders()) {	
				
				for (final Provider.Service PS : P.getServices())
					T.addRow().addAll(PS.getType(), PS.getAlgorithm(), P.getName(), P.getVersion());
			}
			
			Log.log(Level.FINE, "Available Security Providers:\n{0}", T.sortRowsAscending());
		}

	//	Create a global list of all ServiceDataRecoder instances, indexed by the ClassLoader of the web application
	//	for which each ServiceDataRecorder is recording service data.

		SDRList = Collections.synchronizedMap(new WeakHashMap<ClassLoader, WeakReference<ServiceDataRecorder>>());
	}

	void close() {

		try {

		//	Remove our "Qwicap" SecureRandom provider.
			
			RndmProvider.remove();
			
			if (RndmIsRegisteredProvider) {
				
				try {
					Security.removeProvider(Rule30RandomProvider.kProviderName);
				} catch (SecurityException e) {
					Log.log(Level.WARNING, "The security manager has prevented Qwicap from de-registering its \"" + Rule30RandomProvider.kServiceName + "\" service. This should not happen.", e);
				}
			}
			
		//	Clear the list of ServiceDataRecorder instances.
			
			SDRList.clear();
			
		} catch (Exception e) {
			Log.log(Level.WARNING, "Shutdown of QwicapGlobalServices was interrupted by an exception.", e);
		}
	}

	Rule30Random getRandom() {
		return Rndm;
	}

	ServiceDataRecorder addSDR
	(
		final ClassLoader			WebAppClassLoader,
		final ServiceDataRecorder	SDR
	)
	{
		SDRList.put(WebAppClassLoader, new WeakReference<ServiceDataRecorder>(SDR));
		return SDR;
	}
	
	ServiceDataRecorder getSDR
	(
		final ClassLoader			WebAppClassLoader
	)
	{
		final WeakReference<ServiceDataRecorder>	SDRRef = SDRList.get(WebAppClassLoader);
		
		if (SDRRef == null)
			return null;
		
		return SDRRef.get();	//	Can return null.
	}
	
	void removeSDR
	(
		final ClassLoader			WebAppClassLoader
	)
	{
		SDRList.remove(WebAppClassLoader);
	}
}
