/*	QWICAP (Quick Web Interface for Conventional Applications) by Chris W. Johnson
	Copyright (C) 2003-2011 The University of Texas at Austin ("U.T. Austin"). All rights reserved.
	
	U.T. Austin's copyright notice in this evolving version reflects our original authorship. It does not assert 
	copyright over changes contributed by others.
	
	This library is free software; you can redistribute it and/or modify it under the terms of the GNU Lesser General 
	Public License as published by the Free Software Foundation; either version 2.1 of the License, or (at your option) 
	any later version.
	
	This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied 
	warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for 
	more details.
	
	You should have received a copy of the GNU Lesser General Public License along with this library; if not, write to 
	the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/


//  qwicap.java
//  qwicap
//
//  Created by Chris W. Johnson on Wed Jun 04 2003.


package edu.utexas.its.eis.tools.qwicap.servlet;


import edu.utexas.its.eis.tools.qwicap.template.xml.TagException;
import edu.utexas.its.eis.tools.qwicap.template.xml.mutable.MutableMarkup;
import edu.utexas.its.eis.tools.table.MutableTable;
import edu.utexas.its.eis.tools.table.lltable.MutableLLTable;
import edu.utexas.its.eis.tools.table.readonly.ReadOnlyTable;
import edu.utexas.ots.chrisj.util.Hexadecimal;

import javax.servlet.ServletConfig;
import javax.servlet.ServletContext;
import javax.servlet.ServletException;
import javax.servlet.http.HttpServlet;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import javax.servlet.http.HttpSession;
import java.io.IOException;
import java.io.PrintWriter;
import java.lang.reflect.Method;
import java.net.MalformedURLException;
import java.security.AccessControlException;
import java.util.ArrayList;
import java.util.Enumeration;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.logging.Level;
import java.util.logging.Logger;


/**	<code>QwicapServlet</code> is the servlet on which Qwicap (Quick Web Interface for Conventional Applications) is 
 *	built. Qwicap client code runs in a <code>QwicapThread</code> which interacts with this servlet to get input from 
 *	users and send output to them. The client code never has to concern itself with this, or any other servlet, except 
 *	in as much as it might want to use the {@link #getServletConfig} method to get information about the servlet 
 *	environment in which it is operating.
 *	
 *	@author Chris W. Johnson
 */
public class QwicapServlet extends HttpServlet {

	private static final Logger				Log                      = Logger.getLogger(QwicapServlet.class.getName());

//	static final String						DefaultContentType       = "text/html";
	static final String						DefaultContentType       = "application/xhtml+xml";	//	See <http://www.rfc-editor.org/rfc/rfc3236.txt>.
	static final String						DefaultCharset           = "UTF-8";					//	UTF-8 is the default character set for XML, and Qwicap only deals with XHTML pages, so this is the correct default. 
	static final String						DefaultContentTypeHeader = DefaultContentType + "; charset=" + DefaultCharset;
	
	private static final String				kPreservedPostDataSet    = "qwicap-previous-post-data-set";
	
	private static final String				kNoStaticFilesAuthMsg    = "You are not authorized to retrieve static files from this web application.";

	private static final String				kAuth2SchemeFactoryClassNameParam = "QwicapAuth2SchemeFactoryClassName";
	private static final String				kAuth2SchemeAttrNamePostfix       = "QwicapAuth2Scheme";

	private static final Object				StaticInitSynchObject = new Object();			//	Shared by every QwicapServlet within this class loader.
	private static int						InitedServletCount;								//	Shared by every QwicapServlet within this class loader.
	private static QwicapGlobalServices		GlobalServices;									//	Shared by every QwicapServlet within this class loader.
	private static ClassPreloaderAndHolder	DestroyClasses;									//	Shared by every QwicapServlet within this class loader.
	
	private QwicapApplicationServices		ApplicationServices;							//	Specific to each servlet, and therefore an instance variable.
	private Auth2SchemeFactory				A2SFactory;										//	Specific to each servlet, and therefore an instance variable.
	private Class<?>						ClientClass;									//	Specific to each servlet, and therefore an instance variable.
	private String[]						ClientMainArgs;									//	Specific to each servlet, and therefore an instance variable.
	private String							ServletToString = "Servlet not initialized";	//	Specific to each servlet, and therefore an instance variable.
	private String							ContextPath;									//	Specific to each servlet, and therefore an instance variable.
	
	private final AtomicInteger				ConcurrentExecutionCount = new AtomicInteger();
	

/**	<p>Initializes <code>QwicapServlet</code> by invoking {@link #init(Class, Auth2SchemeFactory)}, and passing 
 * 	<code>null</code> for both parameters. If you wish to change the manner in which Qwicap discovers your web 
 * 	application's {@link Auth2SchemeFactory}, or client class (the class that contains the 
 * 	"<code>public static void main(String[] args)</code>" method where Qwicap should begin executing your web 
 * 	application), you may create a subclass of <code>QwicapServlet</code>, override this method, and pass whatever you 
 * 	wish to the {@link #init(Class, Auth2SchemeFactory)} method. In order to make Qwicap use your subclass, you'll need 
 *  to change the web application deployment descriptor ("web.xml" file) such that the tag 
 * 	"<code>&lt;servlet-class&gt;edu.utexas.its.eis.tools.qwicap.servlet.QwicapServlet&lt;/servlet-class&gt;</code>"
 * 	specifies your subclass.
 * 	</p>
 * 
 * 	<p>Note: Creating your own subclass of <code>QwicapServlet</code> to override <code>init</code> is also useful when
 * 	Qwicap is installed as a shared library, for example in Tomcat's "shared/lib" subdirectory. In such cases, the 
 * 	classes for your web application are typically in a class loader that is a "child" of the shared library class 
 * 	loader. Because class loaders only delegate to their parents, and never to their children, your web application's
 * 	classes cannot be loaded by code in the shared libraries. However, by overriding <code>init()</code> with your own 
 * 	implementation that instantiates your {@link Auth2SchemeFactory} (if any), and supplies a reference to the class 
 * 	that contains your application's entry point (a "<code>public static void main(String[] args)</code>" method), you 
 * 	can avoid those class loader issues.
 * 	</p>
 *  
 * 	@see #init(Class, Auth2SchemeFactory)
 */
	@Override public void init() {
		init(null, null);
	}

/** Initializes the <code>QwicapServlet</code>. This method is invoked by the {@link #init()} method.
 * 
 * @param ClientClassSuggestion		A reference to the class that implements your web application's entry point (a 
 * 									"<code>public static void main(String[] args)</code>" method), or <code>null</code>
 * 									if Qwicap should use the class specified by the web application deployment 
 * 									descriptor ("web.xml" file).
 * @param A2SFactorySuggestion		An instance of a class implementing the {@link Auth2SchemeFactory} interface, or
 * 									<code>null</code> if Qwicap should use the factory class specified by the web 
 * 									application deployment descriptor ("web.xml" file). If <code>null</code> is passed,
 * 									and there is no such specification in the deployment descriptor, Qwicap will 
 * 									perform no authentication, or authorization for your application.
 */
	public final void init
	(
		final Class<?>				ClientClassSuggestion,	//	Can be null.
		final Auth2SchemeFactory	A2SFactorySuggestion	//	Can be null.
	)
	{
		ContextPath = getContextPath();
		
		if (Log.isLoggable(Level.FINEST)) {
			
			try {
				Log.log(Level.FINEST, "Qwicap {0} servlet \"init\" method invoked for context \"{4}\". (InitCount: {1}) System properties are as follows:\n{2}\n\nClass loaders are as follows:\n\n{3}", new Object[] { Qwicap.getVersionString(), InitedServletCount, new MutableLLTable(System.getProperties()).sortRowsAscending().toString(), getClass().getClassLoader().toString(), ContextPath } );
			} catch (AccessControlException e) {
				Log.log(Level.FINEST, "Qwicap {0} servlet \"init\" method invoked for context \"{3}\". (InitCount: {1}) Class loaders are as follows:\n\n{2}", new Object[] { Qwicap.getVersionString(), InitedServletCount, getClass().getClassLoader().toString(), ContextPath } );
			}
			
		} else
			Log.log(Level.FINE, "Qwicap {0} servlet \"init\" method invoked for context \"{1}\".", new Object[] { Qwicap.getVersionString(), ContextPath });
		
		synchronized (StaticInitSynchObject) {

			if (InitedServletCount++ == 0) {

			//	Setup objects and services that are shared by all servlets in this class loader. (Things that
			//	are not in any way specific to a particular servlet.) Ordinarily, there's only one servlet in a Qwicap
			//	web application, and that web application has its own, private class loader. However, if Qwicap is 
			//	installed in the application server (e.g. Tomcat) as a shared library, these GlobalServices may be 
			//	shared by many, otherwise separate, web applications.
				
				GlobalServices = new QwicapGlobalServices();
		
			//	In Tomcat 6.0.14 and 6.0.19 (and possibly many others), it has been observed that the re-deployment of 
			//	a web application (but, oddly enough, not the shutdown of the Tomcat server) can trigger 
			//	NoClassDefFoundErrors during the execution of the servlet's "destroy" method. This appears to apply 
			//	only to classes that haven't been loaded prior to the execution of "destroy", and implies that the web 
			//	application's class loader has been instructed to suspend class loading prior to the time "destroy" is 
			//	invoked.
			//
			//	A means to change this behavior of the web application's class loader has not been identified. One 
			//	workaround, however, is to force loading of the problematic classes before "destroy" is invoked. We do 
			//	that here, hoping that we will always pre-load enough classes to prevent the NoClassDefFoundError from 
			//	manifesting under all circumstances in which the "destroy" method may execute. (Further, we have to 
			//	hope that the classes pre-loaded here cannot be garbage-collected before "destroy" is invoked. Of
			//	course, hanging onto references to those classes should preclude the garbage-collection problem.)
				
				DestroyClasses = new ClassPreloaderAndHolder(this, 
					"edu.utexas.its.eis.tools.qwicap.servlet.",
					"edu.utexas.its.eis.tools.qwicap.util.",
					"edu.utexas.its.eis.tools.table."
				);
			}
		}
		
	//	Determine the identity of the client class for which this servlet front-ends.

		ClientClass = ClientClassSuggestion;
		if (ClientClass == null) {

			try {
				ClientClass = QwicapClientClass.get(this);
			} catch (Exception e) {
				throw new RuntimeException("Qwicap " + Qwicap.getVersionString() + " encountered an error while trying to determine the identity of the class in the \"" + ContextPath + "\" web application that contains the application's entry point.", e);
			}
		}
		
	//	Get the array of strings that we'll pass to the ClientClass' "public static void main(String[] Args)" method.
		
		if (ClientMainArgs == null)
			ClientMainArgs = getContextAndInitParameters(ContextPath);
		
	//	Register the client's ClassLoader with the experimental ClassLoaderMonitor.
		
		try {
			final Class<?>			CLMClass    = Class.forName("edu.utexas.its.eis.tools.clm.ClassLoaderMonitor");	
			final Method			GetInstance = CLMClass.getMethod("getInstance", (Class<?>[]) null);
			final Method			Register    = CLMClass.getMethod("register", String.class, ClassLoader.class);
			final Object			CLM         = GetInstance.invoke(null, (Object[]) null);
			
			Register.invoke(CLM, ContextPath, ClientClass.getClassLoader());
			
			Log.info("Registered \"" + ContextPath + "\" with ClassLoaderMonitor.");
			
		} catch (Exception e) {
			Log.log(Level.FINE, "Unable to register with the ClassLoaderMonitor. (It may not be installed.)", e);
		}
		
	//	Setup the standard application services. These are shared by all sessions associated with a particular servlet.
		
		final boolean			DevelopmentMode;
		
		{
		//	This block of code deals with a backwards-compatibility issue between some test builds of Qwicap 1.4b13. 
		//	We should be able to eliminate support for the "qwicap-report-exceptions-to-users" context parameter 
		//	in future betas.
		
			final boolean			HasDevelopmentModeParam         = hasContextParameter("qwicap-development-mode");
			final boolean			HasReportExceptionsToUsersParam = hasContextParameter("qwicap-report-exceptions-to-users");
			
			if (HasDevelopmentModeParam)
				DevelopmentMode = getContextParameter(Level.INFO, "qwicap-development-mode",           false);
			else if (HasReportExceptionsToUsersParam)
				DevelopmentMode = getContextParameter(Level.INFO, "qwicap-report-exceptions-to-users", false);
			else
				DevelopmentMode = false;
			
			if (HasReportExceptionsToUsersParam)
				Log.log(Level.WARNING, "The \"web.xml\" file of the web application \"{0}\" contains the deprecated context parameter \"qwicap-report-exceptions-to-users\". Please use the \"qwicap-development-mode\" context parameter instead.", ContextPath);
			
			if (DevelopmentMode)
				Log.log(Level.WARNING, "Qwicap has been instructed to run web application \"{0}\" in the low-security \"development\" mode. Disable that mode before making this web application publicly accessible.", ContextPath);
		}
		
		try {
			
			ApplicationServices = new QwicapApplicationServices(
				ContextPath,
				getContextParameter(Level.INFO, "qwicap-threads-minimum",          0),
				getContextParameter(Level.INFO, "qwicap-threads-spare-maximum",    0),
				getContextParameter(Level.INFO, "qwicap-threads-maximum",       1000),
				getContextParameter(Level.INFO, "qwicap-thread-stack-kilobytes",   0),
				DevelopmentMode,
				new MIMETypeGuesser(this),
				getServletContext(),
				ClientClass,
				ClientMainArgs.clone()
			);
		
		} catch (NoSuchMethodException e) {
			throw new RuntimeException(e);
		}
		
	//	Register the ServiceDataRecorder in ApplicationServices with the global list of ServiceDataRecorder instances.
		
		GlobalServices.addSDR(ClientClass.getClassLoader(), ApplicationServices.getServiceDataRecorder());
		
	//	Get the authentication & authorization factory associated with this servlet.

		A2SFactory = A2SFactorySuggestion;
		if (A2SFactory == null) {
			final String			Auth2SchemeFactoryClassName = getServletConfig().getInitParameter(kAuth2SchemeFactoryClassNameParam);

			if (Auth2SchemeFactoryClassName != null) {

				try {
					A2SFactory = (Auth2SchemeFactory) Class.forName(Auth2SchemeFactoryClassName).newInstance();
				} catch (ClassNotFoundException e) {
					throw new RuntimeException("Qwicap " + Qwicap.getVersionString() + " could not locate the authentication & authorization factory that the \"" + ContextPath + "\" web application says it requires: " + Auth2SchemeFactoryClassName, e);
				} catch (Exception e) {
					throw new RuntimeException("Qwicap " + Qwicap.getVersionString() + " could not create an instance of the authentication & authorization factory that the \"" + ContextPath + "\" web application says it requires: " + Auth2SchemeFactoryClassName, e);
				}

			} else
				A2SFactory = new Auth2SchemeDefault();
		}
		
	//	Create the string that this servlet will return when its "toString" method is invoked.
		
		ServletToString = "QwicapServlet, version " + Qwicap.getVersionString() + ", executing client class \"" + ClientClass.getName() + "\" at context path \"" + ContextPath + '"';
		
	//	Initialization is complete.
		
		Log.log(Level.INFO, "Qwicap {0} is in use by context \"{1}\", executing client class \"{2}\". (InitCount: {3})", new Object[] { Qwicap.getVersionString(), ContextPath /* getServletContext().getServletContextName() */, ClientClass.getName(), InitedServletCount });
	}

	@Override public void destroy() {
		
		Log.log(Level.INFO, "Qwicap {0} context \"{1}\", executing client class \"{2}\", is about to be destroyed. This servlet is {3,choice,0#not processing any requests|1#processing one request|1<processing {3} concurrent requests}. (InitCount: {4})", new Object[] { Qwicap.getVersionString(), ContextPath, ClientClass.getName(), ConcurrentExecutionCount, InitedServletCount });
		
		try {
			
			ApplicationServices.close();
			ApplicationServices = null;
			
			GlobalServices.removeSDR(ClientClass.getClassLoader());
			
			synchronized (StaticInitSynchObject) {

			//	Do not close the global services unless there are no servlets remaining that use the QwicapServlet 
			//	class in this class loader. Ordinarily, only one servlet uses the QwicapServlet class in any given
			//	class loader, but when Qwicap is installed as a shared library, many servlets may be sharing the
			//	same QwicapServlet class, and therefore the same global services.

				try {

					if (InitedServletCount == 1) {
						
						GlobalServices.close();
						GlobalServices = null;
					}
					
				} finally {
					--InitedServletCount;
				}
			}

	//	The following catch block will only serve to catch RuntimeExceptions (the enclosed code throws no declared
	//	exceptions). It is here to ensure that nothing interferes with the application server's operation.

		} catch (Exception e) {

			Log.log(Level.WARNING, "The execution of the \"destroy\" method for the \"" + ContextPath + "\" web application was interrupted by an exception.", e);
			
		} catch (Throwable e) {
			
			Log.log(Level.SEVERE, "Something very fundamental has interfered with servlet destruction.", e);
			
		} finally {

			Log.log(Level.INFO, "Qwicap {0} context \"{1}\", executing client class \"{2}\", has been destroyed. This servlet is {3,choice,0#not processing any requests|1#is processing one request|1<processing {3} concurrent requests}. (InitCount: {4})", new Object[] { Qwicap.getVersionString(), ContextPath, ClientClass.getName(), ConcurrentExecutionCount, InitedServletCount });
		
			super.destroy();
		}
	}

//	todo Can we provide an implementation of getLastModified?

	@Override protected void doHead
	(
		final HttpServletRequest	Req,
		final HttpServletResponse	Resp
	)
	throws ServletException, IOException
	{
		ConcurrentExecutionCount.getAndIncrement();
		
		try {
			
			logRequest(Req);
			super.doHead(Req, Resp);
			
		} finally {
			ConcurrentExecutionCount.getAndDecrement();
		}
	}
	
	@Override protected void doGet
	(
		final HttpServletRequest	Req,
		final HttpServletResponse	Resp
	)
	throws ServletException, IOException
	{
		ConcurrentExecutionCount.getAndIncrement();
		
		try {
			logRequest(Req);

		//	Get the form data set (if any).
			
			FormDataSet				FDS = null;
			
		//	If this GET was produced by a previous POST operation redirecting to a "clean" URL for this web application,
		//	there will be a session attribute containing the form data set that was supplied with that POST operation.
		//	If there is such an attribute, use the data set it contains. Otherwise, instantiate a new FormDataSet object
		//	to represent the data set (if any) that was supplied as the query string of this GET operation.
			
			if (Req.getQueryString() == null) {
				final HttpSession		Sess = Req.getSession(false);
				
				if (Sess != null) {
					final String			DataSetAttrName = SessionContext.createSessionAttributeName(Req.getContextPath(), kPreservedPostDataSet);
					
					FDS = (FormDataSet) Sess.getAttribute(DataSetAttrName);
					Sess.removeAttribute(DataSetAttrName);
				}
			}
			
			if (FDS == null)
				FDS = new FormDataSet(Req);
	
		//	Authenticate and authorize.
	
			final Auth2SchemeReply		A2SReply = authenticateAndAuthorize(Req, Resp, FDS);
	
			if (A2SReply.qwicapShouldDiscardThisSession())
				return;		//	Do nothing - the authenticateAndAuthorize method has already done all of the important work.
	
			if (A2SReply.qwicapCanInvokeWebApplication() == false
			&&  A2SReply.qwicapCanSendStaticFiles()      == false) {
	
			//	The following sendError invocation screws-up attempts by the Auth2Scheme to do things like redirect 
			//	users to a login page. Therefore, the best thing we can do here is nothing. We make it the responsib-
			//	ility of the Auth2Scheme to send errors, or whatever, when it is refusing all access to the user.	
			//	
			//	Resp.sendError(HttpServletResponse.SC_UNAUTHORIZED, "You are not authenticated and/or authorized to use this web application.");
				
				return;	//	If the user isn't allowed to do anything with this application, return.
			}
	
		//	If the request URI contains the string "/qwicapdl/", which is a reserved phrase (like all identifiers 
		//	beginning with "qwicap"), we interpret it as a request for a dynamically-generated download.
			
			final String			RequestURI = Req.getRequestURI();
			
			if (Downloadable.isDownloadableURI(RequestURI)) {
	
				if (A2SReply.qwicapCanInvokeWebApplication()) {
	
					SessionContext.sendDownloadable(Req, Resp, RequestURI);
					return;
				}
			}
			
			final String			ApplicationPath = getApplicationPath(Req);
			
			if (RequestURI.equals(ApplicationPath) == false) {
				final String			ServletPath = Req.getServletPath();
	
			//	If we're being asked for a static file (including resources like "qwicap.css", which Qwicap serves-out as 
			//	if they were static files), then transmit that file to the client, and return.
			
				if ("/".equals(ServletPath) == false) {
					
					if (ServletPath.endsWith("/") == false) {
					
					//	We can't use a RequestDispatcher to answer this request because the servlet path is probably so 
					//	all-encompassing that whatever we'd send to it would cause a dispatch back to this servlet, thereby
					//	creating an infinite loop.
	
						if (A2SReply.qwicapCanSendStaticFiles()) {
							
							if (sendURLReferent(Req, Resp, ServletPath))
								return;
							
						} else {
							
							Resp.sendError(HttpServletResponse.SC_UNAUTHORIZED, kNoStaticFilesAuthMsg);
							return;
						}
					}
	
				//	If we reach this point, we know that we're not dealing with a valid request for a static file or 
				//	resource, so redirect to the web application's unadulterated URL to get the client back on track.
				//	UNLESS, that is, redirecting would lead to a perpetual redirect loop, in which case we drop through
				//	and process the hit, even though it may be rubbish. (The worst that can happen is the hit will be 
				//	ruled invalid by Qwicap, and the application's current page will be retransmitted to the client.)
					
					Log.log(Level.FINE, "Request URI <{0}> is invalid. Redirecting to <{1}>.", new Object[] { Req.getRequestURI(), ApplicationPath });
					
					Resp.sendRedirect(ApplicationPath);
					return;
				}
			}
			
		//	We were NOT asked for a static file. Pass this hit along to the web application.
			
		//	In the absence of a valid session and/or SessionContext, provide properly constructed, but otherwise 
		//	uninformative, responses to AJAX commands. (Failure to intercept them before they are processed by the 
		//	doQwicap method, can result in a new session being created.)
			
			if (FDS.has(QwicapAJAX.kQwicapAJAXCommandControlName) && (Req.isRequestedSessionIdValid() == false || SessionContext.get(Req) == null)) {
				
				QwicapAJAX.processCommand(ApplicationServices, Req, Resp, FDS);
				return;
			}
			
		//	Process this hit.
			
			doQWICAp(Req, FDS, Resp, A2SReply);
			
		} finally {
			ConcurrentExecutionCount.getAndDecrement();
		}
	}
	
	private static String getApplicationPath
	(
		final HttpServletRequest	Req
	) 
	{
		final String			ContextPath = Req.getContextPath();
		
		if (ContextPath.endsWith("/"))
			return ContextPath;
		
		return ContextPath + '/';
	}
	
	private boolean sendURLReferent
	(
		final HttpServletRequest	Req,
		final HttpServletResponse	Resp,
		final String				ServletPath
	)
	throws IOException
	{
		Log.log(Level.FINE, "Document \"{0}\" has been requested.", ServletPath);
		
		if (SessionContext.sendDocument(Req, Resp, ServletPath)) {
			
			Log.log(Level.FINE, "Document \"{0}\" has been sent.", ServletPath);
			return true;
		}
		
	//	When there is no session, we still need to send language/locale-appropriate documents, when available.
		
		final DocumentSource	Doc = ApplicationServices.getDocumentAccessor().getDocument(i18nLocaleSettings.create(ApplicationServices.getAllKnownLocales(), ApplicationServices.getLocalizations(), new HTTPAcceptLanguageList(Req.getHeader("Accept-Language"))), ServletPath);
		
		if (Doc.exists()) {
			
			Doc.writeTo(Resp);
			Log.log(Level.FINE, "Document \"{0}\" has been sent.", ServletPath);

			return true;
		}
		
		Log.log(Level.FINE, "Document \"{0}\" was not found.", ServletPath);

		return false;
	}

	@Override protected void doPost
	(
		final HttpServletRequest	Req,
		final HttpServletResponse	Resp
	)
	throws ServletException, IOException
	{
		ConcurrentExecutionCount.getAndIncrement();
		
		try {
			
			if (Req.getQueryString() != null) {
			
			//	There should be no query string in the URL used by a POST operation. If there is, get the client back to
			//	using a clean URL (one without a query string) by returning a redirect.
				
			//	Try not to loose the form data set that arrived with this POST operation.
				
				final HttpSession		Sess = Req.getSession(false);
				
				if (Sess != null)
					Sess.setAttribute(SessionContext.createSessionAttributeName(Req.getContextPath(), kPreservedPostDataSet), new FormDataSet(Req));
				
			//	Redirect to the "clean" URL for this web application. This will result in a GET operation being
			//	initiated by the client.
				
				Resp.sendRedirect(Req.getContextPath());
				return;
			}
			
			logRequest(Req);
			
			final FormDataSet			FDS = new FormDataSet(Req);
			
			doQWICAp(Req, FDS, Resp, authenticateAndAuthorize(Req, Resp, FDS));
			
		} finally {
			ConcurrentExecutionCount.getAndDecrement();
		}
	}
	
	private void doQWICAp
	(
		final HttpServletRequest	Req,
		final FormDataSet			FDS,
		final HttpServletResponse	Resp,
		final Auth2SchemeReply		A2SReply
	)
	throws ServletException
	{
		Log.finest("SERVLET: Hit received.");

		try {

			if (A2SReply.qwicapShouldDiscardThisSession()) {
				
			//	Do nothing - the authenticateAndAuthorize method has already done all of the important work.
				
			} else if (A2SReply.qwicapCanInvokeWebApplication()) {

				Resp.setContentType(DefaultContentTypeHeader);
				
				final HitContext		Hit  = new HitContext(this, Req, FDS, Resp, A2SReply);
				SessionContext			Ctxt = SessionContext.get(Req);

				if (Ctxt == null) {
					
				//	If the data set specifies a Qwicap page ID, but there is currently no session, don't process this
				//	data set, and certainly do not create a new session.
					
					if (FDS.has(QwicapAJAX.kQwicapAJAXCommandControlName) == false && FDS.has(Qwicap.kQwicapPageIDControlName)) {
						
						Log.fine("Showing application restart page, because there is no session context, but a Qwicap page ID was included in the data set.");
						
						showRestartPage(Req, Resp);
						return;
					}
					
				//	If there's an old HttpSession from a previous run of this web application still present, invalidate 
				//	it, so we can start-off with a clean slate.
					
					{
						final HttpSession		LegacySession = Req.getSession(false);
						
						if (LegacySession != null)
							LegacySession.invalidate();
					}
					
				//	Create the new SessionContext, which creates our new HttpSession automatically.
					
					Ctxt = new SessionContext(GlobalServices, ApplicationServices, Hit);	//	Note that this hit is not processed by the SessionContext constructor, but some of the information it contains is required by the constructor, or underlying code.	
				}

				Log.finest("SERVLET: Invoking GO.");

				Ctxt.go(Hit);

				Log.finest("SERVLET: GO complete.");

			} else {
				
				if (Resp.isCommitted() == false)
					Resp.sendError(HttpServletResponse.SC_UNAUTHORIZED, "You are not authenticated and/or authorized to execute this web application.");
			}
			
		} catch (Exception e) {

			Log.log(Level.INFO, "SERVLET: Qwicap servlet encountered an exception.", e);
			if (e instanceof ServletException)
				throw (ServletException) e;
			else
				throw new ServletException(e);

		} finally {

			Log.finest("SERVLET: Hit complete.\n\n");
		}
	}
	
	private void showRestartPage
	(
		final HttpServletRequest	Req,
		final HttpServletResponse	Resp
	)
	throws IOException, TagException
	{
		final DocumentSource	Doc = ApplicationServices.getDocumentAccessor().getDocument(i18nLocaleSettings.create(ApplicationServices.getAllKnownLocales(), ApplicationServices.getLocalizations(), new HTTPAcceptLanguageList(Req.getHeader("Accept-Language"))), "/qwicap-restart.html");
		
		if (Doc.exists()) {
			final MutableMarkup		Page = Doc.getXML().getMutable();
			
			Page.get("span.web-app-name").setContent(getWebApplicationName(Req));
			Page.get("head title"       ).setContent(Page.get("h1").getText());
		//	Page.get("form"             ).setAttribute("action", Req.getRequestURL());	//	If we really need to do this, use a method like Form.actionRefersToThisWebApp to determine whether or not each action URL should be changed. (Action URLs that do NOT refer to this web app should be left alone.)
			
			Resp.setContentType("text/html;charset=UTF-8");
			
			final PrintWriter		Out = Resp.getWriter();
			
			Page.write(Out);
			Out.close();
			
		} else
			throw new IllegalStateException("Cannot locate the document \"qwicap-restart.html\", which should be built into Qwicap.");
	}
	
	void showExitPage
	(
		final HttpServletRequest	Req,
		final HttpServletResponse	Resp
	)
	{
		if (Resp.isCommitted()) {
			
			Log.severe("Could not show the \"web application quit\" page, because the response has already been committed.");
			return;
		}
		
		try {
			final DocumentSource	Doc = ApplicationServices.getDocumentAccessor().getDocument(i18nLocaleSettings.create(ApplicationServices.getAllKnownLocales(), ApplicationServices.getLocalizations(), new HTTPAcceptLanguageList(Req.getHeader("Accept-Language"))), "/qwicap-exit.html");
			
			if (Doc.exists()) {
				final MutableMarkup		Page = Doc.getXML().getMutable();
				
				Page.get("span.web-app-name").setContent(getWebApplicationName(Req));
				Page.get("head title"       ).setContent(Page.get("h1").getText());
			//	Page.get("form"             ).setAttribute("action", Req.getRequestURL());	//	If we really need to do this, use a method like Form.actionRefersToThisWebApp to determine whether or not each action URL should be changed. (Action URLs that do NOT refer to this web app should be left alone.)
				
				Resp.setContentType("text/html;charset=UTF-8");
				
				final PrintWriter		Out = Resp.getWriter();
				
				Page.write(Out);
				Out.close();
				
			} else
				throw new IllegalStateException("Cannot locate the document \"qwicap-exit.html\", which should be built into Qwicap.");
			
		} catch (Exception e) {
			Log.log(Level.SEVERE, "Could not show the \"web application quit\" page, due to an exception.", e);
		}
	}
	
	void showCrashPage
	(
		final HttpServletRequest	Req,
		final HttpServletResponse	Resp,
		final IncidentThrowable		Incident
	)
	{
		if (Resp.isCommitted()) {
			
			Log.severe("Could not show the \"web application crashed\" page, because the response has already been committed.");
			return;
		}
		
		try {
			final DocumentSource	Doc = ApplicationServices.getDocumentAccessor().getDocument(i18nLocaleSettings.create(ApplicationServices.getAllKnownLocales(), ApplicationServices.getLocalizations(), new HTTPAcceptLanguageList(Req.getHeader("Accept-Language"))), "/qwicap-crash.html");
			
			if (Doc.exists()) {
				final MutableMarkup		Page = Doc.getXML().getMutable();
				
				Page.get("span.web-app-name").setContent(getWebApplicationName(Req));
				Page.get("head title"       ).setContent(Page.get("h1").getText());
			//	Page.get("form"             ).setAttribute("action", Req.getRequestURL());	//	If we really need to do this, use a method like Form.actionRefersToThisWebApp to determine whether or not each action URL should be changed. (Action URLs that do NOT refer to this web app should be left alone.)
				Page.get("span." + Qwicap.kQwicapIncidentCodeClass).setContent(Incident.getIncidentCode());
				
				if (Incident.getDebugInformation() != null)
					Page.get(".show-exception span.exception").setContent(Incident.getDebugInformation());
				else
					Page.get(".show-exception").delete();
				
				Resp.setContentType("text/html;charset=UTF-8");
				
				final PrintWriter		Out = Resp.getWriter();
				
				Page.write(Out);
				Out.close();
				
			} else
				throw new IllegalStateException("Cannot locate the document \"qwicap-crash.html\", which should be built into Qwicap.");
		
		} catch (Exception e) {
			Log.log(Level.SEVERE, "Could not show the \"web application crashed\" page, due to an exception.", e);
		}
	}
	
	private String getWebApplicationName
	(
		final HttpServletRequest	Req
	) 
	{
		String					WebAppName = getServletContext().getServletContextName();
		
		if (Qwicap.kQwicapWebAppDefaultName.equalsIgnoreCase(WebAppName)) {
			
			WebAppName = Req.getContextPath();
			if (WebAppName.startsWith("/"))
				WebAppName = WebAppName.substring(1);
		}
		
		return WebAppName;
	}

	private Auth2SchemeReply authenticateAndAuthorize
	(
		final HttpServletRequest	Req,
		final HttpServletResponse	Resp,
		final FormDataSet			FDS
	)
	throws ServletException, IOException
	{
		final boolean				CanLogAtFinest = Log.isLoggable(Level.FINEST);
		
		if (CanLogAtFinest)
			Log.log(Level.FINEST, "qwicap-auth2scheme-discards = {0}", Req.getAttribute("qwicap-auth2scheme-discards"));
		
		try {
			final Auth2SchemeReply	A2SReply;
			final String			A2SAttrName = SessionContext.createSessionAttributeName(Req.getContextPath(), kAuth2SchemeAttrNamePostfix);
			HttpSession				Sess        = Req.getSession(false);	//	Get the session, if any. Do not create a new session.
			Auth2Scheme				A2S         = null;

			if (Sess != null)
				A2S = (Auth2Scheme) Sess.getAttribute(A2SAttrName);

			final boolean			ObtainedFromSession;

			if (A2S == null) {

				A2S = A2SFactory.getAuth2SchemeInstance();
				ObtainedFromSession = false;

			} else
				ObtainedFromSession = true;

			A2SReply = A2S.qwicapAuth2SchemeHit(this, Req, Resp, FDS);
			
			if (CanLogAtFinest)
				Log.log(Level.FINEST, "The Auth2Scheme implemented by {0} produced the following reply (0x{1}):\n{2}", new Object[] { A2S.getClass(), Hexadecimal.fromInt(A2SReply.hashCode()), A2SReply });
			
			if (A2SReply.qwicapShouldDiscardThisSession()) {
			
				if (Sess != null) {
					
					if (CanLogAtFinest)
						Log.log(Level.FINEST, "The Auth2Scheme implemented by {0} commanded that the existing session (0x{1}) be discarded.", new Object[] { A2S.getClass(), Hexadecimal.fromInt(Sess.hashCode()) } );
					
					final SessionContext	Ctxt = SessionContext.get(Req);
					
					if (Ctxt != null) {
						
						Ctxt.getQwicap().pleaseDie("The Auth2Scheme commanded that this session be discontinued.");
						Ctxt.close(true);
						
					} else if (Sess != null) {
						
						Log.fine("Invalidating session.");
						Sess.invalidate();	
					}
				}
				
				if (Resp.isCommitted() == false) {
				
				//	The following "if" statement is a sanity-check, intended only to prevent infinite-loops in the 
				//	presence of bugs in this class. At this point, it shouldn't be necessary any longer, but better 
				//	safe than sorry.
					
					if (postIncrementRequestAttribute(Req, "qwicap-auth2scheme-discards") < 2) {
						final String			Method = Req.getMethod();
						
						if (CanLogAtFinest)
							Log.log(Level.FINEST, "QwicapServlet will now attempt to process the current request as the first hit of a new session. To accomplish this, QwicapServlet will reinvoke its \"{0}\" handler.", Method);
						
						if ("POST".equalsIgnoreCase(Method))
							doPost(Req, Resp);
						else if ("GET".equalsIgnoreCase(Method))
							doGet(Req, Resp);
					//	else if ("PUT".equalsIgnoreCase(Method))
					//		doPut(Req, Resp);
					} else
						Resp.sendError(HttpServletResponse.SC_UNAUTHORIZED, "The authentication and authorization system commanded that this session be discontinued.");
				
				} else if (CanLogAtFinest)
					Log.finest("QwicapServlet would have attempted to process the current request as the first hit of a new session, had the response not already been committed.");

			} else if (A2S.qwicapShouldRetainThisAuth2SchemeInstance()) {

				if (!ObtainedFromSession) {

				//	Create a new session, if necessary.

					if (Sess == null)
						Sess = Req.getSession(true);	//	Now we create a new session.

				//	Store the Auth2Scheme instance in the session attribute table.
				
					Sess.setAttribute(A2SAttrName, A2S);
					
					if (CanLogAtFinest)
						Log.log(Level.FINEST, "Added Auth2Scheme instance 0x{0} to session 0x{1}.", new Object[] { Hexadecimal.fromInt(A2S.hashCode()), Hexadecimal.fromInt(Sess.hashCode()) });
				}

			} else {

			//	The Auth2Scheme instance was stored in our session attribute table, but it has just told us NOT to 
			//	retain it any longer, so we remove it from the session data.

				if (ObtainedFromSession) {
					
					Sess.removeAttribute(A2SAttrName);
					
					if (CanLogAtFinest)
						Log.log(Level.FINEST, "Removed Auth2Scheme instance 0x{0} from session 0x{1}.", new Object[] { Hexadecimal.fromInt(A2S.hashCode()), Hexadecimal.fromInt(Sess.hashCode()) });
				}
			}
			
			return A2SReply;

		} catch (IOException e) {
			throw e;	//	There is no need to encapsulate this type of exception.
		} catch (ServletException e) {
			throw e;	//	There is no need to encapsulate this type of exception.
		} catch (RuntimeException e) {
			throw e;	//	There is no need to encapsulate this type of exception.
		} catch (Exception e) {
			throw new ServletException(e);
		}
	}
	
	private static int postIncrementRequestAttribute
	(
		final HttpServletRequest	Req,
		final String				AttrName
	)
	{
		final Object			A2SInvocationsObj = Req.getAttribute(AttrName);
		int						Value = 0;
		
		if (A2SInvocationsObj != null)
			Value = (Integer) A2SInvocationsObj;
			
		Req.setAttribute(AttrName, Value + 1);
		
		return Value;
	}
	
	private boolean hasContextParameter
	(
		final String			Name
	)
	{
		return getServletContext().getInitParameter(Name) != null;
	}

	private int getContextParameter
	(
		final Level				LogLevel,
		final String			Name,
		final int				Default
	)
	{
		final String			Value = getServletContext().getInitParameter(Name);

		if (Value == null) {

			Log.log(LogLevel, "The web application deployment descriptor''s context parameter \"{0}\" is missing. The default value, \"{1}\", will be used instead.", new Object[] { Name, Default });
			return Default;
		}

		try {
			return Integer.parseInt(Value);
		} catch (NumberFormatException e) {

			Log.log(LogLevel, "The web application deployment descriptor''s context parameter \"{0}\" contained the value \"{1}\", which is not a valid integer. The default value, \"{2}\", will be used instead.", new Object[] { Name, Value, Default });
			return Default;
		}
	}

	private boolean getContextParameter
	(
		final Level				LogLevel,
		final String			Name,
		final boolean			Default
	)
	{
		final String			Value = getServletContext().getInitParameter(Name);

		if (Value == null) {

			Log.log(LogLevel, "The web application deployment descriptor''s context parameter \"{0}\" is missing. The default value, \"{1}\", will be used instead.", new Object[] { Name, Default });
			return Default;
		}
		
	//	We could just use Boolean.parseBoolean, but it is too tolerant of bad inputs for our purposes here.
		
		if ("true".equalsIgnoreCase(Value)) {
		
			return true;
			
		} else if ("false".equalsIgnoreCase(Value)) {
			
			return false;
		
		} else {

			Log.log(LogLevel, "The web application deployment descriptor''s context parameter \"{0}\" contained the value \"{1}\", which is not a valid boolean. The default value, \"{2}\", will be used instead.", new Object[] { Name, Value, Default });
			return Default;
		}
	}
	
	private String[] getContextAndInitParameters
	(
		final String				WebAppDeployedName
	) 
	{
		final ArrayList<String>		Params = new ArrayList<String>();
		
	//	Emulate the normal command-line behavior of C and C++ by making the application's name the first parameter.
		
		Params.add(WebAppDeployedName);
		
	//	Get the web application's "context-param" data.
		
		{
			final ServletContext		SCtxt  = getServletContext();
			final Enumeration<String>	Names  = SCtxt.getInitParameterNames();
			
			while (Names.hasMoreElements()) {
				final String			Name  = Names.nextElement();
				final String			Value = SCtxt.getInitParameter(Name);
				
				Params.add(Name);
				Params.add(Value);
			}
		}
		
	//	Get this servlet's "init-param" data.
	
		{
			final ServletConfig			SConf = getServletConfig();
			final Enumeration<String>	Names = SConf.getInitParameterNames();
			
			while (Names.hasMoreElements()) {
				final String			Name  = Names.nextElement();
				final String			Value = SConf.getInitParameter(Name);
				
				Params.add(Name);
				Params.add(Value);
			}
		}
		
		return Params.toArray(new String[Params.size()]);
	}
	
	private void logRequest
	(
		final HttpServletRequest	Req
	)
	{
		if (Log.isLoggable(Level.FINE) == false)
			return;
		
		final boolean			FinerIsLoggable = Log.isLoggable(Level.FINER);
		final StringBuilder		Buff            = new StringBuilder(FinerIsLoggable ? 6 * 1024 : 3 * 1024);
		
		if (FinerIsLoggable) {
			
			Buff.append(ServletToString).append("\n\n");
			
		//	Log the HTTP headers.
			
			{
				final MutableTable			T    = new MutableLLTable().addCol("Header Name").addCol("Header Value");
				final Enumeration<String>	Hdrs = Req.getHeaderNames();
				
				while (Hdrs.hasMoreElements()) {
					final String				HdrName = Hdrs.nextElement();
					final Enumeration<String>	Values  = Req.getHeaders(HdrName);
					
					while (Values.hasMoreElements()) {
						
						if ("cookie".equalsIgnoreCase(HdrName))
							T.addRow().add(HdrName).add(Values.nextElement().replaceAll("; ", ";\n"));
						else
							T.addRow().add(HdrName).add(Values.nextElement());
					}
				}
				
				T.sortRowsAscending();
				
				Buff.append("HTTP Headers:\n").append(T).append('\n');
			}
		}
		
		Buff.append("Request 0x").append(Hexadecimal.fromInt(Req.hashCode())).append(" Information:\n");
		Buff.append(requestToString(Req));
		
//		Including the data set in the material we're logging is a good idea, but, in practice, it won't be available 
//		here, because Qwicap has already had to read and parse it directly in order to ensure that character set 
//		issues are handled correctly.	
//		
//		Buff.append("\nRequest Data Set:\n");
//		Buff.append(new MutableTable(Req.getParameterMap()));
		
		if (FinerIsLoggable) {
			
			Buff.append('\n');
			
		//	Log the request attributes.
			
			{		
				final MutableTable			T         = new MutableLLTable().addCol("Attribute Name").addCol("Attribute Value");
				final Enumeration<String>	Attrs     = Req.getAttributeNames();
				int							AttrCount = 0;
				
				while (Attrs.hasMoreElements()) {
					final String			AttrName = Attrs.nextElement();
					
					T.addRow().add(AttrName).add(Req.getAttribute(AttrName));
					AttrCount++;
				}
				
				T.sortRowsAscending();
				
				Buff.append("Request 0x").append(Hexadecimal.fromInt(Req.hashCode())).append(" Attributes (").append(AttrCount).append("):\n").append(T).append('\n');
			}
			
		//	Log the session (if any) and its attributes.
			
			{
				final HttpSession		Sess = Req.getSession(false);
				
				if (Sess != null) {
					final MutableTable			T         = new MutableLLTable().addCol("Attribute Name").addCol("Attribute Value");
					final Enumeration<String>	Attrs     = Sess.getAttributeNames();
					int							AttrCount = 0;
					
					while (Attrs.hasMoreElements()) {
						final String			AttrName = Attrs.nextElement();
						
						T.addRow().add(AttrName).add(Sess.getAttribute(AttrName));
						AttrCount++;
					}
					
					T.sortRowsAscending();
					
					Buff.append("Session 0x").append(Hexadecimal.fromInt(Sess.hashCode())).append(" Attributes (").append(AttrCount).append("):\n").append(T);
					
				} else
					Buff.append("No Session.\n");
			}
		}
		
		Log.fine(Buff.toString());
	}

	private static String requestToString
	(
		final HttpServletRequest	Req
	)
	{
	//	For background on intrepeting some of this information see the article "Context Roots and Servlet Mappings" at: 
	//	http://access1.sun.com/techarticles/context.html
		
		return new ReadOnlyTable(new Object[][] {
		
			{ "Remote Address",       Req.getRemoteAddr()            },
			{ "Remote Host",          Req.getRemoteHost()            },
			{ "Protocol",             Req.getProtocol()              },
			{ "Method",               Req.getMethod()                },
			{ "Request URL",          Req.getRequestURL()            },
			{ "Content Type",         Req.getContentType()           },
			{ "Content Length",       Req.getContentLength()         },
			{ "Character Encoding",   Req.getCharacterEncoding()     },	//	This is usually (always?) meaningless.
		
			{ "Scheme",               Req.getScheme()                },
			{ "Server Name",          Req.getServerName()            },
			{ "Server Port",          Req.getServerPort()            },
			{ "Request URI",          Req.getRequestURI()            },
			{ "Query String",         Req.getQueryString()           },
			{ "Context Path",         Req.getContextPath()           },
			{ "Servlet Path",         Req.getServletPath()           },
			{ "Path Info",            Req.getPathInfo()              },
			{ "Path Translated",      Req.getPathTranslated()        },
		
			{ "Requested Session ID", Req.getRequestedSessionId()    },
			{ "Session ID Valid",     Req.isRequestedSessionIdValid()},
			{ "Is Secure",            Req.isSecure()                 },
			{ "Auth Type",            Req.getAuthType()              },
			{ "Remote User",          Req.getRemoteUser()            }
		
		}).toString();
	}

	@Override public String toString() {
		return ServletToString;
	}
	
	final Class<?> getClientClass() {
		return ClientClass;
	}
	
	static QwicapGlobalServices getGlobalServices() {
		return GlobalServices;
	}
	
	final QwicapApplicationServices getApplicationServices() {
		return ApplicationServices;	
	}
	
	private String getContextPath() {
		
	//	Note: Contrary to what the documentation suggests, the method Servlet.getServletName() appears to be another 
	//	      source of this same information. For a web application deployed as "guess.war", getServletName() returns 
	//	      "/localhost/guess/".
		
		try {
			String					Path = getServletContext().getResource("/").getPath();
			
			if (Path.startsWith("/localhost/"))
				Path = Path.substring("/localhost/".length());
			if (Path.endsWith("/"))
				Path = Path.substring(0, Path.length() - 1);
			
			return Path;
			
		} catch (MalformedURLException e) {
			Log.log(Level.WARNING, "Cannot get resource \"/\" from servlet context.", e);	
		}
		
		return "?";
	}
	
	
/*	todo Revisit this issue.
	
	private static String[]		ServletURLPatterns;
	
	private static String[] getAllKnownServletURLPatterns() {
		
		if (ServletURLPatterns == null) {
			
			try {
				final Markup			WebXML = new XMLDocument(new File("WEB-INF/web.xml"));
				final Results			URLPats = WebXML.get("web-app servlet-mapping url-pattern");
				
				ServletURLPatterns = new String[URLPats.size()];
				
				for (int Index = 0, Count = ServletURLPatterns.length; Index < Count; Index++)
					ServletURLPatterns[Index] = URLPats.get(Index).getText().trim();
				
			} catch (Exception e) {
				
				e.printStackTrace();
				ServletURLPatterns = new String[0];
			}
		}
		
		return ServletURLPatterns;
	}
*/
}




/*	From <http://access1.sun.com/techarticles/context.html>:

	When a request is processed an attempt is made to match any string that comes after the context root with the URL 
	patterns of any available Servlet mappings. A URL pattern can take any one of the following forms:
	
	1. A string that begins with a '/' and ends with '/*' such as '/Utils/*'. This would match with '/Utils/namelookup' 
	   and '/Utils/test'.
	
	2. A string that begins with '*.' such as '*.asp'. This would match with any string after the context root that 
	   ends with '.asp'.
	
	3. A plain string with no wild cards such as '/order'. This would only match with '/order'.
	
	4. A string containing only the '/' character. This would be the default Java Servlet of the Web Module and would 
	   be the one that gets invoked if the URL ended with the context root followed by '/'.
	
	In the case where we had mappings for two different Java Servlets as follows:
	
		Servlet1 '/Utils/*' Servlet2 '*.asp'
	
	A request that ended with '/Utils/test.asp' would always invoke Servlet1.
*/






