/*	QWICAP (Quick Web Interface for Conventional Applications) by Chris W. Johnson
	Copyright (C) 2003-2010 The University of Texas at Austin ("U.T. Austin"). All rights reserved.
	
	U.T. Austin's copyright notice in this evolving version reflects our original authorship. It does not assert 
	copyright over changes contributed by others.
	
	This library is free software; you can redistribute it and/or modify it under the terms of the GNU Lesser General 
	Public License as published by the Free Software Foundation; either version 2.1 of the License, or (at your option) 
	any later version.
	
	This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied 
	warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for 
	more details.
	
	You should have received a copy of the GNU Lesser General Public License along with this library; if not, write to 
	the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/


package edu.utexas.its.eis.tools.qwicap.servlet;

import edu.utexas.its.eis.tools.qwicap.util.StackTrace;

import javax.servlet.http.HttpSession;
import java.util.Date;
import java.util.logging.Logger;
import java.util.logging.Level;
import java.io.IOException;


/**
 * Created by IntelliJ IDEA.
 * User: chrisj
 * Date: Oct 2, 2006
 * Time: 7:54:34 PM
 * To change this template use File | Settings | File Templates.
 */
final class QwicapThread extends Thread {

	private static final Logger		Log = Logger.getLogger(QwicapThread.class.getName());
	
	QwicapThreadList				List;
	QwicapThread					Prev;
	QwicapThread					Next;
	
	private final QwicapThreadPool	Pool;

	private boolean					PleaseDie;
	private Qwicap					Application;
	private boolean					Active;
	private boolean					Jammed;

	private long					BeginMillis;
	private long					EndMillis;

	private long					LastHitTimeMillis;

	private boolean					Expired;			//	true if this thread's session expired.


	QwicapThread
	(
		final QwicapThreadPool	Pool,
		final long				StackSize
	)
	{
		super(Pool.getGroup(), null, "Qwicap Thread", StackSize);
		this.Pool = Pool;

		Pool.threadBorn();
	}

	long getLastHitTimeMillis() {
		
	//	In the general case, access to LastHitTimeMillis should be synchronized, since access to long primitives isn't 
	//	guaranteed to be atomic. However, this method is only used when killing-off "jammed" threads, and a jammed 
	//	thread, by definition, isn't accepting any hits, so this value can't be in the process of being updated whem it
	//	is accessed here.
		
		return LastHitTimeMillis;
	}

	boolean isActive() {
		return Active;
	}

	boolean isJammed
	(
		final long				CurTimeMillis
	)
	{
		final Qwicap			Application = this.Application;
		
		if (Application == null || LastHitTimeMillis == 0)
			return false;

	//	We want to give Tomcat's expiration mechanism a little bit of leeway before concluding that a thread is jammed, 
	//	so we arbitrarily require a thread to be 120 seconds beyond its expiration time before designating it "jammed".

		final HttpSession		Sess = Application.getSessionContext().getSession();
		
		if (Sess == null)
			return false;
		
		final int				JamMinSecs       = Sess.getMaxInactiveInterval() + 120;
		final int				SecsSinceLastHit = (int) ((CurTimeMillis - LastHitTimeMillis) / 1000);

		if (Jammed) {

			Log.log(Level.INFO, "A Qwicap thread is still jammed after {0} seconds. (Since {1}.) The thread''s current state is {2}. Its stack trace follows:\n{3}", new Object[] { SecsSinceLastHit - JamMinSecs, new Date(LastHitTimeMillis), getState(), StackTrace.toString(this) });
			return true;
		}

	//	Determine whether or not this thread's session should have expired, but hasn't.

		if (SecsSinceLastHit < JamMinSecs)
			return false;
		
		Log.log(Level.INFO, "A Qwicap thread has been jammed for {0} seconds. (Since {1}.) The thread''s current state is {2}. Its stack trace follows:\n{3}", new Object[] { SecsSinceLastHit - JamMinSecs, new Date(LastHitTimeMillis), getState(), StackTrace.toString(this) });
		
		Jammed = true;
		Pool.threadJammed();
		
		return true;
	}

	long getTotalSessionMillis() {
		
	//	In the general case, access to BeginMillis and EndMillis should be synchronized, since access to long primitives 
	//	isn't guaranteed to be atomic. However, this method is only used once a thread has become inactive (it no longer
	//  represents a web application session), so these values can't be in the process of being updated whem they are
	//	accessed here.
		
		return EndMillis - BeginMillis;
	}

	boolean expired() {
		
	//	Accesses to this type of primitive are atomic, so there's no problem doing so in a non-synchronized fashion.
		
		return Expired;
	}

	Qwicap getApplication() {
		return Application;
	}

	void hitDone
	(
		final long				StartMillis
	)
	{
		LastHitTimeMillis = StartMillis;
		Pool.hit(StartMillis);
	}

	void badDataSet() {
		
		Pool.badDataSet();
		Log.fine("Bad form data set received.");
	}

	void sessionExpired() {
		Expired = true;
	}

	void killThread
	(
		final String			Reason
	)
	{
		try {
			PleaseDie = true;
			
		//	This should stop the current Qwicap application from executing, but leave the thread alive and available
		//	for re-use.
			
			stopApplication(Reason);
			
			if (isAlive()) {
				
			//	This should wake-up a thread that is not executing a Qwicap application, and thereby allow it to notice 
			//	that it is supposed to die.
				
				interrupt();
				
			} else if (Thread.State.NEW == getState()) {
				
			//	If this thread hasn't been started, we can proceed directly to declaring this thread dead.
				
				Pool.threadDead(this);
			}
				
	//	The following catch block will only serve to catch RuntimeExceptions (the enclosed code throws no declared
	//	exceptions). It is here to ensure that nothing prevents the thread pool from killing off all of a servlet's
	//	threads, when the servlet is destroyed.
			
		} catch (Exception e) {
			Log.log(Level.WARNING, "An attempt to kill a Qwicap thread produced an exception.", e);
		}
	}

	synchronized QwicapThread startApplication
	(
		final QwicapGlobalServices		GlobalServices,
		final QwicapApplicationServices	ApplicationServices,
		final SessionContext			SessCtxt,
		final HitContext				HitCtxt
	)
	throws QwicapThreadAlreadyActiveException, NoSuchMethodException, IOException
	{
		if (Active)
			throw new QwicapThreadAlreadyActiveException("This thread is already active, and therefore cannot be reused at this time.");

		BeginMillis       = System.currentTimeMillis();
		EndMillis         = 0;
		Expired           = false;
		Jammed            = false;
		LastHitTimeMillis = 0;
		Application       = new Qwicap(GlobalServices, ApplicationServices, this, SessCtxt, HitCtxt);
		Active            = true;
		
		setName("Qwicap \"" + ApplicationServices.getServletContextPath() + "\" Thread");
		
		if (isAlive() == false)
			start();	//	Start this new, not-yet-started thread.
		else
			notify();	//	Wake-up this existing, running thread, which is blocked in its "run" method.

		return this;
	}

	void stopApplication
	(
		final String			Reason
	)
	{
		final Qwicap			Application = this.Application;

		if (Application != null)
			Application.pleaseDie(Reason);
	}

	@Override public void run() {
		
		Log.log(Level.FINEST, "QTHREAD: Qwicap thread <{0}> started.", new Object[] { this });
		
		try {

			while (!PleaseDie) {

				if (Application != null) {
					Class<?>				ClientClass = null;
					
					try {
						
						ClientClass = Application.getHitContext().getServlet().getClientClass();
						
						Log.log(Level.FINEST, "QTHREAD: Qwicap thread <{0}> beginning run of {1}.", new Object[] { this, ClientClass });
						
						Pool.threadActive(this);
						Application.run();

					} catch (Exception e) {
						
					//	Discard exception. Whatever killed the application shouldn't kill this pooled thread.
					//	Also, given all of the exception handling in Qwicap.run, there should never be an exception to 
					//	catch at this point.
						
						Log.log(Level.WARNING, "QTHREAD: Qwicap session died due to an unanticipated exception. Its thread will be returned to the pool.", e);
						
					} finally {
						
						Log.log(Level.FINEST, "QTHREAD: Qwicap thread <{0}> completed run of {1}.", new Object[] { this, ClientClass });

						EndMillis   = System.currentTimeMillis();	//	Set EndMillis before invoking threadInactive.
						Active      = false;
						Application = null;
						
						setName("Qwicap Idle Pooled Thread");
						
						Pool.threadInactive(this);
					}
				}

				synchronized (this) {

				//	Wait for this thread to be assigned a session to run.

					if (Application == null && !PleaseDie)
						try { wait(); } catch (InterruptedException e) { /* Ignore. */ }
				}
			}

		} catch (Exception e) {
			
			Log.log(Level.WARNING, "QTHREAD: Qwicap thread <" + this + "> died due to an exception. This thread will be dropped from the pool.", e);
			
		} finally {
			
			Pool.threadDead(this);
			
			Log.log(Level.FINEST, "QTHREAD: Qwicap thread <{0}> stopped.", this);
		}
	}
}
