/*	QWICAP (Quick Web Interface for Conventional Applications) by Chris W. Johnson
	Copyright (C) 2003-2008 The University of Texas at Austin ("U.T. Austin"). All rights reserved.
	
	U.T. Austin's copyright notice in this evolving version reflects our original authorship. It does not assert 
	copyright over changes contributed by others.
	
	This library is free software; you can redistribute it and/or modify it under the terms of the GNU Lesser General 
	Public License as published by the Free Software Foundation; either version 2.1 of the License, or (at your option) 
	any later version.
	
	This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied 
	warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for 
	more details.
	
	You should have received a copy of the GNU Lesser General Public License along with this library; if not, write to 
	the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/


package edu.utexas.its.eis.tools.qwicap.servlet;


import edu.utexas.its.eis.tools.qwicap.util.StackTrace;

import java.util.Date;
import java.util.EnumMap;
import java.util.logging.Level;
import java.util.logging.Logger;


/**	Represents a list of <code>QwicapThread</code> instances that are either running, or waiting to run, an instance of
 * 	<code>Qwicap</code>. Note that there is no synchronization in this class, because this class is only for use by the
 * 	<code>QwicapThreadList</code> class, which maintains two such lists (one for active threads, and one for inactive)
 * 	and provides all of the necessary synchronization in its own implementation.
 *
 * 	@author Chris W. Johnson
 */
final class QwicapThreadList {

	private static final Logger	Log = Logger.getLogger(QwicapThreadList.class.getName());

	private int					Count;
	private QwicapThread		First;
	private QwicapThread		Last;


	QwicapThread pop() {

		if (Last == null)
			return null;

		final QwicapThread		Popped = Last;

		if (Count == 1) {

			First = null;
			Last  = null;

		} else {

			Last = Popped.Prev;
			Last.Next = null;
		}

		Popped.List = null;
		Popped.Prev = null;
		Popped.Next = null;

		Count--;

		return Popped;
	}

	void addAndTrimTo
	(
		final QwicapThread		QT,
		final int				MaxElemCount,
		final String			Reason
	)
	{
		if (Count < MaxElemCount) {

			add(QT);

		} else {

			QT.killThread(Reason);
			trimTo(MaxElemCount, Reason);
		}
	}

	QwicapThread add
	(
		final QwicapThread		QT
	)
	{
		if (QT.List == this)
			throw new IllegalStateException("Cannot add a thread to a list to which it already belongs.");

		if (QT.List != null)
			throw new IllegalArgumentException("Cannot add a thread to a list while it is an element of another list.");

		if (First == null) {

			QT.Prev = null;
			First = Last = QT;

		} else {

			QT.Prev = Last;
			Last.Next = Last = QT;
		}

		QT.Next = null;
		QT.List = this;

		Count++;

		return QT;
	}

	void remove
	(
		final QwicapThread		QT
	)
	{
		if (QT.List == null)
			throw new IllegalStateException("Cannot remove a thread from a list when it is not a member of any list.");

		if (QT.List != this)
			throw new IllegalArgumentException("Cannot remove a thread from a list to which it does not belong.");

		final QwicapThread		Prev = QT.Prev;
		final QwicapThread		Next = QT.Next;

		QT.List = null;
		QT.Prev = null;
		QT.Next = null;

		if (Prev == null)
			First = Next;	//	We're removing the first item in the list.	
		else
			Prev.Next = Next;

		if (Next == null)
			Last = Prev;	//	We're removing the last item in the list.
		else
			Next.Prev = Prev;

		Count--;
	}

	void trimTo
	(
		final int				MaxElemCount,
		final String			Reason
	)
	{
		while (Count > MaxElemCount)
			pop().killThread(Reason);
	}

	int length() {
		return Count;
	}

	void killAll
	(
		final String			Reason
	)
	{
		QwicapThread			QT = First;

		while (QT != null) {
			final QwicapThread		Next = QT.Next;	//	killThread doesn't change the thread's links, but let's keep things obviously safe.

			QT.killThread(Reason);

			QT = Next;
		}

	//	Those "killThread" invocations should cause the threads to exit. As they do so, they tell the thread pool that
	//	they're dead by invoking QwicapThreadPool.threadDead, and it invokes the "remove" method of the appropriate
	//	QwicapThreadList instance. Note that all of the those QwicapThreadPool.threadDead invocations made by the 
	//	exiting threads will block until the QwicapThreadPool.destroyAll method releases its lock. If the threads take 
	//	a while to exit, QwicapThreadPool.destroyAll may have completed its work, and the exiting threads may only have 
	//	to contend with each other for the lock.
	}

	int killJammedThreads() {
		int						JammedCount   = 0;
		final long				CurTimeMillis = System.currentTimeMillis();

		for (QwicapThread QT = First; QT != null; QT = QT.Next) {

			try {

				if (QT.isJammed(CurTimeMillis)) {

					Log.log(Level.INFO, "Killing a Qwicap thread that has been jammed since {0}.", new Object[] { new Date(QT.getLastHitTimeMillis()) });

					JammedCount++;
					QT.killThread("Thread was jammed.");
				}

			} catch (Exception e) {
				Log.log(Level.INFO, "An attempt to determine a thread's \"jammed\" status, or to kill a \"jammed\" thread, has failed due to an exception.", e);
			}
		}

		return JammedCount;
	}

	void logAllStackTraces
	(
		final Level				LogLevel,
		final String			ThreadListName
	)
	{
		try {
			if (Log.isLoggable(LogLevel) == false)
				return;

			QwicapThread			QT = First;

			while (QT != null) {

				Log.log(LogLevel, "The thread \"{1}\", in the {0} list, is in the state \"{2}\", and its current stack trace is:\n{3}", new Object[] { ThreadListName, QT, QT.getState(), StackTrace.toString(QT.getStackTrace()) });
				QT = QT.Next;
			}

		} catch (Exception e) {
			Log.log(Level.SEVERE, "An attempt to log the current stack traces of \"" + ThreadListName + "\" threads failed.", e);
		}
	}

	EnumMap<Thread.State, Integer> getThreadStates() {
		final Thread.State[]	States   = Thread.State.values();
		final int[]				Counters = new int[States.length];

		for (QwicapThread QT = First; QT != null; QT = QT.Next)
			Counters[QT.getState().ordinal()]++;

		final EnumMap<Thread.State, Integer>	StatesMap = new EnumMap<Thread.State, Integer>(Thread.State.class);

		for (int Index = 0, Count = States.length; Index < Count; Index++)
			StatesMap.put(States[Index], Counters[Index]);

		return StatesMap;
	}
}
