/*	QWICAP (Quick Web Interface for Conventional Applications) by Chris W. Johnson
	Copyright (C) 2003-2008 The University of Texas at Austin ("U.T. Austin"). All rights reserved.
	
	U.T. Austin's copyright notice in this evolving version reflects our original authorship. It does not assert 
	copyright over changes contributed by others.
	
	This library is free software; you can redistribute it and/or modify it under the terms of the GNU Lesser General 
	Public License as published by the Free Software Foundation; either version 2.1 of the License, or (at your option) 
	any later version.
	
	This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied 
	warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for 
	more details.
	
	You should have received a copy of the GNU Lesser General Public License along with this library; if not, write to 
	the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/


package edu.utexas.its.eis.tools.qwicap.servlet;


import edu.utexas.its.eis.tools.table.readonly.ReadOnlyTable;
import edu.utexas.its.eis.tools.qwicap.util.StackTrace;

import java.util.logging.Logger;
import java.util.logging.Level;
import java.util.ArrayList;
import java.io.IOException;


/**
 * Created by IntelliJ IDEA.
 * User: chrisj
 * Date: Oct 2, 2006
 * Time: 7:50:11 PM
 * To change this template use File | Settings | File Templates.
 */
final class QwicapThreadPool {

	private static final Logger			Log = Logger.getLogger(QwicapThreadPool.class.getName());

	private final String				ServletName;
	
	private final int					MaxThreads;
	private final long					ThreadStackSize;
	private final int					MaxSpareThreads;
	
	private final ThreadGroup			Group;

	private final Object				ThreadListsSynchObj = new Object();
	private final QwicapThreadList		Active              = new QwicapThreadList();
	private final QwicapThreadList		Inactive            = new QwicapThreadList();
		
	private final ServiceDataRecorder	SDR;
	
	private boolean						Closed;
	private int							StartCount;


	QwicapThreadPool
	(
		final String				ServletName,
		int							MinThreads,
		int							MaxSpareThreads,
		final int					MaxThreads,
		final long					ThreadStackSize,
		final ServiceDataRecorder	SDR
	)
	{
		this.ServletName = ServletName;
		this.Group       = new ThreadGroup("Qwicap \"" + ServletName + "\" thread group");
		this.SDR         = SDR;
		
		SDR.setThreadPool(this);
		
		if (MaxSpareThreads > MaxThreads)
			MaxSpareThreads = MaxThreads;
		else if (MaxSpareThreads < 0)
			MaxSpareThreads = 0;
		
		if (MinThreads > MaxSpareThreads)
			MinThreads = MaxSpareThreads;
		else if (MinThreads < 0)
			MinThreads = 0;
		
		this.MaxThreads      = MaxThreads;
		this.MaxSpareThreads = MaxSpareThreads;
		this.ThreadStackSize = ThreadStackSize;
		
	//	Create some inactive threads.

		for (int Count = MinThreads; --Count >= 0; )
			Inactive.add(new QwicapThread(this, ThreadStackSize));

		Log.log(Level.INFO, "Thread pool \"{4}\" setup. Minimum threads: {0}; Maximum spare threads: {1}; Maximum threads: {2}; Stack sizes: {3} KB.", new Object[] { MinThreads, MaxSpareThreads, MaxThreads, ThreadStackSize / 1024, ServletName });
	}

	void destroy() {
		
		try {
			
			synchronized (ThreadListsSynchObj) {
				
				Closed = true;
				
				if (Log.isLoggable(Level.FINE))
					Log.log(Level.FINE, "About to kill all threads in Qwicap thread pool \"{2}\". Active: {0}; Inactive: {1}.\n\nActive Pool Thread States:\n{3}\nInactive Pool Thread States:\n{4}", new Object[] { Active.length(), Inactive.length(), ServletName, new ReadOnlyTable(Active.getThreadStates()), new ReadOnlyTable(Inactive.getThreadStates()) });
				
				final int				InitialActive   = Active.length();
				final int				InitialInactive = Inactive.length();
				final int				InitialThreads  = InitialActive + InitialInactive;
				final long				StartMillis     = System.currentTimeMillis();
				int						Iteration       = 0;	
				
				while (Active.length() != 0 || Inactive.length() != 0) {
					
					Log.log(Level.FINE, "Killing all threads in Qwicap thread pool \"{3}\"; iteration {0} begun. Active: {1}; Inactive: {2}.", new Object[] { Iteration, Active.length(), Inactive.length(), ServletName});
					
				//	If, for any reason, we just can't seem to kill these threads, log some useful information about them.
					
					if (++Iteration % 30 == 0) {
					
						Active.logAllStackTraces(Level.INFO, "active");
						Inactive.logAllStackTraces(Level.INFO, "inactive");
					}
					
				//	The following try/catch blocks will only serve to catch RuntimeExceptions (the enclosed code throws 
				//	no declared exceptions). It is here to ensure that nothing prevents the thread pool from killing 
				//	off all of a servlet's threads, when the servlet is destroyed.
					
					try {
						Active.killAll("Web application shutdown (servlet destroyed).");
					} catch (Exception e) {
						Log.log(Level.WARNING, "An attempt to kill all active threads in \"" + ServletName + "\" was interrupted by an exception. Further attempts will be made.", e);
					}
					
					try {
						Inactive.killAll("Web application shutdown (servlet destroyed).");
					} catch (Exception e) {
						Log.log(Level.WARNING, "An attempt to kill all inactive threads in \"" + ServletName + "\" was interrupted by an exception. Further attempts will be made.", e);
					}
					
				//	Release our lock on ThreadListsSynchObj by waiting. During the wait, the threads we killed above,
				//	which may have already reached their invocations of threadDead and blocked waiting for the lock, 
				//	will have a chance to unblock and complete the threadDead invocation. if all threads manage to do 
				//	that during our wait, the enclosing "while" loop will terminate.
					
					try {
						ThreadListsSynchObj.wait(1000);	//	One second.
					} catch (InterruptedException e) {
					//	Ignore.
					}
				}
				
				Log.log(Level.INFO, "{0,choice,0#All zero threads|1#The one thread|1<All {0} threads} ({1} active, {2} inactive) in the Qwicap thread pool \"{5}\" {0,choice,0#were|1#was|1<were} killed after {3,choice,0#zero iterations|1#one iteration|1<{3} iterations}, totalling {4} ms.", new Object[] { InitialThreads, InitialActive, InitialInactive, Iteration, System.currentTimeMillis() - StartMillis, ServletName });
			}
			
		} catch (Exception e) {
			
			Log.log(Level.SEVERE, "The process of killing all threads in Qwicap thread pool \"" + ServletName + "\" failed due to an exception.", e);
		
		} finally {
			
		//	As a final check that all threads have been shutdown, report the status of every thread remaining in our
		//	thread group.
		
			reportThreadGroupStatus();
			
		//	Some code, like the MySQL driver, creates daemon threads in our thread group, and leaves them running. 
		//	While they will not prevent server shutdown, they will retain resources (including the current ClassLoader)
		//	across application redeployments, which is bad. Because those daemon threads have no obvious way of learning
		//	that the web application is being destroyed, it seems like a good idea to interrupt them, in the hope that
		//	doing so will prompt them to exit gracefully. There's nothing else we can do about daemon threads at this 
		//	level. The question is: should we be doing anything about them at all? They're obviously a problem, but 
		//	could we make matters worse by interrupting them? I believe the answer is "no, we won't make matters worse", 
		//	because those daemon threads will be killed one way or another if/when the application server is shutdown, 
		//	so they must be able to die in a graceful (or, at least, a harmless) manner.
		//
		//	There's also the question of whether we should perform this interruption before, or after, the invocation of
		//	reportThreadGroupStatus. If we do the interruption before (and wait a little while), all of the daemon 
		//	threads may exit, and therefore there will be no need for reportThreadGroupStatus to worry a system admin
		//	about the daemon threads. On the other hand, maybe the system admin should be aware of the daemon thread 
		//	issue, in case they can do something in the application code to prevent the daemons from lingering in the 
		//	first place.
				
			try {
				final int				OrigActiveCount = Group.activeCount();
				
				if (OrigActiveCount != 0) {
					
					Group.interrupt();
					Log.log(Level.INFO, "Interrupted all lingering threads in the {0}. Hopefully, that will induce them to exit gracefully.", Group.getName());
				
				//	When interrupting AFTER invoking reportThreadGroupStatus, there's no point in waiting around to 
				//	give the lingering threads time to respond to the interruption. They'll either exit, or they won't,
				//	and waiting around won't change the outcome.
				
					synchronized (ThreadListsSynchObj) {
						ThreadListsSynchObj.wait(1000);
					}
					
					final int				LingeringThreadCount = Group.activeCount();
					
					Log.log(Level.INFO, "After giving the interrupted {2,choice,1#thread|1<threads} in the {0} one second to respond to the interruption, {1,choice,0#all threads have exited|1#one thread remains active|1<{1} threads remain active}.", new Object[] { Group.getName(), LingeringThreadCount, OrigActiveCount });
					
					if (LingeringThreadCount != 0)
						Log.log(Level.WARNING, "There {0,choice,1#is one lingering thread that has resisted all efforts to make it|1<are {0} lingering threads that have resisted all efforts to make them} exit. {0,choice,1#It|1<They} will probably prevent this web application''s ClassLoader from being garbage-collected, and will therefore create a memory leak in the application server.", LingeringThreadCount);
					
				/*	if (LingeringThreadCount != 0) {
						
						if (makeLastDitchEffortToKillLingeringThreads()) {
							
							synchronized (ThreadListsSynchObj) {
								ThreadListsSynchObj.wait(2000);
							}
						}
						
						Log.log(Level.INFO, "After making a last-ditch effort to kill the lingering {2,choice,1#thread|1<threads} in the {0}, {1,choice,0#all threads have exited|1#one thread remains active|1<{1} threads remain active}.", new Object[] { Group.getName(), Group.activeCount(), LingeringThreadCount });
					}
				*/	
					logThreadStates();
				}
				
			} catch (Exception e) {
			//	We could receive a SecurityException or an InterruptedException here. Either way, we don't want the 
			//	exception to propagate, because that would only interfere with the shutdown of the current servlet.
			}
			
		//	todo Should we invoke our thread group's "destroy" method? Would doing so actually help the application server to free-up resources associated with this application, or is doing so irrelevant when a thread group is empty?
		}
	}
	
/*	This method works, in as much as it will kill a TimerTask left behind by the MySQL driver. However, killing that
	task does not appear to make a difference as to whether or not the web application's ClassLoader can be garbage-
	collected. Therefore, there's no point in executing this code, which does, after all, depend on the deprecated 
	Thread.stop method.
	
	private boolean makeLastDitchEffortToKillLingeringThreads() {
		boolean					StoppedSomeThreads = false;
		
		try {
			final Thread[]			Rogues           = new Thread[Group.activeCount() + 10];
			final int				GroupActiveCount = Group.enumerate(Rogues);
			
			for (int Index = 0; Index < GroupActiveCount; Index++) {
				final Thread			Rogue = Rogues[Index];
				
				if ("java.util.TimerThread".equals(Rogue.getClass().getName())) {
				
				//	This issue arises due to the "MySQL Statement Cancellation Timer", which won't quit when 
				//	interrupted, or when all database connections are closed. Therefore, the thread lingers (even 
				//	though it is a daemon thread) after web application destruction, and thereby prevents the web 
				//	application's ClassLoader from being garbage collected. That represents a serious memory leak that 
				//	can eventually choke the application server.
				//
				//	For any less of an issue, I wouldn't resort to using the deprecated, and potentially dangerous, 
				//	Thread.stop() method. However, an examination of the java.util.TimerThread's source code indicates 
				//	that it can be shutdown safely using the "stop" method, because its "run" method has an 
				//	all-encompasing "finally" block that appears to perform all necessary cleanup without regard to the 
				//	circumstances of the thread's death.
					
					try {
						Rogue.stop();
						StoppedSomeThreads = true;
						Log.log(Level.INFO, "In a last-ditch cleanup attempt, the java.util.TimerThread instance \"{0}\" was killed using the Thread.stop method. In this case it should be safe, but this should not have been necessary.", Rogue);
					} catch (Exception e) {
						Log.log(Level.INFO, "The last-ditch attempt to stop the java.util.TimerThread instance \"" + Rogue + "\" failed due to an exception. The Timer will probably prevent this web application''s ClassLoader from being garbage-collected, and will therefore create a memory leak in the application server.", e);
					}
				}
				
			//	Log.log(Level.INFO, "A lingering Timer exists (\"{0}\"). It will probably prevent this web application''s ClassLoader from being garbage-collected, and will therefore create a memory leak in the application server.", Rogue);
			}
			
		} catch (Exception e) {
			Log.log(Level.WARNING, "The last-ditch effort to kill lingering threads failed due to an exception.", e);	
		}
		
		return StoppedSomeThreads;
	}
*/	
	private void reportThreadGroupStatus() {
		
	//	Just to be absolutely certain that things are working as expected, check our thread group for lingering
	//	threads. There should be no way for there to be any, but, if there were some, we'd want to know about it.
		
		try {
			final Thread[]			Rogues           = new Thread[Group.activeCount() + 10];
			final int				GroupActiveCount = Group.enumerate(Rogues);
			
			if (GroupActiveCount == 0) {
				
				Log.log(Level.INFO, "It has been verified that there are no longer any active threads in the {0}. No impediments to the destruction of servlet \"{1}\" should remain.", new Object[] { Group.getName(), ServletName });
				return;
			}
						
		//	There are some lingering threads in the Group thread group. Determine which ones are normal threads that
		//	could interfere with servlet destruction, and which ones are daemons that should create no interference.
			
			final ArrayList<Thread>	NormalThreads = new ArrayList<Thread>(GroupActiveCount);
			final ArrayList<Thread>	DaemonThreads = new ArrayList<Thread>(GroupActiveCount);

			for (int Index = 0; Index < GroupActiveCount; Index++) {
				final Thread			Rogue = Rogues[Index];
				
				if (Rogue.isDaemon())
					DaemonThreads.add(Rogue);
				else
					NormalThreads.add(Rogue);
			}
			
		//	Log as much of the information that we've gathered as is appropriate to the current logging level.
						
			if (NormalThreads.isEmpty()) {
				
				Log.log(Level.INFO, "It has been verified that there are no longer any active, non-daemon threads in the {0}. (However, there {1,choice,1#is one daemon thread|1<are {1} daemon threads}, and {1,choice,1#it is|1<they are} likely to retain resources across application deployments, and thereby cause a memory leak in the application server.) Nonetheless, destruction of servlet \"{2}\" can proceed.", new Object[] { Group.getName(), DaemonThreads.size(), ServletName });
				
			} else {
				
				Log.log(Level.SEVERE, "All non-daemon threads in the {0} should be dead, but {1,choice,1#one is|1<{1} are} still active.", new Object[] { Group.getName(), NormalThreads.size() });
				
				for (final Thread Rogue : NormalThreads)
					logThreadState(Level.SEVERE, Rogue, false);
				
				if (DaemonThreads.size() != 0)
					Log.log(Level.FINE, "There {1,choice,1#is also one daemon thread|1<are also {1} daemon threads} still active in the {0}.", new Object[] { Group.getName(), DaemonThreads.size() });
			}
				
			if (Log.isLoggable(Level.FINE)) {
				
				for (final Thread Rogue : DaemonThreads)
					logThreadState(Level.FINE, Rogue, true);
			}
			
		} catch (Exception e) {
			Log.log(Level.WARNING, "An attempt to report thread group status failed.", e);	
		}
	}
	
	private void logThreadStates() {
	
		try {
			int						GroupActiveCount = Group.activeCount();
			final Thread[]			Rogues           = new Thread[GroupActiveCount + 10];
			
			GroupActiveCount = Group.enumerate(Rogues);
			
			for (int Index = 0; Index < GroupActiveCount; Index++) {
				final Thread			Rogue = Rogues[Index];
				
				logThreadState(Level.INFO, Rogue, Rogue.isDaemon());
			}
		
		} catch (Exception e) {
			Log.log(Level.WARNING, "An attempt to log thread group status failed.", e);	
		}		
	}
	
	private void logThreadState
	(
		final Level				Level,
		final Thread			Rogue,
		final boolean 			IsDaemon
	)
	{
		Log.log(
			Level,
			"The {4}thread \"{1}\" of class \"{5}\" in the {0}, is in the state \"{2}\", and its current stack trace is:\n{3}", 
			new Object[] { 
				Group.getName(), 
				Rogue, 
				Rogue.getState(), 
				StackTrace.toString(Rogue.getStackTrace()).trim(), 
				IsDaemon ? "daemon " : "",
				Rogue.getClass().getName()
			}
		);
	}
	
	int getActiveThreadCount() {
		synchronized (ThreadListsSynchObj) {
			return Active.length();
		}
	}
	
	int getInactiveThreadCount() {
		synchronized (ThreadListsSynchObj) {
			return Inactive.length();
		}
	}
	
	int getJammedThreadCount() {
		synchronized (ThreadListsSynchObj) {
			return Active.killJammedThreads();
		}
	}

	QwicapThread startApplication
	(
		final QwicapGlobalServices		GlobalServices,
		final QwicapApplicationServices	ApplicationServices,
		final SessionContext			SessCtxt,
		final HitContext				HitCtxt
	)
	throws NoSuchMethodException, QwicapThreadPoolException, IOException
	{
		synchronized (ThreadListsSynchObj) {
			
			if (Closed)
				throw new QwicapThreadPoolException("The web application \"" + ServletName + "\" is shutting down. No new sessions will be started.");

			for (int Tries = 0; Tries < 2; Tries++) {
				
				if ((Tries != 0 || StartCount++ >= 1000) && Active.killJammedThreads() != 0) {
					
					StartCount = 0;
					try { ThreadListsSynchObj.wait(500); } catch (InterruptedException e) { /* Ignore. */ }	//	Wait a little while, then try again.
				}
				
				QwicapThread			QT = Inactive.pop();
				
				if (QT == null) {
					
				//	The list of inactive threads is empty. We can only satisfy this request by creating a new thread, if
				//	we're permitted to do so.
					
					if (MaxThreads == 0 || MaxThreads > Active.length())
						QT = new QwicapThread(this, ThreadStackSize);
				}
				
				if (QT != null)
					return QT.startApplication(GlobalServices, ApplicationServices, SessCtxt, HitCtxt);
			}
		}
		
		SDR.threadUnavailable();

		throw new QwicapThreadPoolException("No more Qwicap threads are available to \"" + ServletName + "\". The total number of Qwicap threads is limited to " + MaxThreads + ". That limit has been reached, and all of those threads are active.");
	}

	ThreadGroup getGroup() {
		return Group;
	}

	void threadBorn() {
		
		synchronized (ThreadListsSynchObj) {
		
		//	We add one to the active thread count in order to account for the newly-born thread which is neither 
		//	active, nor inactive, yet, but which will be active any moment now.
		
			SDR.threadBorn(Active.length() + 1, Inactive.length());
		}
	}

	void threadDead
	(
		final QwicapThread		QT	
	) 
	{
		synchronized (ThreadListsSynchObj) {
			
			if (QT.List != null)
				QT.List.remove(QT);
			
			SDR.threadDead(Active.length(), Inactive.length());
		}
	}

	void threadActive
	(
		final QwicapThread		QT
	)
	{
		synchronized (ThreadListsSynchObj) {
			
			Active.add(QT);
			SDR.threadActive(Active.length(), Inactive.length());
		}
	}

	void threadInactive
	(
		final QwicapThread		QT
	)
	{
		final long				SessionMillis = QT.getTotalSessionMillis();
		final boolean			Expired       = QT.expired();
		
		synchronized (ThreadListsSynchObj) {
			
			Active.remove(QT);
			Inactive.addAndTrimTo(QT, MaxSpareThreads, "Excess inactive threads.");
			
			SDR.threadInactive(Active.length(), Inactive.length(), SessionMillis, Expired);
		}
	}

	void threadJammed() {
		SDR.threadJammed();
	}

	void hit
	(
		final long				StartMillis
	)
	{
		SDR.hit(StartMillis);
	}

	void badDataSet() {
		SDR.badDataSet();
	}
}
