/*	QWICAP (Quick Web Interface for Conventional Applications) by Chris W. Johnson
	Copyright (C) 2003-2010 The University of Texas at Austin ("U.T. Austin"). All rights reserved.
	
	U.T. Austin's copyright notice in this evolving version reflects our original authorship. It does not assert 
	copyright over changes contributed by others.
	
	This library is free software; you can redistribute it and/or modify it under the terms of the GNU Lesser General 
	Public License as published by the Free Software Foundation; either version 2.1 of the License, or (at your option) 
	any later version.
	
	This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied 
	warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for 
	more details.
	
	You should have received a copy of the GNU Lesser General Public License along with this library; if not, write to 
	the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/


package edu.utexas.its.eis.tools.qwicap.servlet;


import java.security.SecureRandom;
import java.security.Security;
import java.util.Random;
import java.util.logging.Logger;
import java.util.logging.Level;


/**	<p><code>Rule30Random</code> is a pseudo-random number generator (PRNG) based on a "rule 30", one-dimensional, 
 *	cellular automaton, as suggested by Stephen Wolfram in his book <cite>A New Kind of Science</cite>, 
 *	<a href='http://www.wolframscience.com/nksonline/page-317'>page 317</a> (see also 
 *	<a href='http://www.wolframscience.com/nksonline/page-974b-text'>page 974</a>). As Wolfram points-out, this is the 
 *	same method that Mathematica has always used to generate its random numbers. Relative to other PRNG algorithms,
 * 	about the only weakness of this one is that it will produce a repeating sequence of output bits, if it is seeded 
 * 	with a repeating sequence of bits. This implementation automatically converts such bad seeds into good seeds, by 
 * 	detecting and destroying the repetition in them. The only exception to this rule is the {@link Rule30Random(byte[])}
 * 	constructor which accepts its initial state as-is, for people who just want to watch a Rule 30 automaton do its
 * 	thing.
 *	</p>
 *	
 *	<p>The reason Qwicap has its own random number generator is this: We encountered an intermittent, and apparently 
 *	unsolvable, problem with using the <code>SecureRandom</code> class when using the combination of Solaris and Tomcat.
 *	Specifically, we would intermittently receive the exception <samp>java.security.ProviderException: Could not obtain 
 * 	session</samp> with the following stack trace:<br/>
 *	<br/>
 *	<samp>sun.security.pkcs11.SessionManager.getOpSession(SessionManager.java:122)<br/>
 *	sun.security.pkcs11.Token.getOpSession(Token.java:241)<br/>
 *	sun.security.pkcs11.P11SecureRandom.engineNextBytes(P11SecureRandom.java:92)<br/>
 *	java.security.SecureRandom.nextBytes(SecureRandom.java:413)<br/>
 *	</samp>
 *	</p>
 *
 *	<p>A Google search with the terms "java.security.ProviderException" and "Could not obtain session" showed that
 *	we weren't the first to encounter this problem, and that no one had identified a real solution. So, the only way
 *	to eliminate the problem seemed to be to stop using <code>SecureRandom</code>. On the plus side, Wolfram's
 *	description of the properties of the rule-30 random number generator indicate that it produces much stronger
 *	(less likely to repeat) random sequences than any other PRNG algorithm. 
 *	</p>
 * 
 * 	<p>In the end, a Solaris update eliminated the <code>java.security.ProviderException</code>, but by that time, this
 * 	implementation was long since complete, and I see no point in removing it.
 *	</p>
 * 	
 * 	<p><b>Note:</b> If the application server's security manager permits it, this PRNG algorithm is made available 
 * 	through the {@link Security} class' \"Provider\" mechanism. Thus, a {@link SecureRandom} instance which utilizes 
 * 	this PRNG algorithm, and which, by default, is seeded from Qwicap's internal instance of this PRNG, may be obtained 
 * 	by using the following invocation: <code>SecureRandom.getInstance("Rule30", "Qwicap");</code>.
 * 	</p>
 * 
 *	@author Chris W. Johnson
 */
final class Rule30Random extends Random {

	private static final Logger		Log = Logger.getLogger(Rule30Random.class.getName());
	
	private static final int		kDefaultCellCount = 256;	//	Since we compute in 64-bit chunks, this might as well be a multiple of 64 (though that's not a requirement).
	
	private final Rule30			R30Impl;
	private boolean					Initialized;	//	This should NOT be "final". It is read before it can, or should, be set by our constructor, because the consructor of one of the superclasses invokes the setSeed method.


	Rule30Random() {
		this(kDefaultCellCount);
	}

	Rule30Random
	(
		final int				Width
	)
	{
		this(ensurePatternIsNonRepeating(Entropy.getBits(new byte[Width])));
		
	//	Perform an initial number of runs such that even a worst case input will have wrapped around the edges of the
	//	state array, thereby involving all of the bits of state in the output to some extent.

		for (int Count = Width; --Count >= 0; )
			R30Impl.runOnce();
	}
	
/**	Creates a <code>Rule30Random</code> object with the initial state of the automaton set to a copy of the supplied
 *  initial state. The initial state is accepted as-is, with no effort made to detect or correct initial states that
 * 	will produce repeating output sequences. Therefore, this constructor is not directly suitable for creating 
 * 	<code>Rule30Random</code> objects that will produce useful pseudo-random sequences. It is useful, however, if you
 * 	already have initial state that is known to be good, if you are writing a unit test, or if you just want to 
 * 	experiment with a Rule 30 automaton. 
 * 
 * @param InitialBits			The initial state of the automaton, in which each bit of the automaton's state is 
 * 								represented by a <code>byte</code> containing either the value <code>0</code>, or 
 * 								<code>1</code>.
 */
	Rule30Random
	(
		final byte[]			InitialBits	//	Each byte represents a bit, and therefore has either a value of 0 or 1.
	)
	{
		Log.fine("New Rule30Random instance created.");
		
	//	Note that we do not use the "ensurePatternIsNonRepeating" method on InitialBits, because this constructor is 
	//	used for JUnit tests, and in order for those tests to work, the initial state must be exactly what they set.
		
		final long[]			Seed = bitsToLongs(InitialBits);
		
		R30Impl     = InitialBits.length % 64 == 0 ? new Rule30WidthMultipleOf64(Seed) : new Rule30WidthArbitrary(InitialBits.length, Seed);
		Initialized = true;
	}
	
/**	The <code>Random</code> class supplies this method, which means that this implementation must deal with the 
 * 	possibility that someone will invoke it. This implementation of the method, however, does nothing except throw an 
 * 	exception telling you not to invoke it. The reason is that Qwicap internally depends on its random number generator 
 * 	<strong>not</strong> repeating itself. With the ability to reset the generator's seed, this assumption becomes 
 * 	untenable. 
 * 
 * @param SeedLong							A seed value which will be ignored.
 * @throws UnsupportedOperationException	This exception is always thrown when this method is invoked outside of
 * 											this class' constructor.
 */
	@Override public synchronized void setSeed(final long SeedLong) {

	//	We can't prevent the Random() constructor from calling setSeed, so we just agree to silently ignore calls to
	//	setSeed before this class' constructors have finished their work.

		if (Initialized)
			throw new UnsupportedOperationException("Re-seeding Qwicap's random number generator is likely to cause Qwicap to malfunction, therefore re-seeding is not supported.");
	}

/**	Returns the current state of this cellular automaton as a <code>String</code> of zero and one characters.
 * 
 * @return						The current state of this cellular automaton as a <code>String</code> zero and one 
 * 								characters.
 */
	@Override public String toString() {
		final int				BitCount = R30Impl.getBitCount();
		final char[]			Chars    = new char[BitCount];

		for (int Index = 0; Index < BitCount; Index++)
			Chars[Index] = (char) ('0' + R30Impl.readBit(Index));

		return new String(Chars);
	}

	@Override public int nextInt() {

		synchronized (R30Impl) {
			int						Bits = R30Impl.runOnce().readMiddleBit();

			for (int BitCount = 31; --BitCount >= 0; )
				Bits = (Bits << 1) | R30Impl.runOnce().readMiddleBit();

			return Bits;
		}
	}

	@Override public long nextLong() {

		synchronized (R30Impl) {
			long					Bits = R30Impl.runOnce().readMiddleBit();

			for (int BitCount = 63; --BitCount >= 0; )
				Bits = (Bits << 1) | R30Impl.runOnce().readMiddleBit();

			return Bits;
		}
	}

	@Override public int nextInt
	(
		final int				MaximumValue
	)
	{

	//	This implementation is based directly on the one shown in java.util.Random.nextInt(int)

		if (MaximumValue <= 0)
			throw new IllegalArgumentException("The maximum random value must be greater than zero; " + MaximumValue + " is not.");

		if ((MaximumValue & -MaximumValue) == MaximumValue)	//	i.e., n is a power of 2
			return nextPowerOf2(MaximumValue);

		synchronized (R30Impl) {
			int						Bits;
			int						Value;

			do {

				Bits  = nextBits(31);
				Value = Bits % MaximumValue;

			} while (Bits - Value + (MaximumValue - 1) < 0);

			return Value;
		}
	}

	@Override public boolean nextBoolean() {

		synchronized (R30Impl) {
			return R30Impl.runOnce().readMiddleBit() != 0;
		}
	}

	@Override public int next
	(
		final int				BitCount
	)
	{
		synchronized (R30Impl) {
			return nextBits(BitCount);
		}
	}

	@Override public void nextBytes
	(
		final byte[]			Bytes
	)
	{
		synchronized (R30Impl) {

			for (int Index = 0, Count = Bytes.length; Index < Count; Index++) {
				int						Bits = R30Impl.runOnce().readMiddleBit();

				for (int BitCount = 7; --BitCount >= 0; )
					Bits = (Bits << 1) | R30Impl.runOnce().readMiddleBit();

				Bytes[Index] = (byte) Bits;
			}
		}
	}
	
	Rule30Random runOnce() {
		R30Impl.runOnce();
		return this;
	}
	
	private int nextPowerOf2
	(
		int						MaximumValue	//	Must be a power of two.
	)
	{
		synchronized (R30Impl) {
			int						Bits = 0;

			MaximumValue >>>= 1;

			while (MaximumValue != 0) {

				Bits = (Bits << 1) | R30Impl.runOnce().readMiddleBit();

				MaximumValue >>>= 1;
			}

			return Bits;
		}
	}

	private int nextBits
	(
		int						BitCount
	)
	{
		if (BitCount < 1 || BitCount > 32)
			throw new IllegalArgumentException("The requested number of bits must be between 1 and 32 (inclusive). The number of bits requested was " + BitCount + '.');

		int						Bits = R30Impl.runOnce().readMiddleBit();

		while (--BitCount > 0)
			Bits = (Bits << 1) | R30Impl.runOnce().readMiddleBit();

		return Bits;
	}

	private static long[] bitsToLongs
	(
		final byte[]			Bits	//	Every byte represents a bit, and therefore has a value of 1 or 0.
	)
	{
		final int				BitCount = Bits.length;
		final long[]			Longs    = new long[(BitCount + 63) / 64];
		int 					BitIndex = 0;

		for (int LongIndex = 0, LongCount = Longs.length; LongIndex < LongCount; LongIndex++) {
			long					Long = 0;

			for (int BitCounter = 64; --BitCounter >= 0 && BitIndex < BitCount; )
				Long = (Long << 1) | Bits[BitIndex++];

			Longs[LongIndex] = Long;
		}
		
	//	Note: The behavior of placing undefined bits in the low-order positions is the opposite of the behavior one
	//	      would generally expect, though it is exactly the right behavior in this case. That's why this method is 
	//	      implemented privately here, rather than in the Bits class.
		
		final int				UnusedFinalBits = 64 - (BitCount % 64);

		if (UnusedFinalBits != 0)
			Longs[Longs.length - 1] <<= UnusedFinalBits;

		return Longs;
	}

	static byte[] ensurePatternIsNonRepeating
	(
		final byte[]			Bits	//	Every byte represents a bit, and therefore has a value of 1 or 0.
	)
	{
		if (isRepeatingPattern(Bits, true) == false)
			return Bits;

	//	The pattern repeats, so reverse bits in the pattern until it doesn't repeat.
	//
	//	Is this guaranteed to always solve the problem? I suspect so, but can't be sure. I have run a brute-force test 
	//	on every combination of 32 bits, and this approach was able eliminate repeating patterns from that set of 
	//	values.

		final int				BitCount     = Bits.length;
		int						BitToReverse = 0;

		do {

			Bits[BitToReverse] = (byte) (Bits[BitToReverse] == 0 ? 1 : 0);

		} while (isRepeatingPattern(Bits, true) && ++BitToReverse < BitCount);

		return Bits;	//	Every byte represents a bit, and therefore has a value of 1 or 0.
	}

	private static boolean isRepeatingPattern
	(
		final byte[]			Bits,	//	Every byte represents a bit, and therefore has a value of 1 or 0.
		final boolean			ConsiderAllBits
	)
	{
		for (int WindowLength = 1, WindowLengthLimit = Bits.length >>> 1; WindowLength <= WindowLengthLimit; WindowLength++) {

			if (patternRepeats(Bits, WindowLength, ConsiderAllBits)) {
				
				if (Log.isLoggable(Level.FINE)) {
					
					Log.log(
						Level.FINE, 
						"There is a repeating pattern of {1,choice,1#one bit|1<{1,number,integer} bits} within the bits: {0}.", 
						new Object[] { 
							edu.utexas.its.eis.tools.qwicap.servlet.Bits.toBinaryString(Bits), 
							WindowLength 
						}
					);	
				}
				return true;
			}
		}

		return false;
	}

	private static boolean patternRepeats
	(
		final byte[]			Bits,	//	Every byte represents a bit, and therefore has a value of 1 or 0.
		final int 				WindowLength,
		final boolean			ConsiderAllBits
	)
	{
		final int				StateLength = Bits.length;

	//	Repetitions that matter to the rule 30 automaton probably aren't possible unless the length of State is a whole 
	//	multiple of WindowLength, so if WindowLength doesn't divide evenly into StateLength, we can conclude that 
	//	repetition doesn't exist at this scale.

		if (ConsiderAllBits && StateLength % WindowLength != 0)
			return false;

		for (int StartIndex = WindowLength, WindowCount = StateLength / WindowLength; --WindowCount > 0; StartIndex += WindowLength) {
			int						MasterIndex = 0;
			int						ReadIndex   = StartIndex;

			for (int ByteCount = WindowLength; --ByteCount >= 0;) {

				if (Bits[MasterIndex++] != Bits[ReadIndex++])
					return false;
			}
		}

//		final int				RemainingState = StateLength % WindowLength;
//		
//		if (RemainingState != 0) {
//			int						MasterIndex = 0;
//			int						ReadIndex = StateLength - RemainingState;
//			
//			for (int ByteCount = WindowLength; --ByteCount >= 0;) {
//				
//				if (State[MasterIndex++] != State[ReadIndex++])
//					return false;
//				
//				if (ReadIndex >= StateLength)
//					ReadIndex = 0;
//			}
//		}

		return true;
	}
}
