/*	QWICAP (Quick Web Interface for Conventional Applications) by Chris W. Johnson
	Copyright (C) 2003-2008 The University of Texas at Austin ("U.T. Austin"). All rights reserved.
	
	U.T. Austin's copyright notice in this evolving version reflects our original authorship. It does not assert 
	copyright over changes contributed by others.
	
	This library is free software; you can redistribute it and/or modify it under the terms of the GNU Lesser General 
	Public License as published by the Free Software Foundation; either version 2.1 of the License, or (at your option) 
	any later version.
	
	This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied 
	warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for 
	more details.
	
	You should have received a copy of the GNU Lesser General Public License along with this library; if not, write to 
	the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/


package edu.utexas.its.eis.tools.qwicap.servlet;


import edu.utexas.its.eis.tools.table.Table;
import edu.utexas.its.eis.tools.table.MutableTable;
import edu.utexas.its.eis.tools.table.cellformatter.CellNumberFormatter;
import edu.utexas.its.eis.tools.table.cellformatter.CellAutoByteFormatter;
import edu.utexas.its.eis.tools.table.lltable.MutableLLTable;

import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.logging.Logger;
import java.util.logging.Level;


/**
 * Created by IntelliJ IDEA.
 * User: chrisj
 * Date: Oct 5, 2006
 * Time: 5:08:40 PM
 * To change this template use File | Settings | File Templates.
 */
final class ServiceDataRecorder {

	private static final Logger		Log = Logger.getLogger(ServiceDataRecorder.class.getName());
	
	private QwicapThreadPool		QTP;
	private DocumentCache			DocCache;
	
	private long					PeriodDurationMillis;
	private int						PeriodCount;

	private final Object			DataSynchronizer = new Object();	//	Used to synchronize all access to the "Data" instance variable.
	private ServiceData				Data;
	private long					StopMillis;
	
	private String					Label;
	

	ServiceDataRecorder
	(
		final long				PeriodDurationMillis,
		final int				PeriodCount
	)
	{
	//	todo Be aware that the date formatter stored in kDateFmt assumes that our period duration is one hour. If we
	//	     really want to support other durations, we need to do something like accept the date format as a parameter.

		resetLow(PeriodDurationMillis, PeriodCount);
	}
	
	void setThreadPool
	(
		final QwicapThreadPool	QTP
	)
	{
		this.QTP = QTP;
	}
	
	void setDocumentCache
	(
		final DocumentCache		DocCache
	)
	{
		this.DocCache = DocCache;
	}
	
	void setLabel
	(
		final String			Label
	)
	{
		this.Label = Label;
	}
	
	int getActiveThreadCount() {
		
		synchronized (DataSynchronizer) {
			return getCurrentData().getThreadActiveCountInfo().getCount();
		}
	}
	
	void clear() {
		synchronized (DataSynchronizer) {
			getCurrentData().PrevData = null;	//	Retain the data for the current hour, but discard all previous data.
		}
	}
	
	void reset
	(
		final long				PeriodDurationMillis,
		final int				PeriodCount
	)
	{
	//	todo Be aware that the date formatter stored in kDateFmt assumes that our period duration is one hour. If we
	//	     really want to support other durations, we need to do something like accept the date format as a parameter.

		synchronized (DataSynchronizer) {
			resetLow(PeriodDurationMillis, PeriodCount);
		}
	}

	private void resetLow
	(
		final long				PeriodDurationMillis,
		final int				PeriodCount
	)
	{
		this.PeriodDurationMillis = PeriodDurationMillis;
		this.PeriodCount          = PeriodCount;

		final long				CurTime     = System.currentTimeMillis();
		final long				StartMillis = CurTime - CurTime % PeriodDurationMillis;
		final long				StopMillis  = StartMillis + PeriodDurationMillis;

		synchronized (DataSynchronizer) {

			Data = new ServiceData(StartMillis, StopMillis, Data);
			Data.init(QTP);
			
			this.StopMillis = StopMillis;

		//	Truncate the list, such that the old ThreadData objects are discarded.

			Data.PrevData = null;
		}
	}

	void threadUnavailable() {
		
		try {
			
			synchronized (DataSynchronizer) {
				getCurrentData().threadUnavailable();
			}
			
		} catch (Exception e) {
			Log.log(Level.WARNING, "The service data recorder experienced an internal failure: " + e.getMessage(), e);
		}
	}

	void threadBorn
	(
		final int				ActiveThreads,
		final int				InactiveThreads
	) 
	{
		try {
			
			synchronized (DataSynchronizer) {
				getCurrentData().threadBorn(ActiveThreads, InactiveThreads);
			}
			
		} catch (Exception e) {
			Log.log(Level.WARNING, "The service data recorder experienced an internal failure: " + e.getMessage(), e);
		}
	}

	void threadDead
	(
		final int				ActiveThreads,
		final int				InactiveThreads
	)
	{
		try {
		
			synchronized (DataSynchronizer) {
				getCurrentData().threadDead(ActiveThreads, InactiveThreads);
			}
			
		} catch (Exception e) {
			Log.log(Level.WARNING, "The service data recorder experienced an internal failure: " + e.getMessage(), e);
		}
	}

	void threadActive
	(
		final int				ActiveThreads,
		final int				InactiveThreads
	)
	{
		try {
		
			synchronized (DataSynchronizer) {
				getCurrentData().threadActive(ActiveThreads, InactiveThreads);
			}
			
		} catch (Exception e) {
			Log.log(Level.WARNING, "The service data recorder experienced an internal failure: " + e.getMessage(), e);
		}
	}

	void threadInactive
	(
		final int				ActiveThreads,
		final int				InactiveThreads,
		final long				SessionMillis,
		final boolean			Expired
	)
	{
		try {
			
			synchronized (DataSynchronizer) {
				
				getCurrentData().threadInactive(ActiveThreads, InactiveThreads, SessionMillis, Expired);
				
				if (getActiveThreadCount() == 0 && Log.isLoggable(Level.FINE))
					Log.fine("No active threads. Qwicap status for " + Label + ":\n" + getGatheredData(false) + "\n\nDocument Cache Status (Discards: " + DocCache.getDiscardCount(false) + "):\n" + DocCache.getStatus(false));
			}
			
		} catch (Exception e) {
			Log.log(Level.WARNING, "The service data recorder experienced an internal failure: " + e.getMessage(), e);
		}
	}
	
	void threadJammed() {
		
		try {
			
			synchronized (DataSynchronizer) {
				getCurrentData().threadJammed();
			}
			
		} catch (Exception e) {
			Log.log(Level.WARNING, "The service data recorder experienced an internal failure: " + e.getMessage(), e);
		}
	}

	void hit
	(
		final long				StartMillis
	)
	{
		try {
	
			synchronized (DataSynchronizer) {
				getCurrentData().hit(StartMillis);
			}
			
		} catch (Exception e) {
			Log.log(Level.WARNING, "The service data recorder experienced an internal failure: " + e.getMessage(), e);
		}
	}

	void badDataSet() {
	
		try {
			
			synchronized (DataSynchronizer) {
				getCurrentData().badDataSet();
			}
			
		} catch (Exception e) {
			Log.log(Level.WARNING, "The service data recorder experienced an internal failure: " + e.getMessage(), e);
		}
	}
	
	private ServiceData getCurrentData() {
		final long				CurTime    = System.currentTimeMillis();
		long					StopMillis = this.StopMillis;

		if (CurTime < StopMillis)
			return Data;

		final long				DurationMillis = PeriodDurationMillis;
		long					StartMillis;

		do {
			
			StartMillis = StopMillis;
			StopMillis += DurationMillis;
			
			Data = new ServiceData(StartMillis, StopMillis, Data);
			
		} while (CurTime >= StopMillis);

		this.StopMillis = StopMillis;
		
		pruneDataList();
		
	//	Setup the initial values of the new ServiceData object that describes the current time interval. Note
	//	that any ServiceData objects created in the loop above that describe hours when nothing happened will not
	//	have their init methods invoked. This is intentional. It results in a lot of counters reading "n/a", which
	//	is accurate, since no data was gathered during that period.
		
		Data.init(QTP);
		
	//	Log the current service data, omitting the data for the period that has just begun, since it will be empty.
		
		if (Log.isLoggable(Level.FINE))
			Log.log(Level.FINE, "Qwicap \"new hour\" status for {0}:\n{1}", new Object[] { Label, getGatheredDataLow(Data, true) });	//	We can't use getGatheredData, because it invokes getCurrentData.
		
		return Data;
	}

	private void pruneDataList() {

	//	Note that this method will always leave at least the first element of the list. This is intentional.

		int						Count = PeriodCount;

		for (ServiceData Data = this.Data; Data != null; Data = Data.PrevData) {

			if (--Count == 0) {

				Data.PrevData = null;
				return;
			}
		}
	}
	
	Table getGatheredData
	(
		final boolean			Clear
	)
	{
		synchronized (DataSynchronizer) {
			final Table				GDT = getGatheredDataLow(getCurrentData(), false);
			
			if (Clear)
				clear();
			
			return GDT;
		}
	}
	
	private static Table getGatheredDataLow
	(
		ServiceData				Data,
		final boolean			OmitMostRecentData
	) 
	{
		final SimpleDateFormat	DateFmt   = new SimpleDateFormat("yyyy-MM-dd KK a");
		final long				MaxMemory = Runtime.getRuntime().maxMemory();
		final MutableTable		T         = new MutableLLTable();

	//	We're producing output that looks like this:
	//
	//					   +---------+---------+----------+-----------+-----------+------------+-----------+----------+----------------+-------+------+-----------------+-------------+-------------+-----------+
	//					   | Threads | Threads | Threads  | Threads   | Threads   | Threads    | Sessions  | Sessions | Session Length | Hits  | Hits | Hit Runtime     | RAM Free MB | RAM Used MB | RAM Total |
	//					   | Total   | Active  | Inactive | Born/Died | Jammed    | Unvailable | Completed | Expired  | Min/Avg/Max s. | Total | Bad  | Min/Avg/Max ms. | Min/Avg/Max | Min/Avg/Max |           |
	//	+------------------+---------+---------+----------+-----------+-----------+------------+-----------+----------+----------------+-------+------+-----------------+-------------+-------------+-----------+
	//	| 2008-04-16 08 PM |     0-1 |     0-1 |      0-0 |       1/1 |         0 |          0 |         1 |        0 |           13.0 |    20 |    0 |   8.0/18.2/72.0 | 0.2/2.2/4.0 | 8.5/9.1/9.8 | 396.94 MB |
	//	| 2008-04-16 09 PM |     0-0 |     0-0 |      0-0 |       0/0 |         0 |          0 |         0 |        0 |            N/A |     0 |    0 |             N/A | 2.8/2.8/2.8 | 9.8/9.8/9.8 | 396.94 MB |
	//	+------------------+---------+---------+----------+-----------+-----------+------------+-----------+----------+----------------+-------+------+-----------------+-------------+-------------+-----------+
		
		T.addCol("Threads\nTotal",                 new NumericArrayCellFormatter("#,##0", "-", true, "N/A"))
		 .addCol("Threads\nActive",                new NumericArrayCellFormatter("#,##0", "-", true, "N/A"))
		 .addCol("Threads\nInactive",              new NumericArrayCellFormatter("#,##0", "-", true, "N/A"))
		 .addCol("Threads\nBorn/Died",             new NumericArrayCellFormatter("#,##0", "/", true, "N/A"))
		 .addCol("Threads\nJammed",                new NumericArrayCellFormatter("#,##0", "-", true, "N/A"))
		 .addCol("Threads\nUnvailable",            new CellNumberFormatter("#,##0"))
		 .addCol("Sessions\nCompleted",            new CellNumberFormatter("#,##0"))
		 .addCol("Sessions\nExpired",              new CellNumberFormatter("#,##0"))
		 .addCol("Session Length\nMin/Avg/Max s.", new NumericArrayCellFormatter("#,##0.0", "/", true, "N/A"))
		 .addCol("Hits\nTotal",                    new CellNumberFormatter("#,##0"))
		 .addCol("Hits\nBad",                      new CellNumberFormatter("#,##0"))
		 .addCol("Hit Runtime\nMin/Avg/Max ms.",   new NumericArrayCellFormatter("#,##0.0", "/", true, "N/A"))
		 .addCol("RAM Free MB\nMin/Avg/Max",       new NumericArrayCellFormatter("#,##0.0", "/", true, "N/A"))
		 .addCol("RAM Used MB\nMin/Avg/Max",       new NumericArrayCellFormatter("#,##0.0", "/", true, "N/A"))
		 .addCol("RAM Total",                      new CellAutoByteFormatter());
		
		if (OmitMostRecentData)
			Data = Data.PrevData;
		
		for (; Data != null; Data = Data.PrevData) {

			T.addRow(DateFmt.format(new Date(Data.getDataStartTimeMillis()))).addAll(
				Data.getThreadCountInfo().toRangeArray(),
				Data.getThreadActiveCountInfo().toRangeArray(),
				Data.getThreadInactiveCountInfo().toRangeArray(),
				new int[] { Data.getThreadBirthCount(), Data.getThreadDeathCount() },
				Data.getThreadJammedCountInfo().toRangeArray(),
				Data.getThreadUnavailableCount(),
				Data.getSessionCompletedCount(),
				Data.getSessionExpiredCount(),
				Data.getSessionTotalMillisInfo().toMinAvgMaxArray(1000.0),
				Data.getHitCount(),
				Data.getBadHitCount(),
				Data.getHitRunMillisInfo().toMinAvgMaxArray(1.0),
				Data.getMemoryFreeInfo().toMinAvgMaxArray(1024 * 1024),
				Data.getMemoryTotalInfo().toMinAvgMaxArray(1024 * 1024),
				MaxMemory
			);
		}

		return T.reverseRows();
	}
}
