/*	QWICAP (Quick Web Interface for Conventional Applications) by Chris W. Johnson
	Copyright (C) 2003-2008 The University of Texas at Austin ("U.T. Austin"). All rights reserved.
	
	U.T. Austin's copyright notice in this evolving version reflects our original authorship. It does not assert 
	copyright over changes contributed by others.
	
	This library is free software; you can redistribute it and/or modify it under the terms of the GNU Lesser General 
	Public License as published by the Free Software Foundation; either version 2.1 of the License, or (at your option) 
	any later version.
	
	This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied 
	warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for 
	more details.
	
	You should have received a copy of the GNU Lesser General Public License along with this library; if not, write to 
	the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/


//  SessionContext.java
//  qwicap
//
//  Created by Chris W. Johnson on Wed Jun 04 2003.


package edu.utexas.its.eis.tools.qwicap.servlet;


import javax.servlet.http.*;
import java.io.IOException;
import java.net.URL;
import java.net.URLDecoder;
import java.util.logging.Level;
import java.util.logging.Logger;


/**	An instance of <code>SessionContext</code> is associated with every Qwicap session. It coordinates session activity,
 * 	stores the information about the Qwicap session that survives from one hit to the next, and provides various utility 
 *	methods.
 *	
 *	@author Chris W. Johnson
 */
final class SessionContext {
	
	private static final Logger				Log = Logger.getLogger(SessionContext.class.getName());
	
	private static final String				kSessionContextAttributeNameRoot             = "qwicap-session-context";
	private static final String				kSessionContextAutoTearDownAttributeNameRoot = "qwicap-session-context-auto-tear-down-";

	private final QwicapApplicationServices	AppServices;
	private final SessionID					SessionID;
	private HttpSession						Sess;
	private final String					SessCtxtAttrName;	//	The name of the session attribute that holds a reference to this SessionContext object.
	private QwicapThread					QThread;
	private final Qwicap					Q;
	private final String					ContextPath;
	private final Class<?>					ClientClass;		//	Used only to provide informative logging messages.
	private boolean							FirstHit = true;
	private boolean							Closed;
	

/**	Get the Qwicap <code>SessionContext</code> object associated with the requested servlet in the current session.
 * 
 * @param Req					The current HTTP request. From this, we get the servlet identifiers, and the current 
 * 								session (if any). 
 * @return						The relevant <code>SessionContext</code> object, or <code>null</code>.
 */
	static SessionContext get
	(
		final HttpServletRequest	Req
	)
	{
		final HttpSession			Sess = Req.getSession(false);	//	Get the session, if any. Do not create a new session.
		
		if (Sess != null)
			return (SessionContext) Sess.getAttribute(createSessionAttributeName(Req.getContextPath(), kSessionContextAttributeNameRoot));
		
		return null;
	}
	
	static void sendDownloadable
	(
		final HttpServletRequest	Req,
		final HttpServletResponse	Resp,
		final String				RequestURI
	)
	throws IOException
	{
		final String			RelativeURI = Downloadable.simplifyURI(URLDecoder.decode(RequestURI, "UTF-8"));
		final SessionContext	Ctxt        = get(Req);
		final Downloadable		Download;
		
		if (Ctxt != null) {
			
		//	The relevant Qwicap application is still running, so we get the downloadable in the normal manner.
			
			Download = Ctxt.getDownloadable(RelativeURI);
			
		} else {
			
		//	The relevant Qwicap application exited some time prior to this request. If it had any downloadables when it
		//	exited, they may still be available from a LingeringDownloadables object stored in a session attribute. 
		//	(Typically, that attribute is deleted something like 30 seconds after the application exits, which should 
		//	be plenty of time for the client that received the application's final page to request any embedded 
		//	downloadables, like images. It doesn't matter whether or not the downloads have actually completed by the 
		//	time the session attribute is deleted; as long they were begun before then, they'll complete normally, and 
		//	then the downloadables will become available for garbage collection.)
			
			Download = LingeringDownloadables.getDownloadable(Req, RelativeURI);
		}
		
	//	If we found the requested donwloadable (no matter how we found it), send it to the client. (We're satisfying an
	//	HTTP GET request here.)
		
		if (Download != null)
			Download.send(Resp);
		else
			Resp.sendError(HttpServletResponse.SC_NOT_FOUND, "Downloadable \"" + RelativeURI + "\" not found.");
	}
	
	static boolean sendDocument
	(
		final HttpServletRequest	Req,
		final HttpServletResponse	Resp,
		final URL					DocURL
	)
	throws IOException
	{
		final SessionContext	Ctxt = get(Req);
		
		if (Ctxt != null) {
			final DocumentSource	Doc = Ctxt.getQwicap().geti18n().getDocument(DocURL);
			
			if (Doc.exists()) {
				
				Doc.writeTo(Resp);
				return true;	//	Sent.
			}
		}
		
		return false;	//	Not sent.
	}
	
	static boolean sendDocument
	(
		final HttpServletRequest	Req,
		final HttpServletResponse	Resp,
		final String				DocName		//	The servlet path from the request.
	)
	throws IOException
	{
		final SessionContext	Ctxt = get(Req);
		
		if (Ctxt != null) {
			final DocumentSource	Doc = Ctxt.getQwicap().geti18n().getDocument(DocName);
			
			if (Doc.exists()) {
				
				Doc.writeTo(Resp);
				return true;	//	Sent.
			}
		}
		
		return false;	//	Not sent.
	}
	
	SessionContext
	(
		final QwicapGlobalServices		GlobalServices,
		final QwicapApplicationServices	ApplicationServices,
		final HitContext				Hit
	)
	throws NoSuchMethodException, QwicapThreadPoolException, IOException
	{
		AppServices = ApplicationServices;
		SessionID   = new SessionID(GlobalServices.getRandom());
		ClientClass = Hit.getServlet().getClientClass();
		
		Log.log(Level.FINEST, "SCTXT: Creating SessionContext for application \"{0}\".", ClientClass);
		
		final HttpServletRequest	Req = Hit.getRequest();
		
		ContextPath      = Req.getContextPath();
		Sess             = Req.getSession();	//	Create the session, if it doesn't exist already.
		SessCtxtAttrName = setSessionAttribute(kSessionContextAttributeNameRoot, this);
		
	//	Register a SessionContextAutoTearDown instance as a session attribute. That instance is responsible for closing
	//	this SessionContext instance when the session expires or is invalidated.
		
		setSessionAttribute(kSessionContextAutoTearDownAttributeNameRoot + SessionID, new SessionContextAutoTearDown(this));
		
	//	Remove any lingering downloadables that still exist in this session from previous runs of this servlet. 
	//	We can safely assume the client won't be asking for them again, now that it has re-invoked this servlet.
		
		LingeringDownloadables.remove(this);
		
	//	Setup and start the Qwicap thread that will run the client code.
		
		Log.log(Level.FINEST, "SCTXT: Starting application \"{0}\".", ClientClass);
		
		QThread = ApplicationServices.getThreadPool().startApplication(GlobalServices, ApplicationServices, this, Hit);
		Q       = QThread.getApplication();
		
		Log.log(Level.FINE, "SCTXT: Created SessionContext \"{1}\" for application \"{0}\".", new Object[] { ClientClass, SessionID });
	}
	
	QwicapApplicationServices getApplicationServices() {
		return AppServices;
	}
	
	HttpSession getSession() {
		return Sess;
	}
	
	Object getSessionAttribute
	(
		final String			Name
	)
	{
		if (Sess == null)
			throw new IllegalStateException("There is no session at this time.");
		
		return Sess.getAttribute(createSessionAttributeName(ContextPath, Name));
	}
	
	String setSessionAttribute
	(
		final String			Name,
		final Object			Value
	)
	{
		if (Sess == null)
			throw new IllegalStateException("There is no session at this time.");
		
		final String			QualifiedName = createSessionAttributeName(ContextPath, Name);
		
		Sess.setAttribute(QualifiedName, Value);
		
		return QualifiedName;
	}
	
	void removeSessionAttribute
	(
		final String			Name
	)
	{
		if (Sess == null)
			throw new IllegalStateException("There is no session at this time.");
		
		Sess.removeAttribute(createSessionAttributeName(ContextPath, Name));
	}
	
	static String createSessionAttributeName
	(
		final String			ContextPath,
		final String			Name
	)
	{
		return ContextPath + ':' + Name;
	}
	
/*	We don't want web applications that contain multiple servlets to be overwriting each other's session attributes, so
 *	we have a naming convention that incorporates the name of a servlet into its attribute names. This method creates
 *	those attribute names. Given the extremely state-rich environment that Qwicap provides to client applications, they
 *	shouldn't need this method, but maybe it'll prove necessary to someone.
 * 
 * @param BasicName				The name to which this method will add servlet-specific components.
 * @return						The servlet-specific attribute name.
 *
	String getServletSpecificSessionAttributeName
	(
		final String			BasicName
	)
	{
		return getServletSpecificSessionAttributeName(QThread.getHitContext().getRequest(), BasicName);
	}
*/	
	
	Qwicap getQwicap() {
		return Q;
	}

	Downloadable getDownloadable
	(
		final String			URI
	)
	{
		return getQwicap().getPageStack().getDownloadable(URI);
	}
	
	synchronized void go
	(
		final HitContext		Hit
	)
	{
		final long				HitStartMillis = System.currentTimeMillis();
		
		Log.log(Level.FINEST, "GO: Entered.");
		
		try {
			
			if (FirstHit) {
				
			//	Wait for processing of the first hit to complete.
			
				FirstHit = false;
				
				Log.log(Level.FINEST, "GO: First hit.");
				
				Hit.waitForProcessingToComplete();
				Q.HitCtxt = null;
				
			} else if (QThread != null && QThread.isAlive()) {
				
				Log.log(Level.FINEST, "GO: Attempting to synchronize on Q.");
				
				synchronized (Q) {
				
					Log.log(Level.FINEST, "GO: Synchronized on Q.");
					
					Q.HitCtxt    = Hit;
					Q.ShouldWait = false;	//	This avoids potential "Wait Leaks". See: http://www-128.ibm.com/developerworks/java/library/j-perf01215/index.html
					Q.notify();
					
					Log.log(Level.FINEST, "GO: Qwicap thread notified that input has arrived.");
				
				//	Note that the Qwicap thread that we've just notified can't begin executing again until this
				//	"synchronized" block exits.
				}
				
				Hit.waitForProcessingToComplete();
				Q.HitCtxt = null;
				
			} else
				Log.log(Level.WARNING, "GO: Thread is already dead. Can't reinvoke it!");
			
		} catch (InterruptedException e) {
			
		//	Qwicap.Log.log(Level.FINEST, e.getMessage(), e);
			
		} finally {
			final QwicapThread		QThread = this.QThread;	//	The close() method is going to null this reference, so make a copy of it.
			
		//	If the Qwicap thread has died, tear-down this session.
						
			if (Q.isDone())
				try { close(false); } catch (Exception e) { Log.log(Level.WARNING, "SessionContext.close(false) failed.", e); }
			
		//	We invoke hitDone last in order to capture the cost of the "close" invocation in the hit timing.
			
			if (QThread != null)
				QThread.hitDone(HitStartMillis);
			
			Log.log(Level.FINEST, "GO: Exited.");
		}
	}
	
	synchronized void close
	(
		final boolean			ForceImmediateSessionDestruction
	)
	{
		if (Closed)
			return;
		
	//	We usually arrive here because the Qwicap.run method has exited (in other words the Qwicap instance has already
	//	died), in which case the following "pleaseDie" invocation is redundant, but harmless. However, we can arrive 
	//	here before the Qwicap instance has been told to die - in that case, this invocation of "pleaseDie" is vital.
	//	The only known case in which pleaseDie won't have been invoked when we arrive here is the expiration of a 
	//	session - in which case this method is invoked from the SessionContextAutoTearDown instance present in the
	//	session's attribute list.
		
		if (Q.isDone() == false)
			Q.pleaseDie("Session expired or invalidated.");
		
		final String			SessionDiesWhen;
		final HttpSession		Sess = this.Sess;
		
		if (Sess != null) {
		
			if (ForceImmediateSessionDestruction) {
				
			//	Note that invalidating the session at this point is only valid while Qwicap remains limited to one 
			//	application in each WAR.
				
				Log.fine("Invalidating session (1).");
				
				try {
					
					Closed = true;
					Sess.invalidate();
					
				} catch (Exception e) {
					Log.log(Level.FINEST, "Session has probably already been invalidated. (1)", e);
				} finally {
					this.Sess = null;
				}
				
				SessionDiesWhen = "The HttpSession has been invalidated. (1)";
				
			} else {
				final boolean			HasLingeringDownloadables = LingeringDownloadables.install(this);
				
			//	Remove the session attribute that holds a reference to this SessionContext instance, so that this 
			//	context can die, and (potentially) so that the user can re-run this web application, without it 
			//	tripping over a dead context.
				
				Sess.removeAttribute(SessCtxtAttrName);
				
			//	Note that invalidating the session at this point is only valid while Qwicap 1.4 remains limited to one 
			//	application in each WAR.
				
				if (HasLingeringDownloadables) {
					
				//	Give the client 30 more seconds to begin downloading any downloadables on the final page.
					
					Sess.setMaxInactiveInterval(30);
					SessionDiesWhen = "HttpSession set to expire in 30 seconds, due to lingering downloadables.";
					
				} else {
					
					Log.fine("Invalidating session (2).");
					
					try {
						
						Closed = true;
						Sess.invalidate();
						
					} catch (Exception e) {
						Log.log(Level.FINEST, "Session has probably already been invalidated (2).", e);
					} finally {
						this.Sess = null;
					}
					
					SessionDiesWhen = "The HttpSession has been invalidated. (2)";
				}
			}
			
		} else
			SessionDiesWhen = "No HttpSession exists; it must have been invalidated earlier.";

		Log.log(Level.FINE, "Disposed of SessionContext \"{2}\" for application \"{0}\". {1}", new Object[] { ClientClass, SessionDiesWhen, SessionID });
		
		QThread = null;
		
//		if (getSessionContextCount() <= 1) {
//			
//			if (getSessionLingeringDownloadablesCount() == 0) {
//				
//			//	SessInvalidated = true;
//			//	Sess.invalidate();			//	If we clean out our session attributes properly, do we need this?
//				
//				
//			//	removeAllSessionAttributes();	//	This removes even SessionContext objects that were added after closeALittleLater was called, thereby killing new runs of the application. This is bad.
//				
//				System.out.println("Closed completely");
//				
//				Qwicap.Log.log(Level.FINEST, "CLOSE: Disposing of a Qwicap thread's context and session. (Other Qwicap threads in this session: {0})", new Integer(getSessionContextCount() - 1));
//			
//			} else
//				System.out.println("Couldn't close completely, because there were lingering downloadables.");
//			
//		} else {
//			
//			System.out.println("Partial close: This context was removed.");
//			
//			Sess.removeAttribute(SessCtxtAttrName);
//			
//			Qwicap.Log.log(Level.FINEST, "CLOSE: Disposing of a Qwicap thread's context. The session will persist until the other Qwicap threads in it ({0}) have completed.", new Integer(getSessionContextCount() - 1));
//		}


//		Notify any waiting QwicapServlet threads that the hits they are waiting to have processed, have been processed.
//		In this case, we've actually done nothing with the hits, but because the session has shutdown, there's nothing
//		sensible we could do the hits, except, perhaps, return a canned error message, or the final page/redirect sent
//		out by this session.
//		
//		HitContext				Hit = HitStack.pop();
//		
//		while (Hit != null) {
//			
//		//	Consider sending something back to lingering connections like an error page, or the "goodbye" page/redirect.
//			
//			Hit.processingComplete();
//			Hit = HitStack.pop();
//		}
	}
	
	
	
/*	private static String getSessionContextAttributeName
	(
		final HttpServletRequest	Req
	)
	{
		return getServletSpecificSessionAttributeName(Req, kSessionContextAttributeNamePrefix);
	}
	
/*	We don't want web applications that contain multiple servlets to be overwriting each other's session attributes, so
 *	we have a naming convention that incorporates the name of a servlet into its attribute names. This method creates
 *	those attribute names. Given the extremely state-rich environment that Qwicap provides to client applications, they
 *	shouldn't need this method, but maybe it'll prove necessary to someone.
 * 
 * @param Req					The current HTTP request. From this, we get the servlet identifiers. 
 * @param BasicName				The name to which this method will add servlet-specific components.
 * @return						The servlet-specific attribute name.
 *
	static String getServletSpecificSessionAttributeName
	(
		final HttpServletRequest	Req,
		final String				BasicName
	)
	{
		if (Req == null)
			throw new IllegalArgumentException("The HttpServletRequest must not be null, but is.");
		
		return Req.getContextPath() + ':' + BasicName;	//	Will this change prevent multiple Qwicap servlets from operating in the same web application?
		
	//	Pre 1.3.3 implementation. This didn't allow downloads to find their session when the servlet-mapping's 
	//	url-pattern was especially greedy, like when it was "/", because the URI for the download would be regarded as 
	//	the servlet path, and thus the attribute name this method generated wouldn't correspond to the name of the 
	//	session attribute that stored its relevant context.
	//	
	//	return BasicName + ':' + Req.getContextPath() + Req.getServletPath();
	}
*/	
//	private static boolean isSessionContextAttributeName
//	(
//		final String			AttrName
//	)						
//	{
//		return AttrName.startsWith(kSessionContextAttributeNamePrefixWithSeparator);
//	}
	
//	private static String getCookie
//	(
//		final HttpServletRequest	Req,
//		final String				Name
//	)
//	{
//		final Cookie[]			Cooks = Req.getCookies();
//		
//		for (int Index = 0, Count = Cooks.length; Index < Count; Index++) {
//			final Cookie			C = Cooks[Index];
//			
//			if (Name.equals(C.getName()))
//				return C.getName() + "=" + C.getValue();
//		}
//		
//		return "";
//	}


	
//	private void closeALittleLater() {
//		final String			FinalDownloablesSessAttrName = getSessionFinalDownloadablesAttributeName(Req);
//		
//	//	Get a complete list of downloadables from all of the pages still "in play". The final page might, conceivably,
//	//	ask for any of them.
//		
//		final Hashtable			AllDownloadables = QThread.getPageStack().getAllDownloadables();
//		
//	//	Store the list as a session attribute, in a way that the QwicapServlet will recognize.
//		
//		Sess.setAttribute(FinalDownloablesSessAttrName, AllDownloadables);
//		
//	//	Schedule a timer task to delete the session attribute containing the list of downloadables, so that it won't 
//	//	linger for very long on the server.	Eventually, the list will go away when the session expires (or is 
//	//	invalidated), but expiration could take ten minutes or more. I think we can reasonably do without the list of
//	//	downloadables after something like 30 seconds, since they only have to hang around long enough to give the
//	//	client a chance to ask for any that are embedded in the final page (e.g. graphics referenced by "img" tags).
//		
//		DeferredCloseTimer.schedule(new RemoveLingeringDownloadablesTask(Sess, FinalDownloablesSessAttrName), kFinalDownloadDelaySecs * 1000);
//		
//	//	Remove the session attribute that holds a reference to this context, so that this context can die, and 
//	//	(potentially) so the user can re-run this web application, without it tripping over a dead context.
//		
//		Sess.removeAttribute(SessCtxtAttrName);
//		
//		System.out.println("Because there were lingering downloadables when this Qwicap client application exited, final closing has been deferred.");
//		
//	//	Qwicap.Log.log(Level.FINEST, "CLOSE: Qwicap thread died. Because its current page includes downloadables, disposal of its context will be delayed {0} seconds to allow downloads to begin.");
//	}
	
	
	
//	private void removeAllSessionAttributes() {
//		final Enumeration 		Names = Sess.getAttributeNames();
//		final ArrayList			NameList = new ArrayList();		//	Contains String objects.
//		
//		while (Names.hasMoreElements())
//			NameList.add(Names.nextElement());
//		
//		for (int Index = 0, Count = NameList.size(); Index < Count; Index++)
//			Sess.removeAttribute((String) NameList.get(Index));
//	}
	
	
	
//	private int getSessionAttributeCount
//	(
//		final String			NonServletSpecificName
//	) 
//	{
//		int						AttrCount = 0;
//		
//		if (SessInvalidated == false) {
//			final String			Prefix = NonServletSpecificName + kSessionAttributeNameSeparator;
//			final Enumeration		AttrNames = Sess.getAttributeNames();
//			
//			
//			while (AttrNames.hasMoreElements()) {
//				final String			AttrName = (String) AttrNames.nextElement();
//				
//				if (AttrName.startsWith(Prefix))
//					AttrCount++;
//			}
//		}
//		
//		return AttrCount;
//	}
//	
//	private int getSessionLingeringDownloadablesCount() {
//		return getSessionAttributeCount(kFinalDownloadablesSessAttrNamePrefix);
//	}
//	
//	private int getSessionContextCount() {
//		return getSessionAttributeCount(kSessionContextAttributeNamePrefix);
//	}
}
