/*	QWICAP (Quick Web Interface for Conventional Applications) by Chris W. Johnson
	Copyright (C) 2006-2010 Chris W. Johnson. All rights reserved.
	
	Chris W. Johnson's copyright notice in this evolving version reflects his original authorship. It does not assert 
	copyright over changes contributed by others.
	
	This library is free software; you can redistribute it and/or modify it under the terms of the GNU Lesser General 
	Public License as published by the Free Software Foundation; either version 2.1 of the License, or (at your option) 
	any later version.
	
	This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied 
	warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for 
	more details.
	
	You should have received a copy of the GNU Lesser General Public License along with this library; if not, write to 
	the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/


package edu.utexas.its.eis.tools.qwicap.servlet;


import edu.utexas.its.eis.tools.qwicap.util.StackTrace;
import edu.utexas.its.eis.tools.qwicap.util.ArrayIterator;

import java.util.Iterator;


/**	Represents a stack trace that has been stripped of all leading elements that correspond to locations within 
 *	Qwicap's code.
 * 
 * 	@author Chris W. Johnson
 */
final class SimplifiedStackTrace implements Iterable<StackTraceElement> {

	private final StackTraceElement[]	Trace;
	private String						TraceString;


	SimplifiedStackTrace() {
		Trace = stripQwicapFromStartOfStackTrace(new Throwable().getStackTrace());
	}

	@Override public int hashCode() {
		return toString().hashCode();
	}
	
	@Override public boolean equals
	(
		final Object			Other
	)
	{
		if (Other == this)
			return true;

		if (Other == null)
			return false;

		if (Other instanceof SimplifiedStackTrace == false)
			return false;

		return stackTracesAreIdentical(Trace, ((SimplifiedStackTrace) Other).Trace);
	}

	@Override public Iterator<StackTraceElement> iterator() {
		return new ArrayIterator<StackTraceElement>(Trace);
	}

	StackTraceElement getApplicationTraceElement() {

	//	The application's method that invoked Qwicap (and ultimately led to the creation of this object, and the 
	//	invocation of this method) is always the first element in the Trace array.

		return Trace[0];
	}

	boolean endsWithMethod
	(
		final Class<?>			PromptPatClass,
		final String			PromptPatMethodName
	)
	{
		final StackTraceElement	T = Trace[0];

		return T.getMethodName().equals(PromptPatMethodName) && T.getClassName().equals(PromptPatClass.getName());
	}

	boolean isSameMethodAs
	(
		final SimplifiedStackTrace Other
	)
	{
		return isSameMethod(Trace, Other.Trace);
	}

	boolean isSubsetOf
	(
		final SimplifiedStackTrace Superset
	)
	{
		return isSubsetOf(Trace, Superset.Trace);
	}

	void print
	(
		final int					Indent
	)
	{
		for (final StackTraceElement T : Trace) {
		
			for (int TIndex = 0; TIndex < Indent; TIndex++)
				System.out.print('\t');

			System.out.print(T.getClassName());
			System.out.print('.');
			System.out.print(T.getMethodName());
			System.out.print(':');
			System.out.println(T.getLineNumber());
		}
	}

	@Override public String toString() {
		
		if (TraceString == null)
			TraceString = StackTrace.toString(Trace);
		
		return TraceString;
	}

/*	Reduces a stack trace like:

		edu.utexas.its.eis.tools.qwicap.servlet.Page.<init>:82
		edu.utexas.its.eis.tools.qwicap.servlet.PageStack.autoPushPopReplace:361
		edu.utexas.its.eis.tools.qwicap.servlet.PageStack.set:331
		edu.utexas.its.eis.tools.qwicap.servlet.Qwicap.prompt:675
		edu.utexas.its.eis.tools.qwicap.servlet.Qwicap.promptModelessSynch:572
		edu.utexas.its.eis.tools.demo.HuntTheWumpus.play:79
		edu.utexas.its.eis.tools.demo.WebApplication.<init>:100
		edu.utexas.its.eis.tools.demo.WebApplication.main:50
		sun.reflect.NativeMethodAccessorImpl.invoke0:-2
		sun.reflect.NativeMethodAccessorImpl.invoke:39
		sun.reflect.DelegatingMethodAccessorImpl.invoke:25
		java.lang.reflect.Method.invoke:324
		edu.utexas.its.eis.tools.qwicap.servlet.Qwicap.run:366

	...to a stack trace like:
	
		edu.utexas.its.eis.tools.demo.HuntTheWumpus.play:79
		edu.utexas.its.eis.tools.demo.WebApplication.<init>:100
		edu.utexas.its.eis.tools.demo.WebApplication.main:50
		sun.reflect.NativeMethodAccessorImpl.invoke0:-2
		sun.reflect.NativeMethodAccessorImpl.invoke:39
		sun.reflect.DelegatingMethodAccessorImpl.invoke:25
		java.lang.reflect.Method.invoke:324
		edu.utexas.its.eis.tools.qwicap.servlet.Qwicap.run:366
*/
	private static StackTraceElement[] stripQwicapFromStartOfStackTrace
	(
		final StackTraceElement[]	Traces
	)
	{
		final String			QwicapPackage = Qwicap.kQwicapPackage;
		final int				Count         = Traces.length;
		int						Index         = 0;

		while (Index < Count) {

			if (Traces[Index].getClassName().startsWith(QwicapPackage) == false)
				break;

			Index++;
		}
		
		if (Index == 0)
			return Traces;

		final StackTraceElement[]	SimplifiedTraces = new StackTraceElement[Count - Index];

		System.arraycopy(Traces, Index, SimplifiedTraces, 0, SimplifiedTraces.length);

		return SimplifiedTraces;
	}

/**	Compares all the elements of two <code>StackTraceElement</code> arrays for equality. It is assumed that both 
 *	arrays were produced by the <code>stripQwicapFromStackTrace</code> method.
 * 
 * @param ATraces			An array of traces that was simplified by the <code>stripQwicapFromStackTrace</code> method.
 * @param BTraces			An array of traces that was simplified by the <code>stripQwicapFromStackTrace</code> method.
 * @return					<code>true</code> if the traces in both arrays were identical, <code>false</code> otherwise.
 */
	private static boolean stackTracesAreIdentical
	(
		final StackTraceElement[]	ATraces,	//	Simplified by stripQwicapFromStackTrace
		final StackTraceElement[]	BTraces		//	Simplified by stripQwicapFromStackTrace
	)
	{
		final int				Count = ATraces.length;

	//	These traces can't possibly correspond to the same method, if their lengths are different.

		if (Count != BTraces.length)
			return false;

	//	This loop compares all elements in the two arrays for exact equality.

		for (int Index = 0; Index < Count; Index++) {

			if (ATraces[Index].equals(BTraces[Index]) == false)
				return false;
		}

		return true;
	}

/**	This method is exactly like <code>stackTracesAreIdentical</code>, with one exception: it does NOT compare the line 
 *	numbers of the first <code>StackTraceElement</code> in each array (it compares everything else, however).
 * 
 * @param ATraces			An array of traces that was simplified by the <code>stripQwicapFromStackTrace</code> method.
 * @param BTraces			An array of traces that was simplified by the <code>stripQwicapFromStackTrace</code> method.
 * @return					<code>true</code> if the traces match (except for the final line number), <code>false</code>
 *							otherwise.
 */
	private static boolean isSameMethod
	(
		final StackTraceElement[]	ATraces,	//	Simplified by stripQwicapFromStackTrace
		final StackTraceElement[]	BTraces		//	Simplified by stripQwicapFromStackTrace
	)
	{
		final int				Count = ATraces.length;

	//	These traces can't possibly correspond to the same method, if their lengths are different.

		if (Count != BTraces.length)
			return false;

	//	Compare the first elements, except for their line numbers.

		{
			final StackTraceElement		A = ATraces[0];
			final StackTraceElement		B = BTraces[0];

			if (A.getClassName().equals(B.getClassName()) == false || A.getMethodName().equals(B.getMethodName()) == false)
				return false;
		}

	//	Compare all aspects of the remaining elements in each array.

		for (int Index = 1; Index < Count; Index++) {

			if (ATraces[Index].equals(BTraces[Index]) == false)
				return false;
		}

		return true;
	}

/**	This method is exactly like <code>isSameMethod</code>, with one exception: it doesn't mind if the Superset array
 *	is longer than the Subset array. It just compares every element in Subset to the corresponding element in Superset,
 *  and returns <code>true</code> if those match (with the exception of the line number of the first element in Subset).
 * 
 * @param Subset			An array of traces that was simplified by the <code>stripQwicapFromStackTrace</code> method.
 * @param Superset			An array of traces that was simplified by the <code>stripQwicapFromStackTrace</code> method.
 * @return					<code>true</code> if the traces match (except for the final line number), <code>false</code>
 *							otherwise.
 */
	private static boolean isSubsetOf
	(
		final StackTraceElement[]	Subset,		//	Simplified by stripQwicapFromStackTrace
		final StackTraceElement[]	Superset	//	Simplified by stripQwicapFromStackTrace
	)
	{
		final int				SubCount = Subset.length;
		final int				SupCount = Superset.length;

	//	If the subset is longer than the superset, it isn't a subset at all.

		if (SubCount > SupCount)
			return false;

		int						SubIndex = SubCount - 1;
		int						SupIndex = SupCount - 1;

		while (SubIndex > 0) {

			if (Subset[SubIndex--].equals(Superset[SupIndex--]) == false)
				return false;
		}

	//	Compare the final elements.

		final StackTraceElement		A = Subset[SubIndex];	//	SubIndex will always be zero when we arrive here.
		final StackTraceElement		B = Superset[SupIndex];

		return A.getClassName().equals(B.getClassName()) && A.getMethodName().equals(B.getMethodName());
	}
}




/*	An actual stack trace for an application with three pages (page 1 is mistakenly believed to be "in play").
	These strack traces have been simplified by stripQwicapFromStackTrace method will do.

Page 0:
        Page ID: 1E957BFF18C9ED68
        Markup: /usr/local/tomcat/webapps/wumpus/wumpus-intro.html
        Parameters:
                Form Method:  post
                SessionContext Path: /wumpus
                Servlet Path: /
                Form Data Set:
                        action: "config"
                        qwicap-form-id: "938934539D4AF8C4"
                        qwicap-page-id: "1E957BFF18C9ED68"
        Stack Trace:
                edu.utexas.its.eis.tools.demo.WebApplication.<init>:85
                edu.utexas.its.eis.tools.demo.WebApplication.main:50
                sun.reflect.NativeMethodAccessorImpl.invoke0:-2
                sun.reflect.NativeMethodAccessorImpl.invoke:39
                sun.reflect.DelegatingMethodAccessorImpl.invoke:25
                java.lang.reflect.Method.invoke:324
                edu.utexas.its.eis.tools.qwicap.servlet.Qwicap.run:366
Page 1:
        Page ID: 7CFA6C34B6A13780
        Markup: /usr/local/tomcat/webapps/wumpus/map-config.html
        Parameters:
                Form Method:  post
                SessionContext Path: /wumpus
                Servlet Path: /
                Form Data Set:
                        rooms: "42"
                        codec: "1"
                        qwicap-form-id: "040034D6B923310C"
                        play: "play"
                        qwicap-page-id: "7CFA6C34B6A13780"
                        map: "custom"
                        sides: "7"
                        quality: "0.75"
        Stack Trace:
                edu.utexas.its.eis.tools.demo.WebApplication.configPage:134
                edu.utexas.its.eis.tools.demo.WebApplication.<init>:96
                edu.utexas.its.eis.tools.demo.WebApplication.main:50
                sun.reflect.NativeMethodAccessorImpl.invoke0:-2
                sun.reflect.NativeMethodAccessorImpl.invoke:39
                sun.reflect.DelegatingMethodAccessorImpl.invoke:25
                java.lang.reflect.Method.invoke:324
                edu.utexas.its.eis.tools.qwicap.servlet.Qwicap.run:366
Page 2:
        Page ID: 6395584AA5039FB5
        Markup: /usr/local/tomcat/webapps/wumpus/wumpus.html
        Parameters:
                Form Method:  n/a
                SessionContext Path: 
                Servlet Path: 
                Form Data Set:
        Stack Trace:
                edu.utexas.its.eis.tools.demo.HuntTheWumpus.play:79
                edu.utexas.its.eis.tools.demo.WebApplication.<init>:100
                edu.utexas.its.eis.tools.demo.WebApplication.main:50
                sun.reflect.NativeMethodAccessorImpl.invoke0:-2
                sun.reflect.NativeMethodAccessorImpl.invoke:39
                sun.reflect.DelegatingMethodAccessorImpl.invoke:25
                java.lang.reflect.Method.invoke:324
                edu.utexas.its.eis.tools.qwicap.servlet.Qwicap.run:366
*/



