/*	QWICAP (Quick Web Interface for Conventional Applications) by Chris W. Johnson
	Copyright (C) 2007 The University of Texas at Austin ("U.T. Austin"). All rights reserved.
	
	U.T. Austin's copyright notice in this evolving version reflects our original authorship. It does not assert 
	copyright over changes contributed by others.
	
	This library is free software; you can redistribute it and/or modify it under the terms of the GNU Lesser General 
	Public License as published by the Free Software Foundation; either version 2.1 of the License, or (at your option) 
	any later version.
	
	This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied 
	warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for 
	more details.
	
	You should have received a copy of the GNU Lesser General Public License along with this library; if not, write to 
	the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/


package edu.utexas.its.eis.tools.qwicap.servlet;


import java.nio.charset.CharsetDecoder;
import java.nio.charset.Charset;


/**
 * Created by IntelliJ IDEA.
 * User: chrisj
 * Date: Jul 17, 2007
 * Time: 8:57:21 PM
 * To change this template use File | Settings | File Templates.
 */
final class UTF8Analysis {

	final boolean			CannotBeUTF8;
	final boolean			ProbablyIsUTF8;


	static UTF8Analysis getUnified
	(
		final String				DefaultCharSetName,	//	Can be null.
		final FormControlBytes...	FCBs
	)
	{
		boolean					CannotBeUTF8   = false;
		boolean					ProbablyIsUTF8 = false;

		for (final FormControlBytes FCB : FCBs) {

		//	Test the control name for UTF-8-ness...

			final UTF8Analysis		NameDetection = new UTF8Analysis(FCB.Name);

			if (NameDetection.CannotBeUTF8) {

				CannotBeUTF8 = true;
				break;
			}

		//	Test the control value for UTF-8-ness...

			final UTF8Analysis		ValueDetection = new UTF8Analysis(FCB.Value);

			if (ValueDetection.CannotBeUTF8) {

				CannotBeUTF8 = true;
				break;
			}

		//	Done with this name and value. Update the value of ProbablyIsUTF8.

			ProbablyIsUTF8 |= NameDetection.ProbablyIsUTF8 | ValueDetection.ProbablyIsUTF8;
		}

		if (CannotBeUTF8 == false && ProbablyIsUTF8 && "UTF-8".equalsIgnoreCase(DefaultCharSetName) == false) {

		//	Test whether or not all of the bytes can be decoded in the specified character set. If they can be,
		//	the specified character set is most likely to be correct. If not, the character set is almost 
		//	certainly UTF-8.
			
			try {
				final CharsetDecoder	Decoder = Charset.forName(DefaultCharSetName).newDecoder();
				
			//	The following method invocations will throw an exception if there are any bytes in the input that
			//	cannot be decoded to characters supported by the decoder's character set.

				for (final FormControlBytes FCB : FCBs) {
					
					FCB.Name.toString(Decoder);
					FCB.Value.toString(Decoder);
				}
				
			//	No exceptions were thrown above, so the default character set could decode everything it encountered.
			//	Therefore, we have no basis for assuming that these characters are UTF-8.
				
				ProbablyIsUTF8 = false;
				
			} catch (Exception e) {
			//	Discard exception.
			}
		}
		
		return new UTF8Analysis(CannotBeUTF8, ProbablyIsUTF8);
	}

	UTF8Analysis
	(
		final ByteArray			BA
	)
	{
		this(BA.getArray(), BA.getOffset(), BA.size());
	}

	UTF8Analysis
	(
		final byte[]			Bytes
	)
	{
		this(Bytes, 0, Bytes.length);
	}
	
	private UTF8Analysis
	(
		final boolean			CannotBeUTF8,
		final boolean			ProbablyIsUTF8
	)
	{
		this.CannotBeUTF8   = CannotBeUTF8;
		this.ProbablyIsUTF8 = !CannotBeUTF8 && ProbablyIsUTF8;
	}
	
	private UTF8Analysis
	(
		final byte[]			Bytes,
		int						Index,
		final int				Count
	)
	{
		final int				Limit          = Index + Count;
		boolean					FoundUTF8Chars = false;

		while (Index < Limit) {
			byte					Byte = Bytes[Index++];

			if ((Byte & 0x80) != 0) {

			//	"The bytes 0xFE and 0xFF are never used in the UTF-8 encoding."

				if ((Byte & 0xFF) > 0xFD) {

				//	This would be a basic mistake, if these bytes represented UTF-8.

					CannotBeUTF8   = true;
					ProbablyIsUTF8 = false;
					return;
				}

			//	Count the consecutive high-order '1' bits. Each bit corresponds to an additional byte of information
			//	about the composition of this character.

				int						BytesFollowing = 0;

				Byte <<= 1;

				while ((Byte & 0x80) != 0) {

					BytesFollowing++;
					Byte <<= 1;
				}

				if (BytesFollowing > 5) {

				//	This would be a basic mistake, if these bytes represented UTF-8.

					CannotBeUTF8   = true;
					ProbablyIsUTF8 = false;
					return;
				}

			//	Make sure the following bytes are present, as expected.

				if (BytesFollowing > Limit - Index) {

				//	This would be a basic mistake, if these bytes represented UTF-8.

					CannotBeUTF8   = true;
					ProbablyIsUTF8 = false;
					return;
				}

				while (--BytesFollowing >= 0) {
					final int				ByteUnsigned = Bytes[Index++] & 0xFF;

					if (ByteUnsigned >= 0x80 && ByteUnsigned <= 0xBF) {
					
						FoundUTF8Chars = true;
					
					} else {

					//	This would be a basic mistake, if these bytes represented UTF-8.

						CannotBeUTF8   = true;
						ProbablyIsUTF8 = false;
						return;
					}
				}
			}
		}

		CannotBeUTF8   = false;
		ProbablyIsUTF8 = FoundUTF8Chars;
	}
}