/*	QWICAP (Quick Web Interface for Conventional Applications) by Chris W. Johnson
	Copyright (C) 2011 The University of Texas at Austin ("U.T. Austin"). All rights reserved.
	
	U.T. Austin's copyright notice in this evolving version reflects our original authorship. It does not assert 
	copyright over changes contributed by others.
	
	This library is free software; you can redistribute it and/or modify it under the terms of the GNU Lesser General 
	Public License as published by the Free Software Foundation; either version 2.1 of the License, or (at your option) 
	any later version.
	
	This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied 
	warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for 
	more details.
	
	You should have received a copy of the GNU Lesser General Public License along with this library; if not, write to 
	the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/


package edu.utexas.its.eis.tools.qwicap.servlet;


/*
 * Created by IntelliJ IDEA.
 * User: chrisj
 * Date: Apr 27, 2011
 * Time: 6:45:05 AM
 * To change this template use File | Settings | File Templates.
 */


class XMLContentTypeDefaultCharSet implements ContentTypeDefaultCharSet {
	
//	For the origins of these types, information about them, and the same for any future XML types, see the official
//	list of MIME types at <http://www.iana.org/assignments/media-types/index.html>.
	
// 	Strictly speaking, the MIME type "vnd.moml+xml" makes no reference to RFC 3023, so I'm assuming from its stated
//	use of 7-bit encoding that it is a text type, for purposes of RFC 3023 categorization.
	
	private static final String[][]	kTextXMLTypes        = { { "vnd", "moml+xml" } };
	private static final String[][]	kApplicationXMLTypes = { { "image", "svg+xml" }, { "message", "imdn+xml" }, { "vnd", "collada+xml" } };
	
	
	@Override public String get
	(
		String					Type,	//	Can be null.
		String					Subtype	//	Can be null.
	)
	{
		if (Type != null)
			Type = Type.toLowerCase();
		else
			Type = "";		//	Important when we use "foundIn".
		
		if (Subtype != null)
			Subtype = Subtype.toLowerCase();
		else
			Subtype = "";	//	Important when we use "Subtype.endsWith" and "foundIn".	
		
	//	If we're NOT dealing with content delivered via HTTP, RFC 2616's guidance about default character sets 
	//	(which RFC 3023, <http://www.ietf.org/rfc/rfc3023.txt>, section 3.1, says is wrong), is irrelevant. However, 
	//	both the XML specification and RFC 3023 have guidance to offer.
	//
	//	One trick becomes identifying XML types. And there's a complication: The default character set defined by 
	//	the XML specification (UTF-8) only applies when there's *no* other character set identification information, 
	//	including a BOM, or the "encoding" attribute of the "xml" element. Here, we have no idea how the content 
	//	type or "charset" information was obtained. Not only could they be wrong or absent, but the document could 
	//	still possess a BOM and/or an "encoding" attribute. However, those identifiers fall outside the scope of 
	//	what a Content-Type parser can know about, so this piece of code will always return "UTF-8" as the default 
	//	character set for XML.
	//
	//	From <http://www.w3.org/TR/REC-xml/#charencoding>: "In the absence of information provided by an external 
	//	transport protocol (e.g. HTTP or MIME), it is a fatal error for an entity including an encoding declaration 
	//	to be presented to the XML processor in an encoding other than that named in the declaration, or for an 
	//	entity which begins with neither a Byte Order Mark nor an encoding declaration to use an encoding other 
	//	than UTF-8."
		
	//	With regard to identifying XML types, see <http://www.ietf.org/rfc/rfc3023.txt>, and 
	//	<http://www.iana.org/assignments/media-types/>.
	//
	//	See also <http://www.imc.org/ietf-xml-mime/mail-archive/msg00854.html> for important commentary from one of
	//	the authors of RFC 3023.
		
		if ("text".equals(Type) && ("xml".equals(Subtype) || "xml-external-parsed-entity".equals(Subtype) || Subtype.endsWith("+xml")))
			return "US-ASCII";	//	See RFC 3023, <http://www.ietf.org/rfc/rfc3023.txt>, section 3.1, 3.3, and 7/7.1.
		
		if ("application".equals(Type) && ("xml".equals(Subtype) || "xml-external-parsed-entity".equals(Subtype) || "xml-dtd".equals(Subtype) || Subtype.endsWith("+xml")))
			return "UTF-8";		//	See RFC 3023, <http://www.ietf.org/rfc/rfc3023.txt>, section 3.2, 3.4 3.5, and 7/7.1.
		
		if (Subtype.endsWith("+xml")) {
			
			if (foundIn(Type, Subtype, kTextXMLTypes))
				return "US-ASCII";
			
			if (foundIn(Type, Subtype, kApplicationXMLTypes))
				return "UTF-8";
		}
					
		return null;
	}
	
	private static boolean foundIn
	(
		final String			Type,		//	Non-null, lower-case.
		final String			Subtype,	//	Non-null, lower-case.
		final String[][]		MIMETypeList
	)
	{
		for (final String[] MIMEType : MIMETypeList) {
			
			if (MIMEType[0].equals(Type) && MIMEType[1].equals(Subtype))
				return true;
		}
		
		return false;
	}
}