/*	QWICAP (Quick Web Interface for Conventional Applications) by Chris W. Johnson
	Copyright (C) 2007-2010 The University of Texas at Austin ("U.T. Austin"). All rights reserved.
	
	U.T. Austin's copyright notice in this evolving version reflects our original authorship. It does not assert 
	copyright over changes contributed by others.
	
	This library is free software; you can redistribute it and/or modify it under the terms of the GNU Lesser General 
	Public License as published by the Free Software Foundation; either version 2.1 of the License, or (at your option) 
	any later version.
	
	This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied 
	warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for 
	more details.
	
	You should have received a copy of the GNU Lesser General Public License along with this library; if not, write to 
	the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/


package edu.utexas.its.eis.tools.qwicap.servlet;


import javax.servlet.ServletContext;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.logging.Level;
import java.util.logging.Logger;
import java.util.HashSet;


/**	This class is a front-end to the <code>DocumentCache</code>. It implements the search through the cache for 
 * 	appropriate translations/localizations of each requested document. While the <code>DocumentCache</code> is shared 
 * 	by all threads and sessions of a particular web application, <code>i18nDocumentAccessor</code> instances are 
 * 	specific to each session, since each session may be serving a user with different localization requirements.
 * 
 * 	@author Chris W. Johnson
 */
class i18nDocumentAccessor {
	
	private static final Logger			Log = Logger.getLogger(i18nDocumentAccessor.class.getName());
	
	private final DocumentCache			DocCache;
	private final i18nLocalizationsList	WebAppLangs;
	private final String				WebAppContentURLRoot;
	private final File					WebAppContentPathRoot;
	
	private final HashSet<String>		ResourceSearchPaths = new HashSet<String>();
	
	
	i18nDocumentAccessor
	(
		final DocumentCache			DocCache,
		final i18nLocalizationsList	WebAppLangs,
		final ServletContext		SrvltCtxt
	)
	throws MalformedURLException, FileNotFoundException
	{
		this.DocCache    = DocCache;
		this.WebAppLangs = WebAppLangs;
								
		final URL				WebAppContentURLRoot = SrvltCtxt.getResource("/");
		
		if (WebAppContentURLRoot == null)
			throw new FileNotFoundException("The resource \"/\" could not be located by the servlet container.");
		
		this.WebAppContentURLRoot = WebAppContentURLRoot.toString();
		
		Log.log(Level.FINE, "The contents of this web application can be found at URL <{0}>.", this.WebAppContentURLRoot);
		
		final String			WebAppContentPathRoot = SrvltCtxt.getRealPath("/");
		
		if (WebAppContentPathRoot == null) {
			
			this.WebAppContentPathRoot = null;
			
			Log.log(Level.FINE, "The path to the contents of this web application could not be determined by requesting the real path to \"/\". If this web application is running directly from a WAR file, that is to be expected.");
			
		} else {
			
			this.WebAppContentPathRoot = new File(WebAppContentPathRoot);
			
			if (this.WebAppContentPathRoot.exists() == false)
				throw new FileNotFoundException("The top-level directory of this web application was not found at \"" + this.WebAppContentPathRoot.getAbsolutePath() + "\".");
			
			Log.log(Level.FINE, "The contents of this web application can be found at path \"{0}\".", this.WebAppContentPathRoot.getAbsolutePath());
		}
	}
	
	i18nDocumentAccessor
	(
		final DocumentCache			DocCache,
		final i18nLocalizationsList	WebAppLangs,
		final String				WebAppContentURLRoot,
		final File					WebAppContentPathRoot
	)
	{
		this.DocCache              = DocCache;
		this.WebAppLangs           = WebAppLangs;
		this.WebAppContentURLRoot  = WebAppContentURLRoot;
		this.WebAppContentPathRoot = WebAppContentPathRoot;
	}
	
	DocumentCache getDocumentCache() {
		return DocCache;
	}
	
	i18nLocalizationsList getLocalizations() {
		return WebAppLangs;
	}
	
	String getDocumentURLRoot() {
		return WebAppContentURLRoot;
	}
	
	File getDocumentPathRoot() {
		return WebAppContentPathRoot;
	}
	
	URL getBasicURL
	(
		final String			Document
	)
	throws MalformedURLException
	{
		if (Document.startsWith("/"))
			return new URL(WebAppContentURLRoot + Document.substring(1));
		else
			return new URL(WebAppContentURLRoot + Document);
	}
	
	DocumentSource getDocument
	(
		final i18nLocaleSettings	Settings,
		final String				DocName
	)
	throws IOException
	{
		DocumentSource			DocSrc;
		
		if (DocName.startsWith("/")) {
			
		//	Try to retrieve the named document as a servlet container resource.
			
			DocSrc = getDocumentResource(Settings, DocName);
			
		} else {
			
		//	Try to retrieve the named document as a file, if (and only if) we previously established the path to this
		//	web application's "root" directory. Otherwise, try to access the file as a resource.
			
		//	todo Consider: Is there any point in continuing to support access to documents via direct File I/O?
			
			if (WebAppContentPathRoot != null)
				DocSrc = getDocumentFile(Settings, DocName);
			else
				DocSrc = getDocumentResource(Settings, '/' + DocName);
		}
		
		if (DocSrc.exists() == false) {
			final DocumentSource		STRsrcDS = getDocumentResourceBasedOnStackTrace(Settings, DocName);
			
			if (STRsrcDS != null)
				DocSrc = STRsrcDS;
			
			if (DocSrc.exists() == false) {
				
			//	If all else has failed, see if we're being asked for a document that might be one of Qwicap's internal 
			//	resources.
				
				if (DocName.startsWith("/qwicap")) {
					
				//	todo We should probably construct the following URL ourselves, and skip whatever overhead the getResource method may include.
					
					final URL				RsrcURL = QwicapServlet.class.getResource(DocName.substring(1));	//	Use the ServletPath string, but without its leading slash character.
					
				//	todo Should there be a multi-lingual search scheme for Qwicap's internal resources?
					
					DocSrc = getDocument(Settings, RsrcURL);
					
				} else {
					
				//	Search in locations where we've previously located documents that were stored as resources.
					
					final String			RsrcName = DocName.startsWith("/") ? DocName.substring(1) : DocName;
					
					for (final String RsrcURLStr : ResourceSearchPaths) {
						
						Log.log(Level.FINE, "Looking for \"{1}\" in <{0}>.", new Object[] { RsrcURLStr, RsrcName });
						
						DocSrc = getDocument(Settings, new URL(RsrcURLStr + RsrcName));
						if (DocSrc.exists()) {
							
							Log.log(Level.FINE, "Found <{0}>.", new Object[] { DocSrc.getURL() });
							break;
						}
					}
				}
			}
		}
		
	//	Give-up and return a DocumentSource instance that will tell the caller that the request document does not exist.
		
		return DocSrc;
	}
	
	private DocumentSource getDocumentResourceBasedOnStackTrace
	(
		final i18nLocaleSettings	Settings,
		final String				RsrcName	//	Might begin with an unwanted slash.
	)
	throws IOException 
	{
		final String				DocName     = RsrcName.startsWith("/") ? RsrcName.substring(1) : RsrcName;
		DocumentSource				FirstDocSrc = null;
		
		for (final StackTraceElement TraceElem : new SimplifiedStackTrace()) {
			
			try {
				final String			TraceClassName = TraceElem.getClassName();
				
				if (TraceClassName.startsWith("sun.reflect.") || TraceClassName.startsWith("javax."))
					break;
				
				final Class				TraceClass = Class.forName(TraceClassName);
				final URL				RsrcURL    = TraceClass.getResource(DocName);
				
			//	Example RsrcURL: <jar:file:/usr/local/tomcat/work/Catalina/localhost/docupload/WEB-INF/lib/dbcoreweb-1.0d0-no-libs.jar!/edu/utexas/its/eis/tools/dbcoreweb/namespaces.html>
				
			//	todo If we find a document using this technique, we need to add the place where the document was found to a list of places that will be searched for future documents.
				
			//	todo Implement an internationalization-aware search for the resource.
				
				if (RsrcURL != null) {
					
				//	Make a note of the fact that, in this location, we've found a document (resource) for which we were
				//	searching. If future searches for documents fail, we'll make a last ditch effort to locate the 
				//	documents by searching these locations (which might not otherwise be searched).
					
					ResourceSearchPaths.add(stripTrailingString(RsrcURL.toString(), DocName));
					
					final DocumentSource	DocSrc = getDocument(Settings, RsrcURL);
					
					if (DocSrc.exists())
						return DocSrc;
					
					if (FirstDocSrc == null)
						FirstDocSrc = DocSrc;
				}
				
				Log.log(Level.FINE, "Class resource <{0}> was requested, but it does not exist as a resource of class \"{1}\".", new Object[] { RsrcName, TraceClass });		
				
			} catch (ClassNotFoundException e) {
				Log.log(Level.WARNING, "Class \"" + TraceElem.getClassName() + "\" not found.", e);
			}
		}
		
		return null;	//	This document does not exist, but that's OK.
	}
	
	
/*	private static final Pattern	JarFileURLPattern = Pattern.compile("(jar:file:/.+?[.][jJ][aA][rR]!/)(.*)");

	
	private void preloadSiblingResources
	(
		final i18nLocaleSettings	Settings,
		final Class					MasterClass,
		final URL					MasterRsrcURL,
		final String				MasterRsrcName
	)
	throws IOException
	{
		final String				MasterRsrcURLStr = MasterRsrcURL.toString();
		
		if (!MasterRsrcURLStr.startsWith("jar:"))
			return;
		
		if (!MasterRsrcURLStr.endsWith(MasterRsrcName))
			return;
		
		final Matcher				M = JarFileURLPattern.matcher(MasterRsrcURLStr);
		
		if (M.matches() == false) {
			
			Log.log(Level.INFO, "We do not know how to search for sibling resources in <{0}>.", MasterRsrcURLStr);
			return;
		}
		
		final String				JarFileURLStr = stripTrailingString(MasterRsrcURLStr, MasterRsrcName.startsWith("/") ? MasterRsrcName.substring(1) : MasterRsrcName);

		Log.log(Level.FINE, "Trying to open JAR file: <{0}>", JarFileURLStr);
		
		try {
			final JarInputStream		JarIn = new JarInputStream(new URL(JarFileURLStr).openStream());
	
			try {
				
				Log.log(Level.FINE, "Iterating entries in JAR file: <{0}>", JarFileURLStr);
				
				for (JarEntry Entry = JarIn.getNextJarEntry(); Entry != null; Entry = JarIn.getNextJarEntry()) {
					final String			Name = Entry.getName();
					
					Log.log(Level.FINE, "Jar Entry: {0}", Name);
					
				//	Skip all class files.
					
					if (!Name.endsWith(".class")) {
						final int				LastSlashIndex = Name.lastIndexOf('/');
						final String			FileName       = LastSlashIndex >= 0 ? Name.substring(LastSlashIndex + 1) : Name;
						
						try {
							getDocument(Settings, new URL(JarFileURLStr + FileName));
							Log.log(Level.FINE, "Preloaded resource <{0}{1}>.", new Object[] { JarFileURLStr, FileName });
						} catch (MalformedURLException e) {
							Log.log(Level.WARNING, "The URL <" + JarFileURLStr + Name + "> is malformed.", e);
						}
					}
				}
	
			} finally {
				JarIn.close();
			}
			
		} catch (IOException e) {
			Log.log(Level.WARNING, "The URL <" + JarFileURLStr + "> could not be opened.", e);
		}
	}
*/	
	private static String stripTrailingString
	(
		final String				FullStr,
		final String				TrailingStr
	)
	{
		if (FullStr.endsWith(TrailingStr))
			return FullStr.substring(0, FullStr.length() - TrailingStr.length());
		
		throw new IllegalArgumentException("The string \"" + FullStr + "\" does not end with the string \"" + TrailingStr + "\".");
	}

	
	private DocumentSource getDocumentResource
	(
		final i18nLocaleSettings	Settings,
		final String				RsrcName	//	Must begin with a slash.
	)
	throws MalformedURLException 
	{
		final String			URLPrefix = WebAppContentURLRoot;
		
		for (final String LangDir : Settings.getWebAppTranslationSearchDirs()) {
			final URL				i18nURL = new URL(URLPrefix + LangDir + RsrcName);
			final DocumentSource	DocSrc  = DocCache.getDocumentSource(i18nURL);
			
			if (DocSrc.exists())
				return DocSrc;	
			
			Log.log(Level.FINE, "Servlet container resource \"{0}\" was requested. Localized version <{1}> does not exist.", new Object[] { RsrcName, i18nURL });
		}
		
		return DocCache.getDocumentSource(new URL(URLPrefix + RsrcName.substring(1)));
	}

	private DocumentSource getDocumentFile
	(
		final i18nLocaleSettings	Settings,
		final String				FileName
	)
	throws IOException 
	{
		for (final String LangDir : Settings.getWebAppTranslationSearchDirs()) {
			final String			i18nFileName = LangDir + File.separatorChar + FileName;
			final DocumentSource	DocSrc       = getDocumentFileLow(i18nFileName);
			
			if (DocSrc.exists())
				return DocSrc;
			
			Log.log(Level.FINE, "Document \"{0}\" was requested. Localized version \"{1}\" does not exist.", new Object[] { FileName, i18nFileName });
		}
		
		final DocumentSource	DocSrc = getDocumentFileLow(FileName);
		
		if (DocSrc.exists() == false)
			Log.log(Level.FINE, "Document \"{0}\" was requested. No default version, or applicable localized version, exists.", FileName);
		
		return DocSrc;
	}
	
	private DocumentSource getDocumentFileLow
	(
		final String			FileName
	)
	throws IOException
	{
		final File				DocFile = new File(WebAppContentPathRoot, FileName);
		final DocumentSource	Doc     = DocCache.getDocumentSource(DocFile);

		if (Log.isLoggable(Level.FINE))
			Log.log(Level.FINE, "Document \"{0}\" was requested. It was found at \"{1}\".", new Object[] { FileName, DocFile.getAbsolutePath() });

		return Doc;
	}

	DocumentSource getDocument
	(
		final i18nLocaleSettings	Settings,
		final URL					DocURL
	)
	throws IOException
	{
		final String			DocURLStr = DocURL.toString();
		final String			URLPrefix = WebAppContentURLRoot;
		
		if (DocURLStr.startsWith(URLPrefix)) {
			
			Log.log(Level.FINE, "Document <{0}> was requested. Will attempt to find localized versions.", DocURL);
			
			final String			URLPostfix        = DocURLStr.substring(URLPrefix.length());

			for (final String LangDir : Settings.getWebAppTranslationSearchDirs()) {
				final URL				i18nURL = new URL(URLPrefix + LangDir + '/' + URLPostfix);
				final DocumentSource	DocSrc  = DocCache.getDocumentSource(i18nURL);
				
				if (DocSrc.exists())
					return DocSrc;
				
				Log.log(Level.FINE, "Document <{0}> was requested. Localized version <{1}> does not exist.", new Object[] { DocURLStr, i18nURL });
			}

		} else
			Log.log(Level.FINE, "Document <{0}> was requested.", DocURL);

		final DocumentSource	DocSrc = DocCache.getDocumentSource(DocURL);
		
		if (DocSrc.exists() == false)
			Log.log(Level.FINE, "Document <{0}> was requested, but does not exist.", DocURLStr);
		
		return DocSrc;
	}
}
