/*	QWICAP (Quick Web Interface for Conventional Applications) by Chris W. Johnson
	Copyright (C) 2007-2010 The University of Texas at Austin ("U.T. Austin"). All rights reserved.
	
	U.T. Austin's copyright notice in this evolving version reflects our original authorship. It does not assert 
	copyright over changes contributed by others.
	
	This library is free software; you can redistribute it and/or modify it under the terms of the GNU Lesser General 
	Public License as published by the Free Software Foundation; either version 2.1 of the License, or (at your option) 
	any later version.
	
	This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied 
	warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for 
	more details.
	
	You should have received a copy of the GNU Lesser General Public License along with this library; if not, write to 
	the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/


package edu.utexas.its.eis.tools.qwicap.servlet;


import javax.servlet.ServletContext;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.logging.Level;
import java.util.logging.Logger;


/**	Extends {@link i18nDocumentAccessor} to add a name mapping cache which eliminates the search for previously 
 * 	accessed documents applicable to a particular language setting. If the language setting changes, the cache is
 * 	automatically cleared. Note that while instances of this class are normally associated with a particular session,
 * 	they may be accessed either from the session's thread, or a servlet thread that is handling an HTTP GET request 
 * 	associated with the session. Therefore, some internal synchronization is required in order to protect the integrity
 * 	of the name map.
 * 
 * 	@author Chris W. Johnson
 */
final class i18nDocumentAccessorWithNameMap extends i18nDocumentAccessor {

	private static final Logger				Log             = Logger.getLogger(i18nDocumentAccessorWithNameMap.class.getName());

	private final Object					NameMapSynchObj = new Object();
	private final LRUCacheMap<String, URL>	NameMap;
	private HTTPLanguageTag					CurLang;


	i18nDocumentAccessorWithNameMap
	(
		final i18nDocumentAccessor	DocAcc
	)
	{
		super(DocAcc.getDocumentCache(), DocAcc.getLocalizations(), DocAcc.getDocumentURLRoot(), DocAcc.getDocumentPathRoot());
		
		final DocumentCensus		DocCensus = DocAcc.getDocumentCache().getCensus();
		
		NameMap = new LRUCacheMap<String, URL>(DocCensus.getDocumentNameCount(), DocCensus.getDocumentNameCount() * 2);
	}
	
	i18nDocumentAccessorWithNameMap
	(
		final DocumentCache			DocCache,
		final i18nLocalizationsList	WebAppLangs,
		final ServletContext		SrvltCtxt
	)
	throws MalformedURLException, FileNotFoundException
	{
		super(DocCache, WebAppLangs, SrvltCtxt);
		
		final DocumentCensus		DocCensus = DocCache.getCensus();
		
		NameMap = new LRUCacheMap<String, URL>(DocCensus.getDocumentNameCount(), DocCensus.getDocumentNameCount() * 2);
	}

	i18nDocumentAccessorWithNameMap
	(
		final DocumentCache			DocCache,
		final i18nLocalizationsList	WebAppLangs,
		final String				WebAppContentURLRoot,
		final File					WebAppContentPathRoot
	)
	{
		super(DocCache, WebAppLangs, WebAppContentURLRoot, WebAppContentPathRoot);
		
		final DocumentCensus		DocCensus = DocCache.getCensus();
		
		NameMap = new LRUCacheMap<String, URL>(DocCensus.getDocumentNameCount(), DocCensus.getDocumentNameCount() * 2);
	}

	@Override DocumentSource getDocument
	(
		final i18nLocaleSettings	Settings,
		final String				DocName
	)
	throws IOException
	{
		final HTTPLanguageTag	DocLang = Settings.getLocaleAsLanguageTag();

		if (DocLang.equals(CurLang) == false) {

		//	If the localization we're supposed to be using has been changed, clear the name mapping cache.
			
			synchronized (NameMapSynchObj) {
				
				CurLang = DocLang;
				NameMap.clear();
			}

		} else {
			final URL				RealDocURL;
					
			synchronized (NameMapSynchObj) {
				RealDocURL = NameMap.get(DocName);
			}

			if (RealDocURL != null) {

				Log.log(Level.FINE, "Canonical document \"{0}\" is known to correspond to real document <{1}>.", new Object[] { DocName, RealDocURL });
				return getDocumentCache().getDocumentSource(RealDocURL);
			}
		}
		
		return populateNameMap(super.getDocument(Settings, DocName), DocName);
	}

	@Override DocumentSource getDocument
	(
		final i18nLocaleSettings	Settings,
		final URL					DocURL
	)
	throws IOException
	{
		final HTTPLanguageTag	DocLang = Settings.getLocaleAsLanguageTag();
		
		if (DocLang.equals(CurLang) == false) {

		//	If the localization we're supposed to be using has been changed, clear the name mapping cache.
			
			synchronized (NameMapSynchObj) {
				
				CurLang = DocLang;
				NameMap.clear();
			}
			
		} else {
			final URL				RealDocURL;
					
			synchronized (NameMapSynchObj) {
				RealDocURL = NameMap.get(DocURL.toString());
			}
			
			if (RealDocURL != null) {
				
				Log.log(Level.FINE, "Canonical document <{0}> is known to correspond to real document <{1}>.", new Object[] { DocURL, RealDocURL });
				return getDocumentCache().getDocumentSource(RealDocURL);
			}
		}
		
		return populateNameMap(super.getDocument(Settings, DocURL), DocURL);
	}
	
	private DocumentSource populateNameMap
	(
		final DocumentSource	DocSrc,
		final URL				DocURL
	)
	{
		return populateNameMap(DocSrc, DocURL.toString());
	}

	private DocumentSource populateNameMap
	(
		final DocumentSource	DocSrc,
		final String			DocName
	)
	{
		synchronized (NameMapSynchObj) {
			NameMap.put(DocName, DocSrc.getURL());
		}
		return DocSrc;
	}
}
