/*	QWICAP (Quick Web Interface for Conventional Applications) by Chris W. Johnson
	Copyright (C) 2007-2011 The University of Texas at Austin ("U.T. Austin"). All rights reserved.
	
	U.T. Austin's copyright notice in this evolving version reflects our original authorship. It does not assert 
	copyright over changes contributed by others.
	
	This library is free software; you can redistribute it and/or modify it under the terms of the GNU Lesser General 
	Public License as published by the Free Software Foundation; either version 2.1 of the License, or (at your option) 
	any later version.
	
	This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied 
	warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for 
	more details.
	
	You should have received a copy of the GNU Lesser General Public License along with this library; if not, write to 
	the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/


package edu.utexas.its.eis.tools.qwicap.servlet;


import edu.utexas.its.eis.tools.qwicap.util.ArrayToString;

import java.util.LinkedHashSet;
import java.util.Locale;
import java.util.Set;
import java.util.logging.Level;
import java.util.logging.Logger;


/**	Represents everything we know, at a given time, about the localizations that are to be performed-by/applied-to a
 * 	a particular web application session (or, in the absence of a session, to each HTTP GET operation).
 * 	
 * 	@author Chris W. Johnson
 */
final class i18nLocaleSettings {
	
	private static final Logger		Log = Logger.getLogger(i18nLocaleSettings.class.getName());
	
//	Java Locale in use for number formatting, etc.	

	private final Locale			CurrentLocale;

//	CurrentLocale represented as a language tag.

	private final HTTPLanguageTag	CurrentLocaleTag;
	
//	The translation of this web application that is to be employed.
	
	private final HTTPLanguageTag	WebAppTranslation;
	
//	The intersection of the current language tag's various subtag combinations with the localizations supported by
//	this web application. Each element represents the name of a directory within the web application where we should
//	look for documents when they are requested. The name of the default directory (which is the root directory of the
//	web application) is not included in this list. The document search methods are expected to know that they should 
// 	fall-back to the web app's root directory, if all else fails.

	private final String[]			WebAppTranslationDirs;
	
	private final int				HashCode;

	
	static i18nLocaleSettings create
	(
		final i18nLocaleList			KnownLocales,
		final i18nLocalizationsList		WebAppLangs,	//	Guaranteed to be a subset of KnownLocales.
		final HTTPAcceptLanguageList	ClientLangs
	)
	{
		HTTPLanguageRange		LangRange;
		HTTPLanguageTag			BestLocale      = null;
		HTTPLanguageTag			BestTranslation = null;
		
		for (final HTTPAcceptElement<HTTPLanguageRange> LangRangeElem : ClientLangs) {
			
			if (LangRangeElem.getQualityInt() <= 0)
				break;	//	We've found an "accept" element whose quality is less than or equal to zero, so we're done.
			
			LangRange       = LangRangeElem.getIdentifier();
			BestLocale      = WebAppLangs.getApplicableLocales().compareTagToListGetBest(LangRange);
			BestTranslation = WebAppLangs.getTranslations().compareListToTagGetBest(LangRange);		//	todo This doesn't do what we'd intended.

			if (BestLocale != null)
				break;
		}
		
		if (BestLocale == null) {
			
		//	We have to use some locale, even though none of our locales match those requested by the client, because
		//	it's better to respond to the client with a working web application in an unwanted language than to have to 
		//	return an error to the client and dis-allow access to the web application. So, when all else fails, we use
		//	the default Locale of this JVM.
			
			LangRange = new HTTPLanguageRange(Locale.getDefault());
			
			BestLocale = WebAppLangs.getApplicableLocales().compareTagToListGetBest(LangRange);
			if (BestLocale == null)
				BestLocale = LangRange;
			
			BestTranslation = WebAppLangs.getTranslations().compareListToTagGetBest(LangRange);
			if (BestTranslation == null)
				BestTranslation = new HTTPLanguageTag("");	//	Use the default language in the web app.
		}
		
		return new i18nLocaleSettings(KnownLocales, WebAppLangs, BestLocale, BestTranslation);
	}
	
	i18nLocaleSettings
	(
		final i18nLocaleList		KnownLocales,
		final i18nLocalizationsList	WebAppTranslations,
		final Locale				ClientLocale,		//	The client's ideal locale, limited to locales that at least partially match those supported by the web app.
		final Locale				WebAppTranslation	//	The locale supported by the web app that was closest to the client's ideal locale.
	)
	{
		this(KnownLocales, WebAppTranslations, new HTTPLanguageTag(ClientLocale), new HTTPLanguageTag(WebAppTranslation));
	}
	
	i18nLocaleSettings
	(
		final i18nLocaleList		KnownLocales,
		final i18nLocalizationsList	WebAppTranslations,	//	Guaranteed to be a subset of KnownLocales.
		final HTTPLanguageTag		ClientLocale,		//	The client's ideal locale, limited to locales that at least partially match those supported by the web app.
		final HTTPLanguageTag		WebAppTranslation	//	The locale supported by the web app that was closest to the client's ideal locale.
	)
	{
		this.CurrentLocale         = KnownLocales.getLocale(ClientLocale);
		this.CurrentLocaleTag      = ClientLocale;
		this.WebAppTranslation     = WebAppTranslation;
		this.WebAppTranslationDirs = getSearchDirs(WebAppTranslations, WebAppTranslation);
		this.HashCode              = createHashCode(CurrentLocale, CurrentLocaleTag, WebAppTranslationDirs);
		
		if (CurrentLocale == null)
			throw new IllegalArgumentException("The language tag \"" + ClientLocale + "\" does not correspond to a Locale supported by this Java installation.");
		
		if (Log.isLoggable(Level.FINE)) {
			final StringBuilder			Buff = new StringBuilder(256);
			
			if (WebAppTranslations.getTranslations().size() == 0)
				Buff.append("This web application includes no translations.");
			else {
				Buff.append("This web application includes the following translation(s): ").append(WebAppTranslations.getTranslations()).append('.');
				Buff.append(" It will use the translation \"").append(WebAppTranslation).append("\", and will search for its documents in: ").append(new ArrayToString(WebAppTranslationDirs).alwaysQuote()).append('.');
			}
			
			Buff.append(" It will employ the \"").append(CurrentLocale).append("\" locale for the formatting of numbers, dates, etc.");
			
			Log.fine(Buff.toString());
		}
	}
	
	static String[] getSearchDirs
	(
		final i18nLocalizationsList	WebAppLangs,
		final HTTPLanguageTag		ClientLang
	)
	{
		final HTTPLanguageTagsWithDefault	Translations = WebAppLangs.getTranslations();
		final LinkedHashSet<String>			SearchDirs   = new LinkedHashSet<String>();
		
	//	Get the intersection of the current language tag's various subtag combinations with the localizations supported 
	//	by this web application.
		
		addSupportedTranslationsToList(Translations, ClientLang, SearchDirs);
		
	//	Also include the web application's default translation (and its ancestors) in the search path.
		
		final HTTPLanguageTag	DefaultTranslation = Translations.getDefault();
								
		if (DefaultTranslation != null)
			addSupportedTranslationsToList(Translations, DefaultTranslation, SearchDirs);
		
		return SearchDirs.toArray(new String[SearchDirs.size()]);
	}
	
	private static void addSupportedTranslationsToList
	(
		final HTTPLanguageTagsWithDefault	Translations,
		final HTTPLanguageTag				Translation,
		final Set<String>					SearchDirs
	)
	{
		final HTTPLanguageTag	DefaultTranslation = Translations.getDefault();
		final HTTPLanguageTag[]	SubtagCombos       = Translation.getSubtagCombinations();
			
		for (final HTTPLanguageTag LangTag : SubtagCombos) {
			final String			LangTagStr = LangTag.toString();
				
			if (Translations.getSupportedIdentifier(LangTagStr) != null) {
				
			//	Yes, a REFERENCE equality test is what we want below.
				
				if (LangTag == DefaultTranslation) {
					
				//	LangTag is the same object referenced by DefaultTranslation, so reverse the physical-to-logical
				//	directory name translation that occurred in the i18nLocalizationsList constructor, by adding the 
				//	default translation postfix back onto the logical directory name.
					
					SearchDirs.add(LangTagStr + i18nLocalizationsList.kDefaultTranslationPostfix);
					
				} else
					SearchDirs.add(LangTagStr);
			}	
		}
	}
	
	private static int createHashCode
	(
		final Locale			CurrentLocale,
		final HTTPLanguageTag	CurrentLocaleTag,
		final String[]			WebAppTranslationDirs
	) 
	{
		int						HashCode = CurrentLocale.hashCode() ^ CurrentLocaleTag.hashCode();
		
		for (final String WebAppTransDir : WebAppTranslationDirs)
			HashCode ^= WebAppTransDir.hashCode();
		
		return HashCode;
	}
	
	@Override public int hashCode() {
		return HashCode;
	}
	
	@Override public boolean equals
	(
		final Object			Other	
	)
	{
		if (Other instanceof i18nLocaleSettings)
			return false;
	
		final i18nLocaleSettings	O = (i18nLocaleSettings) Other;
		
		if (CurrentLocale.equals(O.CurrentLocale) == false)
			return false;
		
		if (CurrentLocaleTag.equals(O.CurrentLocaleTag) == false)
			return false;
		
		if (WebAppTranslationDirs.length != O.WebAppTranslationDirs.length)
			return false;
		
		for (int Index = 0, Count = WebAppTranslationDirs.length; Index < Count; Index++) {
			
			if (WebAppTranslationDirs[Index].equals(O.WebAppTranslationDirs[Index]) == false)
				return false;
		}
		
		return true;
	}
	
	Locale getLocale() {
		return CurrentLocale;
	}
	
	HTTPLanguageTag getLocaleAsLanguageTag() {
		return CurrentLocaleTag;
	}
	
	HTTPLanguageTag getWebAppTranslation() {
		return WebAppTranslation;
	}

	String[] getWebAppTranslationSearchDirs() {
		return WebAppTranslationDirs;
	}
}

/*	Pre-implementation ramblings:
 *
 *	This class recognizes the awkward fact that, at any given time, a particular session has three localization 
 * 	settings: (1) The ideal localization(s) specified by the client, (2) the best match to that ideal that can be 
 * 	found among the <code>Locale</code> implementations built into Java, and (3) the localization of a particular web 
 * 	application that best matches the Java <code>Locale</code>. This is further complicated by the case of web 
 * 	applications that lack localizations that would be acceptable to the client. In this case, the JVM's default
 * 	<code>Locale</code> is assumed to be the appropriate localization.
 * 	
 * 	....
 * 
 * 	When there is no intersection between the locale(s) acceptable to the client, the locales supported by the JVM, and
 * 	the locales supported by the web application, we default to using the JVM's default locale. Of course, the JVM's
 * 	default locale may have no intersection with the locales supported by the web application. (That may be because the 
 * 	web application has no specific localizations; all of its files are in the default locations, and therefore are not 
 * 	associated with any locale.) In this worst-case scenario, we use the JVM's default locale for interpreting numbers,
 * 	dates, etc., and the web application's default documents for everything else. 
 * 
 * 	In a related manner, we may (I think) use a Java <code>Locale</code> that is more specific to the client's 
 * 	requirements than the web application localization we are using. For instance, if the client requests "en_GB", 
 * 	we will find a matching Java <code>Locale</code>, but we might find that the web application supports nothing 
 * 	more specific than an "en" localization, or no specific localization at all. The same situation can arise when we
 * 	must fall back on the JVM's default <code>Locale</code>, and it has only a partial intersection, or no intersection,
 * 	with the locales explicitly supported by the web application. In all such cases of locale mis-match, we should
 * 	recognize the mis-match in our document search implementation, such that we do NOT search for localizations that
 * 	we already know are not present in the web application. For example, when there are no localizations in a web 
 * 	application, we'll end-up using the JVM's default, which will often be "en_US". We do NOT want to waste space in
 * 	the document cache by searching for, say, "en-us/index.html", "en/index.html" and finally "index.html", when we
 * 	knew in the first place that the web application didn't have specific "en_US" and "en" localizations. Unfortuantely,
 * 	that's what the current i18nDocumentAccessor implementation does, because it isn't aware of the disparity between
 * 	the Java <code>Locale</code> being employed, and it intersection with the web application's localizations. This
 * 	class should represent all of the localization setting that currently apply to a session.
 */