/*	QWICAP (Quick Web Interface for Conventional Applications) by Chris W. Johnson
	Copyright (C) 2007-2010 The University of Texas at Austin ("U.T. Austin"). All rights reserved.
	
	U.T. Austin's copyright notice in this evolving version reflects our original authorship. It does not assert 
	copyright over changes contributed by others.
	
	This library is free software; you can redistribute it and/or modify it under the terms of the GNU Lesser General 
	Public License as published by the Free Software Foundation; either version 2.1 of the License, or (at your option) 
	any later version.
	
	This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied 
	warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for 
	more details.
	
	You should have received a copy of the GNU Lesser General Public License along with this library; if not, write to 
	the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/


package edu.utexas.its.eis.tools.qwicap.servlet;


import edu.utexas.its.eis.tools.qwicap.util.ArrayToString;

import javax.servlet.ServletContext;
import java.util.*;
import java.util.logging.Logger;
import java.util.logging.Level;


/**	A list of the available translations/localizations of this web application, compiled by examining the contents of 
 * 	the web application, and noting the existence of top-level directories corresponding to known locale identifiers.
 * 
 * 	@author Chris W. Johnson
 */
final class i18nLocalizationsList {

	private static final Logger					Log = Logger.getLogger(i18nLocalizationsList.class.getName());
	
	static final String							kDefaultTranslationPostfix = " (default)";
	
	private final HTTPLanguageTagsWithDefault	Translations;
	private final HTTPLanguageTags				Localizations;
	

	i18nLocalizationsList
	(
		final i18nLocaleList	KnownLocales,
		final ServletContext	SCtxt
	)
	{
		this(KnownLocales, SCtxt.getResourcePaths("/"));
	}

	i18nLocalizationsList
	(
		final i18nLocaleList	KnownLocales,
		final Set<String>		TopLevelRsrcs
	)
	{
		final HashSet<HTTPLanguageTag>	Localizations      = new HashSet<HTTPLanguageTag>(KnownLocales.size());
		final HashSet<HTTPLanguageTag>	Translations       = new HashSet<HTTPLanguageTag>(21);
		HTTPLanguageTag					DefaultTranslation = null;
		
		for (final String TopLevelRsrc : TopLevelRsrcs) {

		//	If the resource's name ends with a slash, it's a directory. If not, it's a file, and we don't care about it.

			if (TopLevelRsrc.endsWith("/") == false)
				continue;

		//	Search for a known HTTPLanguageString matching TopLevelRsrc name, without the trailing slash.
			
			final String			PhysicalDirName = TopLevelRsrc.substring(1, TopLevelRsrc.length() - 1);
			final String			LogicalDirName;
			
		//	Filter DirName for the default translation marker string. If present, remove it, and make a note of the 
		//	default localization.
			
			if (PhysicalDirName.endsWith(kDefaultTranslationPostfix))
				LogicalDirName = PhysicalDirName.substring(0, PhysicalDirName.length() - kDefaultTranslationPostfix.length());
			else
				LogicalDirName = PhysicalDirName;
			
			final HTTPLanguageTag	HTTPLang = KnownLocales.getKnownHTTPLanguageTag(LogicalDirName);
			
			if (HTTPLang != null) {
				
			//	Yes, we really DO want to make a REFERENCE inequality test here.
				
				if (PhysicalDirName != LogicalDirName) {
					
					DefaultTranslation = HTTPLang;
					Log.log(Level.INFO, "The default translation of this web application will be \"{0}\", which is located in the directory named \"{1}\".", new Object[] { DefaultTranslation, PhysicalDirName });
				}
				
			//	DirName/HTTPLang correspond to a translation that is present in this web application.
				
				Translations.add(HTTPLang);
				
			//	Treating LogicalDirName as an HTTP language range, get a list of the known locales that it matches. 
			//	These are the Locales that we would be willing to employ with this translation. For instance, given a 
			//	translation of "en", we'd be willing to use the Locales: "en", "en_US", "en_GB", etc.
				
				Localizations.addAll(KnownLocales.getMatchingTags(new HTTPLanguageRange(LogicalDirName)));
				
			} else {
				
			//	Yes, we really DO want to make a REFERENCE equality test here.
				
				if (PhysicalDirName == LogicalDirName)
					Log.log(Level.FINER, "Directory \"{0}\" is not a localization directory.", PhysicalDirName);
				else
					Log.log(Level.SEVERE, "Directory \"{0}\" is named as if it represents the default translation for this web application (it ends with \"{2}\"), but no locale named \"{1}\" is known to this JVM, so it cannot be employed as an actual translation.", new Object[] { PhysicalDirName, LogicalDirName, kDefaultTranslationPostfix});
			}
		}
				
	//	Convert the Localizations HashSet to an array of HTTPLanguageTag objects, since that'll be more useful in some
	//	situations.
		
		this.Localizations = new HTTPLanguageTags(Localizations.toArray(new HTTPLanguageTag[Localizations.size()]));
		this.Translations  = new HTTPLanguageTagsWithDefault(Translations.toArray(new HTTPLanguageTag[ Translations.size()]), DefaultTranslation);
		
		if (Log.isLoggable(Level.FINE)) {
			
			if (this.Translations.size() == 0)
				Log.log(Level.FINE, "This web application includes no recongized translations. The server''s default locale, \"{0}\", will always be used.", Locale.getDefault());
			else
				Log.log(Level.FINE, "This web application includes the following translations: {0}.", new ArrayToString(Translations).alwaysQuote());
		}
	}
	
	HTTPLanguageTag getDefaultTranslation() {
		return Translations.getDefault();	
	}
	
	HTTPLanguageTagsWithDefault getTranslations() {
		return Translations;
	}
	
	HTTPLanguageTags getApplicableLocales() {
		return Localizations;
	}
	
	@Override public String toString() {
		return "Web application translations: " + Translations + "; Applicable locales: " + Localizations + '.';
	}
}
