/*	QWICAP (Quick Web Interface for Conventional Applications) by Chris W. Johnson
	Copyright (C) 2007 The University of Texas at Austin ("U.T. Austin"). All rights reserved.
	
	U.T. Austin's copyright notice in this evolving version reflects our original authorship. It does not assert 
	copyright over changes contributed by others.
	
	This library is free software; you can redistribute it and/or modify it under the terms of the GNU Lesser General 
	Public License as published by the Free Software Foundation; either version 2.1 of the License, or (at your option) 
	any later version.
	
	This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied 
	warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for 
	more details.
	
	You should have received a copy of the GNU Lesser General Public License along with this library; if not, write to 
	the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/


package edu.utexas.its.eis.tools.qwicap.servlet;


import edu.utexas.its.eis.tools.qwicap.template.xml.TagException;
import edu.utexas.its.eis.tools.qwicap.template.xml.immutable.XMLDocument;
import edu.utexas.its.eis.tools.qwicap.util.ArrayToString;

import java.io.IOException;
import java.math.BigDecimal;
import java.math.BigInteger;
import java.net.URL;
import java.text.*;
import java.util.*;
import java.util.logging.Level;
import java.util.logging.Logger;
import java.sql.Time;


/**	Internationalization (i18n) Manager class, specific to each session. Note that an instance of this class may be
 * 	accessed by code running in the <code>QwicapThread</code> corresponding to a particular session, as well as 
 * 	directly from <code>QwicapServlet</code>, when this session receives a GET request for a document. Therefore, a
 * 	modicum of synchronization is required.
 * 
 * 	@author Chris W. Johnson
 */
final class i18nManager {

	private static final Logger				Log = Logger.getLogger(i18nManager.class.getName());
		
	private final i18nLocaleList					KnownLocales;
	private final i18nDocumentAccessorWithNameMap	DocAcc;
	
	private final ClassLoader				CLoader;

	private final Object					CurrentLocaleSynchObj = new Object();
	private i18nLocaleSettings				CurrentLocaleSettings;
	private NumberFormat					IntFormat;	//	Only used for formatting - never parsing.
	private DecimalFormat					DecFormat;	//	Used for parsing integer and decimal strings, and for formatting decimal strings.
	private DateFormat						DateTimeFmt;
	private DateFormat						DateFmt;
	private DateFormat						TimeFmt;
	
	private final HashMap<Class, i18nProxy> Proxies = new HashMap<Class, i18nProxy>();
	

	i18nManager
	(
		final ClassLoader			CLoader,
		final i18nLocaleList		KnownLocales,
		final i18nDocumentAccessor	DocAccessor,
		final String				HTTPAcceptLanguageHeaderStr
	)
	{
		this.CLoader      = CLoader;
		this.KnownLocales = KnownLocales;
		this.DocAcc       = new i18nDocumentAccessorWithNameMap(DocAccessor);
		
		setLocaleByHTTPAcceptLanguageHeader(HTTPAcceptLanguageHeaderStr);
	}
	
	DocumentCache getDocumentCache() {
		return DocAcc.getDocumentCache();
	}
	
	public i18nServices getServicesForClass
	(
		final Class<?>			C
	)
	{
		i18nProxy				Proxy = Proxies.get(C);

		if (Proxy != null)
			return Proxy;
		
		
		i18nXMLPropertiesLocaleChain	PropsChain;
		MissingResourceException		PropsChainException;
		
		try {
			PropsChain          = i18nXMLPropertiesLocaleChain.get(this, CLoader, C, getLocale());
			PropsChainException = null;
		} catch (MissingResourceException e) {
			PropsChain          = null;
			PropsChainException = e;
		}

		Proxy = new i18nProxy(this, C, PropsChain, PropsChainException);
		Proxies.put(C, Proxy);

		return Proxy;
	}

	public BigInteger parseInteger
	(
		final String			IntStr
	)
	throws ParseException, ArithmeticException	//	ArithmeticException is thrown when there's a fractional component to this would-be intereger quantity.
	{
		return parseDecimal(IntStr).toBigIntegerExact();
	}

	public BigDecimal parseDecimal
	(
		final String			DecStr
	)
	throws ParseException
	{
		final String			TrimmedDecStr = DecStr.trim();
		final ParsePosition		ParsePos      = new ParsePosition(0);
		final BigDecimal		Number;

		synchronized (CurrentLocaleSynchObj) {
			Number = (BigDecimal) DecFormat.parse(DecStr.trim(), ParsePos);
		}

		if (ParsePos.getErrorIndex() >= 0)
			throw new ParseException("The string \"" + TrimmedDecStr + "\" could not be completely interpreted as a number. Error at character " + ParsePos.getErrorIndex() + " ('" + TrimmedDecStr.charAt(ParsePos.getErrorIndex()) + "').", ParsePos.getErrorIndex());

		if (ParsePos.getIndex() < TrimmedDecStr.length())
			throw new ParseException("The string \"" + TrimmedDecStr + "\" could not be completely interpreted as a number. Parsing stopped at character " + ParsePos.getIndex() + " ('" + TrimmedDecStr.charAt(ParsePos.getIndex()) + "').", ParsePos.getIndex());

		return Number;
	}
	
	public String formatNumber
	(
		final Number			No
	)
	{
		if (No instanceof BigDecimal || No instanceof Double || No instanceof Float)
			return formatDecimal(No);
		else
			return formatInteger(No);
	}
	
	public String formatInteger
	(
		final Number			No
	)
	{
		synchronized (CurrentLocaleSynchObj) {
			return IntFormat.format(No);
		}
	}

	public String formatDecimal
	(
		final Number			No
	)
	{
		synchronized (CurrentLocaleSynchObj) {
			return DecFormat.format(No);
		}
	}
	
	public Date parseDateTime
	(
		final String			Text
	)
	throws ParseException
	{
		synchronized (CurrentLocaleSynchObj) {
			return DateTimeFmt.parse(Text);
		}
	}
	
	public java.sql.Date parseDate
	(
		final String			Text
	)
	throws ParseException
	{
		synchronized (CurrentLocaleSynchObj) {
			return new java.sql.Date(DateFmt.parse(Text).getTime());
		}
	}
	
	public Time parseTime
	(
		final String			Text
	)
	throws ParseException
	{
		synchronized (CurrentLocaleSynchObj) {
			return new Time(TimeFmt.parse(Text).getTime());
		}
	}
	
	public String formatDateTime
	(
		final long				DTMillis
	)
	{
		synchronized (CurrentLocaleSynchObj) {
			return DateTimeFmt.format(new Date(DTMillis));
		}
	}
	
	public String formatDateTime
	(
		final Date				DTMillis
	)
	{
		synchronized (CurrentLocaleSynchObj) {
			return DateTimeFmt.format(DTMillis);
		}
	}
	
	public String formatDateTime
	(
		final Calendar			DTMillis
	)
	{
		synchronized (CurrentLocaleSynchObj) {
			return DateTimeFmt.format(DTMillis.getTime());
		}
	}
	
	public String formatDate
	(
		final long				DTMillis
	)
	{
		synchronized (CurrentLocaleSynchObj) {
			return DateFmt.format(new Date(DTMillis));
		}
	}
	
	public String formatDate
	(
		final Date				DTMillis
	)
	{
		synchronized (CurrentLocaleSynchObj) {
			return DateFmt.format(DTMillis);
		}
	}
	
	public String formatDate
	(
		final Calendar			DTMillis
	)
	{
		synchronized (CurrentLocaleSynchObj) {
			return DateFmt.format(DTMillis.getTime());
		}
	}
	
	public String formatTime
	(
		final long				DTMillis
	)
	{
		synchronized (CurrentLocaleSynchObj) {
			return TimeFmt.format(new Date(DTMillis));
		}
	}
	
	public String formatTime
	(
		final Date				DTMillis
	)
	{
		synchronized (CurrentLocaleSynchObj) {
			return TimeFmt.format(DTMillis);
		}
	}
	
	public String formatTime
	(
		final Calendar			DTMillis
	)
	{
		synchronized (CurrentLocaleSynchObj) {
			return TimeFmt.format(DTMillis.getTime());
		}
	}
	
	public DocumentSource getDocument
	(
		final String			DocName
	)
	throws IOException
	{
		return DocAcc.getDocument(CurrentLocaleSettings, DocName);
	}

	public DocumentSource getDocument
	(
		final URL				DocURL
	)
	throws IOException
	{
		return DocAcc.getDocument(CurrentLocaleSettings, DocURL);
	}
	
	public XMLDocument getXMLDocument
	(
		final String			FileName
	)
	throws TagException, IOException
	{
		return getDocument(FileName).getXML();
	}

	public XMLDocument getXMLDocument
	(
		final URL				DocURL
	)
	throws TagException, IOException
	{
		return getDocument(DocURL).getXML();
	}

	public Locale getLocale() {

		synchronized (CurrentLocaleSynchObj) {
			return CurrentLocaleSettings.getLocale();
		}
	}

	public HTTPLanguageTag getHTTPLanguageTag() {

		synchronized (CurrentLocaleSynchObj) {
			return CurrentLocaleSettings.getLocaleAsLanguageTag();
		}
	}

	i18nManager setLocaleByHTTPAcceptLanguageHeader
	(
		final String			AcceptLangHdr	//	Can be null.
	)
	{	
		setLocale(i18nLocaleSettings.create(KnownLocales, DocAcc.getLocalizations(), new HTTPAcceptLanguageList(AcceptLangHdr)));

		return this;
	}

	private void setLocale
	(
		final i18nLocaleSettings	Settings
	)
	{
		synchronized (CurrentLocaleSynchObj) {

			if (CurrentLocaleSettings != null && CurrentLocaleSettings.equals(Settings))
				return;
			
			CurrentLocaleSettings = Settings;

			updateFormats();
			updateProxies();
			
			if (Log.isLoggable(Level.FINE))
				Log.log(Level.FINE, "Locale set to \"{0}\". Language tag subtype combinations: {1}.", new Object[] { CurrentLocaleSettings.getLocale(), new ArrayToString(CurrentLocaleSettings.getWebAppTranslationSearchDirs()).alwaysQuote().toString() });
		}
	}

	private void updateFormats() {
		final Locale			CurrentLocale = CurrentLocaleSettings.getLocale();
		
	//	IntFormat is used only for formatting integer values. We can't use it for parsing, because it includes no 
	//	provision for returning values as BigInteger objects.
		
		IntFormat = NumberFormat.getIntegerInstance(CurrentLocale);
		
	//	DecFormat is used for parsing both integer and decimal values, as well as for formatting decimal values.
		
		DecFormat = (DecimalFormat) NumberFormat.getNumberInstance(CurrentLocale);
		DecFormat.setParseBigDecimal(true);
		DecFormat.setParseIntegerOnly(false);
		DecFormat.setMaximumFractionDigits(50);
		
		DateTimeFmt = DateFormat.getDateTimeInstance(DateFormat.LONG, DateFormat.LONG, CurrentLocale);
		DateFmt     = DateFormat.getDateInstance(DateFormat.LONG, CurrentLocale);
		TimeFmt     = DateFormat.getTimeInstance(DateFormat.LONG, CurrentLocale);
	}

	private void updateProxies() {

	//	Inform every proxy of the change of locale.

		for (final i18nProxy Proxy : Proxies.values()) {
			final Class<?>			C = Proxy.getProxyClass();

			Proxy.setProxyBundle(i18nXMLPropertiesLocaleChain.get(this, C.getClassLoader(), C, getLocale()));
		}
	}
}
