/*	QWICAP (Quick Web Interface for Conventional Applications) by Chris W. Johnson
	Copyright (C) 2003-2007 The University of Texas at Austin ("U.T. Austin"). All rights reserved.
	
	U.T. Austin's copyright notice in this evolving version reflects our original authorship. It does not assert 
	copyright over changes contributed by others.
	
	This library is free software; you can redistribute it and/or modify it under the terms of the GNU Lesser General 
	Public License as published by the Free Software Foundation; either version 2.1 of the License, or (at your option) 
	any later version.
	
	This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied 
	warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for 
	more details.
	
	You should have received a copy of the GNU Lesser General Public License along with this library; if not, write to 
	the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/


package edu.utexas.its.eis.tools.qwicap.template.css;


/**	<code>CSSParser</code> is a utility class that provides various static methods which assist with the more annoying
 *	aspects of parsing CSS patterns. It is not part of Qwicap's public API.
 */
final class CSSParser {
	
	
	static String getIdentifierOrString
	(
		final CharacterIterator	Chars
	)
	{
		final StringBuilder		Buff = new StringBuilder(32);
		
		if (is_string(Chars, Buff) || is_identifier(Chars, Buff))
			return Buff.toString();
		
		throw new IllegalArgumentException("The string is neither a valid CSS identifier, nor a valid CSS string. " +
			"See " + Chars + '.');
	}
	
	static String getIdentifier
	(
		final CharacterIterator	Chars
	)
	{
		final StringBuilder		Buff = new StringBuilder(32);
		
		if (is_identifier(Chars, Buff))
			return Buff.toString();
		
		throw new IllegalArgumentException("CSS2 identifiers (including element names, classes, and IDs in " +
			"selectors) can begin with only the characters [A-Za-z] and ISO 10646 characters 161 and " +
			"higher. After the first character, they may also include hyphens (-), digits (0-9), and characters " +
			"escaped with backslashes ('\\'). See <http://www.w3.org/TR/CSS2/syndata.html#q4>. The supplied " +
			"string does not meet those, or related, requirements. See " + Chars + '.');
	}	
	
	private static boolean is_identifier
	(
		final CharacterIterator	Chars,
		final StringBuilder		Buff
	)
	{
		if (!Chars.hasNext())
			return false;
		
		final char				FirstChar = Chars.next();
		
		if (!(FirstChar >= 'a' && FirstChar <= 'z' || FirstChar >= 'A' && FirstChar <= 'Z' || FirstChar >= 161)) {
			
			Chars.goBack();
			return false;
		}
		
		Buff.append(FirstChar);
		
		while (Chars.hasNext()) {
			final char				Char = Chars.next();
			
			if (Char >= 'a' && Char <= 'z' 
			||  Char >= 'A' && Char <= 'Z' 
			||  Char >= '0' && Char <= '9'
			||  Char == '-' || Char >= 161) {
				
				Buff.append(Char);
				
			} else if (Char == '\\') {
				
				is_escape(Chars.goBack(), Buff);
				
			} else {
				
				Chars.goBack();	//	Make sure the terminating character, which is not a part of this identifier, is not lost.
				break;
			}
		}
		
		return true;
	}
	
	private static boolean is_string
	(
		final CharacterIterator	Chars,
		final StringBuilder		Buff
	)
	{
	//	From <http://www.w3.org/TR/REC-CSS2/syndata.html#strings>:
	//
	//	4.3.10 Strings
	//	
	//	Strings can either be written with double quotes or with single quotes. Double quotes cannot occur inside 
	//	double quotes, unless escaped (as '\"' or as '\22'). Analogously for single quotes ("\'" or "\27").
	//
	//		"this is a 'string'"
	//		"this is a \"string\""
	//		'this is a "string"'
	//		'this is a \'string\''
	//
	//	A string cannot directly contain a newline. To include a newline in a string, use the escape "\A" (hexadecimal A 
	//	is the line feed character in Unicode, but represents the generic notion of "newline" in CSS). See the 'content' 
	//	property for an example.
	//	
	//	It is possible to break strings over several lines, for aesthetic or other reasons, but in such a case the newline 
	//	itself has to be escaped with a backslash (\). For instance, the following two selectors are exactly the same:
	//	
	//		A[TITLE="a not s\
	//		o very long title"] {/*...*/}
	//
	//		A[TITLE="a not so very long title"] {/*...*/}
		
		if (!Chars.hasNext())
			return false;
		
		final char				QuoteChar = Chars.next();
		
		if (QuoteChar != '\'' && QuoteChar != '"')
			return false;
		
		final int				OrigBuffLength     = Buff.length();
		boolean					TerminatedProperly = false;
		
		while (Chars.hasNext()) {
			final char				Char = Chars.next();
			
			if (Char == '\\') {
				
				is_escape(Chars.goBack(), Buff);
				
			} else if (Char == QuoteChar) {
				
				TerminatedProperly = true;
				break;
				
			} else
				Buff.append(Char);
		}
		
		if (TerminatedProperly)
			return true;
		
		Buff.setLength(OrigBuffLength);
		
		throw new IllegalArgumentException("CSS string should have been terminated with the character '" + QuoteChar + 
			"', but wasn't. See " + Chars + '.');
	}
	
/*	The following rules aren't entirely consistent with other parts of the CSS2 specficiation, which are typically more
	liberal, and seemingly more complete, but they do serve as the starting point for the methods below.
	
	From <http://www.w3.org/TR/REC-CSS2/syndata.html#q4>:
	
	MACRO		DEFINITION
	----------- ---------------------------------------------------
	
	ident		{nmstart}{nmchar}*
	
	name		{nmchar}+
	
	nmstart		[a-zA-Z]|{nonascii}|{escape}
	
	nonascii	[^\0-\177]
	
	unicode		\\[0-9a-f]{1,6}[ \n\r\t\f]?
	
	escape		{unicode}|\\[ -~\200-\4177777]
	
	nmchar		[a-z0-9-]|{nonascii}|{escape}
	
	num			[0-9]+|[0-9]*\.[0-9]+
	
	string		{string1}|{string2}
	
	string1		\"([\t !#$%&(-~]|\\{nl}|\'|{nonascii}|{escape})*\"
	
	string2		\'([\t !#$%&(-~]|\\{nl}|\"|{nonascii}|{escape})*\'
	
	nl			\n|\r\n|\r|\f
	
	w			[ \t\r\n\f]*
*/
	
	private static boolean is_escape
	(
		final CharacterIterator	Chars,
		final StringBuilder		Buff
	)
	{
		if (is_unicode(Chars, Buff))
			return true;
		
		if (!Chars.hasNext())
			return false;
		
		if (Chars.next() != '\\') {

			Chars.goBack();
			return false;
		}
		
	//	From <http://www.w3.org/TR/REC-CSS2/syndata.html#q4>:
	//	
	//	[I]nside a string, a backslash followed by a newline is ignored (i.e., the string is deemed not to contain 
	//	either the backslash or the newline).
		
		if (is_nl(Chars, null))
			return true;
		
	//	From <http://www.w3.org/TR/REC-CSS2/syndata.html#q4>:
	//	
	//	[A backslash character] cancels the meaning of special CSS characters. Any character (except a hexadecimal 
	//	digit) can be escaped with a backslash to remove its special meaning. For example, "\"" is a string consisting 
	//	of one double quote. Style sheet preprocessors must not remove these backslashes from a style sheet since that 
	//	would  change the style sheet's meaning.
	//
	//	The definition of "escape" is fairly restrictive: only characters ' ' (space), '-', '~' and characters >=
	//	0x80 and <= 0x10ffff can be escaped, yet it's pretty clear from the above paragraph that any character can be
	//	escaped, except a hex digit. 
		
		if (Chars.hasNext())
			Buff.append(Chars.next());
		
		return true;
	}
	
/*	private static boolean is_w
	(
		final CharacterIterator	Chars,
		final StringBuffer		Buff	//	Can be null.
	)
	{
		if (!Chars.hasNext())
			return false;
		
		if (Chars.next() == ' ') {
			
			if (Buff != null)
				Buff.append(' ');
			
			return true;
		}
		
		return is_nl(Chars, Buff);
	}
*/	
	private static boolean is_nl
	(
		final CharacterIterator	Chars,
		final StringBuilder		Buff	//	Can be null.
	)
	{
		if (!Chars.hasNext())
			return false;
		
		final char				Char = Chars.next();
		
		if (Char == '\n' || Char == '\f') {
			
			if (Buff != null) 
				Buff.append(Char);
			return true;
		}
		
		if (Char == '\r') {
			
			if (Buff != null)
				Buff.append('\r');
			
			if (Chars.hasNext()) {
				
				if (Chars.next() == '\n') {
					
					if (Buff != null)
						Buff.append('\n');
				
				} else
					Chars.goBack();
			}
			
			return true;
		}
		
		Chars.goBack();
		return false;
	}
	
	private static boolean is_unicode
	(
		final CharacterIterator	Chars,
		final StringBuilder		Buff
	)
	{
		if (!Chars.hasNext())
			return false;
		
		if (Chars.next() != '\\') {
			
			Chars.goBack();
			return false;
		}
		
	//	Does the first character following the backslash indicate that this is a hexadecimal sequence?
	//	
	//	From <http://www.w3.org/TR/REC-CSS2/syndata.html#q4>:
	//
	//	[B]ackslash escapes allow authors to refer to characters they can't easily put in a document. In this case, 
	//	the backslash is followed by at most six hexadecimal digits (0..9A..F), which  stand for the ISO 10646 
	//	character with that number. If a digit or letter follows the hexadecimal number,  the end of the number needs 
	//	to be made clear. There are two ways  to do that:
	//
	//	* with a space (or other whitespace character): "\26 B" ("&B")
	//	* by providing exactly 6 hexadecimal digits: "\000026B" ("&B")
	//
	//	In fact, these two methods may be combined. Only one whitespace character is ignored after a hexadecimal 
	//	escape. Note that this means that a "real" space after the escape sequence must itself either be escaped or 
	//	doubled.
		
		char					Char = Chars.next();
		
		if (Char >= 'A' && Char <= 'F' || Char >= '0' && Char <= '9') {
		
		//	Yes, this is a hex sequence.
		
			int						HexValue = nibbleToInt(Char);
			int						HexCharCount = 1;
			
			while (Chars.hasNext()) {
				
				Char = Chars.next();
				
			//	Is Char a hexadecimal character? (Letters 'A' through 'F' must be uppercase.)
				
				if (Char >= 'A' && Char <= 'F' || Char >= '0' && Char <= '9') {
					
					HexValue = HexValue << 4 | nibbleToInt(Char);
					
				//	If this is the 6th hex digit we've gathered, we're done gathering digits.
					
					if (++HexCharCount == 6) {
						
					//	If there's a whitespace character after the final hex digit, discard it.
					//	(See the notes in the next "else if" regarding outstanding questions about this test.)
						
						if (Chars.hasNext() && " \n\r\t\f".indexOf(Chars.next()) == -1)
							Chars.goBack();		//	The next character wasn't whitespace, so backup one character.
						
						break;
					}
					
				} else if (" \n\r\t\f".indexOf(Char) != -1) {
					
				//	We're done gathering hex digits. Eat the whitespace, because it's a terminator for hex-sequences.
				//	(In this case, for hex-sequences of fewer than 6 digits.)
					
				//	If we encounter a terminating sequence of "\r\n", are we meant to eat the '\r' and leave behind
				//	the '\n', or should we understand that the combination constitutes a new-line indicator (any 
				//	new-line counts as whitespace) and therefore eliminate the entire "\r\n" sequence? The CSS2 
				//	spec doesn't really address this question.
					
					break;
					
				} else {
					
				//	We've found a non-hexadecimal character (that also isn't whitespace). That means we're
				//	done. (Ensure that the non-hexadecimal character will not be lost.)
					
					Chars.goBack();
					break;
				}
			}
			
			Buff.append((char) HexValue);
			return true;
			
		} else {
			
		//	Whatever's being escaped, it isn't a hex sequence, so it doesn't concern us.
			
			Chars.goBack();	//	Give-back the character that wasn't a hex digit.
			Chars.goBack();	//	Give-back the backslash character.
		}
		
		return false;
	}
	
	private static int nibbleToInt
	(
		final char				HexChar
	)
	{
		if (HexChar >= '0' && HexChar <= '9')
			return HexChar - '0';
		else
			return HexChar - ('A' - 10);
	}
	
/*	private static boolean is_nonascii
	(
		final CharacterIterator	Chars,
		final StringBuffer		Buff
	)
	{
		if (!Chars.hasNext())
			return false;
		
		final char				Char = Chars.next();
		
		if (Char > 127) {
			
			Buff.append(Char);
			return true;
		}
		
		Chars.goBack();
		return false;
	}
*/
}



