/*	QWICAP (Quick Web Interface for Conventional Applications) by Chris W. Johnson
	Copyright (C) 2003-2010 The University of Texas at Austin ("U.T. Austin"). All rights reserved.
	
	U.T. Austin's copyright notice in this evolving version reflects our original authorship. It does not assert 
	copyright over changes contributed by others.
	
	This library is free software; you can redistribute it and/or modify it under the terms of the GNU Lesser General 
	Public License as published by the Free Software Foundation; either version 2.1 of the License, or (at your option) 
	any later version.
	
	This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied 
	warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for 
	more details.
	
	You should have received a copy of the GNU Lesser General Public License along with this library; if not, write to 
	the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/


//  CSSPattern.java
//  qwicap
//
//  Created by Chris W. Johnson on Mon Jun 23 2003.


package edu.utexas.its.eis.tools.qwicap.template.css;


import edu.utexas.its.eis.tools.qwicap.template.xml.MarkupEnumeration;
import edu.utexas.its.eis.tools.qwicap.template.xml.DescendantEnumerator;
import edu.utexas.its.eis.tools.qwicap.template.xml.structure.Range;
import edu.utexas.its.eis.tools.qwicap.template.xml.structure.TagWithAttributes;


/**	<code>CSSPattern</code> represents a single, arbitrarily complex CSS2 selector pattern.
 *
 * 	@see <a href="http://www.w3.org/TR/REC-CSS2/selector.html">Cascading Style Sheets, level 2,
 *	CSS2 Specification, Selectors</a>
 *	
 *	@author Chris W. Johnson
 */
final class CSSPattern {
	
	CSSPattern					NextPat;
	
	private CSSPatternComponent	List;
	private CSSPatternComponent	ListEnd;
	
	
/**	Creates a <code>CSSPattern</code> object to represent a single, arbitrarily complex CSS2 selector pattern. 
 *	Ordinarily, this constructor is used only by the {@link CSSPatterns} class.
 *
 * @param Patterns		A reference to the larger pattern (combination of patterns) of which this pattern is a part.
 * @param Chars			A CSS2 selector pattern.
 */
	CSSPattern
	(
		final CSSPatterns		Patterns,
		final CharacterIterator	Chars
	)
	{
		Chars.eatWhitespace();	//	Advance past any leading whitespace.
		
		while (Chars.hasNext()) {
			
		//	If we've found the pattern separator character, we've found the end of this pattern. (We peek, because
		//	the CSSPatterns object expects us to return with the comma, if any, as the next character in Chars.)
			
			if (Chars.peek() == ',')
				break;
			
		//	We call Combinator.read first, because it can be handed anything without throwing an exception. If it 
		//	doesn't like what it's given, it rewinds the CharacterIterator to the first non-whitspace character it 
		//	encountered and returns null.
			
			CSSPatternComponent		SearchObj = Combinator.read(Chars);
			
			if (SearchObj == null) {
			
			//	We call Selector.read second, because if it doesn't get what it expects, it ordinarily throws an 
			//	exception. Since a CSSPattern should only be composed of Combinators and Selectors, and terminated
			//	by the end of the string, or a comma marking the start of the next pattern, this arrangement is 
			//	appropriate.
				
				SearchObj = Selector.read(Patterns, Chars);
				if (SearchObj == null)
					throw new IllegalArgumentException("CSS patterns must consist of nothing other than selectors and cominators. This one includes material that falls into neither catagory. See: \"" + Chars + "\".");
			}
			
			addPart(SearchObj);
		}
		
	//	A dangling CombinatorDescendant can be caused by a space trailing the final selector in the pattern, because
	//	the space character is the symbol for the CombinatorDescendant. Here we remove such a dangling combinator.
		
		if (ListEnd instanceof CombinatorDescendant)
			removeLastPart();
	}
	
	private void removeLastPart() {
		
		if (List != null) {
			
			if (List == ListEnd) {
				
				List = null;
				ListEnd = null;
				
			} else {
				CSSPatternComponent		Part = List;
				
				while (Part.NextPart != ListEnd)
					Part = Part.NextPart;
				
				Part.NextPart = null;
				ListEnd = Part;
			}
		}
	}
	
	private void addPart
	(
		final CSSPatternComponent 	NewPart
	)
	{
		if (List == null)
			List = NewPart;
		else
			ListEnd.NextPart = NewPart;
		
		ListEnd = NewPart;
	}
	
/**	Recreates the string representation of this CSS2 pattern.
 *
 * @return		The string representation of this CSS2 pattern.
 */
	@Override public String toString() {
		final StringBuilder		Buff = new StringBuilder(64);
		CSSPatternComponent		Part = List;
		
		while (Part != null) {
			
			Buff.append(Part.toString());
			Part = Part.NextPart;
		}
		
		return Buff.toString();
	}

	void search
	(
		final MarkupEnumeration	Enumerator,
		final Results			Results
	)
	{
		final Selector				Sel   = (Selector) List;	//	The invoking methods all ensure that this will not be null.
		final Combinator			Comb  = (Combinator) (Sel != null ? Sel.NextPart : null);
		final DescendantEnumerator	Elems = new DescendantEnumerator(Enumerator);
		
		while (Elems.hasMoreElements()) {
			final Range					Elem = Elems.nextRange();
			
			if (Elem.isTagWithAttributes() == false)
				continue;
			
			if (Sel.selects((TagWithAttributes) Elem, Elems.getChildNo())) {
				
				if (Comb != null)
					Comb.search(Elems, (Selector) Comb.NextPart, Results);
				else
					Results.addMatch(Elems.getMatch());
			}
		}
	}
}




