/*	QWICAP (Quick Web Interface for Conventional Applications) by Chris W. Johnson
	Copyright (C) 2003-2010 The University of Texas at Austin ("U.T. Austin"). All rights reserved.
	
	U.T. Austin's copyright notice in this evolving version reflects our original authorship. It does not assert 
	copyright over changes contributed by others.
	
	This library is free software; you can redistribute it and/or modify it under the terms of the GNU Lesser General 
	Public License as published by the Free Software Foundation; either version 2.1 of the License, or (at your option) 
	any later version.
	
	This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied 
	warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for 
	more details.
	
	You should have received a copy of the GNU Lesser General Public License along with this library; if not, write to 
	the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/


//  CSSPatterns.java
//  qwicap
//
//  Created by Chris W. Johnson on Mon Jun 23 2003.


package edu.utexas.its.eis.tools.qwicap.template.css;


import edu.utexas.its.eis.tools.qwicap.template.xml.structure.Namespace;
import edu.utexas.its.eis.tools.qwicap.template.xml.immutable.ImNamespace;
import edu.utexas.its.eis.tools.qwicap.template.xml.MarkupEnumeration;

import java.util.*;


/**	<code>CSSPatterns</code> represents one or more <a href='http://www.w3.org/TR/CSS21/selector.html#q1'>CSS2 patterns</a>.
 *	
 *	@author Chris W. Johnson
 */
public final class CSSPatterns {
	
	private static final char[]	     kAtNamespaceURLBegin = "(".toCharArray();
	private static final char[]	     kAtNamespaceURLEnd   = ");".toCharArray();
	private static final Namespace[] kEmptyNamespaceArray = { };	//	Note that this does NOT include the ever-present "xml" and "xmlns" namespaces. The getNamespaceForPrefix method takes care of identifying them.
	
	private final CSSPattern	List;
	private final Namespace[]	Namespaces;		//	Note that this does NOT include the ever-present "xml" and "xmlns" namespaces. The getNamespaceForPrefix method takes care of identifying them.
	private Namespace			DefaultNamespace;
	
	
/**	Creates a <code>CSSPatterns</code> object from a string containing a comma-separated list of 
 *	<a href='http://www.w3.org/TR/CSS21/selector.html#q1'>CSS2 patterns</a>.
 *
 * @param PatternStr	A comma-separated list of CSS2 patterns. If there is no comma in the string, it is assumed
 * 						that it contains only one pattern, which is perfectly acceptable.
 *
 */
	public CSSPatterns
	(
		final String			PatternStr
	)
	{
		final CharacterIterator	Chars = new CharacterIterator(PatternStr);
		
		Namespaces = readAtNamespaceRules(Chars);
		
		List = new CSSPattern(this, Chars);
		
		CSSPattern				ListEnd = List;
		
		while (Chars.hasNext()) {
			
			if (Chars.next() == ',') {
				
				Chars.eatWhitespace();
				
				final CSSPattern		NewPat = new CSSPattern(this, Chars);
				
				ListEnd.NextPat = NewPat;
				ListEnd         = NewPat;
			}
		}
	}

/**	Recreates the string representation of this CSS2 pattern list.
 *
 * @return		The string representation of this CSS2 pattern list.
 */
	@Override public String toString() {
		final StringBuilder		Buff = new StringBuilder(64);
		
	//	Append the namespace "at rules".
		
		for (final Namespace Ns : Namespaces) {
			
//			if (Ns == Namespace.kXMLNamespace)
//				continue;	//	Skip the implicit XML namespace.
			
			Buff.append("@namespace ");
			
			if (Ns.isDefault() == false)
				Buff.append(Ns.getName()).append(' ');
			
			Buff.append("url(").append(Ns.getURI()).append("); ");
		}
		
	//	Append the patterns.
		
		CSSPattern				Pat = List;
		int						Index = 0;
		
		while (Pat != null) {
			
			if (Index++ > 0)
				Buff.append(", ");
			
			Buff.append(Pat.toString());
			
			Pat = Pat.NextPat;
		}
		
		return Buff.toString();
	}
	
	public Namespace getDefaultNamespace() {
		
		if (DefaultNamespace == null) {
			
			for (final Namespace Ns : Namespaces) {
				
				if (Ns.getName() == null) {
					
					DefaultNamespace = Ns;
					break;
				}
			}
			
			if (DefaultNamespace == null)
				DefaultNamespace = Namespace.kNoNamespace;
		}
		
		return DefaultNamespace;
	}
	
	public Namespace getNamespaceForPrefix
	(
		final String			NsPrefix	//	If null, we return the default Namespace, if any.
	)
	{
		if (NsPrefix == null)
			return getDefaultNamespace();
		
	//	The XML and XMLNS namespaces are always recognized in XML, regardless of whether they've been explicitly
	//	declared. Below, we test for references to those namespaces.
			
		if (Namespace.kXMLNamespace.getName().equals(NsPrefix))
			return Namespace.kXMLNamespace;
		
		if (Namespace.kXMLNSNamespace.getName().equals(NsPrefix))
			return Namespace.kXMLNSNamespace;
		
	//	Search the list of namespaces that have been declared (using "@namespace" rules) in these patterns.
		
		for (int Index = 0, Count = Namespaces.length; Index < Count; Index++) {
			final Namespace			Ns = Namespaces[Index];
			
			if (Ns.isDefault() == false && Ns.getName().equals(NsPrefix))
				return Ns;
		}
		
		return null;
	}
	
	private static Namespace[] readAtNamespaceRules
	(
		final CharacterIterator Chars
	)
	{
		final Namespace			FirstNs = readAtNamespaceRule(Chars);
		
		if (FirstNs == null)
			return kEmptyNamespaceArray;
		
		final Namespace			SecondNs = readAtNamespaceRule(Chars);
		
		if (SecondNs == null)
			return new Namespace[] { FirstNs };
		
		final ArrayList<Namespace>	Namespaces = new ArrayList<Namespace>();
		
		Namespaces.add(FirstNs);
		Namespaces.add(SecondNs);
		
		Namespace				NextNs = readAtNamespaceRule(Chars);
		
		while (NextNs != null) {
		
			Namespaces.add(NextNs);
			
			NextNs = readAtNamespaceRule(Chars);
		}
				
		return Namespaces.toArray(new Namespace[Namespaces.size()]);
	}
	
	private static Namespace readAtNamespaceRule
	(
		final CharacterIterator	Chars
	) 
	{
	//	For the description of "@namespace" rules see: <http://www.w3.org/TR/2003/WD-css3-syntax-20030813/#defining>
	//	The examples provided in that description are:
	//		@namespace url(http://www.w3.org/1999/xhtml);
	//		@namespace svg url(http://www.w3.org/2000/svg);
		
		Chars.eatWhitespace();
		
		if (Chars.nextCharIs('@') == false)
			return null;
		
		Chars.next();	//	Advance past the '@' sign.
		
		final String			AtRule = CSSParser.getIdentifier(Chars);
		
		if ("namespace".equals(AtRule) == false)
			throw new IllegalArgumentException("The only \"at rule\" supported is \"@namespace\", so \"@" + AtRule + "\" is not a valid option.");
		
		Chars.eatWhitespace();
		
		final String			NsPrefix;	//	Can be null;
		final String			NsURI;
		String					NextIdentifier = CSSParser.getIdentifier(Chars);
		
		if ("url".equals(NextIdentifier)) {
			
			NsPrefix = null;	//	Default namespace.
			
		} else {
			
			NsPrefix = NextIdentifier;
			
			Chars.eatWhitespace();
			NextIdentifier = CSSParser.getIdentifier(Chars);
			
			if ("url".equals(NextIdentifier) == false)
				throw new IllegalArgumentException("Was expecting the next identifier in the \"@namespace\" rule to be \"url\", but found \"" + NextIdentifier + "\" instead.");
		}
		
		NsURI = getFromBetween(Chars, kAtNamespaceURLBegin, kAtNamespaceURLEnd);
		
 		Chars.eatWhitespace();
		
		return new ImNamespace(NsPrefix, NsURI);
	}	
	
	private static String getFromBetween
	(
		final CharacterIterator	Chars,
		final char[]			BeginStr,
		final char[]			EndStr
	)
	{
		if (Chars.continuesWith(BeginStr)) {
			final StringBuilder		Buff = new StringBuilder(32);
			
			while (Chars.hasNext()) {
				
				if (Chars.continuesWith(EndStr))
					return Buff.toString();
				
				Buff.append(Chars.next());
			}
		}
		
		return null;
	}
	
/**	Searches the markup supplied by a <code>MarkupEnumeration</code> object for elements that match any of the 
 * 	patterns represented by this object.
 *
 * @param Enumerator			A <code>MarkupEnumeration</code> that will supply the markup to be searched.
 * @return						A <code>Results</code> object containing the search results.
 */
	public Results search
	(
		final MarkupEnumeration	Enumerator
	)
	{
		return search(Enumerator, null);
	}
	
/**	Searches the markup supplied by a <code>MarkupEnumeration</code> object for elements that match any of the 
 * 	patterns represented by this object.
 * 
 * @param Enumerator			A <code>MarkupEnumeration</code> that will supply the markup to be searched.
 * @param DestResults			The results of this search will be added to this <code>Results</code> object. If
 *								<code>null</code>, a new <code>Results</code> object is created.
 * @return						The <code>Results</code> object to which the search results were added.
 */
	public Results search
	(
		final MarkupEnumeration	Enumerator,
		Results					DestResults		//	Can be null. The results of this search are added to this Results object.
	) 
	{
		if (DestResults == null)
			DestResults = new Results();
		
		if (List == null)
			return DestResults;
		
		if (List.NextPat == null) {		//	If there's only one pattern...
			
			List.search(Enumerator, DestResults);
			
		} else {
			CSSPattern			Pat = List;
			
		//	Note that this loop is setup such that it will only be entered if there's more than one pattern in the
		//	list. This allows us to avoid cloning the markup enumeration in the commond, single-pattern case.
			
			while (Pat.NextPat != null) {
				
				Pat.search(Enumerator.clone(), DestResults);
				Pat = Pat.NextPat;
			}
			
			Pat.search(Enumerator, DestResults);
		}
		
		return DestResults;
	}
}




