/*	QWICAP (Quick Web Interface for Conventional Applications) by Chris W. Johnson
	Copyright (C) 2003-2010 The University of Texas at Austin ("U.T. Austin"). All rights reserved.
	
	U.T. Austin's copyright notice in this evolving version reflects our original authorship. It does not assert 
	copyright over changes contributed by others.
	
	This library is free software; you can redistribute it and/or modify it under the terms of the GNU Lesser General 
	Public License as published by the Free Software Foundation; either version 2.1 of the License, or (at your option) 
	any later version.
	
	This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied 
	warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for 
	more details.
	
	You should have received a copy of the GNU Lesser General Public License along with this library; if not, write to 
	the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/


package edu.utexas.its.eis.tools.qwicap.template.css;


import edu.utexas.its.eis.tools.qwicap.template.xml.immutable.ImmutableMarkup;
import edu.utexas.its.eis.tools.qwicap.template.xml.mutable.MutableMarkup;

import java.util.LinkedHashMap;
import java.util.Map;


/**	Used internally by {@link ImmutableMarkup} and {@link MutableMarkup} to cache "compiled" CSS patterns that 
 *	have been, or are about to be, used to search their markup. The idea is to avoid having to re-parse and re-compile 
 *	CSS patterns every time they are encountered, especially within loops. "Compiled" CSS patterns are represented by 
 *	{@link CSSPatterns} objects.
 *
 *	<p>This class is not part of the public API, but Java's access specifier scheme being what it is, it has to be
 *	marked as "public" in order to be accessible to both of the classes that use it.
 * 	</p>
 * 
 * 	@author Chris W. Johnson
 */
public final class CSSPatternsCache {
	
	private static final int	kPruneToLength = 100;
	
	private String				MRUPatsStr;
	private CSSPatterns			MRUPats;
	
	private final Object		CacheSynch = new Object();
	private MRUCache			Cache;
	
	
/**	Returns a {@link CSSPatterns} object implementing the pattern(s) described by the supplied CSS pattern string. The 
 *	<code>CSSPatterns</code> object is retrieved from, or created and added to, this cache as necessary.
 * 
 * @param PatsStr				A CSS patterns string.
 * @return						A <code>CSSPatterns</code> object implementing the pattern described by the 
 *								<code>PatsStr</code> parameter. 
 */
	public CSSPatterns getCSSPatterns
	(
		final String			PatsStr
	)
	{
		if (PatsStr.equals(MRUPatsStr) == false) {
			CSSPatterns				Pats = null;
			
			if (Cache != null) {
				
				Pats = Cache.get(PatsStr);
			
			} else {
				
				synchronized (CacheSynch) {
					Cache = new MRUCache(kPruneToLength);
				}
			}
			
			if (Pats == null) {
				
				Pats = new CSSPatterns(PatsStr);
				synchronized (CacheSynch) {
					Cache.put(PatsStr, Pats);
				}
			}
			
			MRUPatsStr = PatsStr;
			MRUPats    = Pats;
		}
		
		return MRUPats;
	}
	
	private static final class MRUCache extends LinkedHashMap<String, CSSPatterns> {
		
		private final int			MaxSize;
		
		
		MRUCache
		(
			final int				MaxSize
		) 
		{
			super((int) (MaxSize * 4.0/3.0 + 1), 0.75f, true);
			this.MaxSize = MaxSize;
		}
		
		@Override protected boolean removeEldestEntry
		(
			final Map.Entry<String, CSSPatterns>	Eldest
		) 
		{
			return size() > MaxSize;
		}
	}
}
