/*	QWICAP (Quick Web Interface for Conventional Applications) by Chris W. Johnson
	Copyright (C) 2003-2010 The University of Texas at Austin ("U.T. Austin"). All rights reserved.
	
	U.T. Austin's copyright notice in this evolving version reflects our original authorship. It does not assert 
	copyright over changes contributed by others.
	
	This library is free software; you can redistribute it and/or modify it under the terms of the GNU Lesser General 
	Public License as published by the Free Software Foundation; either version 2.1 of the License, or (at your option) 
	any later version.
	
	This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied 
	warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for 
	more details.
	
	You should have received a copy of the GNU Lesser General Public License along with this library; if not, write to 
	the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/


package edu.utexas.its.eis.tools.qwicap.template.css;


import java.util.NoSuchElementException;


/**	<code>CharacterIterator</code> is a utility class that is instantiated exclusively by <code>CSSPatterns</code>
 *	and is subsequently used by various combinator and selector classes. It is not part of the public API.
 */
final class CharacterIterator {
	
	private static final char[]	EmptyCharArray = {};
	
	private final char[]		Chars;
	private final int			Limit;
	private final int			StartIndex;
	private int					Index;
	private boolean				HasNext;
	
	
	CharacterIterator
	(
		final String			Str
	)
	{
		this(Str != null ? Str.toCharArray() : EmptyCharArray, 0, Str != null ? Str.length() : 0);
	}
	
	CharacterIterator
	(
		final char[]			Chars,
		final int				Index,
		final int				Limit
	)
	{
		this.Chars = Chars;
		this.Limit = Limit;
		this.StartIndex = Index;
		this.Index = Index;
		this.HasNext = hasAnotherCharacter();
	}
	
	CharacterIterator eatWhitespace() {
		final char[]			Chars = this.Chars;
		final int				Limit = this.Limit;
		int						Index = this.Index;
		
		while (Index < Limit) {
			
			if (Character.isWhitespace(Chars[Index]) == false)
				break;
			
			Index++;
		}
		
		this.Index = Index;
		
		HasNext = hasAnotherCharacter();
		
		return this;
	}
	
	private boolean hasAnotherCharacter() {
		return Index >= StartIndex && Index < Limit;
	}
	
	char peek() {
		
		if (!HasNext)
			throw new NoSuchElementException();
		
		return Chars[Index];
	}
	
	boolean hasNext() {
		return HasNext;
	}
	
	char next() {
		
		if (!HasNext)
			throw new NoSuchElementException();
		
		final char				Char = Chars[Index++];
		
		HasNext = hasAnotherCharacter();
		
		return Char;
	}
	
	CharacterIterator goBack() {
		
		Index--;
		HasNext = hasAnotherCharacter();
		
		return this;
	}
	
	@Override public String toString() {
		
		if (Index >= 0 && Index < Limit)
			return "character '" + Chars[Index - 1] + "' at index " + (Index - 1) + " of sequence \"" + toOriginalString() + '\"';
		else
			return "character at index " + (Index - 1) + " of sequence \"" + toOriginalString() + '\"';
	}
	
	String toOriginalString() {
		return new String(Chars, StartIndex, Limit - StartIndex);
	}
	
	String toRemainingString() {
		
		if (Index >= 0 && Index < Limit)
			return new String(Chars, Index, Limit - Index);
		else
			return "";
	}
	
	boolean continuesWith
	(
		final String			Substring
	)
	{
		boolean					ContinuesWith = false;
		final int				SubLength = Substring.length();
		final char[]			Chars = this.Chars;
		
	//	Are there enough characters remaining in Chars to permit a match with the substring?
		
		if (Limit - Index >= SubLength) {
			int						CIndex = Index;
			
		//	Does the first character match?
			
			if (Chars[CIndex++] == Substring.charAt(0)) {
				
			//	The first character matched. 
				
				ContinuesWith = true;
				
			//	Are there more characters in the substring?
				
				if (SubLength > 1) {
				
				//	There is more than one character in Substring, so compare the remainder of Substring to the
				//	remaining sequence of characters in Chars.
					
					final char[]			Subchars = Substring.toCharArray();
					
					for (int SIndex = 1; SIndex < SubLength; CIndex++, SIndex++) {
						
						if (Chars[CIndex] != Subchars[SIndex]) {
							
							ContinuesWith = false;
							break;
						}
					}	
				}
			}
			
			if (ContinuesWith) {
				
				Index = CIndex;
				HasNext = hasAnotherCharacter();
			}
		}
		
		return ContinuesWith;
	}
	
	boolean continuesWith
	(
		final char[]			Substring
	)
	{
		boolean					ContinuesWith = false;
		final int				SubLength = Substring.length;
		
	//	Are there enough characters remaining in Chars to permit a match with the substring?
		
		if (Limit - Index >= SubLength) {
			int						CIndex = Index;
			final char[]			Chars = this.Chars;
			
			ContinuesWith = true;
			
			for (int SIndex = 0; SIndex < SubLength; CIndex++, SIndex++) {
				
				if (Chars[CIndex] != Substring[SIndex]) {
					
					ContinuesWith = false;
					break;
				}
			}	
			
			if (ContinuesWith) {
				
				Index = CIndex;
				HasNext = hasAnotherCharacter();
			}
		}
		
		return ContinuesWith;
	}
	
	boolean nextCharIs
	(
		final char				TestChar
	)
	{
		return hasNext() && Chars[Index] == TestChar;
	}
}
