/*	QWICAP (Quick Web Interface for Conventional Applications) by Chris W. Johnson
	Copyright (C) 2003-2011 The University of Texas at Austin ("U.T. Austin"). All rights reserved.
	
	U.T. Austin's copyright notice in this evolving version reflects our original authorship. It does not assert 
	copyright over changes contributed by others.
	
	This library is free software; you can redistribute it and/or modify it under the terms of the GNU Lesser General 
	Public License as published by the Free Software Foundation; either version 2.1 of the License, or (at your option) 
	any later version.
	
	This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied 
	warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for 
	more details.
	
	You should have received a copy of the GNU Lesser General Public License along with this library; if not, write to 
	the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/


//  Results.java
//  qwicap
//
//  Created by Chris W. Johnson on Wed Jun 18 2003.


package edu.utexas.its.eis.tools.qwicap.template.css;


import edu.utexas.its.eis.tools.qwicap.servlet.*;
import edu.utexas.its.eis.tools.qwicap.template.xml.Markup;
import edu.utexas.its.eis.tools.qwicap.template.xml.Match;
import edu.utexas.its.eis.tools.qwicap.template.xml.TagException;
import edu.utexas.its.eis.tools.qwicap.template.xml.immutable.ImmutableMarkup;
import edu.utexas.its.eis.tools.qwicap.template.xml.mutable.MutableMarkup;
import edu.utexas.its.eis.tools.qwicap.util.ArrayToString;

import java.math.BigDecimal;
import java.math.BigInteger;
import java.text.ParseException;
import java.util.Iterator;
import java.util.logging.Level;
import java.util.logging.Logger;


/**	The <code>Results</code> class exists to store references to the XML markup elements selected by
 * 	<a href="http://www.w3.org/TR/REC-CSS2/selector.html#q1">CSS2 pattern</a> searches, as well as references to the 
 *	markup elements affected by the various operations that modify XML markup, such as the insert, delete, and 
 *	duplicate operations which <code>Results</code> supplies. These references are referred to as "matches" and are
 * 	represented by instances of the {@link Match} class. A <code>Results</code> object may contain any number of 
 *	matches, including none. Every operation that can be performed on XML markup has a corresponding method in the 
 *	<code>Results</code> class. Whenever such a method is invoked, the operation is individually applied to every 
 *	match in that <code>Results</code> object. In cases where an operation can't logically be applied to some matches, 
 *	the operation fails silenty and leaves untouched the markup referenced by those matches. Almost all operations 
 *	return a new <code>Results</code> object containing matches relevant to the operation that was performed, so method 
 *	chaining is not only possible, but very powerful.
 *	
 *	<p>Search Oriented Methods:</p>
 *	<ul>
 *		<li>{@link #get(String)}</li>
 *		<li>{@link #get(CSSPatterns)}</li>
 *	</ul>
 *
 *	<p>XML Element Oriented Methods:</p>
 *	<ul>
 *		<li>{@link #add(boolean, Object)}</li>
 *		<li>{@link #addAfter(Object)}</li>
 *		<li>{@link #addBefore(Object)}</li>
 *		<li>{@link #copy}</li>
 *		<li>{@link #cut}</li>
 *		<li>{@link #delete}</li>
 *		<li>{@link #duplicate}</li>
 *		<li>{@link #duplicateAfter}</li>
 *		<li>{@link #duplicateBefore}</li>
 *		<li>{@link #extract}</li>
 *		<li>{@link #extractToMarkup}</li>
 *		<li>{@link #getAncestors}</li>
 *		<li>{@link #getCDATA}</li>
 *		<li>{@link #getComments}</li>
 *		<li>{@link #getDeclarations}</li>
 *		<li>{@link #getParent}</li>
 *		<li>{@link #getType}</li>
 *		<li>{@link #replace}</li>
 *		<li>{@link #toForm}</li>
 *		<li>{@link #toMarkup}</li>
 *	</ul>
 *
 *	<p>XML Element Content Oriented Methods:</p>
 *	<ul>
 *		<li>{@link #addContent(boolean, Object)}</li>
 *		<li>{@link #addToStartOfContent(Object)}</li>
 *		<li>{@link #addToEndOfContent(Object)}</li>
 *		<li>{@link #deleteContent}</li>
 * 		<li>{@link #getBigDecimal}</li>
 * 		<li>{@link #getBigInteger}</li>
 * 		<li>{@link #getBoolean}</li>
 *		<li>{@link #getByte}</li>
 *		<li>{@link #getContent}</li>
 *		<li>{@link #getDouble}</li>
 *		<li>{@link #getFloat}</li>
 *		<li>{@link #getInt}</li>
 *		<li>{@link #getLong}</li>
 *		<li>{@link #getShort}</li>
 *		<li>{@link #getText}</li>
 *		<li>{@link #setContent(Object)}</li>
 *		<li>{@link #setContents(Object...)}</li>
 *	</ul>
 *
 *	<p>XML Element Attribute Oriented Methods:</p>
 *	<ul>
 *		<li>{@link #addAttribute(String, Object)}</li>
 *		<li>{@link #deleteAttribute(String)}</li>
 * 		<li>{@link #getAttribute(String)}</li>
 *		<li>{@link #setAttribute(String, Object)}</li>
 *		<li>{@link #setAttributes(Object...)}</li>
 *		<li>{@link #updateAttribute(String, Object)}</li>
 *	</ul>
 *	
 *	<p>XML Element "class" Attribute Oriented Methods:</p>
 *	<ul>
 *		<li>{@link #addClass(String)}</li>
 *		<li>{@link #deleteClass(String)}</li>
 *		<li>{@link #setClass(String)}</li>
 *	</ul>
 *
 *	<p>Filtering Operations on Results Objects:</p>
 *	<ul>
 *		<li>{@link #distinct}</li>
 *		<li>{@link #even}</li>
 * 		<li>{@link #everyNth}</li>
 *		<li>{@link #first}</li>
 *		<li>{@link #last}</li>
 *		<li>{@link #nth}</li>
 *		<li>{@link #odd}</li>
 *	</ul>
 *
 *	<p>Logical Operations on Pairs of Results Objects:</p>
 *	<ul>
 *		<li>{@link #and()}</li>
 *		<li>{@link #and(Results)}</li>
 *		<li>{@link #or()}</li>
 *		<li>{@link #or(Results)}</li>
 *		<li>{@link #xor()}</li>
 *		<li>{@link #xor(Results)}</li>
 *	</ul>
 *
 *	<p>Method Invocation Chain Manipulation Methods:</p>
 *	<ul>
 *		<li>{@link #IF}</li>
 *		<li>{@link #IF_NOT_EMPTY}</li>
 *		<li>{@link #ELSE}</li>
 *		<li>{@link #ENDIF}</li>
 *		<li>{@link #pop}</li>
 *		<li>{@link #pop(int)}</li>
 *	</ul>
 *	
 *	<p>Building, Exploring, and Clearing Results Objects:</p>
 *	<ul>
 *		<li>{@link #addMatch(Match)}</li>
 *		<li>{@link #addResults(Results)}</li>
 *		<li>{@link #addToResults(Results)}</li>
 *		<li>{@link #clear}</li>
 *		<li>{@link #get(int)}</li>
 *		<li>{@link #isEmpty}</li>
 *		<li>{@link #iterator}</li>
 *		<li>{@link #notEmpty}</li>
 *		<li>{@link #nth(int)}</li>
 *		<li>{@link #size}</li>
 *		<li>{@link #toString}</li>
 *	</ul>
 *	
 *	@author Chris W. Johnson
 */
public final class Results implements Iterable<Match> {
	
	private static final Logger	Log = Logger.getLogger(Results.class.getName());
	
	private enum ResultsType { 
		Normal, 
		IfTrue, 
		IfFalse, 
		ElseTrue, 	//	The IF tested true, so the ELSE does nothing.
		ElseFalse	//	The IF tested false, so the ELSE does something.
	}

	private static final int	kInitialMatchesLength = 16;	//	The initial length of the Matches array.

	private final Results		PrevResults;

	private Match[]				Matches;
	private int					MatchesLength;	//	The length of the Matches array. Used to avoid calling Matches.length all of the time.
	private int					MatchesIndex;	//	The index of the next unused element in the Matches array.	

	private final ResultsType	Type;
	
	private static Boolean		CanRequestI18NServices;
	private boolean				i18nServicesRequested;	//	This must NOT be static, because it describes the state of the i18n instance variable.
	private i18nServices		i18n;	//	This must NOT be static - i18nServices instances will vary depending on a user's language preferences.


/**	Instantiates an empty <code>Results</code> object.
 */
	public Results() {
		this(null, ResultsType.Normal, null);
	}

/**	Instantiates an empty <code>Results</code> object, which includes a reference to another "previous" 
 * 	<code>Results</code> object from which this <code>Results</code> object was, presumably, derived. The "previous"
 * 	<code>Results</code> object can be explicitly accessed using the {@link #pop} method, and is implicitly utilized 
 * 	by various unitary operations like {@link #and()}, {@link #or()}, and {@link #xor()}.
 * 
 * @param PrevResults			A "previous" <code>Results</code> object from which this <code>Results</code> object
 * 								was, presumably, derived.
 */
	public Results
	(
		final Results			PrevResults
	)
	{
		this(PrevResults, ResultsType.Normal, null);
	}

	private Results
	(
		final Results			PrevResults,
		final ResultsType		Type,
		final Results			SourceOfMatches		//	Can be null
	)
	{
		this.PrevResults = PrevResults;
		this.Type        = Type;

		if (SourceOfMatches != null) {

			Matches       = SourceOfMatches.Matches;
			MatchesLength = SourceOfMatches.MatchesLength;
			MatchesIndex  = SourceOfMatches.MatchesIndex;

		} else {

			Matches       = new Match[kInitialMatchesLength];
			MatchesLength = kInitialMatchesLength;
			MatchesIndex  = 0;
		}
	}

/**	Returns an <code>Iterator</code> of the matches stored in this <code>Results</code> object.
 *
 * @return			An <code>Iterator</code> of the matches stored in this <code>Results</code> object.
 */
	@Override public Iterator<Match> iterator() {
		return new ResultsEnumerator(this);
	}
	
/**	Returns an <code>Enumeration</code> of the matches stored in this <code>Results</code> object.
 *
 * @return			An <code>Enumeration</code> of the matches stored in this <code>Results</code> object.
 * @deprecated		Use {@link #iterator} instead.
 */
	@Deprecated public ResultsEnumerator enumerate() {
		return new ResultsEnumerator(this);
	}

/**	Returns the <code>Results</code> object, if any, from which this instance was produced.
 *
 * @return	The <code>Results</code> object from which this instance was produced, or <code>null</code>, if there was no 
 *			previous <code>Results</code> object.
 */
	public Results pop() {
		return PrevResults;
	}

/**	Returns the N<sup>th</sup> <code>Results</code> object in this chain of <code>Results</code> objects; in effect 
 *	this invokes the <code>pop</code> method N times and returns the result. 
 * 
 * @param PopCount	The number of <code>pop</code>s to perform.
 * @return			The N<sup>th</sup> <code>Results</code> object in this chain, or <code>null</code> if there weren't 
 *					N <code>Results</code> objects in the chain.
 */
	public Results pop
	(
		final int				PopCount
	)
	{
		Results					NthRes = this;

		for (int Index = 0; Index < PopCount && NthRes != null; Index++)
			NthRes = NthRes.PrevResults;

		return NthRes;
	}

/**	Returns the number of matches in this <code>Results</code> object.
 *
 * @return	The number of matches in this <code>Results</code> object.
 */
	public int size() {
		return MatchesIndex;
	}

/**	Removes all matches from this <code>Results</code> object. (The size method will subsequently return zero.)
 *
 * @return	This <code>Results</code> object, with its contents removed.
 */
	public Results clear() {
		MatchesIndex = 0;
		return this;
	}

/**	Returns <code>true</code> if there are no matches in this <code>Results</code> object.
 *
 * @return	<code>true</code> if empty, <code>false</code> otherwise.
 */
	public boolean isEmpty() {
		return MatchesIndex == 0;
	}

/**	Returns <code>true</code> if there is one, or more, matches in this <code>Results</code> object.
 * 
 * @return	<code>true</code> if one or more matches, <code>false</code> if empty.
 */
	public boolean notEmpty() {
		return !isEmpty();
	}

/**	Appends a <code>Match</code> to this <code>Results</code> object.
 *
 * @param AMatch	The <code>Match</code> to be appended to this list.
 * @return			This <code>Results</code> object.
 */
	public Results addMatch
	(
		final Match				AMatch
	)
	{
		if (AMatch != null) {

			if (MatchesIndex >= MatchesLength)
				expandMatchesArray(1);

			Matches[MatchesIndex++] = AMatch;
		}

		return this;
	}

/**	Appends a <code>Match</code> to this <code>Results</code> object.
 *
 * @param AMatch	The <code>Match</code> to be appended to this list.
 * @return			This <code>Results</code> object.
 * @deprecated		Use {@link #addMatch addMatch} instead.
 */
	@Deprecated public Results add
	(
		final Match				AMatch
	)
	{
		return addMatch(AMatch);
	}

/**	Appends all of the matches in another <code>Results</code> object to this object.
 *
 * @param SomeResults	The <code>Results</code> object whose contents should be appended to the contents of this object.
 * @return				This <code>Results</code> object.
 */
	public Results addResults
	(
		final Results			SomeResults
	)
	{
		expandMatchesArray(SomeResults.MatchesIndex);

		System.arraycopy(SomeResults.Matches, 0, Matches, MatchesIndex, SomeResults.MatchesIndex);
		MatchesIndex += SomeResults.MatchesIndex;

		return this;
	}

/**	Appends all of the matches in another <code>Results</code> object to this object.
 *
 * @param SomeResults	The <code>Results</code> object whose contents should be appended to the contents of this object.
 * @return				This <code>Results</code> object.
 * @deprecated			Use {@link #addResults addResults} instead.
 */
	@Deprecated public Results add
	(
		final Results			SomeResults
	)
	{
		return addResults(SomeResults);
	}

	private void expandMatchesArray
	(
		final int				ExpandByCount
	)
	{
		final int				MinimumLength = MatchesIndex + ExpandByCount;

		if (MinimumLength > MatchesLength) {
			int						NewLength = (MatchesLength * 3) / 2 + 1;

			if (NewLength < MinimumLength)
				NewLength = MinimumLength;

			final Match[]			NewArray = new Match[NewLength];

			System.arraycopy(Matches, 0, NewArray, 0, MatchesIndex);

			Matches = NewArray;
			MatchesLength = NewLength;
		}
	}

	private boolean contains
	(
		final Match				M
	)
	{
		for (int Index = 0, Count = MatchesIndex; Index < Count; Index++) {

			if (Matches[Index].equals(M))
				return true;
		}

		return false;
	}

/**	Appends all of the matches in this <code>Results</code> object to another <code>Results</code> object. This is the 
 *	mirror image of {@link #addMatch addMatch}.
 *
 * @param SomeResults	The <code>Results</code> object to which this object's contents should be appended.
 * @return				This <code>Results</code> object.
 */
	public Results addToResults
	(
		final Results			SomeResults
	)
	{
		SomeResults.addResults(this);
		return this;
	}

/**	Appends all of the matches in this <code>Results</code> object to another <code>Results</code> object. This is the 
 *	mirror image of {@link #addMatch addMatch}.
 *
 * @param SomeResults	The <code>Results</code> object to which this object's contents should be appended.
 * @return				This <code>Results</code> object.
 * @deprecated			Use {@link #addToResults addToResults} instead.
 */
	@Deprecated public Results addTo
	(
		final Results			SomeResults
	)
	{
		return addToResults(SomeResults);
	}

/**	Creates a new <code>Results</code> object containing all of the matches that are present in either this 
 *	<code>Results</code> object, or the previous <code>Results</code> object (as returned by {@link #pop}). No 
 *	<code>Match</code> will appear more than once in the new <code>Results</code> object.
 * 
 * @return				A new <code>Results</code> object containing all of the matches in this and the previous 
 *						<code>Results</code> objects.
 */
	public Results or() {
		return or(pop(), this, new Results(this));
	}

/**	Creates a new <code>Results</code> object containing all of the matches that are present in either this 
 *	<code>Results</code> object, or the supplied <code>Results</code> object, "B". No <code>Match</code> will appear more 
 *	than once in the new <code>Results</code> object.
 * 
 * @param B				The other <code>Results</code> object.
 * @return				A new <code>Results</code> object containing all of the matches in this <code>Results</code>
 *	 					object and "B".
 */
	public Results or
	(
		final Results			B
	)
	{
		return or(this, B, new Results(this));
	}

/**	Adds to the <code>Results</code> object "Out" all of the matches that are present in either <code>Results</code>
 *  objects "A" or "B". No <code>Match</code> will appear more than once in the new <code>Results</code> object.
 * 
 * @param A				One <code>Results</code> object.
 * @param B				Another <code>Results</code> object.
 * @param Out			An empty <code>Results</code> object into which the results of this operation will be placed.
 * @return				Returns "Out", containing all of the matches in <code>Results</code> objects "A" and "B".
 */
	private static Results or
	(
		final Results			A,
		final Results			B,
		final Results			Out
	)
	{
	//	Add all of A to Out. We assume, in fact we demand, that Out is empty.

		Out.addResults(A);

	//	Add anything in B that is not in A to Out.

		final Match[]			BMatches = B.Matches;	//	Using a local variable should be a little more efficient.

		for (int Index = 0, Count = B.MatchesIndex; Index < Count; Index++) {
			final Match				M = BMatches[Index];

			if (A.contains(M) == false)
				Out.addMatch(M);
		}

		return Out;
	}

/**	Creates a new <code>Results</code> object containing all of the matches that are common to this 
 *	<code>Results</code> object, and the previous <code>Results</code> object (as returned by {@link #pop}).
 * 
 * @return				A new <code>Results</code> object containing all of the matches common to this and the 
 *						previous <code>Results</code> objects.
 */
	public Results and() {
		return and(pop(), this, new Results(this));
	}

/**	Creates a new <code>Results</code> object containing all of the matches that are common to this 
 *	<code>Results</code> object, and the supplied <code>Results</code> object, "B".
 * 
 * @param B				The other <code>Results</code> object.
 * @return				A new <code>Results</code> object containing all of the matches common to this 
 *						<code>Results</code> object and "B".
 */
	public Results and
	(
		final Results			B
	)
	{
		return and(this, B, new Results(this));
	}

/** Adds to the <code>Results</code> object "Out" only the matches that are common to <code>Results</code> 
 *	objects "A" and "B".
 * 
 * @param A				One <code>Results</code> object.
 * @param B				Another <code>Results</code> object.
 * @param Out			The <code>Results</code> object into which the results of this operation will be placed.
 * @return				Returns "Out", containing all of the matches common to <code>Results</code> objects "A" and "B".
 */
	private static Results and
	(
		final Results			A,
		final Results			B,
		final Results			Out
	)
	{
		final Match[]			BMatches = B.Matches;

		for (int Index = 0, Count = B.MatchesIndex; Index < Count; Index++) {
			final Match				M = BMatches[Index];

			if (A.contains(M))
				Out.addMatch(M);
		}

		return Out;
	}

/**	Creates a new <code>Results</code> object containing all of the matches that are <b>not</b> common to this 
 *	<code>Results</code> object and the previous <code>Results</code> object (as returned by {@link #pop}).
 * 
 * @return				A new <code>Results</code> object containing all of the matches not common to this and the 
 *						previous <code>Results</code> objects.
 */
	public Results xor() {
		return xor(pop(), this, new Results(this));
	}

/**	Creates a new <code>Results</code> object containing all of the matches that are <b>not</b> common to this and the 
 *	supplied <code>Results</code> object, "B".
 *
 * @param B				The other <code>Results</code> object.
 * @return				A new <code>Results</code> object containing all of the matches not common to this and "B".
 */
	public Results xor
	(
		final Results			B	//	Subtract B from this results object.
	)
	{
		return xor(this, B, new Results(this));
	}

/**	Adds to the <code>Results</code> object "Out" only the matches that are <b>not</b> common to <code>Results</code> 
 *	objects "A" and "B".
 * 
 * @param A				One <code>Results</code> object.
 * @param B				Another <code>Results</code> object.
 * @param Out			The <code>Results</code> object into which the results of this operation will be placed.
 * @return				Returns "Out", containing all of the matches not common to <code>Results</code> objects "A" 
 * 						and "B".
 */
	private static Results xor
	(
		Results					A,
		Results					B,
		final Results			Out	//	The result is only assured to represent an XOR if Out is empty. 
	)
	{
	//	Make sure B contains the fewest matches. If not, swap A and B before proceeding.

		if (A.MatchesIndex < B.MatchesIndex) {
			final Results			Temp = A;

			A = B;
			B = Temp;
		}

	//	Keep a local reference to the Matches array of A, and a local COPY of B.

		final Match[]			AMatches = A.Matches;
		final Match[]			BMatches = new Match[B.MatchesIndex];

		System.arraycopy(B.Matches, 0, BMatches, 0, BMatches.length);

	//	Add every item in A that is NOT in B to the output, while clearing the elements in BMatches (our copy of the 
	// 	B.Matches array) that held the references to the common items.

		for (int AIndex = 0, ACount = A.MatchesIndex; AIndex < ACount; AIndex++) {
			final Match				AM = AMatches[AIndex];
			boolean					Common = false;

			for (int BIndex = 0, BCount = BMatches.length; BIndex < BCount; BIndex++) {

				if (AM.equals(BMatches[BIndex])) {

					BMatches[BIndex] = null;
					Common = true;

				//	Note that we can't exit the "B" loop at this point, because we can't be certain that any given
				//	Match occurs only once in the B array. Ordinarily, that would be the case, but you can't be 
				//	certain short of calling "distinct".
				}
			}

			if (!Common)
				Out.addMatch(AM);
		}

	//	Add to the output every element in B that has not been set to null. The elements that *have* been set to null 
	//	were found to be present in A. Since every element in A was tested against every element in B, there are no
	//	comparisons or searches remaining to be performed.

		for (int BIndex = 0, BCount = BMatches.length; BIndex < BCount; BIndex++) {
			final Match				BM = BMatches[BIndex];

			if (BM != null)
				Out.addMatch(BM);
		}

		return Out;
	}

/*
if TRUE:

	[fred].IF(true).[copy-of-fred(type=kResultsTypeIfTrue)][thing1].ELSE().[empty(type=kResultsTypeElseTrue)].[thing2].ENDIF().[fred]
	   ^                   |                                                  |
       |                   |                                                  |
	   +-------------------+                                                  |
	   +----------------------------------------------------------------------+
	

if FALSE:

	[fred].IF(false).[empty(type=kResultsTypeIfFalse)].[thing1].ELSE().[copy-of-fred(type=kResultsTypeElseFalse)].[thing2].ENDIF().[fred]
       ^                |                                                    |
       |                |                                                    |
       +----------------+                                                    |
       +---------------------------------------------------------------------+
*/

/**	Tests a boolean parameter: If it's <code>true</code>, a new <code>Results</code> object is returned that is the same
 *	as this <code>Results</code> object; if it's <code>false</code>, an empty <code>Results</code> object is returned.
 *	is returned. Note that nesting of <code>IF</code>, or <code>IF_NOT_EMPTY</code>, invocations is not supported.
 * 
 * @param B			<code>boolean</code> value to be tested.
 * @return			If the parameter was <code>true</code>, a <code>Results</code> object that is same as this object, 
 *					otherwise an empty <code>Results</code>.
 */
	public Results IF
	(
		final boolean			B
	)
	{
		if (B)
			return new Results(this, ResultsType.IfTrue,  this);	//	Return a new Results containing the same matches as this one.
		else
			return new Results(this, ResultsType.IfFalse, null);//	Return a new Results containing nothing.
	}

/**	If this <code>Results</code> object is not empty, a new <code>Results</code> object is returned that is the same as
 * 	this <code>Results</code> object. If this <code>Results</code> object is empty, an empty <code>Results</code> object 
 * 	is returned. Note that nesting of <code>IF</code>, or <code>IF_NOT_EMPTY</code>, invocations is not supported.
 *  
 * 	@return			If this <code>Results</code> is not empty, a <code>Results</code> object that is same as this 
 * 					object, otherwise an empty <code>Results</code>.
 */
	public Results IF_NOT_EMPTY() {
		return IF(notEmpty());
	}

/**	Returns an empty <code>Results</code> object if the preceding <code>IF</code> invocation was passed 
 *	<code>true</code>, and otherwise returns a <code>Results</code> object that is the same	as the <code>Results</code> 
 *	object on which the <code>IF</code> was invoked. 
 * 
 * @return			If the preceding <code>IF</code> tested <code>true</code>, an empty <code>Results</code> object,
 *					otherwise a <code>Results</code> object that is same as the <code>Results</code> object on which
 *					<code>IF</code> was invoked.
 */
	public Results ELSE() {

	//	Search backwards through the Results "stack" until we find a Results object produced by an IF invocation.

		Results					Prev = this;

		while (Prev != null) {
			final ResultsType		Type = Prev.Type;

			if (Type == ResultsType.IfTrue) {

				return new Results(Prev.pop(), ResultsType.ElseTrue, null);

			} else if (Type == ResultsType.IfFalse) {

				return new Results(Prev.pop(), ResultsType.ElseFalse, Prev.pop());
			}

			Prev = Prev.pop();
		}

		return new Results(this);
	}

/**	Returns the <code>Results</code> object on which <code>IF</code> was invoked.
 * 
 * @return			The <code>Results</code> object on which <code>IF</code> was invoked.
 */
	public Results ENDIF() {

	//	Search backwards through the Results "stack" until we find a Results object produced by an ELSE or IF invocation.

		Results					Prev = this;

		while (Prev != null) {

			if (Prev.Type != ResultsType.Normal)
				return Prev.pop();

			Prev = Prev.pop();
		}

		return new Results(this);
	}

/**	Returns the <code>Match</code> at the specified index in this <code>Results</code> object.
 *
 * @param Index								The index number of the desired <code>Match</code>.
 * @return									The <code>Match</code> in this <code>Results</code> object.
 * @throws ArrayIndexOutOfBoundsException	If Index is too large or small.
 * @see #nth
 * @see #everyNth
 * @see #size
 * @see #iterator
 */
	public Match get
	(
		final int				Index
	)
	{
		if (Index < 0 || Index >= MatchesIndex)
			throw new ArrayIndexOutOfBoundsException("This Results object contains " + MatchesIndex + " elements, therefore " + Index + " is not a valid index.");

		return Matches[Index];
	}

/**	Applies one or more <a href="http://www.w3.org/TR/REC-CSS2/selector.html#q1">CSS2 selector pattern</a> strings to
 * 	the subsections of XML referenced by the matches in this <code>Results</code> object.
 *
 * @param CSSPatternsStr	<a href="http://www.w3.org/TR/REC-CSS2/selector.html#q1">CSS2 selector pattern</a>(s).
 * @return					A <code>Results</code> object containing matches for each element selected by the CSS2 
 *							pattern.
 */
	public Results get
	(
		final String			CSSPatternsStr
	)
	{
		if (notEmpty())
			return get(getMarkup().getCSSPatterns(CSSPatternsStr));
		else
			return new Results(this);
	}
	
/**	Returns a new <code>Results</code> object listing all of the declaration (<samp>&lt;!</samp> &hellip; <samp>&gt;</samp>) 
 * 	elements in this markup.
 * 
 * @return					A <code>Results</code> object containing matches for declaration elements.
 */
	public Results getDeclarations() {
		final Results			Declarations = new Results(this);
		
		for (int Index = 0, Count = size(); Index < Count; Index++)
			Declarations.addResults(get(Index).getDeclarations());
		
		return Declarations;
	}

/**	Returns a new <code>Results</code> object listing all of the comment (<samp>&lt;!--</samp> &hellip; <samp>--&gt;</samp>) 
 * 	elements in this markup.
 * 
 * @return					A <code>Results</code> object containing matches for comment elements.
 */
	public Results getComments() {
		final Results			Comments = new Results(this);
		
		for (int Index = 0, Count = size(); Index < Count; Index++)
			Comments.addResults(get(Index).getComments());
		
		return Comments;
	}

/**	Returns a new <code>Results</code> object listing all of the CDATA (<samp>&lt;![CDATA[</samp> &hellip; <samp>]]&gt;</samp>) 
 * 	elements in the subsections of XML referenced by the matches in this <code>Results</code> object.
 * 
 * @return					A <code>Results</code> object containing matches for CDATA sections.
 */
	public Results getCDATA() {
		final Results			CDATA = new Results(this);

		for (int Index = 0, Count = size(); Index < Count; Index++)
			CDATA.addResults(get(Index).getCDATA());
		
		return CDATA;
	}

/**	Returns a reference to the <code>Markup</code> of the which the first <code>Match</code> in this 
 *	<code>Results</code> is a member. If this <code>Results</code> is empty, this method can only return 
 *	<code>null</code>. Even when this <code>Results</code> is not empty, the result of this method is not entirely
 * 	what you might expect, because it is possible (though not normal), for a <code>Results</code> object to contain 
 *	references to elements in multiple <code>Markup</code> objects. 
 * 
 * @return					The <code>Markup</code> in which the first <code>Match</code> in this <code>Results</code>
 *							object resides, or <code>null</code> if this <code>Results</code> object is empty.
 */
	private Markup getMarkup() {

		if (size() > 0)
			return get(0).getMarkup();

		return null;
	}

/**	Applies one or more <a href="http://www.w3.org/TR/REC-CSS2/selector.html#q1">CSS2 selector patterns</a> to the
 * 	subsections of XML referenced by the matches in this <code>Results</code> object.
 *
 * @param Patterns	<a href="http://www.w3.org/TR/REC-CSS2/selector.html#q1">CSS2 selector pattern</a>(s) represented
 * 					in their "compiled" form by a <code>CSSPatterns</code> object.
 * @return			A <code>Results</code> object containing matches for each element selected by the CSS2 pattern.
 */
	public Results get
	(
		final CSSPatterns		Patterns
	)
	{
		final Results			NewResults = new Results(this);

		for (int Index = 0, Count = size(); Index < Count; Index++)
			Patterns.search(get(Index).enumerate(), NewResults);

		return NewResults;
	}

/**	Returns a complete list of the ancestors of each element referenced by the matches in this <code>Results</code>
 *	object. This is generally useful only when there is a single match in this <code>Results</code> object.
 * 
 * @return	A new <code>Results</code> object containing a list of the ancestors of each element referenced by the 
 *			matches in this <code>Results</code> object. If there is only one element referenced by this 
 *			<code>Results</code> object, the first item in the <code>Results</code> object that is returned will be 
 *			the immediate parent of the element, and the last element will be the most distant ancestor. If there is 
 *			more than one element referenced by this <code>Results</code> object, this method probably isn't useful.
 */
	public Results getAncestors() {
		final Results			NewResults = new Results(this);

		for (int Index = 0, Count = size(); Index < Count; Index++)
			NewResults.addResults(get(Index).getAncestors());

		return NewResults;
	}

/**	Returns the parent element (if any) of each element referenced by the matches in this <code>Results</code> object. 
 * 
 * @return	A new <code>Results</code> object containing a reference to the parent element of each element referenced 
 *			in this <code>Results</code> object.
 */
	public Results getParent() {
		final Results			NewResults = new Results(this);

		for (int Index = 0, Count = size(); Index < Count; Index++)
			NewResults.addMatch(get(Index).getParent());

		return NewResults;
	}

/**	Returns a <code>Form</code> object for examining and/or manipulating the XHTML "<samp>form</samp>" element (and the 
 * 	controls within it) that is referenced by the first match in this <code>Results</code> object. If you intend to 
 * 	modify the "<samp>form</samp>" element, or its contents, you should already be operating on {@link MutableMarkup}
 * 	before invoking <code>toForm</code>. A mutable version of <code>ImmutableMarkup</code> is created trivially by
 * 	merely invoking the {@link Markup#getMutable()} method.
 * 
 * @return						A <code>Form</code> object for manipulating the "<samp>form</samp>" element 
 *								referenced by the first match in this <code>Results</code> object, or <code>null</code>
 * 								if the <code>Results</code> object is either empty or its first match does not reference
 * 								a "<samp>form</samp>" element.
 * @see MutableMarkup#getForm(FormDataSet)
 * @deprecated Use {@link #toForm}, instead.
 */
	@Deprecated public Form getForm() {
	
		if (notEmpty()) {
			
			try {
				return new Form(get(0));
			} catch (FormNotFoundException e) {
			//	Ignore.
			}
		}
		
		return null;
	}

/**	Returns a <code>Form</code> object for examining and/or manipulating the first XHTML "<samp>form</samp>" element 
 * 	(and the controls within it) that is found in this <code>Results</code> object. If you intend to 
 * 	modify the "<samp>form</samp>" element, or its contents, you should already be operating on {@link MutableMarkup}
 * 	before invoking <code>toForm</code>. A mutable version of <code>ImmutableMarkup</code> is created trivially by
 * 	merely invoking the {@link Markup#getMutable()} method.
 * 
 * @return							A <code>Form</code> object for manipulating the first "<samp>form</samp>" element 
 *									found in this <code>Results</code> object.
 * @throws FormNotFoundException	If this <code>Results</code> object contains no "<samp>form</samp>" element.
 * @see MutableMarkup#getForm(FormDataSet)
 * @see Match#toForm
 */
	public Form toForm() throws FormNotFoundException {

		if (isEmpty())
			throw new FormNotFoundException("This Results object is empty, therefore there was no \"form\" element on which to operate.");
		
		for (final Match Elem : this) {
			
			if ("form".equals(Elem.getType()))
				return new Form(Elem);
		}
		
	//	No "form" elements were found. Create an informative FormNotFoundException exception.
		
		final String[]			ElemTypes = new String[size()];
		
		for (int Index = 0; Index < ElemTypes.length; Index++)
			ElemTypes[Index] = get(Index).getType();
		
		if (ElemTypes.length == 1)
			throw new FormNotFoundException("This Results object contains only one element, which is of type \"" + ElemTypes[0] + "\". There is no \"form\" element.");
		else
			throw new FormNotFoundException("This Results object contains elements of the types " + new ArrayToString(ElemTypes).alwaysQuote() + ", but no \"form\" element.");
	}

/**	Returns a new <code>Results</code> object which eliminates any redundant element references that may have been 
 *	present in this <code>Results</code> object.
 * 
 * @return	A new <code>Results</code> object in which all of the element references are unique.
 */
	public Results distinct() {
		final Results			NewResults = new Results(this);

		for (int OIndex = 0, OCount = size(); OIndex < OCount; OIndex++) {
			final Match				OMatch   = get(OIndex);
			boolean					Distinct = true;

			for (int NIndex = 0, NCount = NewResults.size(); NIndex < NCount; NIndex++) {
				final Match				NMatch = NewResults.get(NIndex);

				if (OMatch.equals(NMatch)) {

					Distinct = false;
					break;
				}
			}

			if (Distinct)
				NewResults.addMatch(OMatch);
		}

		return NewResults;
	}

/**	Returns the first <code>Match</code> in this <code>Results</code> object as the sole element of a new 
 *	<code>Results</code> object. If there were no matches in this <code>Results</code> object, the new 
 *	<code>Results</code> object is empty.
 *
 * @return	A new <code>Results</code> object containing only the first <code>Match</code> in this <code>Results</code> 
 *			object.
 * @see #last
 * @see #even
 * @see #odd
 */
	public Results first() {
		final Results			NewResults = new Results(this);

		if (MatchesIndex > 0)
			NewResults.addMatch(Matches[0]);

		return NewResults;
	}

/**	Returns the final <code>Match</code> in this <code>Results</code> object as the sole element of a new 
 *	<code>Results</code> object. If there were no matches in this <code>Results</code> object, the new 
 *	<code>Results</code> object is empty.
 *
 * @return	A new <code>Results</code> object containing only the final <code>Match</code> in this <code>Results</code> 
 *			object.
 * @see 	#first
 * @see 	#odd
 * @see 	#even
 */
	public Results last() {
		final Results			NewResults = new Results(this);

		if (MatchesIndex > 0)
			NewResults.addMatch(Matches[MatchesIndex - 1]);

		return NewResults;
	}

/**	Returns the <code>Match</code> at the specified index in this <code>Results</code> object (the "N<sup>th</sup>" 
 *	<code>Match</code>) as the sole element of a new <code>Results</code> object. If there is no such element in this 
 *	<code>Results</code> object, the new <code>Results</code> object is empty.
 * 
 * @param Index		The index number of the <code>Match</code> to be returned. (Greater than or equal to zero, less than 
 *					{@link #size}.)
 * @return			A new <code>Results</code> object containing the N<sup>th</sup> <code>Match</code> in this 
 *					<code>Results</code> object.
 * @see #get(int)
 * @see #size
 * @see #everyNth
 */
	public Results nth
	(
		final int				Index
	)
	{
		final Results			NewResults = new Results(this);

		if (Index >= 0 && Index < MatchesIndex)
			NewResults.addMatch(Matches[Index]);

		return NewResults;
	}

/**	Conceptually divides the matches in this <code>Results</code> object into groups of a certain size, then returns
 * 	a new <code>Results</code> object containing the Nth match from each group. For example, the {@link #even} 
 * 	convenience method is implemented as <code>everyNth(2, 0)</code> (return the match at index zero from every group 
 * 	of 2), and the {@link #odd} method is implemented as <code>everyNth(2, 1)</code> (return the match at index 1 from
 * 	every group of 2).
 * 
 * @param GroupSize				The size of the groups into which the matches should be conceptually divided.
 * @param GroupIndex			The index of the match within each group of matches that is to be returned.
 * @return						A new <code>Results</code> object containing the requested <code>Match</code> objects.
 */
	public Results everyNth
	(
		final int				GroupSize,
		final int				GroupIndex
	)
	{
		if (GroupIndex < 0 || GroupIndex >= GroupSize)
			throw new IllegalArgumentException("The everyNth(" + GroupSize + ", " + GroupIndex + ") invocation is invalid because the group index (the second parameter) must be >= zero, and < group size (the first parameter).");

		final Results			NewResults = new Results(this);

		for (int Index = GroupIndex, Count = size(); Index < Count; Index += GroupSize)
			NewResults.addMatch(get(Index));

		return NewResults;
	}

/**	Returns every even-numbered <code>Match</code> in this <code>Results</code> object, beginning with the first 
 *	<code>Match</code> (element 0). If there were no matches in this <code>Results</code> object, the new 
 *	<code>Results</code> object is empty. This method is useful for doing things like color-coding alternating rows of 
 *	a table.
 *
 * @return	A new <code>Results</code> object containing every even-numbered <code>Match</code> in this 
 *			<code>Results</code> object, beginning with the first <code>Match</code>.
 * @see		#odd
 * @see		#everyNth
 * @see		#first
 * @see		#last
 */
	public Results even() {
		return everyNth(2, 0);
	}

/**	Returns every odd-numbered <code>Match</code> in this <code>Results</code> object, beginning with the second 
 *	<code>Match</code> (element 1). If there were less than two matches in this <code>Results</code> object, the new 
 *	<code>Results</code> object is empty. This method is useful for doing things like color-coding alternating rows of 
 *	a table.
 *
 * @return	A new <code>Results</code> object containing every odd-numbered <code>Match</code> in this 
 *			<code>Results</code> object, beginning with the second <code>Match</code>.
 * @see		#even
 * @see		#everyNth
 * @see		#first
 * @see		#last
 */
	public Results odd() {
		return everyNth(2, 1);
	}

/**	Deletes the contents of each XML element referenced by the matches in this <code>Results</code> object. For 
 *	example, if there were a match for an "<samp>h1</samp>" element that looked like: 
 *	"<samp>&lt;h1&gt;War and Peace&lt;/h1&gt;</samp>", then calling <code>deleteContent</code>
 * 	would reduce that "h1" element to: "<samp>&lt;h1&gt;&lt;/h1&gt;</samp>". Elements that have no content are not 
 *	affected.
 *
 * @return					A new <code>Results</code> object containing all of the matches in this <code>Results</code>
 *							object.
 * @throws TagException		If a problem is encountered with the manipulation of the markup.
 */
	public Results deleteContent() throws TagException {
		final Results			NewResults = new Results(this);

		for (int Index = 0, Count = size(); Index < Count; Index++)
			NewResults.addMatch(get(Index).deleteContent());

		return NewResults;
	}

/**	Sets the content of each XML element referenced by the matches in this <code>Results</code> object. For example, 
 *	if there were a match for an "<samp>h1</samp>" element that looked like: "<samp>&lt;h1&gt;War and 
 *	Peace&lt;/h1&gt;</samp>", then calling <code>setContent("Crime and Punishment")</code> would change that 
 *	"<samp>h1</samp>" element to: "<samp>&lt;h1&gt;Crime and Punishment&lt;/h1&gt;</samp>". Elements that have no 
 *	content are not affected.
 *
 * @param Value				An object specifying the new content of the elements referenced by this <code>Results</code>
 * 							object. Any type of object is potentially suitable, see {@link MutableMarkup#insert MutableMarkup.insert} 
 * 							for details.
 * @return					A new <code>Results</code> object containing all of the matches in this <code>Results</code>
 *	 						object. If you would prefer a <code>Results</code> that references all of the content that
 * 							was just set, follow your invocation of this method with an invocation of 
 *							{@link #getContent getContent}.
 * @throws TagException		If a problem is encountered with the manipulation of the markup.
 * @see #getContent
 */
	public Results setContent
	(
		final Object			Value
	)
	throws TagException
	{
		final Results			NewResults = new Results(this);

		if (Value instanceof Results) {
			final Results			Contents = (Results) Value;

			for (int Index = 0, Count = size(); Index < Count; Index++) {
				final Match				Target = get(Index);

				Target.deleteContent();

				for (int CIndex = 0, CCount = Contents.MatchesIndex; CIndex < CCount; CIndex++)
					NewResults.addMatch(Target.addContent(false, Contents.Matches[CIndex]));
			}

		} else {

			for (int Index = 0, Count = size(); Index < Count; Index++)
				NewResults.addMatch(get(Index).setContent(Value));
		}

		return NewResults;
	}

/**	Searches the content of each XML element referenced by the matches in this <code>Results</code> object using one 
 *	or more <a href="http://www.w3.org/TR/REC-CSS2/selector.html#q1">CSS2 selector patterns</a>, represented as 
 *	<code>String</code> or a <code>CSSPatterns</code> object, then sets the content of the matching elements to the 
 *	supplied value. This is equivalent to repeatedly chaining the method invocations: {@link #get(String)} and 
 *	{@link #setContent(Object)}. The various <code>setContents</code> methods are handy for doing things like setting 
 * 	the contents of all of the cells in a table row in one go.
 *
 * @param SelectorsAndValues	An array of alternating CSS2 selector patterns, represented as <code>String</code> or 
 *								<code>CSSPatterns</code> objects, and objects representing the content that should 
 *								replace the existing content of the XML elements that match the preceding CSS2 
 *								selector pattern(s). There may be as many such selector/content pairs as desired.
 * @return						A new <code>Results</code> object containing matches for each element whose contents 
 *								were set. If you would prefer a <code>Results</code> that references all of the content 
 * 								that was just set, follow your invocation of this method with an invocation of 
 *								{@link #getContent getContent}.
 * @throws TagException			If a problem is encountered with the manipulation of the markup.
 * @see #getContent
 */
	public Results setContents
	(
		final Object...			SelectorsAndValues	//	Selector object followed by value object, followed by selector object, followed by value....
	)
	throws TagException
	{
		final int				SVLength = SelectorsAndValues.length;

		if (SVLength % 2 != 0)
			throw new IllegalArgumentException("The array passed to setContents is meant to be composed of selector/value pairs, and therefore should contain an even number of elements. The array passed in this case, however, contains an odd number of elements: " + SVLength + '.');

		final Results			NewResults = new Results(this);

		for (int Index = 0; Index < SVLength; Index += 2) {
			final Object			Selector = SelectorsAndValues[Index];
			final Object			Value = SelectorsAndValues[Index + 1];

			if (Selector instanceof String)
				get((String) Selector).setContent(Value).addToResults(NewResults);
			else if (Selector instanceof CSSPatterns)
				get((CSSPatterns) Selector).setContent(Value).addToResults(NewResults);
			else
				throw new IllegalArgumentException("The element at index " + Index + ", \"" + Selector + "\", must function as a selector, so it must be either a String or a CSSPatterns object. It is actually of type \"" + Selector.getClass().getName() + "\".");
		}

		return NewResults;
	}

/**	Returns the type of the first XML element in this <code>Results</code> objects.
 * 
 * @return				The type of the first XML element in this <code>Results</code> object, or an empty string if 
 *						none of the matches refer to XML elements.
 */
	public String getType() {

		for (int Index = 0, Count = size(); Index < Count; Index++) {
			final String			Type = get(Index).getType();

			if (Type.length() != 0)
				return Type;
		}

		return "";
	}

/**	Returns the human-readable/-visible text, if any, that the elements referenced by this <code>Results</code> object 
 *	would contribute to the final, rendered document. At present, this does not usually include the text contributed 
 *	by "<samp>form</samp>" elements, and CSS style information (for example "<samp>display: none;</samp>") is not taken 
 *	into account.
 * 
 * @return						The human-readable text of the elements referenced by this <code>Results</code> object, 
 * 								or an empty string if there was no such text.
 */
	public String getText() {
		final StringBuilder		Buff = new StringBuilder(size() * 128);

		for (int Index = 0, Count = size(); Index < Count; Index++) {

		//	When the buffer does not end with whitespace, append a space to the buffer before appending the next 
		//	hunk of text. This may be a feature, or a problem, depending on the context, unfortunately.

			if (Index > 0 && Buff.length() > 0 && Character.isWhitespace(Buff.charAt(Buff.length() - 1)) == false)
				Buff.append(' ');

			get(Index).getText(Buff);
		}

		return Buff.toString();
	}

/**	Invokes {@link #getText} and converts the <code>String</code> it returns into a	<code>float</code>, if possible.
 * 
 * @return							The numeric value represented by the text of the element, as a <code>float</code>.
 * @throws NumberFormatException	If the text of this element couldn't be interpreted as a floating-point number.
 * @see Float#parseFloat(String)
 */
	public float getFloat() throws NumberFormatException {
		final i18nServices		i18n = geti18nServices();
		
		if (i18n != null) {
			try {
				return i18n.parseDecimal(getText().trim()).floatValue();
			} catch (ParseException e) {
				throw new NumberFormatParseException(e);
			}
		}
	
		return Float.parseFloat(getText());			//	parseDouble ignores leading and trailing whitespace.
	}

/**	Invokes {@link #getText} and converts the <code>String</code> it returns into a	<code>double</code>, if possible.
 * 
 * @return							The numeric value represented by the text of the element, as a <code>double</code>.
 * @throws NumberFormatException	If the text of this element couldn't be interpreted as a floating-point number.
 * @see Double#parseDouble(String)
 */
	public double getDouble() throws NumberFormatException {
		final i18nServices		i18n = geti18nServices();
			
		if (i18n != null) {
			try {
				return i18n.parseDecimal(getText().trim()).doubleValue();
			} catch (ParseException e) {
				throw new NumberFormatParseException(e);
			}
		}
	
		return Double.parseDouble(getText());		//	parseDouble ignores leading and trailing whitespace.
	}

/**	Invokes {@link #getText} and converts the <code>String</code> it returns into a <code>BigDecimal</code>, if possible.
 * 
 * @return							The numeric value represented by the text of the element, as a <code>BigDecimal</code>.
 * @throws NumberFormatException	If the text of this element couldn't be interpreted as a <code>BigDecimal</code>.
 * @see BigDecimal#BigDecimal(String)
 */
	public BigDecimal getBigDecimal() throws NumberFormatException {
		final i18nServices		i18n = geti18nServices();
			
		if (i18n != null) {
			try {
				return i18n.parseDecimal(getText().trim());
			} catch (ParseException e) {
				throw new NumberFormatParseException(e);
			}
		}
	
		return new BigDecimal(getText().trim());
	}

/**	Invokes {@link #getText} and converts the <code>String</code> it returns into a <code>boolean</code>, if possible.
 * 
 * @return							<code>true</code> if the text of the element was "true" (case insensitive), 
 *									<code>false</code> otherwise.
 * @see Boolean#valueOf(String)
 */
	public boolean getBoolean() {
		return Boolean.valueOf(getText().trim());
	}

/**	Invokes {@link #getText} and converts the <code>String</code> it returns into a <code>byte</code>, if possible.
 * 
 * @return							The numeric value represented by the text of the element, as a <code>byte</code>.
 * @throws NumberFormatException	If the text of this element couldn't be interpreted as a <code>byte</code>.
 * @see Byte#parseByte(String)
 */
	public byte getByte() throws NumberFormatException {
		final i18nServices		i18n = geti18nServices();
			
		if (i18n != null) {
			try {
				return i18n.parseDecimal(getText().trim()).byteValue();
			} catch (ParseException e) {
				throw new NumberFormatParseException(e);
			}
		}
	
		return Byte.parseByte(getText().trim());	//	parseLong does NOT ignore leading or trailing whitespace.
	}

/**	Invokes {@link #getText} and converts the <code>String</code> it returns into a <code>short</code>, if possible.
 * 
 * @return							The numeric value represented by the text of the element, as a <code>short</code>.
 * @throws NumberFormatException	If the text of this element couldn't be interpreted as an <code>short</code>.
 * @see Short#parseShort(String)
 */
	public short getShort() throws NumberFormatException {
		final i18nServices		i18n = geti18nServices();
				
		if (i18n != null) {
			try {
				return i18n.parseDecimal(getText().trim()).shortValue();
			} catch (ParseException e) {
				throw new NumberFormatParseException(e);
			}
		}
	
		return Short.parseShort(getText().trim());	//	parseLong does NOT ignore leading or trailing whitespace.
	}

/**	Invokes {@link #getText} and converts the <code>String</code> it returns into an <code>int</code>, if possible.
 * 
 * @return							The numeric value represented by the text of the element, as an <code>int</code>.
 * @throws NumberFormatException	If the text of this element couldn't be interpreted as an <code>int</code>.
 * @see Integer#parseInt(String)
 */
	public int getInt() throws NumberFormatException {
		final i18nServices		i18n = geti18nServices();
				
		if (i18n != null) {
			try {
				return i18n.parseDecimal(getText().trim()).intValue();
			} catch (ParseException e) {
				throw new NumberFormatParseException(e);
			}
		}
	
		return Integer.parseInt(getText().trim());	//	parseLong does NOT ignore leading or trailing whitespace.
	}

/**	Invokes {@link #getText} and converts the <code>String</code> it returns into a <code>long</code>, if possible.
 * 
 * @return							The numeric value represented by the text of the element, as a <code>long</code>.
 * @throws NumberFormatException	If the text of this element couldn't be interpreted as a <code>long</code>.
 * @see Long#parseLong(String)
 */
	public long getLong() throws NumberFormatException {
		final i18nServices		i18n = geti18nServices();
				
		if (i18n != null) {
			try {
				return i18n.parseDecimal(getText().trim()).longValue();
			} catch (ParseException e) {
				throw new NumberFormatParseException(e);
			}
		}
	
		return Long.parseLong(getText().trim());	//	parseLong does NOT ignore leading or trailing whitespace.
	}

/**	Invokes {@link #getText} and converts the <code>String</code> it returns into a <code>BigInteger</code>, if possible.
 * 
 * @return							The numeric value represented by the text of the element, as a <code>BigInteger</code>.
 * @throws NumberFormatException	If the text of this element couldn't be interpreted as a <code>BigInteger</code>.
 * @see BigInteger#BigInteger(String)
 */
	public BigInteger getBigInteger() throws NumberFormatException {
		final i18nServices		i18n = geti18nServices();
				
		if (i18n != null) {
			try {
				return i18n.parseInteger(getText().trim());
			} catch (ParseException e) {
				throw new NumberFormatParseException(e);
			}
		}
		return new BigInteger(getText().trim());
	}

/**	Returns a new <code>Results</code> object that references every child of every element referenced by this 
 *	<code>Results</code> object. Note that only <em>children</em> are listed, not grandchildren and further 
 *	descendants.
 * 
 * @return	A new <code>Results</code> object referencing every child of every element referenced by this 
 *			<code>Results</code> object.
 * @see #setContent
 */
	public Results getContent() {
		final Results			NewResults = new Results(this);

		for (int Index = 0, Count = size(); Index < Count; Index++)
			NewResults.addResults(get(Index).getContent());

		return NewResults;
	}

/**	Adds new material at the start of the content of each XML element referenced by the matches in this 
 *	<code>Results</code> object.
 * 
 * @param AdditionalContent				Any type of object is potentially suitable, see 
 * 										{@link MutableMarkup#insert MutableMarkup.insert} for details.
 * @return								A new <code>Results</code> object containing references to the content that was 
 *										added to this markup.
 * @throws TagException					If a problem is encountered with the manipulation of the markup.
 * @see #addContent(boolean, Object)
 */
	public Results addToStartOfContent
	(
		final Object			AdditionalContent
	)
	throws TagException
	{
		return addContent(true, AdditionalContent);
	}

/**	Adds new material to the end of the content of each XML element referenced by the matches in this 
 *	<code>Results</code> object.
 * 
 * @param AdditionalContent				Any type of object is potentially suitable, see 
 * 										{@link MutableMarkup#insert MutableMarkup.insert} for details.
 * @return								A new <code>Results</code> object containing references to the content that was 
 *										added to this markup.
 * @throws TagException					If a problem is encountered with the manipulation of the markup.
 * @see #addContent(boolean, Object)
 */
	public Results addToEndOfContent
	(
		final Object			AdditionalContent
	)
	throws TagException
	{
		return addContent(false, AdditionalContent);
	}

/**	Prepends new material to the content of each XML element referenced by the matches in this <code>Results</code> 
 *	object.
 * 
 * @param AdditionalContent				Any type of object is potentially suitable, see 
 * 										{@link MutableMarkup#insert MutableMarkup.insert} for details.
 * @return								A new <code>Results</code> object containing references to the content that was 
 *										added to this markup.
 * @throws TagException					If a problem is encountered with the manipulation of the markup.
 * @deprecated							Use {@link #addToStartOfContent addToStartOfContent} instead.
 */
	@Deprecated public Results addContentAtStart
	(
		final Object			AdditionalContent
	)
	throws TagException
	{
		return addContent(true, AdditionalContent);
	}

/**	Appends new material to the content of each XML element referenced by the matches in this <code>Results</code> 
 *	object.
 * 
 * @param AdditionalContent				Any type of object is potentially suitable, see 
 * 										{@link MutableMarkup#insert MutableMarkup.insert} for details.
 * @return								A new <code>Results</code> object containing references to the content that was 
 *										added to this markup.
 * @throws TagException					If a problem is encountered with the manipulation of the markup.
 * @deprecated							Use {@link #addToEndOfContent addToEndOfContent} instead.
 */
	@Deprecated public Results addContentAtEnd
	(
		final Object			AdditionalContent
	)
	throws TagException
	{
		return addContent(false, AdditionalContent);
	}

/**	Prepends new material to the content of each XML element referenced by the matches in this <code>Results</code> 
 *	object.
 * 
 * @param AdditionalContent				Any type of object is potentially suitable, see 
 * 										{@link MutableMarkup#insert MutableMarkup.insert} for details.
 * @return								A new <code>Results</code> object containing references to the content that was 
 *										added to this markup.
 * @throws TagException					If a problem is encountered with the manipulation of the markup.
 * @deprecated							Use {@link #addToStartOfContent addToStartOfContent} instead.
 */
	@Deprecated public Results addContentBefore
	(
		final Object			AdditionalContent
	)
	throws TagException
	{
		return addContent(true, AdditionalContent);
	}

/**	Appends new material to the content of each XML element referenced by the matches in this <code>Results</code> 
 *	object.
 * 
 * @param AdditionalContent				Any type of object is potentially suitable, see 
 * 										{@link MutableMarkup#insert MutableMarkup.insert} for details.
 * @return								A new <code>Results</code> object containing references to the content that was 
 *										added to this markup.
 * @throws TagException					If a problem is encountered with the manipulation of the markup.
 * @deprecated							Use {@link #addToEndOfContent addToEndOfContent} instead.
 */
	@Deprecated public Results addContentAfter
	(
		final Object			AdditionalContent
	)
	throws TagException
	{
		return addContent(false, AdditionalContent);
	}

/**	Adds to the content of each XML element referenced by the matches in this <code>Results</code> object. For 
 *	example, if there were a match for an "<samp>h1</samp>" element that looked like: 
 *	"<samp>&lt;h1&gt;The Histories&lt;/h1&gt;</samp>", then calling <code>addContent(false, " by Herodotus")</code> 
 *	would change that "<samp>h1</samp>" element to: "<samp>&lt;h1&gt;The Histories by Herodotus&lt;/h1&gt;</samp>".
 *
 * @param InsertBeforeExistingContent	<code>true</code> to insert before the start of the existing content, 
 *										<code>false</code> to insert after it.
 * @param AdditionalContent				Any type of object is potentially suitable, see 
 * 										{@link MutableMarkup#insert MutableMarkup.insert} for details.
 * @return								A new <code>Results</code> object containing references to the content that was 
 *										added to this markup.
 * @throws TagException					If a problem is encountered with the manipulation of the markup.
 * @see #addToStartOfContent(Object)
 * @see #addToEndOfContent(Object)
 */
	public Results addContent
	(
		final boolean			InsertBeforeExistingContent,	//	If false, we insert after the existing content.
		final Object			AdditionalContent
	)
	throws TagException
	{
		final Results			NewResults = new Results(this);

		if (AdditionalContent instanceof Results) {
			final Results			ContentsToAdd = (Results) AdditionalContent;

			for (int Index = 0, Count = size(); Index < Count; Index++) {
				final Match				Target = get(Index);

				for (int CIndex = 0, CCount = ContentsToAdd.MatchesIndex; CIndex < CCount; CIndex++)
					NewResults.addMatch(Target.addContent(InsertBeforeExistingContent, ContentsToAdd.Matches[CIndex]));
			}

		} else {

			for (int Index = 0, Count = size(); Index < Count; Index++)
				NewResults.addMatch(get(Index).addContent(InsertBeforeExistingContent, AdditionalContent));
		}

		return NewResults;
	}

/**	Duplicates each XML element referenced by the matches in this <code>Results</code> object, and places the 
 *	duplicate immediately before the referenced element. The children, if any, of those elements are also duplicated.
 *
 * @return								A new <code>Results</code> object that references each duplicate.
 * @throws TagException					If a problem is encountered with the manipulation of the markup.
 */
	public Results duplicateBefore() throws TagException {
		return duplicate(true);
	}

/**	Duplicates each XML element referenced by the matches in this <code>Results</code> object, and places the 
 *	duplicate immediately after the referenced element. The children, if any, of those elements are also duplicated.
 *
 * @return								A new <code>Results</code> object that references each duplicate.
 * @throws TagException					If a problem is encountered with the manipulation of the markup.
 */
	public Results duplicateAfter() throws TagException {
		return duplicate(false);
	}

/**	Duplicates each XML element referenced by the matches in this <code>Results</code> object, and places the 
 *	duplicate immediately before, or after, the referenced element. The children, if any, of those elements are also 
 *	duplicated.
 *
 * @param InsertDuplicateBefore			<code>true</code> if the duplicate should be inserted before the item that was 
 *										duplicated, <code>false</code> if it should be inserted after.
 * @return								A new <code>Results</code> object that references each duplicate.
 * @throws TagException					If a problem is encountered with the manipulation of the markup.
 */
	public Results duplicate
	(
		final boolean			InsertDuplicateBefore
	)
	throws TagException
	{
		final Results			NewResults = new Results(this);

		for (int Index = 0, Count = size(); Index < Count; Index++)
			NewResults.addMatch(get(Index).duplicate(InsertDuplicateBefore));

		return NewResults;
	}

/**	Replaces each XML element referenced by the matches in this <code>Results</code> object with the supplied material.
 *
 * @param Replacement					Any type of object is potentially suitable, see {@link MutableMarkup#insert 
 * 										MutableMarkup.insert} for details.
 * @return								A new <code>Results</code> object that references each replacement.
 * @throws TagException					If a problem is encountered with the manipulation of the markup.
 */
	public Results replace
	(
		final Object			Replacement
	)
	throws TagException {
		final Results			NewResults = new Results(this);

		if (Replacement instanceof Results) {
			final Results			Contents = (Results) Replacement;

			for (int Index = 0, Count = size(); Index < Count; Index++) {
				final Match				Target = get(Index);

				if (Contents.MatchesIndex > 0) {

				//	The first match is inserted using "replace".

					final Match				ReplacementMatch = Target.replace(Contents.Matches[0]);

					NewResults.addMatch(ReplacementMatch);

				//	All matches other than the first are inserted immediately after the location where the replacement 
				//	occurred.

					for (int CIndex = 1, CCount = Contents.MatchesIndex; CIndex < CCount; CIndex++)
						NewResults.addMatch(ReplacementMatch.add(false, Contents.Matches[CIndex]));
				}
			}

		} else {

			for (int Index = 0, Count = size(); Index < Count; Index++)
				NewResults.addMatch(get(Index).replace(Replacement));
		}

		return NewResults;
	}

/**	Adds new material before each XML element referenced by the matches in this <code>Results</code> object.
 * 
 * @param NewStuff						Any type of object is potentially suitable, see {@link MutableMarkup#insert 
 * 										MutableMarkup.insert} for details.
 * @return								A new <code>Results</code> object that references the inserted material.
 * @throws TagException					If a problem is encountered with the manipulation of the markup.
 */
	public Results addBefore
	(
		final Object			NewStuff
	)
	throws TagException
	{
		return add(true, NewStuff);
	}

/**	Adds new material after each XML element referenced by the matches in this <code>Results</code> object.
 * 
 * @param NewStuff						Any type of object is potentially suitable, see {@link MutableMarkup#insert 
 * 										MutableMarkup.insert} for details.
 * @return								A new <code>Results</code> object that references the inserted material.
 * @throws TagException					If a problem is encountered with the manipulation of the markup.
 */
	public Results addAfter
	(
		final Object			NewStuff
	)
	throws TagException
	{
		return add(false, NewStuff);
	}

/**	Inserts new material before each XML element referenced by the matches in this <code>Results</code> object.
 * 
 * @param NewStuff						Any type of object is potentially suitable, see {@link MutableMarkup#insert 
 * 										MutableMarkup.insert} for details.
 * @return								A new <code>Results</code> object that references the inserted material.
 * @throws TagException					If a problem is encountered with the manipulation of the markup.
 * @deprecated							Use {@link #addBefore addBefore} instead.
 */
	@Deprecated public Results insertBefore
	(
		final Object			NewStuff
	)
	throws TagException
	{
		return add(true, NewStuff);
	}

/**	Inserts new material after each XML element referenced by the matches in this <code>Results</code> object.
 * 
 * @param NewStuff						Any type of object is potentially suitable, see {@link MutableMarkup#insert 
 * 										MutableMarkup.insert} for details.
 * @return								A new <code>Results</code> object that references the inserted material.
 * @throws TagException					If a problem is encountered with the manipulation of the markup.
 * @deprecated							Use {@link #addAfter addAfter} instead.
 */
	@Deprecated public Results insertAfter
	(
		final Object			NewStuff
	)
	throws TagException
	{
		return add(false, NewStuff);
	}

/**	Inserts new material before, or after, each XML element referenced by the matches in this <code>Results</code> 
 *	object.
 *
 * @param InsertBefore					<code>true</code> if the new material should be inserted before the matches, 
 *										<code>false</code> if it should be inserted after.
 * @param NewStuff						Any type of object is potentially suitable, see {@link MutableMarkup#insert 
 * 										MutableMarkup.insert} for details.
 * @return								A new <code>Results</code> object that references the inserted material.
 * @throws TagException					If a problem is encountered with the manipulation of the markup.
 * @deprecated							Use {@link #add add} instead.
 */
	@Deprecated public Results insert
	(
		final boolean			InsertBefore,		//	If false, we insert after each Match.
		final Object			NewStuff
	)
	throws TagException
	{
		return add(InsertBefore, NewStuff);
	}

/**	Inserts new material before, or after, each XML element referenced by the matches in this <code>Results</code> 
 *	object.
 *
 * @param AddBefore						<code>true</code> if the new material should be inserted before the matches, 
 *										<code>false</code> if it should be inserted after.
 * @param NewStuff						Any type of object is potentially suitable, see {@link MutableMarkup#insert 
 * 										MutableMarkup.insert} for details.
 * @return								A new <code>Results</code> object that references the inserted material.
 * @throws TagException					If a problem is encountered with the manipulation of the markup.
 */
	public Results add
	(
		final boolean			AddBefore,		//	If false, we insert after each Match.
		final Object			NewStuff
	)
	throws TagException
	{
		final Results			NewResults = new Results(this);

		if (NewStuff instanceof Results) {
			final Results			Contents = (Results) NewStuff;

			for (int Index = 0, Count = size(); Index < Count; Index++) {
				final Match				Target = get(Index);

				for (int CIndex = 0, CCount = Contents.MatchesIndex; CIndex < CCount; CIndex++)
					NewResults.addMatch(Target.add(AddBefore, Contents.Matches[CIndex]));
			}

		} else {

			for (int Index = 0, Count = size(); Index < Count; Index++)
				NewResults.addMatch(get(Index).add(AddBefore, NewStuff));
		}

		return NewResults;
	}

/**	Deletes each XML element (and any associated children) referenced by the matches in this <code>Results</code> 
 *	object.
 *
 * @return 								A new, empty <code>Results</code> object.
 * @throws TagException					If a problem is encountered with the manipulation of the markup.
 */
	public Results delete() throws TagException {

	//	We delete in reverse order to avoid the performance penalty of 
	//	freshening stale Matches. If we deleted them in normal, first-to-last
	//	order, each delete would make all of the following Matches stale.

		for (int Index = size() - 1; Index >= 0; Index--)
			get(Index).delete();

	//	This Results object no longer has anything to refer to. Empty it.

		clear();

	//	Return a new Results object so that callers can use methods like "pop" to access earlier Results objects that
	// 	refer to document elements that weren't deleted.

		return new Results(this);
	}

/**	Returns a <code>String</code> representation of all of the XML elements (and any associated children) referenced by 
 * 	the matches in this <code>Results</code> object.
 *
 * @return	<code>String</code> representation of all of the XML elements referenced by this <code>Results</code> object.
 */
	@Override public String toString() {
		final StringBuilder		Buff = new StringBuilder(128 * size());
		
		for (int Index = 0, Count = size(); Index < Count; Index++)
			Buff.append(get(Index).toString(false));
		
		return Buff.toString();
	}

/**	Creates a <code>MutableMarkup</code> object containing all of the XML elements (and any associated children) 
 *	referenced by the matches in this <code>Results</code> object. Note that the elements in this markup are the same
 *	elements that are present in the original markup, and changes to these elements in either body of markup will 
 *	result in changes to these elements both bodies of markup.
 *
 * @return						<code>MutableMarkup</code> object containing all of the XML elements referenced by this 
 *								<code>Results</code> object.
 * @see #copy
 */
	public MutableMarkup toMarkup() {
		final MutableMarkup		Subsection = new MutableMarkup();

		for (int Index = 0, Count = size(); Index < Count; Index++)
			Subsection.add(get(Index));

		final Markup			OrigMarkup = getMarkup();

		if (OrigMarkup != null)
			Subsection.setMarkupName("Derivative of \"" + OrigMarkup.getMarkupName() + '"');

		return Subsection;
	}

/**	Moves all of the XML elements (children included) referenced by the matches in this <code>Results</code> object 
 *	into a new <code>MutableMarkup</code> object, then returns a new <code>Results</code> object containing references 
 *	to those elements in their new home. This is useful for extracting a section of a document&mdash;typically a 
 *	template&mdash;to get it out of the way until it is needed at a later time for use in the same document, or in a 
 *	completely different one. Attempts to change the extracted markup by using lingering references to it in 
 *	<code>Match</code> or <code>Results</code> objects will fail. However, if references have been retained to 
 *	individal, mutable <code>Range</code> objects in the markup that was extracted, modifications to the extracted 
 *	markup can be performed using them. That's not a normal usage mode, and you would probably know if you were keeping 
 *	and using such references, so it's usually not worth worrying about, but if you need to be certain that nothing can 
 *	modify the extracted markup through such lingering references, use the {@link #cut} method instead.
 *
 * @return								A new <code>Results</code> object that references all of the extracted material 
 * 										in the new <code>MutableMarkup</code> object that was automatically created to 
 * 										contain it.
 * @throws TagException					If a problem is encountered with the manipulation of the markup.
 * @see #extractToMarkup
 * @see #cut
 */
	public Results extract() throws TagException {
		final Results			NewResults = new Results(this);
		final MutableMarkup		Subsection = new MutableMarkup();

		for (int Index = 0, Count = size(); Index < Count; Index++)
			NewResults.addMatch(Subsection.add(get(Index)));

		delete();

		final Markup			OrigMarkup = getMarkup();

		if (OrigMarkup != null)
			Subsection.setMarkupName("Derivative of \"" + OrigMarkup.getMarkupName() + '"');

		return NewResults;
	}

/**	Moves all of the XML elements (children included) referenced by the matches in this <code>Results</code> object 
 *	into a new <code>MutableMarkup</code> object. This is useful for extracting a section of a document&mdash;typically 
 *	a template&mdash;to get it out of the way until it's needed at a later time for use in the same document, or in a 
 *	completely different one. Attempts to change the extracted markup by using lingering references to it in 
 *	<code>Match</code> or <code>Results</code> objects will fail. However, if references have been retained to 
 *	individal, mutable <code>Range</code> objects in the markup that was extracted, modifications to the extracted 
 *	markup can be performed using them. That's not a normal usage mode, and you would probably know if you were keeping 
 *	and using such references, so it's usually not worth worrying about, but if you need to be certain that nothing can 
 *	modify the extracted markup through such lingering references, use the {@link #cut} method instead.
 *
 * @return								A new <code>MutableMarkup</code> object containing all of the extracted material.
 * @throws TagException					If a problem is encountered with the manipulation of the markup.
 * @see #extract
 * @see #cut
 */
	public MutableMarkup extractToMarkup() throws TagException {
		final MutableMarkup		Subsection = toMarkup();

		delete();

		return Subsection;
	}

/**	Creates a copy of all of the markup represented by the matches (and any children they may have) in this 
 *	<code>Results</code> object. Changes to the original markup will not affect this copy.
 * 
 * @return								A new <code>ImmutableMarkup</code> object that is a copy of all of the markup 
 * 										represented by the matches in this <code>Results</code> object.
 * @throws TagException					If a problem is encountered with the manipulation of the markup.
 * @see #toMarkup
 */
	public ImmutableMarkup copy() throws TagException {
		return toMarkup().getImmutable();
	}

/**	Removes all of the markup represented by the matches (and any children they may have) in this <code>Results</code>
 *	object, and returns it in a new <code>ImmutableMarkup</code> object. Changes to the original markup will not affect
 *	this copy.
 * 	
 * @return								A new <code>MutableMarkup</code> object that holds all of the markup represented 
 * 										by the matches in this <code>Results</code> object.
 * @throws TagException					If a problem is encountered with the manipulation of the markup.
 * @see #extractToMarkup
 * @see #extract
 */
	public ImmutableMarkup cut() throws TagException {
		return extractToMarkup().getImmutable();
	}

/**	Adds one or more CSS class names to the "<samp>class</samp>" attribute of all of the applicable XML elements 
 *	referenced by the matches in this <code>Results</code> object. The "<samp>class</samp>" attribute will be created, 
 *	as necessary. Duplicate class names are detected and omitted, so there's no need to worry about redundancy.
 *
 * @param ClassList						Space-separated list of class names.
 * @return								A new <code>Results</code> object containing all of the matches in this 
 *										<code>Results</code> object.
 * @throws TagException					If a problem is encountered with the manipulation of the markup.
 * @see #setClass(String)
 */
	public Results addClass
	(
		final String			ClassList
	)
	throws TagException
	{
		final Results			NewResults = new Results(this);

		for (int Index = 0, Count = size(); Index < Count; Index++)
			NewResults.addMatch(get(Index).addClass(ClassList));

		return NewResults;
	}

/**	Sets the value of the "<samp>class</samp>" attribute of all of the applicable XML elements referenced by the matches 
 *	in this <code>Results</code> object. The "<samp>class</samp>" attribute will be created, as necessary.
 *
 * @param ClassList						Space-separated list of class names.
 * @return								A new Results object containing all of the matches in this <code>Results</code> 
 * 										object.
 * @throws TagException					If a problem is encountered with the manipulation of the markup.
 * @see #addClass(String)
 */
	public Results setClass
	(
		final String			ClassList
	)
	throws TagException
	{
		final Results			NewResults = new Results(this);

		for (int Index = 0, Count = size(); Index < Count; Index++)
			NewResults.addMatch(get(Index).setClass(ClassList));

		return NewResults;
	}

/**	Removes one or more CSS class names from the "<samp>class</samp>" attribute of all of the applicable XML elements 
 *	referenced by the matches in this <code>Results</code> object. If the "<samp>class</samp>" attribute becomes empty, 
 *	it will be omitted from the resulting markup.
 *
 * @param ClassList						Space-separated list of class names.
 * @return								A new <code>Results</code> object containing all of the matches in this 
 *										<code>Results</code> object.
 * @throws TagException					If a problem is encountered with the manipulation of the markup.
 */
	public Results deleteClass
	(
		final String			ClassList
	)
	throws TagException
	{
		final Results			NewResults = new Results(this);

		for (int Index = 0, Count = size(); Index < Count; Index++)
			NewResults.addMatch(get(Index).deleteClass(ClassList));

		return NewResults;
	}

/**	Returns the value of the specified attribute from the first match in this <code>Results</code> object. If there are
 * 	multiple matches in this <code>Results</code> object and you want the attribute value from each one, you will have 
 * 	to retrieve each match individually (using {@link #get(int)} or {@link #iterator}) and invoke 
 * 	{@link Match#getAttribute Match.getAttribute} on each one.
 * 
 * @param AttrName						Name of the attribute.
 * @return								Value of the attribute, or <code>null</code> if there is no such attribute.
 */
	public String getAttribute
	(
		final String			AttrName
	)
	{
		return size() > 0 ? get(0).getAttribute(AttrName) : null;
	}

/**	Adds an attribute to all of the applicable XML elements referenced by the matches in this <code>Results</code> 
 *	object. If the attribute is already present, the new value is appended to the existing value. Where that appending 
 *	behavior is not wanted, use {@link #setAttribute setAttribute}.
 *
 * @param AttrName						Name of the attribute. (For the "<samp>class</samp>" attribute, {@link #addClass 
 * 										addClass} is more efficient.)
 * @param AttrValue						Value of the attribute. Can be <code>null</code>. If not <code>null</code>, the 
 * 										object's <code>toString</code> method is invoked.
 * @return								A new <code>Results</code> object containing all of the matches in this 
 *										<code>Results</code> object.
 * @throws TagException					If a problem is encountered with the manipulation of the markup.
 * @see #setAttribute(String, Object)
 */
	public Results addAttribute
	(
		final String			AttrName,
		final Object			AttrValue
	)
	throws TagException
	{
		final Results			NewResults = new Results(this);

		for (int Index = 0, Count = size(); Index < Count; Index++)
			NewResults.addMatch(get(Index).addAttribute(AttrName, AttrValue));

		return NewResults;
	}

/**	Sets the value of the specified attribute in all of the applicable XML elements referenced by the matches in this
 * 	<code>Results</code> object. If the attribute does not exist, nothing happens. If it does exist, its value is 
 *	replaced by the new value.
 *
 * @param AttrName						Name of the attribute. (For the "<samp>class</samp>" attribute, {@link #setClass 
 * 										setClass} is more efficient.)
 * @param AttrValue						Value of the attribute. Can be <code>null</code>. If not <code>null</code>, the 
 * 										object's <code>toString</code> method is invoked.
 * @return								A new <code>Results</code> object containing all of the matches in this 
 *										<code>Results</code> object.
 * @throws TagException					If a problem is encountered with the manipulation of the markup.
 */
	public Results updateAttribute
	(
		final String			AttrName,
		final Object			AttrValue	//	Can be null.
	)
	throws TagException
	{
		final Results			NewResults = new Results(this);

		for (int Index = 0, Count = size(); Index < Count; Index++)
			NewResults.addMatch(get(Index).updateAttribute(AttrName, AttrValue));

		return NewResults;
	}

/**	Sets the value of the specified attribute in all of the applicable XML elements referenced by the matches in this
 * 	<code>Results</code> object. If the attribute does not exist, it is created. If it does exist, its value is 
 *	replaced by the new value.
 *
 * @param AttrName						Name of the attribute. (For the "<samp>class</samp>" attribute, {@link #setClass 
 * 										setClass} is more efficient.)
 * @param AttrValue						Value of the attribute. Can be <code>null</code>. If not <code>null</code>, the 
 * 										object's <code>toString</code> method is invoked.
 * @return								A new <code>Results</code> object containing all of the matches in this 
 *										<code>Results</code> object.
 * @throws TagException					If a problem is encountered with the manipulation of the markup.
 */
	public Results setAttribute
	(
		final String			AttrName,
		final Object			AttrValue	//	Can be null.
	)
	throws TagException
	{
		final Results			NewResults = new Results(this);

		for (int Index = 0, Count = size(); Index < Count; Index++)
			NewResults.addMatch(get(Index).setAttribute(AttrName, AttrValue));

		return NewResults;
	}

/**	Sets the value of an arbitrary number of specified attributes in all of the applicable XML elements referenced by 
 *	the matches in this <code>Results</code> object. If the attribute does not exist, it is created. If it does exist, 
 *	its value is replaced by the new value.
 *
 * @param AttrNamesAndValues			Array of name/value pairs specify the attributes to be set. The first element in 
 * 										each pair is the name, and the second is the value. The value can be 
 * 										<code>null</code>.
 * @return								A new <code>Results</code> object containing all of the matches in this 
 *										<code>Results</code> object.
 * @throws TagException					If a problem is encountered with the manipulation of the markup.
 */
	public Results setAttributes
	(
		final Object...			AttrNamesAndValues
	)
	throws TagException
	{
		final int				AVLength = AttrNamesAndValues.length;

		if (AVLength % 2 != 0)
			throw new IllegalArgumentException("The array passed to setAttributes is meant to be composed of attribute-name/value pairs, and therefore should contain an even number of elements. The array passed in this case, however, contains an odd number of elements: " + AVLength + '.');

		final Results			NewResults = new Results(this);

		for (int Index = 0, Count = size(); Index < Count; Index++) {
			final Match				Elem = get(Index);

			for (int AIndex = 0; AIndex < AVLength; AIndex += 2)
				Elem.setAttribute(AttrNamesAndValues[AIndex].toString(), AttrNamesAndValues[AIndex + 1]);

			NewResults.addMatch(Elem);
		}

		return NewResults;
	}

/**	Deletes the specified attribute from all of the applicable XML elements referenced by the matches in this 
 *	<code>Results</code> object.
 *
 * @param AttrName						Name of the attribute to be deleted.
 * @return								A new <code>Results</code> object containing all of the matches in this 
 *										<code>Results</code> object.
 * @throws TagException					If a problem is encountered with the manipulation of the markup.
 */
	public Results deleteAttribute
	(
		final String			AttrName
	)
	throws TagException
	{
		final Results			NewResults = new Results(this);

		for (int Index = 0, Count = size(); Index < Count; Index++)
			NewResults.addMatch(get(Index).deleteAttribute(AttrName));

		return NewResults;
	}

/**	Prints the XML elements referenced by the matches in this <code>Results</code> object to {@link System#out}.
 *
 * @return								This <code>Results</code> object, unaltered.
 * @deprecated Use {@link #toString} instead.
 * @see #toString
 */
	@Deprecated public Results print() {

		print(null);
		return this;
	}

/**	Prints the XML elements referenced by the matches in this <code>Results</code> object to {@link System#out} with 
 *	each line prefixed by the supplied string.
 *
 * @param Prefix						A <code>String</code> to be used a prefix for each line of output. (Handy for 
 * 										indenting.)
 * @return								This <code>Results</code> object, unaltered.
 * @deprecated Use {@link #toString} instead.
 * @see #toString
 */
	@Deprecated public Results print
	(
		String					Prefix
	)
	{
		if (Prefix == null)
			Prefix = "";

		for (int Index = 0, Count = MatchesIndex; Index < Count; Index++) {

			System.out.print(Prefix);
			Matches[Index].print();
		}

		return this;
	}
	
//	Note that the geti18nServices method is present in both the SmartStringConverter and Results classes, and the 
//	implementations should be kept identical. 

	private i18nServices geti18nServices() {
		
		if (i18nServicesRequested)
			return i18n;
		
		if (CanRequestI18NServices == Boolean.FALSE)	//	Yes, a reference equality check is correct.
			return null;
		
		CanRequestI18NServices = Boolean.FALSE;
		
		try {
			
		//	If Class.forName fails, it throws ClassNotFoundException, LinkageError, or ExceptionInInitializerError.
		//	We perform this method invocation on the assumption that a ClassNotFoundException has less impact on the 
		//	JVM than the NoClassDefFoundError potentially produced by the Qwicap.getCurrentInstance invocation below.
			
			Class.forName("javax.servlet.Servlet");
			
		//	If we've arrived here, the javax.servlet.Servlet class must be available (in some form), so the invocation 
		//	of Qwicap.getCurrentInstance should not cause any class loading problems. That's not to say that it will
		//	succeed, but it won't fail due to a problem that will necessarily persist.
			
			CanRequestI18NServices = Boolean.TRUE;
			
		//	If Qwicap.getCurrentInstance fails because some of the classes it references are missing (typically the 
		//	javax.servlet.* classes), a NoClassDefFoundError is thrown. Anyway, that has been reported in the past.
		//	In current tests, when this code is run outside of a Qwicap web application, Qwicap.getCurrentInstance 
		//	throws an IllegalStateException because there is no current instance of Qwicap. Both of these cases, and
		//	many others, are handled by the catch blocks below.
			
		//	Note that we pass "Qwicap.class" below, because, in order to simplify interationalization, we keep all of our 
		//	internationalizable material in a single file, associated with the Qwicap class.
			
			i18nServicesRequested = true;
			i18n = Qwicap.getCurrentInstance().geti18nServicesForClass(Qwicap.class);
			
		} catch (NoClassDefFoundError e) {
			
		//	This happens when only Qwicap's XML engine is being used, and therefore the javax.servlet library is not 
		//	present, which the JVM notices when it loads the Qwicap class.
			
			CanRequestI18NServices = Boolean.FALSE;
			Log.log(Level.FINE, "Cannot get internationalization services. Will continue without them. ({0} message: \"{1}\")", new Object[] { e.getClass().getName(), e.getMessage() });
			
		} catch (Exception e) {
			
			Log.log(Level.FINE, "Cannot get internationalization services. Will continue without them. ({0} message: \"{1}\")", new Object[] { e.getClass().getName(), e.getMessage() });
		}
		
		return i18n;
	}
}





