/*	QWICAP (Quick Web Interface for Conventional Applications) by Chris W. Johnson
	Copyright (C) 2003-2010 The University of Texas at Austin ("U.T. Austin"). All rights reserved.
	
	U.T. Austin's copyright notice in this evolving version reflects our original authorship. It does not assert 
	copyright over changes contributed by others.
	
	This library is free software; you can redistribute it and/or modify it under the terms of the GNU Lesser General 
	Public License as published by the Free Software Foundation; either version 2.1 of the License, or (at your option) 
	any later version.
	
	This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied 
	warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for 
	more details.
	
	You should have received a copy of the GNU Lesser General Public License along with this library; if not, write to 
	the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/


//  Selector.java
//  qwicap
//
//  Created by Chris W. Johnson on Tue Jun 10 2003.


package edu.utexas.its.eis.tools.qwicap.template.css;


import edu.utexas.its.eis.tools.qwicap.template.xml.structure.*;


/**	<code>Selector</code> is the abstract base-class of all selector classes. It also performs the parsing of CSS 
 *	selector strings, and the instantiation of the necessary selector objects. Example CSS selectors: "*", "E",
 *	"E:first-child", "E:lang(c)", "E.myclass[href='http://www.utexas.edu/'][target]", "E#myid".
 *	
 *	@author Chris W. Johnson
 */
abstract class Selector extends CSSPatternComponent {
	
	Selector				NextSel;
	
	
/*	More correct rules for parsing:

	The material before the first '#' or '.' is the tag name. If there is nothing, or an asterisk, before those 
	characters, we're dealing with a universal selector that selects all tag of all types.
	
	After the first '#' or '.' is an id attribute value, or an attribute name, respectively. The HTML 4.01 spec. 
	states: "ID and NAME tokens must begin with a letter ([A-Za-z]) and may be followed by any number of letters, 
	digits ([0-9]), hyphens ("-"), underscores ("_"), colons (":"), and periods (".")."
	See <http://www.w3.org/TR/html4/types.html#type-name>.
	
	The ID or NAME token may be followed by one or more attribute selectors which are bounded by brackets 
	('[' and ']'). If the ID or NAME token is followed by the end of the string, or whitespace, the end of the 
	selector has been reached, and another selector may follow in the latter case.
	
	
	Leading Universal Selector Cases:
	
	*.attribute
	*.attribute[....
	*#id
	*#id[....
	*[....
	
	Leading Implicit Universal Selector Cases:
	
	.attribute
	.attribute[....
	#id
	#id[....
	[....
	
	Tag Selector Cases:
	
	tag.attribute
	tag.attribute[....
	tag#id
	tag#id[....
	tag[....

	A Pseudo-Class selector, like ":first-child" or ":lang", can be appended to any of the above cases.
	Note that IDs and NAMEs are allowed to contain colons.
	
	In XML, element and attribute names must begin with a letter, underscore ('_') or colon (':'), and continue with 
	any number of letters, digits, periods ('.'), dashes ('-'), underscores ('_') or colons (':'). It's actually a 
	little more complicated than that (see <http://www.w3.org/TR/REC-xml/#NT-Name> for the details), but the summary 
	above should be accurate enough for our purposes here.
	
	
*/	

/*	We're decoding a "simple selector" here. The CSS2 spec, <http://www.w3.org/TR/REC-CSS2/selector.html#q2>,
 	defines such a selector as followS:

	"A simple selector is either a type selector or universal selector followed immediately by zero or more 
	attribute selectors, ID selectors, or pseudo-classes, in any order. The simple selector matches if all 
	of its components match."
*/	
	static Selector read
	(
		final CSSPatterns		Patterns,
		final CharacterIterator	Chars
	)
	{
		if (!Chars.hasNext())
			return null;
		
		final SelectorCompound	CompoundSelector = new SelectorCompound();
		String					FirstIdentifier  = null;
		
		switch (Chars.next()) {
		
		//	Handle the optional, leading namespace selectors. See: <http://www.w3.org/TR/2000/WD-css3-selectors-20001005/#typenmsp>.
			
			case '*':
				
				if (Chars.nextCharIs('|'))
					;	//	Add a selector for elements in any namespace, including those without any declared namespace.
				else
					Chars.goBack();	//	This is not part of a namespace selector.
				break;
				
			case '|':
				
				CompoundSelector.add(new SelectorNoDeclaredNamespace());	//	Add a selector for elements without any declared namespace.
				break;
			
		//	These characters are definitely NOT part of a namespace selector.
			
			case '.':	
			case '#':
			case '[':
				
				Chars.goBack();
				break;
				
			default:
				
			//	Read the leading identifier. If the next character is '|' then we know that the identifier is the
			//	name of a namespace. Otherwise we know it's a SelectorType.
				
				FirstIdentifier = CSSParser.getIdentifier(Chars.goBack());
				
				if (Chars.nextCharIs('|')) {
					
					Chars.next();	//	"Read" the '|' character in order to advance the character iterator beyond it.
					
				//	Add a selector for elements in the namespace named by this identifier.
					
					final Namespace			Ns = Patterns.getNamespaceForPrefix(FirstIdentifier);
					
					if (Ns == null)
						throw new IllegalArgumentException("No \"@namespace\" rule has defined a namespace named \"" + FirstIdentifier + "\", as referenced in the CSS pattern \"" + Chars.toOriginalString() + "\".");
					
					CompoundSelector.add(new SelectorNamespace(Ns));
					FirstIdentifier = null;
					
				} else {
					
				//	If a default namespace was specified, add a selector for it. Otherwise, add a selector for 
				//	elements in any namespace, including those without any declared namespace.
					
					final Namespace			DefaultNs = Patterns.getDefaultNamespace();

					if (DefaultNs != Namespace.kNoNamespace)
						CompoundSelector.add(new SelectorNamespace(DefaultNs));	//	Add a selector for the default namespace.
					
//					if (DefaultNs == Namespace.kNoNamespace)
//						CompoundSelector.add(new SelectorNoDeclaredNamespace());
//					else
//						CompoundSelector.add(new SelectorNamespace(DefaultNs));	//	Add a selector for the default namespace.
				}
				break;		
		}
		
		if (FirstIdentifier != null) {
			
			CompoundSelector.add(new SelectorType(FirstIdentifier));
			
		} else if (Chars.hasNext()) {
			
		//	Handle the normal (non-namespace related), leading selectors.
			
			switch (Chars.next()) {
				
				case '*':	//	Begins with Explicit Universal Selector
					
					CompoundSelector.add(new SelectorUniversal());
					break;
					
				case '.':	//	Begins with Implicit Universal Selector
				case '#':	//	Begins with Implicit Universal Selector
				case '[':	//	Begins with Implicit Universal Selector
				
					CompoundSelector.add(new SelectorUniversal());
					Chars.goBack();
					break;
					
				default:	//	Anything else must be a Type Selector.
					
					CompoundSelector.add(new SelectorType(CSSParser.getIdentifier(Chars.goBack())));
					break;
			}
		}
		
	CharacterProcessingLoop:
		
		while (Chars.hasNext()) {
			
			switch (Chars.next()) {
			
				case '.':	//	Class Selector (a convenient shorthand for the "[class~='Attr']" attribute selector).
					
					CompoundSelector.add(new SelectorClass(CSSParser.getIdentifier(Chars)));
					break;
				
				case '#':	//	ID Selector.
					
					CompoundSelector.add(new SelectorID(CSSParser.getIdentifier(Chars)));	//	IntelliJ IDEA 4.5 thinks this is bad, but the java compiler doesn't, and I agree with the compiler.
					break;
					
				case '[':	//	Attribute Selector
					
				//	Note that the left-bracket is expected by SelectorAttribute.create, which is why we call 
				//	Chars.goBack. Similarly, the terminating right-bracket is expected by SelectorAttribute.create,
				//	which is why we don't have to deal with it here.
					
					CompoundSelector.add(SelectorAttribute.create(Patterns, Chars.goBack()));	
					break;
					
				case ':':	//	Pseudo-Class Selector
					
					CompoundSelector.add(SelectorPseudoClass.create(Chars));
					break;
					
				default:	//	We've found a character which is not part of this selector.
					
					Chars.goBack();
					break CharacterProcessingLoop;
			}
		}
		
		return CompoundSelector.simplify();
	}
	
/**	Tests whether, or not, a particular tag matches the criterion represented by this selector.
 * 
 * @param OtherTag			The tag to be tested by this selector.
 * @param OtherTagChildNo	The child number of the tag to be tested by this selector. If the tag is the first child 
 *							of its parent, its child number is zero. Child numbers increase monotonically from there.
 * @return					<code>true</code> if this criterion applies to the specified tag, <code>false</coder> 
 *							otherwise.
 */
	abstract boolean selects(TagWithAttributes OtherTag, int OtherTagChildNo);
	
	@Override public abstract String toString();
}






