/*	QWICAP (Quick Web Interface for Conventional Applications) by Chris W. Johnson
	Copyright (C) 2003-2008 The University of Texas at Austin ("U.T. Austin"). All rights reserved.
	
	U.T. Austin's copyright notice in this evolving version reflects our original authorship. It does not assert 
	copyright over changes contributed by others.
	
	This library is free software; you can redistribute it and/or modify it under the terms of the GNU Lesser General 
	Public License as published by the Free Software Foundation; either version 2.1 of the License, or (at your option) 
	any later version.
	
	This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied 
	warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for 
	more details.
	
	You should have received a copy of the GNU Lesser General Public License along with this library; if not, write to 
	the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/


//  SelectorAttribute.java
//  qwicap
//
//  Created by Chris W. Johnson on Wed Jun 11 2003.


package edu.utexas.its.eis.tools.qwicap.template.css;

import edu.utexas.its.eis.tools.qwicap.template.xml.structure.Namespace;


/**	<code>SelectorAttribute</code> is the abstract base-class of all the attribute selector classes. It also does the 
 *	parsing of the portions of CSS selector strings that deal with the attributes of tags, and the instantiation of the 
 *	necessary attribute selector objects. Example selectors: "E[foo]", "E[foo='thingy']", "E[foo~='thingy']", 
 *	"E[lang|='en']". The portions in brackets are the attribute selectors, they're what this class and its children 
 *	deal with.
 *	
 *	@author Chris W. Johnson
 */
abstract class SelectorAttribute extends Selector {
	
//	E[foo]				--	SelectorAttributeExists
//	E[foo="warning"]	--	SelectorAttributeValueMatches
//	E[foo~="warning"]	--	SelectorAttributeValueIncludes
//	E[lang|="en"]		--	SelectorAttributeValueStartsWith
	
//	For information on namespace-aware selectors, see: <http://www.w3.org/TR/css3-selectors/#attrnmsp>
	
	static Selector create
	(
		final CSSPatterns		Patterns,	//	We only need this because it provides the accumulated namespace information associated with this CSS pattern(s).
		final CharacterIterator	Chars
	)
	{
		if (!Chars.hasNext())
			throw new IllegalArgumentException("There are no characters remaining in the specified selector. See " + Chars + '.');
		
		if (Chars.next() != '[')
			throw new IllegalArgumentException("Attribute selectors are always bounded by brackets. The first character of '" + Chars.goBack() + "' is not a left-bracket ('[').");
		
		String					AttributeName      = CSSParser.getIdentifier(Chars);
		Namespace				AttributeNamespace = null;
		
		if (!Chars.hasNext())
			throw new IllegalArgumentException("Would-be attribute selector \"[" + AttributeName + "\" is incomplete; there's nothing after the attribute identifier. See " + Chars + '.');
		
		Selector				Sel = null;
		boolean					Repeat;
		
		do {
			
			Repeat = false;
			
			switch (Chars.next()) {
				
				case ']':
					
					Sel = new SelectorAttributeExists(AttributeNamespace, AttributeName);
					Chars.goBack();	//	The sanity check at the end of this method expects this right-bracket to be the "next" character.
					break;
					
				case '=':
					
					Sel = new SelectorAttributeValueMatches(AttributeNamespace, AttributeName, CSSParser.getIdentifierOrString(Chars));
					break;
					
				case '~':
					
					if (Chars.hasNext() && Chars.next() == '=') {
						
					//	todo Consider how to handle namespace issues here.
						
						if ("class".equals(AttributeName))
							Sel = new SelectorClass(CSSParser.getIdentifierOrString(Chars));
						else
							Sel = new SelectorAttributeValueIncludes(AttributeNamespace, AttributeName, CSSParser.getIdentifierOrString(Chars));
					} else
						throw new IllegalArgumentException("Expected equal ('=') character not found after tilde ('~') character in would-be CSS attribute selector. See " + Chars + '.');
					break;
					
				case '|':
					
					if (Chars.hasNext()) {
						
						if (Chars.next() == '=') {
							
							Sel = new SelectorAttributeValueStartsWith(AttributeNamespace, AttributeName, CSSParser.getIdentifierOrString(Chars));
							
						} else {
							
						//	If the vertical bar isn't followed by an equal sign ('='), it is the separator between a 
						//	namespace prefix, and the actual name of the attribute. For details, see: 
						//	<http://www.w3.org/TR/css3-selectors/#attrnmsp>.
							
							final String			NamespaceName = AttributeName;
												
							AttributeNamespace = Patterns.getNamespaceForPrefix(NamespaceName);
							AttributeName      = CSSParser.getIdentifier(Chars.goBack());
							Repeat             = true;
					
							if (AttributeNamespace == null)
								throw new IllegalArgumentException("No \"@namespace\" rule has defined a namespace named \"" + NamespaceName + "\", as referenced in the CSS pattern \"" + Chars.toOriginalString() + "\".");
						}
						
					} else
						throw new IllegalArgumentException("The expected equal ('=') character was not found after vertical bar ('|') character in a would-be CSS attribute selector. (There were no characters following the vertical bar.) See " + Chars + '.');
					
					break;
				
				default:
					
					throw new IllegalArgumentException("Invalid separator between attribute selector's attribute name and all that follows. See " + Chars + '.');
			}
		
		} while (Repeat);
		
		if (Chars.hasNext() == false || Chars.next() != ']')
			throw new IllegalArgumentException("The CSS attribute selector \"" + Sel + "\" was not terminated with the required right-bracket (']') character. See " + Chars + '.');
		
		return Sel;
	}
}





