/*	QWICAP (Quick Web Interface for Conventional Applications) by Chris W. Johnson
	Copyright (C) 2003-2010 The University of Texas at Austin ("U.T. Austin"). All rights reserved.
	
	U.T. Austin's copyright notice in this evolving version reflects our original authorship. It does not assert 
	copyright over changes contributed by others.
	
	This library is free software; you can redistribute it and/or modify it under the terms of the GNU Lesser General 
	Public License as published by the Free Software Foundation; either version 2.1 of the License, or (at your option) 
	any later version.
	
	This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied 
	warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for 
	more details.
	
	You should have received a copy of the GNU Lesser General Public License along with this library; if not, write to 
	the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/


//  DescendantEnumerator.java
//  qwicap
//
//  Created by Chris W. Johnson on Fri Jun 13 2003.


package edu.utexas.its.eis.tools.qwicap.template.xml;


import edu.utexas.its.eis.tools.qwicap.template.xml.structure.*;

import java.util.NoSuchElementException;
import java.text.DecimalFormat;


/**	<code>DescendantEnumerator</code> provides a convenient means of enumerating some, or all, of a body of XML markup
 *	while maintaining an awareness of each tag's context in the tag hierarchy. The elements it enumerates are 
 *	determined by the <code>MarkupEnumeration</code> object passed to its constructor.
 *	
 *	@author Chris W. Johnson
 */
public final class DescendantEnumerator implements MarkupEnumeration {
	
	private final MarkupEnumeration	Elems;
	private final boolean			ReturnEndTags;
	
	private int						ParentCount;
	private int						SiblingNo;
	
	private int[]					SiblingNoStack;
	private int						SiblingNoStackLength;
	private int						SiblingNoStackIndex;
	
	private Element					CurrElement;
	private Element					NextElement;
	
	
	private DescendantEnumerator
	(
		final DescendantEnumerator	Master
	)
	{
		Elems                = Master.Elems.clone();
		ReturnEndTags        = Master.ReturnEndTags;
		
		ParentCount          = Master.ParentCount;
		SiblingNo            = Master.SiblingNo;
		
		SiblingNoStack       = new int[Master.SiblingNoStackLength];
		System.arraycopy(Master.SiblingNoStack, 0, SiblingNoStack, 0, Master.SiblingNoStackIndex);
		SiblingNoStackLength = Master.SiblingNoStackLength;
		SiblingNoStackIndex  = Master.SiblingNoStackIndex;
		
		CurrElement          = new Element(Master.CurrElement);
		NextElement          = new Element(Master.NextElement);
	}
	
	public DescendantEnumerator
	(
		final MarkupEnumeration	Elems
	)
	{
		this(Elems, false);
	}
		
	public DescendantEnumerator
	(
		final MarkupEnumeration	Elems,
		final boolean			ReturnEndTags
	)
	{
		this.Elems           = Elems;
		this.ReturnEndTags   = ReturnEndTags;
		
		SiblingNo            = -1;
		
		SiblingNoStack       = new int[50];
		SiblingNoStackLength = SiblingNoStack.length;
		
		CurrElement          = new Element();
		NextElement          = new Element();
		
		advance();
	}
	
	@Override public Markup getMarkup() {
		return Elems.getMarkup();
	}
	
	public int getParentCount() {
		return CurrElement.ParentCount;
	}
	
	public int getChildNo() {
		return CurrElement.SiblingNo;
	}
	
	@Override public int getIndex() {
		return CurrElement.ElemIndex;
	}
	
	@Override public Match getMatch() {
		return Match.create(getMarkup(), CurrElement.Elem, CurrElement.ElemIndex);
	}
	
	@Override public boolean hasNext() {
		return hasMoreElements();
	}
	
	@Override public Range next() {
		return nextRange();
	}
	
	@Override public void remove() {
		throw new UnsupportedOperationException();
	}
	
	@Override public boolean hasMoreElements() {
		return NextElement.Elem != null;
	}
	
	@Override public Range nextElement() {
		return nextRange();
	}
	
	@Override public Range nextRange() {
		
		if (NextElement.Elem == null)
			throw new NoSuchElementException();
		
		advance();
		
		return CurrElement.Elem;
	}
	
	private void advance() {
		final Element			Next = CurrElement;
		
		CurrElement = NextElement;
		NextElement = Next;

		Next.Elem = null;	//	If there are no more elements, this will cause hasMoreElements to return false.
		
		while (Elems.hasMoreElements()) {
			final Range				Elem = Elems.nextRange();
			
			Next.Elem      = Elem;
			Next.ElemIndex = Elems.getIndex();
			
			if (Elem.isStartTag()) {
				
				Next.ParentCount = ParentCount;
				Next.SiblingNo   = ++SiblingNo;
				
				push(SiblingNo);
				
				ParentCount++;
				SiblingNo = -1;
				
				return;
				
			} else if (Elem.isEndTag()) {
				
				SiblingNo = pop();
				
				Next.ParentCount = --ParentCount;
				Next.SiblingNo   = -1;	//	End tags have no sibling number because they aren't considered "real" tags for selector purposes.
				
				if (ReturnEndTags)
					return;
				
			} else if (Elem.isEmptyTag()) {
				
				Next.ParentCount = ParentCount;
				Next.SiblingNo   = ++SiblingNo;
				
				return;
				
			} else {
				
				Next.ParentCount = ParentCount;
				Next.SiblingNo   = -1;		//	Only start tags and empty tags have a valid sibling number.
				
				return;
			}
		}
	}
	
	private void push
	(
		final int				No
	)
	{
		if (SiblingNoStackIndex >= SiblingNoStackLength) {
			
		//	Allocate the new array, and copy the contents of the previous array into it, then replace the previous
		//	array with the new one.
			
			final int				NewLength = (SiblingNoStackLength * 3) / 2 + 1;
			final int[]				NewArray  = new int[NewLength];
			
			System.arraycopy(SiblingNoStack, 0, NewArray, 0, SiblingNoStackLength);
			
			SiblingNoStack       = NewArray;
			SiblingNoStackLength = NewLength;
		}
		
		SiblingNoStack[SiblingNoStackIndex++] = No;
	}
	
	private int pop() {
		return SiblingNoStack[--SiblingNoStackIndex];
	}
	
	@Override public DescendantEnumerator clone() {
		return new DescendantEnumerator(this);
	}

	public void print() {
		print(false);
	}
	
	public void print
	(
		final boolean				PrintChildNos
	) 
	{
		final DecimalFormat			DFmt  = new DecimalFormat("0000");
		final DescendantEnumerator	Elems = clone();
		
		while (Elems.hasMoreElements()) {
			final Range					Elem = Elems.nextRange();
			
		//	Ignore regions of whitespace.
			
			if (Elem instanceof Whitespace)
				continue;
			
		//	Print the child no, where applicable.
			
			if (PrintChildNos) {
				final int					ChildNo = Elems.getChildNo();
				
				if (ChildNo < 0)
					System.out.print("----");
				else
					System.out.print(DFmt.format(ChildNo));
			}
			
		//	Indent according to the number of parents.
			
			for (int Index = 0, Count = Elems.getParentCount(); Index < Count; Index++)
				System.out.print('\t');
			
		//	Print out the item, sans leading and trailing whitespace.
			
			System.out.println(Elem.toString().trim());
		}
	}
	
	private static final class Element {
		
		Range					Elem;
		int						ElemIndex;
		int						ParentCount;
		int						SiblingNo;
		
		Element() {
			
			ElemIndex   = -1;
			ParentCount = -1;
			SiblingNo   = -1;
		}
		
		Element
		(
			final Element			Master
		)
		{
			Elem        = Master.Elem;
			ElemIndex   = Master.ElemIndex;
			ParentCount = Master.ParentCount;
			SiblingNo   = Master.SiblingNo;
		}
	}
}
