/*	QWICAP (Quick Web Interface for Conventional Applications) by Chris W. Johnson
	Copyright (C) 2008-2010 The University of Texas at Austin ("U.T. Austin"). All rights reserved.
	
	U.T. Austin's copyright notice in this evolving version reflects our original authorship. It does not assert 
	copyright over changes contributed by others.
	
	This library is free software; you can redistribute it and/or modify it under the terms of the GNU Lesser General 
	Public License as published by the Free Software Foundation; either version 2.1 of the License, or (at your option) 
	any later version.
	
	This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied 
	warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for 
	more details.
	
	You should have received a copy of the GNU Lesser General Public License along with this library; if not, write to 
	the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/


package edu.utexas.its.eis.tools.qwicap.template.xml;


import edu.utexas.its.eis.tools.qwicap.template.css.Results;
import edu.utexas.its.eis.tools.qwicap.template.css.CSSPatterns;
import edu.utexas.its.eis.tools.qwicap.template.xml.structure.*;

import java.io.PrintWriter;
import java.io.IOException;


/**	Represents the content inside of a CDATA, comment or declaration element, as though the content were a child of the 
 * 	element. (Internally, the content is <i>not</i> separate from the element, so treating the content as a child is 
 * 	merely a convenient fiction intended to keep the markup manipulation API simple and relatively consistent.) This 
 * 	class is not part of Qwicap's public API, but could not be hidden due to limitations in Java's access control
 * 	scheme.
 * 
 * 	@author Chris W. Johnson
 */
public abstract class IntraElementContentMatch extends Match {

	protected final Match		ElementMatch;


	public IntraElementContentMatch
	(
		final Match				ElementMatch
	)
	{
		this.ElementMatch = ElementMatch;
	}
	
	@Override public boolean exists() {
		return ElementMatch.exists();
	}
	
	@Override public Range get() {
		return ElementMatch.get();
	}

	@Override public int getIndex() {
		return ElementMatch.getIndex();
	}

	@Override public Markup getMarkup() {
		return ElementMatch.getMarkup();
	}

	@Override public Namespace getNamespace() {
		return ElementMatch.getNamespace();
	}

	@Override public Match getParent() {
		return ElementMatch;
	}
	
	@Override public boolean equals
	(
		final Object			Other
	)
	{
		if (this == Other)
			return true;
		
		if (Other instanceof IntraElementContentMatch)
			return ElementMatch.equals(((IntraElementContentMatch) Other).ElementMatch);
		
		return false;
	}
	
	@Override public Results get
	(
		final CSSPatterns		Patterns,
		final Results			PrevResults
	)
	{
		return new Results(PrevResults);	//	Return an empty Results instance.
	}

	@Override public Results get
	(
		final String			Patterns
	)
	{
		return new Results();
	}

	@Override public Results get
	(
		final String			Patterns,
		final Results			PrevResults
	)
	{
		return new Results(PrevResults);
	}

	@Override public Results getAncestors
	(
		final boolean			IncludeThis,
		final Results			PrevResults
	)
	{
		final Results			Ancestors = super.getAncestors(true, PrevResults);
		
		if (IncludeThis == false)
			return Ancestors;
		
		return new Results(PrevResults).addMatch(this).addResults(Ancestors);
	}

/**	Returns an empty <code>Results</code> object, because the content of CDATA and comment elements does possess 
 * 	content. (Content <i>is</i> content; it doesn't <i>have</i> content of its own.)
 * 
 * @return		A <code>Results</code> object containin references to all of the children of this markup element.
 */
	@Override public Results getContent() {
		return getContent(null);
	}

/**	Returns an empty <code>Results</code> object, because the content of CDATA and comment elements does possess 
 * 	content. (Content <i>is</i> content; it doesn't <i>have</i> content of its own.)
 * 
 * @param PrevResults	The previous <code>Results</code> object in a call chain. This is the <code>Results</code> 
 * 						object that will be returned when the pop method is invoked on the new <code>Results</code> 
 * 						object returned by this method. 
 * @return				A <code>Results</code> object containin references to all of the children of this markup 
 * 						element.
 */
	@Override public Results getContent
	(
		final Results			PrevResults
	)
	{
		return new Results(PrevResults);
	}

	@Override public double getDouble() throws NumberFormatException {
		return Double.parseDouble(toString().trim());
	}

	@Override public long getLong() throws NumberFormatException {
		return Long.parseLong(toString().trim());
	}

	@Override public String getType() {
		return "";	//	Intra-element content has no type.
	}

	@Override public String toString() {
		return ((Item) get()).getContent().toString();
	}

	@Override public Match print() {
		System.out.println(toString());
		return this;
	}

	@Override public void write
	(
		final PrintWriter		Out
	)
	throws IOException
	{
		((Item) get()).getContent().write(Out);
	}
}
