/*	QWICAP (Quick Web Interface for Conventional Applications) by Chris W. Johnson
	Copyright (C) 2003-2010 The University of Texas at Austin ("U.T. Austin"). All rights reserved.
	
	U.T. Austin's copyright notice in this evolving version reflects our original authorship. It does not assert 
	copyright over changes contributed by others.
	
	This library is free software; you can redistribute it and/or modify it under the terms of the GNU Lesser General 
	Public License as published by the Free Software Foundation; either version 2.1 of the License, or (at your option) 
	any later version.
	
	This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied 
	warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for 
	more details.
	
	You should have received a copy of the GNU Lesser General Public License along with this library; if not, write to 
	the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/


//  Markup.java
//  qwicap
//
//  Created by Chris W. Johnson on Wed Jun 18 2003.


package edu.utexas.its.eis.tools.qwicap.template.xml;


import edu.utexas.its.eis.tools.qwicap.template.css.*;
import edu.utexas.its.eis.tools.qwicap.template.xml.structure.*;
import edu.utexas.its.eis.tools.qwicap.template.xml.mutable.MutableMarkup;
import edu.utexas.its.eis.tools.qwicap.template.xml.immutable.ImmutableMarkup;
import edu.utexas.its.eis.tools.qwicap.util.ALStack;
import edu.utexas.its.eis.tools.qwicap.util.ArrayToString;

import java.util.logging.Logger;
import java.util.logging.Level;
import java.util.Iterator;
import java.io.StringWriter;
import java.io.PrintWriter;
import java.io.IOException;


/**	<code>Markup</code> represents a body of XML markup. The markup represented may be an entire XML document, or 
 *	merely a self-contained fragment. In either case, well-formedness is required to the extent that all "start" tags
 * 	must have matching "end" tags, and the "end" tags must be correctly ordered.
 *
 *	@author Chris W. Johnson
 */
public abstract class Markup implements Cloneable, Iterable<Range> {
	
	private static final Logger		Log = Logger.getLogger(Markup.class.getName());
	
	private String					MarkupName;


/**	Returns the name of the character set in which this markup was encoded when it was read. Once the markup has been
 * 	read into memory, the characters are represented as standard Java (Unicode) characters. When written, they will
 * 	be encoded in the character set in which they were originally encoded.
 * 
 * @return							The name of the character set in which this markup was found to be encoded when it 
 * 									was read.
 */
	public abstract String getCharacterSet();
	
/**	If this markup is already immutable, this method does nothing, and returns a reference to this object. If this
 *	markup is mutable, this method creates and returns an immutable representation of this mutable markup.
 * 
 * @return							An immutable version of this markup.
 * @throws TagException				If an immutable form of this markup could not be made available.
 */	
	public abstract ImmutableMarkup getImmutable() throws TagException;
	
/**	If this markup is already mutable, this method does nothing, and returns a reference to this object. If this markup is
 *  immutable, this method creates and returns a mutable copy of this immutable markup. Initially, all elements in the 
 *	newly created <code>MutableMarkup</code> object are references to the immutable markup elements of the original
 *	<code>ImmutableMarkup</code> object, which makes the creation of the mutable copy of this markup extremely fast 
 *	and memory efficient. Only the elements that are eventually modified are converted into mutable forms, and that 
 *	conversion is performed both automatically, and transparently.
 *
 * @return							A mutable version of this markup.
 * @throws TagException				If a mutable form of this markup could not be made available.
 */
	public abstract MutableMarkup getMutable() throws TagException;
	
/**	Assigns an arbitrary name to this body of markup. This name is purely for your reference, and has no functional 
 *	significance; the only method that accesses it is getMarkupName. For markup loaded from a file, the name is
 *	typically set to the absolute path of the file. For markup loaded from a URL, the name is typically set to the URL.
 * 
 * @param Name						The arbitrary name to be assigned to this body of markup.
 */	
	public final void setMarkupName
	(
		final String		Name
	)
	{
		MarkupName = Name;
	}
	
/**	Returns the name, if any, of this body of markup. The name is arbitrary, but when markup is loaded from a file,
 *	or URL, the name is typically set to the file name, or URL, respectively.
 * 
 * @return							The name assigned to this body of markup, or <code>null</code>.
 */	
	public final String getMarkupName() {
		return MarkupName;
	}
	
/**	<b>This method is not part of the public API</b>, but could not be hidden due to limitations in Java's access 
 * 	control system. It returns the underlying <code>List</code> object that stores all of the elements of this markup.
 *
 * @return 							<code>List</code> object that stores all the elements in this markup. Do not 
 * 									modify this <code>List</code>.
 * @see #get(int Index) 
 */
	public abstract RangeList getList();

/**	Returns a monotonically increasing integer which is incremented every time a modification
 *  is made to the markup list. This allows objects like <code>MutableMatch</code> to detect changes
 *  to the markup list that require them to re-verify the presence and location of the markup
 *  elements they reference. Users of this method should detect changes by testing a cached
 * 	change count value for equality with the current change count value. Inequality means there's
 * 	been a change. No further assumptions about the values should be made.
 *
 * @return An integer that is incremented each time the markup list is altered.
 */
	public abstract int getChangeCount();

/**	Returns the number of elements in this markup.
 *
 * @return							The number of elements in this markup.
 */
	public final int size() {
		return getList().size();
	}

/**	Returns an element from this markup.
 *
 * @param Index						The index number of the desired markup element.
 * @return							The markup element, a subclass of {@link Range}.
 */
	public final Range get
	(
		final int				Index
	)
	{
		return getList().get(Index);
	}
	
/**	Performs a pattern match on this markup using the Cascading Style Sheets level 2 (CSS2) selector scheme, and
 * 	returns a list of all elements that matched the pattern(s). See the
 * 	<a href="http://www.w3.org/TR/REC-CSS2/selector.html">Selectors</a> section of the
 * 	<a href="http://www.w3.org/TR/REC-CSS2/cover.html">Cascading Style Sheets, level 2, CSS2 Specification</a>. Note
 * 	that those basic selectors can be used in complex combinations, and that this method supports all valid
 * 	combinations, including coma-separated lists of independent selectors.
 *
 * @param CSSPatterns				One or more CSS2 selector patterns represented as a <code>String</code>. See 
 *									<a href="http://www.w3.org/TR/REC-CSS2/selector.html#q1">section 5.1</a> of the 
 * 									CSS2 specification for basic examples.
 * @return							A {@link Results} list containing references to each matching element in this 
 * 									markup.
 */
	public final Results get
	(
		final String			CSSPatterns
	)
	{
		return getCSSPatterns(CSSPatterns).search(enumerate());
	}
	
/**	Performs a pattern match on this markup using the Cascading Style Sheets level 2 (CSS2) selector scheme, and
 * 	returns a list of all elements that matched the pattern(s).
 *
 * @param CSSPatts					One or more CSS2 selector pattern represented as a <code>CSSPatterns</code> 
 * 									instance.
 * @return							A {@link Results} list containing references to each matching element in this 
 * 									markup.
 */
	public final Results get
	(
		final CSSPatterns		CSSPatts
	)
	{
		return CSSPatts.search(enumerate());
	}
	
/**	Returns a new <code>Results</code> object listing all of the declaration (<samp>&lt;!</samp> &hellip; <samp>&gt;</samp>) 
 * 	elements in this markup.
 * 
 * @return							A <code>Results</code> object containing matches for declaration elements.
 */
	public Results getDeclarations() {
		final Results			Comments  = new Results();
		final RangeList			List      = getList();
		final Range[]			ListArray = List.Array;
		
		for (int Index = 0, Count = List.size(); Index < Count; Index++) {
			final Range				Thing = ListArray[Index];
			
			if (Thing instanceof Declaration)
				Comments.addMatch(createMatch(Thing, Index));
		}
		
		return Comments;
	}
	
/**	Returns a new <code>Results</code> object listing all of the comment (<samp>&lt;!--</samp> &hellip; <samp>--&gt;</samp>) 
 * 	elements in this markup.
 * 
 * @return							A <code>Results</code> object containing matches for comment elements.
 */
	public Results getComments() {
		final Results			Comments  = new Results();
		final RangeList			List      = getList();
		final Range[]			ListArray = List.Array;
		
		for (int Index = 0, Count = List.size(); Index < Count; Index++) {
			final Range				Thing = ListArray[Index];
			
			if (Thing instanceof Comment)
				Comments.addMatch(createMatch(Thing, Index));
		}
		
		return Comments;
	}
	
/**	Returns a new <code>Results</code> object listing all of the CDATA (<samp>&lt;![CDATA[</samp> &hellip; <samp>]]&gt;</samp>) 
 * 	elements in this markup.
 * 
 * @return							A <code>Results</code> object containing matches for CDATA elements.
 */
	public Results getCDATA() {
		final Results			CDATAs    = new Results();
		final RangeList			List      = getList();
		final Range[]			ListArray = List.Array;
		
		for (int Index = 0, Count = List.size(); Index < Count; Index++) {
			final Range				Thing = ListArray[Index];
			
			if (Thing instanceof CDATA)
				CDATAs.addMatch(createMatch(Thing, Index));
		}
		
		return CDATAs;
	}
	
/**	Returns a <code>Results</code> object containing a reference to the first start tag in the document.	
 * 
 * @return							The first start tag in the document.
 */	
	public final Results first() {
		final Results			First     = new Results();
		final RangeList			List      = getList();
		final Range[]			ListArray = List.Array;
		
		for (int Index = 0, Count = List.size(); Index < Count; Index++) {
			final Range				Thing = ListArray[Index];
			
			if (Thing.isStartTag()) {
				
				First.addMatch(createMatch(Thing, Index));
				break;
			}
		}
		
		return First;
	}
	
/**	Retrieves a compiled version of the specified CSS pattern. Where possible, a previously compiled version of the 
 *	pattern is retreived from an internal cache, to eliminate the expense of parsing the pattern and allocating the
 *	objects that describe it.
 * 
 * @param PatsStr		A CSS2 selector pattern string. See <a href="http://www.w3.org/TR/REC-CSS2/selector.html#q1">
 * 						section 5.1</a> of the CSS2 specification for basic examples.
 * @return				A {@link CSSPatterns} object that implements the request search pattern.
 */
	public abstract CSSPatterns getCSSPatterns(String PatsStr);
	
	public abstract Match createMatch(Range TheElement, int IndexOfElementInThisMarkup);
		
/**	Returns <code>true</code> if this markup contains no elements.
 *
 * @return	<code>true</code> if this markup contains no elements, <code>false</code> otherwise.
 */
	public final boolean isEmpty() {
		return getList().isEmpty();
	}

/**	Returns an enumeration for the elements in this markup.
 *
 * @return	A {@link MarkupEnumeration} for the elements in this markup.
 */
	public final MarkupEnumeration enumerate() {
		return new RangeListEnumerator(this);
	}
	
/**	Returns an <code>Iterator</code> for the elements in this markup.
 *
 * @return	An iterator for the <code>Range</code> objects that comprise this markup.
 */
	@Override public Iterator<Range> iterator() {
		return enumerate();
	}

/**	Performs a simple verification of the tag hierarchy in the XHTML markup represented by this object. It does
 *  nothing more than verify that every start-tag has a matching end-tag. This tells you that the XML is more-or-less
 *  correctly formed, which is sufficient to make this templating package happy, but does nothing more to verify
 * 	well-formedness.
 *
 * @throws TagException		If that tag hierarchy is incomplete or corrupt.
 */
 	public void checkHierarchy() throws TagException {
		final TagHierarchyVerifier	Hierarchy = new TagHierarchyVerifier();
		final RangeList				List      = getList();
		final Range[]				ListArray = List.Array;
		
		for (int Index = 0, Count = List.size(); Index < Count; Index++) {
			final Range				Thing = ListArray[Index];
			
			if (Thing.isStartTag()) {
				
				Hierarchy.startTag((StartTag) Thing);
				
			} else if (Thing.isEndTag()) {
				
				Hierarchy.endTag((EndTag) Thing);
			}
		}
		
		Hierarchy.done();
	}

/**	Returns a <code>String</code> representing all the markup stored in this object.
 *
 * @return	A <code>String</code> representing all the markup stored in this object.
 */
	@Override public String toString() {
		final StringWriter		SOut = new StringWriter(size() * 16);	//	We guess that items in this markup will average around 16 characters in length.
		final PrintWriter		Out  = new PrintWriter(SOut);
		
		try {
			write(Out);
		} catch (IOException e) {
			Out.println();
			Out.print(e.getClass().getName());
			Out.print(": ");
			Out.println(e.getMessage());
		}
		
		Out.close();
		
		return SOut.toString();
	}
	
/**	Prints this markup to <code>System.out</code> for convenience in debugging.
 * 
 * @throws IOException			If there was a problem writing to <code>System.out</code>.
 * @see #write
 */	
	public void print() throws IOException {
		final PrintWriter		Out = new PrintWriter(System.out);
		
		write(Out);
		Out.flush();
	}

/**	Writes all of the markup stored in this object to a <code>PrintWriter</code>.
 *
 * @param Out					The <code>PrintWriter</code> to which this markup will be written.
 * @return						The <code>PrintWriter</code> that was received as the <code>Out</code> parameter.
 * @throws IOException			If the supplied <code>PrintWriter</code> has any problem writing.
 */
	public PrintWriter write
	(
		final PrintWriter		Out
	)
	throws IOException
	{
		final long				StartMillis = System.currentTimeMillis();
		final MarkupEnumeration	Elems       = new RangeListEnumerator(this);
		
		while (Elems.hasMoreElements())
			Elems.nextRange().write(Out);
		
		if (Log.isLoggable(Level.FINEST))
			Log.log(Level.FINEST, "Markup written in {0} ms.", System.currentTimeMillis() - StartMillis);
		
		return Out;
	}
	
	private static final class TagHierarchyVerifier {
		
		private final ALStack<StartTag>	StartTags = new ALStack<StartTag>();	//	Contains StartTag objects.
		
		
		TagHierarchyVerifier() { }
		
		void done() throws TagException {
			
		//	The item pushed by the constructor (to represent the XML namespace) will always be present, but there 
		//	shouldn't be anything else.
			
			if (StartTags.isEmpty())
				return;
			
			throw new TagException("XML Hierarchy Error: The following tags (" + StartTags.size() + ") are not terminated: " + new ArrayToString(StartTags));
		}
		
		void startTag
		(
			final StartTag			Tag
		)
		{
			StartTags.push(Tag);
		}
		
		void endTag
		(
			final EndTag			Tag
		)
		throws TagException
		{
			
		//	Verify tag hierarchy.
			
			if (StartTags.isEmpty())
				throw new TagException("XML Hierarchy Error: Isolated end tag \"" + Tag + "\".");
			
			final StartTag			StartTag = StartTags.pop();
			
			if (StartTag.nameEquals(Tag) == false)
				throw new TagException("XML Hierarchy Error: Found end tag \"" + Tag + "\" where there should have been an end tag for \"" + StartTag + "\".");
		}
	}
}

