/*	QWICAP (Quick Web Interface for Conventional Applications) by Chris W. Johnson
	Copyright (C) 2003-2011 The University of Texas at Austin ("U.T. Austin"). All rights reserved.
	
	U.T. Austin's copyright notice in this evolving version reflects our original authorship. It does not assert 
	copyright over changes contributed by others.
	
	This library is free software; you can redistribute it and/or modify it under the terms of the GNU Lesser General 
	Public License as published by the Free Software Foundation; either version 2.1 of the License, or (at your option) 
	any later version.
	
	This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied 
	warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for 
	more details.
	
	You should have received a copy of the GNU Lesser General Public License along with this library; if not, write to 
	the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/


//  Match.java
//  qwicap
//
//  Created by Chris W. Johnson on Wed Jun 18 2003.


package edu.utexas.its.eis.tools.qwicap.template.xml;


import edu.utexas.its.eis.tools.qwicap.template.css.CSSPatterns;
import edu.utexas.its.eis.tools.qwicap.template.css.Results;
import edu.utexas.its.eis.tools.qwicap.template.xml.immutable.ImContent;
import edu.utexas.its.eis.tools.qwicap.template.xml.immutable.ImmutableMarkup;
import edu.utexas.its.eis.tools.qwicap.template.xml.immutable.ImmutableMatch;
import edu.utexas.its.eis.tools.qwicap.template.xml.mutable.AttributeUtilities;
import edu.utexas.its.eis.tools.qwicap.template.xml.mutable.MutableMarkup;
import edu.utexas.its.eis.tools.qwicap.template.xml.mutable.MutableMatch;
import edu.utexas.its.eis.tools.qwicap.template.xml.structure.*;
import edu.utexas.its.eis.tools.qwicap.util.Characters;
import edu.utexas.its.eis.tools.qwicap.servlet.Form;
import edu.utexas.its.eis.tools.qwicap.servlet.FormNotFoundException;
import edu.utexas.its.eis.tools.qwicap.servlet.FormDataSet;

import java.io.IOException;
import java.io.PrintWriter;
import java.io.StringWriter;
import java.util.NoSuchElementException;
import java.util.StringTokenizer;
import java.util.Iterator;


/**	The <code>Match</code> class uniquely identifies a particular element within a body of XHTML markup, and provides 
 *	methods for performing every type of operation possible on any element. (Note that not all operations make sense 
 *	for all elements.) It is created to represent an element selected by a 
 * 	<a href='http://www.w3.org/TR/CSS21/selector.html#q1'>CSS2 pattern</a> applied to a body of markup by the 
 * 	{@link CSSPatterns} class, which returns the zero, or more, matches produced by its search in a {@link Results} 
 * 	object.
 *
 *	<p>The <code>Match</code> class is abstract. Separate implementations handle the simple case of elements in 
 *	immutable markup (<code>ImmutableMatch</code>) where none of the modification functions do anything but throw 
 * 	exceptions, and the complex case of elements in mutable markup (<code>MutableMatch</code>) where the modification 
 * 	functions actually work.
 *	</p>	
 * 
 *	@author Chris W. Johnson
 */
public abstract class Match implements Iterable<Range> {
	
/**	Creates the type of <code>Match</code> object (<code>ImmutableMatch</code>} or {<code>MutableMatch</code>) 
 * 	appropriate to the type of <code>Markup</code> object ({@link ImmutableMarkup} or {@link MutableMarkup}) in which 
 * 	the specified element occurs.
 *
 * @param Markup						The <code>Markup</code> object in which the specified element is stored.
 * @param Item							The markup element to be represented by this class.
 * @param Index							The index of the markup element within the <code>Markup</code> object.
 * @return								A <code>Match</code> subclass uniquely identifying the specified markup element.
 */
	public static Match create
	(
		final Markup			Markup,
		final Range				Item,
		final int				Index
	)
	{
		return Markup.createMatch(Item, Index);
	}
	
	protected Match() { }

/**	Returns the <code>Markup</code> object in which this match occurs. This will always be the same <code>Markup</code> 
 * 	object that was passed to the constructor.
 *
 * @return								The <code>Markup</code> object in which this <code>Match</code> occurs.
 */
	public abstract Markup	getMarkup();

/**	The index of this element in the markup that contains it. When the containing markup is immutable, this index is
 * 	a constant, but when the markup is mutable, the index can change. In the latter case, this method verifies the
 * 	index of the element, and corrects it if there's been a change.
 *
 * @return								The index of this element within the markup.
 * @throws NoSuchElementException		If this element is no longer present in the markup.
 */
	public abstract int		getIndex() throws NoSuchElementException;

/**	Returns the markup element represented by this <code>Match</code> object.
 *
 * @return								The markup element represented by this <code>Match</code> object.
 * @throws NoSuchElementException		If this element is no longer present in the markup.
 */
	public abstract Range	get() throws NoSuchElementException;
	
/**	Creates a {@link Results} object containing only this <code>Match</code>.
 * 
 * @return								A <code>Results</code> object containing only this <code>Match</code>.
 */
	public Results toResults() {
		return new Results().addMatch(this);
	}

/**	Returns a <code>Form</code> object for examining and/or manipulating the XHTML "<samp>form</samp>" element (and the 
 * 	controls within it) that is referenced by the first match in this <code>Results</code> object. If you intend to 
 * 	modify the "<samp>form</samp>" element, or its contents, you should already be operating on {@link MutableMarkup}
 * 	before invoking <code>toForm</code>. A mutable version of <code>ImmutableMarkup</code> is created trivially by
 * 	merely invoking the {@link Markup#getMutable()} method.
 * 
 * @return								A <code>Form</code> object for manipulating the "<samp>form</samp>" element 
 *										referenced by the first match in this <code>Results</code> object.
 * @throws FormNotFoundException		If the this <code>Match</code> instances does not refer to a "<samp>form</samp>" 
 * 										element.
 * @see MutableMarkup#getForm(FormDataSet)
 * @see Results#toForm
 */
	public Form toForm() throws FormNotFoundException {

		if ("form".equals(getType()))
			return new Form(this);
		
		throw new FormNotFoundException("The toForm method can only be invoked on a Match object describing a \"form\" element. This Match object refers to: " + this);
	}
	
/**	Makes a copy of this markup element, and any children it may have, then inserts the copy immediately before, or 
 * 	after, this element.
 *
 * @param InsertDuplicateBefore			<code>true</code> if the duplicate should be inserted before the item that was 
 *										duplicated, <code>false</code> if it should be inserted after.
 * @return								A <code>Match</code> object identifying the duplicate markup element.
 * @throws TagException					If a problem is encountered with the manipulation of the markup.
 */
	public Match duplicate
	(
		final boolean			InsertDuplicateBefore
	)
	throws TagException
	{
		throw new TagException(notSupportedMessage());
	}
	
/**	Removes this element from the markup, and inserts the supplied replacement object in its place. Refer to the
 * 	documentation for {@link MutableMarkup#insert MutableMarkup.insert} for an explanation of how the Replacement 
 * 	object is handled.
 *
 * @param Replacement					The material that will replace this element in the markup. See 
 * 										{@link MutableMarkup#insert MutableMarkup.insert} for details.
 * @return								A <code>Match</code> object identifying the material that replaced this element.
 * @throws TagException					If a problem is encountered with the manipulation of the markup.
 */
	public Match replace
	(
		final Object			Replacement
	)
	throws TagException
	{
		throw new TagException(notSupportedMessage());
	}

/**	Adds the supplied material before, or after, the element specified by this <code>Match</code> object. If the element
 * 	specified by this <code>Match</code> object has children, and the new material is to be inserted after this element, 
 * 	the new material is inserted immediately after this element's end-tag (for example: "&lt;/a&gt;", "&lt;/div&gt;" or
 * 	"&lt;/p&gt;").
 *
 * @param AddBefore						<code>true</code> if the new material should be inserted before this element, 
 *										<code>false</code> if it should be inserted after it.
 * @param NewStuff						The material that will be inserted before/after this element in the markup. See
 * 										{@link MutableMarkup#insert MutableMarkup.insert} for details.
 * @return								A <code>Match</code> object identifying the inserted material.
 * @throws TagException					If a problem is encountered with the manipulation of the markup.
 */
	public Match add
	(
		final boolean			AddBefore,
		final Object			NewStuff
	)
	throws TagException
	{
		throw new TagException(notSupportedMessage());
	}
	
/**	Inserts the supplied material before, or after, the element specified by this <code>Match</code> object. If the 
 * 	element specified by this <code>Match</code> object has children, and the new material is to be inserted after this 
 * 	element, the new material is inserted immediately after this element's end-tag (for example: "&lt;/a&gt;", 
 * 	"&lt;/div&gt;" or "&lt;/p&gt;").
 *
 * @param InsertBefore					<code>true</code> if the new material should be inserted before this element, 
 *										<code>false</code> if it should be inserted after it.
 * @param NewStuff						The material that will be inserted before/after this element in the markup. See
 * 										{@link MutableMarkup#insert MutableMarkup.insert} for details.
 * @return								A <code>Match</code> object identifying the material that replaced this element.
 * @throws TagException					If a problem is encountered with the manipulation of the markup.
 * @deprecated							Use {@link #add} instead.
 */
	public Match insert
	(
		final boolean			InsertBefore,
		final Object			NewStuff
	)
	throws TagException
	{
		throw new TagException(notSupportedMessage());
	}

/**	Deletes this markup element and any children it may have.
 *
 * @throws TagException					If a problem is encountered with the manipulation of the markup.
 */
	public void delete() throws TagException {
		throw new TagException(notSupportedMessage());
	}

/**	Returns <code>true</code> if the markup element referenced by this <code>Match</code> still exists, 
 *	 <code>false</code> otherwise. This is only relevant when the markup containing this <code>Match</code> is mutable.
 * 
 * @return								<code>true</code> if the markup element referenced by this <code>Match</code> 
 * 										still exists, <code>false</code> otherwise.
 */
	public abstract boolean exists();
	
/**	Returns the type of the XML element referenced by this match.
 * 
 * @return								The type of the XML element, or an empty <code>String</code> if this match 
 * 										references something other than an element.
 */
	public String getType() {
		final Range				R = get();
		
	//	todo Consider: We could eliminate all of this conditional logic by adding a getType method to the Range interface.
		
		if (R.isTagWithAttributes() || R.isEndTag())
			return ((Tag) R).getNameString();
		else if (R instanceof CDATA)
			return "!CDATA";
		else if (R instanceof Comment)
			return "!--";
		else if (R instanceof Declaration)
			return "!" + ((Declaration) R).getName();
		else
			return "";
	}
	
/**	Returns the namespace of the XML element referenced by this match.	
 * 
 * @return								The namespace of the XML element, or <code>null</code> if no namespace.
 */	
	public Namespace getNamespace() {
		Namespace				Ns = null;
		final Range				R  = get();
		
		if (R.isTagWithAttributes() || R.isEndTag())
			Ns = ((Tag) R).getNamespace();
		
		return Ns == Namespace.kNoNamespace ? null : Ns;
	}
	
/**	Invokes {@link #getText} and converts the <code>String</code> it returns into a	<code>double</code>, if possible.
 * 
 * @return								The numeric value represented by the text of the element, as a 
 * 										<code>double</code>.
 * @throws NumberFormatException		If the text of this element couldn't be interpreted as a floating-point number.
 * @see Double#parseDouble
 */	
	public double getDouble() throws NumberFormatException {
		return Double.parseDouble(getText());		//	parseDouble ignores leading and trailing whitespace.
	}
	
/**	Invokes {@link #getText} and converts the <code>String</code> it returns into a	<code>long</code>, if possible.
 * 	
 * @return								The numeric value represented by the text of the element, as a 
 * 										<code>long</code>.
 * @throws NumberFormatException		If the text of this element couldn't be interpreted as an integer number.
 * @see Long#parseLong
 */	
	public long getLong() throws NumberFormatException {
		return Long.parseLong(getText().trim());	//	parseLong does NOT ignore leading or trailing whitespace.
	}
	
/**	Returns the human-readable/-visible text, if any, that this element would contribute to the final, rendered 
 * 	document.
 * 
 * @return								A <code>String</code> containing the text of the element, or an empty string if 
 * 										no such text was present.
 */	
 	public String getText() {
		return getText(new StringBuilder(48)).toString();
	}
	
/**	Returns the human-readable/-visible text, if any, that this element would contribute to the final, rendered 
 * 	document.
 * 
 * @param Buff							<code>StringBuffer</code> to which the text will be appended.
 * @return								The <code>StringBuffer</code> that was passed-in as the Buff parameter.
 */	
	public StringBuffer getText
	(
		final StringBuffer		Buff
	) 
	{
		final MarkupEnumeration	Contents = enumerate(false);
	
		while (Contents.hasMoreElements()) {
			final Range				Element = Contents.nextRange();
		
		//	todo Note that including all content will include the text inside things like "option" tags (e.g. "<option value='.2'>20%</option>") that, in the case of pop-up menus, mostly won't be visible to the user.
			
			if (Element instanceof Content) {
			
				if (Element instanceof ImContent) {
					final ImContent			IC = (ImContent) Element;
				
					if (IC.HadLeadingWhitespace)
						appendAtMostOneSpace(Buff);
				
					MutableMarkup.htmlDecode(((Content) Element).getContent(), Buff);
				
					if (IC.HadTrailingWhitespace)
						appendAtMostOneSpace(Buff);
				
				} else {
				
				//	The getContent method returns the content stripped of leading and trailing whitespace. This might be a problem in some cases.
			
					MutableMarkup.htmlDecode(((Content) Element).getContent(), Buff);	
				}
			
			} else if (Element instanceof Whitespace) {
			
				appendAtMostOneSpace(Buff);	//	HTML/XHTML rendering is indifferent to the amount of whitespace, and so are we.
			}
		
		//	@todo Consider: What other elements can contribute visible text to the final, rendered document? 
		//	The "alt" attributes of "img" tags? Some of the form elements, perhaps? And do we care?
		}
	
	//	We will have left a trailing space on the end of the text, if any. Remove that space.
	
		final int				Length = Buff.length();
	
		if (Length > 0 && Buff.charAt(Length - 1) == ' ')
			Buff.setLength(Length - 1);
	
	//	Done.
	
		return Buff;
	}
	
/**	Returns the human-readable/-visible text, if any, that this element would contribute to the final, rendered 
 * 	document.
 * 
 * @param Buff							<code>StringBuilder</code> to which the text will be appended.
 * @return								The <code>StringBuilder</code> that was passed-in as the Buff parameter.
 */	
	public StringBuilder getText
	(
		final StringBuilder		Buff
	) 
	{
		final MarkupEnumeration	Contents = enumerate(false);
	
		while (Contents.hasMoreElements()) {
			final Range				Element = Contents.nextRange();

			if (Element instanceof Content) {
			
				if (Element instanceof ImContent) {
					final ImContent			IC = (ImContent) Element;
				
					if (IC.HadLeadingWhitespace)
						appendAtMostOneSpace(Buff);
				
					MutableMarkup.htmlDecode(((Content) Element).getContent(), Buff);
				
					if (IC.HadTrailingWhitespace)
						appendAtMostOneSpace(Buff);
				
				} else {
				
				//	The getContent method returns the content stripped of leading and trailing whitespace. This might be a problem in some cases.
			
					MutableMarkup.htmlDecode(((Content) Element).getContent(), Buff);	
				}
			
			} else if (Element instanceof Whitespace) {
			
				appendAtMostOneSpace(Buff);	//	HTML/XHTML rendering is indifferent to the amount of whitespace, and so are we.
			}
		
		//	@todo Consider: What other elements can contribute visible text to the final, rendered document? 
		//	The "alt" attributes of "img" tags? Some of the form elements, perhaps? And do we care?
		}
	
	//	We will have left a trailing space on the end of the text, if any. Remove that space.
	
		final int				Length = Buff.length();
	
		if (Length > 0 && Buff.charAt(Length - 1) == ' ')
			Buff.setLength(Length - 1);
	
	//	Done.
	
		return Buff;
	}
	
	private static void appendAtMostOneSpace
	(
		final StringBuffer		Buff
	)
	{
		final int				Length = Buff.length();
		
		if (Length > 0) {
			
			if (Character.isWhitespace(Buff.charAt(Length - 1)) == false)
				Buff.append(' ');
		}
	}
	
	private static void appendAtMostOneSpace
	(
		final StringBuilder		Buff
	)
	{
		final int				Length = Buff.length();
		
		if (Length > 0) {
			
			if (Character.isWhitespace(Buff.charAt(Length - 1)) == false)
				Buff.append(' ');
		}
	}
	
/**	Returns a new <code>Results</code> object listing all of the CDATA (<samp>&lt;![CDATA[</samp> &hellip; <samp>]]&gt;</samp>) 
 * 	elements in this markup.
 * 
 * @return								A <code>Results</code> object containing matches for CDATA sections.
 */
	public Results getCDATA() {
		return getCDATA(null);
	}
	
/**	Returns a new <code>Results</code> object listing all of the CDATA (<samp>&lt;![CDATA[</samp> &hellip; <samp>]]&gt;</samp>) 
 * 	elements in this markup.
 * 
 * @param PrevResults					The previous <code>Results</code> object in a call chain. This is the 
 * 										<code>Results</code> object that will be returned when the <code>pop</code> 
 * 										method is invoked on the new <code>Results</code> object returned by this 
 * 										method. 
 * @return								A <code>Results</code> object containing matches for CDATA sections.
 */
	public Results getCDATA
	(
		final Results			PrevResults
	) 
	{
		final Results			NewResults = new Results(PrevResults);
		final MarkupEnumeration	Contents   = enumerate(false);
	
		while (Contents.hasMoreElements()) {
			
			if (Contents.nextRange() instanceof CDATA)
				NewResults.addMatch(Contents.getMatch());
		}
		
		return NewResults;
	}
	
/**	Returns a new <code>Results</code> object listing all of the comment (<samp>&lt;!--</samp> &hellip; <samp>--&gt;</samp>) 
 * 	elements in this markup.
 * 
 * @return								A <code>Results</code> object containing matches for comment elements.
 */
	public Results getComments() {
		return getComments(null);
	}
		
/**	Returns a new <code>Results</code> object listing all of the comment (<samp>&lt;!--</samp> &hellip; <samp>--&gt;</samp>) 
 * 	elements in this markup.
 * 
 * @param PrevResults					The previous <code>Results</code> object in a call chain. This is the 
 * 										<code>Results</code> object that will be returned when the <code>pop</code> 
 * 										method is invoked on the new <code>Results</code> object returned by this 
 * 										method. 
 * @return								A <code>Results</code> object containing matches for comment elements.
 */
	public Results getComments
	(
		final Results			PrevResults
	) 
	{
		final Results			NewResults = new Results(PrevResults);
		final MarkupEnumeration	Contents   = enumerate(false);
	
		while (Contents.hasMoreElements()) {
			
			if (Contents.nextRange() instanceof Comment)
				NewResults.addMatch(Contents.getMatch());
		}
		
		return NewResults;
	}
	
/**	Returns a new <code>Results</code> object listing all of the declaration (<samp>&lt;!</samp> &hellip; <samp>&gt;</samp>) 
 * 	elements in this markup.
 * 
 * @return								A <code>Results</code> object containing matches for declaration elements.
 */
	public Results getDeclarations() {
		return getDeclarations(null);
	}
		
/**	Returns a new <code>Results</code> object listing all of the declaration (<samp>&lt;!</samp> &hellip; <samp>&gt;</samp>) 
 * 	elements in this markup.
 * 
 * @param PrevResults					The previous <code>Results</code> object in a call chain. This is the 
 * 										<code>Results</code> object that will be returned when the <code>pop</code> 
 * 										method is invoked on the new <code>Results</code> object returned by this 
 * 										method. 
 * @return								A <code>Results</code> object containing matches for declaration elements.
 */
	public Results getDeclarations
	(
		final Results			PrevResults
	) 
	{
		final Results			NewResults = new Results(PrevResults);
		final MarkupEnumeration	Contents   = enumerate(false);
	
		while (Contents.hasMoreElements()) {
			
			if (Contents.nextRange() instanceof Declaration)
				NewResults.addMatch(Contents.getMatch());
		}
		
		return NewResults;
	}
	
/**	Return references to all of the children (but not grandchildren, etc.) of this markup element. 
 * 
 * @return								A <code>Results</code> object containin references to all of the children of 
 * 										this markup element.
 */
	public Results getContent() {
		return getContent(null);
	}
	
/**	Return references to all of the children (but not grandchildren, etc.) of this markup element. 
 * 
 * @param PrevResults					The previous <code>Results</code> object in a call chain. This is the 
 * 										<code>Results</code> object that will be returned when the <code>pop</code> 
 * 										method is invoked on the new <code>Results</code> object returned by this 
 * 										method. 
 * @return								A <code>Results</code> object containin references to all of the children of 
 * 										this markup element.
 */
	public Results getContent
	(
		final Results				PrevResults
	)
	{
		final Results				Content  = new Results(PrevResults);
		final DescendantEnumerator	Contents = new DescendantEnumerator(enumerate());
		
		while (Contents.hasMoreElements()) {
			
			Contents.nextRange();
			
		//	We don't want level 0 elements, because they'll be this tag (and its end tag), and we're after the 
		//	*contents* of this tag. Also, we don't want elements with more than 1 parent because, by convention, 
		//	parent tags and their children aren't listed in the same Results object. (The convention states that
		//	operations on a parent implicitly apply to its children, where appropriate.)
			
			if (Contents.getParentCount() == 1)
				Content.addMatch(Contents.getMatch());
		}
		
		return Content;
	}
	
/**	Inserts the supplied material into the content of this markup element, provided that this element is a start-tag,
 * 	since no other elements can contain content (children). The supplied material may be inserted before the first
 * 	child, or after the last.
 *
 * @param InsertBeforeExistingContent	<code>true</code> if the new material should be inserted before the first child 
 *										of this element, <code>false</code> if it should be inserted after the final 
 *										child.
 * @param Content						The material that will be inserted before/after this element's content. See
 * 										{@link MutableMarkup#insert MutableMarkup.insert} for details.
 * @return								A <code>Match</code> object identifying the inserted material.
 * @throws TagException					If a problem is encountered with the manipulation of the markup.
 */
	public Match addContent
	(
		final boolean			InsertBeforeExistingContent,	//	If false, we insert after the existing content.
		final Object			Content
	)
	throws TagException
	{
		throw new TagException(notSupportedMessage());
	}

/**	Deletes the content of this markup element, provided that this element is a start-tag, since no other elements
 * 	can contain content (children).
 *
 * @return								A <code>Match</code> object identifying this markup element. (Not the deleted 
 * 										material.)
 * @throws TagException					If a problem is encountered with the manipulation of the markup.
 */
	public Match deleteContent() throws TagException {
		throw new TagException(notSupportedMessage());
	}

/**	Replaces any existing contents of this markup element with the supplied material.
 *
 * @param Content						The material that will replace this element's content. See
 * 										{@link MutableMarkup#insert MutableMarkup.insert} for details.
 * @return								A <code>Match</code> object identifying this markup element. (Not the inserted 
 * 										material.)
 * @throws TagException					If a problem is encountered with the manipulation of the markup.
 */
	public Match setContent
	(
		final Object			Content
	)
	throws TagException
	{
		throw new TagException(notSupportedMessage());
	}
	
/**	Tests the "class" attribute of this element, if it has one, for the presence of the named class or classes.
 * 
 * @param ClassList						The name of a class, or a space-separated list of class names.
 * @return								<code>true</code> if all of the named classes are present in the "class" 
 * 										attribute of this element, <code>false</code> otherwise.
 * @throws TagException					If a problem is encountered with the manipulation of the markup.
 */
	public boolean hasClass
	(
		final String			ClassList
	)
	throws TagException
	{
		final Range				R = get();
		
		if (R.isTagWithAttributes() == false)
			return false;
		
		final TagWithAttributes	Tag = (TagWithAttributes) R;
		
	//	If this is a list of classes (space separated), every listed class must be present. 
		
		if (ClassList.indexOf(' ') >= 0) {
			final StringTokenizer	ST = new StringTokenizer(ClassList, " ");
			
			while (ST.hasMoreTokens()) {
				
				if (Tag.isOfClass(new Characters(ST.nextToken())) == false)
					return false;
			}
			
			return true;
			
		} else
			return Tag.isOfClass(new Characters(ClassList));
	}
	
/**	Adds one or more class names to the "class" attribute of this markup element. If this markup element has no
 * 	"class" attribute, one is created. Redundant additions of class names are detected and omitted, so duplicates will
 * 	not be created. Calling {@link #addAttribute addAttribute} and supplying "class" as the attribute name has exactly 
 * 	the same effect as this method, but is less efficient.
 *
 * @param ClassList						A space-separated list of class names.
 * @return								A <code>Match</code> object identifying this markup element. If this element 
 * 										was immutable, a mutable version of it will have been created, substitued into 
 * 										this markup (assuming the markup itself is mutable), and the <code>Match</code>
 * 										returned will refer to the new mutable form of this element.
 * @throws TagException					If a problem is encountered with the manipulation of the markup.
 */
	public Match addClass
	(
		final String			ClassList
	)
	throws TagException
	{
		throw new TagException(notSupportedMessage());
	}

/**	Replaces any existing value of the "class" attribute of this markup element with the supplied space-separated list
 * 	of class names. Redundant additions of class names are detected and omitted, so duplicates will not be created.
 * 	Calling {@link #setAttribute setAttribute} and supplying "class" as the attribute name has exactly the same effect 
 * 	as this method, but is less efficient.
 *
 * @param ClassList						A space-separated list of class names.
 * @return								A <code>Match</code> object identifying this markup element. If this element was 
 * 										immutable, a mutable version of it will have been created, substitued into this 
 * 										markup (assuming the markup itself is mutable), and the <code>Match</code> 
 * 										returned will refer to the new mutable form of this element.
 * @throws TagException					If a problem is encountered with the manipulation of the markup.
 */
	public Match setClass
	(
		final String			ClassList
	)
	throws TagException
	{
		throw new TagException(notSupportedMessage());
	}

/**	Deletes the supplied space-separated list of class names from the value of this markup element's "class" attribute.
 * 	Deleting class names that aren't actually present in the value of this markup element's "class" attribute is no
 * 	problem. If you want to remove all class names from this element, use {@link #deleteAttribute deleteAttribute} and 
 * 	pass "class" as the attribute name.
 *
 * @param ClassList						A space-separated list of class names.
 * @return								A <code>Match</code> object identifying this markup element. If this element was 
 * 										immutable, a mutable version of it will have been created, substitued into this 
 * 										markup (assuming the markup itself is mutable), and the <code>Match</code> 
 * 										returned will refer to the new mutable form of this element.
 * @throws TagException					If a problem is encountered with the manipulation of the markup.
 */
	public Match deleteClass
	(
		final String			ClassList
	)
	throws TagException
	{
		throw new TagException(notSupportedMessage());
	}
	
/**	Returns <code>true</code> if this markup element has an attribute of the specified name and value.	
 * 
 * @param AttrName						Name of the attribute.
 * @param AttrValue						Value of the attribute.
 * @return								<code>true</code> if the attribute is present and is of the specified value, 
 *										<code>false</code> otherwise.
 */	
	public boolean hasAttributeOfValue
	(
		final String			AttrName,
		final String			AttrValue
	)
	{
		final Range				Tag = get();
		
		if (Tag.isTagWithAttributes()) {
			final String			Value = ((TagWithAttributes) Tag).getAttribute(AttributeUtilities.encodeAttributeName(AttrName));
			
			if (Value != null)
				return Value.equals(AttrValue);
		}
		
		return false;
	}
	
/**	Returns <code>true</code> if this markup element includes the specified attribute.
 * 
 * @param AttrName						Name of the attribute.
 * @return								<code>true</code> if the attribute is present, <code>false</code> otherwise.
 */
	public boolean hasAttribute
	(
		final String			AttrName
	)
	{
		final Range				Tag = get();
		
		if (Tag.isTagWithAttributes())
			return ((TagWithAttributes) Tag).getAttribute(AttributeUtilities.encodeAttributeName(AttrName)) != null;
		
		return false;
	}
	
/**	Returns <code>true</code> if this markup element includes the specified attribute, and the value of the attribute
 * 	is of a non-zero length. Otherwise, this method returns <code>false</code>.	
 * 
 * @param AttrName						Name of the attribute.
 * @return								<code>true</code> if this markup element includes the specified attribute, and 
 * 										the value of the attribute is of a non-zero length, <code>false</code> 
 * 										otherwise.
 */	
	public boolean hasNonEmptyAttribute
	(
		final String			AttrName
	)
	{
		final Range				Tag = get();
		
		if (Tag.isTagWithAttributes()) {
			final String			Value = ((TagWithAttributes) Tag).getAttribute(AttributeUtilities.encodeAttributeName(AttrName));
			
			if (Value != null)
				return Value.length() > 0;
		}
		
		return false;
	}
	
/**	Returns the value of the named attribute, or <code>null</code> if there is no such attribute in this markup element.	
 * 
 * @param AttrName						Name of the attribute.
 * @return								Value of the attribute, or <code>null</code> if no such attribute.
 */	
	public String getAttribute
	(
		final String			AttrName
	)
	{
		String					Value = null;
		final Range				Tag   = get();
		
		if (Tag.isTagWithAttributes()) {
			
			Value = ((TagWithAttributes) Tag).getAttribute(AttributeUtilities.encodeAttributeName(AttrName));
			if (Value != null)
				Value = AttributeUtilities.decodeAttributeValue(Value);
		}
		
		return Value;
	}
	
/**	Adds an attribute to any markup element that is allowed to have attributes. If the markup element already has an
 * 	attribute of the specified name, the new value is appended to the existing value. Note that nothing clever is done
 * 	when appending, such as placing spaces between the original value and the new material that is appended to it.
 * 	Instead, this is a plain and simple append operation. Because of that, it is possible to use this method to, for
 * 	instance, progressively build-up a URL in an "a" (anchor) element's "href" attribute. When appending is not
 * 	desired, use the {@link #setAttribute setAttribute} method instead.
 *
 * 	<p>If the attribute name is "class", this method performs the same function as the {@link #addClass addClass} 
 * 	method, but somewhat less efficiently.
 *	</p>
 * 	
 * @param AttrName						Name of the attribute.
 * @param AttrValue						Value of a new attribute, or the string to be appended to the value of an
 * 										existing attribute.
 * @return								A <code>Match</code> object identifying this markup element. If this element was 
 * 										immutable, a mutable version of it will have been created, substitued into this 
 * 										markup (assuming the markup itself is mutable), and the <code>Match</code> 
 * 										returned will refer to the new mutable form of this element.
 * @throws TagException					If a problem is encountered with the manipulation of the markup.
 */
	public Match addAttribute
	(
		final String			AttrName,
		final Object			AttrValue
	)
	throws TagException
	{
		throw new TagException(notSupportedMessage());
	}

/**	Replaces the value of an existing attribute, or creates a new attribute, in any markup element that is allowed to
 * 	have attributes. If the attribute name is "class", this method performs the same function as the 
 * 	{@link #setClass setClass} method, but somewhat less efficiently.
 *
 * @param AttrName						Name of the attribute.
 * @param AttrValue						Value of the attribute.
 * @return								A <code>Match</code> object identifying this markup element. If this element was 
 * 										immutable, a mutable version of it will have been created, substitued into this 
 * 										markup (assuming the markup itself is mutable), and the <code>Match</code> 
 * 										returned will refer to the new mutable form of this element.
 * @throws TagException					If a problem is encountered with the manipulation of the markup.
 */
	public Match setAttribute
	(
		final String			AttrName,
		final Object			AttrValue
	)
	throws TagException
	{
		throw new TagException(notSupportedMessage());
	}
	
/**	Replaces the value of an existing attribute, but never creates a new attribute, in any markup element that is 
 *	allowed to have attributes. If the attribute name is "class", this method performs the same function as the 
 *	{@link #setClass setClass} method, but somewhat less efficiently.
 *
 * @param AttrName						Name of the attribute.
 * @param AttrValue						Value of the attribute.
 * @return								A <code>Match</code> object identifying this markup element. If this element was 
 * 										immutable, a mutable version of it will have been created, substitued into this 
 * 										markup (assuming the markup itself is mutable), and the <code>Match</code> 
 * 										returned will refer to the new mutable form of this element.
 * @throws TagException					If a problem is encountered with the manipulation of the markup.
 */
	public Match updateAttribute
	(
		final String			AttrName,
		final Object			AttrValue
	)
	throws TagException
	{
		throw new TagException(notSupportedMessage());
	}

/**	Removes an attribute from this markup element.
 *
 * @param AttrName						Name of the attribute.
 * @return								A <code>Match</code> object identifying this markup element. If this element was 
 * 										immutable, a mutable version of it will have been created, substitued into this 
 * 										markup (assuming the markup itself is mutable), and the <code>Match</code> 
 * 										returned will refer to the new mutable form of this element.
 * @throws TagException					If a problem is encountered with the manipulation of the markup.
 */
	public Match deleteAttribute
	(
		final String			AttrName
	)
	throws TagException
	{
		throw new TagException(notSupportedMessage());
	}
	
/**	Returns an <code>Iterator</code> for this subsection of the larger document. If this <code>Match</code> object
 *	specifies an empty tag, only that tag will be returned by this <code>Iterator</code>. If this <code>Match</code> 
 * 	object specifies a start tag, that tag and all of its children (including all "end" tags) will be returned by the
 * 	<code>Iterator</code>.
 *
 * @return								An <code>Iterator</code> of this element and any children it may possess.
 */
	@Override public Iterator<Range> iterator() {
		return enumerate();
	}

/**	Returns an enumeration of this subsection of the larger document. If this <code>Match</code> object
 *	specifies an empty tag, only that tag will be returned by this enumeration. If this <code>Match</code> object 
 *	specifies a start tag, that tag and all of its children (including all "end" tags) will be enumerated.
 *
 * @return								An enumeration of this element and any children it may possess.
 */
	public MarkupEnumeration enumerate() {
		return enumerate(true);
	}

/**	Returns an enumeration of this subsection of the larger document. If this <code>Match</code> object
	specifies a self-contained tag, only that tag will be returned by this enumeration. If
 	this <code>Match</code> object specifies a start-tag, that tag and all of its children will be enumerated.

	@param ReturnEndTags				Set to <code>true</code> if the <code>MarkupEnumeration</code> object that
 										is returned should include end tags (example: "&lt;/a&gt;")
 										among the elements it enumerates. If <code>false</code>, end tags
 										are transparently skipped-over.
 	@return 							An enumeration of this element and any children it may possess.
*/
	public MarkupEnumeration enumerate
	(
		final boolean			ReturnEndTags
	)
	{
		if (get().isStartTag())
			return new MatchEnumerator(this, ReturnEndTags);
		else
			return new OneItemEnumerator(this);
	}
	
/**	Searches this element, and its children, for matches with the specified 
 * 	<a href='http://www.w3.org/TR/CSS21/selector.html#q1'>CSS2 pattern</a>(s). If there are no matches, the 
 * 	<code>Results</code> object returned is empty.
 *
 * @param Patterns						One or more CSS2 patterns to be applied to this element's children.
 * @return								A <code>Results</code> object containing a list of elements that matched the 
 * 										specified pattern(s), if any.
 */
	public Results get
	(
		final String			Patterns
	)
	{
		return get(Patterns, null);
	}
	
/**	Searches this element, and its children, for matches with the specified 
 * 	<a href='http://www.w3.org/TR/CSS21/selector.html#q1'>CSS2 pattern</a>(s). If there are no matches, the 
 * 	<code>Results</code> object returned is empty.
 *
 * @param Patterns						One or more CSS2 patterns to be applied to this element's children.
 * @return								A <code>Results</code> object containing a list of elements that matched the 
 * 										specified pattern(s), if any.
 */
	public Results get
	(
		final CSSPatterns		Patterns
	)
	{
		return get(Patterns, null);
	}
	
/**	Searches this element, and its children, for matches with the specified 
 * 	<a href='http://www.w3.org/TR/CSS21/selector.html#q1'>CSS2 pattern</a>(s). If there are no matches, the 
 * 	<code>Results</code> object returned is empty.
 *
 * @param Patterns						One or more CSS2 patterns to be applied to this element's children.
 * @param PrevResults					The previous <code>Results</code> object in a call chain. This is the 
 * 										<code>Results</code> object that will be returned when the <code>pop</code> 
 * 										method is invoked on the new <code>Results</code> object returned by this 
 * 										method. 
 * @return								A <code>Results</code> object containing a list of elements that matched the 
 * 										specified pattern(s), if any.
 */
	public Results get
	(
		final String			Patterns,
		final Results			PrevResults
	)
	{
		return get(getMarkup().getCSSPatterns(Patterns), PrevResults);
	}
	
/**	Searches this element, and its children, for matches with the specified 
 * 	<a href='http://www.w3.org/TR/CSS21/selector.html#q1'>CSS2 pattern</a>(s). If there are no matches, the 
 * 	<code>Results</code> object returned is empty.
 *
 * @param Patterns						One or more CSS2 patterns to be applied to this element's children.
 * @param PrevResults					The previous <code>Results</code> object in a call chain. This is the 
 * 										<code>Results</code> object that will be returned when the <code>pop</code> 
 * 										method is invoked on the new <code>Results</code> object returned by this 
 * 										method. 
 * @return								A <code>Results</code> object containing a list of elements that matched the 
 * 										specified pattern(s), if any.
 */
	public Results get
	(
		final CSSPatterns		Patterns,
		final Results			PrevResults
	)
	{
		return Patterns.search(enumerate(), new Results(PrevResults));
	}
	
/**	Get a complete list of all the ancestors of the XHTML element referenced by this <code>Match</code>.
 * 
 * @return								A new <code>Results</code> object containing a list of the ancestors of this 
 * 										tag. The first item will be the immediate parent of this element, the last 
 * 										element will be the most distant ancestor.
 */
	public Results getAncestors() {
		return getAncestors(false, null);
	}
	
/**	Get a complete list of all the ancestors of the XHTML element referenced by this <code>Match</code>.
 * 
 * @param IncludeThis					<code>true</code> if this <code>Match</code> should be included in the list, 
 * 										<code>false</code> otherwise.
 * @param PrevResults					The previous <code>Results</code> object in a call chain. This is the 
 * 										<code>Results</code> object that will be returned when the <code>pop</code> 
 * 										method is invoked on the new <code>Results</code> object returned by this 
 * 										method. Can be <code>null</code>.
 * @return								A new <code>Results</code> object containing a list of the ancestors of this 
 * 										tag. The first item will be the immediate parent of this element, the last 
 * 										element will be the most distant ancestor.
 */	
	public Results getAncestors
	(
		final boolean			IncludeThis,
		final Results			PrevResults
	) 
	{
		final Results			Ancestors   = new Results(PrevResults);
		final Markup			List        = getMarkup();
		int						EndTagCount = 0;
		
		if (IncludeThis)
			Ancestors.addMatch(this);
		
		for (int Index = getIndex(); --Index >= 0;) {
			final Range				Elem = List.get(Index);
			
			if (Elem.isEndTag()) {
				
				EndTagCount++;
			
			} else if (Elem.isStartTag()) {
				
				if (EndTagCount == 0)
					Ancestors.addMatch(create(List, Elem, Index));
				else
					EndTagCount--;
			}
		}
		
		return Ancestors;
	}
	
/**	Get the most recent ancestor of a particular type, for example "form" or "label".
 * 
 * @param TagType						The type of tag to look for amongst this element's ancestors.
 * @return								The most recent ancestor of the specified type.
 */
	public Match getAncestorOfType
	(
		final String			TagType
	)
	{
		Match					Ancestor = getParent();
		
		while (Ancestor != null) {
			
			if (TagType.equals(Ancestor.getType()))
				return Ancestor;
			
			Ancestor = Ancestor.getParent();
		}
		
		return null;
	}
	
/**	Get the immediate parent of the XHTML element referenced by this <code>Match</code>.
 * 
 * @return								A Match referencing the immediate parent of the XHTML reference by this 
 * 										<code>Match</code> object.
 */
	public Match getParent() {
		final Markup			List        = getMarkup();
		int						EndTagCount = 0;
		
		for (int Index = getIndex(); --Index >= 0;) {
			final Range				Elem = List.get(Index);
			
			if (Elem.isEndTag()) {
				
				EndTagCount++;
			
			} else if (Elem.isStartTag()) {
				
				if (EndTagCount == 0)
					return create(List, Elem, Index);
				
				EndTagCount--;
			}
		}
		
		return null;
	}
	
/**	Prints a nicely formatted representation of this element, and any children it may possess, to {@link System#out}.
 *	
 *	@return								A reference to this <code>Match</code> instance.
 */
	public Match print() {
		new DescendantEnumerator(enumerate(true), true).print();
		return this;
	}
	
/**	Compares another object to this one, in order to determine if it refers to the same component of the same body of
 *	markup. (Note that the subclasses that supply the implementations of this method, <code>ImmutableMatch</code> and
 *	<code>MutableMatch</code>, use very different means of testing for equality.)
 * 
 * @param Other							Another object to be tested for equality with this object.
 * @return								<code>true</code> if the objects match, <code>false</code> otherwise.
 */
	@Override public abstract boolean equals
	(
		final Object			Other
	);
	
	@Override public int hashCode() {
		return get().hashCode();
	}
	
/**	Returns a <code>String</code> containing the XML markup for the element referenced by this <code>Match</code>. 
 * 	Child elements are not included.	
 * 
 * @return					The XML markup for this element.
 * @see #toString(boolean)
 */	
	@Override public String toString() {
		return toString(true);
	}
	
/**	Returns a <code>String</code> containing the XML markup for the element referenced by this <code>Match</code>,
 *	optionally including any child elements, and the terminating "end" tag, if any. 
 * 
 * @param ElementOnly					<code>true</code> if nothing but this element should be represented in the 
 * 										<code>String</code>, or <code>false</code> if child elements and the terminating 
 * 										"end" tag, if present, should also be included in the <code>String</code>.
 * @return								This element's XML markup, represented as a <code>String</code>.
 */
	public String toString
	(
		final boolean			ElementOnly
	) 
	{
		if (ElementOnly) {
			
			return get().toString();
			
		} else {
			final StringWriter		SOut = new StringWriter(1024);
			final PrintWriter		Out  = new PrintWriter(SOut);
			
			try {
				write(Out);
			} catch (IOException e) {
				Out.println();
				Out.print(e.getClass().getName());
				Out.print(": ");
				Out.println(e.getMessage());
			}
			
			Out.close();
			
			return SOut.toString();
		}
	}
	
/**	Writes-out the XML markup element referenced by this <code>Match</code> object, including all child elements, and
 *	the terminating "end" tag, if any.
 * 
 * @param Out							The <code>PrintWriter</code> to which the markup should be written.
 * @throws IOException					If the <code>PrintWriter</code> had a problem.
 */
	public void write
	(
		final PrintWriter		Out
	)
	throws IOException
	{
		final MarkupEnumeration	Elems = enumerate();
		
		while (Elems.hasMoreElements())
			Elems.nextRange().write(Out);
	}
	
	private String notSupportedMessage() {
		final StringBuilder		Buff = new StringBuilder(192);
		
		Buff.append("Not supported. ");
		
		if (this instanceof ImmutableMatch)
			Buff.append("The markup \"").append(getMarkup().getMarkupName()).append("\" is immutable, and therefore cannot be modified. (The markup was represented by the class \"").append(getMarkup().getClass().getName()).append("\".) A modifiable form of the markup can be obtained by invoking its \"getMutable\" method.");
		else if (this instanceof MutableMatch)
			Buff.append("The MutableMatch class does not supply an implementation of this method. This should never happen.");
		else
			Buff.append("An unknown subclass of Match, \"").append(getClass().getName()).append("\", does not supply an implementation of this method.");
		
		return Buff.toString();
	}
}






