/*	QWICAP (Quick Web Interface for Conventional Applications) by Chris W. Johnson
	Copyright (C) 2003-2010 The University of Texas at Austin ("U.T. Austin"). All rights reserved.
	
	U.T. Austin's copyright notice in this evolving version reflects our original authorship. It does not assert 
	copyright over changes contributed by others.
	
	This library is free software; you can redistribute it and/or modify it under the terms of the GNU Lesser General 
	Public License as published by the Free Software Foundation; either version 2.1 of the License, or (at your option) 
	any later version.
	
	This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied 
	warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for 
	more details.
	
	You should have received a copy of the GNU Lesser General Public License along with this library; if not, write to 
	the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/


package edu.utexas.its.eis.tools.qwicap.template.xml;


import edu.utexas.its.eis.tools.qwicap.template.xml.structure.Range;

import java.util.NoSuchElementException;


/**	Used by the {@link Match#enumerate} method to enumerate any <code>Match</code> that represents a 
 *	<code>StartTag</code>. This enumeration will return that <code>StartTag</code> and all of its children, along with 
 *	(optionally) the associated <code>EndTag</code>(s). Matches that represent anything other than 
 *	<code>StartTag</code>s should be enumerated using the <code>OneItemEnumerator</code>, since they have no contents.
 *	
 *	<p>When configured to return end tags, this enumerator may be passed to <code>DescendantEnumerator</code> which 
 *	assists code with comprehending the context of tags during enumeration.</p>
 *	
 *	@author Chris W. Johnson
 */
final class MatchEnumerator implements MarkupEnumeration {
	
	private final Markup		Doc;
	private final int			DocOrigChangeCount;
	private final boolean		ReturnEndTags;
	private final RangeList		List;
	private final int			FinalIndex;		//	Index of the final Range we will return (ReturnEndTags has been taken into account).

	private int					NextIndex;
	private Range				NextRange;

	private int					CurrIndex;
	private Range				CurrRange;
	
	
	MatchEnumerator
	(
		final Match				M,
		final boolean			ReturnEndTags
	)
	{
		if (M.get().isStartTag() == false)
			throw new IllegalArgumentException("MatchEnumerator must be given a Match object specifying a StartTag. This Match object specifies a \"" + M.get().getClass().getName() + "\".");
		
		Doc = M.getMarkup();
		DocOrigChangeCount = Doc.getChangeCount();
		
		this.ReturnEndTags = ReturnEndTags;
		
		List       = M.getMarkup().getList();
		
		FinalIndex = findFinalIndex(M);
	
		NextIndex  = M.getIndex();
		NextRange  = M.get();
		
		CurrIndex  = -1;
		CurrRange  = null;
	}
	
	private MatchEnumerator
	(
		final MatchEnumerator	Master
	)
	{
		Doc                = Master.Doc;
		DocOrigChangeCount = Master.DocOrigChangeCount;
		ReturnEndTags      = Master.ReturnEndTags;
		List               = Master.List;
		FinalIndex         = Master.FinalIndex;
		
		NextIndex          = Master.NextIndex;
		NextRange          = Master.NextRange;
		
		CurrIndex          = Master.CurrIndex;
		CurrRange          = Master.CurrRange;
	}
	
	@Override public MatchEnumerator clone() {
		return new MatchEnumerator(this);
	}

	@Override public int getIndex() {
		return CurrIndex;
	}

	@Override public Markup getMarkup() {
		return Doc;
	}

	@Override public Match getMatch() {
		return Match.create(getMarkup(), CurrRange, CurrIndex);
	}
	
	@Override public boolean hasNext() {
		return hasMoreElements();
	}
	
	@Override public Range next() {
		return nextRange();
	}
	
	@Override public void remove() {
		throw new UnsupportedOperationException();
	}
	
	@Override public boolean hasMoreElements() {
		return NextRange != null;
	}
	
	@Override public Range nextElement() {
		return nextRange();
	}

	@Override public Range nextRange() throws NoSuchElementException {
		
		if (NextRange == null)
			throw new NoSuchElementException();
		
		if (Doc.getChangeCount() != DocOrigChangeCount)
			throw new IllegalStateException("The markup \"" + Doc.getMarkupName() + "\" has been modified while being enumerated. This enumeration cannot continue with assured accuracy.");
		
		CurrIndex = NextIndex;
		CurrRange = NextRange;
		
		if (NextIndex < FinalIndex) {
			
			NextRange = List.Array[++NextIndex];
			
			if (ReturnEndTags == false) {
			
			//	Advance until we find something other than an end-tag.
						
				while (NextIndex <= FinalIndex && NextRange.isEndTag())
					NextRange = List.Array[++NextIndex];
			}
				
		} else {
			
			NextIndex = -1;
			NextRange = null;
		}
		
		return CurrRange;
	}
	
	private int findFinalIndex
	(
		final Match				InitialMatch
	) 
	{
		final Range[]			Ranges     = List.Array;
		final int				RangeCount = List.size();
		int						Index      = InitialMatch.getIndex();
		int						StartCount = 0;
		
		do {
			final Range				R = Ranges[Index];
			
			if (R.isStartTag()) {
				
				StartCount++;
				
			} else if (R.isEndTag()) {
				
				if (--StartCount == 0)
					return ReturnEndTags ? Index : Index - 1;
			}
			
		} while (++Index < RangeCount);
	
		throw new RuntimeException(new TagException("The markup \"" + getMarkup().getMarkupName() + "\" is not well-formed: There is no terminating tag for the tag \"" + InitialMatch.toString(true) + "\"."));
	}
}
