/*	QWICAP (Quick Web Interface for Conventional Applications) by Chris W. Johnson
	Copyright (C) 2003-2010 The University of Texas at Austin ("U.T. Austin"). All rights reserved.
	
	U.T. Austin's copyright notice in this evolving version reflects our original authorship. It does not assert 
	copyright over changes contributed by others.
	
	This library is free software; you can redistribute it and/or modify it under the terms of the GNU Lesser General 
	Public License as published by the Free Software Foundation; either version 2.1 of the License, or (at your option) 
	any later version.
	
	This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied 
	warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for 
	more details.
	
	You should have received a copy of the GNU Lesser General Public License along with this library; if not, write to 
	the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/


package edu.utexas.its.eis.tools.qwicap.template.xml;


import edu.utexas.its.eis.tools.qwicap.template.xml.structure.Range;

import java.util.List;


/**	Allows <code>ImmutableMarkup</code> and <code>MutableMarkup</code> to avoid some of the overhead in using an 
 *	<code>ArrayList</code> object for this purpose. There are two main advantages: (1) direct access to the array, so 
 *	there is less overhead than calling <code>ArrayList.get</code> all the time, and (2) the array is of type 
 *	<code>Range</code>, so there's somewhat less typecasting to be done. Also, methods like <code>contains</code> and 
 *	<code>indexOf</code> only test for object reference equality, which makes them faster than their counterparts in 
 *	the <code>List</code> interface, though less flexible (Qwicap doesn't need that flexibility).
 *	
 *	<p>Note that this class is conceptually similar to <code>MatchList</code>, and where applicable their 
 *	implementations should be maintained in parallel.
 * 	</p>
 *
 *	<p>This class is not part of Qwicap's public API, but it could not be hidden because uses of it span package 
 * 	boundaries.
 * 	</p>
 *
 *	@author Chris W. Johnson
 */
public final class RangeList implements Cloneable {
	
	public Range[]				Array;
	private int					LogicalLength;
	
	
	public RangeList
	(
		final int				InitialCapacity
	)
	{
		Array = new Range[InitialCapacity];
		LogicalLength = 0;
	}
	
	public void clear() {
		LogicalLength = 0;
	}
	
	public void trimToSize() {
		
		if (LogicalLength < Array.length) {
			final Range[]			TrimmedArray = new Range[LogicalLength];
			
			System.arraycopy(Array, 0, TrimmedArray, 0, LogicalLength);
			
			Array = TrimmedArray;
		}
	}
	
	public int size() {
		return LogicalLength;
	}
	
	public boolean isEmpty() {
		return LogicalLength == 0;
	}
	
//	For backwards compatibility with ArrayList.
	
	public Range get
	(
		final int				Index
	)
	{
		if (Index < 0 || Index >= LogicalLength)
			throw new ArrayIndexOutOfBoundsException("Invalid array index. We were asked to retrieve element " + Index + ", but the valid indices range from 0 to " + (LogicalLength - 1) + " (inclusive).");
		
		return Array[Index];
	}
	
//	Note that this implementation differs from the one in List in that it tests only for object reference equality, 
//	whereas the implementation in List uses the "equals" method of one of the objects.
	
	public boolean contains
	(
		final Range				SearchForObj
	)
	{
		return indexOf(SearchForObj) != -1;
	}
	
//	Note that this implementation differs from the one in List in that it tests only for object reference equality, 
//	whereas the implementation in List uses the "equals" method of one of the objects.
	
	public int indexOf
	(
		final Range				SearchForObj
	)
	{
		final Range[]			Array = this.Array;
		
		for (int Index = 0, Count = LogicalLength; Index < Count; Index++) {
			
			if (Array[Index] == SearchForObj)
				return Index;
		}
		
		return -1;
	}
	
//	For backwards compatibility with ArrayList.
	
	public void set
	(
		final int				Index,
		final Range				Value
	)
	{
		Array[Index] = Value;
	}
	
	public void add
	(
		final Range				Value
	)
	{
		growIfNecessary(1);
		Array[LogicalLength++] = Value;
	}
	
	public void add
	(
		final int				Index,
		final Range				Value
	)
	{
		growIfNecessary(1);
		makeRoom(Index, 1);
		Array[Index] = Value;
	}
	
	public void addSubsection
	(
		final int				DestIndex,
		final RangeList			SrcList,
		final int				SrcIndex,
		final int				SrcCount
	)
	{
		if (SrcIndex + SrcCount > SrcList.size())
			throw new IllegalArgumentException("The size of the source list is " + SrcList.size() + ". Therefore, it is too small to copy " + SrcCount + " elements from index " + SrcIndex + '.');
		
		growIfNecessary(SrcCount);
		makeRoom(DestIndex, SrcCount);
		System.arraycopy(SrcList.Array, SrcIndex, Array, DestIndex, SrcCount);
	}
	
	public void addAll
	(
		final int				Index,
		final RangeList			Values
	)
	{
		final int				ValuesCount = Values.size();
		
		growIfNecessary(ValuesCount);
		makeRoom(Index, ValuesCount);
		System.arraycopy(Values.Array, 0, Array, Index, ValuesCount);
	}
	
	public void addAll
	(
		final int				Index,
		final List<Range>		Values
	)
	{
		final int				ValuesCount = Values.size();
		
		growIfNecessary(ValuesCount);
		makeRoom(Index, ValuesCount);
		System.arraycopy(Values.toArray(), 0, Array, Index, ValuesCount);
	}
	
	public void remove
	(
		final Range				Obj
	)
	{
		remove(indexOf(Obj));
	}
	
	public void remove
	(
		final int				Index
	)
	{
		remove(Index, 1);
	}
	
	public void remove
	(
		final int				Index,
		final int				Count
	)
	{
		final int				RemainderIndex = Index + Count;
		final int				RemainderLength = LogicalLength - RemainderIndex;
		
		if (Index < 0 || Index >= LogicalLength)
			throw new ArrayIndexOutOfBoundsException("Invalid array index. We were asked to remove elements beginning at " + Index + ", but the valid indices range from 0 to " + (LogicalLength - 1) + '.');
		if (Count < 0)
			throw new ArrayIndexOutOfBoundsException("The number of elements to be removed from the array cannot be less than zero, so " + Count + " won't work.");
		if (RemainderIndex > LogicalLength)
			throw new ArrayIndexOutOfBoundsException("Invalid array index. We were asked to remove " + Count + " elements from the array beginning at " + Index + ", but because the array's logical length is " + LogicalLength + " we can only remove at most " + (LogicalLength - Index) + " elements from that index.");
		
		if (RemainderLength > 0)
			System.arraycopy(Array, RemainderIndex, Array, Index, RemainderLength);
		
		LogicalLength -= Count;
	}
	
	private void makeRoom
	(
		final int				DestIndex,
		final int				NewElements
	)
	{
		if (DestIndex < 0 || DestIndex > LogicalLength)
			throw new ArrayIndexOutOfBoundsException("Invalid array index. We were asked to shift array elements beginning with " + DestIndex + ", but the valid indices range from 0 to " + LogicalLength + '.');
		if (NewElements < 0)
			throw new IllegalArgumentException("Can't make room for a negative number (" + NewElements + ") of new array elements.");
		
		final int				ElementsToMove = LogicalLength - DestIndex;
		
		if (ElementsToMove > 0)
			System.arraycopy(Array, DestIndex, Array, DestIndex + NewElements, ElementsToMove);
		
		LogicalLength += NewElements;
	}
	
	private void growIfNecessary
	(
		final int				NewElementsNeeded
	)
	{
		if (NewElementsNeeded < 0)
			throw new IllegalArgumentException("Can't use growIfNecessary to \"grow\" an array by a negative amount (" + NewElementsNeeded + ").");
		
		final int				PyLen = Array.length;
		final int				NewLgLen = LogicalLength + NewElementsNeeded;
		
		if (NewLgLen > PyLen) {
			int						NewPyLen = (PyLen * 3 >> 1) + 1;
			
			if (NewPyLen < NewLgLen)
				NewPyLen = NewLgLen;
						
		//	Allocate the new array, and copy the contents of the previous array into it, then replace the previous
		//	array with the new one.
			
			final Range[]			NewArray = new Range[NewPyLen];
			
			System.arraycopy(Array, 0, NewArray, 0, LogicalLength);
			Array = NewArray;
		}
	}
	
	@Override public RangeList clone() {
		
		try {
			final RangeList			NewList = (RangeList) super.clone();
			
			NewList.Array = Array.clone();
			
			return NewList;
			
		} catch (CloneNotSupportedException e) {
			throw new RuntimeException(e);
		}
	}
}
