/*	QWICAP (Quick Web Interface for Conventional Applications) by Chris W. Johnson
	Copyright (C) 2003-2010 The University of Texas at Austin ("U.T. Austin"). All rights reserved.
	
	U.T. Austin's copyright notice in this evolving version reflects our original authorship. It does not assert 
	copyright over changes contributed by others.
	
	This library is free software; you can redistribute it and/or modify it under the terms of the GNU Lesser General 
	Public License as published by the Free Software Foundation; either version 2.1 of the License, or (at your option) 
	any later version.
	
	This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied 
	warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for 
	more details.
	
	You should have received a copy of the GNU Lesser General Public License along with this library; if not, write to 
	the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/


//  Attribute.java
//  qwicap
//
//  Created by Chris W. Johnson on Sat Jun 07 2003.


package edu.utexas.its.eis.tools.qwicap.template.xml.immutable;


import edu.utexas.its.eis.tools.qwicap.template.xml.structure.Attribute;
import edu.utexas.its.eis.tools.qwicap.template.xml.structure.Namespace;
import edu.utexas.its.eis.tools.qwicap.util.Characters;

import java.util.*;


/**	Represents an immutable attribute in an XML document.
 * 
 *	@author Chris W. Johnson
 */
final class ImAttribute extends ImNamedItem implements Attribute {
	
	private static final ImAttribute[]	kEmptyImAttributeArray = {};
	
//	private boolean			DefinesNamespace;
	private Namespace		DefinedNamespace;
	
	
//	Format: name="value"  
//	Quotes (single or double) are mandatory in XML/XHTML. Space may, or may not, 
//	be optional around the equal sign.
	
	private ImAttribute
	(
		final char[]			Chars,
		final int				PyStart,
		final int				PyEnd,
		final int				NameStart,
		final int				NameEnd,
		final int				ValueStart,
		final int				ValueEnd
	)
	{
		super(Chars, PyStart, PyEnd);
		
		this.ContentStart = ValueStart;
		this.ContentEnd   = ValueEnd;
		
		this.NameStart    = NameStart;
		this.NameEnd      = NameEnd;
	}
	
	@Override public boolean valueEquals
	(
		final Characters		Other
	)
	{
		return contentEquals(Other);
	}
	
	@Override public boolean valueIncludes
	(
		final Characters		Other
	)
	{
		final int				ValueLength = ContentEnd - ContentStart;
		final int				OtherLength = Other.length();
		
		if (OtherLength == ValueLength) {
			
			return valueEquals(Other);
			
		} else if (OtherLength < ValueLength) {
			final char[]			OtherChars    = Other.getArray();
			final int				OtherOffset   = Other.getArrayOffset();
			boolean					WasWhitespace = true;
			
			for (int TIndex = ContentStart, TLimit = ContentEnd; TIndex < TLimit; TIndex++) {
				boolean					IsWhitespace = isWhitespace(Chars[TIndex]);
				
				if (WasWhitespace && IsWhitespace == false) {
					
				//	We've found the start of a value in a space-separated list of attribute values.
					
					if (TLimit - TIndex < OtherLength)
						return false;		//	There aren't enough characters remaining for a match to be possible.
					
				//	Compare the characters in Other and the Content of this attribute.
					
					int						OCount = OtherLength;
					
					for (int OIndex = OtherOffset; OCount > 0; OIndex++, TIndex++, OCount--) {
						
						if (OtherChars[OIndex] != Chars[TIndex])
							break;
					}
					
				//	OK. Comparison complete. Arriving here, however, we don't know the outcome of the 
				//	comparison, i.e. we can't say with certainty whether the value we're looking for 
				//	is present in this attribute's content. We acquire that certainty from the tests below.
					
				//	Since the comparison loop may have advanced TIndex, the value of IsWhitespace could be 
				//	stale. Bring it up to date.
					
					IsWhitespace = isWhitespace(Chars[TIndex]);
					
				//	We only have a complete match if all the characters were compared, and we've either
				//	exhausted all of the characters in our Content, or the next character in our Content is 
				//	whitespace.
					
					if (OCount == 0 && (TIndex == TLimit || IsWhitespace))
						return true;
				}
				
				WasWhitespace = IsWhitespace;
			}
		}
		
		return false;
	}

/*	This method performs the type of test specified by the Cascading Style Sheets (CSS)
	"|=" attribute selector. That selector matches the start of a hyphen-sepearated value
	list. For the match to be successful, the search string must either match everything
	from the start of the value list up to a hyphen character, or must match the entire
	value list. The search string MUST NOT include a trailing hyphen.
	
	An example from the CSS2 specification, section 5.8.1:
	
	The following rule will match for values of the "lang" attribute that begin with "en", 
	including "en", "en-US", and "en-cockney":

		*[LANG|="en"] { color : red }
*/	
	@Override public boolean valueStartsWith
	(
		final Characters		Other
	)
	{
		final int				ValueLength = ContentEnd - ContentStart;
		int						OtherLength = Other.length();
		
		if (OtherLength == ValueLength) {
			
			return valueEquals(Other);
			
		} else if (OtherLength < ValueLength) {
			final char[]			OtherChars = Other.getArray();
			int						OIndex     = Other.getArrayOffset();
			int						CIndex     = ContentStart;
			
			while (--OtherLength >= 0) {
				
				if (Chars[CIndex++] != OtherChars[OIndex++])
					return false;
			}
			
			return CIndex == ContentEnd || Chars[CIndex] == '-';
		}
		
		return false;
	}
	
	@Override public Characters getValue() {
		return getContent();
	}
	
	@Override public StringBuffer append
	(
		final StringBuffer		Buff
	)
	{
		getName().append(Buff);
		Buff.append("=\"");
		getValue().append(Buff);
		Buff.append('"');
		
		return Buff;
	}
	
	@Override public StringBuilder append
	(
		final StringBuilder		Buff
	)
	{
		getName().append(Buff);
		Buff.append("=\"");
		getValue().append(Buff);
		Buff.append('"');
		
		return Buff;
	}
	
	@Override public boolean beNamespaceAware() {
		final boolean			HasNamespaceInfo = super.beNamespaceAware();
		
		if (HasNamespaceInfo) {
		
		//	This attribute's name was found to contain a namespace prefix, e.g. "prefix:name".
			
			final Characters		Prefix = getNamespacePrefix();
			
			if (Namespace.kNamespaceAttributeName.equals(Prefix)) {
				
			//	The attribute's prefix is "xmlns", which means that it defines a namespace. Amongst this tag's children,
			//	this namespace will be specified by including this attribute's name as the prefix of tag/attribute 
			//	names.
				
			//	Note that although this attribute defines a namespace, it is not *in* the namespace it defines. Its 
			//	"xmlns"	prefix places it in the reserved "http://www.w3.org/2000/xmlns/" namespace. See:
			//	<http://www.w3.org/TR/REC-xml-names/#ns-decl>.
				
				setNamespace(Namespace.kXMLNSNamespace);
				DefinedNamespace = new ImNamespace(getName(), getValue());
				
			//	setNamespace(new ImNamespace(getName(), getValue()));
			//	DefinesNamespace = true;
				
			} // else The attribute's namespace prefix references a namespace defined previously in the document. (Or 
			//	the namespace could be defined outside the document, if the document was intended for use with a 
			//	validating XML parser, in which case the document shouldn't be used with Qwicap, because Qwicap's 
			//	parser is non-validating.)
			
		} else if (nameEquals(Namespace.kNamespaceAttributeName)) {
			
		//	This attribute is named "xmlns" and has no prefix, so it defines the default namespace for this tag's
		//	children.
		//
		//	Note from "Namespaces in XML", <http://www.w3.org/TR/REC-xml-names/#defaulting>:
		//	"If the URI reference in a default namespace declaration is empty, then unprefixed elements in the scope 
		//	of the declaration are not considered to be in any namespace."
		//	
		//	We set the namespace of elements that are not in any namespace to the static constant Namespace.kNoNamespace.
			
		//	Note that although this attribute defines a namespace, it is not *in* the namespace it defines. Because it
		//	lacks a namespace prefix, and because it is a reserved word in XML, the "xmlns" attribute is not in any 
		//	namespace.
			
			setNamespace(Namespace.kNoNamespace);
	
			final Characters		URI = getValue();
			
			if (URI.length() == 0)
				DefinedNamespace = Namespace.kNoNamespace;
			else
				DefinedNamespace = new ImNamespace(null, getValue());
			
		} else {
			
		//	From "Namspaces in XML", <http://www.w3.org/TR/REC-xml-names/#defaulting>:
		//	"Note that default namespaces do not apply directly to attributes."
		//	
		//	This attribute has no namespace prefix, and its name is other than "xmlns", therefore it has no namespace.
			
			setNamespace(Namespace.kNoNamespace);
		}
		
		return HasNamespaceInfo;
	}
	
	boolean definesNamespace() {
		return DefinedNamespace != null;
	}
	
	Namespace getDefinedNamespace() {
		return DefinedNamespace;
	}
	
/**	Read any number of attributes from a subset of the specified character array.
	Return the attributes in a List. If no attributes are found, return null.
*/	
	static ImAttribute[] readAll
	(
		final char[]			Chars,
		int						Index,
		final int				Limit
	)
	{
		ArrayList<ImAttribute>	Attrs = null;
		
		while (Index < Limit) {
			int						NameStart            = -1;
			int						NameEnd              = -1;
			int						ValueStart           = -1;
			boolean					LookingForName       = true;
			boolean					LookingForEqual      = false;
			boolean					LookingForOpenQuote  = false;
		//	boolean					LookingForCloseQuote = false;
			char					CloseQuoteChar       = 0;
			
			while (Index < Limit) {
				final char				Char = Chars[Index];
				
				if (LookingForName) {
					
					if (isWhitespace(Char) == false) {
						
						NameStart = Index;
						
						LookingForName  = false;
						LookingForEqual = true;
					}
					
				} else if (LookingForEqual) {
					
					if (Char == '=') {
						
						NameEnd = Index;
						
						LookingForEqual     = false;
						LookingForOpenQuote = true;
					}
					
				} else if (LookingForOpenQuote) {
					
					if (Char == '"' || Char == '\'') {
						
						CloseQuoteChar = Char;
						ValueStart = Index + 1;
						
						LookingForOpenQuote  = false;
					//	LookingForCloseQuote = true;
					}
					
				} else /* if (LookingForCloseQuote) */ {
					
					if (Char == CloseQuoteChar) {
						
						if (Attrs == null)
							Attrs = new ArrayList<ImAttribute>();
						
						Attrs.add(new ImAttribute(Chars, NameStart, Index + 1, NameStart, NameEnd, ValueStart, Index));
						Index++;
						break;
					}
				}
				
				Index++;
			}
		}
		
		if (Attrs != null) 
			return Attrs.toArray(new ImAttribute[Attrs.size()]);
		
		return kEmptyImAttributeArray;
	}
}









