/*	QWICAP (Quick Web Interface for Conventional Applications) by Chris W. Johnson
	Copyright (C) 2003-2010 The University of Texas at Austin ("U.T. Austin"). All rights reserved.
	
	U.T. Austin's copyright notice in this evolving version reflects our original authorship. It does not assert 
	copyright over changes contributed by others.
	
	This library is free software; you can redistribute it and/or modify it under the terms of the GNU Lesser General 
	Public License as published by the Free Software Foundation; either version 2.1 of the License, or (at your option) 
	any later version.
	
	This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied 
	warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for 
	more details.
	
	You should have received a copy of the GNU Lesser General Public License along with this library; if not, write to 
	the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/


//  NamedItem.java
//  qwicap
//
//  Created by Chris W. Johnson on Sat Jun 07 2003.


package edu.utexas.its.eis.tools.qwicap.template.xml.immutable;


import edu.utexas.its.eis.tools.qwicap.template.xml.structure.NamedItem;
import edu.utexas.its.eis.tools.qwicap.template.xml.structure.Namespace;
import edu.utexas.its.eis.tools.qwicap.util.Characters;


/**	Common ancestor of all immutable named items in XML documents. This is not part of Qwicap's public API, but it
 *	could not be hidden because uses of it span package boundaries.
 * 
 *	@author Chris W. Johnson
 */
public abstract class ImNamedItem extends ImItem implements NamedItem {
	
	int						NameStart = -1;
	int						NameEnd   = -1;
	
	private boolean			NsResolved;
	private Namespace		Ns;
	private int				NsPrefixStart = -1;
	private int				NsPrefixEnd   = -1;
	
	
	ImNamedItem
	(
		final char[]			Chars,
		final int				PyStart,
		final int				PyEnd
	)
	{
		super(Chars, PyStart, PyEnd);
	}
	
	final boolean nameEquals
	(
		final ImNamedItem		Other
	)
	{
		int						N1    = NameStart;
		int						N2    = Other.NameStart;
		int						Count = NameEnd - N1;
		
		if (Count != Other.NameEnd - N2)
			return false;
		
		while (--Count >= 0) {
			
			if (Chars[N1++] != Other.Chars[N2++])
				return false;
		}
		
		return true;
	}
	
	@Override public final boolean nameEquals
	(
		final NamedItem			Other
	)
	{
		return nameEquals(Other.getName());
	}
	
	@Override public final boolean nameEquals
	(
		final Characters		Other
	)
	{
		return Other.equals(Chars, NameStart, NameEnd - NameStart);
	}
	
	@Override public final Characters getName() {
		return new Characters(Chars, NameStart, NameEnd - NameStart);
	}
	
	@Override public final String getNameString() {
		return new String(Chars, NameStart, NameEnd - NameStart);
	}
	
	boolean beNamespaceAware() {
		
		for (int Index = NameStart; Index < NameEnd; Index++) {
			
			if (Chars[Index] == ':') {
				
				NsPrefixStart = NameStart;
				NsPrefixEnd   = Index;
				NameStart     = Index + 1;
				break;
			}
		}
		
		return NsPrefixStart >= 0;	//	true if we have namespace information of any kind, false if we don't.
	}
	
	void restoreNamespacePrefixToName() {
		
		if (NsPrefixStart < 0)
			return;
		
	//	This undoes the work done by our beNamespaceAware method.
		
		NameStart     = NsPrefixStart;
		NsPrefixStart = -1;
		NsPrefixEnd   = -1;
	}
	
	final Characters getNamespacePrefix() {
		
		if (NsPrefixStart < 0)
			return null;
		
	//	If our namespace has been resolved, we can be more efficient by returning its name object, which will be 
	//	the same as our prefix.
		
		if (NsResolved && Ns != null)
			return Ns.getName();	
		else
			return new Characters(Chars, NsPrefixStart, NsPrefixEnd - NsPrefixStart);
	}
	
	@Override public Namespace getNamespace() {
		return Ns;
	}
	
	ImNamedItem setNamespace
	(
		final Namespace			TheNs	//	Can be null.
	)
	{
		if (NsResolved) {
			
		//	Be tolerant of redundant settings, for now.
			
			if (Ns != null && Ns.equals(TheNs) == false)
				throw new IllegalArgumentException("The namespace of this item (\"" + toString() + "\") has already been set (to \"" + this.Ns.getURI() + "\"), so it can't be set again.");
		}
		
		NsResolved = true;
		Ns         = TheNs;
		
		return this;
	}
	
	boolean namespaceResolved() {
		return NsResolved;
	}
	
/**	Create a clone of this object, but assigns the clone to the specified namespace.
 * 
 * @param NewNs					The namespace of the clone.
 * @return						The clone.
 */
	public ImNamedItem clone
	(
		final Namespace			NewNs
	)
	{
		return ((ImNamedItem) clone()).setNamespace(NewNs);
	}
}





