/*	QWICAP (Quick Web Interface for Conventional Applications) by Chris W. Johnson
	Copyright (C) 2003-2010 The University of Texas at Austin ("U.T. Austin"). All rights reserved.
	
	U.T. Austin's copyright notice in this evolving version reflects our original authorship. It does not assert 
	copyright over changes contributed by others.
	
	This library is free software; you can redistribute it and/or modify it under the terms of the GNU Lesser General 
	Public License as published by the Free Software Foundation; either version 2.1 of the License, or (at your option) 
	any later version.
	
	This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied 
	warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for 
	more details.
	
	You should have received a copy of the GNU Lesser General Public License along with this library; if not, write to 
	the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/


//  Range.java
//  qwicap
//
//  Created by Chris W. Johnson on Sun Jun 08 2003.


package edu.utexas.its.eis.tools.qwicap.template.xml.immutable;


import edu.utexas.its.eis.tools.qwicap.template.xml.structure.Range;
import edu.utexas.its.eis.tools.qwicap.util.Characters;

import java.io.PrintWriter;


/*	Class hierarchy:
	
	Range  [abstract]
	  |
	  +-- Whitespace
	  |
	  +-- Item  [abstract]			** Consider renaming: Element. Unfortunately, "Content" doesn't exactly fit in such a catagory.
	        |
	        +-- XMLDeclaration
	        |
	        +-- Declaration
	        |
	        +-- Comment
			|
			+-- Content
	        |
	        +-- CDATA
	        |
	        +-- NamedItem  [abstract]
	              |
	              +-- Attribute
	              |
				  +-- Tag  [abstract]
	                    |
				        +-- EndTag
			            |
				        +-- TagWithAttributes  [abstract]
				              |
	                          +-- StartTag
	                          |
	                          +-- EmptyTag
	
	
	Tag with Attributes parsed:
	
	<a href="http://www.thing.edu/">
	||||                           ||
	|||+-- ContentStart            |+-- PyEnd
	||+--- NameEnd                 +--- ContentEnd
	|+---- NameStart
    +----- PyStart
	
	
	Attribute parsed:
	
	href="http://www.thing.edu/"
	|   | |                    ||
	|   | +-- ContentStart     |+-- PyEnd
	|   +---- NameEnd          +--- ContentEnd
	+-------- PyStart, NameStart
	
	
	End Tag parsed:
	
	</table>
	| |    ||
	| |    |+-- PyEnd
	| |    +--- NameEnd, ContentStart, ContentEnd
	| +-------- NameStart
	+---------- PyStart
	
	
	Declaration parsed:
	
	<?xml version="1.0" encoding="UTF-8"?>
	| |                                  ||
	| +-- ContentStart                   |+-- PyStop
	+---- PyStart                        +---ContentEnd
	
	
	Comment parsed:
	
	<!-- Fred the Wonder Llama bids you greetings. -->
	|    |                                        |   |
	|    +-- ContentStart                         |   +-- PyStop
	+---- PyStart                                 +------ ContentEnd
	
	
	Content parsed:
	
	\r\t\tFred the Wonder Llama bids you greetings. \r\r
	|     |                                        |    |
	|     +-- ContentStart                         |    +-- PyStop
	+---- PyStart                                  +---ContentEnd
	
	
*/

/**	Common ancestor of all sections of any type of immutable material in XML documents.
 * 
 *	@author Chris W. Johnson
 */
abstract class ImRange implements Range, CharSequence {

	final char[]			Chars;
	private final int		PyStart;	//	Offset of first character, inclusive.
	final int				PyEnd;		//	Offset of final character, exclusive.
	
	
	ImRange
	(
		final char[]			Chars,
		final int				PyStart,
		final int				PyEnd
	) 
	{
		this.Chars   = Chars;
		this.PyStart = PyStart;
		this.PyEnd   = PyEnd;
	}

/**	Returns the length of this character sequence. The length is the number of 16-bit chars in the sequence.
 * 
 * @return						The length of this character sequence. The length is the number of 16-bit chars in the 
 * 								sequence.
 */
	@Override public int length() {
		return PyEnd - PyStart;	
	}
	
/**	Returns the <code>char</code> value at the specified index. An index ranges from zero to <code>length() - 1</code>. 
 * 	The first <code>char</code> value of the sequence is at index zero, the next at index one, and so on, as for array 
 * 	indexing.
 * 
 *	<p>If the char value specified by the index is a surrogate, the surrogate value is returned.
 * 	</p>
 *
 * @param Index							The index of the character to be returned.
 * @return								The <code>char</code> at the specified index.
 * @throws IndexOutOfBoundsException	If the index is invalid.
 */
	@Override public char charAt
	(
		final int				Index
	)
	{
		if (Index < 0)
			throw new IndexOutOfBoundsException("Index values cannot be less than zero. This one, " + Index + ", is less than zero.");
		if (Index >= length())
			throw new IndexOutOfBoundsException("Index values cannot fall beyond the end of the range. This range is " + length() + " characters in length, so this index, " + Index + ", is invalid.");
		
		return Chars[PyStart + Index];
	}
	
/**	Returns a <code>Characters</code> instance representing a subsequence of this object's characters.
 * 
 * @param FirstCharOffset		Offset of the first character, inclusive.
 * @param LastCharOffset		Offset of the final character, exclusive.
 * @return						The requested subsequence, represented by an instance of <code>Characters</code>.
 */
	@Override public CharSequence subSequence
	(
		int 					FirstCharOffset,
		int						LastCharOffset
	)
	{
		final int				Length = length();
		
		if (FirstCharOffset < 0)
			FirstCharOffset = 0;
		else if (FirstCharOffset >= Length)
			return new Characters(Chars, PyEnd, 0);
		
		if (LastCharOffset < 0)
			return new Characters(Chars, PyStart, 0);
		else if (LastCharOffset > Length)
			LastCharOffset = Length;
		
		return new Characters(Chars, PyStart + FirstCharOffset, LastCharOffset - FirstCharOffset);
	}
	
	@Override public final boolean rangeEquals
	(
		final Characters		Other
	)
	{
		return Other.equals(Chars, PyStart, PyEnd - PyStart);
	}
	
	@Override public final String toString() {
		return new String(Chars, PyStart, PyEnd - PyStart);
	}
	
	@Override public final void write
	(
		final PrintWriter		Out
	)
	{
		Out.write(Chars, PyStart, PyEnd - PyStart);
	}
	
	static boolean isWhitespace
	(
		final char				Char
	)
	{
	/*	The XML 1.0 (Third Edition) specification, section 2.3, <http://www.w3.org/TR/REC-xml/#sec-common-syn>, defines 
		whitespace as follows:
			
			S (white space) consists of one or more space (#x20) characters, carriage returns, line feeds, or tabs.
			
			White Space
			
			[3]   	S	   ::=   	(#x20 | #x9 | #xD | #xA)+
			Note:
			
			The presence of #xD in the above production is maintained purely for backward compatibility with the First 
			Edition. As explained in 2.11 End-of-Line Handling, all #xD characters literally present in an XML document 
			are either removed or replaced by #xA characters before any other processing is done. The only way to get 
			a #xD character to match this production is to use a character reference in an entity value literal.
			
		Perhaps we should use that definition, seeing as how we're parsing XML. However, the XHTML 1.0 specification 
		expands on that definition in Section 3.2, "User Agent Conformance":
		
			The XHTML user agent in addition, must treat the following characters as whitespace:

				Form feed (&#x000C;)
				Zero-width space (&#x200B;)
		
		It's also hard to know whether the Character class, which is meant to know Unicode's definitions of character
		types, agrees with either of the specifications referenced above, or even with Unicode's own elusive defintion 
		of whitespace, let alone all three. And then there's the matter of what people/programmers recongize as 
		whitespace (characters without glyphs would be a good first approximation), which may not be technically 
		relevant, but which is likely to intrude on the reality of this issue at some stage, if only in document markup.
		
		All things considered, I have no idea what defintion of whitespace to use, so the following will have to do....
	*/
	
		if (Char <= ' ')
			return true;
		
		if (Char < 127)		//	127 is "del" and is almost certainly considered whitespace.
			return false;
		
		return Character.isWhitespace(Char);
	}
	
	static int eatWhitespace
	(
		final char[]			Chars,
		int						Index,
		final int				Limit
	)
	{
		while (Index < Limit && isWhitespace(Chars[Index]))
			Index++;
		
		return Index;
	}
	
	static int readUntil
	(
		final char[]			Chars,
		int						Index,
		final int				Limit,
		final char				UntilChar
	)
	{
		while (Index < Limit && Chars[Index] != UntilChar)
			Index++;
		
		return Index;
	}
	
	static int readUntilWhitespace
	(
		final char[]			Chars,
		int						Index,
		final int				Limit
	)
	{
		while (Index < Limit && isWhitespace(Chars[Index]) == false)
			Index++;
		
		return Index;
	}
	
	static int readUntilWhitespaceOr
	(
		final char[]			Chars,
		int						Index,
		final int				Limit,
		final char				OrChar1
	)
	{
		while (Index < Limit) {
			final char				Char = Chars[Index];
			
			if (Char == OrChar1 || isWhitespace(Char))
				break;
			
			Index++;
		}
		
		return Index;
	}
	
	static int readUntilWhitespaceOr
	(
		final char[]			Chars,
		int						Index,
		final int				Limit,
		final char				OrChar1,
		final char				OrChar2
	)
	{
		while (Index < Limit) {
			final char				Char = Chars[Index];
			
			if (Char == OrChar1 || Char == OrChar2 || isWhitespace(Char))
				break;
			
			Index++;
		}
		
		return Index;
	}
	
	static boolean equals
	(
		final char[]			Chars,
		int						Index,
		final int				Limit,
		final char[]			TestSeq
	)
	{
		int						SeqIndex = 0;
		int						SeqLength = TestSeq.length;
		
	//	Are there enough characters remaining in Chars to match the entire
	//	sequence of characters in TestSeq?
		
		if (Limit - Index >= SeqLength) {
			
			while (--SeqLength >= 0) {
				
				if (Chars[Index++] != TestSeq[SeqIndex++])
					return false;
			}
			
			return true;
		}
		
		return false;
	}
	
	@Override public final Object clone() {
		try {
			return super.clone();
		} catch (CloneNotSupportedException e) {
			throw new RuntimeException(e);
		}
	}
	
	@Override public boolean isStartTag()          { return false; }
	@Override public boolean isEndTag()            { return false; }
	@Override public boolean isEmptyTag()          { return false; }
	@Override public boolean isTagWithAttributes() { return false; }
}





