/*	QWICAP (Quick Web Interface for Conventional Applications) by Chris W. Johnson
	Copyright (C) 2003-2010 The University of Texas at Austin ("U.T. Austin"). All rights reserved.
	
	U.T. Austin's copyright notice in this evolving version reflects our original authorship. It does not assert 
	copyright over changes contributed by others.
	
	This library is free software; you can redistribute it and/or modify it under the terms of the GNU Lesser General 
	Public License as published by the Free Software Foundation; either version 2.1 of the License, or (at your option) 
	any later version.
	
	This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied 
	warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for 
	more details.
	
	You should have received a copy of the GNU Lesser General Public License along with this library; if not, write to 
	the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/


//  TagWithAttributes.java
//  qwicap
//
//  Created by Chris W. Johnson on Sat Jun 07 2003.


package edu.utexas.its.eis.tools.qwicap.template.xml.immutable;


import edu.utexas.its.eis.tools.qwicap.template.xml.*;
import edu.utexas.its.eis.tools.qwicap.template.xml.structure.*;
import edu.utexas.its.eis.tools.qwicap.util.Characters;

import java.util.*;


/**	Common ancestor of all immutable tags that can have attributes in XML documents.
 *	
 *	@author Chris W. Johnson
 */
abstract class ImTagWithAttributes extends ImTag implements TagWithAttributes {
	
	private static final Namespace[]	kEmptyNamespaceArray = {};
	
	private ImAttribute[]				Attributes;
	private ArrayList<Characters>		Classes;
	private boolean						DefinesNamespaces;
	private Namespace[]					DefinedNamespaces = kEmptyNamespaceArray;
	
	
	ImTagWithAttributes
	(
		final char[]			Chars,
		final int				PyStart,
		final int				PyEnd
	)
	{
		super(Chars, PyStart, PyEnd);
	}
	
	final void locateName() {
		
		NameStart = ContentStart;
		NameEnd   = readUntilWhitespaceOr(Chars, ContentStart, ContentEnd, '>', '/');
		
		ContentStart = NameEnd;
		trimContent();
	}
	
	final void locateAttributes() {
		final char[]			Chars = this.Chars;
		
	//	The locateName method must have already been invoked.
		
		Attributes = ImAttribute.readAll(Chars, ContentStart, ContentEnd);
		
	//	The "class" attribute is special (later we may wish to regard "id" as special,
	//	too), so we offer special access to its contents...
		
		final ImAttribute		CAttr = (ImAttribute) getAttribute(kClassAttributeName);
		
		if (CAttr != null) {
			ArrayList<Characters>	Classes = new ArrayList<Characters>();
			
		//	Parse the contents of the "class" attribute, making each class name contained
		//	therein an element of the Classes vector, so we can easily search by class names
		//	in the future.
			
			int						Index           = eatWhitespace(Chars, CAttr.ContentStart, CAttr.ContentEnd);
			int						ClassCount      = 0;
			final int				CAttrContentEnd = CAttr.ContentEnd;
			
			while (Index < CAttrContentEnd) {
				final int				Start = Index;
				final int				End   = readUntilWhitespace(Chars, Index, CAttrContentEnd);
				
				Classes.add(new Characters(Chars, Start, End - Start));
				ClassCount++;
				
				Index = eatWhitespace(Chars, End, CAttrContentEnd);
			}
			
			if (ClassCount == 0)
				Classes = null;
			else
				Classes.trimToSize();
				
			this.Classes = Classes;
		}
	}
	
	@Override public final boolean isOfClass
	(
		final Characters		ClassName
	)
	{
		final ArrayList<Characters>	Classes = this.Classes;
		
		if (Classes != null) {
			
			for (int Index = 0, Count = Classes.size(); Index < Count; Index++) {
				
				if (ClassName.equals(Classes.get(Index)))
					return true;
			}
		}
		
		return false;
	}
	
	@Override public final String getAttribute
	(
		final String			AttrName
	)
	{
		String					Value = null;
		final Attribute			Attr  = getAttribute(new Characters(AttrName));
		
		if (Attr != null)
			Value = Attr.getValue().toString();
		
		return Value;
	}
	
	@Override public final Attribute getAttribute
	(
		final Characters		AttrName
	)
	{
		final ImAttribute[]		Attributes = this.Attributes;
		
		for (int Index = 0, Count = Attributes.length; Index < Count; Index++) {
			final ImAttribute		Attr = Attributes[Index];
			
			if (Attr.nameEquals(AttrName))
				return Attr;
		}
		
		return null;
	}

	@Override public final AttributeEnumeration enumerateAttributes() {
		
		if (Attributes.length == 0)
			return NoAttributeEnumerator.kTheOnlyInstanceYouWillEverNeed;
		else
			return new AttributeEnumeratorArray(Attributes);
	}
	
	@Override public Iterator<Attribute> iterator() {
		return enumerateAttributes();
	}
	
	@Override public boolean beNamespaceAware() {
		boolean					HasNamespaceInfo = super.beNamespaceAware();
		final ImAttribute[]		Attributes       = this.Attributes;
		final int				AttrCount        = Attributes.length;
		
		if (AttrCount > 0) {
			final Namespace[]		NsDefs   = new Namespace[AttrCount];
			int						NsDefIdx = 0;
			
		//	Tell the attributes of this tag to be namespace aware, and build a list of the namespaces they define, if any.
			
			for (int Index = 0; Index < AttrCount; Index++) {
				final ImAttribute		Attr = Attributes[Index];
				
				HasNamespaceInfo |= Attr.beNamespaceAware();
				
			//	If this attribute defines a namespace, add it to the list of namespace definitions.
				
				if (Attr.definesNamespace())
					NsDefs[NsDefIdx++] = Attr.getDefinedNamespace();
			}
			
		//	If the attributes included any namespace definitions....
			
			if (NsDefIdx > 0) {
				
				DefinesNamespaces = true;
				HasNamespaceInfo  = true;
				
				if (NsDefIdx == NsDefs.length) {
					
					DefinedNamespaces = NsDefs;
					
				} else {
					
				//	Copy the used portion of the NsDefs array (which was allocated for the worst case in which all of
				//	this tag's attributes defined namespaces) to the a new array of exactly the right length.
					
					DefinedNamespaces = new Namespace[NsDefIdx];
					System.arraycopy(NsDefs, 0, DefinedNamespaces, 0, NsDefIdx);
				}
			}
		}
		
		return HasNamespaceInfo;
	}
	
	boolean definesNamespaces() {
		return DefinesNamespaces;
	}
	
	Namespace[] getDefinedNamespaces() {
		return DefinedNamespaces;
	}
	
	final void resolveAttributeNamespaces
	(
		final TagHierarchy			TH
	)
	{
		final ImAttribute[]		Attributes   = this.Attributes;
		final Namespace			kNoNamespace = Namespace.kNoNamespace;
		
		for (int Index = 0, Count = Attributes.length; Index < Count; Index++) {
			final ImAttribute		Attr = Attributes[Index];
			
			if (Attr.namespaceResolved() == false) {
				final Characters		Prefix = Attr.getNamespacePrefix();
				
				if (Prefix == null) {
					
				//	The attribute has no namespace. (The XML spec. say attributes without prefixes are NOT assigned to 
				//	the default namespace, unlike tags.)
					
					Attr.setNamespace(kNoNamespace);
					
				} else {
					
				//	Search up through the tag hierarchy for a namespace definition that corresponds to the namespace
				//	prefix used on this attribute. If no namespace definition is found, the attribute is assigned to 
				//	the null namespace represented by the static Namespace.kNoNamespace object.
					
					Attr.setNamespace(TH.getNamespaceForPrefix(Prefix));
				}
			}
		}
	}
	
	@Override public boolean isTagWithAttributes() { return true; }
}





