/*	QWICAP (Quick Web Interface for Conventional Applications) by Chris W. Johnson
	Copyright (C) 2003-2011 The University of Texas at Austin ("U.T. Austin"). All rights reserved.
	
	U.T. Austin's copyright notice in this evolving version reflects our original authorship. It does not assert 
	copyright over changes contributed by others.
	
	This library is free software; you can redistribute it and/or modify it under the terms of the GNU Lesser General 
	Public License as published by the Free Software Foundation; either version 2.1 of the License, or (at your option) 
	any later version.
	
	This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied 
	warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for 
	more details.
	
	You should have received a copy of the GNU Lesser General Public License along with this library; if not, write to 
	the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/


//  ImmutableMarkup.java
//  qwicap
//
//  Created by Chris W. Johnson on Wed Jun 18 2003.


package edu.utexas.its.eis.tools.qwicap.template.xml.immutable;


import edu.utexas.its.eis.tools.qwicap.template.css.CSSPatterns;
import edu.utexas.its.eis.tools.qwicap.template.css.CSSPatternsCache;
import edu.utexas.its.eis.tools.qwicap.template.xml.*;
import edu.utexas.its.eis.tools.qwicap.template.xml.mutable.Mutable;
import edu.utexas.its.eis.tools.qwicap.template.xml.mutable.MutableMarkup;
import edu.utexas.its.eis.tools.qwicap.template.xml.structure.Content;
import edu.utexas.its.eis.tools.qwicap.template.xml.structure.NamedItem;
import edu.utexas.its.eis.tools.qwicap.template.xml.structure.Range;
import edu.utexas.its.eis.tools.qwicap.template.xml.structure.XMLDeclaration;
import edu.utexas.its.eis.tools.qwicap.util.Characters;
import edu.utexas.its.eis.tools.qwicap.servlet.HTTPContentType;

import java.io.IOException;
import java.io.PrintWriter;
import java.io.StringWriter;


/**	<code>ImmutableMarkup</code> represents a body of XML markup that cannot be altered in any way. The markup 
 *	represented may be an entire XML document, or merely a fragment of XML. In either case, well-formedness is required 
 *	to the extent that all "start" tags must have matching "end" tags.
 *	
 *	@author Chris W. Johnson
 */
public class ImmutableMarkup extends Markup {
	
	private static final CSSPatterns	kCSSHTMLHeadMetaEquivContent = new CSSPatterns("html > head > meta[http-equiv][content]");
	private static final int			kAverageCharactersPerRange   = 20;
	
	private Characters					Input;
	private RangeList					List;
	private boolean						IsCompound;
	
	private boolean						CharacterSetHasBeenSet;
	private String						CharacterSet = "UTF-8";	//	In memory the markup is always in Unicode (UTF-16); this variable tells us how the markup was encoded when it was read, and how it should be encoded when it is written.
	
	private final CSSPatternsCache		CSSPatsCache = new CSSPatternsCache();
	
	
	ImmutableMarkup() {
		List = new RangeList(25);
	}
	
/**	Parses a character array containing XML markup; either a well-formed fragment, or an entire XML document.
 * 
 * @param Chars					The array of characters containing the XML markup to be parsed. The array
 *								is retained as the backing store for this body of markup, so under no circumstances
 *								should the contents of the array be modified after invoking this constructor.
 * @throws TagException			If there's a problem parsing the XML, verifying its tag hierarchy, or any number of
 *								related problems.
 */
	public ImmutableMarkup
	(
		final char[]			Chars
	)
	throws TagException
	{
		this(Chars, 0, Chars.length);
	}
	
/**	Parses a subsection of a character array containing XML markup; either a well-formed fragment, or an entire XML
 *	document.
 * 
 * @param Chars					The array of characters containing the XML markup to be parsed. The array
 *								is retained as the backing store for this body of markup, so under no circumstances
 *								should the contents of the array be modified after invoking this constructor.
 * @param Index					The offset of the first character in the array that should be parsed.
 * @param Limit					The offset of the character <b>immediately after</b> the last character that should be 
 *								parsed.
 * @throws TagException			If there's a problem parsing the XML, verifying its tag hierarchy, or any number of
 *								related problems.
 */
	public ImmutableMarkup
	(
		final char[]			Chars,
		final int				Index,
		final int				Limit
	)
	throws TagException
	{
		read(Chars, Index, Limit);
	}
	
	public ImmutableMarkup
	(
		final MutableMarkup		MDoc
	)
	throws TagException
	{
		setMarkupName(MDoc.getMarkupName());
		setCharacterSet(MDoc.getCharacterSet());
		
		List = new RangeList(MDoc.getList().size());
		
		IsCompound = true;	//	This ImmutableMarkup is potentially made-up of pieces from many documents.
		
		final RangeList			MList      = MDoc.getList();
		final Range[]			MListArray = MList.Array;
		final RangeList			List       = this.List;
		
		for (int Index = 0, Count = MList.size(); Index < Count; Index++) {
			final Range				R = MListArray[Index];
			
			if (R instanceof Mutable) {
				final char[]			Chars = R.toString().toCharArray();
			
				if (R instanceof Content) {
					
				//	This is a small optimization. Better ones are possible, but we'd have to expose the MuContent class
				//	in the public API.
						
					List.add(new ImContent(Chars, 0, Chars.length));
					
				} else {
					final Range				NewR = readOneRange(Chars, 0, Chars.length, TagHierarchyDummy.kTheOnlyInstanceYouWillEverNeed);
					
					if (R instanceof NamedItem)
						((ImNamedItem) NewR).setNamespace(((NamedItem) R).getNamespace());
					
					List.add(NewR);
				}
				
			} else
				List.add(R);
		}
		
		checkHierarchyAndResolveNamespaces(true);
	}
	
	@Override public String getCharacterSet() {
		
		if (CharacterSetHasBeenSet == false) {
		
			if (List.size() != 0) {
				
			//	First, try to get the character set name from the XML declaration.
				
				{
					final Range				FirstElem = List.Array[0];
					
					if (FirstElem instanceof XMLDeclaration) {
						final String			Encoding = ((XMLDeclaration) FirstElem).getEncoding();
						
						if (Encoding != null)
							setCharacterSet(Encoding);
					}
				}
				
			//	Second, try to get the character set name from the XHTML "meta"/"http-equiv"/"content-type" tag in the 
			//	"head" of the document, if present.
				
				if (CharacterSetHasBeenSet == false) {
					
					for (final Match Meta : get(kCSSHTMLHeadMetaEquivContent)) {
						
						if ("Content-Type".equalsIgnoreCase(Meta.getAttribute("http-equiv"))) {
							final String			CharSetName = new HTTPContentType(Meta.getAttribute("content")).getCharacterSet();
							
							if (CharSetName != null) {
								
								setCharacterSet(CharSetName);
								break;
							}
						}
					}
				}
			}
			
			if (CharacterSetHasBeenSet == false)
				setCharacterSet("UTF-8");
		}
		
		return CharacterSet;
	}
	
	void setCharacterSet
	(
		final String			CharacterSet
	)
	{
		if (CharacterSetHasBeenSet)
			throw new IllegalStateException("The ImmutableMarkup.setCharacterSet method can be invoked only once.");
		
		this.CharacterSet           = CharacterSet;
		this.CharacterSetHasBeenSet = true;
	}
	
	@Override public ImmutableMarkup clone() throws CloneNotSupportedException {
		
	//	This implementation is present only in the interest of code-clarity.
	//
	//	A shallow copy is fine here. The only possible source of conflict is CSSPatsCache, because its contents are
	//	modifiable, but its implementation has sufficient synchronization to make it shareable. In the future, we
	//	may wish to consider whether sharing it is appropriate, but it's not clear why anyone would clone an immutable
	//	object in the first place.
		
		return (ImmutableMarkup) super.clone();	
	}
	
	@Override public CSSPatterns getCSSPatterns
	(
		final String			PatsStr
	)
	{
		synchronized (CSSPatsCache) {
			return CSSPatsCache.getCSSPatterns(PatsStr);
		}
	}
	
	@Override public final int getChangeCount() {
		return 0;
	}
	
	@Override public final RangeList getList() {
		return List;
	}
	
	@Override public final Match createMatch
	(
		final Range				Element,
		final int				Index
	)
	{
		return new ImmutableMatch(this, Element, Index);
	}
	
	@Override public ImmutableMarkup getImmutable() throws TagException {
		return this;
	}
	
	@Override public final MutableMarkup getMutable() {
		final MutableMarkup		MutDoc = new MutableMarkup(List.clone(), getCharacterSet());
		
		MutDoc.setMarkupName(getMarkupName());
		
		return MutDoc;
	}
	
	final void read
	(
		final Characters		Chars
	)
	throws TagException
	{
		read(Chars, true);
	}
	
	final void read
	(
		final Characters		Chars,
		final boolean			ResolveNamespaces
	)
	throws TagException
	{
		Input = Chars;
		read(Chars.getArray(), Chars.getArrayOffset(), Chars.getArrayOffset() + Chars.length(), ResolveNamespaces);
	}
	
	final void read
	(
		final char[]			Chars,
		final int				Index,
		final int				Limit
	)
	throws TagException
	{
		read(Chars, Index, Limit, true);
	}
	
	final void read
	(
		final char[]			Chars,
		final int				Index,
		final int				Limit,
		final boolean			ResolveNamespaces
	)
	throws TagException
	{
		if (List == null)
			List = new RangeList(((Limit - Index) + (kAverageCharactersPerRange - 1)) / kAverageCharactersPerRange);
		
		if (Input == null)
			Input = new Characters(Chars, Index, Limit - Index);
		
		try {
			final RangeList			Ranges    = getList();
			final TagHierarchy		Hierarchy = new TagHierarchy(getMarkupName(), ResolveNamespaces);
			ImRange					Thing     = readOneRange(Chars, Index, Limit, Hierarchy);
			
			while (Thing != null) {
				
				Ranges.add(Thing);
				
			//	Get the next range of the document, starting from the end of the previous range.
				
				Thing = readOneRange(Chars, Thing.PyEnd, Limit, Hierarchy);
			}
			
			Hierarchy.done();
		//	checkHierarchyAndResolveNamespaces(ResolveNamespaces);
			
		} catch (TagException e) {
			e.setMarkup(Chars, Index, Limit);
			throw e;
		}
	}
	
	private void checkHierarchyAndResolveNamespaces
	(
		final boolean			ResolveNamespaces
	) 
	throws TagException 
	{
		final TagHierarchy		Hierarchy = new TagHierarchy(getMarkupName(), ResolveNamespaces);
		final Range[]			ListArray = this.List.Array;
		
		for (int Index = 0, Count = List.size(); Index < Count; Index++) {
			final Range				Thing = ListArray[Index];
			
			if (Thing.isStartTag()) {
				
				Hierarchy.startTag((ImStartTag) Thing);
			
			} else if (Thing.isEndTag()) {
				
				Hierarchy.endTag((ImEndTag) Thing);
			
			} else if (Thing.isEmptyTag()) {
				
				Hierarchy.emptyTag((ImEmptyTag) Thing);
			}
		}
		
		Hierarchy.done();
	}
	
/*                        LEADING            TRAILING
                          -------            --------
    * XML Declaration:      <?                  ?>
    * Declaration:          <!                  >
    * Comment:              <!--                -->
    * CDATA:                <![CDATA[           ]]>
    * End Tag:              </                  >
    * Start Tag:            <                   >
    * Self-Contained Tag:   <                   />
*/

	static ImRange readOneRange
	(
		final char[]				Chars,
		int							Index,
		final int					Limit,
		final TagHierarchyValidator	Hierarchy
	)
	throws TagException
	{
		if (Index < Limit) {
			final char				FirstChar = Chars[Index];
			
			if (FirstChar == '<') {
				
				if (Limit - Index > 2) {
					
					switch (Chars[Index + 1]) {
						
						case '/':
							return Hierarchy.endTag(new ImEndTag(Chars, Index, locateTagTerminator(Chars, Index, Limit, ImEndTag.kEndSequence)));
							
						case '!':	//	Could be declaration, comment, or CDATA.
							
							if (ImRange.equals(Chars, Index, Limit, ImCDATA.kBeginSequence))
								return new ImCDATA(Chars, Index, locateTagTerminator(Chars, Index, Limit, ImCDATA.kEndSequence));
							else if (ImRange.equals(Chars, Index, Limit, ImComment.kBeginSequence))
								return new ImComment(Chars, Index, locateTagTerminator(Chars, Index, Limit, ImComment.kEndSequence));
							else
								return new ImDeclaration(Chars, Index, locateTagTerminator(Chars, Index, Limit, ImDeclaration.kEndSequence));
							
						case '?':
							return new ImXMLDeclaration(Chars, Index, locateTagTerminator(Chars, Index, Limit, ImXMLDeclaration.kEndSequence));
							
						default: {
								final int				PyEnd = locateTagTerminator(Chars, Index, Limit, ImStartTag.kEndSequence);
								
								if (Chars[PyEnd - 2] == '/')
									return Hierarchy.emptyTag(new ImEmptyTag(Chars, Index, PyEnd));
								else
									return Hierarchy.startTag(new ImStartTag(Chars, Index, PyEnd));
							}
					}
				}
				
			} else {
				final int				StartIndex = Index;
				
				if (ImRange.isWhitespace(FirstChar))
					Index = ImRange.eatWhitespace(Chars, StartIndex, Limit);
				
				if (Index == Limit || Chars[Index] == '<')
					return new ImWhitespace(Chars, StartIndex, Index);
				else
					return new ImContent(Chars, StartIndex, ImRange.readUntil(Chars, Index, Limit, '<'));
			}
		}
		
		return null;
	}

	private static int locateTagTerminator
	(
		final char[]			Chars,
		int						Index,		//	Should be equivalent of PyStart
		int						Limit,
		final char[]			EndSeq		//	Last character must be '>', but we don't check it.
	)
	throws IncompleteTagException
	{
		final char				EndSeqChar0  = EndSeq[0];
		final int				EndSeqLength = EndSeq.length;
		final int				OrigIndex    = Index;
		
	//	Reduce the limit by the number of characters (less one) in the end-sequence for which we're looking. This
	//	way, when we find a match for the first character in the end-sequence, we don't have to check whether there
	//	are enough characters remaining before the limit to accomodate the entire end-sequence; we know there are at
	//	least that many.
		
		Limit -= EndSeqLength - 1;
		
		while (Index < Limit) {
			
			if (Chars[Index] == EndSeqChar0) {
				int						EIndex = 1;
				
				for (int CIndex = Index + 1; EIndex < EndSeqLength; CIndex++, EIndex++) {
					
					if (Chars[CIndex] != EndSeq[EIndex])
						break;
				}
				
				if (EIndex == EndSeqLength)
					return Index + EndSeqLength;
			}
			
			Index++;
		}
		
		throw new IncompleteTagException("The element terminator sequence \"" + new String(EndSeq) + "\" was not found in the remaining markup: " + new String(Chars, OrigIndex, Limit - OrigIndex));
	}
	
	@Override public String toString() {
		
		if (IsCompound && Input == null) {
			final StringWriter		SW = new StringWriter(16 * 1024);
			
			try {
				write(new PrintWriter(SW));
			} catch (IOException e) {
				throw new RuntimeException(e);
			}
			
			Input = new Characters(SW.toString());
		}
		
		return Input.toString();
	}
	
	@Override public PrintWriter write
	(
		final PrintWriter		Out
	)
	throws IOException
	{
		if (IsCompound && Input == null) {
			
		//	We can't assume that all the elements of this document share a common backing store, or, even if they happen
		//	to do so, that their order in the list is consistent with their order in the backing store, or that all the
		//	elements described in the backing store markup are present in the list.
			
			for (int Index = 0, Count = List.size(); Index < Count; Index++)
				List.Array[Index].write(Out);
			
		} else
			Input.write(Out);
		
		return Out;
	}
}




