/*	QWICAP (Quick Web Interface for Conventional Applications) by Chris W. Johnson
	Copyright (C) 2003-2010 The University of Texas at Austin ("U.T. Austin"). All rights reserved.
	
	U.T. Austin's copyright notice in this evolving version reflects our original authorship. It does not assert 
	copyright over changes contributed by others.
	
	This library is free software; you can redistribute it and/or modify it under the terms of the GNU Lesser General 
	Public License as published by the Free Software Foundation; either version 2.1 of the License, or (at your option) 
	any later version.
	
	This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied 
	warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for 
	more details.
	
	You should have received a copy of the GNU Lesser General Public License along with this library; if not, write to 
	the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/


//  ImmutableMatch.java
//  qwicap
//
//  Created by Chris W. Johnson on Sat Jun 21 2003.


package edu.utexas.its.eis.tools.qwicap.template.xml.immutable;


import edu.utexas.its.eis.tools.qwicap.template.xml.Markup;
import edu.utexas.its.eis.tools.qwicap.template.xml.Match;
import edu.utexas.its.eis.tools.qwicap.template.xml.mutable.MutableMatch;
import edu.utexas.its.eis.tools.qwicap.template.xml.structure.Range;
import edu.utexas.its.eis.tools.qwicap.template.xml.structure.Comment;
import edu.utexas.its.eis.tools.qwicap.template.xml.structure.CDATA;
import edu.utexas.its.eis.tools.qwicap.template.xml.structure.Declaration;
import edu.utexas.its.eis.tools.qwicap.template.css.Results;

import java.util.NoSuchElementException;


/**	<code>ImmutableMatch</code> uniquely identifies a particular element within a body of XHTML markup represented by 
 *	an {@link ImmutableMarkup} object.
 *	
 *	@author Chris W. Johnson
 */
public final class ImmutableMatch extends Match {
	
	private final ImmutableMarkup	Markup;
	private final Range				Item;
	private final int				Index;
	
	
	ImmutableMatch
	(
		final ImmutableMarkup	Markup,
		final Range				Item,
		final int				Index
	)
	{
		this.Markup = Markup;
		this.Item   = Item;
		this.Index  = Index;
	}
	
	@Override public boolean exists() {
		return true;
	}
	
/**	Tests an object for equality with this one. This means establishing that the other object is an 
 *	<code>ImmutableMatch</code>, and that it refers to the same body of markup, and the same component within that 
 *	markup. (Note that {@link MutableMatch#equals(Object)} works very differently.)
 * 
 * @param Other					Another object to be tested for equality with this object.
 * @return						<code>true</code> if the objects match, <code>false</code> otherwise.
 */
	@Override public boolean equals
	(
		final Object			Other
	)
	{
		if (Other == this)
			return true;
		
		if (Other instanceof ImmutableMatch) {
			final Match				OtherMatch = (Match) Other;
			
			if (OtherMatch.getMarkup() == getMarkup()) {
				
			//	Immutable markup does not maintain a match cache, so multiple ImmutableMatch instances may refer to
			//	the same range. Nonetheless, we want such instances to be considered equal. So, here we test to see
			//	if the two matches refer to the same range.
				
				if (OtherMatch.get() == get()) {
					
					return true;
				}
			}
		}
		
		return false;
	}
	
	@Override public Markup getMarkup() {
		return Markup;
	}
	
	@Override public int getIndex() throws NoSuchElementException {
		return Index;
	}
	
	@Override public Range get() throws NoSuchElementException {
		return Item;
	}
	
	@Override public Results getContent
	(
		final Results			PrevResults
	)
	{
		final Range				Item = get();
		
		if (Item.isStartTag()) {
			
			return super.getContent(PrevResults);
		
		} else {
			final Results			Content = new Results(PrevResults);
			
			if (Item instanceof Comment || Item instanceof CDATA || Item instanceof Declaration)
				Content.addMatch(new ImmutableIntraElementContentMatch(this));
			
			return Content;
		}
	}
}





