/*	QWICAP (Quick Web Interface for Conventional Applications) by Chris W. Johnson
	Copyright (C) 2003-2011 The University of Texas at Austin ("U.T. Austin"). All rights reserved.
	
	U.T. Austin's copyright notice in this evolving version reflects our original authorship. It does not assert 
	copyright over changes contributed by others.
	
	This library is free software; you can redistribute it and/or modify it under the terms of the GNU Lesser General 
	Public License as published by the Free Software Foundation; either version 2.1 of the License, or (at your option) 
	any later version.
	
	This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied 
	warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for 
	more details.
	
	You should have received a copy of the GNU Lesser General Public License along with this library; if not, write to 
	the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/


package edu.utexas.its.eis.tools.qwicap.template.xml.immutable;


import edu.utexas.its.eis.tools.qwicap.template.xml.structure.*;
import edu.utexas.its.eis.tools.qwicap.template.xml.TagException;
import edu.utexas.its.eis.tools.qwicap.util.Characters;
import edu.utexas.its.eis.tools.qwicap.util.ALStack;
import edu.utexas.its.eis.tools.qwicap.util.ArrayToString;
import edu.utexas.its.eis.tools.qwicap.util.Null;

import java.util.EmptyStackException;


/**	Verifies an XML document's tag hierarchy, and optionally resolves the namespaces of all elements and attributes.
 *	
 *	@author Chris W. Johnson
 */
final class TagHierarchy implements TagHierarchyValidator {

	private final String			DocumentName;				//	Can be null, although it really shouldn't be.
	private final boolean			BeNamespaceAware;
	private final ALStack<TagInfo>	Scopes = new ALStack<TagInfo>();	//	Stores TagInfo objects.
	private TagInfo					CurScope;
	
	
	TagHierarchy
	(
		final String			DocumentName,		//	Can be null, although it really shouldn't be.
		final boolean			BeNamespaceAware
	) 
	{
		this.DocumentName     = DocumentName;
		this.BeNamespaceAware = BeNamespaceAware;
		
	//	Ensure that the default "xml" namespace is always defined.
		
		push(new TagInfo(null, null, new Namespace[] { Namespace.kXMLNamespace }));
	}
	
	@Override public void done() throws TagException {
		
	//	The item pushed by the constructor (to represent the XML namespace) will always be present, but there 
	//	shouldn't be anything else.
		
		if (Scopes.size() == 1)
			return;
		
		Scopes.remove(0);	//	Remove that first item for clarity in the following error message.
		
		throw new TagException("XML Hierarchy Error: In the document \"" + DocumentName + "\", the following tags (" + Scopes.size() + ") are not terminated: " + new ArrayToString(Scopes));
	}
	
	@Override public ImStartTag startTag
	(
		final ImStartTag		Tag
	)
	{
		if (BeNamespaceAware) {
			
			Tag.beNamespaceAware();
			
			push(new TagInfo(Tag, CurScope.getDefaultNamespace()));
			
			resolveTagNamespace(Tag);
			Tag.resolveAttributeNamespaces(this);
			
		} else
			push(new TagInfo(Tag, CurScope.getDefaultNamespace()));
		
		return Tag;
	}
	
	@Override public ImEndTag endTag
	(
		final ImEndTag			Tag
	)
	throws TagException
	{
	
	//	Resolve tag namespace, if necessary.
	
		if (BeNamespaceAware) {
			
			Tag.beNamespaceAware();
			resolveTagNamespace(Tag);
		}
		
	//	Verify tag hierarchy.
		
		try {
			final TagInfo			TInfo    = pop();
			final ImStartTag		StartTag = TInfo.getStartTag();
			
			if (Tag.nameEquals(StartTag) == false)
				throw new TagException("XML Hierarchy Error: In the document \"" + DocumentName + "\", end tag \"" + Tag + "\" was found where there should have been an end tag for \"" + StartTag + "\".");
			
			if (BeNamespaceAware && Null.safeEquals(StartTag.getNamespace(), Tag.getNamespace()) == false)
				throw new TagException("XML Namespace Error: In the document \"" + DocumentName + "\", end tag \"" + Tag + "\" is in the \"" + Tag.getNamespace() + "\" namespace, but its corresponding start tag \"" + StartTag + "\" is in the \"" + StartTag.getNamespace() + "\" namespace.");
		
		} catch (EmptyStackException e) {
			throw new TagException("XML Hierarchy Error: In the document \"" + DocumentName + "\", the end tag \"" + Tag + "\" has no corresponding start tag.");	
		}
		
		return Tag;
	}
	
	@Override public ImEmptyTag emptyTag
	(
		final ImEmptyTag		Tag
	)
	{
		if (BeNamespaceAware) {
			
			Tag.beNamespaceAware();
			resolveTagNamespace(Tag);
			Tag.resolveAttributeNamespaces(this);
		}
		
		return Tag;
	}
	
	private void resolveTagNamespace
	(
		final ImTag				Tag
	)
	{
		if (Tag.namespaceResolved() == false) {
			final Characters		Prefix = Tag.getNamespacePrefix();
			
			if (Prefix == null) {
				
			//	Assign this tag to the current default namespace, even if that's null.	
				
				Tag.setNamespace(CurScope.getDefaultNamespace());
				
			} else {
				
			//	Search up through the tag hierarchy for a namespace definition that corresponds to the namespace
			//	prefix used on this tag. If no namespace definition is found, the tag is assigned to the null 
			//	namespace, which may, or may not be, a good idea.
				
				final Namespace			Ns = getNamespaceForPrefix(Prefix);
				
				if (Ns != Namespace.kNoNamespace) {	//	Yes, a REFERENCE equality test is appropriate here.
					
					Tag.setNamespace(Ns);
					
				} else {
				
				//	There is no namespace definition corresponding to the namespace prefix attached to this tag's name.
				
				//	Note that without the following "restoreNamespacePrefixToName" invocation, the namespace prefix is 
				//	lost from the tag's name if the tag is ever converted to its mutable form, because the constructor 
				//	for the mutable form of the tag can only get the namespace prefix by invoking this tag's
				//	getNamespace().getName() method, which won't work, because getNamespace will return null.
					
				//	Since the namespace for the prefix can't be found, restore the prefix to the tag's name, so that it 
				//	won't be lost if the tag is converted to a mutable form later. 
					
					Tag.restoreNamespacePrefixToName();
				}
			}
		}
	}
	
	Namespace getNamespaceForPrefix
	(
		final Characters		Prefix
	)
	{
		for (int Index = Scopes.size(); --Index >= 0; ) {
			final Namespace			Ns = Scopes.get(Index).getNamespaceForPrefix(Prefix);
			
			if (Ns != null)
				return Ns;
		}
		
		return Namespace.kNoNamespace;	
	}
	
	private void push
	(
		final TagInfo			NewScope
	)
	{
		CurScope = NewScope;
		Scopes.push(NewScope);
	}
	
	private TagInfo pop() {
		final TagInfo			PoppedScope = Scopes.pop();
		
	//	Revert to the previous default namespace. (This works even in the case of the final end-tag because we
	//	pushed a fake TagInfo element onto Scopes in our constructor in order to define the "xml" namespace.)

		CurScope = Scopes.peek();
		
		return PoppedScope;
	}
	
	
	private static final class TagInfo {
		
		private final ImStartTag	Tag;
		private final Namespace[]	Namespaces;
		private final boolean		DefinesNamespaces;
		private Namespace			DefaultNamespace;
		
		
		TagInfo
		(
			final ImStartTag		Tag,
			final Namespace			CurrentDefaultNs
		)
		{
			this(Tag, CurrentDefaultNs, Tag.getDefinedNamespaces());
		}
		
		TagInfo
		(
			final ImStartTag		Tag,
			final Namespace			CurrentDefaultNs,
			final Namespace[]		Namespaces
		)
		{
			this.Tag               = Tag;
			this.Namespaces        = Namespaces;
			this.DefaultNamespace  = CurrentDefaultNs;
			this.DefinesNamespaces = Namespaces.length > 0;
			
		//	See if one of the attributes defines a new default namespace. If so, update the DefaultNamespace variable.
			
			for (final Namespace Ns : Namespaces) {
				
				if (Ns.isDefault()) {
					
					DefaultNamespace = Ns;
					break;
				}
			}		
		}
		
		ImStartTag getStartTag() {
			return Tag;
		}
		
		Namespace getDefaultNamespace() {
			return DefaultNamespace;
		}
		
		Namespace getNamespaceForPrefix
		(
			final Characters		Prefix	//	Can be null
		)
		{
			if (DefinesNamespaces) {
				
				if (Prefix == null) {
					
					return DefaultNamespace;	//	Will return null if the default namespace isn't defined here.
					
				} else {
					
					for (final Namespace Ns : Namespaces) {
						
						if (Prefix.equals(Ns.getName()))
							return Ns;
					}
				}
			}
			
			return null;
		}
		
		@Override public String toString() {
			return Tag == null ? "null" : Tag.toString();
		}
	}
}
