/*	QWICAP (Quick Web Interface for Conventional Applications) by Chris W. Johnson
	Copyright (C) 2003-2010 The University of Texas at Austin ("U.T. Austin"). All rights reserved.
	
	U.T. Austin's copyright notice in this evolving version reflects our original authorship. It does not assert 
	copyright over changes contributed by others.
	
	This library is free software; you can redistribute it and/or modify it under the terms of the GNU Lesser General 
	Public License as published by the Free Software Foundation; either version 2.1 of the License, or (at your option) 
	any later version.
	
	This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied 
	warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for 
	more details.
	
	You should have received a copy of the GNU Lesser General Public License along with this library; if not, write to 
	the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/


package edu.utexas.its.eis.tools.qwicap.template.xml.immutable;


import edu.utexas.its.eis.tools.qwicap.servlet.HTTPContentType;
import edu.utexas.its.eis.tools.qwicap.template.xml.TagException;
import edu.utexas.its.eis.tools.qwicap.template.xml.util.xcsd.XMLCharacterSetDetector;
import edu.utexas.its.eis.tools.qwicap.template.xml.structure.Attribute;
import edu.utexas.its.eis.tools.qwicap.util.Characters;

import java.io.*;
import java.util.logging.Level;
import java.util.logging.Logger;


/**	<code>XHTMLReader</code> handles the reading of XML/XHTML documents whose character sets are not known in advance, 
 * 	and therefore must be determined on-the-fly when the XML declaration, or, in the case of XHTML documents, the 
 * 	"<samp>&lt;<meta equiv='ContentType' content='....'/&gt;</samp> tag, is read. All characters up-to and including
 * 	the tag that specifies the character set are treated as members of the ISO-8859-1 character set. All characters 
 * 	afterward are treated as members of the specified character set, provided that that character set is recognized by 
 * 	Java. For background, see <a href='http://www.joelonsoftware.com/articles/Unicode.html'>The Absolute Minimum Every 
 * 	Software Developer Absolutely, Positively Must Know About Unicode and Character Sets (No Excuses!)</a>.
 * 
 * 	<p>Note that the read efficiency of this class is low until the character set declaration is discovered. Since
 * 	that discovery should take place very early in the reading of any document, this shouldn't be a problem. However,
 * 	if large documents that omit character set declarations are routinely read, the performance problem might 
 * 	become noticeable. The obvious fix is to make the documents well-formed by adding an XML declaration that includes
 * 	a character set specification.
 * 	</p>
 * 
 * @author Chris W. Johnson
 */
final class XHTMLReader extends Reader {

	private static final Logger		Log = Logger.getLogger(XHTMLReader.class.getName());

/**	Absent other markings, "UTF-8" is the default character set required by the XML specification. See: 
 * 	<a href='http://www.w3.org/TR/REC-xml/#charencoding'>http://www.w3.org/TR/REC-xml/#charencoding</a>.
 */
 	private static final String		DefaultCharacterSetName = "UTF-8";

	private static final Characters	MetaTagName             = new Characters("meta");
	private static final Characters HttpEquivAttrName       = new Characters("http-equiv");
	private static final Characters ContentAttrName         = new Characters("content");
	private static final Characters HeadTagName             = new Characters("head");
	
	static final int				kParseBufferSize        = 4 * 1024;
		
	private BufferedInputStream		In;
	private Reader					RIn;

	private boolean 				CharacterSetSpecifiedByDocument;
	private String					CharacterSetName;
	private String					CharacterSetTagStr;
	
	
	XHTMLReader
	(
		final String			MarkupName,
		final InputStream		OrigIn
	)
	throws IOException
	{
		super(OrigIn);
		
		In = new BufferedInputStream(OrigIn, kParseBufferSize);
		
		final XMLCharacterSetDetector CharSetInfo = new XMLCharacterSetDetector(MarkupName, In);		
		
		readMarkupForCharacterSetInfo(CharSetInfo, In);
		
		if (CharSetInfo.getCharSetDetectionIsDefinitive()) {
			
			if (CharacterSetSpecifiedByDocument) {
				
				if (CharSetInfo.getCharSetName().equalsIgnoreCase(CharacterSetName) == false) {
					
				//	UTF-16 and UTF-16BE are, apart from the presence/absence of a byte-order-mark, synonymous. Don't 
				//	complain about this case.
					
					if (bothStringsMatchOneOf(CharSetInfo.getCharSetName(), CharacterSetName, "UTF-16", "UTF-16BE") == false) {
						
						Log.log(Level.WARNING, "The markup \"{0}\" contains conflicting character set specifications. " +
							"The byte-order-mark (BOM) at the start of the markup identifies the character set as \"{1}\", " +
							"while the markup itself, in the tag \"{3}\", specifies the character set as \"{2}\". The " +
							"BOM will be honored. Please resolve this conflict in the markup.", 
							new Object[] { MarkupName, CharSetInfo.getCharSetName(), CharacterSetName, CharacterSetTagStr } 
						);
						
					//	Honor the BOM, like we said we would.
						
						CharacterSetName = CharSetInfo.getCharSetName();
					}
				}
				
			} else
				CharacterSetName = CharSetInfo.getCharSetName();
			
			RIn = CharSetInfo.getReader(In);
			
		} else {
		
			if (CharacterSetSpecifiedByDocument) {
				
				try {
					RIn = new InputStreamReader(In, CharacterSetName);
				} catch (UnsupportedEncodingException e) {
					throw new UnsupportedEncodingForXHTMLDocumentException(MarkupName, DefaultCharacterSetName, CharacterSetName, CharacterSetTagStr, e);
				}
				
			} else {
			
				Log.log(Level.INFO, "The markup \"{0}\" was searched for a character set specification, but no such " +
					"specification was found. Therefore, the markup was interpreted as if it was in the \"{1}\" character set. " +
					"The two best methods to specify the character set of an XML or XHTML document are: (1) Include " +
					"a byte-order-mark (BOM) as the first bytes of the markup. (This is particularly useful when dealing " +
					"with markup fragments.) (2) Include an XML declaration, for example \"<?xml version=\"1.0\" " +
					"encoding=\"UTF-8\"?>\", as the very first characters of the document (after any BOM, of course). " +
					"However, in XHTML documents it is also acceptable to supply a content type specification within " +
					"the \"head\" tag of the document, for example \"<meta http-equiv=''Content-Type'' content=''text/html; " +
					"charset=UTF-8''/>\". If more than one specification is supplied, Qwicap will first honor the BOM, " +
					"then the XML declaration, and finally the XHTML \"meta\" tag. Note that browsers may behave " +
					"differently, so all character set specifications should be kept consistent with each other in " +
					"order to ensure consistent results. For reference, see section C.9 of the XHTML 1.0 specification " +
					"at <http://www.w3.org/TR/2000/REC-xhtml1-20000126/#guidelines>, and appendix F of the XML 1.0 " +
					"specification at <http://www.w3.org/TR/REC-xml/#sec-guessing>.", 
					new Object[] { MarkupName, DefaultCharacterSetName }
				);
				
				RIn = new InputStreamReader(In, DefaultCharacterSetName);
			}
		}
	}
	
	private static boolean bothStringsMatchOneOf
	(
		final String			StringA,
		final String			StringB,
		final String...			PossibleValues
	)
	{
		return stringMatchesOneOf(StringA, PossibleValues) && stringMatchesOneOf(StringB, PossibleValues);
	}
	
	private static boolean stringMatchesOneOf
	(
		final String			TheString,
		final String...			PossibleValues
	)
	{
		for (final String PossibleMatch : PossibleValues) {
		
			if (TheString.equalsIgnoreCase(PossibleMatch))
				return true;
		}
		
		return false;
	}
	
	@Override public void close() {
		
		if (RIn != null) {

			try { RIn.close(); } catch (IOException e) { Log.log(Level.WARNING, "An exception was thrown while closing the Reader.", e); }
			RIn = null;
		}
		
		if (In != null) {
			
			try { In.close(); } catch (IOException e) { Log.log(Level.WARNING, "An exception was thrown while closing the InputStream.", e); }
			In = null;
		}
	}
	
	@Override public int read
	(
		final char[] 			Chars,
		final int				CharsOff,
		final int				CharCount
	)
	throws IOException
	{
		return RIn.read(Chars, CharsOff, CharCount);
	}
	
	boolean getCharacterSetSpecifiedByDocument() {
		return CharacterSetSpecifiedByDocument;
	}
	
	String getCharacterSetName() {
		return CharacterSetName;
	}
	
	private boolean readMarkupForCharacterSetInfo
	(
		final XMLCharacterSetDetector	CSD,
		final BufferedInputStream		In
	)
	throws IOException
	{
		final byte[]			LeadingBytes = new byte[kParseBufferSize];
		final int				BytesInBuff  = readLeadingBytes(In, LeadingBytes);
								
		return parseLeadingCharsForCharSet(CSD, LeadingBytes, BytesInBuff);
	}
	
	private static int readLeadingBytes
	(
		final BufferedInputStream	In,
		final byte[]				ParseBuff
	)
	throws IOException
	{	
		final int				ByteCount = ParseBuff.length;
		
		In.mark(ByteCount);
		
		int						TotalBytesRead = 0;
		int						BytesRead      = In.read(ParseBuff, TotalBytesRead, ByteCount);
		
		while (BytesRead >= 0) {
			
			TotalBytesRead += BytesRead;
			
			final int 				BytesToRead = ByteCount - TotalBytesRead;
			
			if (BytesToRead == 0)
				break;
			
			BytesRead = In.read(ParseBuff, TotalBytesRead, BytesToRead);
		}
		
		In.reset();
		
		return TotalBytesRead;
	}
	
	private boolean parseLeadingCharsForCharSet
	(
		final XMLCharacterSetDetector CSD,
		final byte[]					LeadingBytes,
		final int						ByteCount
	)
	throws IOException
	{
		final Reader			In             = CSD.getReader(new ByteArrayInputStream(LeadingBytes, 0, ByteCount));
		final int				ParseBuffLen   = ByteCount / CSD.getTypicalBytesPerCharacter();
		final char[]			ParseBuff      = new char[ParseBuffLen];
		int						TotalCharsRead = 0;
		int						CharsRead      = In.read(ParseBuff, TotalCharsRead, ParseBuffLen);
		
		while (CharsRead >= 0) {
			
			TotalCharsRead += CharsRead;
			if (TotalCharsRead >= ParseBuffLen)
				break;
			
			CharsRead = In.read(ParseBuff, TotalCharsRead, ParseBuffLen - TotalCharsRead);
		}
		
		return parseLeadingCharsForCharSet(ParseBuff, 0, TotalCharsRead);
	}

	private boolean parseLeadingCharsForCharSet
	(
		final char[]			Chars,
		int						Offset,
		final int				Length
	)
	{
		try {

			do {
				final ImRange			Thing = ImmutableMarkup.readOneRange(Chars, Offset, Length, TagHierarchyDummy.kTheOnlyInstanceYouWillEverNeed);
				
				if (Thing == null)
					break;
				
				Offset = Thing.PyEnd;
				
				if (Thing instanceof ImXMLDeclaration) {
					
					CharacterSetName = ((ImXMLDeclaration) Thing).getEncoding();
					if (CharacterSetName != null) {
						
						CharacterSetTagStr = Thing.toString();
						CharacterSetSpecifiedByDocument = true;
						return true;
					}
					
				} else if (Thing.isTagWithAttributes()) {
					
				//	We are looking for a tag like this: <META http-equiv="Content-Type" content="text/html; charset=UTF-8"/>
					
					final ImTagWithAttributes	Tag = (ImTagWithAttributes) Thing;

					if (Tag.nameEquals(MetaTagName)) {
						final Attribute				HttpEquivAttr = Tag.getAttribute(HttpEquivAttrName);

						if (HttpEquivAttr != null) {

							if ("Content-Type".equalsIgnoreCase(HttpEquivAttr.getValue().toString())) {
								final Attribute			ContentAttr = Tag.getAttribute(ContentAttrName);

								if (ContentAttr != null) {

									CharacterSetTagStr = Tag.toString();
									CharacterSetName   = new HTTPContentType(ContentAttr.getValue().toString()).getCharacterSet();
									CharacterSetSpecifiedByDocument = true;
									return true;
								}
							}
						}
					}

				} else if (Thing.isEndTag() && ((ImTag) Thing).getName().equals(HeadTagName)) {

				//	There was no character set specification in the document header. Give-up the search.
					
					CharacterSetName = DefaultCharacterSetName;
					return true;
				}
				
			} while (true);
			
		} catch (TagException e) {

		//	Usually, this exception will mean that the last tag in HeaderBuff has not yet been read fully.
		//
		//	It could also mean that the document isn't well-formed, but that's a problem to be dealt with by
		//	the ImmutableMarkup class.
		}

		return false;
	}

//	This is duplicated in MutableMarkup.

	static String getCharSetFromContentType
	(
		final String			ContentType
	)
	{
		if (ContentType == null)
			return null;
		
		return new HTTPContentType(ContentType).getCharacterSet();
	}
}
