/*	QWICAP (Quick Web Interface for Conventional Applications) by Chris W. Johnson
	Copyright (C) 2003-2011 The University of Texas at Austin ("U.T. Austin"). All rights reserved.
	
	U.T. Austin's copyright notice in this evolving version reflects our original authorship. It does not assert 
	copyright over changes contributed by others.
	
	This library is free software; you can redistribute it and/or modify it under the terms of the GNU Lesser General 
	Public License as published by the Free Software Foundation; either version 2.1 of the License, or (at your option) 
	any later version.
	
	This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied 
	warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for 
	more details.
	
	You should have received a copy of the GNU Lesser General Public License along with this library; if not, write to 
	the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/


package edu.utexas.its.eis.tools.qwicap.template.xml.immutable;


import edu.utexas.its.eis.tools.qwicap.servlet.Qwicap;
import edu.utexas.its.eis.tools.qwicap.template.xml.TagException;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.StringReader;
import java.net.URL;
import java.util.Hashtable;
import java.util.logging.Level;
import java.util.logging.Logger;


/**	Provides a thread-safe cache of immutable XML documents.
 * 
 *	@author Chris W. Johnson
 * 	@deprecated While this class may be useful to some applications, it is no longer used by Qwicap, whose 
 *              {@link Qwicap#getDocument(URL)} and {@link Qwicap#getDocument(String)} methods now use an internal, web 
 *              app specific, LRU cache, that is pre-sized by a census of the documents present in a web app. It is 
 *              believed that applications not served by the caching built into those methods would be better off 
 *              implementing one appropriate to their own caching needs and/or policies, because this implememtation 
 *              creates a global (static) cache from which documents are never removed. Since any one web app tends to 
 *              deal with a modest number of documents, typically all of which are built-into the web app, this was OK 
 *              before Qwicap was required to be able to function as a shared library, but it is now an opportunity for 
 *              memory leaks, both because it can hold references to items in undeployed web apps, and because, 
 *              potentially being shared by any number of web apps, its growth is unbounded.
 */
@Deprecated public final class XMLCache {
	
	private static final Logger								Log   = Logger.getLogger(XMLCache.class.getName());
	
	private static final Hashtable<String, XMLCacheEntry>	Cache = new Hashtable<String, XMLCacheEntry>();	//	Contains XMLDocument objects.
	private static final XMLDocument						DocNotFound;
	
	static {	
		try {
			DocNotFound = new XMLDocument(new StringReader("Document not found."), 25);
		} catch (Exception e) {
			throw new RuntimeException(e);
		}
	}
	
		
	public static XMLDocument getDocument
	(
		final File				DocRef
	)
	throws TagException, IOException
	{
		return getDocument(DocRef.getAbsolutePath());
	}
	
	public static XMLDocument getDocument
	(
		final String			DocName
	)
	throws TagException, IOException
	{
		final XMLCacheEntry		XCE = Cache.get(DocName);
		XMLDocument				Doc = XCE != null ? XCE.getXMLDocument() : null;
		
		if (Doc == DocNotFound) {
			
			Log.log(Level.FINE, "Document \"{0}\" was not found on the first occasion it was requested. That fact was cached, and a FileNotFoundException is being thrown on that basis.", DocName);
			throw new FileNotFoundException("Document \"" + DocName + "\" was not found on the first occasion it was requested. That fact was cached, and is being reported on that basis.");
		}
		
		try {
			
			if (Doc == null) {
				
				Doc = new XMLDocument(new File(DocName));
				Cache.put(DocName, new XMLCacheEntry(Doc));
				
				Log.log(Level.FINE, "XML document \"{0}\" has been read from disk for the first time, and cached.", DocName);
				
			} else {
				
			//	If the document on disk has been modified since a representation of it was placed in the cache,
			//	update the cache.
				
				final XMLDocument		CurDoc = Doc.getLatest();
				
				if (CurDoc != Doc) {
					
					Cache.put(DocName, new XMLCacheEntry(CurDoc));
					Doc = CurDoc;
					
					Log.log(Level.FINE, "XML document \"{0}\" has been reloaded from disk, and cached.", DocName);
				
				} else
					Log.log(Level.FINE, "XML document \"{0}\" was found in the cache.", DocName);
			}
			
			return Doc;
			
		} catch (FileNotFoundException e) {
			
		//	Record in the cache the fact that this document does not exist.
			
			Log.log(Level.FINE, "XML document \"{0}\" was not found. That fact has been cached for future reference.", DocName);
			
			Cache.put(DocName, new XMLCacheEntry(DocNotFound));
			throw e;
		}
	}
	
	public static XMLDocument getDocument
	(
		final URL				DocURL
	)
	throws TagException, IOException
	{
		final String			DocIDStr = DocURL.toString();
		final XMLCacheEntry		XCE      = Cache.get(DocIDStr);
		XMLDocument				Doc      = XCE != null ? XCE.getXMLDocument() : null;
		
		if (Doc == DocNotFound) {
			
			Log.log(Level.FINE, "Document <{0}> was not found on the first occasion it was requested. That fact was cached, and a FileNotFoundException is being thrown on that basis.", DocIDStr);
			throw new FileNotFoundException("Document <" + DocURL + "> was not found on the first occasion it was requested. That fact has been cached, and is being reported on that basis.");
		}
		
		try {
			
			if (Doc == null) {
				
				Doc = new XMLDocument(DocURL);
				Cache.put(DocIDStr, new XMLCacheEntry(Doc));
				
				Log.log(Level.FINE, "XML document <{0}> has been read, and permanently cached.", DocIDStr);
				
			} else
				Log.log(Level.FINE, "XML document <{0}> was found in the cache.", DocIDStr);
			
			return Doc;
	
		} catch (FileNotFoundException e) {
			
		//	Record in the cache the fact that this document does not exist.
			
			Log.log(Level.FINE, "XML document <{0}> was not found. That fact has been cached for future reference.", DocIDStr);
			
			Cache.put(DocIDStr, new XMLCacheEntry(DocNotFound));
			throw e;
		}
	}
}

