/*	QWICAP (Quick Web Interface for Conventional Applications) by Chris W. Johnson
	Copyright (C) 2003-2011 The University of Texas at Austin ("U.T. Austin"). All rights reserved.
	
	U.T. Austin's copyright notice in this evolving version reflects our original authorship. It does not assert 
	copyright over changes contributed by others.
	
	This library is free software; you can redistribute it and/or modify it under the terms of the GNU Lesser General 
	Public License as published by the Free Software Foundation; either version 2.1 of the License, or (at your option) 
	any later version.
	
	This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied 
	warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for 
	more details.
	
	You should have received a copy of the GNU Lesser General Public License along with this library; if not, write to 
	the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/


//  XMLDocument.java
//  qwicap
//
//  Created by Chris W. Johnson on Fri Jun 06 2003.


package edu.utexas.its.eis.tools.qwicap.template.xml.immutable;


import edu.utexas.its.eis.tools.qwicap.template.xml.*;
import edu.utexas.its.eis.tools.qwicap.util.Characters;
import edu.utexas.its.eis.tools.qwicap.servlet.HTTPContentType;
import edu.utexas.its.eis.tools.qwicap.servlet.XMLContentType;
import edu.utexas.its.eis.tools.qwicap.servlet.ContentType;

import java.io.*;
import java.net.URL;
import java.net.URLConnection;
import java.net.HttpURLConnection;


/** <code>XMLDocument</code> creates a body of immutable markup from the XML document specified by a <code>File</code> 
 *	or <code>URL</code> object, or, at a lower level, any <code>InputStream</code> or <code>Reader</code> object. 
 *	Whatever the source of the XML markup, it should be well-formed, in the sense that for every "start" tag there is a 
 *	corresponding "end" tag. Other than that, <code>XMLDocument</code> doesn't care what you feed to it, whether that 
 *	be tiny fragments of XML, or entire documents. Be aware, however, that fragments will typically lack any of the 
 * 	markup necessary to identify their character set, which means that they cannot be read in a manner that is assured
 * 	to be correct unless they are read using one of the constructors that accepts a parameter specifying the character
 * 	set. If you have any doubts about the significance of character set issues, please read Joel Spolsky's article
 * 	<a href='http://www.joelonsoftware.com/articles/Unicode.html'><cite>The Absolute Minimum Every Software Developer 
 * 	Absolutely, Positively Must Know About Unicode and Character Sets (No Excuses!)</cite></a>.
 *	
 *	@author Chris W. Johnson
 */
public class XMLDocument extends ImmutableMarkup {
	
	private final File			XDoc;
	private final long			XDocModified;
	private final int			XDocLength;		//	The length of the document, in bytes. This assists in determining when the document on disk has changed.
	
	
/**	Reads an XML document from a file. The character set of the document read by this constructor is determined by 
 * 	either the XML declaration (for example: "<samp>&lt;?xml version="1.0" encoding="UTF-8"?&gt;</samp>") at the 
 * 	beginning of the document, or by a "Content-Type" meta tag (for example: "<samp>&lt;meta http-equiv='Content-Type' 
 * 	content='text/html; charset=UTF-8'/&gt;</samp>") within the "head" tag, assuming the document is XHTML. If both 
 * 	character set specifications are present, the one in the XML declaration takes precedence. To programmatically 
 * 	specify the document's character set, use the {@link #XMLDocument(File, String)} constructor.
 * 
 * @param XDoc						The XML file that should be read.
 * @throws TagException				If there was a parsing error.
 * @throws IOException				If there was an I/O error.
 */	
	public XMLDocument
	(
		final File				XDoc
	)
	throws TagException, IOException
	{
		this(XDoc, null);
	}
	
/**	Reads an XML document from a file. The file is interpreted as though it is encoded in the specified character set, 
 * 	no matter what the document may, or may not, claim. If your document specifies its character set in its markup, and
 * 	you prefer to have that specification honored, use the {@link #XMLDocument(File)} constructor instead.
 * 
 * @param XDoc						The XML file that should be read.
 * @param Charset					The character set of the XML file, or <code>null</code> if the character set should
 * 									be determined by examining the XML file.
 * @throws TagException				If there's a parsing error.
 * @throws IOException				If there's an I/O error.
 */	
	public XMLDocument
	(
		final File				XDoc,
		final String			Charset		//	Can be null.
	)
	throws TagException, IOException
	{
	//	If XDoc doesn't exist, manually throw a more informative FileNotFoundException than would be thrown 
	//	automatically.
		
		if (XDoc.exists() == false)	
			throw new FileNotFoundException(XDoc.getAbsolutePath());
		
		this.XDoc         = XDoc;
		this.XDocModified = XDoc.lastModified();
		this.XDocLength   = (int) XDoc.length();
		
		setMarkupName(XDoc.getAbsolutePath());
		
		if (Charset != null)
			initFromReader(new InputStreamReader(new FileInputStream(XDoc), Charset), true, XDocLength);			
		else
			initFromReader(new XHTMLReader(XDoc.getAbsolutePath(), new FileInputStream(XDoc)), true, XDocLength);
	}
	
/**	Reads an XML document from a URL. The URL object's internal connection to the document specified by the URL is 
 * 	closed after the document has been read. This constructor does <i>not</i> check the 
 *	<a href='http://www.iana.org/assignments/media-types/'>MIME type</a> of the document specified by the URL, so you 
 *	can't receive any MIME-type-mismatch surprises. Conversely, though, if you're passing arbitrary URLs
 *	to this constructor, be aware that discovering that a document does <i>not</i> contain XML by feeding it to
 *	this constructor and seeing whether or not it throws an exception, is a needlessly expensive process. In that
 * 	situation, you should be checking the MIME type, and using the <code>XMLDocument(URLConnection)</code> constructor, 
 *	if the type is appropriate (most XHTML documents arrive as type "text/html").
 * 
 * @param DocURL					The URL of the XML document to read.
 * @throws TagException				If there's a parsing error.
 * @throws IOException				If there's an I/O error.
 */	
	public XMLDocument
	(
		final URL				DocURL
	)
	throws TagException, IOException
	{	
		this(DocURL.openConnection());
	}
	
/**	Reads an XML document from a <code>URLConnection</code>. The <code>URLConnection</code> object's 
 * 	<code>InputStream</code> is closed after the document has been read. The character set of the document read by this 
 * 	constructor is, ideally, determined by the connection's content type (for example: "<samp>text/html; charset=UTF-8</samp>"), 
 * 	as returned by the {@link URLConnection#getContentType} method. If the character set is not specfied by the content
 * 	type, then it is determined by either the XML byte-order-mark (BOM), the XML declaration (for example: 
 * 	"<samp>&lt;?xml version="1.0" encoding="UTF-8"?&gt;</samp>") at the beginning of the document, or by a 
 * 	"Content-Type" meta tag (for example: "<samp>&lt;meta http-equiv='Content-Type' content='text/html; charset=UTF-8'/&gt;</samp>") 
 * 	within the "head" tag, assuming the document is XHTML. If both character set specifications are present, the one in 
 * 	the XML declaration takes precedence. To programmatically specify the document's character set, get the 
 * 	<code>URLConnection</code> instance's <code>InputStream</code> instance and wrap it in an instance of 
 * 	{@link InputStreamReader#InputStreamReader(InputStream, String) InputStreamReader(InputStream, String)}, then pass 
 * 	that <code>Reader</code> to the {@link XMLDocument#XMLDocument(Reader, long) XMLDocument(Reader, int)} constructor.
 * 
 * @param Conn						A <code>URLConnection</code> from which the XML document can be read.
 * @throws TagException				If there's a parsing error.
 * @throws IOException				If there's an I/O error.
 */	
	public XMLDocument
	(
		final URLConnection		Conn
	)
	throws TagException, IOException
	{
		XDoc         = null;
		XDocModified = 0;
		XDocLength   = 0;
		
		setMarkupName(Conn.getURL().toString());
		
		final ContentType		ConType;
		
		if (Conn instanceof HttpURLConnection)
			ConType = new HTTPContentType(Conn.getContentType());
		else
			ConType = new XMLContentType(Conn.getContentType());
						
		final int				Length = Conn.getContentLength();
		final InputStream		In     = Conn.getInputStream();
		
		if (ConType.getCharacterSetWasSpecified())
			initFromReader(new InputStreamReader(In, ConType.getCharacterSet()), true, Length);			
		else
			initFromReader(new XHTMLReader(getMarkupName(), In), true, Length);
	}
	
/**	Reads an XML document from an <code>InputStream</code>. The <code>InputStream</code> is <i>not</i> closed after the
 * 	document has been read.	The character set of the document read by this constructor is determined by either the XML 
 * 	declaration (for example: "<samp>&lt;?xml version="1.0" encoding="UTF-8"?&gt;</samp>") at the beginning of the document, or by
 * 	a "Content-Type" meta tag (for example: "<samp>&lt;meta http-equiv='Content-Type' content='text/html; charset=UTF-8'/&gt;</samp>") 
 * 	within the "head" tag, assuming the document is XHTML. If both character set specifications are present, the 
 * 	one in the XML declaration takes precedence. To programmatically specify the document's character set, wrap the
 * 	<code>InputStream</code> instance in an instance of {@link InputStreamReader#InputStreamReader(InputStream, String)
 * 	InputStreamReader(InputStream, String)} and pass that <code>Reader</code> to the 
 * 	{@link XMLDocument#XMLDocument(Reader, long) XMLDocument(Reader, int)} constructor.
 * 
 * @param In					A <code>InputStream</code> from which the XML document can be read.
 * @param LengthHint			The length of the XML document, or your best guess. If you can't even guess, pass zero
 *								or a negative value.
 * @throws TagException			If there's a parsing error.
 * @throws IOException			If there's an I/O error.
 */
	public XMLDocument
	(
		final InputStream		In,
		final long				LengthHint	//	Set to zero, or negative, if you have no idea how much there is to read from In.
	)
	throws TagException, IOException
	{
		this(new XHTMLReader("InputStream", In), LengthHint, null);
	}
	
/**	Reads an XML document from an <code>InputStream</code>. The <code>InputStream</code> is <i>not</i> closed after the
 * 	document has been read.	The character set of the document read by this constructor is determined by either the XML 
 * 	declaration (for example: "<samp>&lt;?xml version="1.0" encoding="UTF-8"?&gt;</samp>") at the beginning of the 
 * 	document, or by a "Content-Type" meta tag (for example: 
 * 	"<samp>&lt;meta http-equiv='Content-Type' content='text/html; charset=UTF-8'/&gt;</samp>") 
 * 	within the "head" tag, assuming the document is XHTML. If both character set specifications are present, the 
 * 	one in the XML declaration takes precedence. To programmatically specify the document's character set, wrap the
 * 	<code>InputStream</code> instance in an instance of {@link InputStreamReader#InputStreamReader(InputStream, String)
 * 	InputStreamReader(InputStream, String)} and pass that <code>InputStreamReader</code> to the 
 * 	{@link XMLDocument#XMLDocument(Reader, long) XMLDocument(Reader, int)} constructor.
 * 
 * @param In					A <code>InputStream</code> from which the XML document can be read.
 * @param LengthHint			The length of the XML document, or your best guess. If you can't even guess, pass zero
 *								or a negative value.
 * @param DocumentName			The name to be associated with this document, or <code>null</code>. If not 
 * 								<code>null</code>, this name is used to make error messages more informative, and is 
 * 								available from the {@link #getMarkupName} method.
 * @throws TagException			If there's a parsing error.
 * @throws IOException			If there's an I/O error.
 */
	public XMLDocument
	(
		final InputStream		In,
		final long				LengthHint,	//	Set to zero, or negative, if you have no idea how much there is to read from In.
		final String			DocumentName
	)
	throws TagException, IOException
	{
		this(new XHTMLReader("InputStream", In), LengthHint, DocumentName);
	}
	
/**	Reads an XML document from a <code>Reader</code>. The <code>Reader</code> is <i>not</i> closed after the
 * 	document has been read.	Note that when you create the <code>Reader</code> you should explicitly specify the 
 * 	document's character set (see {@link InputStreamReader#InputStreamReader(InputStream, String) 
 * 	InputStreamReader(InputStream, String)}), in order to ensure consistent behavior across platforms.
 * 
 * @param In					A <code>Reader</code> from which the XML document can be read.
 * @param LengthHint			The length of the XML document, or your best guess. If you can't even guess, pass zero
 *								or a negative value.
 * @throws TagException			If there's a parsing error.
 * @throws IOException			If there's an I/O error.
 */
	public XMLDocument
	(
		final Reader			In,
		final long				LengthHint	//	Set to zero, or negative, if you have no idea how much there is to read from In.
	)
	throws TagException, IOException
	{
		this(In, LengthHint, null);
	}
	
/**	Reads an XML document from a <code>Reader</code>. The <code>Reader</code> is <i>not</i> closed after the
 * 	document has been read.	Note that when you create the <code>Reader</code> you should explicitly specify the 
 * 	document's character set (see {@link InputStreamReader#InputStreamReader(InputStream, String) 
 * 	InputStreamReader(InputStream, String)}), in order to ensure consistent behavior across platforms.
 * 
 * @param In					A <code>Reader</code> from which the XML document can be read.
 * @param LengthHint			The number of characters in the XML document, or your best guess. If you can't even 
 * 								guess, pass zero or a negative value.
 * @param DocumentName			The name to be associated with this document, or <code>null</code>. If not 
 * 								<code>null</code>, this name is used to make error messages more informative, and is 
 * 								available from the {@link #getMarkupName} method.
 * @throws TagException			If there's a parsing error.
 * @throws IOException			If there's an I/O error.
 */
	public XMLDocument
	(
		final Reader			In,
		final long				LengthHint,	//	Set to zero, or negative, if you have no idea how much there is to read from In.
		final String			DocumentName
	)
	throws TagException, IOException
	{
		XDoc         = null;
		XDocModified = 0;
		XDocLength   = 0;
		
		setMarkupName(DocumentName);
		initFromReader(In, false, LengthHint);
	}
	
	private void initFromReader
	(
		final Reader			In,
		final boolean			CloseIn,
		final long				LengthHint	//	Set to zero, or negative, if you have no idea how much there is to read from In.
	)
	throws TagException, IOException
	{
		try {
			final int				BuffLen;
			
			if (LengthHint <= 0)
				BuffLen = 64 * 1024;
			else if (LengthHint >= 1024 * 1024)
				BuffLen = 1024 * 1024;
			else
				BuffLen = (int) LengthHint;
			
			final CharArrayWriter	CharsOut  = new CharArrayWriter(BuffLen);
			final char[]			Buff      = new char[BuffLen];
			int						CharsRead = In.read(Buff);
			
			while (CharsRead != -1) {
				
				CharsOut.write(Buff, 0, CharsRead);
				CharsRead = In.read(Buff);
			}
			
			read(new Characters(CharsOut.toCharArray()));
			
			if (In instanceof XHTMLReader)
				setCharacterSet(((XHTMLReader) In).getCharacterSetName());
			else if (In instanceof InputStreamReader)
				setCharacterSet(((InputStreamReader) In).getEncoding());
			
		} finally {
			
			if (CloseIn) 
				In.close();
		}
	}
	
/**	If this <code>XMLDocument</code> read its markup from a file, and that file has changed since this immutable 
 *	representation was created, this method returns a new object representing the current contents of the file.
 *	If the file has not been modified, this object is returned. If this <code>XMLDocument</code> read its markup from a 
 *	<code>URL</code>, <code>InputStream</code> or <code>Reader</code>, nothing happens, and this object is returned. 
 *	That's because, in the case of <code>URL</code>s, it is presumed to be too expensive to check for updates, and in 
 *	the case of <code>InputStream</code>s and <code>Reader</code>s, there's no way this class can determine the original 
 *	source of the material in order to open a new <code>InputStream</code> or <code>Reader</code> to re-read it (also, 
 *	re-reading a document to determine whether, or not, it has been changed wouldn't be efficient).
 *
 * @throws TagException			If there's a parsing error.
 * @throws IOException			If there's an I/O error.
 * @return						The latest version of this <code>XMLDocument</code>.
 */	
	public XMLDocument getLatest() throws TagException, IOException {
		
		if (XDoc != null && (XDoc.lastModified() != XDocModified || XDoc.length() != XDocLength))
			return new XMLDocument(XDoc);
		
		return this;
	}
	
	@Override public XMLDocument clone() throws CloneNotSupportedException {
		
	//	This implementation is present only in the interest of code-clarity.
	//
	//	A shallow copy is fine here.
		
		return (XMLDocument) super.clone();	
	}
}


