/*	QWICAP (Quick Web Interface for Conventional Applications) by Chris W. Johnson
	Copyright (C) 2003-2006 The University of Texas at Austin ("U.T. Austin"). All rights reserved.
	
	U.T. Austin's copyright notice in this evolving version reflects our original authorship. It does not assert 
	copyright over changes contributed by others.
	
	This library is free software; you can redistribute it and/or modify it under the terms of the GNU Lesser General 
	Public License as published by the Free Software Foundation; either version 2.1 of the License, or (at your option) 
	any later version.
	
	This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied 
	warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for 
	more details.
	
	You should have received a copy of the GNU Lesser General Public License along with this library; if not, write to 
	the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/


//  XMLString.java
//  qwicap
//
//  Created by Chris W. Johnson on Thu Jun 26 2003.


package edu.utexas.its.eis.tools.qwicap.template.xml.immutable;


import edu.utexas.its.eis.tools.qwicap.template.xml.*;
import edu.utexas.its.eis.tools.qwicap.util.Characters;


/**	<code>XMLString</code> creates a body of immutable markup from any <code>String</code> containing XML markup, 
 *	provided that the string is well-formed, in the sense that for every "start" tag there is a corresponding "end" tag.
 * 	Other than that, <code>XMLString</code> doesn't care what you feed to it, whether that be tiny fragments of XML, or 
 *	entire documents.
 *	
 *	@author Chris W. Johnson
 */
public class XMLString extends ImmutableMarkup {
	
	private final int			Length;
	
	
/**	Creates a body of immutable markup from any object whose <code>toString</code> method will produce a 
 *	<code>String</code> of XML markup, whether that string is a well-formed fragment, or an entire XML document.
 * 
 * @param XMLStringProducer		A string of XML, either a well-formed fragment, or an entire XML document.
 * @throws TagException			If there's a problem parsing the supplied string as XML, validating its tag
 *								hierarchy, or any number of related problems.
 */	
	public XMLString
	(
		final Object			XMLStringProducer
	)
	throws TagException
	{
		final Characters		Chars = new Characters(XMLStringProducer.toString());
		
		Length = Chars.length();
		
		read(Chars, false);
	}
	
/**	Creates a body of immutable markup from any <code>String</code> of XML markup, whether that string is a 
 *	well-formed fragment, or an entire XML document.
 * 
 * @param XStr					A string of XML, either a well-formed fragment, or an entire XML document.
 * @throws TagException			If there's a problem parsing the supplied string as XML, validating its tag
 *								hierarchy, or any number of related problems.
 */	
	public XMLString
	(
		final String			XStr
	)
	throws TagException
	{
		final Characters		Chars = new Characters(XStr);
		
		Length = Chars.length();
		
		read(Chars, false);
	}
	
/**	Creates a body of immutable markup from any object whose <code>toString</code> method will produce a 
 *	<code>String</code> of XML markup, whether that string is a well-formed fragment, or an entire XML document.
 * 
 * @param XMLStringProducer		A string of XML, either a well-formed fragment, or an entire XML document.
 * @param ResolveNamespaces		<code>true</code> if the namespaces of the elements in this fragment should be 
 *								determined.
 * @throws TagException			If there's a problem parsing the supplied string as XML, validating its tag
 *								hierarchy, or any number of related problems.
 */	
	public XMLString
	(
		final Object			XMLStringProducer,
		final boolean			ResolveNamespaces
	)
	throws TagException
	{
		final Characters		Chars = new Characters(XMLStringProducer.toString());
		
		Length = Chars.length();
		
		read(Chars, ResolveNamespaces);
	}
	
/**	Creates a body of immutable markup from any <code>String</code> of XML markup, whether that string is a 
 *	well-formed fragment, or an entire XML document.
 * 
 * @param XStr					A string of XML, either a well-formed fragment, or an entire XML document.
 * @param ResolveNamespaces		<code>true</code> if the namespaces of the elements in this fragment should be 
 *								determined.
 * @throws TagException			If there's a problem parsing the supplied string as XML, validating its tag
 *								hierarchy, or any number of related problems.
 */	
	public XMLString
	(
		final String			XStr,
		final boolean			ResolveNamespaces
	)
	throws TagException
	{
		final Characters		Chars = new Characters(XStr);
		
		Length = Chars.length();
		
		read(Chars, ResolveNamespaces);
	}
	
	public int length() {
		return Length;
	}
}



