/*	QWICAP (Quick Web Interface for Conventional Applications) by Chris W. Johnson
	Copyright (C) 2003-2010 The University of Texas at Austin ("U.T. Austin"). All rights reserved.
	
	U.T. Austin's copyright notice in this evolving version reflects our original authorship. It does not assert 
	copyright over changes contributed by others.
	
	This library is free software; you can redistribute it and/or modify it under the terms of the GNU Lesser General 
	Public License as published by the Free Software Foundation; either version 2.1 of the License, or (at your option) 
	any later version.
	
	This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied 
	warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for 
	more details.
	
	You should have received a copy of the GNU Lesser General Public License along with this library; if not, write to 
	the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/


package edu.utexas.its.eis.tools.qwicap.template.xml.mutable;


import edu.utexas.its.eis.tools.qwicap.util.Characters;
import edu.utexas.its.eis.tools.qwicap.util.HTMLEntityCodec;


/**	Utility methods for encoding and decoding attribute values and names. This is not part of Qwicap's public API, but
 *	it is used in several internal packages, and therefore can't be properly hidden.
 *
 *	@author Chris W. Johnson
 */
public final class AttributeUtilities {
	
	private static final char[]		kNoCharacters         = {};
	
	static final HTMLEntityCodec	DoubleQuoteNonEncoder = new HTMLEntityCodec(false);
	static final HTMLEntityCodec	DoubleQuoteEncoder    = new HTMLEntityCodec(true);
	
	static {
	
	//	Instruct the "normal" encoder not to encode double quote characters.
	
		DoubleQuoteNonEncoder.setEncoding('"', null);
	}
	
/**	Encodes a would-be attribute name such that it uses only characters that are permitted in XML attribute names -- this 
 * 	method currently does nothing because I've never seen anyone use an invalid attribute name, and therefore I'm
 * 	concerned about the cost/benefit ratio of performing this encoding every time an attribute name is passed
 * 	into Qwicap's public API. This method is used appropriately throughout Qwicap's implementation, so the 
 *	encoding functionality can be easily enabled by uncommenting the commented-out portion of this method.
 * 
 * @param UnName				The would-be attribute name to be encoded, if encoding is required.
 * @return						The attribute name, encoded as necessary.
 */	
	public static String encodeAttributeName
	(
		final String			UnName
	)
	{
		return UnName;
	
	//	There's nothing wrong with the following line, but I'm concerned that its cumulative cost in execution 
	//	speed and memory usage is not proportional to its actual value, given that, as best I can tell, people 
	//	basically never use attribute names that require encoding/decoding.	
	//
	//	return MutableMarkup.HTML.encode(UnName);
	}
	
/**	Encodes a would-be attribute name such that it uses only characters that are permitted in XML attribute names -- this 
 * 	method currently does nothing because I've never seen anyone use an invalid attribute name, and therefore I'm
 * 	concerned about the cost/benefit ratio of performing this encoding every time an attribute name is passed
 * 	into Qwicap's public API. This method is used appropriately throughout Qwicap's implementation, so the 
 *	encoding functionality can be easily enabled by uncommenting the commented-out portion of this method.
 * 
 * @param UnName				The would-be attribute name to be encoded, if encoding is required.
 * @return						The attribute name, encoded as necessary.
 */	
	public static Characters encodeAttributeName
	(
		final Characters		UnName
	)
	{
		return UnName;	
	
	//	There's nothing wrong with the following line, but I'm concerned that its cumulative cost in execution 
	//	speed and memory usage is not proportional to its actual value, given that, as best I can tell, people 
	//	basically never use attribute names that require encoding/decoding.	
	//
	//	return MutableMarkup.HTML.encode(UnName);
	}
	
/**	Decodes an attribute name into normal, unencoded Unicode representations of all of its characters -- this 
 * 	method currently does nothing because I've never seen anyone use an invalid attribute name, and therefore I'm
 * 	concerned about the cost/benefit ratio of performing this decoding every time an attribute name is passed
 * 	out of Qwicap's public API. This method is used appropriately throughout Qwicap's implementation, so the 
 *	decoding functionality can be easily enabled by uncommenting the commented-out portion of this method.
 * 
 * @param EncName				The encoded attribute name to be decoded, if decoding is required.
 * @return						The attribute name, decoded as necessary.
 */	
	public static String decodeAttributeName
	(
		final String			EncName
	)
	{	
		return EncName;
	
	//	There's nothing wrong with the following line, but I'm concerned that its cumulative cost in execution 
	//	speed and memory usage is not proportional to its actual value, given that, as best I can tell, people 
	//	basically never use attribute names that require encoding/decoding.	
	//
	//	return MutableMarkup.HTML.decode(EncName);
	}
	
/**	Decodes an attribute name into normal, unencoded Unicode representations of all of its characters -- this 
 * 	method currently does nothing because I've never seen anyone use an invalid attribute name, and therefore I'm
 * 	concerned about the cost/benefit ratio of performing this decoding every time an attribute name is passed
 * 	out of Qwicap's public API. This method is used appropriately throughout Qwicap's implementation, so the 
 *	decoding functionality can be easily enabled by uncommenting the commented-out portion of this method.
 * 
 * @param EncName				The encoded attribute name to be decoded, if decoding is required.
 * @return						The attribute name, decoded as necessary.
 */	
	public static Characters decodeAttributeName
	(
		final Characters		EncName
	)
	{
		return EncName;
	
	//	There's nothing wrong with the following line, but I'm concerned that its cumulative cost in execution 
	//	speed and memory usage is not proportional to its actual value, given that, as best I can tell, people 
	//	basically never use attribute names that require encoding/decoding.	
	//
	//	return MutableMarkup.HTML.decode(EncName);
	}
	
/**	Encodes a would-be attribute value such that it uses only characters that are permitted in XML attribute values.
 *	Note that unlike the current versions of the <code>encodeAttributeName</code> methods, this method actually does 
 *	perform encoding.
 * 
 * @param UnValue				The would-be attribute value to be encoded, if encoding is required.
 * @return						The attribute value, encoded as necessary.
 */	
	public static String encodeAttributeValue
	(
		final String			UnValue
	)
	{
		if (UnValue == null)
			return "";
		
		final boolean			HasSingleQuote = UnValue.indexOf('\'') >= 0;
		final boolean			HasDoubleQuote = UnValue.indexOf('"') >= 0;
		HTMLEntityCodec			Encoder        = DoubleQuoteNonEncoder;
		
		if (HasSingleQuote && HasDoubleQuote) {
			
		//	Cause the double quote characters in the value to encoded, so we can use double quotes to bound the value.
			
			Encoder = DoubleQuoteEncoder;
		}
		
		return Encoder.encode(UnValue);
	}
	
/**	Encodes a would-be attribute value such that it uses only characters that are permitted in XML attribute values.
 *	Note that unlike the current versions of the <code>encodeAttributeName</code> methods, this method actually does 
 *	perform encoding.
 * 
 * @param UnValue				The would-be attribute value to be encoded, if encoding is required.
 * @return						The attribute value, encoded as necessary.
 */	
	public static Characters encodeAttributeValue
	(
		final Characters		UnValue	//	Unicode without any HTML entity encoding
	)
	{
		if (UnValue == null)
			return new Characters(kNoCharacters);
		
		final boolean			HasSingleQuote = UnValue.indexOf('\'') >= 0;
		final boolean			HasDoubleQuote = UnValue.indexOf('"') >= 0;
		HTMLEntityCodec			Encoder        = DoubleQuoteNonEncoder;
		
		if (HasSingleQuote && HasDoubleQuote) {
			
		//	Cause the double quote characters in the value to encoded, so we can use double quotes to bound the value.
			
			Encoder = DoubleQuoteEncoder;
		}
		
		return Encoder.encode(UnValue);
	}
	
/**	Decodes an attribute value into normal, unencoded Unicode representations of all of its characters.
 *	Note that unlike the current versions of the <code>decodeAttributeName</code> methods, this method actually does 
 *	perform decoding.
 * 
 * @param EncValue				The encoded attribute value to be decoded, if decoding is required.
 * @return						The attribute value, decoded as necessary.
 */	
	public static String decodeAttributeValue
	(
		final String			EncValue
	)
	{
		return MutableMarkup.htmlDecode(EncValue);
	}
	
/**	Decodes an attribute value into normal, unencoded Unicode representations of all of its characters.
 *	Note that unlike the current versions of the <code>decodeAttributeName</code> methods, this method actually does 
 *	perform decoding.
 * 
 * @param EncValue				The encoded attribute value to be decoded, if decoding is required.
 * @return						The attribute value, decoded as necessary.
 */	
	public static Characters decodeAttributeValue
	(
		final Characters		EncValue
	)
	{
		return MutableMarkup.htmlDecode(EncValue);
	}
}
