/*	QWICAP (Quick Web Interface for Conventional Applications) by Chris W. Johnson
	Copyright (C) 2003-2010 The University of Texas at Austin ("U.T. Austin"). All rights reserved.
	
	U.T. Austin's copyright notice in this evolving version reflects our original authorship. It does not assert 
	copyright over changes contributed by others.
	
	This library is free software; you can redistribute it and/or modify it under the terms of the GNU Lesser General 
	Public License as published by the Free Software Foundation; either version 2.1 of the License, or (at your option) 
	any later version.
	
	This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied 
	warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for 
	more details.
	
	You should have received a copy of the GNU Lesser General Public License along with this library; if not, write to 
	the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/


package edu.utexas.its.eis.tools.qwicap.template.xml.mutable;


import java.util.Arrays;


/**	Allows <code>MutableMarkup</code> to avoid some of the overhead in using an <code>ArrayList</code> object for this 
 *	purpose. There are two main advantages: (1) direct access to the array, so there is less overhead than calling 
 *	<code>ArrayList.get</code> all the time, and (2) the array is of type <code>MutableMatch</code>, so there's 
 *	somewhat less typecasting to be done. Also, methods like <code>contains</code> and <code>indexOf</code> only test 
 *	for object reference equality, which makes them faster than their counterparts in the <code>List</code> interface, 
 *	though less flexible (Qwicap doesn't need that flexibility).
 *	
 *	<p>Note that this class is conceptually similar to <code>RangeList</code>, and where applicable their 
 *	implementations should be maintained in parallel.
 *
 *	@author Chris W. Johnson
 */
final class MatchList implements Cloneable {
	
	public MutableMatch[]		Array;
	private int					LogicalLength;
	
	
	MatchList
	(
		final int				InitialCapacity
	)
	{
		Array = new MutableMatch[InitialCapacity];
		LogicalLength = 0;
	}
	
	public void clear() {
		LogicalLength = 0;
	}
	
	public int size() {
		return LogicalLength;
	}
	
	public boolean isEmpty() {
		return LogicalLength == 0;
	}
	
//	For backwards compatibility with ArrayList.
	
	public MutableMatch get
	(
		final int				Index
	)
	{
		if (Index < 0 || Index >= LogicalLength)
			throw new ArrayIndexOutOfBoundsException("Invalid array index. We were asked to retrieve element " + Index + ", but the valid indices range from 0 to " + (LogicalLength - 1) + " (inclusive).");
		
		return Array[Index];
	}

/*	
//	Note that this implementation differs from the one in List in that it tests only for object reference equality, 
//	whereas the implementation in List uses the "equals" method of one of the objects.
	
	public boolean contains
	(
		final Range				SearchForObj
	)
	{
		return indexOf(SearchForObj) != -1;
	}
*/	

/*
//	Note that this implementation differs from the one in List in that it tests only for object reference equality, 
//	whereas the implementation in List uses the "equals" method of one of the objects.
	
	public int indexOf
	(
		final Range				SearchForObj
	)
	{
		for (int Index = 0, Count = LogicalLength; Index < Count; Index++) {
			
			if (Array[Index] == SearchForObj)
				return Index;
		}
		
		return -1;
	}
*/	
//	For backwards compatibility with ArrayList.
	
	public void set
	(
		final int				Index,
		final MutableMatch		Value
	)
	{
		Array[Index] = Value;
	}
	
/*	private void add
	(
		final MutableMatch		Value
	)
	{
		growIfNecessary(1);
		Array[LogicalLength++] = Value;
	}
	
	private void add
	(
		final int				Index,
		final MutableMatch		Value
	)
	{
		growIfNecessary(1);
		makeRoom(Index, 1);
		Array[Index] = Value;
	}
	
	private void addAll
	(
		final int				Index,
		final List				Values
	)
	{
		final int				ValuesCount = Values.size();
		
		growIfNecessary(ValuesCount);
		makeRoom(Index, ValuesCount);
		System.arraycopy(Values.toArray(), 0, Array, Index, ValuesCount);
	}
*/	
	public void add
	(
		final int				Index
	)
	{
		add(Index, 1);
	}
	
	public void add
	(
		final int				Index,
		final int				Count
	)
	{
		growIfNecessary(Count);
		makeRoom(Index, Count);
		Arrays.fill(Array, Index, Index + Count, null);
		
		fixMatchCacheIndices(Index + Count);
	}
	
	public void remove
	(
		final int				Index
	)
	{
		remove(Index, 1);
	}
	
	public void remove
	(
		final int				Index,
		final int				Count
	)
	{
		final int				RemainderIndex = Index + Count;
		final int				RemainderLength = LogicalLength - RemainderIndex;
		
		if (Index < 0 || Index >= LogicalLength)
			throw new ArrayIndexOutOfBoundsException("Invalid array index. We were asked to remove elements beginning at " + Index + ", but the valid indices range from 0 to " + (LogicalLength - 1) + '.');
		if (Count < 0)
			throw new ArrayIndexOutOfBoundsException("The number of elements to be removed from the array cannot be less than zero, so " + Count + " won't work.");
		if (RemainderIndex > LogicalLength)
			throw new ArrayIndexOutOfBoundsException("Invalid array index. We were asked to remove " + Count + " elements from the array beginning at " + Index + ", but because the array's logical length is " + LogicalLength + " we can only remove at most " + (LogicalLength - Index) + " elements from that index.");
		
		if (RemainderLength > 0)
			System.arraycopy(Array, RemainderIndex, Array, Index, RemainderLength);
		
		LogicalLength -= Count;
		
		fixMatchCacheIndices(Index);
	}
	
	private void makeRoom
	(
		final int				DestIndex,
		final int				NewElements
	)
	{
		if (DestIndex < 0 || DestIndex > LogicalLength)
			throw new ArrayIndexOutOfBoundsException("Invalid array index. We were asked to shift array elements beginning with " + DestIndex + ", but the valid indices range from 0 to " + LogicalLength + '.');
		if (NewElements < 0)
			throw new IllegalArgumentException("Can't make room for a negative number (" + NewElements + ") of new array elements.");
		
		final int				ElementsToMove = LogicalLength - DestIndex;
		
		if (ElementsToMove > 0)
			System.arraycopy(Array, DestIndex, Array, DestIndex + NewElements, ElementsToMove);
		
		LogicalLength += NewElements;
	}
	
	private void growIfNecessary
	(
		final int				NewElementsNeeded
	)
	{
		if (NewElementsNeeded < 0)
			throw new IllegalArgumentException("Can't use growIfNecessary to \"grow\" an array by a negative amount (" + NewElementsNeeded + ").");
		
		final int				PyLen = Array.length;
		final int				NewLgLen = LogicalLength + NewElementsNeeded;
		
		if (NewLgLen > PyLen) {
			int						NewPyLen = (PyLen * 3) / 2 + 1;
			
			if (NewPyLen < NewLgLen)
				NewPyLen = NewLgLen;
			
		//	Double the length of our array, until we find a length that is >= the new logical length.
		//	
		//	int						NewPyLen = PyLen << 1;
		//	
		//	while (NewPyLen < NewLgLen)
		//		NewPyLen <<= 1;
			
		//	Allocate the new array, and copy the contents of the previous array into it, then replace the previous
		//	array with the new one.
			
			final MutableMatch[]	NewArray = new MutableMatch[NewPyLen];
			
			System.arraycopy(Array, 0, NewArray, 0, LogicalLength);
			Array = NewArray;
		}
	}
	
	@Override public Object clone() throws CloneNotSupportedException {
		final MatchList			NewList = (MatchList) super.clone();
		
		NewList.Array = Array.clone();
		
		return NewList;
	}
	
	private void fixMatchCacheIndices
	(
		final int				StartIndex
	)
	{
	//	Strictly speaking, we don't have to have this method at all, because MutableMatch objects detect when their
	//	indices become stale, and update them automatically. However, that update requires a linear search of the 
	//	markup element list, so it seems like a good idea to avoid that where possible.
		
		final MutableMatch[]	Array = this.Array;
		
		for (int Index = StartIndex, Count = LogicalLength; Index < Count; Index++) {
			final MutableMatch		MuMa = Array[Index];
			
			if (MuMa != null)
				MuMa.setIndex(Index);
		}
	}
		
/*	public void validate() {
		Markup					TheMarkup = null;
		
		for (int Index = 0, Count = LogicalLength; Index < Count; Index++) {
			final MutableMatch		MuMa = Array[Index];
			
			if (MuMa != null) {
				
				if (TheMarkup == null)
					TheMarkup = MuMa.getMarkup();
				else if (MuMa.getMarkup() != TheMarkup)
					throw new IllegalStateException("There is a Match for markup \"" + MuMa.getMarkup().getMarkupName() + "\" in the Match cache for markup \"" + TheMarkup.getMarkupName() + "\".");
				
				MuMa.validate();
			}
		}
	}
*/	
	
/*	public void remove
	(
		final int				Index,
		final int				Count
	)
	{
		remove(Index, Count);
		fixMatchCacheIndices(Index);
	}
*/	

//	private void contractMatchCache
//	(
//		final int				FirstIndex,
//		final int				Count
//	)
//	{
//	//	There's no useful bulk removal method in the List interface (but if we subclass ArrayList, we can gain access
//	//	to its protected remove(fromIndex, toIndex) method), so we have to remove elements one at a
//	//	time. Note that it's more efficient to remove by working up from the bottom, than down from the top.
//	
//		for (int Index = FirstIndex + (Count - 1); Index >= FirstIndex; Index--)
//			MatchCache.remove(Index);
//		
//	//	Fix all the cached matches that refer to elements beneath the point from which we removed elements.
//	
//		fixMatchCacheIndices(FirstIndex);
//	}


	
//	private void expandMatchCache
//	(
//		final int				FirstIndex,
//		final int				Count
//	)
//	{
//		if (Count == 1)
//			MatchCache.add(FirstIndex, null);	
//		else
//			MatchCache.addAll(FirstIndex, Arrays.asList(new Object[Count]));
//		
//	//	Fix all the cached matches that refer to elements beneath the point at which we inserted the new elements.
//		
//		fixMatchCacheIndices(FirstIndex + Count);
//	}

}
