/*	QWICAP (Quick Web Interface for Conventional Applications) by Chris W. Johnson
	Copyright (C) 2003-2010 The University of Texas at Austin ("U.T. Austin"). All rights reserved.
	
	U.T. Austin's copyright notice in this evolving version reflects our original authorship. It does not assert 
	copyright over changes contributed by others.
	
	This library is free software; you can redistribute it and/or modify it under the terms of the GNU Lesser General 
	Public License as published by the Free Software Foundation; either version 2.1 of the License, or (at your option) 
	any later version.
	
	This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied 
	warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for 
	more details.
	
	You should have received a copy of the GNU Lesser General Public License along with this library; if not, write to 
	the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/


//  MuClassAttribute.java
//  qwicap
//
//  Created by Chris W. Johnson on Tue Jul 15 2003.


package edu.utexas.its.eis.tools.qwicap.template.xml.mutable;


import edu.utexas.its.eis.tools.qwicap.template.xml.structure.*;
import edu.utexas.its.eis.tools.qwicap.util.Characters;

import java.util.*;


/**	Represents a mutable "class" attribute in mutable XML documents.
 * 
 *	@author Chris W. Johnson
 */
final class MuClassAttribute extends MuAttribute {
	
	private ArrayList<Characters>	Classes = new ArrayList<Characters>();
	
	
	MuClassAttribute
	(
		final Attribute			Attr
	)
	{
		super(TagWithAttributes.kClassAttributeName, null, Namespace.kNoNamespace);
		addClass(Attr);
	}
	
	MuClassAttribute
	(
		final Characters		Value
	)
	{
		super(TagWithAttributes.kClassAttributeName, null, Namespace.kNoNamespace);
		parseAndAddNonRedundantClasses(Value);
	}
	
	int size() {
		return Classes.size();
	}
	
	void addClass
	(
		final Characters		NewClassName
	)
	{
		if (NewClassName.indexOf(' ') < 0)
			addNonRedundantClass(NewClassName);
		else
			parseAndAddNonRedundantClasses(NewClassName);	//	There's a space in NewClass, so it's likely to be a list of class names to be added.
	}
	
	void addClass
	(
		final Attribute			Attr
	)
	{
		if (Attr instanceof MuClassAttribute) {
			
			addNonRedundantClasses(((MuClassAttribute) Attr).Classes);
			
		} else if (Attr.nameEquals(TagWithAttributes.kClassAttributeName)) {
			
			parseAndAddNonRedundantClasses(Attr.getValue());
			
		} else
			throw new IllegalArgumentException("Attribute passed to addClass must be named \"class\". This one was named \"" + Attr.getName() + "\".");
	}
	
	void setClass
	(
		final Characters		NewClassList
	)
	{
		Classes.clear();
		addClass(NewClassList);
	}
	
	void deleteClass
	(
		final Characters		ClassList
	)
	{
		if (ClassList.indexOf(' ') >= 0) {
			final StringTokenizer	ST = new StringTokenizer(ClassList.toString());
			
			while (ST.hasMoreTokens())
				deleteClassLow(new Characters(ST.nextToken()));
			
		} else
			deleteClassLow(ClassList);
	}
	
	private void deleteClassLow
	(
		final Characters		IndividualClassName
	)
	{
		for (int Index = 0, Count = Classes.size(); Index < Count; Index++) {
			
			if (IndividualClassName.equals(Classes.get(Index))) {
				
				Classes.remove(Index);
				break;
			}
		}
	}
	
	boolean contains
	(
		final Characters		Class
	)
	{
		for (int Index = 0, Count = Classes.size(); Index < Count; Index++) {
			
			if (Class.equals(Classes.get(Index)))
				return true;
		}
		
		return false;
	}
	
	private void parseAndAddNonRedundantClasses
	(
		final Characters		Value
	) 
	{
		final StringTokenizer	ST = new StringTokenizer(Value.toString());
		
		while (ST.hasMoreTokens())
			addNonRedundantClass(new Characters(ST.nextToken()));
	}
	
	private void addNonRedundantClasses
	(
		final List<Characters>	OtherClasses
	)
	{
		for (int Index = 0, Count = OtherClasses.size(); Index < Count; Index++)
			addNonRedundantClass(OtherClasses.get(Index));
	}
	
	private void addNonRedundantClass
	(
		final Characters		Class
	)
	{
		if (contains(Class) == false)
			Classes.add(Class);
	}
	
	@Override public void setValue
	(
		final Characters		ClassList
	)
	{
		if (ClassList != null)
			setClass(ClassList);
	}
	
	@Override public Characters getValue() {
		final StringBuilder		Buff = new StringBuilder(64);
		
		for (int Index = 0, Count = Classes.size(); Index < Count; Index++) {
			
			if (Index > 0)
				Buff.append(' ');
			
			Classes.get(Index).append(Buff);
		}
		
		return new Characters(Buff);
	}

	@Override public String toString() {

		if (Classes.size() != 0)
			return append(new StringBuilder(64)).toString();
		else
			return "";
	}

	@Override public Object clone() {
		final MuClassAttribute	Clone = (MuClassAttribute) super.clone();
		
		Clone.Classes = (ArrayList<Characters>) Classes.clone();
		
		return Clone;
	}
}








