/*	QWICAP (Quick Web Interface for Conventional Applications) by Chris W. Johnson
	Copyright (C) 2003-2010 The University of Texas at Austin ("U.T. Austin"). All rights reserved.
	
	U.T. Austin's copyright notice in this evolving version reflects our original authorship. It does not assert 
	copyright over changes contributed by others.
	
	This library is free software; you can redistribute it and/or modify it under the terms of the GNU Lesser General 
	Public License as published by the Free Software Foundation; either version 2.1 of the License, or (at your option) 
	any later version.
	
	This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied 
	warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for 
	more details.
	
	You should have received a copy of the GNU Lesser General Public License along with this library; if not, write to 
	the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/


//  MuTagWithAttributes.java
//  qwicap
//
//  Created by Chris W. Johnson on Tue Jul 15 2003.


package edu.utexas.its.eis.tools.qwicap.template.xml.mutable;


import edu.utexas.its.eis.tools.qwicap.template.xml.AttributeEnumeration;
import edu.utexas.its.eis.tools.qwicap.template.xml.structure.*;
import edu.utexas.its.eis.tools.qwicap.template.xml.RangeList;
import edu.utexas.its.eis.tools.qwicap.util.Characters;

import java.io.PrintWriter;
import java.io.IOException;
import java.util.Iterator;

//	todo Consider how we need to special-case the handling of attributes so that the comings and goings of namespace definitions are recognized.

/**	Common ancestor of all mutable tags that can have attributes in mutable XML documents.
 * 
 *	@author Chris W. Johnson
 */
abstract class MuTagWithAttributes extends MuTag implements TagWithAttributes {
	
	private RangeList			Attributes;
	private MuClassAttribute	ClassAttr;
	
	
	MuTagWithAttributes
	(
		final TagWithAttributes	Master
	)
	{
		setName(Master.getName());
		setNamespace(Master.getNamespace());
		
	//	todo Note that the line above will loose namespace information if the Master tag's name began with a namespace
	//	     prefix that could never be associated with a namespace declaration. The loss occurs because the prefix  
	//	     (for example "wsdl:") has been removed from the Master tag's name, but its getNamespace method returns 
	//	     null, so we don't get the prefix information back through the normal mechanism. This can only happen in 
	//	     incomplete markup that includes namespace prefixes that are not defined by a namespace declaration.
	//	     Nonetheless, it's bad to have this problem.
	//
	//	     This behavior begins with the use of the TagHierarchy class (in the "immutable" package), when it is 
	//	     instantiated with its BeNamespaceAware parameter set to 'true'. This is when the separation of the 
	//	     namespace prefixes from the tag names occurrs. 
		
	//	Add all of the Master's attributes to our attribute list.
	//	(And isolate the "class" attribute, if any, while we're at it.)
		
		final AttributeEnumeration	Attrs = Master.enumerateAttributes();
		
		Attributes = new RangeList(Attrs.size() + 2);	//	We arbitrarily add 2 to the attribute count to allow for the efficient addition of two attributes. This represents nothing more than a guess about the optimal behavior.
		
		while (Attrs.hasMoreAttributes())
			addAttribute(Attrs.nextAttribute());
	}
	
	@Override public final boolean isOfClass
	(
		final Characters		ClassName
	)
	{
		return ClassAttr != null && ClassAttr.contains(ClassName);

	}
	
	private void addClass
	(
		final Attribute			Attr
	)
	{
		if (ClassAttr == null) {
			
			ClassAttr = new MuClassAttribute(Attr);
			Attributes.add(ClassAttr);
			
		} else
			ClassAttr.addClass(Attr);
	}
	
	final void addClass
	(
		final String			ClassList
	)
	{
		addClass(new Characters(ClassList));
	}
	
	final void addClass
	(
		final Characters		ClassList
	)
	{
		if (ClassAttr == null) {
			
			ClassAttr = new MuClassAttribute(ClassList);
			Attributes.add(ClassAttr);
			
		} else
			ClassAttr.addClass(ClassList);
	}
	
	final void setClass
	(
		final String			ClassList
	)
	{
		setClass(new Characters(ClassList));
	}
	
	final void setClass
	(
		final Characters		NewClassList
	)
	{
		if (ClassAttr == null)
			addClass(NewClassList);
		else
			ClassAttr.setClass(NewClassList);
	}
	
	final void deleteClass
	(
		final String			ClassList
	)
	{
		deleteClass(new Characters(ClassList));
	}
	
	final void deleteClass
	(
		final Characters		ClassName
	)
	{
		if (ClassAttr != null)
			ClassAttr.deleteClass(ClassName);
	}
	
	@Override public final Attribute getAttribute
	(
		final Characters		AttrName
	)
	{
		final RangeList			Attributes = this.Attributes;
		
		for (int Index = 0, Count = Attributes.size(); Index < Count; Index++) {
			final Attribute			Attr = (Attribute) Attributes.get(Index);
			
			if (Attr.nameEquals(AttrName))
				return Attr;
		}
		
		return null;
	}

	@Override public final String getAttribute
	(
		final String			AttrName
	)
	{
		final Attribute			Attr = getAttribute(new Characters(AttrName));
		
		if (Attr != null)
			return Attr.getValue().toString();
		else
			return null;
	}
	
/*	public boolean hasAttribute
	(
		final String			AttrName
	)
	{
		return (getAttribute(new Characters(AttrName)) != null);
	}
*/	
	private boolean hasAttributes() {
		final int				AttrCount = Attributes.size();
		final boolean			HasAttrs;
		
		if (AttrCount == 0) {
			
			HasAttrs = false;
			
		} else if (AttrCount == 1 && ClassAttr != null) {
			
		//	If the only attribute this tag possesses is an empty "class" attribute, return false.
			
			HasAttrs = ClassAttr.size() > 0;
			
		} else
			HasAttrs = true;
		
		return HasAttrs;
	}
	
	private void addAttribute
	(
		final Attribute			Attr
	)
	{
		if (Attr.nameEquals(kClassAttributeName))
			addClass(Attr);
		else
			addAttributeLow(Attr, Attr.getName(), Attr.getValue(), null);	//	The namespace of Attr will be used.
	}
	
	final void addAttribute
	(
		final String			Name,
		final String			Value
	)
	{
		addAttribute(new Characters(Name), new Characters(Value));
	}
	
	final void addAttribute
	(
		final Characters		Name,
		final Characters		Value
	)
	{
		if (Name.equals(kClassAttributeName))
			addClass(Value);
		else
			addAttributeLow(null, Name, Value, Namespace.kNoNamespace);
	}
	
/**	Add a new attribute to our list of attributes, or append to the value of an existing attribute. This low-level 
 * 	method does NOT provide special-case handling for the "class" attribute, so it should never be used to create/alter 
 * 	a "class" attribute.
 * 
 * @param Attr							The <code>Attribute</code> to be added. Can be <code>null</code>.
 * @param AttrName						The name of the attribute. The name must not be "class".
 * @param AttrValue						The values of the attribute.
 * @param AttrNs						The namespace of the attribute. This is only used when <code>Attr</code> is 
 * 										<code>null</code>, and there is no pre-existing attribute of the specified name.
 */	
	private void addAttributeLow
	(
		final Attribute			Attr,		//	Can be null.
		final Characters		AttrName,	//	Must NOT be "class"    @todo Consider supporting a namespace prefix in the attribute name.
		final Characters		AttrValue,
		final Namespace			AttrNs		//	Only used when Attr is null, and there is no pre-existing attribute of the specified name.
	)
	{
		final int				AttrIndex = findAttribute(AttrName);
		
		if (AttrIndex < 0) {
			
		//	Attribute does NOT already exist in our attribute list.
			
			if (Attr != null) {
				
				if (Attr instanceof Mutable)
					Attributes.add((Attribute) Attr.clone());
				else
					Attributes.add(Attr);
			} else
				Attributes.add(new MuAttribute(AttrName, AttrValue, AttrNs));
			
		} else {
			
		//	Attribute is already present in our attribute list.
			
			final Attribute			CurAttr = (Attribute) Attributes.get(AttrIndex);
			
			if (CurAttr instanceof Mutable)
				((MuAttribute) CurAttr).addValue(AttrValue);
			else
				Attributes.set(AttrIndex, new MuAttribute(AttrName, CurAttr.getValue().append(AttrValue), CurAttr.getNamespace()));
		}
	}
	
	final void setAttribute
	(
		final String			Name,
		final String			Value
	)
	{
		setAttribute(new Characters(Name), new Characters(Value));
	}
	
	final void setAttribute
	(
		final Characters		Name,
		final Characters		Value
	)
	{
		if (Name.equals(kClassAttributeName)) {
			
			setClass(Value);
			
		} else {
			final int				AttrIndex = findAttribute(Name);
			
			if (AttrIndex < 0) {
				
			//	Attribute does NOT already exist in our attribute list.
				
				Attributes.add(new MuAttribute(Name, Value, Namespace.kNoNamespace));
				
			} else {
				
			//	Attribute is already present in our attribute list.
				
				final Attribute			CurAttr = (Attribute) Attributes.get(AttrIndex);
				
				if (CurAttr instanceof Mutable)
					((MuAttribute) CurAttr).setValue(Value);
				else
					Attributes.set(AttrIndex, new MuAttribute(Name, Value, CurAttr.getNamespace()));
			}
		}
	}
	
	final void updateAttribute
	(
		final String			Name,
		final String			Value
	)
	{
		updateAttribute(new Characters(Name), new Characters(Value));
	}
	
	final void updateAttribute
	(
		final Characters		Name,
		final Characters		Value
	)
	{
		if (Name.equals(kClassAttributeName)) {
			
			setClass(Value);
			
		} else {
			final int				AttrIndex = findAttribute(Name);
			
			if (AttrIndex >= 0) {
				
			//	Attribute is already present in our attribute list.
				
				final Attribute			CurAttr = (Attribute) Attributes.get(AttrIndex);
				
				if (CurAttr instanceof Mutable)
					((MuAttribute) CurAttr).setValue(Value);
				else
					Attributes.set(AttrIndex, new MuAttribute(Name, Value, Namespace.kNoNamespace));
			}
		}
	}
	
	final void deleteAttribute
	(
		final String			Name
	)
	{
		deleteAttribute(new Characters(Name));
	}
	
	final void deleteAttribute
	(
		final Characters		Name
	)
	{
		if (Name.equals(kClassAttributeName)) {
			
			if (ClassAttr != null) {
				
				Attributes.remove(ClassAttr);
				ClassAttr = null;
			}
			
		} else {
			final int				AttrIndex = findAttribute(Name);
			
			if (AttrIndex >= 0)
				Attributes.remove(AttrIndex);
		}
	}
	
//	@todo Make findAttribute namespace-aware. Apart from taking a Namespace parameter, it might also be useful to accept a namespace prefix on the name.
	
	private int findAttribute
	(
		final Characters		AttrName
	)
	{
		for (int Index = 0, Count = Attributes.size(); Index < Count; Index++) {
			final Attribute			Attr = (Attribute) Attributes.get(Index);
			
			if (Attr.nameEquals(AttrName))
				return Index;
		}
		
		return -1;
	}
	
	@Override public final AttributeEnumeration enumerateAttributes() {
		return new AttributeEnumeratorRangeList(Attributes);
	}
	
	@Override public Iterator<Attribute> iterator() {
		return enumerateAttributes();
	}
	
	@Override public final String toString() {
	
	//	Keep this implementation synchronized with that of "write".
	
		final StringBuilder		Buff = new StringBuilder(128);
		
		Buff.append('<');
		
		final Namespace			Ns = getNamespace();
		
		if (Ns.isDefault() == false)
			Buff.append(Ns.getName()).append(':');
		
		getName().append(Buff);
		
		if (hasAttributes()) {
			
			Buff.append(' ');
			appendAttributes(Buff);
		}
		
		if (this instanceof EmptyTag)
			Buff.append(" />");
		else
			Buff.append('>');
		
		return Buff.toString();
	}
	
	private StringBuilder appendAttributes
	(
		final StringBuilder		Buff
	)
	{
	//	Keep this implementation synchronized with that of "writeAttributes".
	
		final AttributeEnumeration	Attrs     = enumerateAttributes();
		int							AttrCount = 0;
		
		while (Attrs.hasMoreAttributes()) {
			final Attribute			Attr = Attrs.nextAttribute();
			
			if (AttrCount++ > 0)
				Buff.append(' ');
			
			Attr.append(Buff);
		}
		
		return Buff;
	}
	
	@Override public final void write
	(
		final PrintWriter		Out
	)
	throws IOException
	{
	//	Keep this implementation synchronized with that of "toString".
	
		Out.write('<');
		
		final Namespace			Ns = getNamespace();
		
		if (Ns != null && Ns.isDefault() == false) {
			
			Ns.getName().write(Out);
			Out.write(':');
		}
		
		getName().write(Out);
		
		if (hasAttributes()) {
			
			Out.write(' ');
			writeAttributes(Out);
		}
		
		if (this instanceof EmptyTag)
			Out.write(" />");
		else
			Out.write('>');
	}
	
	private void writeAttributes
	(
		final PrintWriter		Out
	)
	throws IOException
	{
	//	Keep this implementation synchronized with that of "appendAttributes".
	
		final AttributeEnumeration	Attrs     = enumerateAttributes();
		int							AttrCount = 0;
		
		while (Attrs.hasMoreAttributes()) {
			final Attribute			Attr = Attrs.nextAttribute();
			
			if (AttrCount++ > 0)
				Out.write(' ');
			
			Attr.write(Out);
		}	
	}
	
	@Override public final Characters getContent() {
		return new Characters(appendAttributes(new StringBuilder(128)));
	}
	
	@Override public final void setContent
	(
		final Characters		NewContent
	)
	{
		throw new IllegalArgumentException("Not supported.");
	}
	
	@Override public final Object clone() {
		final MuTagWithAttributes	Clone            = (MuTagWithAttributes) super.clone();
		final RangeList				ClonedAttributes = Attributes.clone();
		
		Clone.Attributes = ClonedAttributes;
		
	//	Clone all mutable Attribute objects in the clone of the Attribute list.
	//	Remember to update the clone's ClassAttr field if/when we clone a 
	//	MuClassAttribute object.
		
		for (int Index = 0, Count = ClonedAttributes.size(); Index < Count; Index++) {
			Attribute				Attr = (Attribute) ClonedAttributes.get(Index);
			
			if (Attr instanceof Mutable) {
				
				Attr = (Attribute) Attr.clone();
				ClonedAttributes.set(Index, Attr);
				
				if (Attr instanceof MuClassAttribute)
					Clone.ClassAttr = (MuClassAttribute) Attr;
			}
		}
		
		return Clone;
	}

	@Override public boolean isTagWithAttributes() { return true; }
}







