/*	QWICAP (Quick Web Interface for Conventional Applications) by Chris W. Johnson
	Copyright (C) 2003-2010 The University of Texas at Austin ("U.T. Austin"). All rights reserved.
	
	U.T. Austin's copyright notice in this evolving version reflects our original authorship. It does not assert 
	copyright over changes contributed by others.
	
	This library is free software; you can redistribute it and/or modify it under the terms of the GNU Lesser General 
	Public License as published by the Free Software Foundation; either version 2.1 of the License, or (at your option) 
	any later version.
	
	This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied 
	warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for 
	more details.
	
	You should have received a copy of the GNU Lesser General Public License along with this library; if not, write to 
	the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/


package edu.utexas.its.eis.tools.qwicap.template.xml.mutable;


import edu.utexas.its.eis.tools.qwicap.template.xml.structure.XMLDeclaration;
import edu.utexas.its.eis.tools.qwicap.template.xml.structure.Attribute;
import edu.utexas.its.eis.tools.qwicap.template.xml.structure.Namespace;
import edu.utexas.its.eis.tools.qwicap.template.xml.AttributeEnumeration;
import edu.utexas.its.eis.tools.qwicap.template.xml.RangeList;
import edu.utexas.its.eis.tools.qwicap.util.Characters;

import java.io.PrintWriter;
import java.io.IOException;
import java.util.Iterator;


/** Represents the mutable XML declaration tag ("<samp>&lt;?xml version="1.0" encoding="UTF-8"?&gt;</samp>") at the 
 * 	start of any XML document.
 * 
 *	@author Chris W. Johnson
 */
final class MuXMLDeclaration extends MuItem implements XMLDeclaration {

	private RangeList			Attributes;


	MuXMLDeclaration
	(
		final XMLDeclaration	Master
	)
	{

	//	Add all of the Master's attributes to our attribute list.
	//	(And isolate the "class" attribute, if any, while we're at it.)

		final AttributeEnumeration	Attrs = Master.enumerateAttributes();

		Attributes = new RangeList(Attrs.size());

		while (Attrs.hasMoreAttributes())
			addAttribute(Attrs.nextAttribute());
	}

	@Override public final void setContent
	(
		final Characters		NewContent
	)
	{
		throw new IllegalArgumentException("Not supported.");
	}

	@Override public final Characters getContent() {
		return new Characters(appendAttributes(new StringBuilder(64)));
	}

	private StringBuilder appendAttributes
	(
		final StringBuilder		Buff
	)
	{
	//	Keep this implementation synchronized with that of "writeAttributes".

		final AttributeEnumeration	Attrs = enumerateAttributes();
		int							AttrCount = 0;

		while (Attrs.hasMoreAttributes()) {
			final Attribute			Attr = Attrs.nextAttribute();

			if (AttrCount++ > 0)
				Buff.append(' ');

			Attr.append(Buff);
		}

		return Buff;
	}

	@Override public final AttributeEnumeration enumerateAttributes() {
		return new AttributeEnumeratorRangeList(Attributes);
	}
	
	@Override public Iterator<Attribute> iterator() {
		return enumerateAttributes();
	}

	private void addAttribute
	(
		final Attribute			Attr
	)
	{
		addAttributeLow(Attr, Attr.getName(), Attr.getValue(), null);	//	The namespace of Attr will be used.
	}

/**	Add a new attribute to our list of attributes, or append to the value of an existing attribute. This low-level 
 * 	method does NOT provide special-case handling for the "class" attribute, so it should never be used to create/alter 
 * 	a "class" attribute.
 * 
 * @param Attr							The <code>Attribute</code> to be added. Can be <code>null</code>.
 * @param AttrName						The name of the attribute. The name must not be "class".
 * @param AttrValue						The values of the attribute.
 * @param AttrNs						The namespace of the attribute. This is only used when <code>Attr</code> is 
 * 										<code>null</code>, and there is no pre-existing attribute of the specified name.
 */	
	private void addAttributeLow
	(
		final Attribute			Attr,		//	Can be null.
		final Characters		AttrName,	//	Must NOT be "class"    @todo Consider supporting a namespace prefix in the attribute name.
		final Characters		AttrValue,
		final Namespace			AttrNs		//	Only used when Attr is null, and there is no pre-existing attribute of the specified name.
	)
	{
		final int				AttrIndex = findAttribute(AttrName);

		if (AttrIndex < 0) {

		//	Attribute does NOT already exist in our attribute list.

			if (Attr != null) {

				if (Attr instanceof Mutable)
					Attributes.add((Attribute) Attr.clone());
				else
					Attributes.add(Attr);
			} else
				Attributes.add(new MuAttribute(AttrName, AttrValue, AttrNs));

		} else {

		//	Attribute is already present in our attribute list.

			final Attribute			CurAttr = (Attribute) Attributes.get(AttrIndex);

			if (CurAttr instanceof Mutable)
				((MuAttribute) CurAttr).addValue(AttrValue);
			else
				Attributes.set(AttrIndex, new MuAttribute(AttrName, CurAttr.getValue().append(AttrValue), CurAttr.getNamespace()));
		}
	}

	private int findAttribute
	(
		final Characters		AttrName
	)
	{
		for (int Index = 0, Count = Attributes.size(); Index < Count; Index++) {
			final Attribute			Attr = (Attribute) Attributes.get(Index);

			if (Attr.nameEquals(AttrName))
				return Index;
		}

		return -1;
	}

	@Override public final Attribute getAttribute
	(
		final Characters		AttrName
	)
	{
		final RangeList			Attributes = this.Attributes;

		for (int Index = 0, Count = Attributes.size(); Index < Count; Index++) {
			final Attribute			Attr = (Attribute) Attributes.get(Index);

			if (Attr.nameEquals(AttrName))
				return Attr;
		}

		return null;
	}

	@Override public final String getAttribute
	(
		final String			AttrName
	)
	{
		final Attribute			Attr = getAttribute(new Characters(AttrName));

		if (Attr != null)
			return Attr.getValue().toString();
		else
			return null;
	}
	
	final void setAttribute
	(
		final String			Name,
		final String			Value
	)
	{
		setAttribute(new Characters(Name), new Characters(Value));
	}
	
	final void setAttribute
	(
		final Characters		Name,
		final Characters		Value
	)
	{
		final int				AttrIndex = findAttribute(Name);
		
		if (AttrIndex < 0) {
			
		//	Attribute does NOT already exist in our attribute list.
			
			Attributes.add(new MuAttribute(Name, Value, Namespace.kNoNamespace));
			
		} else {
			
		//	Attribute is already present in our attribute list.
			
			final Attribute			CurAttr = (Attribute) Attributes.get(AttrIndex);
			
			if (CurAttr instanceof Mutable)
				((MuAttribute) CurAttr).setValue(Value);
			else
				Attributes.set(AttrIndex, new MuAttribute(Name, Value, CurAttr.getNamespace()));
		}
	}

	@Override public String getEncoding() {
		return getAttribute("encoding");
	}

	@Override public String getVersion() {
		return getAttribute("version");
	}

	@Override public void write
	(
		final PrintWriter		Out
	)
	throws IOException
	{
		Out.write("<?xml");

		final RangeList			Attributes = this.Attributes;

		for (int Index = 0, Count = Attributes.size(); Index < Count; Index++) {
			
			Out.write(' ');
			Attributes.get(Index).write(Out);
		}
		
		Out.write("?>");
	}
	
	@Override public String toString() {
		final StringBuilder		Buff = new StringBuilder(64);
		
		Buff.append("<?xml");

		final RangeList			Attributes = this.Attributes;

		for (int Index = 0, Count = Attributes.size(); Index < Count; Index++) {
			
			if (Index > 0)
				Buff.append(' ');
			
			Buff.append(Attributes.get(Index).toString());
		}
		
		Buff.append("?>");
		
		return Buff.toString();
	}
}
