/*	QWICAP (Quick Web Interface for Conventional Applications) by Chris W. Johnson
	Copyright (C) 2003-2011 The University of Texas at Austin ("U.T. Austin"). All rights reserved.
	
	U.T. Austin's copyright notice in this evolving version reflects our original authorship. It does not assert 
	copyright over changes contributed by others.
	
	This library is free software; you can redistribute it and/or modify it under the terms of the GNU Lesser General 
	Public License as published by the Free Software Foundation; either version 2.1 of the License, or (at your option) 
	any later version.
	
	This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied 
	warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for 
	more details.
	
	You should have received a copy of the GNU Lesser General Public License along with this library; if not, write to 
	the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/


//  MutableMarkup.java
//  qwicap
//
//  Created by Chris W. Johnson on Wed Jun 18 2003.


package edu.utexas.its.eis.tools.qwicap.template.xml.mutable;


import edu.utexas.its.eis.tools.qwicap.servlet.*;
import edu.utexas.its.eis.tools.qwicap.template.css.CSSPatterns;
import edu.utexas.its.eis.tools.qwicap.template.css.CSSPatternsCache;
import edu.utexas.its.eis.tools.qwicap.template.css.Results;
import edu.utexas.its.eis.tools.qwicap.template.xml.*;
import edu.utexas.its.eis.tools.qwicap.template.xml.immutable.ImNamedItem;
import edu.utexas.its.eis.tools.qwicap.template.xml.immutable.ImmutableMarkup;
import edu.utexas.its.eis.tools.qwicap.template.xml.immutable.XMLString;
import edu.utexas.its.eis.tools.qwicap.template.xml.structure.*;
import edu.utexas.its.eis.tools.qwicap.util.*;

import java.nio.CharBuffer;
import java.util.Collection;
import java.util.logging.Level;
import java.util.logging.Logger;
import java.util.regex.Pattern;


/**	<code>MutableMarkup</code> represents a body of XML markup that can be altered. The markup represented may be an 
 *	entire XML document, or merely a self-contained fragment of XML. In either case, well-formedness is required to the 
 *	extent that all "start" tags must have matching "end" tags, and the "end" tags must be correctly ordered.
 *	
 *	@author Chris W. Johnson
 */
public final class MutableMarkup extends Markup {

	private static final Logger				Log = Logger.getLogger(MutableMarkup.class.getName());

	private static final HTMLEntityCodec	HTML                      = new HTMLEntityCodec().setEncoding('\n', "<br/>\n");
	private static final HTMLEntityCodec	HTMLNewLineNotEncodedAsBR = new HTMLEntityCodec();

	private static final Pattern			DashDashPattern = Pattern.compile("--");	//	"--" sequences are not permitted inside comments.
	
	private static final char[]				kEmptyCharArray = {};

	private final RangeList					List;
	private final MatchList					MatchCache;				//	Anytime a Match is generated for an item in List, it is retained here, at the same index as the item in List.
	private int								ChangeCount;			//	Incremented every time List is changed.
	private String							CharacterSet = "UTF-8";	//	In memory the markup is always in Unicode; this tells us how it was encoded when it was read, and how it should be encoded when it is written.
	private final CSSPatternsCache			CSSPatsCache = new CSSPatternsCache();
	
	private SmartStringConverter			StringConverter;


	public MutableMarkup() {
		List       = new RangeList(128);
		MatchCache = new MatchList(128);
	}

	public MutableMarkup
	(
		final RangeList			RangeList,
		final String			CharacterSet
	)
	{
		final int				Size = RangeList.size();

		List       = RangeList;
		MatchCache = new MatchList(Size);
		MatchCache.add(0, Size);
		this.CharacterSet = CharacterSet;
	}

	@Override public MutableMarkup getMutable() {
		return this;
	}

	@Override public ImmutableMarkup getImmutable() throws TagException {
		return new ImmutableMarkup(this);
	}

	@Override public MutableMarkup clone() {
		try {
			return getImmutable().getMutable();
		} catch (TagException e) {
			throw new RuntimeException(e);
		}
	}

	@Override public String getCharacterSet() {
		return CharacterSet;
	}

	public MutableMarkup setCharacterSet
	(
		final String			CharSetName
	)
	throws TagException
	{
		CharacterSet = CharSetName;

		if (List.isEmpty())
			return this;

	//	Deal with the XML declaration, if there is one.

		{
			final Range				FirstItem = List.Array[0];

			if (FirstItem instanceof XMLDeclaration) {

				if (FirstItem instanceof Mutable) {

					((MuXMLDeclaration) FirstItem).setAttribute("encoding", CharSetName);

				} else {
					final XMLDeclaration	XMLDec = (XMLDeclaration) FirstItem;

					if (CharSetName.equalsIgnoreCase(XMLDec.getEncoding()) == false) {
						final MuXMLDeclaration	MuXMLDec = new MuXMLDeclaration(XMLDec);

						MuXMLDec.setAttribute("encoding", CharSetName);

						List.Array[0] = MuXMLDec;	//	Replace the immutable XML declaration with the new, mutable declaration.
						ChangeCount++;
					}
				}
			}
		}

	//	Deal with the XHTML "meta"/"http-equiv"/"content-type" tag in the "head" of the document, if present.
	
		{
			final Results			Head = get("html > head");
			
			if (Head.notEmpty()) {
				boolean					SetCharSetWithMeta = false;
				
				for (final Match Meta : Head.get("meta[http-equiv][content]")) {
					
				//	todo There's better code for interpreting Content-Type values in the Page.PageContentTypeAndCharSet, so find a way to use that instead.
					
					if ("Content-Type".equalsIgnoreCase(Meta.getAttribute("http-equiv"))) {
						final ContentType		ContentType = new XMLContentType(Meta.getAttribute("content"));
						final String			CurCharSet  = ContentType.getCharacterSet();
						
						if (CurCharSet == null || !ContentType.getCharacterSetWasSpecified() || !CharSetName.equalsIgnoreCase(CurCharSet))
							Meta.setAttribute("content", ContentType.getMIMEType() + "; charset=" + CharSetName);
						
						SetCharSetWithMeta = true;
					}
				}
				
				if (SetCharSetWithMeta == false)
					Head.addToStartOfContent(new XMLString("<meta http-equiv='Content-Type' content='text/html; charset=" + CharSetName + "' />"));
			}
		}

		return this;
	}

	@Override public CSSPatterns getCSSPatterns
	(
		final String			PatsStr
	)
	{
		return CSSPatsCache.getCSSPatterns(PatsStr);
	}

/**	Searches the document for a form that could have supplied a specified form data set, and, if found, embeds the
 *	values from the data set into the form. If such a form is not found, an exception is thrown.
 * 
 * @param DataSet				A form data set produced by a form in this document. If <code>null</code>, nothing 
 *								happens.
 * @return						This <code>MutableMarkup</code> document (for convenience).
 * @throws QwicapException		If no form could be found in the document that could have supplied the specified form
 *								data set.
 */
	public MutableMarkup embedFormDataSet
	(
		final FormDataSet		DataSet	//	Can be null.
	)
	throws QwicapException
	{
		if (DataSet != null)
			new Form(this, DataSet).setDataSet(DataSet);

		return this;
	}

/**	Searches the document for a form that could have supplied a specified form data set, and, if found, embeds the
 *	values from the data set into the form. If such a form is not found, nothing happens.
 * 
 * @param DataSet				A form data set produced by a form in this document. If <code>null</code>, nothing 
 *								happens.
 * @return						This <code>MutableMarkup</code> document (for convenience).
 */
	public MutableMarkup embedFormDataSetIfPossible
	(
		final FormDataSet		DataSet	//	Can be null.
	)
	{
		try {
			embedFormDataSet(DataSet);
		} catch (QwicapException e) {
		//	Do nothing with this exception.
		}

		return this;
	}

/**	Searches the document for a form that could have supplied a specified form data set, and returns a 
 *	<code>Form</code> object for manipulating it.
 * 
 * @param DataSet					A form data set supplied by a form in this document.
 * @return							A <code>Form</code> object that can be used to manipulate the relevant form.
 * @throws FormNotFoundException	If no form could be found in the document that could have supplied the specified 
 *									set of parameters.
 * @throws QwicapException
 * @see Results#toForm
 */
	public Form getForm
	(
		final FormDataSet		DataSet
	)
	throws QwicapException
	{
		if (DataSet == null)
			throw new FormNotFoundException("The form data set reference is null.");

		return new Form(this, DataSet);
	}

/**	Converts all "button" controls of type "submit" to "input" controls of type "submit", in order to work around a 
 * 	hideous bug in Internet Explorer (Windows and Mac OS).	
 * 
 * @return 						The number of "button" elements that were converted into "input" elements.
 * @throws TagException			If there was a problem with the markup manipulation.
 */
	public int convertSubmitButtonsToInputs() throws QwicapException {
		final boolean			FinestIsLoggable = Log.isLoggable(Level.FINEST);
		int						TotalConversions = 0;
		int						Index            = 0;
		
		for (final Match Elem : get("form")) {
			final int				Conversions = new Form(Elem).convertSubmitButtonsToInputs();
			
			TotalConversions += Conversions;
			
			if (FinestIsLoggable)
				Log.log(Level.FINEST, "Converted {2} submit buttons of type \"button\" in form no. {0} of page \"{1}\" to \"input\" elements.", new Object[] { Index++, getMarkupName(), Conversions } );
		}

		return TotalConversions;
	}

/**	Converts all "input" controls of type "submit" that were "button" controls before 
 *	{@link #convertSubmitButtonsToInputs} was invoked back into "button" controls of type "submit". In other words, 
 *	this method undoes what {@link #convertSubmitButtonsToInputs} does.
 * 
 * @return 						The number of "input" elements that were converted back into "button" elements.
 * @throws TagException			If there was a problem with the markup manipulation.
 */
	public int convertSubmitInputsToButtons() throws QwicapException {
		final boolean			FinestIsLoggable = Log.isLoggable(Level.FINEST);
		int						TotalConversions = 0;
		int						Index            = 0;
		
		for (final Match Elem : get("form")) {
			final int				Conversions = new Form(Elem).convertSubmitInputsToButtons();
			
			TotalConversions += Conversions;

			if (FinestIsLoggable)
				Log.log(Level.FINEST, "Converted {2} submit buttons of type \"input\" in form no. {0} of page \"{1}\" to \"button\" elements.", new Object[] { Index++, getMarkupName(), Conversions } );
		}

		return TotalConversions;
	}

	@Override public int getChangeCount() {
		return ChangeCount;
	}

	@Override public RangeList getList() {
		return List;
	}

	@Override public Match createMatch
	(
		final Range				Element,
		final int				ElemIndex
	)
	{
		if (List.Array[ElemIndex] != Element)	//	Yes, a reference equality test is desired here.
			throw new IllegalArgumentException("The createMatch method has been invoked in order to create a match " +
				"for element \"" + Element + "\" at index " + ElemIndex + " of markup \"" + getMarkupName() +
				"\", but that element is not located at that index in this markup.");

		MutableMatch			MuMa = MatchCache.Array[ElemIndex];

	//	Perform a quick sanity check to see if we've corrupted the cache at some point in the past.

		if (MuMa != null && MuMa.get() != Element) {	//	Yes, a reference equality test is desired here.

			Log.severe("The MutableMarkup match cache is corrupt. There must be a bug.");
			throw new IllegalStateException("The MutableMarkup match cache is corrupt. There must be a bug.");
		}

		if (MuMa == null) {

			MuMa = new MutableMatch(this, Element, ElemIndex);
			MatchCache.Array[ElemIndex] = MuMa;
		}
		
		return MuMa;
	}
	
//	public final Match createMatch
//	(
//		final Range				Element,
//		final int				ElemIndex,
//		final boolean			IsIntraElementContent
//	)
//	{
//		if (List.Array[ElemIndex] != Element)
//			throw new IllegalArgumentException("The createMatch method has been invoked in order to create a match " +
//				"for element \"" + Element + "\" at index " + ElemIndex + " of markup \"" + getMarkupName() +
//				"\", but that element is not located at that index in this markup.");
//
//		MutableMatch			MuMa = MatchCache.Array[ElemIndex];
//
//	//	Perform a quick sanity check to see if we've corrupted the cache at some point in the past.
//
//		if (MuMa != null && MuMa.get() != Element) {
//
//		//	MuMa = null;
//			Log.severe("The MutableMarkup match cache is corrupt. There must be a bug.");
//			throw new IllegalStateException("The MutableMarkup match cache is corrupt. There must be a bug.");
//		}
//
//		if (MuMa == null) {
//
//			MuMa = new MutableMatch(this, Element, ElemIndex);
//			MatchCache.Array[ElemIndex] = MuMa;
//		}
//		
//		if (IsIntraElementContent)
//			return new MutableIntraElementContentMatch(MuMa);
//		
//		return MuMa;
//	}

/**	Remove all material from this markup object.
 */
	public void clear() {

		List.clear();
		MatchCache.clear();
		ChangeCount++;
	}

/**	Add material at the end of this markup object.
 * 
 * @param Source		The material to be added to this markup object.
 * @return				The material added to this markup object.
 */
	public Match add
	(
		final Match				Source
	)
	{
		return insertLowList(Source.enumerate(), 0, size());
	}

/**	Creates a duplicate of the specified material within this markup object and inserts it either immediately before,
 *	or after, the duplicated material.
 * 
 * @param Target						The materail within this markup object to be duplicated.
 * @param InsertDuplicateBeforeTarget	<code>true</code> if the duplicate should be inserted before the original
 *										material, or <code>false</code> if it should be inserted after.
 * @return								The duplicate.
 */
	public Match duplicate
	(
		final Match				Target,
		final boolean			InsertDuplicateBeforeTarget
	)
	{
		final Match				Duplicate;
		final Range				Tag         = Target.get();
		final int				TargetIndex = Target.getIndex();

		if (Tag.isStartTag()) {
			final RangeList			DupedItems = new RangeList(32);
			final MarkupEnumeration	Elems      = Target.enumerate();
			int						Index      = TargetIndex;

			while (Elems.hasMoreElements()) {
				final Range				Elem = Elems.nextRange();

				if (Elem instanceof Mutable)
					DupedItems.add((Range) Elem.clone());
				else
					DupedItems.add(Elem);

				Index++;
			}

			final int				InsertIndex = InsertDuplicateBeforeTarget ? TargetIndex : Index;

			Duplicate = insertListReturnMatch(DupedItems, InsertIndex);

		} else {
			final Range				NewTag      = Tag instanceof Mutable ? (Range) Tag.clone() : Tag;
			final int				InsertIndex = InsertDuplicateBeforeTarget ? TargetIndex : TargetIndex + 1;

			Duplicate = insertLowRange(NewTag, InsertIndex);
		}

		ChangeCount++;

		return Duplicate;
	}

/**	Replaces the specified material with other material.	
 * 
 * @param Target				The material to be replaced.
 * @param Replacement			The material to be added. See the {@link #insert insert} description for details on how 
 * 								this object is handled.
 * @return						The material added.
 */
	public Match replace
	(
		final Match				Target,
		final Object			Replacement
	)
	{
		final Match				ReplacementMatch = insert(Target, false, Replacement);

		delete(Target);

		return ReplacementMatch;
	}

/**	Deletes the specified material from this markup.
 * 
 * @param Target				The material to be deleted.
 */
	public void delete
	(
		final Match				Target
	)
	{
		final Range				Tag        = Target.get();
		final int				Index      = Target.getIndex();
		final RangeList			List       = this.List;
		final MatchList			MatchCache = this.MatchCache;

		if (Tag.isStartTag()) {
			int						GenerationCount = 0;

			ChangeCount++;

			do {
				final Range				Elem = List.get(Index);

				if (Elem.isStartTag())
					GenerationCount++;

			//	todo The repeated removals used in this loop are inefficient. Use a bulk removal strategy.

				List.remove(Index);
				MatchCache.remove(Index);

				if (Elem.isEndTag())
					GenerationCount--;

			} while (Index < List.size() && GenerationCount > 0);

		} else {

			ChangeCount++;

			List.remove(Index);
			MatchCache.remove(Index);
		}
	}

/**	Adds one or more classes to the specified tag, converting the tag from immutable to mutable, if necessary.
 * 
 * @param Target				The tag to which the class is to be added.
 * @param ClassList				A space-separated list of classes to be added to the target tag.
 * @return						The tag to which the class was added.
 * @throws TagException			If there was a problem with the markup manipulation.
 */
	public Match addClass
	(
		Match					Target,
		final String			ClassList
	)
	throws TagException
	{
		Target = makeTagMutable(Target);

		final Range				R = Target.get();

		if (R.isTagWithAttributes())
			((MuTagWithAttributes) R).addClass(AttributeUtilities.encodeAttributeValue(ClassList));

		return Target;
	}

	public Match setClass
	(
		Match					Target,
		final String			ClassList
	)
	throws TagException
	{
		Target = makeTagMutable(Target);

		final Range				R = Target.get();

		if (R.isTagWithAttributes())
			((MuTagWithAttributes) R).setClass(AttributeUtilities.encodeAttributeValue(ClassList));

		return Target;
	}

/**	Delete from this markup element all of the specified classes.
 * 
 * @param Target				The markup element.
 * @param ClassList				A space-separated list of class names.
 * @return						A Match referencing the modified markup element.
 * @throws TagException			If there was a problem with the markup manipulation.
 */
	public Match deleteClass
	(
		Match					Target,
		final String			ClassList
	)
	throws TagException
	{
		Target = makeTagMutable(Target);

		final Range				R = Target.get();

		if (R.isTagWithAttributes())
			((MuTagWithAttributes) R).deleteClass(AttributeUtilities.encodeAttributeValue(ClassList));

		return Target;
	}

/**	Adds an attribute to any markup element that is allowed to have attributes. If the markup element already has an
 * 	attribute of the specified name, the new value is appended to the existing value. Note that nothing clever is done
 * 	when appending, such as placing spaces between the original value and the new material that is appended to it.
 * 	Instead, this is a plain and simple append operation. Because of that, it is possible to use this method to, for
 * 	instance, progressively build-up a URL in an "a" (anchor) element's "href" attribute. When appending is not
 * 	desired, use the {@link #setAttribute setAttribute} method instead.
 *
 * 	<p>If the attribute name is "class", this method performs the same function as the {@link #addClass addClass} 
 * 	method, but somewhat less efficiently.
 *	</p>
 * 	
 * @param Target						The <code>Match</code> object identifying the markup element on which this 
 * 										method will operate.
 * @param AttrName						Name of the attribute.
 * @param AttrValue						Value of a new attribute, or the string to be appended to the value of an
 * 										existing attribute.
 * @return								A <code>Match</code> object identifying this markup element. If this element was 
 * 										immutable, a mutable version of it will have been created, substitued into this 
 * 										markup (assuming the markup itself is mutable), and the <code>Match</code> 
 * 										returned will refer to the new mutable form of this element.
 * @throws TagException					If there was a problem with the markup manipulation.
 */
	public Match addAttribute
	(
		Match					Target,
		final String			AttrName,
		final Object			AttrValue	//	Can be null.
	)
	throws TagException
	{
	//	todo Consider namespace implications here.

		Target = makeTagMutable(Target);

		final Range				R = Target.get();

		if (R.isTagWithAttributes())
			((MuTagWithAttributes) R).addAttribute(AttributeUtilities.encodeAttributeName(AttrName), AttributeUtilities.encodeAttributeValue(AttrValue == null ? null : AttrValue.toString()));

		return Target;
	}

/**	Replaces the value of an existing attribute, or creates a new attribute, in any markup element that is allowed to
 * 	have attributes. If the attribute name is "class", this method performs the same function as the 
 * 	{@link #setClass setClass} method, but somewhat less efficiently.
 *
 * @param Target						The <code>Match</code> object identifying the markup element on which this 
 * 										method will operate.
 * @param AttrName						Name of the attribute.
 * @param AttrValue						Value of the attribute.
 * @return								A <code>Match</code> object identifying this markup element. If this element was 
 * 										immutable, a mutable version of it will have been created, substitued into this 
 * 										markup (assuming the markup itself is mutable), and the <code>Match</code> 
 * 										returned will refer to the new mutable form of this element.
 * @throws TagException					If there was a problem with the markup manipulation.
 */
	public Match setAttribute
	(
		Match					Target,
		final String			AttrName,
		final Object			AttrValue	//	Can be null.
	)
	throws TagException
	{

	//	todo Consider namespace implications here.

		Target = makeTagMutable(Target);
		
		final Range				R = Target.get();

		if (R.isTagWithAttributes())
			((MuTagWithAttributes) R).setAttribute(AttributeUtilities.encodeAttributeName(AttrName), AttributeUtilities.encodeAttributeValue(AttrValue == null ? null : AttrValue.toString()));

		return Target;
	}

/**	Replaces the value of an existing attribute, but never creates a new attribute, in any markup element that is 
 *	allowed to have attributes. If the attribute name is "class", this method performs the same function as the 
 *	{@link #setClass setClass} method, but somewhat less efficiently.
 *
 * @param Target						The <code>Match</code> object identifying the markup element on which this 
 * 										method will operate.
 * @param AttrName						Name of the attribute.
 * @param AttrValue						Value of the attribute.
 * @return								A <code>Match</code> object identifying this markup element. If this element was 
 * 										immutable, a mutable version of it will have been created, substitued into this 
 * 										markup (assuming the markup itself is mutable), and the <code>Match</code> 
 * 										returned will refer to the new mutable form of this element.
 * @throws TagException					If there was a problem with the markup manipulation.
 */
	public Match updateAttribute
	(
		Match					Target,
		final String			AttrName,
		final Object			AttrValue	//	Can be null.
	)
	throws TagException
	{

	//	todo Consider namespace implications here.

		if (Target.hasAttribute(AttrName)) {

			Target = makeTagMutable(Target);

			final Range				R = Target.get();

			if (R.isTagWithAttributes())
				((MuTagWithAttributes) R).updateAttribute(AttributeUtilities.encodeAttributeName(AttrName), AttributeUtilities.encodeAttributeValue(AttrValue == null ? null : AttrValue.toString()));
		}

		return Target;
	}

/**	Removes an attribute from this markup element.
 *
 * @param Target						The <code>Match</code> object identifying the markup element on which this 
 * 										method will operate.
 * @param AttrName						Name of the attribute.
 * @return								A <code>Match</code> object identifying this markup element. If this element was 
 * 										immutable, a mutable version of it will have been created, substitued into this 
 * 										markup (assuming the markup itself is mutable), and the <code>Match</code> 
 * 										returned will refer to the new mutable form of this element.
 * @throws TagException					If there was a problem with the markup manipulation.
 */
	public Match deleteAttribute
	(
		Match					Target,
		final String			AttrName
	)
	throws TagException
	{

	//	todo Consider namespace implications here. For instance, if we've just deleted the attribute that defined a namespace what do we do with the affected tags?

		Target = makeTagMutable(Target);

	//	todo Consider that we should handle HTML entity encoding of the attribute name here, if we start doing that in addAttribute, setAttribute and updateAttribute.

		final Range				R = Target.get();

		if (R.isTagWithAttributes())
			((MuTagWithAttributes) R).deleteAttribute(AttributeUtilities.encodeAttributeName(AttrName));

		return Target;
	}

	private Match makeTagMutable
	(
		final Match				Target		//	Always a MutableMatch.
	)
	throws TagException
	{
		final Range				ImTag = Target.get();

		if (ImTag instanceof Mutable)
			return Target;

		final Range				MuTag;

		if (ImTag.isStartTag()) {

			MuTag = new MuStartTag((StartTag) ImTag);

		} else if (ImTag.isEmptyTag()) {

			MuTag = new MuEmptyTag((EmptyTag) ImTag);

		} else if (ImTag instanceof Content) {

			MuTag = new MuContent((Content) ImTag);

		} else if (ImTag instanceof Comment) {
			
			MuTag = new MuComment((Comment) ImTag);
			
		} else if (ImTag instanceof CDATA) {
			
			MuTag = new MuCDATA((CDATA) ImTag);
			
		} else if (ImTag instanceof Declaration) {
			
			MuTag = new MuDeclaration((Declaration) ImTag);
			
		} else
			throw new TagException("This method does not know how to convert an immutable object (\"" + ImTag +
				"\") of class \"" + ImTag.getClass().getName() + "\" into a mutable class.");

	//	Replace the immutable tag in the list with the new mutable tag, and return a MutableMatch object
	//	referencing the new mutable tag. Note that the original Match object, Target, is no longer valid,
	//	so any attempt to use it for anything will fail.

		final int				TagIndex = Target.getIndex();

		List.set(TagIndex, MuTag);
		ChangeCount++;				//	We just altered the list, so the counter must be incremented.

	//	The Match object we receive is always a MutableMatch, because that's the only kind of Match that is ever created
	//	to refer to elements in MutableMarkup. Here we update it such that it refers to the new, mutable flavor of this 
	//	markup element. Thus, other objects holding references to that MutableMatch object will continue to have a 
	//	valid reference to the markup element.

		return ((MutableMatch) Target).set(MuTag);
	}

/**	Deletes the content of the specified markup element, provided that the element is a start-tag or comment.
 *
 * @param Target			The <code>Match</code> object whose contents should be deleted.
 * @return					A <code>Match</code> object identifying the markup element whose content should be deleted. 
 * 							(Not the deleted material.)
 * @throws TagException		If there was a problem with the markup manipulation.
 */
	public Match deleteContent
	(
		final Match				Target
	)
	throws TagException
	{
		final Range				Tag = Target.get();

		if (Tag.isStartTag()) {
			final RangeList			List            = this.List;
			final MatchList			MatchCache      = this.MatchCache;
			int						Index           = Target.getIndex();
			int						GenerationCount = 0;

			ChangeCount++;

			do {
				final Range				Elem = List.get(Index);

				if (Elem.isEndTag())
					GenerationCount--;

			//	todo The repeated removals used in this loop are inefficient. Use a bulk removal strategy.

				if (GenerationCount > 0) {

					List.remove(Index);
					MatchCache.remove(Index);

				} else
					Index++;

				if (Elem.isStartTag())
					GenerationCount++;

			} while (Index < List.size() && GenerationCount > 0);

		} else if (Tag instanceof Comment || Tag instanceof CDATA || Tag instanceof Declaration) {
			
			makeTagMutable(Target);
			
			((MuItem) Target.get()).setContent(new Characters(kEmptyCharArray));
			ChangeCount++;
			
		} else
			throw new TagException("The target Match should specify a StartTag or Comment object, but instead specifies a \"" +
				Target.get().getClass().getName() + "\" object, specifically \"" + Tag + "\". That type of object " +
				"cannot have content and therefore is not a valid target for a content-oriented operation.");

		return Target;
	}

/**	Replaces any existing contents of the specified markup element with the supplied material.
 *
 * @param Target						The <code>Match</code> object identifying the markup element on which this 
 * 										method will operate.
 * @param NewContent					The material that will replace this element's content. See
 * 										{@link MutableMarkup#insert MutableMarkup.insert} for details.
 * @return								A <code>Match</code> object identifying this markup element. (Not the inserted 
 * 										material.)
 * @throws TagException					If there was a problem with the markup manipulation.
 */
	public Match setContent
	(
		final Match				Target,
		final Object			NewContent
	)
	throws TagException
	{
		deleteContent(Target);
		addContent(Target, true, NewContent);

		return Target;
	}

/**	Inserts the supplied material into the content of the specified markup element, provided that the element is a 
 * 	start-tag, or comment. The supplied material may be inserted before the existing contents, or after it.
 *
 * @param Target						The <code>Match</code> object identifying the markup element on which this 
 * 										method will operate.
 * @param InsertBeforeExistingContent	<code>true</code> if the new material should be inserted before the existing 
 * 										content of this element, or <code>false</code> if it should be inserted after 
 * 										the existing content.
 * @param NewContent					The material that will be inserted before/after this element's content. See
 * 										{@link MutableMarkup#insert MutableMarkup.insert} for details.
 * @return								A <code>Match</code> object identifying the inserted material.
 * @throws TagException					If there was a problem with the markup manipulation.
 */
	public Match addContent
	(
		final Match				Target,
		final boolean			InsertBeforeExistingContent,	//	If false, we insert after the existing content.
		final Object			NewContent						//	Markup, String, or something that can be converted to a string.
	)
	throws TagException
	{
		final Range				Tag = Target.get();
		final Match				InsertedContent;

		if (Tag.isStartTag()) {

			if (InsertBeforeExistingContent) {

				InsertedContent = insertLow(Target, false, NewContent);

			} else {

			//	Find the final element of the Target.

				final MarkupEnumeration	Elems = Target.enumerate();

				while (Elems.hasMoreElements())
					Elems.nextRange();

			//	Inset before it.

				InsertedContent = insertLow(Elems.getMatch(), true, NewContent);
			}

		} else if (Tag instanceof Comment) {
			
			makeTagMutable(Target);
			
			final MuComment				MuTag         = (MuComment) Target.get();
			final String				RawContentStr = otherToString(new StringBuilder(256), NewContent, 0).toString();
			final StringBuilder			ContentBuff   = new StringBuilder(MuTag.getContent().length() + RawContentStr.length());
			
			if (InsertBeforeExistingContent) {
				
				appendFilteredAndEncodedComment(RawContentStr, ContentBuff);
				MuTag.getContent().append(ContentBuff);
				
			} else {
				
				MuTag.getContent().append(ContentBuff);
				appendFilteredAndEncodedComment(RawContentStr, ContentBuff);
			}
			
			MuTag.setContent(new Characters(ContentBuff));
			
			InsertedContent = Target;	//	Or, I suppose we could return null, but that would introduce a new case for all callers.
			
		} else if (Tag instanceof CDATA) {
			
			makeTagMutable(Target);
			
			final MuCDATA				MuTag         = (MuCDATA) Target.get();
			final String				RawContentStr = otherToString(new StringBuilder(256), NewContent, 0).toString();
			final StringBuilder			ContentBuff   = new StringBuilder(MuTag.getContent().length() + RawContentStr.length());
			
			if (InsertBeforeExistingContent) {
				
				appendFilteredAndEncodedCDATA(RawContentStr, ContentBuff);
				MuTag.getContent().append(ContentBuff);
				
			} else {
				
				MuTag.getContent().append(ContentBuff);
				appendFilteredAndEncodedCDATA(RawContentStr, ContentBuff);
			}
			
			MuTag.setContent(new Characters(ContentBuff));
			
			InsertedContent = Target;	//	Or, I suppose we could return null, but that would introduce a new case for all callers.
			
		} else if (Tag instanceof Declaration) {
			
			makeTagMutable(Target);
			
			final MuDeclaration			MuTag         = (MuDeclaration) Target.get();
			final String				RawContentStr = otherToString(new StringBuilder(256), NewContent, 0).toString();
			final StringBuilder			ContentBuff   = new StringBuilder(MuTag.getContent().length() + RawContentStr.length());
			
			if (InsertBeforeExistingContent) {
				
				appendFilteredAndEncodedDeclaration(RawContentStr, ContentBuff);
				MuTag.getContent().append(ContentBuff);
				
			} else {
				
				MuTag.getContent().append(ContentBuff);
				appendFilteredAndEncodedDeclaration(RawContentStr, ContentBuff);
			}
			
			MuTag.setContent(new Characters(ContentBuff));
			
			InsertedContent = Target;	//	Or, I suppose we could return null, but that would introduce a new case for all callers.
			
		} else
			throw new TagException("The target Match should specify a StartTag, Comment or CDATA object, but instead specifies a \"" +
				Target.get().getClass().getName() + "\" object, specifically \"" + Target.get() + "\". That type of " +
				"object cannot have content and therefore is not a valid target for a content-oriented operation.");

		return InsertedContent;
	}

/**	<p>Inserts material into this markup before or after the element identified by the specified {@link Match}. If the 
 *	element has children, and the insert is specified to take place after the element, the insertion is performed 
 *	immediately after the element's end tag. The material to be inserted may be represented by any type of object. 
 *	That object is handled according to the following rules:
 *	</p>
 *
 * 	<ol>
 *
 *	<li>If the new object is a <code>String</code>, a new mutable content object is created to 
 *	represent the <code>String</code>, and that object is inserted into this markup.
 *	</li>
 *
 *	<li>If the new object is a {@link Match}, the markup element it represents, and any children of that element, 
 *	are inserted into this markup. The "Cloning Rule" applies here.
 *	</li>
 *
 *	<li>If the new object is a {@link Results}, it is handled as if it were a sequence of {@link Match} 
 *	objects being inserted (see above).
 *	</li>
 *
 *	<li>If the new object is a markup element (any subclass of {@link Range})
 *	it is inserted into this markup. The "Cloning Rule" applies here.
 *	</li>
 *
 *	<li>If the new object is {@link Markup}, or any subclass thereof, every element in that <code>Markup</code> is 
 *	inserted into this markup. The "Cloning Rule" applies here. <strong>Note</strong>, however, that because a 
 *	{@link Match} object can reference only one markup element (and all of its children, by implication) the 
 *	<code>Match</code> object returned by this method will reference only the first
 * 	element in the inserted material. So, if the inserted material includes more than one parent tag, only the first
 * 	parent and its children will be referenced by the <code>Match</code> object that is returned. (If this limitation 
 *	ever becomes important, it can be fixed by having the high-level insert-related methods in the {@link Results} 
 *	class test for instances of <code>Markup</code>, whereupon they would submit the top-level elements of the 
 *	<code>Markup</code> as individual <code>Match</code> objects.)
 * 	</li>
 *	
 *	<li>If the new object is a {@link Throwable}, like the ever-popular class {@link Exception},
 *	its message is inserted. If it does not have a message, the fully-qualified name of the <code>Throwable</code> 
 * 	object's class is inserted. In either case, the <code>Throwable</code> object's stack trace is then marked-up as an 
 * 	XHTML comment block and inserted, so the end-user will not see it, but the developer can view the page source to 
 * 	retrieve the stack trace.
 *	</li>
 *
 *	<li>If the new object is a {@link Collection} or an array, the contents will be interated through, with each 
 *	element converted to a string, and commas inserted between elements. Whenever an element is found to be an array, or 
 *	a <code>Collection</code>, the conversion process proceeds recursively, with the new material enclosed by brackets. 
 *	(The elements of <code>byte</code> or {@link Byte} arrays, will be displayed as unsigned decimal quantities. This 
 *	seemed like it would be more useful to most developers than displaying signed quantities.)
 *	</li>
 *
 *	<li>If the new object is not covered by any of the preceding rules, it is converted to a <code>String</code> by 
 *	invoking its <code>toString</code> method, and the rule for inserting <code>String</code> objects is applied.
 *	</li>
 *	
 *	</ol>
 *
 *	<p><strong>The Cloning Rule</strong>: No mutable element may exist more than once within a particular {@link Markup} 
 *	object. If an attempt is made to add a mutable element to markup in which it is already present, the element is 
 *	cloned, and the clone is added. (Note, by way of contrast, that the same mutable element is permitted to exist in 
 *	any number of <code>Markup</code> objects, which is very efficient and convenient, but could lead to surprising 
 *	results if the element is modified.)
 *	</p>
 *
 * @param Target			The location of the insert is relative to this element in this markup.
 * @param InsertBefore		If <code>true</code>, the new material is inserted before the <code>Target</code> element. 
 * 							If <code>false</code>, is inserted after <code>Target</code>. If <code>Target</code> has 
 * 							children, the insert is performed after its end tag.
 * @param NewStuff			The new material to be inserted. Any type of object is potentially valid.
 * @return					A <code>Match</code> object identifying the newly inserted material.
 */
	public Match insert
	(
		final Match				Target,
		final boolean			InsertBefore,
		final Object			NewStuff
	)
	{
		final Match				InsertedContent;

		if (InsertBefore) {

		//	Insert before the first element of the specified Match.

			InsertedContent = insertLow(Target, true, NewStuff);

		} else {

		//	Insert after the final element of the specified Match.

			final MarkupEnumeration	Elems = Target.enumerate();

			while (Elems.hasMoreElements())
				Elems.nextRange();

			InsertedContent = insertLow(Elems.getMatch(), false, NewStuff);
		}

		return InsertedContent;
	}

/**	Inserts before, or after, the specified item ({@link Match}) in this {@link Markup}. The high-level version of this
 *	method should respect tag hierarchies; given a match for a {@link StartTag}, an "insert before" operation
 *	should insert the NewStuff before the <code>StartTag</code> (<code>insertLow</code> behaves this way), and, more 
 *	significantly, an "insert after" operation should insert the NewStuff after the {@link EndTag} that matches with 
 *	the specified <code>StartTag</code>. <code>insertLow</code> does NOT behave this way, and is NOT meant to do so.
 *
 *	@param Target				The <code>Match</code> relative to which the new material will be inserted.
 *	@param InsertBefore			<code>true</code> if the new material should be inserted before Target, <code>false</code>
 *								if it should be inserted after Target.
 *	@param NewStuff				An object representing the new material. A number of different types are explicitly
 *								supported. If the object is of a type not explicitly supported, its <code>toString</code>
 *								method is invoked, and the resulting string is inserted. If the object is <code>null</code>,
 *								the string "null" is inserted.
 *	@return						A <code>Match</code> instance for the inserted material.
 */
	private Match insertLow
	(
		final Match				Target,
		final boolean			InsertBefore,
		final Object			NewStuff
	)
	{
		try {
			final int				TargetIndex   = Target.getIndex();
			final int				InsertIndex   = InsertBefore ? TargetIndex : TargetIndex + 1;
			Match					NewStuffMatch = Target;		//	When all else fails, this is a decent default to return.

			if (NewStuff == null) {

				NewStuffMatch = insertLowString("null", InsertIndex);

			} else if (NewStuff instanceof String) {

				NewStuffMatch = insertLowString((String) NewStuff, InsertIndex);

			} else if (NewStuff instanceof Match) {

			//	todo Consider namespace implications here.

				NewStuffMatch = insertLowList(((Match) NewStuff).enumerate(), 0, InsertIndex);

			} else if (NewStuff instanceof Results) {

			//	Note: In practice, this case doesn't come-up much because Results.setContent, Results.replace, etc. 
			//	already detect cases in which the content is a Results object, and process such cases by initiating
			//	separate operations for each of the matches contained in the Results object. Doing so allows those
			//	Results methods to return new Results objects that reference all of the inserted content, unlike this
			//	method, which can only return a single Match.

			//	todo Consider namespace implications here.
				
				final RangeList			InsertableList = new RangeList(32);
				
				for (final Match Elem : (Results) NewStuff)
					buildInsertableList(Elem.enumerate(), InsertableList);
 
				NewStuffMatch = insertListReturnMatch(InsertableList, InsertIndex);
				
			} else if (NewStuff instanceof Markup) {
				final Markup			NewMarkup = (Markup) NewStuff;

				if (NewMarkup.isEmpty() == false) {
				
					NewStuffMatch = insertListReturnMatch(NewMarkup.getList(), InsertIndex);
					
				//	if (NewStuff instanceof MutableMarkup)
				//		NewStuffMatch = insertListReturnMatch(((MutableMarkup) NewMarkup).getList(), InsertIndex);
				//	else
				//		NewStuffMatch = insertLowList(NewMarkup.enumerate(), NewMarkup.size(), InsertIndex);
				}
				
			} else if (NewStuff instanceof Range) {

				insertLowRange((Range) NewStuff, InsertIndex);

			} else if (NewStuff instanceof Throwable) {
				final Throwable			Problem = (Throwable) NewStuff;
				String					Message = Problem.getMessage();

				if (Null.orEmpty(Message))
					Message = Problem.getClass().getName();

				final String			StackTraces = StackTrace.toString(Problem);
				final StringBuilder		XMLBuff     = new StringBuilder(StackTraces.length() + 1024);
				
				HTML.encode(Message, XMLBuff);
				XMLBuff.append(" <!-- ");
				appendFilteredAndEncodedComment(StackTraces, XMLBuff);

			//	From <http://www.w3.org/TR/REC-xml/#sec-comments>:
			//	"Note that the grammar does not allow a comment ending in --->."

				XMLBuff.append(" -->");	//	The leading space prevents a comment from ever ending in "--->".

				try {
					insertLow(Target, InsertBefore, new XMLString(XMLBuff.toString()));
				} catch (TagException e) {
					throw new RuntimeException(e);
				}

			} else
				NewStuffMatch = insertLowOther(NewStuff, InsertIndex);
			
		//	todo Try to support Downloadable instances here. Ideally, inserting a Downloadable would automatically invoke Qwicap.addDownloadable, thereby avoiding a common mistake.
			
			return NewStuffMatch;

		} finally {
			
		//	Whatever happened above, the list has almost certainly been changed in some way.
		//	Here we guarantee that that fact is duly noted.
			
			ChangeCount++;
		}
	}
	
	
/*	Namespace maintenance:

	* We probably want to keep up-to-date a master list of namespaces used in any body of markup.
	
	* Whenever an item is inserted into markup, its namespace must be compared with the parent tag's namespace. If
		they differ, a namespace definition attribute will have to added to it. Note that the inserted item may already
		have such an attribute, but that the attribute might employ a prefix that is not unique in the document into 
		which it has been inserted, and therefore may have to changed. If the prefix has changed, all uses of that 
		prefix in the inserted material will have to be replaced, too.
	
	* Do we need to anything when material is deleted? We'll certainly want to update the master list of namespaces,
		since we may be able to remove some items from the list.
*/


	private static StringBuilder appendFilteredAndEncodedDeclaration
	(
		final String			RawMaterial,
		final StringBuilder		DestBuff
	)
	{
	//	It's not clear at the moment what limitations apply to the content of a declaration. It shouldn't include a '>'
	//	character, but does it make a difference if the '>' character is inside quoted material in the content?
	
		DestBuff.append(RawMaterial);

		return DestBuff;
	}
	
	private static StringBuilder appendFilteredAndEncodedCDATA
	(
		final String			RawMaterial,
		final StringBuilder		DestBuff
	)
	{
	//	Anything is allowed inside a CDATA block except the sequence "]]>". Here we encode that sequence, if it is present.
	
		DestBuff.append(RawMaterial.replace("]]>", "]]&gt;"));

		return DestBuff;
	}

	private static StringBuilder appendFilteredAndEncodedComment
	(
		final String			RawMaterial,
		final StringBuilder		DestBuff
	)
	{
	//	Remove all control characters (characters less than 0x20), except for 0x09, 0x0A, and 0x0D.

		final char[]			Chars      = RawMaterial.toCharArray();
		int						WriteIndex = 0;

		for (int ReadIndex = 0, Count = Chars.length; ReadIndex < Count; ReadIndex++) {
			final char				Char = Chars[ReadIndex];

			if (Char >= 0x20 || Char == 0x09 || Char == 0x0A || Char == 0x0D) {

				if (WriteIndex != ReadIndex)
					Chars[WriteIndex] = Char;

				WriteIndex++;
			}
		}

	//	From <http://www.w3.org/TR/REC-xml/#sec-comments>:
	//
	//	Definition: Comments MAY appear anywhere in a document outside other markup; in addition, they MAY appear 
	//	within the document type declaration at places allowed by the grammar. They are not part of the document's 
	//	character data; an XML processor MAY, but need not, make it possible for an application to retrieve the text of 
	//	comments. For compatibility, the string "--" (double-hyphen) MUST NOT occur within comments.] Parameter entity 
	//	references MUST NOT be recognized within comments.
		
		DestBuff.append(DashDashPattern.matcher(CharBuffer.wrap(Chars, 0, WriteIndex)).replaceAll("__"));

		return DestBuff;
	}

	private Match insertLowOther
	(
		final Object			ElemObj,
		final int				InsertIndex
	)
	{
		return insertLowString(otherToString(new StringBuilder(512), ElemObj, 0).toString(), InsertIndex);
	}

	private StringBuilder otherToString
	(
		final StringBuilder		Buff,
		final Object			ElemObj,
		final int				ListLevel
	)
	{
		if (StringConverter == null)
			StringConverter = new SmartStringConverter();
		
		return StringConverter.appendLow(Buff, ElemObj, ListLevel);
	}

	private Match insertLowString
	(
		final String			ElemStr,
		final int				InsertIndex
	)
	{
	//	todo Consider being aware of context. Specifically, do not perform the HTML encoding if we're inserting into a CDATA element.

		HTMLEntityCodec			Encoder = HTML;

		if (InsertIndex < List.size()) {
			final Match				Parent = createMatch(List.Array[InsertIndex], InsertIndex).getParent();

		//	todo Consider being aware of namespaces here. Only use the following special case code when the namespace is XHTML or null.

			if (Parent != null) {
				final String			ParentType = Parent.getType();

				if ("pre".equals(ParentType) || "textarea".equals(ParentType))
					Encoder = HTMLNewLineNotEncodedAsBR;
			}
		}

	//	todo Consider that it might be preferable to create an ImContent object, since inserted strings are seldom changed.
	//       However, the MuContent object can use ElemStr (or the encoded version of it, if encoding is necessary) as its
	//	     backing store. Whereas using the ImContent object would compel us to convert ElemStr (or the encoded string) to
	//	     an array of characters.

		final Content			NewContent = new MuContent(Encoder.encode(ElemStr));

		List.add(InsertIndex, NewContent);
		MatchCache.add(InsertIndex, 1);

		return createMatch(NewContent, InsertIndex);
	}

	private Match insertLowRange
	(
		final Range				NewStuff,
		final int				InsertIndex
	)
	{
		final Range				InsertableItem = getInsertableItem(NewStuff);

		List.add(InsertIndex, InsertableItem);
		MatchCache.add(InsertIndex, 1);

		final Match				NewStuffMatch = createMatch(InsertableItem, InsertIndex);

		if (NewStuff instanceof NamedItem) {
			final NamedItem			NI = (NamedItem) NewStuff;

			if (NI.getNamespace() == null) {
				final Namespace			Ns = getEnclosingNamespace(InsertIndex);

				if (Ns != null) {

					if (NI instanceof Mutable) {

						((MuNamedItem) NI).setNamespace(Ns);

					} else {

					//	Create an immutable clone of this immutable item, where the clone is in the Ns namespace. 
					//	Then replace the old immutable item in List.Array with the new one.

						final ImNamedItem		NIClone = ((ImNamedItem) NI).clone(Ns);

						List.Array[InsertIndex] = NIClone;
						((MutableMatch) NewStuffMatch).set(NIClone);
					}
				}
			}
		}

		return NewStuffMatch;
	}

	private Match insertLowList
	(
		final MarkupEnumeration	Elems,
		final int				ElemsCountHint,
		final int				InsertIndex
	)
	{
		return insertListReturnMatch(buildInsertableList(Elems, ElemsCountHint), InsertIndex);
	}

	private Match insertListReturnMatch
	(
		final RangeList			SrcList,
		final int				InsertIndex
	)
	{
		final Match				NewStuff;
		final int				NewItemCount;
		
		{
			final Range[]			SrcListArray         = SrcList.Array;
			final boolean			SrcBeginsWithXMLDecl = SrcList.size() > 0 && SrcListArray[0] instanceof XMLDeclaration;
			final int				SrcListIndex;
			
			if (InsertIndex == 0) {
				
			//	We're inserting at the start of this markup. If both lists begin with an XML declaration, omit the one from
			//	the source list.
				
				if (SrcBeginsWithXMLDecl && List.size() > 0 && List.Array[0] instanceof XMLDeclaration)
					SrcListIndex = 1;	//	Both lists begin with an XML declaration, so do NOT copy the XML declaration from the source material. It would be redundant.
				else	
					SrcListIndex = 0;	//	This markup lacks an XML declaration, but the source includes one. Include the one from the source.
				
			} else {
				
			//	We're inserting into this markup at some place other than the start. If the source material includes an
			//	XML declaration, omit it.
				
				if (SrcBeginsWithXMLDecl)
					SrcListIndex = 1;
				else
					SrcListIndex = 0;
			}
			
		//	Insert the source material into this markup, and expand our match cache to accomodate it.
			
			NewItemCount = SrcList.size() - SrcListIndex;
				
			List.addSubsection(InsertIndex, SrcList, SrcListIndex, NewItemCount);
			MatchCache.add(InsertIndex, NewItemCount);
			
			NewStuff = createMatch(SrcListArray[SrcListIndex], InsertIndex);
		}
		
	//	Assign the new items that have null namespaces to the enclosing namespace.

		{
			final Namespace			Ns = getEnclosingNamespace(InsertIndex);

			if (Ns != null) {
				final Range[]			DestListArray = List.Array;
				
				for (int Index = InsertIndex, Count = NewItemCount; --Count >= 0; Index++) {
					final Range				R = DestListArray[Index];

					if (R instanceof NamedItem) {
						final NamedItem			RN = (NamedItem) R;

						if (RN.getNamespace() == null) {

							if (RN instanceof Mutable) {

								((MuNamedItem) RN).setNamespace(Ns);

							} else {

							//	Create an immutable clone of this immutable item, where the clone is in the Ns namespace. 
							//	Then replace the old immutable item in List.Array with the new one.

								DestListArray[Index] = ((ImNamedItem) RN).clone(Ns);
							}
						}
					}
				}

			//	If we cloned the first item that was inserted, the MutableMatch stored in NewStuff is no longer valid. 
			//	Here we detect that case, and repair the MutableMatch.

				final Range			MatchRange = DestListArray[InsertIndex];

				if (NewStuff.get() != MatchRange)	//	Yes, a reference equality comparison is desired here.
					((MutableMatch) NewStuff).set(MatchRange);
			}
		}

		return NewStuff;
	}

	private Namespace getEnclosingNamespace
	(
		final int				InsertIndex
	)
	{
		int						EndTagCount = 0;
		final Range[]			ListArray   = List.Array;

		for (int Index = InsertIndex; --Index >= 0;) {
			final Range				Elem = ListArray[Index];

			if (Elem.isEndTag()) {

				EndTagCount++;

			} else if (Elem.isStartTag()) {

				if (EndTagCount == 0) {

					if (Elem instanceof NamedItem) {
						final Namespace			Ns = ((NamedItem) Elem).getNamespace();

						if (Ns != null)
							return Ns;
					}

				} else
					EndTagCount--;
			}
		}

		return null;
	}

	private RangeList buildInsertableList
	(
		final MarkupEnumeration	Elems,
		final int				SizeHint	//	Hint about the number of items Elems will return. Pass zero if you have no idea.
	)
	{
		return buildInsertableList(Elems, SizeHint <= 0 ? new RangeList(32) : new RangeList(SizeHint));
	}

	private RangeList buildInsertableList
	(
		final MarkupEnumeration	Elems,
		final RangeList			InsertableList
	)
	{
	//	MutableMarkup objects may contain a mix of mutable and immutable
	//	elements, distinguished by whether or not they implement the
	//	Mutable interface. We can directly insert immutable objects into
	//	this list, even if they are already present in the list, but mutable 
	//	objects must be cloned in order to avoid unintended consequences 
	//	(changing the mutable object in one body of markup would alter all 
	//	other bodies of markup in which it is present - a behavior that 
	//	would be both interesting, and very bad).

		while (Elems.hasMoreElements())
			InsertableList.add(getInsertableItem(Elems.nextRange()));

		return InsertableList;
	}

	private Range getInsertableItem
	(
		final Range				OriginalItem
	)
	{

	//	todo Consider that because this implementation does not always clone mutable items, it violates the rule described in buildInsertableList.

		if (OriginalItem instanceof Mutable && List.contains(OriginalItem))
	//	if (List.contains(OriginalItem))			//	Original implementation, when we cloned all duplicate items.
			return (Range) OriginalItem.clone();
		else
			return OriginalItem;
	}

/**	Encodes a string such that it can be safely embedded in an XHTML document without risk of the document's syntax
 * 	being corrupted, or the contents of the string being interpreted as XHTML markup (a potential security hole).
 * 
 * @param RawStr		The string to be encoded.
 * @return				A string with all reserved characters encoded, or the original string object if no encoding was
 *						necessary.
 */
	public static String htmlEncode
	(
		final String			RawStr
	)
	{
		return HTML.encode(RawStr);
	}

/**	Decodes a string such any HTML character entities in it are translated into corresponding Unicode characters.
 * 
 * @param PotentiallyEncodedStr	The string to be decoded.
 * @return						A string with all HTML characters entities decoded into corresponding Unicode characters.
 */
	public static String htmlDecode
	(
		final String			PotentiallyEncodedStr
	)
	{
		return HTML.decode(PotentiallyEncodedStr);
	}

/**	Encodes a string such that it can be safely embedded in an XHTML document without risk of the document's syntax
 * 	being corrupted, or the contents of the string being interpreted as XHTML markup (a potential security hole).
 * 
 * @param RawStr		The string to be encoded.
 * @return				A string with all reserved characters encoded, or the original string object if no encoding was
 *						necessary.
 */
	public static Characters htmlEncode
	(
		final Characters		RawStr
	)
	{
		return HTML.encode(RawStr);
	}

/**	Decodes a string such any HTML character entities in it are translated into corresponding Unicode characters.
 * 
 * @param PotentiallyEncodedStr	The string to be decoded.
 * @return						A string with all HTML characters entities decoded into corresponding Unicode characters.
 */
	public static Characters htmlDecode
	(
		final Characters		PotentiallyEncodedStr
	)
	{
		return HTML.decode(PotentiallyEncodedStr);
	}

/**	Encodes a string such that it can be safely embedded in an XHTML document without risk of the document's syntax
 * 	being corrupted, or the contents of the string being interpreted as XHTML markup (a potential security hole).
 * 
 * @param RawStr		The string to be encoded.
 * @param Buff			The <code>StringBuffer</code> to which the encoded string should be appended.
 * @return				A string with all reserved characters encoded, or the original string object if no encoding was
 *						necessary.
 */
	public static boolean htmlEncode
	(
		final Characters		RawStr,
		final StringBuffer		Buff
	)
	{
		return HTML.encode(RawStr, Buff);
	}

/**	Decodes a string such any HTML character entities in it are translated into corresponding Unicode characters.
 * 
 * @param PotentiallyEncodedStr	The string to be decoded.
 * @param Buff					The <code>StringBuffer</code> to which the decoded string should be appended.
 * @return						A string with all HTML characters entities decoded into corresponding Unicode characters.
 */
	public static boolean htmlDecode
	(
		final Characters		PotentiallyEncodedStr,
		final StringBuffer		Buff
	)
	{
		return HTML.decode(PotentiallyEncodedStr, Buff);
	}

/**	Encodes a string such that it can be safely embedded in an XHTML document without risk of the document's syntax
 * 	being corrupted, or the contents of the string being interpreted as XHTML markup (a potential security hole).
 * 
 * @param RawStr		The string to be encoded.
 * @param Buff			The <code>StringBuffer</code> to which the encoded string should be appended.
 * @return				A string with all reserved characters encoded, or the original string object if no encoding was
 *						necessary.
 */
	public static boolean htmlEncode
	(
		final Characters		RawStr,
		final StringBuilder		Buff
	)
	{
		return HTML.encode(RawStr, Buff);
	}

/**	Decodes a string such any HTML character entities in it are translated into corresponding Unicode characters.
 * 
 * @param PotentiallyEncodedStr	The string to be decoded.
 * @param Buff					The <code>StringBuffer</code> to which the decoded string should be appended.
 * @return						A string with all HTML characters entities decoded into corresponding Unicode characters.
 */
	public static boolean htmlDecode
	(
		final Characters		PotentiallyEncodedStr,
		final StringBuilder		Buff
	)
	{
		return HTML.decode(PotentiallyEncodedStr, Buff);
	}
}







