/*	QWICAP (Quick Web Interface for Conventional Applications) by Chris W. Johnson
	Copyright (C) 2003-2010 The University of Texas at Austin ("U.T. Austin"). All rights reserved.
	
	U.T. Austin's copyright notice in this evolving version reflects our original authorship. It does not assert 
	copyright over changes contributed by others.
	
	This library is free software; you can redistribute it and/or modify it under the terms of the GNU Lesser General 
	Public License as published by the Free Software Foundation; either version 2.1 of the License, or (at your option) 
	any later version.
	
	This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied 
	warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for 
	more details.
	
	You should have received a copy of the GNU Lesser General Public License along with this library; if not, write to 
	the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/


//  MutableMatch.java
//  qwicap
//
//  Created by Chris W. Johnson on Sat Jun 21 2003.


package edu.utexas.its.eis.tools.qwicap.template.xml.mutable;


import edu.utexas.its.eis.tools.qwicap.template.css.Results;
import edu.utexas.its.eis.tools.qwicap.template.xml.Markup;
import edu.utexas.its.eis.tools.qwicap.template.xml.Match;
import edu.utexas.its.eis.tools.qwicap.template.xml.TagException;
import edu.utexas.its.eis.tools.qwicap.template.xml.immutable.ImmutableMarkup;
import edu.utexas.its.eis.tools.qwicap.template.xml.immutable.ImmutableMatch;
import edu.utexas.its.eis.tools.qwicap.template.xml.structure.Comment;
import edu.utexas.its.eis.tools.qwicap.template.xml.structure.Range;
import edu.utexas.its.eis.tools.qwicap.template.xml.structure.CDATA;
import edu.utexas.its.eis.tools.qwicap.template.xml.structure.Declaration;

import java.util.NoSuchElementException;


/**	<code>MutableMatch</code> uniquely identifies a particular element within a body of XHTML markup represented by a 
 *	{@link MutableMarkup} object, while supplying implementations of all the standard {@link Match} methods. 
 * 	<code>MutableMatch</code> objects are used to identify <i>all</i> elements within <code>MutableMarkup</code>, even 
 * 	immutable elements; the important issue isn't the mutability of the element, but the unique behaviors required of 
 * 	matches in a mutable document (and, conversely, the extremely simple behaviors that are sufficient for matches in 
 * 	an immutable document).
 *	
 *	@author Chris W. Johnson
 */
public final class MutableMatch extends Match {
	
	private final MutableMarkup		Markup;
	private int						PrevChangeCount;
	private Range					Item;
	private int						Index;
	
	
	MutableMatch
	(
		final MutableMarkup		Markup,
		final Range				Item,
		final int				Index
	)
	{
		this.Markup          = Markup;
		this.PrevChangeCount = Markup.getChangeCount();
		this.Item            = Item;
		this.Index           = Index;
	}
	
/**	Tests for reference equality with another object. Reference equality is sufficient because each 
 *	<code>MutableMarkup</code> instance maintains a complete list of the <code>MutableMatch</code> objects that refer 
 *	to its markup, always re-uses those, and never creates duplicates. Also, <code>MutableMatch</code> objects are used 
 *	to identify all components in mutable markup, even the <i>immutable</i> ones. So testing for equality of matches 
 * 	associated with mutable markup is effortless. (Note that the story is very different for the 
 *	{@link ImmutableMatch} objects employed by {@link ImmutableMarkup}.)
 * 
 * @param Other					Another object to be tested for equality with this object.
 * @return						<code>true</code> if the objects match, <code>false</code> otherwise.
 */
	@Override public boolean equals
	(
		final Object			Other
	)
	{
	//	Mutable markup does maintain a match cache, so only one MutableMatch instance will ever be created for any 
	//	element in a given body of mutable markup. Therefore, we can establish equality by making a simple test of
	//	reference equality.
		
		return this == Other;
	}
	
	@Override public Markup getMarkup() {
		return Markup;
	}
	
	@Override public int getIndex() throws NoSuchElementException {
		
		check();
		return Index;
	}
	
	@Override public Range get() throws NoSuchElementException {
		
		check();
		return Item;
	}
	
	Match set
	(
		final Range				NewItem
	)
	{
		Item = NewItem;
		PrevChangeCount = 0;	//	Force a check() call to be made the next time get() is invoked.
		
		return this;
	}
	
/**	Change the index at which the element referenced by this MutableMatch object is believed to occur in the Markup.
 *	Due to the current package layout, this method has to be public, but it is <strong>not</strong> meant for public 
 *	use. It should only be used by the MutableMarkup match cache maintenance methods.
 * 
 * @param NewIndex				The new index of the element.
 */
	void setIndex
	(
		final int				NewIndex
	)
	{
		Index = NewIndex;
	}
	
	@Override public Match duplicate
	(
		final boolean			InsertDuplicateBefore
	)
	{
		return Markup.duplicate(this, InsertDuplicateBefore);
	}
	
	@Override public Match replace
	(
		final Object			Replacement
	)
	{
		return Markup.replace(this, Replacement);
	}
	
	@Override public Match add
	(
		final boolean			InsertBefore,
		final Object			NewStuff
	)
	{
		return Markup.insert(this, InsertBefore, NewStuff);
	}
	
	@Override public void delete() {
		Markup.delete(this);
	}
	
	@Override public Match deleteContent() throws TagException {
		return Markup.deleteContent(this);
	}
	
	@Override public Results getContent
	(
		final Results			PrevResults
	)
	{
		final Range				Item = get();
		
		if (Item.isStartTag()) {
			
			return super.getContent(PrevResults);
		
		} else {
			final Results			Content = new Results(PrevResults);
			
			if (Item instanceof Comment || Item instanceof CDATA || Item instanceof Declaration)
				Content.addMatch(new MutableIntraElementContentMatch(this));
			
			return Content;
		}
	}
	
	@Override public Match setContent
	(
		final Object			Value
	)
	throws TagException
	{
		return Markup.setContent(this, Value);
	}
	
	@Override public Match addContent
	(
		final boolean			InsertBeforeExistingContent,	//	If false, we insert after the existing content.
		final Object			Content
	)
	throws TagException
	{
		return Markup.addContent(this, InsertBeforeExistingContent, Content);
	}
	
	@Override public Match addClass
	(
		final String			ClassList
	)
	throws TagException
	{
		return Markup.addClass(this, ClassList);
	}
	
	@Override public Match setClass
	(
		final String			ClassList
	)
	throws TagException
	{
		return Markup.setClass(this, ClassList);
	}
	
	@Override public Match deleteClass
	(
		final String			ClassList
	)
	throws TagException
	{
		return Markup.deleteClass(this, ClassList);
	}
	
	@Override public Match addAttribute
	(
		final String			AttrName,
		final Object			AttrValue	//	Can be null.
	)
	throws TagException
	{
		return Markup.addAttribute(this, AttrName, AttrValue);
	}
	
/*	public Match addAttribute
	(
		final Characters		AttrName,
		final Characters		AttrValue
	)
	throws TagException
	{
		return (Markup.addAttribute(this, AttrName, AttrValue));
	}
*/	
	@Override public Match setAttribute
	(
		final String			AttrName,
		final Object			AttrValue	//	Can be null.
	)
	throws TagException
	{
		return Markup.setAttribute(this, AttrName, AttrValue);
	}
	
	@Override public Match updateAttribute
	(
		final String			AttrName,
		final Object			AttrValue	//	Can be null.
	)
	throws TagException
	{
		return Markup.updateAttribute(this, AttrName, AttrValue);
	}
	
	@Override public Match deleteAttribute
	(
		final String			AttrName
	)
	throws TagException
	{
		return Markup.deleteAttribute(this, AttrName);
	}
	
	private void check() throws NoSuchElementException {
		
		if (!exists())
			throw new NoSuchElementException("Qwicap was asked to operate on an element (\"" + Item + 
				"\") that no longer exists in the markup \"" + Markup.getMarkupName() + "\". One situation that " +
				"causes this is doing a delete on a Results object that contains duplicate matches for one or " +
				"more elements, thereby causing an element to be \"deleted\" after it has already been deleted. " +
				"That problem might be prevented by refining the CSS search pattern so that duplicates do not " +
				"occur, and can always be avoided by using the \"Results.distinct\" method to eliminate any " +
				"duplicate matches before using methods like \"delete\". Another way to cause this problem is to " +
				"operate on a retained reference to this MutableMatch object (possibly within a Results object), " +
				"after it has been deleted, extracted, cut, or otherwise removed from its markup."
			);
	}
	
	@Override public boolean exists() {
		final int				CurrChangeCount = Markup.getChangeCount();
		
		if (CurrChangeCount != PrevChangeCount) {
			
		//	Ensure that the item index is up-to-date. There are no longer any situations where it's allowed to be stale,
		//	so this is just bug detection code at this stage.
			
			if (Markup.get(Index) != Item)
				return false;
			
			PrevChangeCount = CurrChangeCount;	
		}
		
		return true;
	}
}







