/*	QWICAP (Quick Web Interface for Conventional Applications) by Chris W. Johnson
	Copyright (C) 2003-2010 The University of Texas at Austin ("U.T. Austin"). All rights reserved.
	
	U.T. Austin's copyright notice in this evolving version reflects our original authorship. It does not assert 
	copyright over changes contributed by others.
	
	This library is free software; you can redistribute it and/or modify it under the terms of the GNU Lesser General 
	Public License as published by the Free Software Foundation; either version 2.1 of the License, or (at your option) 
	any later version.
	
	This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied 
	warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for 
	more details.
	
	You should have received a copy of the GNU Lesser General Public License along with this library; if not, write to 
	the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/


//  Range.java
//  qwicap
//
//  Created by Chris W. Johnson on Mon Jul 14 2003.


package edu.utexas.its.eis.tools.qwicap.template.xml.structure;


import edu.utexas.its.eis.tools.qwicap.util.Characters;

import java.io.PrintWriter;
import java.io.IOException;


/**	<p>Super-interface of a hierarchy of interfaces that provide for the classification, and 
 * 	manipulation, of the complete contents of any XML markup. The hierarchy of interfaces looks like this:
 *	</p>
 * 
 *	<ul>
 *		<li>{@link Range}
 *			<ul>
 *				<li>{@link Whitespace}</li>
 *				<li>{@link Item}
 *					<ul>
 *						<li>{@link XMLDeclaration}</li>
 *						<li>{@link Comment}</li>
 *						<li>{@link Content}</li>
 *						<li>{@link CDATA}</li>
 *						<li>{@link NamedItem}
 *							<ul>
 *								<li>{@link Attribute}</li>
 *								<li>{@link Declaration}</li>
 *								<li>{@link Tag}
 *									<ul>
 *										<li>{@link EndTag}</li>
 *										<li>{@link TagWithAttributes}
 *											<ul>
 *												<li>{@link StartTag}</li>
 *												<li>{@link EmptyTag}</li>
 *											</ul>
 *										</li>
 *									</ul>
 *								</li>
 *							</ul>
 *						</li>
 *					</ul>
 *				</li>
 *			</ul>
 *		</li>
 *	</ul>
 * 
 *	<p>The name "Range" comes from the idea that this represents a segment, an area, a span of consecutive characters, 
 *	or, in other words, a range of characters, within an XML document. It makes no assertion about what the characters 
 * 	in its range of the document represent - that's what the sub-interfaces are for.
 * 	</p>
 *
 *	@author Chris W. Johnson
 */

public interface Range extends Cloneable {

/**	Test this range of characters for equality with another set of characters. The comparison is sensitive to 
 * 	case.
 * 
 * @param Chars				The set of characters to be tested for equality with this range of characters.
 * @return					<code>true</code> if the characters in this range exactly match the characters in
 * 							the <code>Chars</code> parameter, <code>false</code> otherwise.
 */	
	public boolean		rangeEquals(Characters Chars);

/**	Returns a <code>String</code> representation of this range of characters.
 * 
 * @return					A <code>String</code> representation of this range of characters.
 */	
	@Override public String		toString();
	
	public Object		clone();

/**	Write this range of characters to the specified {@link PrintWriter}. 
 * 
 * @param Out				This range of characters is written to this <code>PrintWriter</code>.
 * @throws IOException		If anything goes wrong while writing.
 */
	public void			write(PrintWriter Out) throws IOException;

/**	Returns <code>true</code> if this range of characters constitutes a "start" tag, for example: 
 * 	"<samp>&lt;body></samp>". The same determination can be made with the test "<code>instanceof StartTag</code>",
 * 	but using this method is faster.
 * 
 * @return					<code>true</code> if this range of characters constitutes a "start" tag,
 * 							<code>false</code> otherwise.
 */
	public boolean		isStartTag();

/**	Returns <code>true</code> if this range of characters constitutes an "end" tag, for example: 
 * 	"<samp>&lt;/body></samp>". The same determination can be made with the test "<code>instanceof EndTag</code>",
 * 	but using this method is faster.
 * 
 * @return					<code>true</code> if this range of characters constitutes an "end" tag,
 * 							<code>false</code> otherwise.
 */
	public boolean		isEndTag();

/**	Returns <code>true</code> if this range of characters constitutes an "empty" tag, for example: 
 * 	"<samp>&lt;br/></samp>". The same determination can be made with the test "<code>instanceof EmptyTag</code>",
 * 	but using this method is faster.
 * 
 * @return					<code>true</code> if this range of characters constitutes an "empty" tag,
 * 							<code>false</code> otherwise.
 */
	public boolean		isEmptyTag();

/**	Returns <code>true</code> if this range of characters constitutes a "start" or an "empty" tag; in other words, if
 * 	this is a tag that <i>can</i> possess attributes. The same determination can be made with the test 
 * 	"<code>instanceof TagWithAttributes</code>", but using this method is faster. Note that this method returns 
 * 	<code>true</code> for all "start" and "empty" tags; whether or not they actually <i>do</i> possess attributes is 
 * 	irrelevant.
 * 	
 * @return					<code>true</code> if this range of characters constitutes a "start" or an "empty" tag,
 * 							<code>false</code> otherwise.
 */
	public boolean		isTagWithAttributes();
}


/*

Class hierarchy:
	
	Range  [abstract]
	  |
	  +-- Whitespace
	  |
	  +-- Item  [abstract]			** Consider renaming "Element". Unfortunately, "Content" doesn't exactly fit in such a catagory.
	        |
	        +-- XMLDeclaration
	        |
	        +-- Declaration
	        |
	        +-- Comment
			|
			+-- Content
	        |
	        +-- CDATA
	        |
	        +-- NamedItem  [abstract]
	        	  |
	        	  +-- Attribute
	              |
				  +-- Tag  [abstract]
	                    |
				        +-- EndTag
			            |
				        +-- TagWithAttributes  [abstract]
				              |
	                          +-- StartTag
	                          |
	                          +-- EmptyTag
							  
							  
	Tag with Attributes parsed:
	
	<a href="http://www.thing.edu/">
	||||                           ||
	|||+-- ContentStart            |+-- PyEnd
	||+--- NameEnd                 +--- ContentEnd
	|+---- NameStart
    +----- PyStart
	
	
	Attribute parsed:
	
	href="http://www.thing.edu/"
	|   | |                    ||
	|   | +-- ContentStart     |+-- PyEnd
	|   +---- NameEnd          +--- ContentEnd
	+-------- PyStart, NameStart
	
	
	End Tag parsed:
	
	</table>
	| |    ||
	| |    |+-- PyEnd
	| |    +--- NameEnd, ContentStart, ContentEnd
	| +-------- NameStart
	+---------- PyStart
	
	
	Declaration parsed:
	
	<?xml version="1.0" encoding="UTF-8"?>
	| |                                  ||
	| +-- ContentStart                   |+-- PyStop
	+---- PyStart                        +---ContentEnd
	
	
	Comment parsed:
	
	<!-- Fred the Wonder Llama bids you greetings. -->
	|    |                                        |   |
	|    +-- ContentStart                         |   +-- PyStop
	+---- PyStart                                 +------ ContentEnd
	
	
	Content parsed:
	
	\r\t\tFred the Wonder Llama bids you greetings. \r\r
	|     |                                        |    |
	|     +-- ContentStart                         |    +-- PyStop
	+---- PyStart                                  +---ContentEnd
	
	
*/




