/*	QWICAP (Quick Web Interface for Conventional Applications) by Chris W. Johnson
	Copyright (C) 2003-2010 The University of Texas at Austin ("U.T. Austin"). All rights reserved.
	
	U.T. Austin's copyright notice in this evolving version reflects our original authorship. It does not assert 
	copyright over changes contributed by others.
	
	This library is free software; you can redistribute it and/or modify it under the terms of the GNU Lesser General 
	Public License as published by the Free Software Foundation; either version 2.1 of the License, or (at your option) 
	any later version.
	
	This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied 
	warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for 
	more details.
	
	You should have received a copy of the GNU Lesser General Public License along with this library; if not, write to 
	the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/


package edu.utexas.its.eis.tools.qwicap.util;


import java.util.Collection;
import java.util.Iterator;
import java.util.logging.Logger;
import java.lang.reflect.Array;


/**	Converts an array or {@link Collection} to a neatly formatted String when its <code>toString</code> method is 
 *	invoked. This class is useful anytime an array or collection of objects is to be converted to a string, but it is 
 *	particularly useful in conjunction with the {@link Logger}, since it allows an array of objects 
 *	to be incorporated into a log message in a human-readable fashion. Moreover, because it defers all of the 
 *	nitty-gritty work of the string conversions, formatting and appending until its <code>toString</code> method is 
 *	invoked, an instance of this class will do no work if the Logger discards the log entry in which it is used.
 *	
 *	<p>Note that when performing quoting this class attempts to avoid redundant quoting, so, before any item is quoted it 
 *	is checked to see if it already begins and ends with the relevant opening- and closing-quote sequences. If it does, 
 *	no additional quotes are added.
 * 	</p>
 *
 *	<p>Also note that any type of object may be passed to this class' constructors in place of an array or 
 *	<code>Collection</code>. Therefore, this class can be used to supply deferred string conversion for any other class, 
 * 	and from time-to-time the quoting logic might come in handy, too.
 * 	</p>
 *	
 *	@author Chris W. Johnson
 */
public class ArrayToString implements DeferredString {
	
	private static final char[]		kNullChars      = { 'n','u','l','l' };
	private static final char[][]	kNullCharsArray = { kNullChars };
	
	private final Object		Objs;
	private String				ObjsStr;
	private String				SepStr          = ", ";
	private String				FinalSepStr     = " and ";
	private String				OpenQuoteStr    = "\"";
	private String				CloseQuoteStr   = "\"";
	private boolean				AlwaysUseQuotes;
	private boolean				TrimStrings;
	private int					Elements;
	
	
/**	Constructs a new <code>ArrayToString</code> object that will create a <code>String</code> representation of the 
 *	supplied array or <code>Collection</code> object when this object's <code>toString</code> method is invoked, and not
 *	before then. Commas will be placed between all elements except for the final two which will be separated by the
 *	string " and ". Elements that will benefit from quoting are surrounded by the double-quote ('"') character. An 
 *	item is judged to benefit from being quoted if it contains any whitespace, or a comma.
 * 
 * @param A						An array or {@link Collection} object whose contents should be converted to a 
 *								<code>String</code>. At this time, arrays or <code>Collection</code> objects whose elements
 *								are other arrays or <code>Collection</code> objects are not supported in a useful manner.
 */
	public ArrayToString
	(
		final Object			A
	)
	{
		Objs = A;
	}
	
/**	Constructs a new <code>ArrayToString</code> object that will create a <code>String</code> representation of the 
 *	supplied array or <code>Collection</code> object when this object's <code>toString</code> method is invoked, and not
 *	before then. Commas will be placed between all elements except for the final two which will be separated by the
 *	string specified in the <code>FinalSeparatorStr</code> parameter. Elements that will benefit from quoting are 
 *	surrounded by the double-quote ('"') character. An item is judged to benefit from being quoted if it contains
 *	any whitespace, or a comma.
 * 
 * @param A						An array or {@link Collection} object whose contents should be converted to a 
 *								<code>String</code>. At this time, arrays or <code>Collection</code> objects whose elements
 *								are other arrays or <code>Collection</code> objects are not supported in a useful manner.
 * @param FinalSeparatorStr		The string that should be placed between the final two elements in the array or 
 *								<code>Collection</code>, for example: <samp>" and "</samp> or <samp>" or "</samp>.
 */
	public ArrayToString
	(
		final Object			A,
		final String			FinalSeparatorStr
	)
	{
		Objs        = A;
		FinalSepStr = FinalSeparatorStr;
	}
	
/**	Constructs a new <code>ArrayToString</code> object that will create a <code>String</code> representation of the 
 *	supplied array or <code>Collection</code> object when this object's <code>toString</code> method is invoked, and not
 *	before then.
 * 
 * @param A						An array or {@link Collection} object whose contents should be converted to a 
 *								<code>String</code>. At this time, arrays or <code>Collection</code> objects whose elements
 *								are other arrays or <code>Collection</code> objects are not supported in a useful manner.
 * @param SeparatorStr			The string that should be placed between elements in the array or <code>Collection</code>,
 *								for example: a comma with trailing space (<samp>", "</samp>).
 * @param FinalSeparatorStr		The string that should be placed between the final two elements in the array or 
 *								<code>Collection</code>, for example: <samp>" and "</samp> or <samp>" or "</samp>.
 * @param QuoteStr				The string to be placed around items in the array or <code>Collection</code> that would
 *								benefit from being quoted. An item is judged to benefit from being quoted if it contains
 *								any whitespace, or a comma. To disable all quoting behavior, pass <code>null</code> in
 *								this parameter. To provide separate open- and close-quote strings, use {@link #setQuotes}.
 */
	public ArrayToString
	(
		final Object			A,
		final String			SeparatorStr,
		final String			FinalSeparatorStr,
		final String			QuoteStr	//	Can be null, in which case no quoting is performed.
	)
	{
		this.Objs          = A;
		this.SepStr        = SeparatorStr;
		this.FinalSepStr   = FinalSeparatorStr;
		this.OpenQuoteStr  = QuoteStr;
		this.CloseQuoteStr = QuoteStr;
	}
	
/**	Sets the separator string to be placed between all elements except the final two (see {@link #setFinalSeparator setFinalSeparator}).	
 * 
 * @param SeparatorStr			The string that should be placed between elements in the array or <code>Collection</code>,
 *								for example: a comma with trailing space (<samp>", "</samp>).
 * @return						A reference to this instance of the <code>ArrayToString</code> class.
 */	
	public ArrayToString setSeparator
	(
		final String			SeparatorStr
	)
	{
		SepStr  = SeparatorStr;
		ObjsStr = null;		//	Cause the string to be generated afresh the next time toString is invoked.
		
		return this;
	}
	
/**	Sets the string that should be placed between the final two elements in the array or <code>Collection</code>.
 * 
 * @param FinalSeparatorStr		The string that should be placed between the final two elements in the array or 
 *								<code>Collection</code>, for example: <samp>" and "</samp> or <samp>" or "</samp>.
 * @return						A reference to this instance of the <code>ArrayToString</code> class.
 */
	public ArrayToString setFinalSeparator
	(
		final String			FinalSeparatorStr
	)
	{
		FinalSepStr = FinalSeparatorStr;
		ObjsStr     = null;		//	Cause the string to be generated afresh the next time toString is invoked.
		
		return this;
	}

/**	Sets the string that should be placed around elements that would benefit from being quoted. An item is judged to 
 *	benefit from being quoted if it contains any whitespace, or a comma. To disable all quoting behavior, pass 
 *	<code>null</code>, or invoke {@link #neverQuote}. To force quoting under all circumstances, invoke 
 *	{@link #alwaysQuote}. To provide separate open- and close-quote strings, use {@link #setQuotes setQuotes}.
 * 
 * @param QuoteStr				The string to be placed around items in the array or <code>Collection</code> that would
 *								benefit from being quoted. To disable all quoting behavior, pass <code>null</code> in
 *								this parameter.
 * @return						A reference to this instance of the <code>ArrayToString</code> class.
 */	
	public ArrayToString setQuote
	(
		final String			QuoteStr	//	Can be null, in which case no quoting is performed.
	)
	{
		OpenQuoteStr  = QuoteStr;
		CloseQuoteStr = QuoteStr;
		ObjsStr       = null;		//	Cause the string to be generated afresh the next time toString is invoked.
		
		return this;
	}
	
/**	Sets the strings that should be placed to the left and right of elements that would benefit from being quoted. An 
 *	item is judged to benefit from being quoted if it contains any whitespace, or a comma. To disable all quoting 
 *	behavior, pass <code>null</code> in both parameters, or invoke {@link #neverQuote}. To force quoting under all
 *	circumstances, invoke {@link #alwaysQuote}.
 * 
 * @param OpenQuote				The string to be placed to the left of items in the array or <code>Collection</code> 
 *								that would benefit from being quoted.
 * @param CloseQuote			The string to be placed to the right of items in the array or <code>Collection</code> 
 *								that would benefit from being quoted.
 * @return						A reference to this instance of the <code>ArrayToString</code> class.
 */	
	public ArrayToString setQuotes
	(
		final String			OpenQuote,
		final String			CloseQuote
	)
	{
		OpenQuoteStr  = OpenQuote;
		CloseQuoteStr = CloseQuote;
		ObjsStr       = null;		//	Cause the string to be generated afresh the next time toString is invoked.
		
		return this;		
	}
	
/**	Sets the characters that should be placed to the left and right of elements that would benefit from being quoted. 
 *	An item is judged to benefit from being quoted if it contains any whitespace, or a comma. To disable all quoting 
 *	behavior, or invoke {@link #neverQuote}. To force quoting under all circumstances, invoke {@link #alwaysQuote}.
 * 
 * @param OpenQuote				The character to be placed to the left of items in the array or <code>Collection</code> 
 *								that would benefit from being quoted.
 * @param CloseQuote			The character to be placed to the right of items in the array or <code>Collection</code> 
 *								that would benefit from being quoted.
 * @return						A reference to this instance of the <code>ArrayToString</code> class.
 */	
	public ArrayToString setQuotes
	(
		final char				OpenQuote,
		final char				CloseQuote
	)
	{
		return setQuotes(Character.toString(OpenQuote), Character.toString(CloseQuote));
	}
	
/**	Causes all elements to be quoted, whether or not this class would ordinarily judge quoting to be necessary.
 * 
 * @return						A reference to this instance of the <code>ArrayToString</code> class.
 */
	public ArrayToString alwaysQuote() {
		
		AlwaysUseQuotes = true;
		
		if (OpenQuoteStr == null)
			OpenQuoteStr = "\"";
		if (CloseQuoteStr == null)
			CloseQuoteStr = "\"";
		
		ObjsStr = null;		//	Cause the string to be generated afresh the next time toString is invoked.
		
		return this;
	}
	
/**	Prevents all elements from being quoted, whether or not this class would ordinarily judge quoting to be necessary.
 * 
 * @return						A reference to this instance of the <code>ArrayToString</code> class.
 */
	public ArrayToString neverQuote() {
		
		OpenQuoteStr    = null;
		CloseQuoteStr   = null;
		AlwaysUseQuotes = false;
		ObjsStr         = null;		//	Cause the string to be generated afresh the next time toString is invoked.
		
		return this;
	}	
	
/**	Causes the whitespace at the beginning or end of each element in the array or <code>Collection</code> to be 
 * 	trimmed away. By default, trimming does not occur.
 * 
 * @return						A reference to this instance of the <code>ArrayToString</code> class.
 */	
	public ArrayToString trim() {
		
		if (!TrimStrings) {
			
			TrimStrings = true;
			ObjsStr     = null;		//	Cause the string to be generated afresh the next time toString is invoked.
		}
		
		return this;
	}
	
/**	The number of elements in the array or <code>Collection</code> that was converted to a <code>String</code> by this 
 *	object.	Note that this information isn't available until this object's <code>toString</code> method has been 
 *	called once, so <code>length</code> calls <code>toString</code> automatically, if it hasn't already been called.
 *
 * @return						The number of elements in the array or <code>Collection</code> that was passed to this
 *								object's constructor.
 */	
	public int length() {
		
		if (ObjsStr == null)
			toString();
		
		return Elements;
	}
	
/**	Returns a <code>String</code> representation of the array or <code>Collection</code> object passed to the 
 *	constructor. The <code>String</code> representation is not created until this method is called for the first time,
 *	and is thereafter cached, so that all future <code>toString</code> invocations will return the same 
 *	<code>String</code> object as the first invocation, and will do so essentially instantly. Note, however, that if
 *	any of the methods that control the way the string is generated are invoked (for example {@link #setQuotes setQuotes} 
 * 	or {@link #trim}), they will discard the cached <code>String</code> object to ensure that the next time this 
 *	<code>toString</code> method is invoked, the <code>String</code> that is returned will be formatted in a manner 
 *	consistent with the caller's expressed intent.
 * 
 * @return 		A <code>String</code> representation of the array or <code>Collection</code> object passed to the 
 *				constructor.
 */	
	@Override public String toString() {
		
		if (ObjsStr == null) {
			final int				Count;
			final char[][]			Chars;
			int						CharCount = 0;	//	This is only meant to be an estimate of the final length. Underestimating, isn't good, but modest overestimating is no problem.
			boolean					StringsShouldBeQuoted = false;
			
			if (Objs == null) {
				
				Count                 = 1;
				Chars                 = kNullCharsArray;	//	Do not modify anything in this array; it's static, so changes will persist.
				CharCount             = 4;
				StringsShouldBeQuoted = false;
				
			} else if (Objs instanceof Collection) {
			
			//	Handle collections of arbitrary objects.
			
				final Collection<?>		C = (Collection<?>) Objs;
				
				Count = C.size();
				Chars = new char[Count][];
				
				final Iterator<?>		Elems = C.iterator();
				int						Index = 0;
				
				while (Elems.hasNext()) {
					final Object			Elem   = Elems.next();
					final char[]			AChars = Elem == null ? kNullChars : trim(Elem.toString()).toCharArray();
					
					Chars[Index] = AChars;
					CharCount += AChars.length;
					
					if (Index > 0)
						CharCount += SepStr.length();
					
					if (StringsShouldBeQuoted == false)
						StringsShouldBeQuoted = shouldBeQuoted(AChars);
					
					Index++;
				}
				
			} else if (Objs.getClass().isArray()) {
				
			//	Handle arrays of primitive types, or arbitrary objects.
				
				Count = Array.getLength(Objs);
				Chars = new char[Count][];
				
				for (int Index = 0; Index < Count; Index++) {
					final Object			Elem   = Array.get(Objs, Index);
					final char[]			AChars = Elem == null ? kNullChars : trim(Elem.toString()).toCharArray();
					
					Chars[Index] = AChars;
					CharCount += AChars.length;
					
					if (Index > 0)
						CharCount += SepStr.length();
					
					if (StringsShouldBeQuoted == false)
						StringsShouldBeQuoted = shouldBeQuoted(AChars);
				}
				
			} else {
			
			//	Handle an arbitrary object.
			
				Count                 = 1;
				Chars                 = new char[][] { trim(Objs.toString()).toCharArray() };
				CharCount             = Chars[0].length;
				StringsShouldBeQuoted = shouldBeQuoted(Chars[0]);
			}
			
		//	Update our estimate of the final string's length based on what we knoe about the number of elements we're 
		//	dealing with, the separators in use, the open- and close-quotes, and whether we're adding quotes at all.
		//	Overestimating the length isn't a problem, but underestimating will cause the StringBuffer used below to 
		//	reallocate and copy it's internal character array at least once, and that would be an unfortunate waste of
		//	time and memory. 
			
			if (StringsShouldBeQuoted)
				CharCount += Count * (OpenQuoteStr.length() + CloseQuoteStr.length());
			if (Count > 1)
				CharCount += FinalSepStr.length();		//	Characters of the string " and ".
			
		//	Create the final string.
			
			final StringBuilder		Buff = new StringBuilder(CharCount);
			
			for (int Index = 0, Limit = Count - 1; Index <= Limit; Index++) {
				
				if (Index > 0)
					Buff.append(Index != Limit ? SepStr : FinalSepStr);
				
				final char[]			ElemChars = Chars[Index];
				
				if (StringsShouldBeQuoted && !alreadyQuoted(ElemChars))
					Buff.append(OpenQuoteStr).append(ElemChars).append(CloseQuoteStr);
				else
					Buff.append(ElemChars);
			}
			
		//	Retain the final string, and the number of elements from which it was composed.
			
			ObjsStr  = Buff.toString();
			Elements = Count;
		}
		
		return ObjsStr;
	}
	
	private boolean shouldBeQuoted
	(
		final char[]			Chars
	)
	{
		if (OpenQuoteStr == null && CloseQuoteStr == null)
			return false;
		if (AlwaysUseQuotes)
			return true;
		
	//	If it's already quoted in the same way we'd quote it, do not add any quotes.
		
		if (alreadyQuoted(Chars))
			return false;
		
	//	If the items contains commas or whitespace, quote it. It might be useful to quote items containing quote 
	//	characters, but then we'd also have to come-up with a quote escaping scheme, and one that was human-readable.
		
		for (int Index = 0, Count = Chars.length; Index < Count; Index++) {
			final char				Char = Chars[Index];
			
			if (Char == ',' || Character.isWhitespace(Char))
				return true;
		}
		
		return false;
	}
	
	private boolean alreadyQuoted
	(
		final char[]			Chars
	)
	{
	//	If it's already quoted in the same way we'd quote it, do not add any quotes.
		
		return startsWith(Chars, OpenQuoteStr) && endsWith(Chars, CloseQuoteStr);
	}
	
	private static boolean startsWith
	(
		final char[]			Chars,
		final String			PrefixStr
	)
	{
		final int				CLen = Chars.length;
		final int				PLen = PrefixStr.length();
		
		if (PLen > CLen)
			return false;
		
		for (int Index = 0; Index < PLen; Index++) {
			
			if (Chars[Index] != PrefixStr.charAt(Index))
				return false;
		}
		
		return true;
	}
	
	private static boolean endsWith
	(
		final char[]			Chars,
		final String			PostfixStr
	)
	{
		final int				CLen = Chars.length;
		final int				PLen = PostfixStr.length();
		
		if (PLen > CLen)
			return false;
		
		for (int CIndex = CLen - 1, EIndex = PLen - 1; EIndex >= 0; CIndex--, EIndex--) {
			
			if (Chars[CIndex] != PostfixStr.charAt(EIndex))
				return false;
		}
		
		return true;
	}
	
	private String trim
	(
		final String			StringToTrim
	)
	{
		return TrimStrings ? StringToTrim.trim() : StringToTrim;
	}
}
