/*	QWICAP (Quick Web Interface for Conventional Applications) by Chris W. Johnson
	Copyright (C) 2003-2010 The University of Texas at Austin ("U.T. Austin"). All rights reserved.
	
	U.T. Austin's copyright notice in this evolving version reflects our original authorship. It does not assert 
	copyright over changes contributed by others.
	
	This library is free software; you can redistribute it and/or modify it under the terms of the GNU Lesser General 
	Public License as published by the Free Software Foundation; either version 2.1 of the License, or (at your option) 
	any later version.
	
	This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied 
	warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for 
	more details.
	
	You should have received a copy of the GNU Lesser General Public License along with this library; if not, write to 
	the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/


//  Characters.java
//  qwicap
//
//  Created by Chris W. Johnson on Fri Jul 11 2003.


package edu.utexas.its.eis.tools.qwicap.util;


import java.io.*;


/**	<p>The <code>Characters</code> class transparently represents either a <code>String</code>, or a subsection of a 
 * 	character array. Conversions from <code>String</code> to character array, and vice versa, are avoided wherever 
 * 	possible, and if a conversion must be performed to satisfy a caller, the conversion is performed only once, and the 
 * 	result is cached. This means that modifications to a character array passed to the constructor, or returned by the 
 * 	{@link #getArray} method, have the potential to invalidate the cache. So do NOT modify those arrays.
 * 	</p>
 * 	
 * 	<p>Convenience methods such as write, println and append are provided so that those common operations can be 
 * 	performed without the caller having to be aware of the presence/absence of <code>char[]</code> or <code>String</code> 
 * 	forms in any instance of this class, and in order to avoid conversions between those forms unless absolutely 
 * 	necessary.
 *	</p>
 * 
 *	@author Chris W. Johnson
 */
public final class Characters implements CharSequence {

	private String				S;
	private char[]				C;
	private final int			COffset;
	private final int			CLength;
	private boolean				HashCodeComputed;
	private int					HashCode;
	
/**	Creates a <code>Characters</code> object that represents a <code>String</code>.
 *
 * @param S		The <code>String</code> to be represented by this <code>Characters</code> object.
 */
	public Characters
	(
		final String			S
	)
	{
		this.S       = S;
		this.COffset = 0;
		this.CLength = S.length();
	}

/**	Creates a <code>Characters</code> object that represents the <code>String</code> supplied by the specified 
 *	<code>StringBuffer</code> object.
 *
 * @param SB	The <code>StringBuffer</code> object containing the String to be represented by this 
 *				<code>Characters</code> object.
 */
	public Characters
	(
		final StringBuffer		SB
	)
	{
		this(SB.toString());
	}

/**	Creates a <code>Characters</code> object that represents the <code>String</code> supplied by the specified 
 *	<code>StringBuffer</code> object.
 *
 * @param SB	The <code>StringBuilder</code> object containing the String to be represented by this 
 *				<code>Characters</code> object.
 */
	public Characters
	(
		final StringBuilder		SB
	)
	{
		this(SB.toString());
	}

/**	Creates a <code>Characters</code> object that represents an entire char array.
 *
 * @param C		The char array to be represented by this <code>Characters</code> object.
 */
	public Characters
	(
		final char[]			C
	)
	{
		this(C, 0, C.length);
	}

/**	Creates a <code>Characters</code> object that represents a subsection of a char array.
 *
 * @param C			The char array to be represented by this <code>Characters</code> object.
 * @param COffset	Index of the first character in the C array to be represented by this object.
 * @param CLength	Number of characters, beginning at COffset, to be represented by this object.
 */
	public Characters
	(
		final char[]			C,
		final int				COffset,
		final int				CLength
	)
	{
		this.C       = C;
		this.COffset = COffset;
		this.CLength = CLength;
	}

/**	Returns the length of this character sequence. The length is the number of 16-bit chars in the sequence.
 * 
 * @return						The length of this character sequence. The length is the number of 16-bit chars in the 
 * 								sequence.
 */
	@Override public int length() {
		return CLength;
	}

/**	Returns the charactes of this object represented as an array of characters. If only a <code>String</code> 
 *	representation of the characters exists, a char array is created and cached for future use.
 *
 * 	<p>For efficiency, the array returned is not a clone or
 * 	copy of the character array used within this object, it is that actual array. Consequently, alterations to the
 * 	contents of the array are possible, but <strong>strongly discouraged</strong>. Among other things, alterations to
 * 	the array will leave the array out of synch with the <code>String</code> representation of these same characters 
 *	that may also exist. If so, callers treating this <code>Characters</code> object as a source for an array of chars 
 *	will get different results from those treating this <code>Characters</code> object as a source for a 
 *	<code>String</code>. All kinds of bad things might come from this. If this class is needed in a context where 
 *	programmers can't be depended upon to respect this limitation, a subclass should be created which overrides this 
 *	method to return a copy of the array (and overrides <code>getArrayOffset</code> to always return 0).
 *
 * @return The array containing the characters in this object. The contents of this array should not be altered.
 */
	public char[] getArray() {
		return makeCharArray();
	}

/**	Returms the offset of the first character represented by this object in the char array returned by the 
 *	<code>getArray</code> method.
 *
 * @return	Offset of the first character represented by this object in the char array returned by <code>getArray</code>.
 */
	public int getArrayOffset() {
		
	//	We do NOT have to call makeCharArray here to be safe.
	//
	//	(1) If our constructor was fed a String, COffset was left
	//	at its default value of zero, which will be correct if the
	//	String is ever converted to char[] in order to satisfy
	//	getArray(), or equals() under some conditions.
	//	
	//	(2) If our constructor was fed a char[], it goes without
	//	saying that our value of COffset is valid.
	
		return COffset;
	}
	
/**	Compares this object to another object. Useful comparisons may be performed only if the other object is a 
 *	<code>String</code> or another <code>Characters</code> object. In all other cases, this method returns false.
 * 
 * @param OtherObj				Another object to compare with this one.
 * @return						<code>true</code> if this object represents a sequence of characters equivalent to the
 *								supplied object, <code>false</code> otherwise.
 */
	@Override public boolean equals
	(
		final Object			OtherObj
	)
	{
		if (this == OtherObj)
			return true;
		
		if (OtherObj instanceof String) {
			
			makeString();
			return S.equals(OtherObj);
			
		} else if (OtherObj instanceof Characters) {
			final Characters		Other    = (Characters) OtherObj;
			boolean					AreEqual = false;
			
			if (C != null && Other.C != null) {
				
				if (CLength == Other.CLength) {
					final char[]			C      = this.C;	//	Local copy for performance.
					final char[]			OtherC = Other.C;	//	Local copy for performance.
					
					AreEqual = true;
					
					for (int ThisIndex = COffset, OtherIndex = Other.COffset, Count = CLength; --Count >= 0; ThisIndex++, OtherIndex++) {
						
						if (C[ThisIndex] != OtherC[OtherIndex]) {
							
							AreEqual = false;
							break;
						}
					}
				}
				
			} else if (S != null && Other.S != null) {
				
				AreEqual = S.equals(Other.S);
				
			} else {
				
				if (C == null)
					AreEqual = equals(Other.C, Other.COffset, Other.CLength);
				else
					AreEqual = Other.equals(C, COffset, CLength);
			}
			
			return AreEqual;
		}
		
	//	An obvious extension at this point would be to call OtherObj.toString() and test for equality with that 
	//	String. Not sure if that would actually be useful in practice, however.
		
		return false;
	}

/**	Compares the characters in this object to the specified subsection of a char array. The comparison is
 * 	case-sensitive.
 *
 * @param OtherChars	The array of characters to be compared to the characters in this object.
 * @param OtherOffset	The index into <code>OtherChars</code> at which comparison is to begin.
 * @param OtherLength	The number of characters in <code>OtherChars</code>, starting at <code>OtherOffset</code>, to 
 * 						be compared.
 * @return				<code>true</code> if the characters match, <code>false</code> otherwise.
 */
	public boolean equals
	(
		final char[]			OtherChars,
		final int				OtherOffset,
		final int				OtherLength
	)
	{
		boolean					AreEqual = false;
		
		makeCharArray();
		
		if (CLength == OtherLength) {
			final char[]			C = this.C;		//	Local copy for performance.
			
			AreEqual = true;
			
			for (int ThisIndex = COffset, OtherIndex = OtherOffset, Count = CLength; --Count >= 0; ThisIndex++, OtherIndex++) {
				
				if (C[ThisIndex] != OtherChars[OtherIndex]) {
					
					AreEqual = false;
					break;
				}
			}
		}
		
		return AreEqual;
	}
	
/**	Return the hash code for these characters. For the any given set of characters, the hash code is identical 
 *	whether the characters were supplied as a <code>String</code> object, or as an array of characters.
 * 
 * @return				The hash code for the characters represented by this object. 
 */
	@Override public int hashCode() {
		
		if (!HashCodeComputed) {
			
			HashCode = S != null ? S.hashCode() : hashCode(C, COffset, CLength);
			HashCodeComputed = true;
		}
		
		return HashCode;
	}
		
/**	Compute a hash code for an arbitrary subsection of an array of characters using the same algorithm employed by the 
 *	hashCode method of the <code>java.lang.String</code> class.
 * 
 * @param Chars			The array of characters.
 * @param Offset		The offset of the first character in the subsection for which the hash code is to be computed.
 * @param Count			The number of characters in the subsection for which the hash code is to be computed.
 * @return				The hash code for the array of characters.
 */
	private static int hashCode
	(
		final char[]			Chars,
		final int				Offset,
		final int				Count
	)
	{
	/*	The hash code for a String object is computed as
		
			s[0]*31^(n-1) + s[1]*31^(n-2) + ... + s[n-1]
	 
		using int arithmetic, where s[i] is the  ith character of the string, 
		n is the length of  the string, and ^ indicates exponentiation. (The 
		hash value of the empty string is zero.)
	*/	
		if (Count == 0)
			return 0;
		
		int						Hash = 0;
		
		for (int Index = Offset, Limit = Offset + Count; Index < Limit; Index++)
			Hash = 31 * Hash + Chars[Index];
		
		return Hash;
	}
	
/**	Compares the initial characters of this object with all the characters of another <code>Characters</code> object. 
 *	The comparison is case-sensitive.
 *
 * @param Other	The <code>Characters</code> object to be compared with the start of this one.
 * @return		<code>true</code> if this object begins with the characters of the Other object, <code>false</code> 
 * 				otherwise.
 */
	public boolean startsWith
	(
		final Characters		Other
	)
	{
		boolean					StartsWith = false;
		
		if (C != null && Other.C != null) {
			
			if (CLength >= Other.CLength) {
				final char[]			C = this.C;			//	Local copy for performance.
				final char[]			OtherC = Other.C;	//	Local copy for performance.
				
				StartsWith = true;
				
				for (int ThisIndex = COffset, OtherIndex = Other.COffset, Count = Other.CLength; --Count >= 0; ThisIndex++, OtherIndex++) {
					
					if (C[ThisIndex] != OtherC[OtherIndex]) {
						
						StartsWith = false;
						break;
					}
				}
			}
			
		} else if (S != null && Other.S != null) {
			
			StartsWith = S.startsWith(Other.S);
			
		} else {
			
			if (C == null)
				makeCharArray();
			else
				Other.makeCharArray();
			
			StartsWith = startsWith(Other);
		}
		
		return StartsWith;
	}

/**	Retrieves a substring of this object's characters, beginning with the character at the specified offset, and 
 * 	including all of the characters that follow it.
 * 
 * @param FirstCharOffset		Offset of the first character, inclusive.
 * @return						The requested substring.
 */	
	public String substring
	(
		final int				FirstCharOffset
	)
	{
		return substring(FirstCharOffset, CLength);
	}

/**	Retrieves a substring of this object's characters.
 * 
 * @param FirstCharOffset		Offset of the first character, inclusive.
 * @param LastCharOffset		Offset of the final character, exclusive.
 * @return						The requested substring.
 */
	public String substring
	(
		final int				FirstCharOffset,
		final int				LastCharOffset
	)
	{
		if (S != null)
			return S.substring(FirstCharOffset, LastCharOffset);
		
		final int				Length;
		
		if (LastCharOffset > CLength)
			Length = CLength - FirstCharOffset;
		else
			Length = LastCharOffset - FirstCharOffset;
		
		final StringBuilder		Buff = new StringBuilder(Length);
		
		Buff.append(C, COffset + FirstCharOffset, Length);
		
		return Buff.toString();
	}
	
/**	Returns a <code>Characters</code> instance representing a subsequence of this object's characters.
 * 
 * @param FirstCharOffset		Offset of the first character, inclusive.
 * @param LastCharOffset		Offset of the final character, exclusive.
 * @return						The requested subsequence, represented by an instance of <code>Characters</code>.
 */
	@Override public CharSequence subSequence
	(
		int						FirstCharOffset,
		int						LastCharOffset
	)
	{
		if (FirstCharOffset < 0)
			FirstCharOffset = 0;
		else if (FirstCharOffset >= CLength)
			return new Characters(C, COffset + CLength, 0);		//	Return an empty sequence.
		
		if (LastCharOffset <= 0)
			return new Characters(C, COffset, 0);				//	Return an empty sequence.
		else if (LastCharOffset > CLength)
			LastCharOffset = CLength;
		
		return new Characters(C, COffset + FirstCharOffset, LastCharOffset - FirstCharOffset);
	}

/**	If necessary, creates a <code>String</code> representation of this object's characters.
 *
 * @return	A <code>String</code> representation of this object's characters.
 */
	private String makeString() {
		
		if (S == null)
			S = new String(C, COffset, CLength);
		
		return S;
	}

/**	If necessary, create char array represention of this object's characters.
 *
 * @return	A char array representing this object's characters. As usual, the contents should not be modified.
 */
	private char[] makeCharArray() {
		
		if (C == null)
			C = S.toCharArray();
		
		return C;
	}

/**	Returns the <code>char</code> value at the specified index. An index ranges from zero to <code>length() - 1</code>. 
 * 	The first <code>char</code> value of the sequence is at index zero, the next at index one, and so on, as for array 
 * 	indexing.
 * 
 *	<p>If the char value specified by the index is a surrogate, the surrogate value is returned.
 * 	</p>
 *
 * @param Index							The index of the character to be returned.
 * @return								The <code>char</code> at the specified index.
 * @throws IndexOutOfBoundsException	If the index is invalid.
 */
	@Override public char charAt
	(
		final int				Index
	)
	throws IndexOutOfBoundsException
	{
		if (Index < 0 || Index >= CLength)
			throw new IndexOutOfBoundsException(Index + " not in valid range 0-" + (CLength - 1));

		if (C != null)
			return C[Index + COffset];
		else
			return S.charAt(Index);
	}

/**	Returns the index of the first occurrence of the specified character, or -1 if the character was not found.
 *
 * @param Char	Character to be searched for.
 * @return		Index of the first occurrence of Char, or -1 if it was not found.
 */
	public int indexOf
	(
		final int				Char
	)
	{
		return indexOf(Char, 0);
	}
	
/**	Returns the index of the first occurrence of the specified character, or -1 if the character was not found.
 *
 * @param Char					Character to be searched for.
 * @param Offset				The offset at which to begin the search.
 * @return						Index of the first occurrence of <code>Char</code>, or -1 if it was not found.
 */
	public int indexOf
	(
		final int				Char,
		final int				Offset
	)
	{
		if (S != null)
			return S.indexOf(Char, Offset);
		
		final char[]			C = this.C;
		final int				COffset = this.COffset;
		
		for (int Index = COffset + Offset, Limit = COffset + CLength; Index < Limit; Index++) {
			
			if (C[Index] == Char)
				return Index - COffset;
		}
		
		return -1;
	}
	
/**	Returns the <code>String</code> representation of the characters represented by this <code>Characters</code> object. 
 *	This method always returns the same <code>String</code> object for efficiency. If only a character array 
 *	representation of the characters exists, a new <code>String</code> object is created and cached for future use.
 *
 * @return	The characters of this object represented as a <code>String</code>.
 */
	@Override public String toString() {
		return makeString();
	}

/**	Returns <code>true</code> if this object already contains a <code>String</code> representation of its characters, or
 * 	false if such a representation has not yet been created.
 *
 * @return	<code>true</code> if a <code>String</code> representation of this object has already exists, 
 * 			<code>false</code> otherwise.
 */
	public boolean hasString() {
		return S != null;
	}

/**	Returns <code>true</code> if this object already contains a char array representation of its characters,
 * 	or <code>false</code> if such a representation has not yet been created.
 *
 * @return	<code>true</code> if a char array representation of this object has already exists, <code>false</code> 
 * 			otherwise.
 */
	public boolean hasArray() {
		return C != null;
	}

/**	Creates a new <code>Characters</code> object representing the concatenation of first this object, and then the 
 *	specified <code>Characters</code> object.
 *
 * @param Other		The characters to be appended to this object.
 * @return			A new <code>Characters</code> object representing this + <code>Other</code>.
 */
	public Characters append
	(
		final Characters		Other
	)
	{
		if (C != null && Other.C != null) {

		//	Here we check for the case of adjacent subsections of the same array. In this unlikely case,
		//	we avoid creating a new char array.

			if (C == Other.C && COffset + CLength == Other.COffset) {

				return new Characters(C, COffset, CLength + Other.CLength);

			}else {
				final char[]			Combined = new char[CLength + Other.CLength];

				System.arraycopy(C, COffset, Combined, 0, CLength);
				System.arraycopy(Other.C, Other.COffset, Combined, CLength, Other.CLength);

				return new Characters(Combined);
			}

		} else if (S != null && Other.S != null) {
			
			return new Characters(S + Other.S);
			
		} else {
			final char[]			Combined = new char[CLength + Other.CLength];
			
			if (C == null)
				S.getChars(0, CLength, Combined, 0);
			else
				System.arraycopy(C, COffset, Combined, 0, CLength);
			
			if (Other.C == null)
				Other.S.getChars(0, Other.CLength, Combined, CLength);
			else
				System.arraycopy(Other.C, Other.COffset, Combined, CLength, Other.CLength);
			
			return new Characters(Combined);
		}
	}

/**	Appends the characters represented by this object to the specified <code>StringBuffer</code>.
 *
 * @param Buff		The <code>StringBuffer</code> to which this object's characters should be appended.
 */
	public void append
	(
		final StringBuffer		Buff
	)
	{
		if (C != null)
			Buff.append(C, COffset, CLength);
		else
			Buff.append(S);
	}

/**	Appends the characters represented by this object to the specified <code>StringBuilder</code>.
 *
 * @param Buff		The <code>StringBuilder</code> to which this object's characters should be appended.
 */
	public void append
	(
		final StringBuilder		Buff
	)
	{
		if (C != null)
			Buff.append(C, COffset, CLength);
		else
			Buff.append(S);
	}
	
/**	Prints the characters of this object to the specified <code>PrintStream</code>.
 *
 * @param Out			<code>PrintStream</code> to which this object's characters should be printed.
 */
	public void print
	(
		final PrintStream		Out
	) 
	{
		if (S != null) {
			
			Out.print(S);
		
		} else {
			
			if (COffset == 0 && CLength == C.length)
				Out.print(C);
			else
				Out.print(makeString());
		}
	}

/**	Uses the specified <code>PrintStream</code> object's <code>println</code> method to print the characters of this 
 * 	object.
 *
 * @param Out			<code>PrintStream</code> to which this object's characters should be printed.
 */
	public void println
	(
		final PrintStream		Out
	) 
	{
		if (S != null) {
			
			Out.println(S);
		
		} else {
			
			if (COffset == 0 && CLength == C.length)
				Out.println(C);
			else
				Out.println(makeString());
		}
	}
	
/**	Prints the characters of this object to the specified <code>PrintWriter</code>.
 *
 * @param Out			<code>PrintWriter</code> to which this object's characters should be printed.
 * @throws IOException	If thers's a problem printing to the <code>PrintWriter</code>.
 */
	public void print
	(
		final PrintWriter		Out
	)
	throws IOException
	{
		write(Out);
	}

/**	Uses the specified <code>PrintWriter</code> object's <code>println</code> method to print the characters of this 
 * 	object.
 *
 * @param Out			<code>PrintWriter</code> to which this object's characters should be printed.
 * @throws IOException	If thers's a problem printing to the <code>PrintWriter</code>.
 */
	public void println
	(
		final PrintWriter		Out
	)
	throws IOException
	{
		if (S != null) {
			
			Out.println(S);
		
		} else {
			
			if (COffset == 0 && CLength == C.length) {
				
				Out.println(C);
				
			} else {
				
				write(Out);
				Out.println();
			}
		}
	}
	
/**	Writes the characters of this object to the specified <code>Writer</code>.
 *
 * @param Out			<code>Writer</code> to which this object's characters should be written.
 * @throws IOException	If thers's a problem writing to the <code>Writer</code>.
 */
	public void write
	(
		final Writer			Out
	)
	throws IOException
	{
		if (C != null)
			Out.write(C, COffset, CLength);
		else
			Out.write(S);
	}
}







