/*	QWICAP (Quick Web Interface for Conventional Applications) by Chris W. Johnson
	Copyright (C) 2003-2011 The University of Texas at Austin ("U.T. Austin"). All rights reserved.
	
	U.T. Austin's copyright notice in this evolving version reflects our original authorship. It does not assert 
	copyright over changes contributed by others.
	
	This library is free software; you can redistribute it and/or modify it under the terms of the GNU Lesser General 
	Public License as published by the Free Software Foundation; either version 2.1 of the License, or (at your option) 
	any later version.
	
	This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied 
	warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for 
	more details.
	
	You should have received a copy of the GNU Lesser General Public License along with this library; if not, write to 
	the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/


package edu.utexas.its.eis.tools.qwicap.util;


import java.util.HashMap;


/**	<code>HTMLEntityCodec</code> converts most control characters, various reserved characters (ampersands, 
 *	double-quotes, chevrons, etc.), and all Unicode characters greater than or equal to 127, to and from HTML 
 *	character entity references. When encoding, the symbolic form of HTML character entities (e.g. "&amp;middot;") are 
 *	always used in preference to the numeric forms (e.g. "&amp;#183;"). All character entity references specified in 
 *	the HTML 4 specification are supported. 
 *
 *	@see <a href="http://www.w3.org/TR/html4/sgml/entities.html">Character entity references in HTML 4</a>
 *	@author Chris W. Johnson
 * 	@version 1.3
 */
public final class HTMLEntityCodec {

//	todo Consider adding the entities defined by <http://www.w3.org/TR/2010/REC-xml-entity-names-20100401/>.

	
//	Programmatically extracted from "Character entity references in HTML 4", 
//	<http://www.w3.org/TR/html4/sgml/entities.html>.
	
	private static final HTMLEntity[] Mappings = {
		new HTMLEntity("&#0;",         0),		//	Unused character - we encode it to ensure a consistent, obvious representation.
		new HTMLEntity("&#1;",         1),		//	Unused character - we encode it to ensure a consistent, obvious representation.
		new HTMLEntity("&#2;",         2),		//	Unused character - we encode it to ensure a consistent, obvious representation.
		new HTMLEntity("&#3;",         3),		//	Unused character - we encode it to ensure a consistent, obvious representation.
		new HTMLEntity("&#4;",         4),		//	Unused character - we encode it to ensure a consistent, obvious representation.
		new HTMLEntity("&#5;",         5),		//	Unused character - we encode it to ensure a consistent, obvious representation.
		new HTMLEntity("&#6;",         6),		//	Unused character - we encode it to ensure a consistent, obvious representation.
		new HTMLEntity("&#7;",         7),		//	Unused character - we encode it to ensure a consistent, obvious representation.
		new HTMLEntity("&#8;",         8),		//	Unused character - we encode it to ensure a consistent, obvious representation.

	//	new HTMLEntity("&#9;",         9),		//	tab     - by commenting-out this line, tab characters pass-through unchanged.
	//	new HTMLEntity("&#10;",       10),		//	newline - by commenting-out this line, newline characters pass-through unchanged.

		new HTMLEntity("&#11;",       11),		//	Unused character - we encode it to ensure a consistent, obvious representation.
		new HTMLEntity("&#12;",       12),		//	Unused character - we encode it to ensure a consistent, obvious representation.

	//	new HTMLEntity("&#13;",       13),		//	carriage-return - by commenting-out this line, carriage-return characters pass-through unchanged.

		new HTMLEntity("&#14;",       14),		//	Unused character - we encode it to ensure a consistent, obvious representation.
		new HTMLEntity("&#15;",       15),		//	Unused character - we encode it to ensure a consistent, obvious representation.
		new HTMLEntity("&#16;",       16),		//	Unused character - we encode it to ensure a consistent, obvious representation.
		new HTMLEntity("&#17;",       17),		//	Unused character - we encode it to ensure a consistent, obvious representation.
		new HTMLEntity("&#18;",       18),		//	Unused character - we encode it to ensure a consistent, obvious representation.
		new HTMLEntity("&#19;",       19),		//	Unused character - we encode it to ensure a consistent, obvious representation.
		new HTMLEntity("&#20;",       20),		//	Unused character - we encode it to ensure a consistent, obvious representation.
		new HTMLEntity("&#21;",       21),		//	Unused character - we encode it to ensure a consistent, obvious representation.
		new HTMLEntity("&#22;",       22),		//	Unused character - we encode it to ensure a consistent, obvious representation.
		new HTMLEntity("&#23;",       23),		//	Unused character - we encode it to ensure a consistent, obvious representation.
		new HTMLEntity("&#24;",       24),		//	Unused character - we encode it to ensure a consistent, obvious representation.
		new HTMLEntity("&#25;",       25),		//	Unused character - we encode it to ensure a consistent, obvious representation.
		new HTMLEntity("&#26;",       26),		//	Unused character - we encode it to ensure a consistent, obvious representation.
		new HTMLEntity("&#27;",       27),		//	Unused character - we encode it to ensure a consistent, obvious representation.
		new HTMLEntity("&#28;",       28),		//	Unused character - we encode it to ensure a consistent, obvious representation.
		new HTMLEntity("&#29;",       29),		//	Unused character - we encode it to ensure a consistent, obvious representation.
		new HTMLEntity("&#30;",       30),		//	Unused character - we encode it to ensure a consistent, obvious representation.
		new HTMLEntity("&#31;",       31),		//	Unused character - we encode it to ensure a consistent, obvious representation.
		
		new HTMLEntity("&quot;",      34),		//	quotation mark = APL quote, U+0022 ISOnum
		new HTMLEntity("&amp;",       38),		//	ampersand, U+0026 ISOnum
		new HTMLEntity("&lt;",        60),		//	less-than sign, U+003C ISOnum
		new HTMLEntity("&gt;",        62),		//	greater-than sign, U+003E ISOnum

		new HTMLEntity("&#127;",     127),		//	Unused character - we encode it to ensure a consistent, obvious representation.
		new HTMLEntity("&#128;",     128),		//	Unused character - we encode it to ensure a consistent, obvious representation.
		new HTMLEntity("&#129;",     129),		//	Unused character - we encode it to ensure a consistent, obvious representation.
		new HTMLEntity("&#130;",     130),		//	Unused character - we encode it to ensure a consistent, obvious representation.
		new HTMLEntity("&#131;",     131),		//	Unused character - we encode it to ensure a consistent, obvious representation.
		new HTMLEntity("&#132;",     132),		//	Unused character - we encode it to ensure a consistent, obvious representation.
		new HTMLEntity("&#133;",     133),		//	Unused character - we encode it to ensure a consistent, obvious representation.
		new HTMLEntity("&#134;",     134),		//	Unused character - we encode it to ensure a consistent, obvious representation.
		new HTMLEntity("&#135;",     135),		//	Unused character - we encode it to ensure a consistent, obvious representation.
		new HTMLEntity("&#136;",     136),		//	Unused character - we encode it to ensure a consistent, obvious representation.
		new HTMLEntity("&#137;",     137),		//	Unused character - we encode it to ensure a consistent, obvious representation.
		new HTMLEntity("&#138;",     138),		//	Unused character - we encode it to ensure a consistent, obvious representation.
		new HTMLEntity("&#139;",     139),		//	Unused character - we encode it to ensure a consistent, obvious representation.
		new HTMLEntity("&#140;",     140),		//	Unused character - we encode it to ensure a consistent, obvious representation.
		new HTMLEntity("&#141;",     141),		//	Unused character - we encode it to ensure a consistent, obvious representation.
		new HTMLEntity("&#142;",     142),		//	Unused character - we encode it to ensure a consistent, obvious representation.
		new HTMLEntity("&#143;",     143),		//	Unused character - we encode it to ensure a consistent, obvious representation.
		new HTMLEntity("&#144;",     144),		//	Unused character - we encode it to ensure a consistent, obvious representation.
		new HTMLEntity("&#145;",     145),		//	Unused character - we encode it to ensure a consistent, obvious representation.
		new HTMLEntity("&#146;",     146),		//	Unused character - we encode it to ensure a consistent, obvious representation.
		new HTMLEntity("&#147;",     147),		//	Unused character - we encode it to ensure a consistent, obvious representation.
		new HTMLEntity("&#148;",     148),		//	Unused character - we encode it to ensure a consistent, obvious representation.
		new HTMLEntity("&#149;",     149),		//	Unused character - we encode it to ensure a consistent, obvious representation.
		new HTMLEntity("&#150;",     150),		//	Unused character - we encode it to ensure a consistent, obvious representation.
		new HTMLEntity("&#151;",     151),		//	Unused character - we encode it to ensure a consistent, obvious representation.
		new HTMLEntity("&#152;",     152),		//	Unused character - we encode it to ensure a consistent, obvious representation.
		new HTMLEntity("&#153;",     153),		//	Unused character - we encode it to ensure a consistent, obvious representation.
		new HTMLEntity("&#154;",     154),		//	Unused character - we encode it to ensure a consistent, obvious representation.
		new HTMLEntity("&#155;",     155),		//	Unused character - we encode it to ensure a consistent, obvious representation.
		new HTMLEntity("&#156;",     156),		//	Unused character - we encode it to ensure a consistent, obvious representation.
		new HTMLEntity("&#157;",     157),		//	Unused character - we encode it to ensure a consistent, obvious representation.
		new HTMLEntity("&#158;",     158),		//	Unused character - we encode it to ensure a consistent, obvious representation.
		new HTMLEntity("&#159;",     159),		//	Unused character - we encode it to ensure a consistent, obvious representation.

		new HTMLEntity("&nbsp;",     160),		//	no-break space = non-breaking space, U+00A0 ISOnum
		new HTMLEntity("&iexcl;",    161),		//	inverted exclamation mark, U+00A1 ISOnum
		new HTMLEntity("&cent;",     162),		//	cent sign, U+00A2 ISOnum
		new HTMLEntity("&pound;",    163),		//	pound sign, U+00A3 ISOnum
		new HTMLEntity("&curren;",   164),		//	currency sign, U+00A4 ISOnum
		new HTMLEntity("&yen;",      165),		//	yen sign = yuan sign, U+00A5 ISOnum
		new HTMLEntity("&brvbar;",   166),		//	broken bar = broken vertical bar, U+00A6 ISOnum
		new HTMLEntity("&sect;",     167),		//	section sign, U+00A7 ISOnum
		new HTMLEntity("&uml;",      168),		//	diaeresis = spacing diaeresis, U+00A8 ISOdia
		new HTMLEntity("&copy;",     169),		//	copyright sign, U+00A9 ISOnum
		new HTMLEntity("&ordf;",     170),		//	feminine ordinal indicator, U+00AA ISOnum
		new HTMLEntity("&laquo;",    171),		//	left-pointing double angle quotation mark = left pointing guillemet, U+00AB ISOnum
		new HTMLEntity("&not;",      172),		//	not sign, U+00AC ISOnum
		new HTMLEntity("&shy;",      173),		//	soft hyphen = discretionary hyphen, U+00AD ISOnum
		new HTMLEntity("&reg;",      174),		//	registered sign = registered trade mark sign, U+00AE ISOnum
		new HTMLEntity("&macr;",     175),		//	macron = spacing macron = overline = APL overbar, U+00AF ISOdia
		new HTMLEntity("&deg;",      176),		//	degree sign, U+00B0 ISOnum
		new HTMLEntity("&plusmn;",   177),		//	plus-minus sign = plus-or-minus sign, U+00B1 ISOnum
		new HTMLEntity("&sup2;",     178),		//	superscript two = superscript digit two = squared, U+00B2 ISOnum
		new HTMLEntity("&sup3;",     179),		//	superscript three = superscript digit three = cubed, U+00B3 ISOnum
		new HTMLEntity("&acute;",    180),		//	acute accent = spacing acute, U+00B4 ISOdia
		new HTMLEntity("&micro;",    181),		//	micro sign, U+00B5 ISOnum
		new HTMLEntity("&para;",     182),		//	pilcrow sign = paragraph sign, U+00B6 ISOnum
		new HTMLEntity("&middot;",   183),		//	middle dot = Georgian comma = Greek middle dot, U+00B7 ISOnum
		new HTMLEntity("&cedil;",    184),		//	cedilla = spacing cedilla, U+00B8 ISOdia
		new HTMLEntity("&sup1;",     185),		//	superscript one = superscript digit one, U+00B9 ISOnum
		new HTMLEntity("&ordm;",     186),		//	masculine ordinal indicator, U+00BA ISOnum
		new HTMLEntity("&raquo;",    187),		//	right-pointing double angle quotation mark = right pointing guillemet, U+00BB ISOnum
		new HTMLEntity("&frac14;",   188),		//	vulgar fraction one quarter = fraction one quarter, U+00BC ISOnum
		new HTMLEntity("&frac12;",   189),		//	vulgar fraction one half = fraction one half, U+00BD ISOnum
		new HTMLEntity("&frac34;",   190),		//	vulgar fraction three quarters = fraction three quarters, U+00BE ISOnum
		new HTMLEntity("&iquest;",   191),		//	inverted question mark = turned question mark, U+00BF ISOnum
		new HTMLEntity("&Agrave;",   192),		//	latin capital letter A with grave = latin capital letter A grave, U+00C0 ISOlat1
		new HTMLEntity("&Aacute;",   193),		//	latin capital letter A with acute, U+00C1 ISOlat1
		new HTMLEntity("&Acirc;",    194),		//	latin capital letter A with circumflex, U+00C2 ISOlat1
		new HTMLEntity("&Atilde;",   195),		//	latin capital letter A with tilde, U+00C3 ISOlat1
		new HTMLEntity("&Auml;",     196),		//	latin capital letter A with diaeresis, U+00C4 ISOlat1
		new HTMLEntity("&Aring;",    197),		//	latin capital letter A with ring above = latin capital letter A ring, U+00C5 ISOlat1
		new HTMLEntity("&AElig;",    198),		//	latin capital letter AE = latin capital ligature AE, U+00C6 ISOlat1
		new HTMLEntity("&Ccedil;",   199),		//	latin capital letter C with cedilla, U+00C7 ISOlat1
		new HTMLEntity("&Egrave;",   200),		//	latin capital letter E with grave, U+00C8 ISOlat1
		new HTMLEntity("&Eacute;",   201),		//	latin capital letter E with acute, U+00C9 ISOlat1
		new HTMLEntity("&Ecirc;",    202),		//	latin capital letter E with circumflex, U+00CA ISOlat1
		new HTMLEntity("&Euml;",     203),		//	latin capital letter E with diaeresis, U+00CB ISOlat1
		new HTMLEntity("&Igrave;",   204),		//	latin capital letter I with grave, U+00CC ISOlat1
		new HTMLEntity("&Iacute;",   205),		//	latin capital letter I with acute, U+00CD ISOlat1
		new HTMLEntity("&Icirc;",    206),		//	latin capital letter I with circumflex, U+00CE ISOlat1
		new HTMLEntity("&Iuml;",     207),		//	latin capital letter I with diaeresis, U+00CF ISOlat1
		new HTMLEntity("&ETH;",      208),		//	latin capital letter ETH, U+00D0 ISOlat1
		new HTMLEntity("&Ntilde;",   209),		//	latin capital letter N with tilde, U+00D1 ISOlat1
		new HTMLEntity("&Ograve;",   210),		//	latin capital letter O with grave, U+00D2 ISOlat1
		new HTMLEntity("&Oacute;",   211),		//	latin capital letter O with acute, U+00D3 ISOlat1
		new HTMLEntity("&Ocirc;",    212),		//	latin capital letter O with circumflex, U+00D4 ISOlat1
		new HTMLEntity("&Otilde;",   213),		//	latin capital letter O with tilde, U+00D5 ISOlat1
		new HTMLEntity("&Ouml;",     214),		//	latin capital letter O with diaeresis, U+00D6 ISOlat1
		new HTMLEntity("&times;",    215),		//	multiplication sign, U+00D7 ISOnum
		new HTMLEntity("&Oslash;",   216),		//	latin capital letter O with stroke = latin capital letter O slash, U+00D8 ISOlat1
		new HTMLEntity("&Ugrave;",   217),		//	latin capital letter U with grave, U+00D9 ISOlat1
		new HTMLEntity("&Uacute;",   218),		//	latin capital letter U with acute, U+00DA ISOlat1
		new HTMLEntity("&Ucirc;",    219),		//	latin capital letter U with circumflex, U+00DB ISOlat1
		new HTMLEntity("&Uuml;",     220),		//	latin capital letter U with diaeresis, U+00DC ISOlat1
		new HTMLEntity("&Yacute;",   221),		//	latin capital letter Y with acute, U+00DD ISOlat1
		new HTMLEntity("&THORN;",    222),		//	latin capital letter THORN, U+00DE ISOlat1
		new HTMLEntity("&szlig;",    223),		//	latin small letter sharp s = ess-zed, U+00DF ISOlat1
		new HTMLEntity("&agrave;",   224),		//	latin small letter a with grave = latin small letter a grave, U+00E0 ISOlat1
		new HTMLEntity("&aacute;",   225),		//	latin small letter a with acute, U+00E1 ISOlat1
		new HTMLEntity("&acirc;",    226),		//	latin small letter a with circumflex, U+00E2 ISOlat1
		new HTMLEntity("&atilde;",   227),		//	latin small letter a with tilde, U+00E3 ISOlat1
		new HTMLEntity("&auml;",     228),		//	latin small letter a with diaeresis, U+00E4 ISOlat1
		new HTMLEntity("&aring;",    229),		//	latin small letter a with ring above = latin small letter a ring, U+00E5 ISOlat1
		new HTMLEntity("&aelig;",    230),		//	latin small letter ae = latin small ligature ae, U+00E6 ISOlat1
		new HTMLEntity("&ccedil;",   231),		//	latin small letter c with cedilla, U+00E7 ISOlat1
		new HTMLEntity("&egrave;",   232),		//	latin small letter e with grave, U+00E8 ISOlat1
		new HTMLEntity("&eacute;",   233),		//	latin small letter e with acute, U+00E9 ISOlat1
		new HTMLEntity("&ecirc;",    234),		//	latin small letter e with circumflex, U+00EA ISOlat1
		new HTMLEntity("&euml;",     235),		//	latin small letter e with diaeresis, U+00EB ISOlat1
		new HTMLEntity("&igrave;",   236),		//	latin small letter i with grave, U+00EC ISOlat1
		new HTMLEntity("&iacute;",   237),		//	latin small letter i with acute, U+00ED ISOlat1
		new HTMLEntity("&icirc;",    238),		//	latin small letter i with circumflex, U+00EE ISOlat1
		new HTMLEntity("&iuml;",     239),		//	latin small letter i with diaeresis, U+00EF ISOlat1
		new HTMLEntity("&eth;",      240),		//	latin small letter eth, U+00F0 ISOlat1
		new HTMLEntity("&ntilde;",   241),		//	latin small letter n with tilde, U+00F1 ISOlat1
		new HTMLEntity("&ograve;",   242),		//	latin small letter o with grave, U+00F2 ISOlat1
		new HTMLEntity("&oacute;",   243),		//	latin small letter o with acute, U+00F3 ISOlat1
		new HTMLEntity("&ocirc;",    244),		//	latin small letter o with circumflex, U+00F4 ISOlat1
		new HTMLEntity("&otilde;",   245),		//	latin small letter o with tilde, U+00F5 ISOlat1
		new HTMLEntity("&ouml;",     246),		//	latin small letter o with diaeresis, U+00F6 ISOlat1
		new HTMLEntity("&divide;",   247),		//	division sign, U+00F7 ISOnum
		new HTMLEntity("&oslash;",   248),		//	latin small letter o with stroke, = latin small letter o slash, U+00F8 ISOlat1
		new HTMLEntity("&ugrave;",   249),		//	latin small letter u with grave, U+00F9 ISOlat1
		new HTMLEntity("&uacute;",   250),		//	latin small letter u with acute, U+00FA ISOlat1
		new HTMLEntity("&ucirc;",    251),		//	latin small letter u with circumflex, U+00FB ISOlat1
		new HTMLEntity("&uuml;",     252),		//	latin small letter u with diaeresis, U+00FC ISOlat1
		new HTMLEntity("&yacute;",   253),		//	latin small letter y with acute, U+00FD ISOlat1
		new HTMLEntity("&thorn;",    254),		//	latin small letter thorn, U+00FE ISOlat1
		new HTMLEntity("&yuml;",     255),		//	latin small letter y with diaeresis, U+00FF ISOlat1
		new HTMLEntity("&OElig;",    338),		//	latin capital ligature OE, U+0152 ISOlat2
		new HTMLEntity("&oelig;",    339),		//	latin small ligature oe, U+0153 ISOlat2
		new HTMLEntity("&Scaron;",   352),		//	latin capital letter S with caron, U+0160 ISOlat2
		new HTMLEntity("&scaron;",   353),		//	latin small letter s with caron, U+0161 ISOlat2
		new HTMLEntity("&Yuml;",     376),		//	latin capital letter Y with diaeresis, U+0178 ISOlat2
		new HTMLEntity("&fnof;",     402),		//	latin small f with hook = function = florin, U+0192 ISOtech
		new HTMLEntity("&circ;",     710),		//	modifier letter circumflex accent, U+02C6 ISOpub
		new HTMLEntity("&tilde;",    732),		//	small tilde, U+02DC ISOdia
		new HTMLEntity("&Alpha;",    913),		//	greek capital letter alpha, U+0391
		new HTMLEntity("&Beta;",     914),		//	greek capital letter beta, U+0392
		new HTMLEntity("&Gamma;",    915),		//	greek capital letter gamma, U+0393 ISOgrk3
		new HTMLEntity("&Delta;",    916),		//	greek capital letter delta, U+0394 ISOgrk3
		new HTMLEntity("&Epsilon;",  917),		//	greek capital letter epsilon, U+0395
		new HTMLEntity("&Zeta;",     918),		//	greek capital letter zeta, U+0396
		new HTMLEntity("&Eta;",      919),		//	greek capital letter eta, U+0397
		new HTMLEntity("&Theta;",    920),		//	greek capital letter theta, U+0398 ISOgrk3
		new HTMLEntity("&Iota;",     921),		//	greek capital letter iota, U+0399
		new HTMLEntity("&Kappa;",    922),		//	greek capital letter kappa, U+039A
		new HTMLEntity("&Lambda;",   923),		//	greek capital letter lambda, U+039B ISOgrk3
		new HTMLEntity("&Mu;",       924),		//	greek capital letter mu, U+039C
		new HTMLEntity("&Nu;",       925),		//	greek capital letter nu, U+039D
		new HTMLEntity("&Xi;",       926),		//	greek capital letter xi, U+039E ISOgrk3
		new HTMLEntity("&Omicron;",  927),		//	greek capital letter omicron, U+039F
		new HTMLEntity("&Pi;",       928),		//	greek capital letter pi, U+03A0 ISOgrk3
		new HTMLEntity("&Rho;",      929),		//	greek capital letter rho, U+03A1
		new HTMLEntity("&Sigma;",    931),		//	greek capital letter sigma, U+03A3 ISOgrk3
		new HTMLEntity("&Tau;",      932),		//	greek capital letter tau, U+03A4
		new HTMLEntity("&Upsilon;",  933),		//	greek capital letter upsilon, U+03A5 ISOgrk3
		new HTMLEntity("&Phi;",      934),		//	greek capital letter phi, U+03A6 ISOgrk3
		new HTMLEntity("&Chi;",      935),		//	greek capital letter chi, U+03A7
		new HTMLEntity("&Psi;",      936),		//	greek capital letter psi, U+03A8 ISOgrk3
		new HTMLEntity("&Omega;",    937),		//	greek capital letter omega, U+03A9 ISOgrk3
		new HTMLEntity("&alpha;",    945),		//	greek small letter alpha, U+03B1 ISOgrk3
		new HTMLEntity("&beta;",     946),		//	greek small letter beta, U+03B2 ISOgrk3
		new HTMLEntity("&gamma;",    947),		//	greek small letter gamma, U+03B3 ISOgrk3
		new HTMLEntity("&delta;",    948),		//	greek small letter delta, U+03B4 ISOgrk3
		new HTMLEntity("&epsilon;",  949),		//	greek small letter epsilon, U+03B5 ISOgrk3
		new HTMLEntity("&zeta;",     950),		//	greek small letter zeta, U+03B6 ISOgrk3
		new HTMLEntity("&eta;",      951),		//	greek small letter eta, U+03B7 ISOgrk3
		new HTMLEntity("&theta;",    952),		//	greek small letter theta, U+03B8 ISOgrk3
		new HTMLEntity("&iota;",     953),		//	greek small letter iota, U+03B9 ISOgrk3
		new HTMLEntity("&kappa;",    954),		//	greek small letter kappa, U+03BA ISOgrk3
		new HTMLEntity("&lambda;",   955),		//	greek small letter lambda, U+03BB ISOgrk3
		new HTMLEntity("&mu;",       956),		//	greek small letter mu, U+03BC ISOgrk3
		new HTMLEntity("&nu;",       957),		//	greek small letter nu, U+03BD ISOgrk3
		new HTMLEntity("&xi;",       958),		//	greek small letter xi, U+03BE ISOgrk3
		new HTMLEntity("&omicron;",  959),		//	greek small letter omicron, U+03BF NEW
		new HTMLEntity("&pi;",       960),		//	greek small letter pi, U+03C0 ISOgrk3
		new HTMLEntity("&rho;",      961),		//	greek small letter rho, U+03C1 ISOgrk3
		new HTMLEntity("&sigmaf;",   962),		//	greek small letter final sigma, U+03C2 ISOgrk3
		new HTMLEntity("&sigma;",    963),		//	greek small letter sigma, U+03C3 ISOgrk3
		new HTMLEntity("&tau;",      964),		//	greek small letter tau, U+03C4 ISOgrk3
		new HTMLEntity("&upsilon;",  965),		//	greek small letter upsilon, U+03C5 ISOgrk3
		new HTMLEntity("&phi;",      966),		//	greek small letter phi, U+03C6 ISOgrk3
		new HTMLEntity("&chi;",      967),		//	greek small letter chi, U+03C7 ISOgrk3
		new HTMLEntity("&psi;",      968),		//	greek small letter psi, U+03C8 ISOgrk3
		new HTMLEntity("&omega;",    969),		//	greek small letter omega, U+03C9 ISOgrk3
		new HTMLEntity("&thetasym;", 977),		//	greek small letter theta symbol, U+03D1 NEW
		new HTMLEntity("&upsih;",    978),		//	greek upsilon with hook symbol, U+03D2 NEW
		new HTMLEntity("&piv;",      982),		//	greek pi symbol, U+03D6 ISOgrk3
		new HTMLEntity("&ensp;",    8194),		//	en space, U+2002 ISOpub
		new HTMLEntity("&emsp;",    8195),		//	em space, U+2003 ISOpub
		new HTMLEntity("&thinsp;",  8201),		//	thin space, U+2009 ISOpub
		new HTMLEntity("&zwnj;",    8204),		//	zero width non-joiner, U+200C NEW RFC 2070
		new HTMLEntity("&zwj;",     8205),		//	zero width joiner, U+200D NEW RFC 2070
		new HTMLEntity("&lrm;",     8206),		//	left-to-right mark, U+200E NEW RFC 2070
		new HTMLEntity("&rlm;",     8207),		//	right-to-left mark, U+200F NEW RFC 2070
		new HTMLEntity("&ndash;",   8211),		//	en dash, U+2013 ISOpub
		new HTMLEntity("&mdash;",   8212),		//	em dash, U+2014 ISOpub
		new HTMLEntity("&lsquo;",   8216),		//	left single quotation mark, U+2018 ISOnum
		new HTMLEntity("&rsquo;",   8217),		//	right single quotation mark, U+2019 ISOnum
		new HTMLEntity("&sbquo;",   8218),		//	single low-9 quotation mark, U+201A NEW
		new HTMLEntity("&ldquo;",   8220),		//	left double quotation mark, U+201C ISOnum
		new HTMLEntity("&rdquo;",   8221),		//	right double quotation mark, U+201D ISOnum
		new HTMLEntity("&bdquo;",   8222),		//	double low-9 quotation mark, U+201E NEW
		new HTMLEntity("&dagger;",  8224),		//	dagger, U+2020 ISOpub
		new HTMLEntity("&Dagger;",  8225),		//	double dagger, U+2021 ISOpub
		new HTMLEntity("&bull;",    8226),		//	bullet = black small circle, U+2022 ISOpub
		new HTMLEntity("&hellip;",  8230),		//	horizontal ellipsis = three dot leader, U+2026 ISOpub
		new HTMLEntity("&permil;",  8240),		//	per mille sign, U+2030 ISOtech
		new HTMLEntity("&prime;",   8242),		//	prime = minutes = feet, U+2032 ISOtech
		new HTMLEntity("&Prime;",   8243),		//	double prime = seconds = inches, U+2033 ISOtech
		new HTMLEntity("&lsaquo;",  8249),		//	single left-pointing angle quotation mark, U+2039 ISO proposed
		new HTMLEntity("&rsaquo;",  8250),		//	single right-pointing angle quotation mark, U+203A ISO proposed
		new HTMLEntity("&oline;",   8254),		//	overline = spacing overscore, U+203E NEW
		new HTMLEntity("&frasl;",   8260),		//	fraction slash, U+2044 NEW
		new HTMLEntity("&euro;",    8364),		//	euro sign, U+20AC NEW
		new HTMLEntity("&image;",   8465),		//	blackletter capital I = imaginary part, U+2111 ISOamso
		new HTMLEntity("&weierp;",  8472),		//	script capital P = power set = Weierstrass p, U+2118 ISOamso
		new HTMLEntity("&real;",    8476),		//	blackletter capital R = real part symbol, U+211C ISOamso
		new HTMLEntity("&trade;",   8482),		//	trade mark sign, U+2122 ISOnum
		new HTMLEntity("&alefsym;", 8501),		//	alef symbol = first transfinite cardinal, U+2135 NEW
		new HTMLEntity("&larr;",    8592),		//	leftwards arrow, U+2190 ISOnum
		new HTMLEntity("&uarr;",    8593),		//	upwards arrow, U+2191
		new HTMLEntity("&rarr;",    8594),		//	rightwards arrow, U+2192 ISOnum
		new HTMLEntity("&darr;",    8595),		//	downwards arrow, U+2193 ISOnum
		new HTMLEntity("&harr;",    8596),		//	left right arrow, U+2194 ISOamsa
		new HTMLEntity("&crarr;",   8629),		//	downwards arrow with corner leftwards = carriage return, U+21B5 NEW
		new HTMLEntity("&lArr;",    8656),		//	leftwards double arrow, U+21D0 ISOtech
		new HTMLEntity("&uArr;",    8657),		//	upwards double arrow, U+21D1 ISOamsa
		new HTMLEntity("&rArr;",    8658),		//	rightwards double arrow, U+21D2 ISOtech
		new HTMLEntity("&dArr;",    8659),		//	downwards double arrow, U+21D3 ISOamsa
		new HTMLEntity("&hArr;",    8660),		//	left right double arrow, U+21D4 ISOamsa
		new HTMLEntity("&forall;",  8704),		//	for all, U+2200 ISOtech
		new HTMLEntity("&part;",    8706),		//	partial differential, U+2202 ISOtech
		new HTMLEntity("&exist;",   8707),		//	there exists, U+2203 ISOtech
		new HTMLEntity("&empty;",   8709),		//	empty set = null set = diameter, U+2205 ISOamso
		new HTMLEntity("&nabla;",   8711),		//	nabla = backward difference, U+2207 ISOtech
		new HTMLEntity("&isin;",    8712),		//	element of, U+2208 ISOtech
		new HTMLEntity("&notin;",   8713),		//	not an element of, U+2209 ISOtech
		new HTMLEntity("&ni;",      8715),		//	contains as member, U+220B ISOtech
		new HTMLEntity("&prod;",    8719),		//	n-ary product = product sign, U+220F ISOamsb
		new HTMLEntity("&sum;",     8721),		//	n-ary sumation, U+2211 ISOamsb
		new HTMLEntity("&minus;",   8722),		//	minus sign, U+2212 ISOtech
		new HTMLEntity("&lowast;",  8727),		//	asterisk operator, U+2217 ISOtech
		new HTMLEntity("&radic;",   8730),		//	square root = radical sign, U+221A ISOtech
		new HTMLEntity("&prop;",    8733),		//	proportional to, U+221D ISOtech
		new HTMLEntity("&infin;",   8734),		//	infinity, U+221E ISOtech
		new HTMLEntity("&ang;",     8736),		//	angle, U+2220 ISOamso
		new HTMLEntity("&and;",     8743),		//	logical and = wedge, U+2227 ISOtech
		new HTMLEntity("&or;",      8744),		//	logical or = vee, U+2228 ISOtech
		new HTMLEntity("&cap;",     8745),		//	intersection = cap, U+2229 ISOtech
		new HTMLEntity("&cup;",     8746),		//	union = cup, U+222A ISOtech
		new HTMLEntity("&int;",     8747),		//	integral, U+222B ISOtech
		new HTMLEntity("&there4;",  8756),		//	therefore, U+2234 ISOtech
		new HTMLEntity("&sim;",     8764),		//	tilde operator = varies with = similar to, U+223C ISOtech
		new HTMLEntity("&cong;",    8773),		//	approximately equal to, U+2245 ISOtech
		new HTMLEntity("&asymp;",   8776),		//	almost equal to = asymptotic to, U+2248 ISOamsr
		new HTMLEntity("&ne;",      8800),		//	not equal to, U+2260 ISOtech
		new HTMLEntity("&equiv;",   8801),		//	identical to, U+2261 ISOtech
		new HTMLEntity("&le;",      8804),		//	less-than or equal to, U+2264 ISOtech
		new HTMLEntity("&ge;",      8805),		//	greater-than or equal to, U+2265 ISOtech
		new HTMLEntity("&sub;",     8834),		//	subset of, U+2282 ISOtech
		new HTMLEntity("&sup;",     8835),		//	superset of, U+2283 ISOtech
		new HTMLEntity("&nsub;",    8836),		//	not a subset of, U+2284 ISOamsn
		new HTMLEntity("&sube;",    8838),		//	subset of or equal to, U+2286 ISOtech
		new HTMLEntity("&supe;",    8839),		//	superset of or equal to, U+2287 ISOtech
		new HTMLEntity("&oplus;",   8853),		//	circled plus = direct sum, U+2295 ISOamsb
		new HTMLEntity("&otimes;",  8855),		//	circled times = vector product, U+2297 ISOamsb
		new HTMLEntity("&perp;",    8869),		//	up tack = orthogonal to = perpendicular, U+22A5 ISOtech
		new HTMLEntity("&sdot;",    8901),		//	dot operator, U+22C5 ISOamsb
		new HTMLEntity("&lceil;",   8968),		//	left ceiling = apl upstile, U+2308 ISOamsc
		new HTMLEntity("&rceil;",   8969),		//	right ceiling, U+2309 ISOamsc
		new HTMLEntity("&lfloor;",  8970),		//	left floor = apl downstile, U+230A ISOamsc
		new HTMLEntity("&rfloor;",  8971),		//	right floor, U+230B ISOamsc
		new HTMLEntity("&lang;",    9001),		//	left-pointing angle bracket = bra, U+2329 ISOtech
		new HTMLEntity("&rang;",    9002),		//	right-pointing angle bracket = ket, U+232A ISOtech
		new HTMLEntity("&loz;",     9674),		//	lozenge, U+25CA ISOpub
		new HTMLEntity("&spades;",  9824),		//	black spade suit, U+2660 ISOpub
		new HTMLEntity("&clubs;",   9827),		//	black club suit = shamrock, U+2663 ISOpub
		new HTMLEntity("&hearts;",  9829),		//	black heart suit = valentine, U+2665 ISOpub
		new HTMLEntity("&diams;",   9830),		//	black diamond suit, U+2666 ISOpub
	};
	
	private static final HashMap<String, HTMLEntity>	DefaultEntities = new HashMap<String, HTMLEntity>(Mappings.length);
	
	static {
		
	//	Populate the Entities hashtable.
		
		final HashMap<String, HTMLEntity>	DefaultEntities = HTMLEntityCodec.DefaultEntities;
		
		for (final HTMLEntity Map : Mappings)
			DefaultEntities.put(Map.Markup, Map);
	}	
	
	private HashMap<String, HTMLEntity>	Entities        = DefaultEntities;
	private final String[]				SingleByteTable = new String[256];
	private final int					FirstMultiByteMappingIndex;
	
	
/**	Creates an <code>HTMLEntityCodec</code> instance for translating back and forth between Unicode characters and HTML 
 *	character entity references. Double-quote characters are translated to/from the HTML character entity reference 
 *	"&amp;quot;". Unprintable characters are omitted. Line separator characters ('\n') are left unaltered.
 */
	public HTMLEntityCodec() {	
		this(true);
	}
	
/**	Creates an <code>HTMLEntityCodec</code> instance for translating back and forth between Unicode characters and HTML 
 *	character entity references.
 * 
 * @param EncodeDoubleQuotes	<code>true</code> if the double-quote character (") should be replaced by an HTML
 *								character entity reference, <code>false</code> if it should be left unchanged.
 */
	public HTMLEntityCodec
	(
		final boolean			EncodeDoubleQuotes	
	)
	{
		
	//	Place any mappings for characters 0-255 into the single byte character lookup table.
		
		final HTMLEntity[]		Mappings                   = HTMLEntityCodec.Mappings;
		final String[]			SingleByteTable            = this.SingleByteTable;
		int						FirstMultiByteMappingIndex = 0;
		
		for (int Index = 0, Count = Mappings.length; Index < Count; Index++) {
			final HTMLEntity		Map         = Mappings[Index];
			final int				MapCharCode = Map.UnicodeCodePoint;
			
			if (MapCharCode > 255) {
				
				FirstMultiByteMappingIndex = Index;
				break;
			}
			
			SingleByteTable[MapCharCode] = Map.Markup;
		}
		
		this.FirstMultiByteMappingIndex = FirstMultiByteMappingIndex;
		
	//	The table already includes a mapping for the double-quote character, so our only choice is whether or not to 
	//	remove that mapping.
		
		if (EncodeDoubleQuotes == false)
			SingleByteTable['"'] = null;	
	}
	
/**	Returns the encoded form of the specified Unicode character as a <code>String</code>, or returns <code>null</code>, 
 *	if the character will be left unchanged.
 * 
 * @param Char			The character to be encoded.
 * @return				The encoded form the character as a <code>String</code>, or <code>null</code> if the character 
 *						doesn't need to be encoded.
 */	
	public String getEncoding
	(
		final char				Char
	)
	{
		if (Char <= 255)
			return SingleByteTable[Char];
		else
			return getMultiByteMapping(Char);
	}
	
/**	Sets the encoded form of the specified Unicode character.
 * 
 * @param Char			The character for which an encoding is being specified.
 * @param Encoded		The encoding to be used for the specified character, or <code>null</code>, if the character 
 *						should be left unchanged.
 * @return 				A reference to this <code>HTMLEntityCodec</code> instance.
 */
	public HTMLEntityCodec setEncoding
	(
		final char				Char,
		final String			Encoded
	)
	{
	//	Update the single-byte table, if applicable.
	
		if (Char <= 255)
			SingleByteTable[Char] = Encoded;
		
	//	We don't want to alter the original, static HashMap of encodings, so, if we find we're still using a
	//	reference to it, we make a clone of it, and then we alter, and use, the clone from now on.
		
		if (Entities == DefaultEntities)
			Entities = (HashMap<String, HTMLEntity>) DefaultEntities.clone();
		
	//	Update the Entities hashtable.
		
		if (Encoded != null)
			Entities.put(Encoded, new HTMLEntity(Encoded, (int) Char));
		else
			Entities.remove(Encoded);
			
		return this;
	}
	
/**	Encodes a Unicode string by translating characters into HTML character entity references, as needed.
 * 
 * @param UnStr					A Unicode string.
 * @return						A string in which all characters that were supposed to be translated into HTML entity
 *								references have been.
 */
	public String encode
	(
		final String			UnStr
	)
	{
		final StringBuilder		Buff = new StringBuilder(UnStr.length());
		
		if (encode(UnStr, Buff))
			return Buff.toString();
		else
			return UnStr;	//	Nothing changed, so return the original String.
	}
	
/**	Encodes a Unicode character sequence by translating characters into HTML character entity references, as needed.	
 * 
 * @param UnStr					A Unicode character sequence.
 * @param Buff					A <code>StringBuffer</code> to which the output of this method is to be appended. The
 *								buffer is <b>not</b> cleared.
 * @return						<code>true</code> if any characters were encoded, <code>false</code> if no characters
 *								were encoded.
 */
	public boolean encode
	(
		final CharSequence		UnStr,
		final StringBuffer		Buff
	)
	{
		boolean					Encoded         = false;
		final String[]			SingleByteTable = this.SingleByteTable;
		
		for (int Index = 0, Count = UnStr.length(); Index < Count; Index++) {
			final char				Char = UnStr.charAt(Index);
			final String			Mapping;
			
			if (Char <= 255)
				Mapping = SingleByteTable[Char];
			else
				Mapping = getMultiByteMapping(Char);
			
			if (Mapping == null) {
				
				if (Char < 127) {
					
					Buff.append(Char);
					
				} else {
					
				//	Note: Representing Char as hexadecimal digits would reduce document size, but would slow-down 
				//	encoding by requiring invocation of Integer.toHexString with the associated creation of many 
				//	String objects.
					
					Buff.append("&#").append((int) Char).append(';');
					Encoded = true;
				}
				
			} else {
				
				Buff.append(Mapping);
				Encoded = true;
			}
		}
		
		return Encoded;
	}
	
/**	Encodes a Unicode character sequence by translating characters into HTML character entity references, as needed.	
 * 
 * @param UnStr					A Unicode character sequence.
 * @param Buff					A <code>StringBuilder</code> to which the output of this method is to be appended. The
 *								buffer is <b>not</b> cleared.
 * @return						<code>true</code> if any characters were encoded, <code>false</code> if no characters
 *								were encoded.
 */
	public boolean encode
	(
		final CharSequence		UnStr,
		final StringBuilder		Buff
	)
	{
		boolean					Encoded         = false;
		final String[]			SingleByteTable = this.SingleByteTable;
	
		for (int Index = 0, Count = UnStr.length(); Index < Count; Index++) {
			final char				Char = UnStr.charAt(Index);
			final String			Mapping;
		
			if (Char <= 255)
				Mapping = SingleByteTable[Char];
			else
				Mapping = getMultiByteMapping(Char);
		
			if (Mapping == null) {
			
				if (Char < 127) {
				
					Buff.append(Char);
				
				} else {
					
				//	Note: Representing Char as hexadecimal digits would reduce document size, but would slow-down 
				//	encoding by requiring invocation of Integer.toHexString with the associated creation of many 
				//	String objects.
				
					Buff.append("&#").append((int) Char).append(';');
					Encoded = true;
				}
			
			} else {
			
				Buff.append(Mapping);
				Encoded = true;
			}
		}
	
		return Encoded;
	}
	
/**	Encodes a Unicode string, represented as a {@link Characters} object, by translating characters into HTML 
 *	character entity references, as needed.	
 * 
 * @param UnStr					A Unicode string.
 * @return						A sequence of characters in which all characters that were supposed to be translated 
 *								into HTML entity references have been.
 */
	public Characters encode
	(
		final Characters		UnStr
	)
	{
		if (canSkipEncoding(UnStr))
			return UnStr;
		
		final StringBuilder		Buff = new StringBuilder(UnStr.length());
		
		if (encode(UnStr, Buff))
			return new Characters(Buff);
		else
			return UnStr;	//	Nothing changed, so return the original String.
	}
	
/**	Encodes a Unicode string, represented as a {@link Characters} object, by translating characters into HTML character 
 *	entity references, as needed.	
 * 
 * @param UnStr					A Unicode string.
 * @param Buff					A <code>StringBuffer</code> to which the output of this method is to be appended. The
 *								buffer is <b>not</b> cleared.
 * @return						<code>true</code> if any characters were encoded, <code>false</code> if no characters
 *								were encoded.
 */
	public boolean encode
	(
		final Characters		UnStr,
		final StringBuffer		Buff
	)
	{
		boolean					Encoded = false;
		
		if (UnStr.hasArray()) {
			final char[]			Chars           = UnStr.getArray();
			final String[]			SingleByteTable = this.SingleByteTable;
			
			for (int Index = UnStr.getArrayOffset(), Count = UnStr.length(); --Count >= 0; Index++) {
				final char				Char = Chars[Index];
				final String			Mapping;
				
				if (Char <= 255)
					Mapping = SingleByteTable[Char];
				else
					Mapping = getMultiByteMapping(Char);
				
				if (Mapping == null) {
				
					if (Char < 127) {
						
						Buff.append(Char);
					
					} else {
						
					//	Note: Representing Char as hexadecimal digits would reduce document size, but would slow-down 
					//	encoding by requiring invocation of Integer.toHexString with the associated creation of many 
					//	String objects.
						
						Buff.append("&#").append((int) Char).append(';');
						Encoded = true;
					}
				
				} else {
					
					Buff.append(Mapping);
					Encoded = true;
				}
			}
			
		} else
			Encoded = encode(UnStr.toString(), Buff);
		
		return Encoded;		
	}
	
/**	Encodes a Unicode string, represented as a {@link Characters} object, by translating characters into HTML character 
 *	entity references, as needed.	
 * 
 * @param UnStr					A Unicode string.
 * @param Buff					A <code>StringBuilder</code> to which the output of this method is to be appended. The
 *								buffer is <b>not</b> cleared.
 * @return						<code>true</code> if any characters were encoded, <code>false</code> if no characters
 *								were encoded.
 */
	public boolean encode
	(
		final Characters		UnStr,
		final StringBuilder		Buff
	)
	{
		boolean					Encoded = false;
	
		if (UnStr.hasArray()) {
			final char[]			Chars           = UnStr.getArray();
			final String[]			SingleByteTable = this.SingleByteTable;
		
			for (int Index = UnStr.getArrayOffset(), Count = UnStr.length(); --Count >= 0; Index++) {
				final char				Char = Chars[Index];
				final String			Mapping;
			
				if (Char <= 255)
					Mapping = SingleByteTable[Char];
				else
					Mapping = getMultiByteMapping(Char);
			
				if (Mapping == null) {
			
					if (Char < 127) {
					
						Buff.append(Char);
				
					} else {

					//	Note: Representing Char as hexadecimal digits would reduce document size, but would slow-down 
					//	encoding by requiring invocation of Integer.toHexString with the associated creation of many 
					//	String objects.
					
						Buff.append("&#").append((int) Char).append(';');
						Encoded = true;
					}
			
				} else {
				
					Buff.append(Mapping);
					Encoded = true;
				}
			}
		
		} else
			Encoded = encode(UnStr.toString(), Buff);
	
		return Encoded;		
	}
	
	private boolean canSkipEncoding
	(
		final Characters		UnStr
	)
	{
		if (UnStr.hasArray()) {
			final char[]			Chars           = UnStr.getArray();
			final String[]			SingleByteTable = this.SingleByteTable;
			
			for (int Index = UnStr.getArrayOffset(), Count = UnStr.length(); --Count >= 0; Index++) {
				final char				Char = Chars[Index];
				
				if (Char > 255 || SingleByteTable[Char] != null)
					return false;
			}
			
			return true;
		}
		
		return false;
	}
	
	private static boolean canSkipDecoding
	(
		final Characters		EnStr
	)
	{
		return EnStr.indexOf('&') == -1;
	}
	
	private static boolean canSkipDecoding
	(
		final String			EnStr
	)
	{
		return EnStr.indexOf('&') == -1;
	}
	
/**	Decodes a string by translating HTML character entity references into Unicode characters.
 * 
 * @param EnStr				The string to be decoded.
 * @return					A <code>String</code> in which all of HTML character entity references in the original
 *							string have been translated into their Unicode equivalents.
 */	
	public String decode
	(
		final String			EnStr
	)
	{
		if (canSkipDecoding(EnStr))
			return EnStr;
		
		final StringBuilder		Buff = new StringBuilder(EnStr.length());
		
		if (decodeLow(EnStr, Buff))
			return Buff.toString();
		else
			return EnStr;	//	Nothing changed, so return the original String.
	}
	
/**	Decodes a string by translating HTML character entity references into corresponding Unicode characters.
 * 
 * @param EnStr					The string to be decoded.
 * @param Buff					A <code>StringBuffer</code> to which the output of this method is to be appended. The
 *								buffer is <b>not</b> cleared.
 * @return						<code>true</code> if any characters were decoded, <code>false</code> if no characters
 *								were decoded.
 */
	public boolean decode
	(
		final String			EnStr,
		final StringBuffer		Buff
	)
	{
		if (canSkipDecoding(EnStr)) {
			
			Buff.append(EnStr);
			return false;
		}
		
		return decodeLow(EnStr, Buff);
	}
	
/**	Decodes a string by translating HTML character entity references into corresponding Unicode characters.
 * 
 * @param EnStr					The string to be decoded.
 * @param Buff					A <code>StringBuilder</code> to which the output of this method is to be appended. The
 *								buffer is <b>not</b> cleared.
 * @return						<code>true</code> if any characters were decoded, <code>false</code> if no characters
 *								were decoded.
 */
	public boolean decode
	(
		final String			EnStr,
		final StringBuilder		Buff
	)
	{
		if (canSkipDecoding(EnStr)) {
			
			Buff.append(EnStr);
			return false;
		}
	
		return decodeLow(EnStr, Buff);
	}
	
/**	Decodes a string by translating HTML character entity references into corresponding Unicode characters.
 * 
 * @param EnStr					The string to be decoded.
 * @param Buff					A <code>StringBuffer</code> to which the output of this method is to be appended. The
 *								buffer is <b>not</b> cleared.
 * @return						<code>true</code> if any characters were decoded, <code>false</code> if no characters
 *								were decoded.
 */
	private boolean decodeLow
	(
		final String			EnStr,
		final StringBuffer		Buff
	)
	{
		boolean								Decoded           = true;
		int									UnencodedRunStart = 0;
		int									AmpIdx            = EnStr.indexOf('&');
		final HashMap<String, HTMLEntity>	Entities          = this.Entities;
		
		while (AmpIdx >= 0) {
			final int				SCIdx = EnStr.indexOf(';', AmpIdx + 1);
			
			if (SCIdx < 0)
				break;
			
			final String			PossibleEntityName = EnStr.substring(AmpIdx, SCIdx + 1);
			int						DecodedCodePoint   = 0;
			boolean					DecodedThisChar    = false;
			
		//	See <http://www.w3.org/TR/html4/charset.html#h-5.3.1>.
			
			if (PossibleEntityName.charAt(1) == '#') {
				final char				Char2 = PossibleEntityName.charAt(2);
				
				if (Char2 == 'x' || Char2 == 'X') {
					
					DecodedCodePoint = Integer.parseInt(PossibleEntityName.substring(3, PossibleEntityName.length() - 1), 16);
					DecodedThisChar  = true;
					
				} else {
					
					DecodedCodePoint = Integer.parseInt(PossibleEntityName.substring(2, PossibleEntityName.length() - 1));
					DecodedThisChar  = true;
				}
				
			} else {
				final HTMLEntity		Map = Entities.get(PossibleEntityName);
				
				if (Map != null) {
					
					DecodedCodePoint = Map.UnicodeCodePoint;
					DecodedThisChar  = true;
				}
			}
			
			if (DecodedThisChar) {
				
				if (UnencodedRunStart < AmpIdx)
					Buff.append(EnStr.substring(UnencodedRunStart, AmpIdx));
				
				Buff.appendCodePoint(DecodedCodePoint);
				
				UnencodedRunStart = SCIdx + 1;
			}
			
		//	Next iteration.
			
			AmpIdx = EnStr.indexOf('&', SCIdx + 1);
		}
		
		if (UnencodedRunStart == 0) {
			
			Buff.append(EnStr);
			Decoded = false;
			
		} else
			Buff.append(EnStr.substring(UnencodedRunStart));
		
		return Decoded;
	}
	
/**	Decodes a string by translating HTML character entity references into corresponding Unicode characters.
 * 
 * @param EnStr					The string to be decoded.
 * @param Buff					A <code>StringBuilder</code> to which the output of this method is to be appended. The
 *								buffer is <b>not</b> cleared.
 * @return						<code>true</code> if any characters were decoded, <code>false</code> if no characters
 *								were decoded.
 */
	private boolean decodeLow
	(
		final String			EnStr,
		final StringBuilder		Buff
	)
	{
		boolean								Decoded           = true;
		int									UnencodedRunStart = 0;
		int									AmpIdx            = EnStr.indexOf('&');
		final HashMap<String, HTMLEntity>	Entities          = this.Entities;
	
		while (AmpIdx >= 0) {
			final int				SCIdx = EnStr.indexOf(';', AmpIdx + 1);
		
			if (SCIdx < 0)
				break;
		
			final String			PossibleEntityName = EnStr.substring(AmpIdx, SCIdx + 1);
			int						DecodedCodePoint   = 0;
			boolean					DecodedThisChar    = false;

		//	See <http://www.w3.org/TR/html4/charset.html#h-5.3.1>.
		
			if (PossibleEntityName.charAt(1) == '#') {
				final char				Char2 = PossibleEntityName.charAt(2);
				
				if (Char2 == 'x' || Char2 == 'X') {
					
					DecodedCodePoint = Integer.parseInt(PossibleEntityName.substring(3, PossibleEntityName.length() - 1), 16);
					DecodedThisChar  = true;
					
				} else {
					
					DecodedCodePoint = Integer.parseInt(PossibleEntityName.substring(2, PossibleEntityName.length() - 1));
					DecodedThisChar  = true;
				}
			
			} else {
				final HTMLEntity		Map = Entities.get(PossibleEntityName);
			
				if (Map != null) {
				
					DecodedCodePoint = Map.UnicodeCodePoint;
					DecodedThisChar  = true;
				}
			}
		
			if (DecodedThisChar) {
			
				if (UnencodedRunStart < AmpIdx)
					Buff.append(EnStr.substring(UnencodedRunStart, AmpIdx));
			
				Buff.appendCodePoint(DecodedCodePoint);
			
				UnencodedRunStart = SCIdx + 1;
			}
		
		//	Next iteration.
		
			AmpIdx = EnStr.indexOf('&', SCIdx + 1);
		}
		
		if (UnencodedRunStart == 0) {
		
			Buff.append(EnStr);
			Decoded = false;
		
		} else
			Buff.append(EnStr.substring(UnencodedRunStart));
	
		return Decoded;
	}
	
/**	Decodes a string, represented as a {@link Characters} object, by translating HTML character entity references into 
 *	corresponding Unicode characters.
 * 
 * @param EnStr				The string to be decoded.
 * @return					A <code>Characters</code> object in which all of HTML character entity references in the 
 *							original string have been translated into their Unicode equivalents.
 */	
	public Characters decode
	(
		final Characters		EnStr
	)
	{
		if (canSkipDecoding(EnStr))
			return EnStr;
		
		final StringBuilder		Buff = new StringBuilder(EnStr.length());
		
		if (decodeLow(EnStr, Buff))
			return new Characters(Buff);
		else
			return EnStr;	//	Nothing changed, so return the original String.
	}
	
/**	Decodes a string, represented as a {@link Characters} object, by translating HTML character entity references into 
 *	corresponding Unicode characters.
 * 
 * @param EnStr					The string to be decoded.
 * @param Buff					A <code>StringBuffer</code> to which the output of this method is to be appended. The
 *								buffer is <b>not</b> cleared.
 * @return						<code>true</code> if any characters were decoded, <code>false</code> if no characters
 *								were decoded.
 */
	public boolean decode
	(
		final Characters		EnStr,
		final StringBuffer		Buff
	)
	{
		if (canSkipDecoding(EnStr)) {
			
			Buff.append(EnStr);
			return false;
		}
		
		return decodeLow(EnStr, Buff);
	}
	
/**	Decodes a string, represented as a {@link Characters} object, by translating HTML character entity references into 
 *	corresponding Unicode characters.
 * 
 * @param EnStr					The string to be decoded.
 * @param Buff					A <code>StringBuffer</code> to which the output of this method is to be appended. The
 *								buffer is <b>not</b> cleared.
 * @return						<code>true</code> if any characters were decoded, <code>false</code> if no characters
 *								were decoded.
 */
	public boolean decode
	(
		final Characters		EnStr,
		final StringBuilder		Buff
	)
	{
		if (canSkipDecoding(EnStr)) {
		
			Buff.append(EnStr);
			return false;
		}
	
		return decodeLow(EnStr, Buff);
	}
	
/**	Decodes a string, represented as a {@link Characters} object, by translating HTML character entity references into 
 *	corresponding Unicode characters.
 * 
 * @param EnStr					The string to be decoded.
 * @param Buff					A <code>StringBuffer</code> to which the output of this method is to be appended. The
 *								buffer is <b>not</b> cleared.
 * @return						<code>true</code> if any characters were decoded, <code>false</code> if no characters
 *								were decoded.
 */
	private boolean decodeLow
	(
		final Characters		EnStr,
		final StringBuffer		Buff
	)
	{
		boolean					Decoded = true;
		
		if (EnStr.hasArray()) {
			final char[]						Chars             = EnStr.getArray();
			final int							CharsStartOffset  = EnStr.getArrayOffset();
			int									UnencodedRunStart = 0;
			int									AmpIdx            = EnStr.indexOf('&');
			final HashMap<String, HTMLEntity>	Entities          = this.Entities;
			
			while (AmpIdx >= 0) {
				final int				ScIdx = EnStr.indexOf(';', AmpIdx + 1);
				
				if (ScIdx < 0)
					break;
				
				final String			PossibleEntityName = EnStr.substring(AmpIdx, ScIdx + 1);	//	Isolate entity name with leading '&' and trailing ';'
				int						DecodedCodePoint   = 0;
				boolean					DecodedThisChar    = false;
				
			//	See <http://www.w3.org/TR/html4/charset.html#h-5.3.1>.
				
				if (PossibleEntityName.charAt(1) == '#') {
					final char				Char2 = PossibleEntityName.charAt(2);
					
					if (Char2 == 'x' || Char2 == 'X') {
						
						DecodedCodePoint = Integer.parseInt(PossibleEntityName.substring(3, PossibleEntityName.length() - 1), 16);
						DecodedThisChar  = true;
						
					} else {
						
						DecodedCodePoint = Integer.parseInt(PossibleEntityName.substring(2, PossibleEntityName.length() - 1));
						DecodedThisChar  = true;
					}
					
				} else {
					final HTMLEntity		Map = Entities.get(PossibleEntityName);
					
					if (Map != null) {
						
						DecodedCodePoint = Map.UnicodeCodePoint;
						DecodedThisChar  = true;
					}
				}
			
				if (DecodedThisChar) {
					
					if (UnencodedRunStart < AmpIdx)
						Buff.append(Chars, CharsStartOffset + UnencodedRunStart, AmpIdx - UnencodedRunStart);
					
					Buff.appendCodePoint(DecodedCodePoint);
					
					UnencodedRunStart = ScIdx + 1;
				}
				
			//	Next iteration.
				
				AmpIdx = EnStr.indexOf('&', ScIdx + 1);
			}
			
			if (UnencodedRunStart == 0) {
				
				EnStr.append(Buff);
				Decoded = false;
				
			} else
				Buff.append(Chars, CharsStartOffset + UnencodedRunStart, EnStr.length() - UnencodedRunStart);
		} else
			Decoded = decode(EnStr.toString(), Buff);
		
		return Decoded;
	}

/**	Decodes a string, represented as a {@link Characters} object, by translating HTML character entity references into 
 *	corresponding Unicode characters.
 * 
 * @param EnStr					The string to be decoded.
 * @param Buff					A <code>StringBuilder</code> to which the output of this method is to be appended. The
 *								buffer is <b>not</b> cleared.
 * @return						<code>true</code> if any characters were decoded, <code>false</code> if no characters
 *								were decoded.
 */
	private boolean decodeLow
	(
		final Characters		EnStr,
		final StringBuilder		Buff
	)
	{
		boolean					Decoded = true;
	
		if (EnStr.hasArray()) {
			final char[]						Chars             = EnStr.getArray();
			final int							CharsStartOffset  = EnStr.getArrayOffset();
			int									UnencodedRunStart = 0;
			int									AmpIdx            = EnStr.indexOf('&');
			final HashMap<String, HTMLEntity>	Entities          = this.Entities;
		
			while (AmpIdx >= 0) {
				final int				ScIdx = EnStr.indexOf(';', AmpIdx + 1);
			
				if (ScIdx < 0)
					break;
			
				final String			PossibleEntityName = EnStr.substring(AmpIdx, ScIdx + 1);	//	Isolate entity name with leading '&' and trailing ';'
				int						DecodedCodePoint   = 0;
				boolean					DecodedThisChar    = false;
				
			//	See <http://www.w3.org/TR/html4/charset.html#h-5.3.1>.
				
				if (PossibleEntityName.charAt(1) == '#') {
					final char				Char2 = PossibleEntityName.charAt(2);
					
					if (Char2 == 'x' || Char2 == 'X') {
						
						DecodedCodePoint = Integer.parseInt(PossibleEntityName.substring(3, PossibleEntityName.length() - 1), 16);
						DecodedThisChar  = true;
						
					} else {
						
						DecodedCodePoint = Integer.parseInt(PossibleEntityName.substring(2, PossibleEntityName.length() - 1));
						DecodedThisChar  = true;
					}
				
				} else {
					final HTMLEntity		Map = Entities.get(PossibleEntityName);
				
					if (Map != null) {
					
						DecodedCodePoint = Map.UnicodeCodePoint;
						DecodedThisChar  = true;
					}
				}
		
				if (DecodedThisChar) {
				
					if (UnencodedRunStart < AmpIdx)
						Buff.append(Chars, CharsStartOffset + UnencodedRunStart, AmpIdx - UnencodedRunStart);
				
					Buff.appendCodePoint(DecodedCodePoint);
				
					UnencodedRunStart = ScIdx + 1;
				}
			
			//	Next iteration.
			
				AmpIdx = EnStr.indexOf('&', ScIdx + 1);
			}
		
			if (UnencodedRunStart == 0) {
			
				EnStr.append(Buff);
				Decoded = false;
			
			} else
				Buff.append(Chars, CharsStartOffset + UnencodedRunStart, EnStr.length() - UnencodedRunStart);
		} else
			Decoded = decode(EnStr.toString(), Buff);
	
		return Decoded;
	}
	
	private String getMultiByteMapping
	(
		final int				Char
	)
	{
	
	//	Perform a binary search for a mapping for Char in the portion of the Mappings array that deals with multi-byte 
	//	characters.
		
		final HTMLEntity[]		Mappings = HTMLEntityCodec.Mappings;
		int						MinIndex = FirstMultiByteMappingIndex;
		int						MaxIndex = Mappings.length - 1;
		
		if (Char < Mappings[MinIndex].UnicodeCodePoint || Char > Mappings[MaxIndex].UnicodeCodePoint)
			return null;
		
		int						ElementsRemaining = (MaxIndex - MinIndex) + 1;
		
		do {
			final int				Index       = MinIndex + (ElementsRemaining >> 1);
			final HTMLEntity		Map         = Mappings[Index];
			final int				MapCharCode = Map.UnicodeCodePoint;
			
			if (MapCharCode == Char)
				return Map.Markup;
			
			if (MapCharCode < Char)
				MinIndex = Index + 1;
			else
				MaxIndex = Index - 1;
			
			ElementsRemaining = (MaxIndex - MinIndex) + 1;
		
		} while (ElementsRemaining > 0);
		
		return null;
	}
}
