/*	QWICAP (Quick Web Interface for Conventional Applications) by Chris W. Johnson
	Copyright (C) 2003-2007 The University of Texas at Austin ("U.T. Austin"). All rights reserved.
	
	U.T. Austin's copyright notice in this evolving version reflects our original authorship. It does not assert 
	copyright over changes contributed by others.
	
	This library is free software; you can redistribute it and/or modify it under the terms of the GNU Lesser General 
	Public License as published by the Free Software Foundation; either version 2.1 of the License, or (at your option) 
	any later version.
	
	This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied 
	warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for 
	more details.
	
	You should have received a copy of the GNU Lesser General Public License along with this library; if not, write to 
	the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/


package edu.utexas.its.eis.tools.qwicap.util;


/**	This class provides a set of static utility methods that perform common tasks associated with the handling of
 *	<code>null</code> values.
 * 	
 * 	@author Chris W. Johnson
 */
public final class Null {
	
	private Null() { }

/**	Returns <code>true</code> if passed <code>null</code>, or an empty (zero length) <code>String</code> object; returns
 * 	<code>false</code> otherwise.
 * 
 * @param PotentiallyNullOrEmptyStr		<code>null</code> or a <code>String</code> object.
 * @return								<code>true</code> if passed <code>null</code>, or an empty (zero length) 
 * 										<code>String</code> object; returns <code>false</code> otherwise.
 */
	public static boolean orEmpty
	(
		final String			PotentiallyNullOrEmptyStr
	)
	{
		return PotentiallyNullOrEmptyStr == null || PotentiallyNullOrEmptyStr.length() == 0;
	}
	
/**	Returns <code>null</code> if passed an empty (zero-length) <code>String</code> (or <code>null</code>), otherwise 
 *	it returns the same <code>String</code> object that was passed to it.
 * 
 * @param PotentiallyEmptyStr	A <code>String</code> that may be empty, non-empty, or <code>null</code>.
 * @return						<code>null</code> if the <code>String</code> was empty or <code>null</code>, otherwise 
 *								the <code>String</code> object that was passed-in.
 */
	public static String fromEmptyString
	(
		final String			PotentiallyEmptyStr
	)
	{
		return fromSomeString(PotentiallyEmptyStr, "");
	}

/**	If the supplied string literally reads "null", in other words it consists of the characters 'n', 'u', 'l', 'l' and
 * 	no others, the value <code>null</code> is returned. Otherwise, the supplied string is returned. 
 * 
 * @param PotentiallyNullStr	Any <code>String</code> object, or <code>null</code>. 
 * @return						<code>null</code> if the characters in the string spell the word "null", otherwise the
 * 								value passed-in as <code>PotentiallyNullStr</code> is returned.
 */
	public static String fromNullString
	(
		final String			PotentiallyNullStr
	)
	{
		return fromSomeString(PotentiallyNullStr, "null");
	}

/**	If the <code>PotentiallyRepresentingNullStr</code> parameter is a string that matches the string supplied in the
 * 	<code>NullRepresentationStr</code> parameter, the value <code>null</code> is returned. Otherwise, the value of the
 * 	<code>PotentiallyRepresentingNullStr</code> parameter is returned.
 * 
 * @param PotentiallyRepresentingNullStr	Any <code>String</code> object, or <code>null</code>.
 * @param NullRepresentationStr				The string that should be considered to represent a null value.
 * @return									<code>null</code>, or the value passed-in as 
 * 											<code>PotentiallyRepresentingNullStr</code>.
 */
	public static String fromSomeString
	(
		final String			PotentiallyRepresentingNullStr,
		final String			NullRepresentationStr
	)
	{
		return safeEquals(NullRepresentationStr, PotentiallyRepresentingNullStr) ? null : PotentiallyRepresentingNullStr;
	}
	
/**	Returns an empty (zero-length) <code>String</code> if passed <code>null</code>, otherwise it returns the object 
 * 	reference that was passed-in. 
 * 
 * @param PotentiallyNullObject	Any object reference, or <code>null</code>.
 * @return						An empty (zero-length) <code>String</code> if this method was passed <code>null</code>,
 *								otherwise the object reference that was passed-in.
 */
	public static String toEmptyString
	(
		final Object			PotentiallyNullObject
	)
	{
		return toSomeString(PotentiallyNullObject, "");
	}

/**	Returns a string reading "null", if the supplied object reference is <code>null</code>. Otherwise, the supplied 
 * 	object is converted to a <code>String</code> by invoking its {@link Object#toString} method, and that 
 * 	<code>String</code> is returned.
 * 
 * @param PotentiallyNullObject	Any object reference, or <code>null</code>.
 * @return						A <code>String</code> object containing the word "null", or the <code>String</code>
 * 								returned by invoking <code>PotentiallyNullObject.toString()</code>.
 */
	public static String toNullString
	(
		final Object			PotentiallyNullObject
	)
	{
		return toSomeString(PotentiallyNullObject, "null");
	}
	
/**	Returns the supplied <code>NullSubstituteStr</code>, if the object reference supplied in 
 * 	<code>PotentiallyNullObject</code> is <code>null</code>. Otherwise, <code>PotentiallyNullObject</code>
 * 	is converted to a <code>String</code> by invoking its {@link Object#toString} method, and that 
 * 	<code>String</code> is returned.
 * 
 * @param PotentiallyNullObject	Any object reference, or <code>null</code>.
 * @param NullSubstituteString	The <code>String</code> to returned when <code>PotentiallyNullObject</code> is 
 * 								<code>null</code>.
 * @return						The <code>String</code> object passed-in as the <code>PotentiallyNullObject</code>
 * 								parameter, or the <code>String</code> returned by invoking 
 * 								<code>PotentiallyNullObject.toString()</code>.
 */
	public static String toSomeString
	(
		final Object			PotentiallyNullObject,
		final String			NullSubstituteString
	)
	{
		return PotentiallyNullObject == null ? NullSubstituteString : PotentiallyNullObject.toString();
	}
	
/**	Safely compares two objects for equality, where one, both, or none of the objects may be <code>null</code>.	
 * 
 * @param A		The first object to be used in the comparison. May be <code>null</code>.
 * @param B		The second object to be used in the comparison. May be <code>null</code>.
 * @return		<code>true</code> if the parameters are equal, either by reference or value, otherwise <code>false</code>.
 */	
	public static boolean safeEquals
	(
		final Object			A,
		final Object			B
	)
	{
		if (A == B)
			return true;
		
		if (A == null || B == null)
			return false;
			
		return A.equals(B);
	}	
}
