/*	QWICAP (Quick Web Interface for Conventional Applications) by Chris W. Johnson
	Copyright (C) 2008 Chris W. Johnson. All rights reserved.
	
	Chris W. Johnson's copyright notice in this evolving version reflects his original authorship. It does not assert 
	copyright over changes contributed by others.
	
	This library is free software; you can redistribute it and/or modify it under the terms of the GNU Lesser General 
	Public License as published by the Free Software Foundation; either version 2.1 of the License, or (at your option) 
	any later version.
	
	This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied 
	warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for 
	more details.
	
	You should have received a copy of the GNU Lesser General Public License along with this library; if not, write to 
	the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/


package edu.utexas.its.eis.tools.qwicap.util;


import edu.utexas.its.eis.tools.qwicap.servlet.i18nServices;
import edu.utexas.its.eis.tools.qwicap.servlet.Qwicap;
import edu.utexas.its.eis.tools.qwicap.template.xml.Match;

import java.sql.Time;
import java.util.Date;
import java.util.Calendar;
import java.util.Collection;
import java.util.Iterator;
import java.util.logging.Logger;
import java.util.logging.Level;
import java.lang.reflect.Array;


/**	Converts a number of common types of objects into <code>String</code> objects better than their 
 * 	<code>toString</code> methods would ordinarily do so. <b>This class is not part of Qwicap's public API, but could 
 * 	not be hidden due to limitations in Java's access control scheme.</b>
 * 
 * 	@author Chris W. Johnson
 */
public class SmartStringConverter {
	
	private static final Logger	Log = Logger.getLogger(SmartStringConverter.class.getName());
	
	private static Boolean		CanRequestI18NServices;
	private boolean				i18nServicesRequested;	//	This must NOT be static, because it describes the state of the i18n instance variable.
	private i18nServices		i18n;	//	This must NOT be static - i18nServices instances will vary depending on a user's language preferences.
	
	
	public SmartStringConverter() { }
	
	public SmartStringConverter
	(
		final i18nServices		i18n
	)
	{
		this.i18n = i18n;
	}
	
	public String toString
	(
		final Object			Thing
	)
	{
		return appendLow(new StringBuilder(32), Thing, 0).toString();
	}
	
	public StringBuilder append
	(
		final StringBuilder		Buff,
		final Object			Thing
	)
	{
		return appendLow(Buff, Thing, 0);
	}
	
	public StringBuilder appendLow
	(
		final StringBuilder		Buff,
		final Object			ElemObj,
		final int				ListLevel
	)
	{
	//	todo Consider whether we can replace the Collection and array handling below with the ArrayToString class.

		if (ElemObj == null) {

			Buff.append("null");
			return Buff;
		}
		
		if (ElemObj instanceof Match) {
			
			Buff.append(((Match) ElemObj).toString(false));
			
		} else if (ElemObj instanceof Number) {
			
			if (ElemObj instanceof Byte) {
				
				Buff.append(((Byte) ElemObj).intValue() & 0xFF);	//	Write bytes as unsigned values.
				
			} else {
				final i18nServices		i18n = geti18nServices();
				
				if (i18n != null)
					Buff.append(i18n.formatNumber((Number) ElemObj));
				else
					Buff.append(ElemObj);
			}
			
		} else if (ElemObj instanceof java.sql.Time) {
			final i18nServices		i18n = geti18nServices();
			
			if (i18n != null)
				Buff.append(i18n.formatTime((Time) ElemObj));
			else
				Buff.append(ElemObj);
			
		} else if (ElemObj instanceof java.sql.Date) {
			final i18nServices		i18n = geti18nServices();
			
			if (i18n != null)
				Buff.append(i18n.formatDate((Date) ElemObj));
			else
				Buff.append(ElemObj);
			
		} else if (ElemObj instanceof java.util.Date) {
			final i18nServices		i18n = geti18nServices();
			
			if (i18n != null)
				Buff.append(i18n.formatDateTime((Date) ElemObj));
			else
				Buff.append(ElemObj);
			
		} else if (ElemObj instanceof Calendar) {
			final i18nServices		i18n = geti18nServices();
			
			if (i18n != null)
				Buff.append(i18n.formatDateTime((Calendar) ElemObj));
			else
				Buff.append(ElemObj);
			
		} else if (ElemObj instanceof Throwable) {
			
			Buff.append(StackTrace.toString((Throwable) ElemObj));
			
		} else if (ElemObj instanceof Collection) {
			final Iterator<?> Elems     = ((Collection<?>) ElemObj).iterator();
			int						ElemCount = 0;

			if (ListLevel > 0)
				Buff.append("[ ");

			while (Elems.hasNext()) {

				if (ElemCount++ > 0)
					Buff.append(", ");

				appendLow(Buff, Elems.next(), ListLevel + 1);
			}

			if (ListLevel > 0)
				Buff.append(']');

		} else if (ElemObj.getClass().isArray()) {

			if (ListLevel > 0)
				Buff.append("[ ");

			for (int Index = 0, Count = Array.getLength(ElemObj); Index < Count; Index++) {

				if (Index > 0)
					Buff.append(", ");

				appendLow(Buff, Array.get(ElemObj, Index), ListLevel + 1);
			}

			if (ListLevel > 0)
				Buff.append(']');

		} else
			Buff.append(ElemObj);
		
		return Buff;
	}

//	Note that the geti18nServices method is present in both the SmartStringConverter and Results classes, and the 
//	implementations should be kept identical. 

	private i18nServices geti18nServices() {
		
		if (i18nServicesRequested)
			return i18n;
		
		if (CanRequestI18NServices == Boolean.FALSE)
			return null;
		
		CanRequestI18NServices = Boolean.FALSE;
		
		try {
			
		//	If Class.forName fails, it throws ClassNotFoundException, LinkageError, or ExceptionInInitializerError.
		//	We perform this method invocation on the assumption that a ClassNotFoundException has less impact on the 
		//	JVM than the NoClassDefFoundError potentially produced by the Qwicap.getCurrentInstance invocation below.
			
			Class.forName("javax.servlet.Servlet");
			
		//	If we've arrived here, the javax.servlet.Servlet class must be available (in some form), so the invocation 
		//	of Qwicap.getCurrentInstance should not cause any class loading problems. That's not to say that it will
		//	succeed, but it won't fail due to a problem that will necessarily persist.
			
			CanRequestI18NServices = Boolean.TRUE;
			
		//	If Qwicap.getCurrentInstance fails because some of the classes it references are missing (typically the 
		//	javax.servlet.* classes), a NoClassDefFoundError is thrown. Anyway, that has been reported in the past.
		//	In current tests, when this code is run outside of a Qwicap web application, Qwicap.getCurrentInstance 
		//	throws an IllegalStateException because there is no current instance of Qwicap. Both of these cases, and
		//	many others, are handled by the catch blocks below.
			
		//	Note that we pass "Qwicap.class" below, because, in order to simplify interationalization, we keep all of our 
		//	internationalizable material in a single file, associated with the Qwicap class.
			
			i18nServicesRequested = true;
			i18n = Qwicap.getCurrentInstance().geti18nServicesForClass(Qwicap.class);
			
		} catch (NoClassDefFoundError e) {
			
		//	This happens when only the XML engine is being used, and therefore the javax.servlet library is not 
		//	present, which the JVM notices when it loads the Qwicap class.
			
			CanRequestI18NServices = Boolean.FALSE;
			
			Log.log(Level.FINE, "Cannot get internationalization services. Will continue without them. ({0} message: \"{1}\")", new Object[] { e.getClass().getName(), e.getMessage() });
			
		} catch (Exception e) {
			
			Log.log(Level.FINE, "Cannot get internationalization services. Will continue without them. ({0} message: \"{1}\")", new Object[] { e.getClass().getName(), e.getMessage() });
		}
		
		return i18n;
	}
}
