/*	QWICAP (Quick Web Interface for Conventional Applications) by Chris W. Johnson
	Copyright (C) 2003-2011 The University of Texas at Austin ("U.T. Austin"). All rights reserved.
	
	U.T. Austin's copyright notice in this evolving version reflects our original authorship. It does not assert 
	copyright over changes contributed by others.
	
	This library is free software; you can redistribute it and/or modify it under the terms of the GNU Lesser General 
	Public License as published by the Free Software Foundation; either version 2.1 of the License, or (at your option) 
	any later version.
	
	This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied 
	warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for 
	more details.
	
	You should have received a copy of the GNU Lesser General Public License along with this library; if not, write to 
	the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/


package edu.utexas.ots.chrisj.util;


import java.util.Arrays;
import java.io.*;


/**	<code>Hexadecimal</code> provides a suite of static methods for conveniently converting values to and from 
 * 	hexadecimal character sequences.
 *	
 *	@author		Chris W. Johnson
 *	@version	1.2.7
 */
public final class Hexadecimal {
	
	private static final int		kDefaultBytesPerRow    = 16;
	private static final String		kDefaultLinePrefixStr  = null;
	private static final String		kDefaultIntraHexSepStr = " ";
	private static final String		kDefaultHexASCIISepStr = "    ";
	private static final String		kDefaultLineSepStr     = null;
	
	private static final char[]		HexNibbles = { '0', '1', '2', '3', '4', '5', '6', '7', '8', '9', 'A', 'B', 'C', 'D', 'E', 'F' };
	private static final char[][]	HexBytes   = new char[256][2];

	static {

	//	Build the contents of the HexBytes array.

		int						Byte = 0;

		for (int HiNibble = 0; HiNibble < 16; HiNibble++) {
			final char				HiHex = HexNibbles[HiNibble];

			for (int LoNibble = 0; LoNibble < 16; LoNibble++) {
				final char[]			Hex = new char[2];
				
				Hex[0] = HiHex;
				Hex[1] = HexNibbles[LoNibble];

				HexBytes[Byte++] = Hex;
			}
		}
	}

	private static final int[]	NibbleValues = new int['f' + 1];	//	'f' is the highest character code we'll ever need to look-up.

	static {

	//	Build the contents of the NibbleValues array.

		Arrays.fill(NibbleValues, -1);	//	Make it easy to test fro invalid values.

		for (int Index = '0', Value = 0; Index <= '9'; Index++, Value++)
			NibbleValues[Index] = Value;

		for (int Index = 'A', Value = 10; Index <= 'F'; Index++, Value++)
			NibbleValues[Index] = Value;

		for (int Index = 'a', Value = 10; Index <= 'f'; Index++, Value++)
			NibbleValues[Index] = Value;
	}


/**	Evaluates a <code>String</code> to determine whether its contents can be interpreted as a hexadecimal number. 
 * 	Specifically, all of the characters must fall into one the following ranges: 0-9, A-F, or a-f.
 * 
 * @param Hex					The sequence to be evaluated.
 * 
 * @return						<code>true</code> if the sequence can be interpreted as a hexadecimal number, 
 *								<code>false</code> otherwise.
 */
	public static boolean isHex
	(
		final String			Hex
	)
	{
		return isHex(Hex, 0, Hex.length());
	}

/**	Evaluates a subsection of a <code>String</code> to determine whether its contents can be interpreted as a 
 * 	hexadecimal number. Specifically, all of the characters must fall into one the following ranges: 0-9, A-F, or a-f.
 * 
 * @param Hex					The sequence to be evaluated.
 * @param Offset				The offset of the first character to be evaluated.
 * @param Count					The number of characters to be evaluated.
 * 
 * @return						<code>true</code> if the sequence can be interpreted as a hexadecimal number, 
 *								<code>false</code> otherwise.
 */
	public static boolean isHex
	(
		final String			Hex,
		final int				Offset,
		int						Count
	)
	{
		final char[]			Chars     = Hex.toCharArray();
		final int				MaxNibble = NibbleValues.length;

		for (int Index = Offset; Count > 0; Index++, Count--) {
			final int				Char = Chars[Index];

			if (Char >= MaxNibble || NibbleValues[Char] < 0)
				return false;
		}

		return true;
	}

/**	Evaluates a subsection of a <code>String</code> to determine whether its contents can be interpreted as a 
 * 	hexadecimal number. Specifically, all of the characters must fall into one the following ranges: 0-9, A-F, or a-f. 
 * 	If <code>WhitespacePermitted</code> is <code>true</code>, whitespace may be scattered anywhere amongst the 
 * 	hexadecimal characters.
 * 
 *	@param Hex					The sequence to be evaluated.
 *	@param Offset				The offset of the first character to be evaluated.
 *	@param Count					The number of characters to be evaluated.
 *	@param WhitespacePermitted	<code>true</code> if whitespace between hexadecimal characters is permitted, 
 * 								<code>false</code> otherwise.
 * 
 *	@return						<code>true</code> if the sequence can be interpreted as a hexadecimal number, 
 *								<code>false</code> otherwise.
 */
	public static boolean isHex
	(
		final String			Hex,
		final int				Offset,
		int						Count,
		final boolean			WhitespacePermitted
	)
	{
		final char[]			Chars        = Hex.toCharArray();
		final int				MaxNibble    = NibbleValues.length;
		int						HexCharCount = 0;

		for (int Index = Offset; Count > 0; Index++, Count--) {
			final int				Char = Chars[Index];

			if (WhitespacePermitted && Character.isWhitespace(Char))
				continue;

			if (Char >= MaxNibble || NibbleValues[Char] < 0)
				return false;

			HexCharCount++;
		}

		return HexCharCount > 0;
	}

/** Converts a sequence of hexadecimal digits to a <code>byte</code>.
 *
 *	@param Hex					Sequence of hexadecimal digits.
 * 
 *	@return						<code>byte</code> value.
 *
 *	@throws NumberFormatException	If the sequence could not be interpreted as a hexadecimal value.
 */
	public static byte toByte
	(
		final String			Hex
	)
	throws NumberFormatException
	{
		return (byte) toNumber(Hex, Byte.MIN_VALUE, Byte.MAX_VALUE, "Byte");
	}

/** Converts a sequence of hexadecimal digits to a <code>byte</code> array. If the number of hexadecimal digits in the 
 *	sequence is odd, the high-nibble of the first byte of the array will be zero.
 *
 *	@param Hex						Sequence of hexadecimal digits.
 * 	@param WhitespacePermitted		<code>true</code> if whitespace scattered amongst the hexadecimal digits should be
 * 									ignored, <code>false</code> if the presence of whitespace should be considered an 
 * 									error (<code>NumberFormatException</code> will be thrown in this case).
 *	@return							<code>byte</code> array.
 *	
 *	@throws NumberFormatException	If the sequence could not be interpreted as a hexadecimal value.
 */
	public static byte[] toBytes
	(
		final String			Hex,
		final boolean			WhitespacePermitted
	)
	throws NumberFormatException
	{
		final char[]			HexChars    = Hex.toCharArray();
		final int				HexCount    = HexChars.length;
		final int				ByteCount   = HexCount + 1 >> 1;
		int						ByteIndex   = 0;
		final byte[]			Bytes       = new byte[ByteCount];
		int						Byte        = 0;
		int						NibbleCount = 0;

		for (int HexIndex = 0; HexIndex < HexCount; HexIndex++) {
			final char				Char = HexChars[HexIndex];

			if (WhitespacePermitted && Character.isWhitespace(Char))
				continue;

			if (NibbleCount++ == 0) {

				Byte = toNibble(Char);

			} else {

				Bytes[ByteIndex++] = (byte) (Byte << 4 | toNibble(Char));
				NibbleCount = 0;
			}
		}

	//	If we acquired an odd number of nibbles, right shift all of the bits in the byte array by four bits, so that the
	//	"padding" zero bits end-up in the high-order nibble of the first byte, rather than the low-order nibble of the 
	//	last byte.

		if (NibbleCount != 0) {

		//	Add the final nibble to the byte array, putting the final nibble in the high-order nibble of the last byte.

			Bytes[ByteIndex++] = (byte) (Byte << 4);

		//	Right shift all the bits in the array by 4.

			int						CarryNibble = 0;

			for (int Index = 0; Index < ByteIndex; Index++) {

				Byte         = Bytes[Index] & 0xFF;
				Bytes[Index] = (byte) (CarryNibble << 4 | Byte >>> 4);

				CarryNibble  = Byte & 0x0F;
			}
		}

	//	The Bytes array was sized for the worst-case number of bytes we'd receive. If we actually received that many,
	//	we can return Bytes and declare victory.

		if (ByteIndex == ByteCount)
			return Bytes;

	//	If we received fewer bytes than the worst-case number for which the Bytes array was sized, then return a new
	//	array representing only the subsection of Bytes into which we placed data.

		final byte[]			FinalBytes = new byte[ByteIndex];
		
		System.arraycopy(Bytes, 0, FinalBytes, 0, ByteIndex);

		return FinalBytes;
	}

/** Converts a sequence of hexadecimal digits to a <code>byte</code> array. If the number of hexadecimal digits in the 
 *	sequence is odd, the high-nibble of the first byte of the array will be zero.
 *   
 * 	@param Hex					Sequence of hexadecimal digits.
 * 
 *	@return						<code>byte</code> array.
 *	
 *	@throws NumberFormatException	If the sequence could not be interpreted as a hexadecimal value.
 */
	public static byte[] toBytes
	(
		final String			Hex
	)
	throws NumberFormatException
	{
		final char[]			HexChars  = Hex.toCharArray();
		final int				HexCount  = HexChars.length;
		int						HexIndex  = 0;
		final int				ByteCount = HexCount + 1 >> 1;
		int						ByteIndex = 0;
		final byte[]			Bytes     = new byte[ByteCount];

		if (HexCount % 2 != 0)
			Bytes[ByteIndex++] = (byte) toNibble(HexChars[HexIndex++]);

		while (HexIndex < HexCount)
			Bytes[ByteIndex++] = (byte) (toNibble(HexChars[HexIndex++]) << 4 | toNibble(HexChars[HexIndex++]));

		return Bytes;
	}

/** Converts a sequence of hexadecimal digits to a <code>short</code>.
 *
 *	@param Hex					Sequence of hexadecimal digits.
 * 
 *	@return						<code>short</code> value.
 *
 *	@throws NumberFormatException	If the sequence could not be interpreted as a hexadecimal value.
 */
	public static short toShort
	(
		final String			Hex
	)
	throws NumberFormatException
	{
		return (short) toNumber(Hex, Short.MIN_VALUE, Short.MAX_VALUE, "Short");
	}

/** Converts a sequence of hexadecimal digits to a <code>char</code>.
 *
 *	@param Hex					Sequence of hexadecimal digits.
 * 
 *	@return						<code>char</code> value.
 *
 *	@throws NumberFormatException	If the sequence could not be interpreted as a hexadecimal value.
 */
	public static char toChar
	(
		final String			Hex
	)
	throws NumberFormatException
	{
		return (char) toNumber(Hex, Character.MIN_VALUE, Character.MAX_VALUE, "Character");
	}

/** Converts a sequence of hexadecimal digits to an <code>int</code>.
 *
 *	@param Hex					Sequence of hexadecimal digits.
 * 
 *	@return						<code>int</code> value.
 *
 *	@throws NumberFormatException	If the sequence could not be interpreted as a hexadecimal value.
 */
	public static int toInt
	(
		final String			Hex
	)
	throws NumberFormatException
	{
		return (int) toNumber(Hex, Integer.MIN_VALUE, Integer.MAX_VALUE, "Integer");
	}

/** Converts a sequence of hexadecimal digits to a <code>long</code>.
 *
 *	@param Hex					Sequence of hexadecimal digits.
 * 
 *	@return						<code>long</code> value.
 *
 *	@throws NumberFormatException	If the sequence could not be interpreted as a hexadecimal value.
 */
	public static long toLong
	(
		final String			Hex
	)
	throws NumberFormatException
	{
		return toNumber(Hex, Long.MIN_VALUE, Long.MAX_VALUE, "Long");
	}

/** Converts a string of hexadecimal digits to a <code>long</code> confined to a specified range of values.
 *
 *	@param Hex					Sequence of hexadecimal digits.
 *	@param MinValue				Minimum permissible value.
 *	@param MaxValue				Maximum permissible value.
 *	@param TypeName				Name of the datatype whose range limits are described by <code>MinValue</code> and 
 *								<code>MaxValue</code>. This is used to make more informative certain exceptions we 
 * 								throw, and serves no other purpose.
 * 
 *	@return						<code>long</code> value.
 *
 *	@throws NumberFormatException	If the string could not be interpreted as a hexadecimal value within the specified 
 * 									range.
 */
	private static long toNumber
	(
		final String			Hex,
		final long				MinValue,
		final long				MaxValue,
		final String			TypeName
	)
	throws NumberFormatException
	{
		final int				Length = Hex.length();

		if (Length == 0)
			throw new NumberFormatException("The hexadecimal sequence is empty (zero-length).");

		if (Length > 16)
			throw new NumberFormatException("The hexadecimal sequence \"" + Hex + "\" is too long to be converted into a long integer.");

		long					Long     = 0;
		final char[]			HexChars = Hex.toCharArray();

		for (int Index = 0; Index < Length; Index++)
			Long = Long << 4 | toNibble(HexChars[Index]);

		if (Long < MinValue || Long > MaxValue)
			throw new NumberFormatException("The value " + Long + ", from the hexadecimal sequence \"" + Hex + "\", is beyond the range of the " + TypeName + " data type.");

		return Long;
	}

/** Converts a hexadecimal character to an <code>int</code>.
 *
 *	@param Hex					Single hexadecimal character represented as an <code>char</code>.
 * 
 *	@return						Value of the hexadecimal character represented as an <code>int</code>.
 *
 *	@throws NumberFormatException	If the string could not be interpreted as a hexadecimal digit.
 */
	public static int toNibble
	(
		final char				Hex
	)
	throws NumberFormatException
	{
		return toNibble((int) Hex);
	}

/** Converts a hexadecimal character to an <code>int</code>.
 *
 *	@param Hex					Single hexadecimal character represented as an <code>int</code>.
 * 
 *	@return						Value of the hexadecimal character represented as an <code>int</code>.
 *
 *	@throws NumberFormatException	If the string could not be interpreted as a hexadecimal digit.
 */
	public static int toNibble
	(
		final int				Hex
	)
	throws NumberFormatException
	{
		if (Hex >= NibbleValues.length)
			throw new NumberFormatException("'" + (char) Hex + "' is not a hexadecimal character.");

		final int				Value = NibbleValues[Hex];

		if (Value < 0)
			throw new NumberFormatException("'" + (char) Hex + "' is not a hexadecimal character.");

		return Value;
	}

/**	Converts an 8-character-long hexadecimal sequence to a <code>double</code>.
 * 
 *	@param Hex					An 8 character hexadecimal sequence.
 * 
 *	@return						<code>float</code> value.
 * 
 *	@throws NumberFormatException	If the hexadecimal sequence could not be interpreted as a <code>float</code>.
 */
	public static float toFloat
	(
		final String			Hex
	)
	throws NumberFormatException
	{
		final ByteArrayInputStream	BIn = new ByteArrayInputStream(toBytes(Hex));
		final DataInputStream		DIn = new DataInputStream(BIn);
		
		try {
			final float				Value = DIn.readFloat();
			
			DIn.close();
			
			return Value;
		
		} catch (IOException e) {
			throw (NumberFormatException) new NumberFormatException("The hexadecimal sequence \"" + Hex + "\" could not be interpreted as a float.").initCause(e);
		}
	}

/**	Converts a 16-character-long hexadecimal sequence to a <code>double</code>.
 * 
 *	@param Hex					A 16 character hexadecimal sequence.
 * 
 *	@return						<code>double</code> value.
 * 
 *	@throws NumberFormatException	If the hexadecimal sequence could not be interpreted as a <code>double</code>.
 */
	public static double toDouble
	(
		final String			Hex
	)
	throws NumberFormatException
	{
		final ByteArrayInputStream	BIn = new ByteArrayInputStream(toBytes(Hex));
		final DataInputStream		DIn = new DataInputStream(BIn);
		
		try {
			final double			Value = DIn.readDouble();
			
			DIn.close();
			
			return Value;
		
		} catch (IOException e) {
			throw (NumberFormatException) new NumberFormatException("The hexadecimal sequence \"" + Hex + "\" could not be interpreted as a double.").initCause(e);
		}
	}

/** Converts a <code>float</code> to an 8-character-long hexadecimal sequence.
 *
 *	@param Value				<code>float</code> value.
 * 
 *	@return						Hexadecimal sequence.
 */
	public static String fromFloat
	(
		final float				Value
	)
	{
		final ByteArrayOutputStream	BOut = new ByteArrayOutputStream(8);
		
		try {
			final DataOutputStream		DOut = new DataOutputStream(BOut);
			
			DOut.writeFloat(Value);
			DOut.close();
			
		} catch (IOException e) {
		//	An IOException should be impossible when writing to a ByteArrayOutputStream.
		}
		
		return fromBytes(BOut.toByteArray());
	}

/** Converts a <code>double</code> to a 16-character-long hexadecimal sequence.
 *
 *	@param Value				<code>double</code> value.
 * 
 *	@return						Hexadecimal sequence.
 */
	public static String fromDouble
	(
		final double			Value
	)
	{
		final ByteArrayOutputStream	BOut = new ByteArrayOutputStream(8);
		
		try {
			final DataOutputStream		DOut = new DataOutputStream(BOut);
			
			DOut.writeDouble(Value);
			DOut.close();
			
		} catch (IOException e) {
		//	An IOException should be impossible when writing to a ByteArrayOutputStream.
		}
		
		return fromBytes(BOut.toByteArray());
	}

/** Converts a <code>long</code> to a 16-character-long hexadecimal sequence.
 *
 *	@param Value				<code>long</code> value.
 * 
 *	@return						Hexadecimal sequence.
 */
	public static String fromLong
	(
		long					Value
	)
	{
		final char[]			HexBuff = new char[16];
		char[]					Hex;

		Hex         = HexBytes[(int) Value & 0xFF];
		HexBuff[14] = Hex[0];
		HexBuff[15] = Hex[1];

		Value >>= 8;

		Hex         = HexBytes[(int) Value & 0xFF];
		HexBuff[12] = Hex[0];
		HexBuff[13] = Hex[1];

		Value >>= 8;

		Hex         = HexBytes[(int) Value & 0xFF];
		HexBuff[10] = Hex[0];
		HexBuff[11] = Hex[1];

		Value >>= 8;

		Hex        = HexBytes[(int) Value & 0xFF];
		HexBuff[8] = Hex[0];
		HexBuff[9] = Hex[1];

		Value >>= 8;

		Hex        = HexBytes[(int) Value & 0xFF];
		HexBuff[6] = Hex[0];
		HexBuff[7] = Hex[1];

		Value >>= 8;

		Hex        = HexBytes[(int) Value & 0xFF];
		HexBuff[4] = Hex[0];
		HexBuff[5] = Hex[1];

		Value >>= 8;

		Hex        = HexBytes[(int) Value & 0xFF];
		HexBuff[2] = Hex[0];
		HexBuff[3] = Hex[1];

		Value >>= 8;

		Hex        = HexBytes[(int) Value & 0xFF];
		HexBuff[0] = Hex[0];
		HexBuff[1] = Hex[1];

		return new String(HexBuff);
	}

/** Converts an <code>int</code> to an 8-character-long hexadecimal sequence.
 *
 *	@param Value				<code>int</code> value.
 * 
 *	@return						Hexadecimal sequence.
 */
	public static String fromInt
	(
		int						Value
	)
	{
		final char[]			HexBuff = new char[8];
		char[]					Hex;

		Hex        = HexBytes[Value & 0xFF];
		HexBuff[6] = Hex[0];
		HexBuff[7] = Hex[1];

		Value >>= 8;

		Hex        = HexBytes[Value & 0xFF];
		HexBuff[4] = Hex[0];
		HexBuff[5] = Hex[1];

		Value >>= 8;

		Hex        = HexBytes[Value & 0xFF];
		HexBuff[2] = Hex[0];
		HexBuff[3] = Hex[1];

		Value >>= 8;

		Hex        = HexBytes[Value & 0xFF];
		HexBuff[0] = Hex[0];
		HexBuff[1] = Hex[1];

		return new String(HexBuff);
	}

/** Converts a <code>short</code> to a 4-character-long hexadecimal sequence.
 *
 *	@param Value				<code>short</code> value.
 * 
 *	@return						Hexadecimal sequence.
 */
	public static String fromShort
	(
		short					Value
	)
	{
		final char[]			HexBuff = new char[4];
		char[]					Hex;

		Hex        = HexBytes[Value & 0xFF];
		HexBuff[2] = Hex[0];
		HexBuff[3] = Hex[1];

		Value >>= 8;

		Hex        = HexBytes[Value & 0xFF];
		HexBuff[0] = Hex[0];
		HexBuff[1] = Hex[1];

		return new String(HexBuff);
	}

/** Converts a <code>char</code> to a 4-character-long hexadecimal sequence.
 *
 *	@param Value				<code>char</code> value.
 * 
 *	@return						Hexadecimal sequence.
 */
	public static String fromChar
	(
		final char				Value
	)
	{
		return fromShort((short) Value);	//	"short" and "char" are both 16 bits, so this works.
	}

/** Converts the low-order nibble of an <code>int</code> into a single hexadecimal character.
 *
 *	@param Nibble				Nibble value.
 *
 *	@return						Nibble represented as a hexadecimal character.
 */
	public static char fromNibble
	(
		final int				Nibble
	)
	{
		return HexNibbles[Nibble & 0x0F];
	}

/** Converts a <code>byte</code> to a two character, hexadecimal sequence.
 *
 *	@param Value				<code>byte</code> value.
 *	
 *	@return						Hexadecimal sequence.
 */
	public static String fromByte
	(
		final byte				Value
	)
	{
		return new String(HexBytes[Value & 0xFF]);
	}

/** Converts a <code>byte</code> array into an uninterrupted hexadecimal sequence.
 *
 *	@param Bytes				<code>byte</code> array.
 *	
 *	@return						Hexadecimal sequence.
 */
	public static String fromBytes
	(
		final byte[]			Bytes
	)
	{
		if (Bytes == null)
			return "null";
		
		return fromBytes(Bytes, 0, Bytes.length);
	}

/** Converts a <code>byte</code> array into an uninterrupted hexadecimal sequence.
 *
 *	@param Bytes				<code>byte</code> array.
 *	@param Offset				The offset into the array at which to begin reading bytes.
 *	@param Length				The number of bytes to be read.
 *	
 *	@return						Hexadecimal sequence.
 */
	public static String fromBytes
	(
		final byte[]			Bytes,
		final int				Offset,
		final int				Length
	)
	{
		if (Bytes == null)
			return "null";
		
		final char[]			HexBuff  = new char[Length << 1];
		int						HexIndex = 0;

		for (int ByteIndex = Offset, ByteLimit = Offset + Length; ByteIndex < ByteLimit; ByteIndex++) {
			final char[]			Hex = HexBytes[Bytes[ByteIndex] & 0xFF];

			HexBuff[HexIndex++] = Hex[0];
			HexBuff[HexIndex++] = Hex[1];
		}

		return new String(HexBuff);
	}

/** Converts a <code>byte</code> array into a multi-line, hexadecimal sequence.
 *
 *	@param Bytes				<code>byte</code> array.
 *	@param BytesPerRow			The number of hexadecimal bytes to display on each line.
 *	@param LinePrefix			Sequence to precede each line of output. Can be <code>null</code>.
 *	
 *	@return						Hexadecimal sequence.
 */
	public static String fromBytes
	(
		final byte[]			Bytes,
		final int				BytesPerRow,
		final String			LinePrefix			//	Can be null.
	)
	{
		if (Bytes == null)
			return "null";
		
		return fromBytes(Bytes, 0, Bytes.length, BytesPerRow, LinePrefix, null);
	}

/** Converts a <code>byte</code> array into a multi-line, hexadecimal sequence.
 *
 *	@param Bytes				<code>byte</code> array.
 *	@param BytesPerRow			The number of hexadecimal bytes to display on each line.
 *	@param LinePrefix			Sequence to precede each line of output. Can be <code>null</code>.
 *	@param IntraHexSeparator	Sequence to place between hexadecimal bytes. Can be <code>null</code>.
 *	
 *	@return						Hexadecimal sequence.
 */
	public static String fromBytes
	(
		final byte[]			Bytes,
		final int				BytesPerRow,
		final String			LinePrefix,			//	Can be null.
		final String			IntraHexSeparator	//	Can be null.
	)
	{
		if (Bytes == null)
			return "null";
		
		return fromBytes(Bytes, 0, Bytes.length, BytesPerRow, LinePrefix, IntraHexSeparator);
	}

/** Converts a <code>byte</code> array into a multi-line, hexadecimal sequence.
 *
 *	@param Bytes				<code>byte</code> array.
 *	@param ByteIndex			Offset into <code>Bytes</code> at which to begin.
 *	@param ByteCount			Number of bytes, beginning at <code>ByteIndex</code>, to convert into hex.
 *	@param BytesPerRow			The number of hexadecimal bytes to display on each line.
 *	@param LinePrefix			Sequence to precede each line of output. Can be <code>null</code>.
 *	@param IntraHexSeparator	Sequence to place between hexadecimal bytes. Can be <code>null</code>.
 *	
 *	@return						Hexadecimal sequence.
 */
	public static String fromBytes
	(
		final byte[]			Bytes,
		final int				ByteIndex,
		final int				ByteCount,
		final int				BytesPerRow,
		final String			LinePrefix,			//	Can be null.
		final String			IntraHexSeparator	//	Can be null.
	)
	{
		return andASCIIFromBytes(Bytes, ByteIndex, ByteCount, BytesPerRow, LinePrefix, IntraHexSeparator, null, false, kDefaultLineSepStr);
	}

/** Converts a <code>byte</code> array into a multi-line, hexadecimal string, where each line begins with a hex
 * 	representation of 16 bytes from the array and ends with an ASCII representation of those same bytes. Non-printable 
 * 	ASCII characters are printed as periods ('.').
 *
 *	@param Bytes				<code>byte</code> array.
 *	
 *	@return						Hexadecimal and ASCII sequence.
 */
	public static String andASCIIFromBytes
	(
		final byte[]			Bytes
	)
	{
		if (Bytes == null)
			return "null";
		
		return andASCIIFromBytes(Bytes, 0, Bytes.length, kDefaultBytesPerRow, kDefaultLinePrefixStr, kDefaultIntraHexSepStr, kDefaultHexASCIISepStr, true, kDefaultLineSepStr);
	}

/** Converts a <code>byte</code> array into a multi-line, hexadecimal string, where each line begins with a hex
 * 	representation of 16 bytes from the array and ends with an ASCII representation of those same bytes. Non-printable 
 * 	ASCII characters are printed as periods ('.').
 *
 *	@param Bytes				<code>byte</code> array.
 *	@param ByteIndex			Offset into <code>Bytes</code> at which to begin.
 *	@param ByteCount			Number of bytes, beginning at <code>ByteIndex</code>, to convert into hex.
 *	
 *	@return						Hexadecimal and ASCII sequence.
 */
	public static String andASCIIFromBytes
	(
		final byte[]			Bytes,
		final int				ByteIndex,
		final int				ByteCount
	)
	{
		return andASCIIFromBytes(Bytes, ByteIndex, ByteCount, kDefaultBytesPerRow, kDefaultLinePrefixStr, kDefaultIntraHexSepStr, kDefaultHexASCIISepStr, true, kDefaultLineSepStr);
	}

/** Converts a <code>byte</code> array into a multi-line, hexadecimal string, where each line begins with a hex
 * 	representation of the row and ends with an ASCII representation. Non-printable ASCII characters are printed as 
 * 	periods ('.').
 *
 *	@param Bytes				<code>byte</code> array.
 *	@param ByteIndex			Offset into <code>Bytes</code> at which to begin.
 *	@param ByteCount			Number of bytes, beginning at <code>ByteIndex</code>, to convert into hex.
 *	@param BytesPerRow			The number of hexadecimal bytes to display on each line.
 *	
 *	@return						Hexadecimal and ASCII sequence.
 */
	public static String andASCIIFromBytes
	(
		final byte[]			Bytes,
		final int				ByteIndex,
		final int				ByteCount,
		final int				BytesPerRow
	)
	{
		return andASCIIFromBytes(Bytes, ByteIndex, ByteCount, BytesPerRow, kDefaultLinePrefixStr, kDefaultIntraHexSepStr, kDefaultHexASCIISepStr, true, kDefaultLineSepStr);
	}

/** Converts a <code>byte</code> array into a multi-line, hexadecimal string, where each line begins with a hex
 * 	representation of the row and ends with an ASCII representation. Non-printable ASCII characters are printed as 
 * 	periods ('.').
 *
 *	@param Bytes				<code>byte</code> array.
 *	@param ByteIndex			Offset into <code>Bytes</code> at which to begin.
 *	@param ByteCount			Number of bytes, beginning at <code>ByteIndex</code>, to convert into hex.
 *	@param BytesPerRow			The number of hexadecimal bytes to display on each line.
 *	@param LinePrefix			Sequence to precede each line of output. Can be <code>null</code>.
 *	
 *	@return						Hexadecimal and ASCII sequence.
 */
	public static String andASCIIFromBytes
	(
		final byte[]			Bytes,
		final int				ByteIndex,
		final int				ByteCount,
		final int				BytesPerRow,
		final String			LinePrefix			//	Can be null.
	)
	{
		return andASCIIFromBytes(Bytes, ByteIndex, ByteCount, BytesPerRow, LinePrefix, kDefaultIntraHexSepStr, kDefaultHexASCIISepStr, true, kDefaultLineSepStr);
	}

/** Converts a <code>byte</code> array into a multi-line, hexadecimal string, where each line begins with a hex
 * 	representation of the row and ends with an ASCII representation. Non-printable ASCII characters are printed as 
 * 	periods ('.').
 *
 *	@param Bytes				<code>byte</code> array.
 *	@param ByteIndex			Offset into <code>Bytes</code> at which to begin.
 *	@param ByteCount			Number of bytes, beginning at <code>ByteIndex</code>, to convert into hex.
 *	@param BytesPerRow			The number of hexadecimal bytes to display on each line.
 *	@param LinePrefix			Sequence to precede each line of output. Can be <code>null</code>.
 *	@param IntraHexSeparator	Sequence to place between hexadecimal bytes. Can be <code>null</code>.
 *	
 *	@return						Hexadecimal and ASCII sequence.
 */
	public static String andASCIIFromBytes
	(
		final byte[]			Bytes,
		final int				ByteIndex,
		final int				ByteCount,
		final int				BytesPerRow,
		final String			LinePrefix,			//	Can be null.
		final String			IntraHexSeparator	//	Can be null.
	)
	{
		return andASCIIFromBytes(Bytes, ByteIndex, ByteCount, BytesPerRow, LinePrefix, IntraHexSeparator, kDefaultHexASCIISepStr, true, kDefaultLineSepStr);
	}

/** Converts a <code>byte</code> array into a multi-line, hexadecimal string, where each line begins with a hex
 * 	representation of the row and ends with an ASCII representation. Non-printable ASCII characters are printed as 
 * 	periods ('.').
 *
 *	@param Bytes				<code>byte</code> array.
 *	@param ByteIndex			Offset into <code>Bytes</code> at which to begin.
 *	@param ByteCount			Number of bytes, beginning at <code>ByteIndex</code>, to convert into hex.
 *	@param BytesPerRow			The number of hexadecimal bytes to display on each line.
 *	@param LinePrefix			Sequence to precede each line of output. Can be <code>null</code>.
 *	@param IntraHexSeparator	Sequence to place between hexadecimal bytes. Can be <code>null</code>.
 *  @param HexASCIISeparator	Sequence to place between the hexadecimal and ASCII representations. Can be 
 * 								<code>null</code>.
 * 	@param IncludeASCII			<code>true</code> if the ASCII representation of the bytes should be include,
 * 								<code>false</code> otherwise.
 * 	@param LineSeparator		Sequence to place at the end of lines. If <code>null</code> the platform's default
 * 								line separator sequence is used.
 * 
 *	@return						Hexadecimal and ASCII sequence.
 */
	public static String andASCIIFromBytes
	(
		final byte[]			Bytes,
		final int				ByteIndex,
		final int				ByteCount,
		final int				BytesPerRow,
		final String			LinePrefix,			//	Can be null.
		final String			IntraHexSeparator,	//	Can be null.
		final String			HexASCIISeparator,	//	Can be null.
		final boolean			IncludeASCII,
		String					LineSeparator		//	Can be null.
	)
	{
		if (LineSeparator == null)
			LineSeparator = System.getProperty("line.separator");

		final int				RowCount       = (ByteCount + (BytesPerRow - 1)) / BytesPerRow;
		final int				LinePrefixLen  = LinePrefix        != null ? LinePrefix.length()        : 0;
		final int				IntraHexLen    = IntraHexSeparator != null ? IntraHexSeparator.length() : 0;
		final int				HexASCIISepLen = HexASCIISeparator != null ? HexASCIISeparator.length() : 0;
		final int				HexLen         = (BytesPerRow << 1) + IntraHexLen * (BytesPerRow - 1);
		final int				ASCIILen       = IncludeASCII ? BytesPerRow : 0;

		final int				OutputLen      = RowCount * (LinePrefixLen + HexLen + HexASCIISepLen + ASCIILen + LineSeparator.length());
		final StringWriter		Out            = new StringWriter(OutputLen);

		try {
			andASCIIFromBytes(Bytes, ByteIndex, ByteCount, BytesPerRow, LinePrefix, IntraHexSeparator, HexASCIISeparator, IncludeASCII, LineSeparator, Out);
		} catch (IOException e) {
		//	It should be impossible to have an IOException when writing to a StringWriter.
		}

		return Out.toString();
	}

/** Writes a <code>byte</code> array to a {@link Writer Writer} in a multi-line, hexadecimal representation, where each 
 * 	line of output begins with a hex representation of the row and ends with an ASCII representation. Non-printable 
 * 	ASCII characters are printed as periods ('.').
 *
 *	@param Bytes				<code>byte</code> array.
 *	@param ByteIndex			Offset into <code>Bytes</code> at which to begin.
 *	@param ByteCount			Number of bytes, beginning at <code>ByteIndex</code>, to convert into hex.
 *	@param BytesPerRow			The number of hexadecimal bytes to display on each line.
 *	@param LinePrefix			Sequence to precede each line of output. Can be <code>null</code>.
 *	@param IntraHexSeparator	Sequence to place between hexadecimal bytes. Can be <code>null</code>.
 *  @param HexASCIISeparator	Sequence to place between the hexadecimal and ASCII representations. Can be 
 * 								<code>null</code>.
 * 	@param IncludeASCII			<code>true</code> if the ASCII representation of the bytes should be include,
 * 								<code>false</code> otherwise.
 * 	@param LineSeparator		Sequence to place at the end of lines. If <code>null</code> the platform's default
 * 								line separator sequence is used.
 * 	@param Out					The <code>Writer</code> to which the hexadecimal and ASCII sequence should be written.
 * 
 * 	@throws IOException			If the <code>Writer</code> throws an exception.
 */
	public static void andASCIIFromBytes
	(
		final byte[]			Bytes,
		final int				ByteIndex,
		final int				ByteCount,
		final int				BytesPerRow,
		final String			LinePrefix,			//	Can be null.
		final String			IntraHexSeparator,	//	Can be null.
		final String			HexASCIISeparator,	//	Can be null.
		final boolean			IncludeASCII,
		String					LineSeparator,		//	Can be null.
		final Writer			Out
	)
	throws IOException
	{
		if (Bytes == null) {
			
			Out.write("null");
			return;
		}
		
		if (LineSeparator == null)
			LineSeparator = System.getProperty("line.separator");

		final char[][]			HexBytes = Hexadecimal.HexBytes;
		final int				RowCount = (ByteCount + (BytesPerRow - 1)) / BytesPerRow;
		int						Index    = 0;

		for (int Row = 0; Row < RowCount; Row++) {

		//	Print any supplied line prefix string. This allows for the output to be
		//	indented, for instance.

			if (LinePrefix != null)
				Out.write(LinePrefix);

		//	Generate a line of output.

			int						Limit = Index + BytesPerRow;

		//	Print the hex version of this row.

			for (int HIndex = Index, RowRemaining = BytesPerRow; HIndex < Limit; HIndex++) {

				if (HIndex < ByteCount)
					Out.write(HexBytes[Bytes[ByteIndex + HIndex] & 0xFF]);
				else
					Out.write("  ");

			//	Print separating space.

				if (--RowRemaining > 0 && IntraHexSeparator != null)
					Out.write(IntraHexSeparator);
			}

		//	Print the ASCII version of this row.

			if (IncludeASCII) {

				if (HexASCIISeparator != null)
					Out.write(HexASCIISeparator);

				if (Limit > ByteCount)
					Limit = ByteCount;

				for (int AIndex = Index; AIndex < Limit; AIndex++) {
					final int				This = Bytes[ByteIndex + AIndex];

					if (This < 32 || This >= 127)
						Out.write('.');
					else
						Out.write((char) This);
				}
			}

		//	Terminate this line of output.

			Out.write(LineSeparator);

		//	Next iteration.

			Index = Limit;
		}
	}
}
